#include "bndfile.h"
#include "revfile.h"
#include "seqfile.h"
#include "cmafile.h"
#include "getopts.h"
#include "utils.h"
#include "progress.h"

static bool g_AlnRequired = false;
static CMAFile *g_CMA = 0;
static unsigned g_CountWithAln = 0;
static unsigned g_CountWithoutAln = 0;

static void StrVecToStdStrVec(const StrVec &cv, StdStrVec &sv)
	{
	const unsigned N = (unsigned) cv.size();
	sv.resize(N);
	for (unsigned i = 0; i < N; ++i)
		{
		std::string tmp(cv[i]);
		sv[i] = tmp;
		}
	}

static void AIToCoords(const AlnInfo &AI, CoordVec &Coords)
	{
	Coords.resize(2);

	CoordData &C0 = Coords[0];
	CoordData &C1 = Coords[1];

	const GCSeg &Seg0 = AI.Segs[0];
	const GCSeg &Seg1 = AI.Segs[1];

	C0.ChrIndex = Seg0.ChrId;
	C0.StartPos = Seg0.lo;
	C0.EndPos = Seg0.hi;
	C0.Reversed = false;

	C1.ChrIndex = Seg1.ChrId;
	C1.StartPos = Seg1.lo;
	C1.EndPos = Seg1.hi;
	C1.Reversed = (Seg0.Plus != Seg1.Plus);
	}

// Set chr length from seq files
static void GIsFromSeqFiles(std::vector<GenInfo> &GIs)
	{
	const char *Seq1FileName = RequiredValueOpt("seq1");
	const char *Seq2FileName = RequiredValueOpt("seq2");

	for (unsigned GenomeIndex = 0; GenomeIndex < 2; ++GenomeIndex)
		{
		const char *SeqFileName = (GenomeIndex == 0 ? Seq1FileName : Seq2FileName);

		SeqFile Seq;
		Seq.Open(SeqFileName);

		std::vector<ChrInfo> CIs;
		const ChrNameIndex &CNI = Seq.GetChrNameIndex();
		unsigned SeqFileChrId = UINT_MAX;
		const unsigned ChrCount = CNI.GetChrCount();
		for (unsigned i = 0; i < ChrCount; ++i)
			{
			if (i == 0)
				SeqFileChrId = CNI.GetFirstChrIndex();
			else
				SeqFileChrId = CNI.GetNextChrIndex(SeqFileChrId);

			const std::string &ChrName = CNI.GetName(SeqFileChrId);

			unsigned SeqFileChrId = Seq.GetChrIndex(ChrName.c_str());
			coord_t ChrLength = Seq.GetChrLength(SeqFileChrId);

			ChrInfo CI;
			CI.Create(SeqFileChrId, ChrName, ChrLength);

			CIs.push_back(CI);
			}

		Seq.Close();

		const std::string &GenomeName = Seq.GetGenomeName();

		GenInfo GI;
		GI.Create(GenomeName, CIs);

		GIs.push_back(GI);
		}
	}

static unsigned GetChrLength(unsigned ChrId,
  const std::map<unsigned, coord_t> &ChrIdToChrLength)
	{
	std::map<unsigned, coord_t>::const_iterator p = ChrIdToChrLength.find(ChrId);
	if (p == ChrIdToChrLength.end())
		Die("GetChrLength(%u), not found", ChrId);
	return p->second;
	}

// Seq files not specified, set chromosome length to max coord
// found in SPs.
static void GIsFromBndFile(std::vector<GenInfo> &GIs, const BndFile &Bnd)
	{
	std::map<unsigned, coord_t> ChrIdToChrLength1;
	std::map<unsigned, coord_t> ChrIdToChrLength2;

	const ChrPairIndex &CPI = Bnd.GetChrPairIndex();
	for (ChrPairIndex::const_iterator p = CPI.begin(); p != CPI.end(); ++p)
		{
		const ChrPairIndexEntry &E = *p;
		unsigned ChrId1 = E.ChrIndex1;
		unsigned ChrId2 = E.ChrIndex2;

		SPVec SPs;
		VOL Bundles;
		VOL Edges;
		Bnd.ReadChrPair(ChrId1, ChrId2, SPs, Bundles, Edges);

		coord_t MaxCoord1 = 0;
		coord_t MaxCoord2 = 0;
		const unsigned SPCount = (unsigned) SPs.size();
		for (unsigned SPIndex = 0; SPIndex < SPCount; ++SPIndex)
			{
			const SegPair &SP = *(SPs[SPIndex]);
			coord_t hi1 = SP.Segs[0].hi;
			coord_t hi2 = SP.Segs[1].hi;

			MaxCoord1 = std::max(MaxCoord1, hi1);
			MaxCoord2 = std::max(MaxCoord2, hi2);
			}

		ChrIdToChrLength1[ChrId1] = MaxCoord1 + 1;
		ChrIdToChrLength2[ChrId2] = MaxCoord2 + 1;

		FreeSPs(SPs);
		}

// OK
	for (unsigned GenomeIndex = 0; GenomeIndex < 2; ++GenomeIndex)
		{
		const ChrNameIndex &CNI =
		  (GenomeIndex == 0 ? Bnd.GetChrNameIndex1() : Bnd.GetChrNameIndex2());

		std::vector<ChrInfo> CIs;
		unsigned ChrId = UINT_MAX;
		const unsigned ChrCount = CNI.GetChrCount();
		for (unsigned i = 0; i < ChrCount; ++i)
			{
			if (i == 0)
				ChrId = CNI.GetFirstChrIndex();
			else
				ChrId = CNI.GetNextChrIndex(ChrId);

			const std::string &ChrName = CNI.GetName(ChrId);

			coord_t ChrLength = COORD_T_MAX;
			if (GenomeIndex == 0)
				ChrLength = GetChrLength(ChrId, ChrIdToChrLength1);
			else
				ChrLength = GetChrLength(ChrId, ChrIdToChrLength2);

			ChrInfo CI;
			CI.Create(ChrId, ChrName, ChrLength);

			CIs.push_back(CI);
			}

		const char *cstrGenomeName = 
		  (GenomeIndex == 0 ? Bnd.GetGenomeName1() : Bnd.GetGenomeName2());
		const std::string GenomeName = cstrGenomeName;
		// return; // OK

		GenInfo GI;
		GI.Create(GenomeName, CIs);
		// return; // FAIL

		GIs.push_back(GI);
		// FAIL
		}
	}

static void Add(const SegPair &SP, unsigned ChrIdA, unsigned ChrIdB,
  unsigned BundleIndex, RevFile &Rev)
	{
	if (!BndFile::IsActiveHit(SP))
		return;

	AlnInfo AI;
	AI.SetBundleIndex(BundleIndex);
	if (SP.AlnLength > 0)
		AI.SetAlnLength(SP.AlnLength);

	AI.Score = (ai_score_t) SP.Score;

	const Seg &S0 = SP.Segs[0];
	const Seg &S1 = SP.Segs[1];

	GCSeg S;

	S.GenomeIndex = 0;
	S.ChrId = ChrIdA;
	S.lo = S0.lo;
	S.hi = S0.hi;
	S.Plus = true;

	AI.Segs.push_back(S);

	S.GenomeIndex = 1;
	S.ChrId = ChrIdB;
	S.lo = S1.lo;
	S.hi = S1.hi;
	S.Plus = !SP.Reverse;

	AI.Segs.push_back(S);

	if (g_CMA == 0)
		{
		Rev.AddAlnInfo(AI);
		++g_CountWithoutAln;
		}
	else
		{
		CoordVec Coords;
		AIToCoords(AI, Coords);

		StrVec cstrRows;
		bool Found = g_CMA->GetCMA(Coords, cstrRows);

		if (!Found && g_AlnRequired)
			{
			Log("SP: ");
			LogSPN(SP);
			Log(" SPType=%s Subtype=%s\n",
			  SPTypeToStr(SP.Type), SPSubtypeToStr(SP.Subtype));
			Log("AI: ");
			LogAI(AI);
			Log("\n");
			Log("Coords: ");
			LogCoords(Coords);
			Log("\n");
			Die("Alignment not found in CMA file");
			}

		if (!Found)
			{
			Rev.AddAlnInfo(AI);
			++g_CountWithoutAln;
			}
		else
			{
			StdStrVec Rows;
			StrVecToStdStrVec(cstrRows, Rows);

			Rev.AddAln(AI, Rows);

			++g_CountWithAln;
			}
		}
	}

void CvtBnd2Rev()
	{
	const char *BndFileName = RequiredValueOpt("frombnd");
	const char *RevFileName = RequiredValueOpt("torev");
	const char *CMAFileName = ValueOpt("cma");
	g_AlnRequired = FlagOpt("aln_required");

	BndFile Bnd;
	RevFile Rev;

	Progress("Opening %s", BndFileName);
	Bnd.Open(BndFileName);

	if (CMAFileName)
		{
		Progress("Opening %s", CMAFileName);
		g_CMA = new CMAFile;
		g_CMA->Open(CMAFileName);
		}

	std::vector<GenInfo> GIs;
	if (ValueOpt("seq1"))
		{
		Progress("Opening seq files");
		GIsFromSeqFiles(GIs);
		}
	else
		{
		Progress("Getting genome info from bnd file");
		GIsFromBndFile(GIs, Bnd);
		}

	Progress("Creating new rev file");
	Rev.Create(RevFileName, GIs);

	const ChrPairIndex &CPI = Bnd.GetChrPairIndex();
	const unsigned ChrPairCount = (unsigned) CPI.size();

	ProgressStart("Converting");
	for (unsigned ChrPairIndex = 0; ChrPairIndex < ChrPairCount; ++ChrPairIndex)
		{
		if (ChrPairIndex%100 == 0)
			ProgressStep(ChrPairIndex, ChrPairCount);

		const ChrPairIndexEntry &E = CPI[ChrPairIndex];

		VOL Bundles;
		VOL Edges;
		SPVec SPs;
		bool Found = Bnd.ReadChrPair(E.ChrIndex1, E.ChrIndex2, SPs,
		  Bundles, Edges);
		assert(Found);

		const unsigned SPCount = (unsigned) SPs.size();
		const unsigned BundleCount = Bundles.GetVectorLength();
		for (unsigned BundleIndex = 0; BundleIndex < BundleCount; ++BundleIndex)
			{
			const IntList &SPIndexes = Bundles.GetList(BundleIndex);
			for_CIntList(SPIndexes, p)
				{
				unsigned SPIndex = *p;
				assert(SPIndex < SPCount);

				const SegPair &SP = *(SPs[SPIndex]);
				Add(SP, E.ChrIndex1, E.ChrIndex2, BundleIndex, Rev);
				}
			}

		FreeSPs(SPs);
		}
	ProgressDone();

	if (g_CMA != 0)
		{
		g_CMA->Close();
		g_CMA = 0;
		}
	Bnd.Close();

	Progress("Writing data to rev file");
	Rev.Close();

	Progress("Converted %u hits with alignments, %u without",
	  g_CountWithAln, g_CountWithoutAln);
	}
