#define TRACE	0

#include "evo.h"

/***
Conversion of an ancestral and evolved genome to a .rev alignment file
works as follows.

The ancestral genome (.evo file) is read serially from start to finish,
saving the sequence into g_AnSeq as each base is read (using the GotBase
callback). No more than one MUL is accumulated at one time; at the end
of each MUL it is cleared.

A MUL in the evolved genome (.evo file) is read when the ancestral read
hits the position corresponding to the end of that MUL. At this point,
we have both the ancestral and evolved sequences for that MUL.

MicroEvolver::FindAlignments() is then called on the evolved genome,
which reports alignments (maximal sequences of bases that are not novel
and have contiguous ancestral and evolved coordinates).
***/

static unsigned g_MinLength = 8;

static vector<MULIndexEntryData> g_EvIEs;
static vector<gpos_t> g_EvStartPosVec;
static IntVec g_EvMULIndexes;
static unsigned g_MULIndexesSub = UINT_MAX;
static unsigned g_AnMULIndex = UINT_MAX;
static gpos_t g_AnStartPos = GPOS_T_MAX;
static gpos_t g_AnEndPos = GPOS_T_MAX;
static gpos_t g_AnLastPos = GPOS_T_MAX;
static gpos_t g_EvStartPos = GPOS_T_MAX;
static string g_AnSeq;
static gpos_t g_BaseCount = 0;
static gpos_t g_TotalBaseCount = 0;
static FILE *g_fEv = 0;
static RevFile *g_ptrOut;
static gpos_t g_ContigBaseCount = 0;
static unsigned g_DoMULIndex = UINT_MAX;
static map<coord_t, unsigned> g_AnPosMap;
static map<coord_t, unsigned> g_EvPosMap;
static unsigned g_TotalAlnCount = 0;
static unsigned g_AlnCount = 0;
static unsigned g_TotalShortAlnCount = 0;
static unsigned g_ShortAlnCount = 0;
static bool g_EvolvedMULPlus = true;
static gpos_t g_EvolvedMULLength = GPOS_T_MAX;
static bool g_Done = false;

static void MakePosMap(const GenInfo &Genome, map<coord_t, unsigned> &PosMap)
	{
	const unsigned ChrCount = Genome.GetChrCount();
	coord_t Pos = 0;
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		const ChrInfo &CI = Genome.GetChrInfo(ChrIndex);
		unsigned ChrLength = CI.GetLength();
		PosMap[Pos] = ChrIndex;
		Pos += ChrLength;
		}
	}

static void LogPosMap(const map<coord_t, unsigned> &PosMap)
	{
	Log("PosMap.size=%u\n", SIZE(PosMap));
	for (map<coord_t, unsigned>::const_iterator p = PosMap.begin();
	  p != PosMap.end(); ++p)
		Log("Pos %10u  Chr %4u\n", (unsigned) p->first, p->second);
	}

static void PosToChrPos(coord_t Pos, const map<coord_t, unsigned> &PosMap,
  unsigned *ptrChrIndex, coord_t *ptrChrPos)
	{
	map<coord_t, unsigned>::const_iterator p = PosMap.upper_bound(Pos);
	if (p == PosMap.begin())
		{
		Log("\n");
		LogPosMap(PosMap);
		Log("upper_bound=begin, Pos=%u\n", (unsigned) Pos);
		Die("PosToChrPos, not found");
		}
	--p;
	asserta(Pos >= p->first);
	*ptrChrPos = Pos - p->first;
	*ptrChrIndex = p->second;
	}

static void PosToChrPos(AlnInfo &AI)
	{
	const unsigned SegCount = AI.GetSegCount();
	asserta(SegCount == 2);
	GCSeg &AnSeg = AI.Segs[0];
	GCSeg &EvSeg = AI.Segs[1];

	coord_t AnSegLength = AnSeg.GetLength();
	coord_t EvSegLength = EvSeg.GetLength();

	asserta(AnSegLength > 0);
	asserta(EvSegLength > 0);

	PosToChrPos(AnSeg.lo, g_AnPosMap, &AnSeg.ChrId, &AnSeg.lo);
	PosToChrPos(EvSeg.lo, g_EvPosMap, &EvSeg.ChrId, &EvSeg.lo);

	AnSeg.hi = AnSeg.lo + AnSegLength - 1;
	EvSeg.hi = EvSeg.lo + EvSegLength - 1;
	}

static unsigned GetCurrentMULIndex()
	{
	assert(g_MULIndexesSub < SIZE(g_EvMULIndexes));
	return g_EvMULIndexes[g_MULIndexesSub];
	}

static bool LTIE(unsigned IEIndex1, unsigned IEIndex2)
	{
	assert(IEIndex1 < SIZE(g_EvIEs));
	assert(IEIndex2 < SIZE(g_EvIEs));

	const MULIndexEntryData &IE1 = g_EvIEs[IEIndex1];
	const MULIndexEntryData &IE2 = g_EvIEs[IEIndex2];

	return IE1.MI.MinPos < IE2.MI.MinPos;
	}

// The GotAlignment callback receives the ancestral row with gaps but Xs
// instead of letters (because MicroEvolver doesn't know the ancestral
// letters, only ancestral coordinates). MakeRow replaces Xs with AGCTs
// from the ungapped ancestral sequence.
static void MakeRow(const StdStr &GappedSeqX, const StdStr &UngappedSeq,
  StdStr &GappedSeq)
	{
	GappedSeq.clear();
	StdStr::const_iterator pU = UngappedSeq.begin();
	for (StdStr::const_iterator pX = GappedSeqX.begin();
	  pX != GappedSeqX.end(); ++pX)
		{
		char cX = *pX;
		if (cX == '-')
			GappedSeq.clear();
		else
			{
			asserta(pU != UngappedSeq.end());
			char cU = *pU++;
			GappedSeq.push_back(cU);
			}
		}
	}

// GlobalAI.Seg[0] has global coordinates for the ancestral genome,
// which are all considered to be in "fake" chromosome zero.
// GlobalAI.Seg[1] similarly has coordinates relative to the start
// of the ancestral genome.
static void GotAlignment(const MicroEvolver &Micro, const AlnInfo &GlobalAI,
  const StdStrVec &Rows_)
	{
	++g_AlnCount;
TRC
	Log("\n");
	Log("GotAlignment AI=");
	LogAIN(GlobalAI);
	Log("Row0: %s\n", Rows_[0].c_str());
	Log("Row1: %s\n", Rows_[1].c_str());
ENDTRC
	asserta(SIZE(GlobalAI.Segs) == 2);
	asserta(SIZE(Rows_) == 2);

	if (SIZE(Rows_[0]) < g_MinLength)
		{
		++g_TotalShortAlnCount;
		++g_ShortAlnCount;
		return;
		}
	++g_TotalAlnCount;

	asserta(g_AnStartPos == Micro.GetMinPos() && g_AnEndPos == Micro.GetMaxPos());
	asserta(SIZE(g_AnSeq) == (unsigned) (g_AnEndPos - g_AnStartPos + 1));

	const GCSeg &GlobalAnSeg = GlobalAI.GetSeg(0);
	if (GlobalAnSeg.lo < g_AnStartPos || GlobalAnSeg.hi > g_AnEndPos)
		{
		Log("Global AI:\n");
		LogAIN(GlobalAI);
		Die("AnSeg.lo %u  AnSeg.hi %u  Start %"PF"  End %"PF,
		  (unsigned) GlobalAnSeg.lo,
		  (unsigned) GlobalAnSeg.hi,
		  g_AnStartPos,
		  g_AnEndPos);
		}

// Relative to start of the evolved MUL
	unsigned RelativeAnLo = (unsigned) (GlobalAnSeg.lo - g_AnStartPos);
	asserta(RelativeAnLo < SIZE(g_AnSeq));

	StdStrVec Rows;
	Rows = Rows_;
	const StdStr &AnRowX = Rows_[0];

	unsigned AnUngappedLength = GetUngappedSeqLength(AnRowX);
	const string &AnSeq = g_AnSeq.substr(RelativeAnLo, AnUngappedLength);
	StdStr &AnRow = Rows[0];
	MakeRow(AnRowX, AnSeq, AnRow);
TRC
	Log("After MakeRow, AnRow=%s\n", AnRow.c_str());
ENDTRC

	if (!GlobalAnSeg.Plus)
		{
		RevComp(AnRow);
TRC
		Log("Seg minus: RevComp AnRow=%s\n", AnRow.c_str());
ENDTRC
		}

// Fix up coordinates in GlobalAI, giving ChrAI.
	AlnInfo ChrAI;
	ChrAI = GlobalAI;
	GCSeg &ChrEvSeg = ChrAI.Segs[1];
	asserta(ChrEvSeg.lo <= ChrEvSeg.hi);
	asserta(ChrEvSeg.lo < g_EvolvedMULLength);
	asserta(ChrEvSeg.hi < g_EvolvedMULLength);

	if (!g_EvolvedMULPlus)
		{
		ChrEvSeg.Plus = !ChrEvSeg.Plus;
// Don't complement sequence here, MUL was read from ancestral
// genome as positive strand.
//		RevComp(AnRow);
		gpos_t LoOffset = g_EvolvedMULLength - ChrEvSeg.hi - 1;
		gpos_t HiOffset = g_EvolvedMULLength - ChrEvSeg.lo - 1;
		ChrEvSeg.lo = LoOffset;
		ChrEvSeg.hi = HiOffset;
TRC
		Log("MUL minus, change ev seg strand\n");
ENDTRC
		}

	coord_t EvStartPos = g_EvStartPos;
	ChrEvSeg.lo += EvStartPos;
	ChrEvSeg.hi += EvStartPos;

	coord_t ctAnPos = GlobalAnSeg.lo;
	ChrAI.SetAttr(ALN_ATTR_AncestralPos, ctAnPos);

// ChrAI now has global coordinates.
	PosToChrPos(ChrAI);
TRC
	Log("Converted from global to chr coords:\n");
	LogAIRows(ChrAI, Rows);
ENDTRC

	g_ContigBaseCount += AnUngappedLength;

	g_ptrOut->AddAln(ChrAI, Rows);
	}

static void SetStartEndPos()
	{
	unsigned MULIndex = GetCurrentMULIndex();

	asserta(MULIndex < SIZE(g_EvIEs));
	asserta(MULIndex < SIZE(g_EvStartPosVec));

	g_EvStartPos = g_EvStartPosVec[MULIndex];

	const MULIndexEntryData &IE = g_EvIEs[MULIndex];
	g_AnStartPos = IE.MI.MinPos;
	g_AnEndPos = IE.MI.MinPos + IE.MI.AncestralLength - 1;

TRC
	Log("IE:");
	IE.LogMe();
	Log(" Start %"PF, g_AnStartPos);
	Log(" End %"PF, g_AnEndPos);
	Log("\n");
ENDTRC
	}

static void GotEnd()
	{
TRC
	Log("GotEnd\n");
ENDTRC

	unsigned MULIndex = GetCurrentMULIndex();
	const MULIndexEntryData &EvIE = g_EvIEs[MULIndex];
	g_EvolvedMULPlus = EvIE.MI.GetPlus();
	g_EvolvedMULLength = EvIE.MI.EvolvedLength;

	SetProgressMsg("Ancestral MUL %u, evolved MUL %u, %u alns",
	  g_AnMULIndex, MULIndex, g_TotalAlnCount);

TRC
	Log("MUL Plus=%c Len=%"PF"\n", pom(g_EvolvedMULPlus), g_EvolvedMULLength);
ENDTRC

	if (g_DoMULIndex == UINT_MAX || MULIndex == g_DoMULIndex)
		{
		MicroEvolver EvMicro;
		EvMicro.FromFile(g_fEv, EvIE);
		EvMicro.FindAlignments(GotAlignment);
		if (MULIndex == g_DoMULIndex)
			g_Done = true;
TRC
		gpos_t MinPos = EvMicro.GetMinPos();
		gpos_t AnLen = EvMicro.GetAncestralLength();
		gpos_t MaxPos = MinPos + AnLen - 1;
		Log("Evolved MUL %u MinPos %"PF" MaxPos %"PF" Len %"PF"  ",
		  MULIndex, MinPos, MaxPos, AnLen);
		Log("%u alns\n", g_AlnCount - g_ShortAlnCount);
ENDTRC
		g_AlnCount = 0;
		g_ShortAlnCount = 0;
		}

	g_AnSeq.clear();

	const unsigned MULCount = SIZE(g_EvMULIndexes);
	for (; g_MULIndexesSub < MULCount; ++g_MULIndexesSub)
		{
		unsigned MULIndex = GetCurrentMULIndex();
		const MULIndexEntryData &IE = g_EvIEs[MULIndex];
		if (IE.MI.MinPos > g_AnStartPos)
			{
			const MULIndexEntryData &IE = g_EvIEs[MULIndex];
			gpos_t EndPos = IE.MI.MinPos + IE.MI.AncestralLength - 1;
			asserta(EndPos > g_AnEndPos);
			SetStartEndPos();
			return;
			}
		}
	}

static void GotBase(const BaseInfo &Base)
	{
	gpos_t Pos = Base.Pos;
	if (g_AnLastPos != GPOS_T_MAX)
		{
		if (g_AnLastPos + 1 != Pos)
			Die("Ancestral genome not consecutive (%"PF", %"PF")",
			  g_AnLastPos, Pos);
		}

	if (++g_BaseCount%50000 == 0)
		{
		unsigned uTotalBaseCount = (unsigned) (g_TotalBaseCount/100);
		unsigned uBaseCount = (unsigned) (g_BaseCount/100);

		ProgressStep(uBaseCount, uTotalBaseCount);
		}

	g_AnLastPos = Pos;
	if (Pos == g_AnStartPos)
		{
		asserta(g_AnEndPos >= g_AnStartPos);
		size_t Length = (size_t) (g_AnEndPos - g_AnStartPos + 1);
		g_AnSeq.reserve(Length);
		}
	if (Pos >= g_AnStartPos && Pos <= g_AnEndPos)
		{
		if (!Base.Plus)
			{
			Log("\n");
			Base.LogMe();
			Die("Minus strand base in ancestral genome");
			}
		g_AnSeq.push_back(Base.Letter);
		}
	if (Pos == g_AnEndPos)
		GotEnd();
	}

static void GetEvStartPositions(const vector<MULIndexEntryData> &IEs,
  unsigned FirstMULIndex, vector<gpos_t> &EvStartPos)
	{
	unsigned MULCount = SIZE(IEs);
	EvStartPos.resize(MULCount, GPOS_T_MAX);

	gpos_t EvPos = 0;
	unsigned MULIndex = FirstMULIndex;
	for (;;)
		{
		if (MULIndex == UINT_MAX)
			break;
		const MULIndexEntryData &IE = IEs[MULIndex];
		EvStartPos[MULIndex] = EvPos;
		EvPos += IE.MI.EvolvedLength;
		MULIndex = IE.Next;
		}
	}

void DoEvoToRev(const char *AncestralFileName, const char *EvolvedFileName,
  const char *OutFileName)
	{
	IntOpt("minlength", &g_MinLength);

	g_fEv = OpenStdioFile(EvolvedFileName);
	FILE *fAn = OpenStdioFile(AncestralFileName);

	EvoFileHeader EvHdr;
	EvoFileHeader AnHdr;

	vector<MULIndexEntryData> AnIEs;

	vector<GenInfo> Genomes;
	Genomes.resize(2);

	ReadEvoIndex(fAn, AnHdr, AnIEs, Genomes[0]);
	ReadEvoIndex(g_fEv, EvHdr, g_EvIEs, Genomes[1]);

	MakePosMap(Genomes[0], g_AnPosMap);
	MakePosMap(Genomes[1], g_EvPosMap);

	Log("\n");
	Log("AnPosMap:\n");
	LogPosMap(g_AnPosMap);
	Log("\n");
	Log("EvPosMap:\n");
	LogPosMap(g_EvPosMap);

	GetEvStartPositions(g_EvIEs, EvHdr.FirstMULIndex, g_EvStartPosVec);

// Build vector of active evolved MUL indexes
	unsigned MULIndex = EvHdr.FirstMULIndex;
	for (;;)
		{
		if (MULIndex == UINT_MAX)
			break;
		g_EvMULIndexes.push_back(MULIndex);
		const MULIndexEntryData &IE = g_EvIEs[MULIndex];
		MULIndex = IE.Next;
		}

// Sort evolved MUL indexes in ancestral position order
	sort(g_EvMULIndexes.begin(), g_EvMULIndexes.end(), LTIE);

	g_MULIndexesSub = 0;
	SetStartEndPos();

	g_TotalBaseCount = AnHdr.LeafBaseCount;

	RevFile Out;
	Out.Create(OutFileName, Genomes);
	g_ptrOut = &Out;

	MicroFromFileOpts Opts;
	Opts.ScanOnly = true;
	Opts.GotBaseCallback = GotBase;
	Opts.Dump = FlagOpt("dump");

	ProgressStart("Converting");

// Sequential read through ancestral genome
	const unsigned MULCount = SIZE(g_EvIEs);
	g_AnMULIndex = AnHdr.FirstMULIndex;
	unsigned Count = 0;
	for (;;)
		{
		if (g_AnMULIndex == UINT_MAX)
			break;
		SetProgressMsg("Ancestral MUL %u", g_AnMULIndex);
		ProgressStep(Count++, MULCount);
		g_EvMULIndexes.push_back(g_AnMULIndex);
		const MULIndexEntryData &IE = AnIEs[g_AnMULIndex];
		MicroEvolver MicroAn;
		Log("MUL Index %u\n", g_AnMULIndex);
		Log("IE=");
		IE.LogMe();
		Log("\n");
		MicroAn.FromFile(fAn, IE, &Opts);
		if (g_Done)
			break;
		g_AnMULIndex = IE.Next;
		}
	SetProgressMsg("Converting");
	ProgressDone();

	g_EvIEs.clear();
	g_AnLastPos = GPOS_T_MAX;

	CloseStdioFile(g_fEv);
	CloseStdioFile(fAn);
	fAn = 0;
	g_fEv = 0;

	Progress("%u total alns", g_TotalAlnCount);
	Progress("%u short alns", g_TotalShortAlnCount);
	Progress("Finalizing output file");
	Out.Close();
	}

void EvoToRev()
	{
	const char *EvolvedFileName = RequiredValueOpt("evo2rev");
	const char *AncestralFileName = RequiredValueOpt("anc");
	const char *OutFileName = RequiredValueOpt("out");
	IntOpt("mul", &g_DoMULIndex);
	DoEvoToRev(AncestralFileName, EvolvedFileName, OutFileName);
	}
