#include "evo.h"

static vector<ChrInfo> g_CIs;
static unsigned g_ChrId;
static vector<MULIndexEntryData> g_IEs;
static gpos_t g_BaseCount;
static BitFile *g_BF;

static void GotLabel(const std::string &/*Label*/)
	{
	}

static void GetGenInfoOnSeq(const std::string &Label, const std::string &Seq)
	{
	if (Seq.size() > COORD_T_MAX)
		Die("Seq size > COORD_T_MAX");

	coord_t Length = (coord_t) Seq.size();

	ChrInfo CI;
	CI.Create(g_ChrId++, Label, Length);
	g_CIs.push_back(CI);
	}

static void GetGenInfo(const char *FASTAFileName, const char *GenomeName,
  GenInfo &Genome)
	{
	g_CIs.clear();
	g_ChrId = 0;

	ReadFasta(FASTAFileName, GetGenInfoOnSeq);
	Genome.Create(GenomeName, g_CIs);
	}

static void CvtOnSeq(const std::string &Label, const std::string &Seq)
	{
	if (Seq.size() > GPOS_T_MAX)
		Die("Seq length > GPOS_T_MAX");

	gpos_t Length = (gpos_t) Seq.size();
	unsigned ChrCount = SIZE(g_CIs);
	unsigned ChrIndex = SIZE(g_IEs);
	asserta(ChrIndex < ChrCount);

	MULIndexEntryData IE;

	IE.Magic = MULIEMagic;
	IE.MI.EvolvedLength = Length;
	IE.MI.AncestralLength = Length;
	IE.MI.MinPos = g_BaseCount;
	IE.MI.MaxIndex = IE.MI.MinPos + Length - 1;
	IE.FileBytePos = g_BF->GetBytePos();
	IE.FileBitPos = g_BF->GetBitPos();
	IE.InMULIndex = ChrIndex;
	IE.MI.ChrIndex = ChrIndex;
	IE.Leaf = true;
	IE.Next = (ChrIndex == ChrCount - 1 ? UINT_MAX : ChrIndex + 1);
	g_IEs.push_back(IE);

	g_BaseCount += Length;

	for (std::string::const_iterator p = Seq.begin(); p != Seq.end(); ++p)
		{
		char c = *p;
		if (!isacgt(c))
			{
			if (isprint(c))
				Die("Invalid letter '%c' found in sequence %s",
				  c, Label.c_str());
			else
				Die("Invalid byte 0x%02x in sequence %s",
				  (unsigned char) c, Label.c_str());
			}
		MicroEvolver::ZeroOffsetToFile(*g_BF);
		MicroEvolver::LetterToFile(*g_BF, c);
		}
	}

void FASTA2Evo()
	{
	const char *InFileName = RequiredValueOpt("fasta2evo");
	const char *OutFileName = RequiredValueOpt("out");
	const char *GenomeName = RequiredValueOpt("genome");
	DoFASTA2Evo(InFileName, OutFileName, GenomeName);
	}

void DoFASTA2Evo(const char *InFileName, const char *OutFileName,
  const char *GenomeName)
	{
	FILE *fOut = CreateStdioFile(OutFileName);

	GenInfo Genome;
	Progress("First pass to scan for chromosomes");
	GetGenInfo(InFileName, GenomeName, Genome);
	const unsigned ChrCount = Genome.GetChrCount();
	asserta(ChrCount == SIZE(g_CIs));

// Write placeholder header and index entry for now, has
// invalid magics so will fail unless finalized.
	InitEvoFile(fOut, ChrCount, Genome);

	g_BF = new BitFile;
	g_BF->OpenWriting(fOut);

	Progress("Second pass to convert");
	g_IEs.clear();

	ClearInvalidCounts();

	ReadFasta(InFileName, CvtOnSeq, GotLabel);

	MicroEvolver::EndOfFileToFile(*g_BF);

// Construct correct header
	asserta(ChrCount == SIZE(g_IEs));
	asserta(SIZE(g_IEs) == ChrCount);

	EvoFileHeader EvoHdr;
	EvoHdr.StartMagic = EvoStartHdrMagic;
	EvoHdr.TotalMULCount = ChrCount;
	EvoHdr.LeafMULCount = ChrCount;
	EvoHdr.FirstMULIndex = 0;
	EvoHdr.TotalBaseCount = g_BaseCount;
	EvoHdr.LeafBaseCount = g_BaseCount;
	EvoHdr.EndMagic = EvoEndHdrMagic;

// Finalize index
	WriteEvoIndex(fOut, EvoHdr, g_IEs, Genome);

	WarnInvalids();

	g_BF->Close();
	delete g_BF;

	CloseStdioFile(fOut);

	ProgressNoRes("%10u  chromosomes", ChrCount);
	ProgressNoRes("%10u  bases", g_BaseCount);

// Length checks
	coord_t GenomeLength = Genome.GetLength();
	if (GenomeLength != g_BaseCount)
		Die("Length mismatch: base count %"PF", GenInfo %"PF,
		  GenomeLength, EvoHdr.TotalBaseCount);
	}
