#include "utils.h"
#include "intlist.h"
#include "myrand.h"
#include "base.h"
#include "evoparams.h"

struct CEInfo
	{
	vector<pair<unsigned, double> > CCWeightPairs;
	vector<unsigned> WeightVec;
	byte MaxCC;
	unsigned DecayLength;
	double ClusterFactor;
	};

static CEInfo *g_CEIs[MAX_CEC+1];

#define IsFunc(Kind, x)					\
bool Is##Kind##x(uint8 x)				\
	{									\
	return x >= MIN_##Kind##_##x &&		\
	  x <= MAX_##Kind##_##x;			\
	}

IsFunc(Coding, CC)
IsFunc(Coding, CEC)
IsFunc(Noncoding, CC)
IsFunc(Noncoding, CEC)

#undef IsFunc

void GenClusteredSeq(const vector<unsigned> &WeightVec, byte MaxInt,
  unsigned DecayLength, double F, ByteVec &Seq, unsigned SeqLength);

void GetCCs(byte CEC, unsigned Length, ByteVec &CCs, unsigned RandSeed)
	{
	asserta(CEC <= MAX_CEC);
	const CEInfo *CEI = g_CEIs[CEC];
	if (CEI == 0)
		Die("CEC %u used but not defined", CEC);

	SetRandSeed(RandSeed);
	GenClusteredSeq(CEI->WeightVec, CEI->MaxCC, CEI->DecayLength,
	  CEI->ClusterFactor, CCs, Length);
	}

bool IsUserSettableCEC(unsigned CEC)
	{
	if (CEC > 19)
		return false;

	switch (CEC)
		{
	case CEC_Unconstrained:
	case CEC_GenericCDS:
	case CEC_GenericNCCE:
		return false;
		}
	return true;
	}

/***
Syntax of CEC definition:

CEC<n>=<def><sep><def><sep>...
//

where <sep> is one or more semi-colons and/or white-space
characters, including new-line, and <def> is one of:

<CC>,<w>
F=<f>
L=<l>
***/

void ReadCECDef(FILE *f, const string &Line)
	{
	unsigned CEC;
	int n = sscanf(Line.c_str(), "CE%u=", &CEC);
	if (n != 1)
		Die("Expected CE<n>=, got '%s'", Line.c_str());

	if (!IsUserSettableCEC(CEC))
		Die("CEC%u is not valid or is not user-settable", CEC);

	asserta(CEC < 20);
	CEInfo *CEI = new CEInfo;
	g_CEIs[CEC] = CEI;

	CEI->ClusterFactor = P_CEClusterFactor;
	CEI->DecayLength = P_CEDecayLength;
	CEI->MaxCC = 0;
	CEI->WeightVec.clear();
	CEI->CCWeightPairs.clear();
	for (;;)
		{
		string Line;
		bool ReadEvoParamsLine(FILE *f, string &Line);
		bool EndOfFile = ReadEvoParamsLine(f, Line);
		if (EndOfFile)
			Die("Premature EOF in CEC%u definition, missing '//'", CEC);

		if (Line.empty())
			continue;

		if (Line == "//")
			break;

		string sLine;
		void DeleteWhiteSpace(const string &In, string &Out);
		DeleteWhiteSpace(Line, sLine);

		vector<string> Pairs;
		void ParseStrVec(const string &In, vector<string> &Out, char Sep);
		ParseStrVec(sLine, Pairs, ';');

		for (vector<string>::const_iterator p = Pairs.begin();
		  p != Pairs.end(); ++p)
			{
			const string &Pair = *p;

			if (Pair.find("F=") == 0)
				{
				double F = atof(Pair.substr(2).c_str());
				if (F < 0.0 || F > 1.0)
					Die("Invalid cluster factor '%s' in CEC%u", Pair.c_str(), CEC);
				CEI->ClusterFactor = F;
				continue;
				}
			else if (Pair.find("L=") == 0)
				{
				CEI->DecayLength = (unsigned) atoi(Pair.substr(2).c_str());
				continue;
				}

			vector<string> v;
			ParseStrVec(Pair, v, ',');
			if (v.size() != 2)
				Die("Invalid specification for CEC%u, '%s'", CEC, Line.c_str());
			char *NotUsed;
			unsigned CC = strtoul(v[0].c_str(), &NotUsed, 10);
			if (CC >= 20)
				Die("Invalid CC%u in CEC%u", CC, CEC);
			if (CC > CEI->MaxCC)
				CEI->MaxCC = (uint8) CC;

			if (IsCodingCEC((uint8) CEC))
				{
				if (!IsCodingCC((uint8) CC))
					Die("Coding CEC%u defined with noncoding CC%u", CEC, CC);
				}
			else
				{
				if (!IsNoncodingCEC((uint8) CEC))
					Die("CEC%u invalid, must be coding CE (2-9) or noncoding CE (12-19)",
					  CEC);
				if (!IsNoncodingCC((uint8) CC))
					Die("Noncoding CEC%u defined with coding or unconstrained CC%u",
					  CEC, CC);
				}

			double Weight = atof(v[1].c_str());
			CEI->CCWeightPairs.push_back(pair<unsigned, double>(CC, Weight));
			}
		}
	if (CEI->CCWeightPairs.empty())
		Die("CEC%u has no CCs", CEC);

	vector<double> Weights(CEI->MaxCC + 1, 0);
	const unsigned N = SIZE(CEI->CCWeightPairs);
	for (unsigned i = 0; i < N; ++i)
		{
		const pair<unsigned, double> &Pair = CEI->CCWeightPairs[i];
		unsigned CC = Pair.first;
		double w = Pair.second;
		asserta(CC < CEI->MaxCC);
		if (Weights[CC] != 0)
			Die("Weight for CC%u specified twice in CEC%u", CC, CEC);
		Weights[CC] = w;
		}
	SetWeightVec(Weights, CEI->WeightVec, 100);
	}

void LogCE(unsigned CEC, const CEInfo &CEI)
	{
	Log("\n");
	Log("CE%u=\n", CEC);
	if (CEI.ClusterFactor != P_CEClusterFactor)
		Log("    F=%5.3f\n", CEI.ClusterFactor);
	if (CEI.DecayLength != P_CEDecayLength)
		Log("    L=%u\n", CEI.DecayLength);
	const unsigned N = SIZE(CEI.CCWeightPairs);
	for (unsigned i = 0; i < N; ++i)
		{
		const pair<unsigned, double> &Pair = CEI.CCWeightPairs[i];
		Log("    %u,%.3g;\n", Pair.first, Pair.second);
		}
	Log("# Examples:\n");
	for (unsigned i = 0; i < 3; ++i)
		{
		const unsigned L = 64;
		ByteVec CCs;
		GetCCs((uint8) CEC, L, CCs, i);
		asserta(SIZE(CCs) == L);
		for (unsigned j = 0; i < L; ++j)
			Log("%u", CCs[j]%10);
		Log("\n");
		}
	Log("//\n");
	}

void LogCEParams()
	{
	for (unsigned CEC = 0; CEC <= MAX_CEC; ++CEC)
		{
		CEInfo *CEI = g_CEIs[CEC];
		if (CEI == 0)
			continue;
		LogCE(CEC, *CEI);
		}
	}
