#include "evo.h"

static void LogBigDist(const char *Msg, const vector<pair<gpos_t, double> > &Rates,
  const double &IntegratedBigRate, const vector<BigLHF> &BigLHFs,
  const vector<unsigned> &BigBins)
	{
	double MaxMicroRate = 0;
	const unsigned N = SIZE(Rates);
	for (unsigned i = 0; i < N; ++i)
		{
		const pair<gpos_t, double> &LengthRate = Rates[i];
		if (LengthRate.first <= P_MicroThreshold)
			MaxMicroRate = max(MaxMicroRate, LengthRate.second);
		}

	const unsigned LHFCount = SIZE(BigLHFs);
	if (MaxMicroRate == 0)
		{
		Log("\n");
		Log("%s rates:\n", Msg);
		}
	else
		{
		for (unsigned i = 0; i < LHFCount; ++i)
			{
			const BigLHF &LHF = BigLHFs[i];
			gpos_t Length = LHF.HiLength;
			double f = LHF.HiRate;
			Log("\n");
			Log("%10"PF"  %10.8f                       ",
			  Length, f);
			unsigned h = (unsigned) ((f/MaxMicroRate)*60);
			if (h == 0)
				Log(".");
			for (unsigned i = 0; i < h; ++i)
				Log("*");
			Log("\n");
			}
		}

	double SumBigRate = 0;
	for (unsigned i = 0; i < LHFCount; ++i)
		{
		const BigLHF &LHF = BigLHFs[i];
		SumBigRate += LHF.IntegratedRate;
		}

	vector<double> BigBinRates;
	GetWeightsFromWeightVec(BigBins, BigBinRates);
	asserta(SIZE(BigBinRates) == SIZE(BigLHFs));
	Normalize(BigBinRates, SumBigRate);

	Log("\n");
	Log("Big ranges:\n");
	Log(" Lo Length   Hi Length     Lo Rate     Hi Rate   Int. Rate       Bases     BinRate  BinErr%\n");
	Log("==========  ==========  ==========  ==========  ==========  ==========  ==========  ======\n");
	for (unsigned i = 0; i < LHFCount; ++i)
		{
		const BigLHF &LHF = BigLHFs[i];
		double IR = LHF.IntegratedRate;
		if (IR == 0)
			IR = 1;
		double BinAccPct = (BigBinRates[i]/IR)*100.0 - 100.0;
		double DiffRate = abs(LHF.HiRate - LHF.LoRate);
		gpos_t Range = (LHF.HiLength - LHF.LoLength + 1);
		double BasesAffected = 
		  ComputeBasesAffected(LHF.LoLength, LHF.HiLength, LHF.LoRate, LHF.HiRate);
		Log("%10"PF"  %10"PF"  %10.3g  %10.3g  %10.3g  %10.3g  %10.3g  %6.1f%%\n",
		  LHF.LoLength,
		  LHF.HiLength,
		  LHF.LoRate,
		  LHF.HiRate,
		  LHF.IntegratedRate,
		  BasesAffected,
		  BigBinRates[i],
		  BinAccPct);
		}
	Log("\n");
	Log("IntegratedBigRate %.3g\n", IntegratedBigRate);
	}

static void LogMicroDist(const char *Msg, const vector<double> &MicroRates,
  const double &IntegratedMicroRate, const vector<unsigned> &MicroBins)
	{
	Log("\n");
	Log("%s rates.\n", Msg);
	Log("\n");
	Log("IntegratedMicroRate %.3g\n", IntegratedMicroRate);
	Log("\n");
	Log("    Length        Rate     BinRate  BinErr%%\n");
	Log("==========  ==========  ==========  ======\n");

	double MaxRate = 0;
	double SumRate = 0;
	for (gpos_t Length = 1; Length <= P_MicroThreshold; ++Length)
		{
		double f = MicroRates[Length];
		SumRate += f;
		MaxRate = max(MaxRate, f);
		}

	vector<double> MicroBinRates;
	GetWeightsFromWeightVec(MicroBins, MicroBinRates);
	unsigned N = SIZE(MicroBinRates);
	asserta(N <= P_MicroThreshold + 1);
	Normalize(MicroBinRates, SumRate);

	for (gpos_t Length = 1; Length < N; ++Length)
		{
		double f = MicroRates[Length];
		double b = MicroBinRates[Length];
		double BinAccPct = f == 0 ? 0 : (b/f - 1.0)*100.0;
		Log("%10"PF"  %10.8f  %10.8f  %6.1f%%  ",
		  Length, f, b, BinAccPct);
		unsigned h = (unsigned) ((f/MaxRate)*60);
		if (h == 0)
			Log(".");
		for (unsigned i = 0; i < h; ++i)
			Log("*");
		Log("\n");
		}
	}

#define LOGMICRODIST(Event)												\
	LogMicroDist(#Event, g_Micro##Event##Rates,							\
	  g_IntegratedMicro##Event##Rate, g_Micro##Event##Bins);

#define LOGBIGDIST(Event)												\
	LogBigDist(#Event, g_##Event##Rates, g_IntegratedBig##Event##Rate,	\
	  g_Big##Event##LHFs, g_Big##Event##Bins);

#define LOGDIST(Event) LOGMICRODIST(Event) LOGBIGDIST(Event)

void LogDists()
	{
	LOGDIST(Delete)
	LOGMICRODIST(Insert)
	LOGDIST(Invert)
	LOGDIST(Copy)
	LOGDIST(Move)
	LOGDIST(NCCE)
	}

static void LogDistParams(const char *Event,
  const vector<pair<gpos_t, double> > &Rates)
	{
	Log("\n");
	Log("%sRates=\n", Event);
	Log("#      Length        Rate\n");
	Log("#  ==========  ==========\n");
	unsigned N = SIZE(Rates);
	for (unsigned i = 0; i < N; ++i)
		{
		const pair<gpos_t, double> &p = Rates[i];
		Log("   %10"PF", %10.4g;\n", p.first, p.second);
		}
	Log("//\n");
	}

static void LogRateSummary(const char *Event, double IntegratedMicroRate,
  double IntegratedBigRate, const vector<double> &MicroRates,
  const vector<BigLHF> &BigLHFs)
	{
	double Rate1to10 = 0;
	unsigned N = min((unsigned) 10, SIZE(MicroRates));
	for (unsigned L = 1; L <= N; ++L)
		Rate1to10 += MicroRates[L];

	double BasesAffected1to10 = 0;
	for (unsigned L = 1; L <= N; ++L)
		BasesAffected1to10 += L*MicroRates[L];

	double MicroBasesAffected = 0;
	for (unsigned L = 1; L < SIZE(MicroRates); ++L)
		MicroBasesAffected += L*MicroRates[L];

	double BigBasesAffected = 0;
	for (unsigned i = 0; i < SIZE(BigLHFs); ++i)
		{
		const BigLHF &LHF = BigLHFs[i];
		BigBasesAffected += ComputeBasesAffected(LHF.LoLength, LHF.HiLength,
		  LHF.LoRate, LHF.HiRate);
		}

	double TotalRate = IntegratedMicroRate + IntegratedBigRate;
	double TotalBases = MicroBasesAffected + BigBasesAffected;

	Log("# %16.16s", Event);

	if (SIZE(MicroRates) > 0)
		{
		Log("  %10.6f", Rate1to10);
		Log("  %10.6f", IntegratedMicroRate);
		}
	else
		{
		Log("            ");
		Log("            ");
		}

	if (SIZE(BigLHFs) > 0)
		Log("  %10.6f", IntegratedBigRate);
	else
		Log("            ");

	Log("  %10.6f", TotalRate);

	if (SIZE(MicroRates) > 0)
		{
		Log("  %10.6f", BasesAffected1to10);
		Log("  %10.6f", MicroBasesAffected);
		}
	else
		{
		Log("            ");
		Log("            ");
		}

	if (SIZE(BigLHFs) > 0)
		Log("  %10.6f", BigBasesAffected);
	else
		Log("            ");

	Log("  %10.6f", TotalBases);

	Log("\n");
	}

static void LogSubstRateSummary()
	{
	Log("# %16.16s", "Substitute");
	Log("  %10.6f", 1.0);
	Log("  %10.6f", 1.0);
	Log("            ");
	Log("  %10.6f", 1.0);
	Log("  %10.6f", 1.0);
	Log("\n");
	}

void LogDistsParams()
	{
	LogDistParams("Delete", g_DeleteRates);
	LogDistParams("Insert", g_InsertRates);
	LogDistParams("Invert", g_InvertRates);
	LogDistParams("Tandem", g_InvertRates);
	LogDistParams("Copy", g_CopyRates);
	LogDistParams("Move", g_MoveRates);
	LogDistParams("InterChrCopy", g_InterChrCopyRates);
	LogDistParams("InterChrMove", g_InterChrCopyRates);
	LogNCCECWeights();

	Log("\n");
	Log("#            Event   1-10 Rate   MicroRate   MacroRate   TotalRate  1-10 Bases  MicroBases  MacroBases  TotalBases\n");
	Log("# ================  ==========  ==========  ==========  ==========  ==========  ==========  ==========  ==========\n");

	vector<double> NoRates;
	vector<BigLHF> NoLHFs;
	LogSubstRateSummary();
	LogRateSummary("Delete", g_IntegratedMicroDeleteRate, g_IntegratedBigDeleteRate,
	  g_MicroDeleteRates, g_BigDeleteLHFs);
	LogRateSummary("Insert", g_IntegratedMicroInsertRate, 0,
	  g_MicroInsertRates, NoLHFs);
	LogRateSummary("Invert", g_IntegratedMicroInvertRate, g_IntegratedBigInvertRate,
	  g_MicroInvertRates, g_BigInvertLHFs);
	LogRateSummary("Tandem", g_IntegratedMicroTandemRate, g_IntegratedBigTandemRate,
	  g_MicroTandemRates, g_BigTandemLHFs);
	LogRateSummary("Copy", g_IntegratedMicroCopyRate, g_IntegratedBigCopyRate,
	  g_MicroCopyRates, g_BigCopyLHFs);
	LogRateSummary("Move", g_IntegratedMicroMoveRate, g_IntegratedBigMoveRate,
	  g_MicroMoveRates, g_BigMoveLHFs);
	LogRateSummary("InterChrCopy", 0, g_IntegratedBigInterChrCopyRate,
	  NoRates, g_BigInterChrCopyLHFs);
	LogRateSummary("InterChrMove", 0, g_IntegratedBigInterChrMoveRate,
	  NoRates, g_BigInterChrMoveLHFs);

	double GrandTotalMicroRate =
	  1 + // substitutions
	  g_IntegratedMicroDeleteRate +
	  g_IntegratedMicroInsertRate +
	  g_IntegratedMicroInvertRate +
	  g_IntegratedMicroTandemRate +
	  g_IntegratedMicroCopyRate +
	  g_IntegratedMicroMoveRate;

	double GrandTotalBigRate =
	  g_IntegratedBigDeleteRate +
	  g_IntegratedBigInvertRate +
	  g_IntegratedBigTandemRate +
	  g_IntegratedBigCopyRate +
	  g_IntegratedBigMoveRate +
	  g_IntegratedBigInterChrCopyRate +
	  g_IntegratedBigInterChrMoveRate;

	double GrandTotalRate = GrandTotalMicroRate + GrandTotalBigRate;

	Log("#                               ==========  ==========  ==========\n");
	Log("#                               %10.6f  %10.6f  %10.6f\n",
	  GrandTotalMicroRate,
	  GrandTotalBigRate,
	  GrandTotalRate);
	}

void LogDistsAlt()
	{
	vector<string> EventNames;
	EventNames.push_back("Delete");
	EventNames.push_back("Insert");
	EventNames.push_back("Invert");
	EventNames.push_back("Tandem");
	EventNames.push_back("Move");
	EventNames.push_back("Copy");

	vector<vector<double> *> Rates;
	Rates.push_back(&g_MicroDeleteRates);
	Rates.push_back(&g_MicroInsertRates);
	Rates.push_back(&g_MicroInvertRates);
	Rates.push_back(&g_MicroTandemRates);
	Rates.push_back(&g_MicroMoveRates);
	Rates.push_back(&g_MicroCopyRates);

	unsigned N = SIZE(EventNames);

	vector<double> SumRates(N, 0);
	vector<double> TotalSumRates(N, 0);

////////
// Rates
////////
	Log("\n");
	Log("Micro Rates:\n");
	Log("     Length");

	for (unsigned i = 0; i < N; ++i)
		Log("  %10.10s", EventNames[i].c_str());
	Log("\n");

	Log("===========");
	for (unsigned i = 0; i < N; ++i)
		Log("  ==========");
	Log("\n");

	for (gpos_t L = 1; L <= 10; ++L)
		{
		Log("%11"PF, L);
		for (unsigned i = 0; i < N; ++i)
			{
			double Rate = (*Rates[i])[L];
			Log("  %10.7f", Rate);
			SumRates[i] += Rate;
			TotalSumRates[i] += Rate;
			}
		Log("\n");
		}

	Log("\n");
	Log("       1-10");
	for (unsigned i = 0; i < N; ++i)
		Log("  %10.7f", SumRates[i]);
	Log("\n");
	
	SumRates.clear();
	SumRates.resize(N, 0);

	for (gpos_t L = 11; L <= P_MicroThreshold; ++L)
		{
		for (unsigned i = 0; i < N; ++i)
			{
			double Rate = (*Rates[i])[L];
			SumRates[i] += Rate;
			TotalSumRates[i] += Rate;
			}
		}
	Log("\n");
	Log("     11-%"PF, P_MicroThreshold);
	for (unsigned i = 0; i < N; ++i)
		Log("  %10.7f", SumRates[i]);
	Log("\n");

	Log("-----------");
	for (unsigned i = 0; i < N; ++i)
		Log("  ----------");
	Log("\n");

	Log("      1-%"PF, P_MicroThreshold);
	for (unsigned i = 0; i < N; ++i)
		Log("  %10.7f", TotalSumRates[i]);
	Log("\n");

	Log("-----------");
	for (unsigned i = 0; i < N; ++i)
		Log("  ----------");
	Log("\n");
	Log("\n");

/////////////////
// Bases affected
/////////////////
	SumRates.clear();
	SumRates.resize(N, 0);
	TotalSumRates.clear();
	TotalSumRates.resize(N, 0);

	Log("\n");
	Log("Micro bases affected:\n");
	Log("     Length");

	for (unsigned i = 0; i < N; ++i)
		Log("  %10.10s", EventNames[i].c_str());
	Log("\n");

	Log("===========");
	for (unsigned i = 0; i < N; ++i)
		Log("  ==========");
	Log("\n");

	for (gpos_t L = 1; L <= 10; ++L)
		{
		Log("%11"PF, L);
		for (unsigned i = 0; i < N; ++i)
			{
			double Rate = L*(*Rates[i])[L];
			Log("  %10.7f", Rate);
			SumRates[i] += Rate;
			TotalSumRates[i] += Rate;
			}
		Log("\n");
		}

	Log("\n");
	Log("       1-10");
	for (unsigned i = 0; i < N; ++i)
		Log("  %10.7f", SumRates[i]);
	Log("\n");
	
	SumRates.clear();
	SumRates.resize(N, 0);

	for (gpos_t L = 11; L <= P_MicroThreshold; ++L)
		{
		for (unsigned i = 0; i < N; ++i)
			{
			double Rate = L*(*Rates[i])[L];
			SumRates[i] += Rate;
			TotalSumRates[i] += Rate;
			}
		}
	Log("\n");
	Log("     11-%"PF, P_MicroThreshold);
	for (unsigned i = 0; i < N; ++i)
		Log("  %10.7f", SumRates[i]);
	Log("\n");

	Log("-----------");
	for (unsigned i = 0; i < N; ++i)
		Log("  ----------");
	Log("\n");

	Log("      1-%"PF, P_MicroThreshold);
	for (unsigned i = 0; i < N; ++i)
		Log("  %10.7f", TotalSumRates[i]);
	Log("\n");

	Log("-----------");
	for (unsigned i = 0; i < N; ++i)
		Log("  ----------");
	Log("\n");
	}
