#include "stack.h"

#define TRACE	0

void OutputBlockTop(const RevFile &In, unsigned BPIndex, const AIVec &AIs,
  const AIBPVec &BPs, coord_t lo, coord_t hi, RevFile &Out)
	{
/***
Return if hi < lo.

This takes care of some legitimate special cases, and also one case
that we don't handle properly.

1. If two blocks have the same lo and the second has a higher confidence
score, nothing to output here, then OutputBlock is called with hi = lo - 1.

2. If two alignments A and B such that hi(A) = hi(B), conf(A) < conf(B)
and hi(A) appears first in the sort order, then OutputBlock is called 
for A with hi = lo - 1.

3. In the following more complex case, OutputBlock is called with hi = lo - 2. 


       ----------------------- A
                       -------------------- B
------------------------ C

Suppose conf(A) < conf(B) < conf(C), and also hi(C) = lo(B), and hi(C) is
sorted before lo(B). Then OutputBlock is called for A with lo=hi(C)+1 and
hi=lo(B)-1=hi(C)-1.

Note that in example (3), we output the last column of C and not the first
column in B. Strictly speaking, we should output the first column of B
and delete the last column of C. We consider this error to be insignificant
and not worth the additional complications needed to fix it.
***/
	if (hi < lo)
		{
		asserta(hi + 1 == lo || hi + 2 == lo);
		return;
		}

#if	TRACE
	Log("OutputBlock BPIndex=%u lo=%u hi=%u\n",
	  BPIndex, (unsigned) lo, (unsigned) hi);
#endif
	const AIBP &BP = BPs[BPIndex];
	unsigned AIIndex = BP.AlnInfoIndex;
	unsigned SegIndex = BP.SegIndex;
	unsigned ChrId = BP.ChrId;
	const AlnInfo &AI = AIs[AIIndex];

	asserta(SegIndex < SIZE(AI.Segs));

	const coord_t ChrLength = In.GetChrLength(g_GenIndex, ChrId);
	asserta(lo < ChrLength);
	asserta(hi < ChrLength);
	asserta(lo <= hi);

	const GCSeg &Seg = AI.Segs[SegIndex];
	if (lo < Seg.lo || hi > Seg.hi)
		{
		Log("\n");
		LogAI(AI);
		Log("\n");
		Log("SegIndex=%u Gen=%u ChrId=%u lo=%u hi=%u\n",
		  SegIndex, g_GenIndex, ChrId, (unsigned) lo, (unsigned) hi);
		Die("OutputBlock: coords out of range");
		}

	unsigned AlnIndex = In.GetAlnIndex(AIIndex);

	StdStrVec Rows;
	In.GetAlnRows(AlnIndex, Rows);
	const unsigned RowCount = SIZE(Rows);

	IntVec ColToOffset;
	IntVec OffsetToCol;
	BoolVec ColHasGap;
	const StdStr &Row = Rows[SegIndex];
	MakeColOffsetVecs(Row, ColToOffset, OffsetToCol, ColHasGap, Seg.Plus);

	bool Plus = Seg.Plus;
	unsigned StartOffset = (lo - Seg.lo);
	unsigned EndOffset = (hi - Seg.lo);

	asserta(StartOffset < SIZE(OffsetToCol));
	asserta(EndOffset < SIZE(OffsetToCol));

	unsigned StartCol = Plus ? OffsetToCol[StartOffset] : OffsetToCol[EndOffset];
	unsigned EndCol = Plus ? OffsetToCol[EndOffset] : OffsetToCol[StartOffset];
	asserta(StartCol <= EndCol);

	StdStrVec OutRows;
	AlnInfo OutAI;

	OutAI.Attrs = AI.Attrs;
	OutAI.Values = AI.Values;

	const GCSegVec &Segs = AI.Segs;
	for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
		{
		const GCSeg &Seg = Segs[RowIndex];
		const StdStr &Row = Rows[RowIndex];
		MakeColOffsetVecs(Row, ColToOffset, OffsetToCol, ColHasGap, Seg.Plus);

		GCSeg OutSeg = Seg;
		unsigned loOffset = (Seg.Plus ? ColToOffset[StartCol] : ColToOffset[EndCol]);
		unsigned hiOffset = (Seg.Plus ? ColToOffset[EndCol] : ColToOffset[StartCol]);

		if (ColHasGap[StartCol] && !Seg.Plus)
			--hiOffset;
		if (ColHasGap[EndCol] && Seg.Plus)
			--hiOffset;

		OutSeg.lo = Seg.lo + loOffset;
		OutSeg.hi = Seg.lo + hiOffset;

		StdStr OutRow;
		for (unsigned ColIndex = StartCol; ColIndex <= EndCol; ++ColIndex)
			{
			char c = Row[ColIndex];
			OutRow.push_back(c);
			}
		if (GetRowLength(OutRow) > 0)
			{
			OutAI.Segs.push_back(OutSeg);
			OutRows.push_back(OutRow);
			ValidateSegLength(OutSeg, OutRow);
			}
		}

	if (SIZE(OutRows) < 2)
		return;

	coord_t SegLength = Seg.GetLength();
	coord_t OutLength = (hi - lo + 1);
	asserta(OutLength > 0);
	double Fraction = OutLength / SegLength;
	asserta(Fraction <= 1.0);

	double ConfScore = AI.GetFloatAttr(ALN_ATTR_ConfScore);
	OutAI.SetFloatAttr(ALN_ATTR_ConfScore, ConfScore*Fraction);
	OutAI.Score = (float) (AI.Score*Fraction);

	Out.AddAln(OutAI, OutRows);
	}
