#include "evo.h"

/***
Parameters are specified in a human read/write-able text file,
formatted as follows.

White space at the beginning and end of a line is ignored.

Empty lines are ignored.

Comments begin with '#' and extend to the end of the line.

Scalar parameters are specified as:

<name>[white]'='[white]<value>[white][EOL]
***/

static void StripLeadingWhitespace(string &Line)
	{
	string Tmp;
	size_t N = Line.size();
	bool Deleting = true;
	for (size_t i = 0; i < N; ++i)
		{
		char c = Line[i];
		if (!isspace(c))
			Deleting = false;
		if (!Deleting)
			Tmp.push_back(c);
		}
	Line = Tmp;
	}

static void StripTrailingWhitespace(string &Line)
	{
	size_t N = Line.size();
	size_t TruncatedSize = 0;
	for (size_t i = 0; i < N; ++i)
		{
		char c = Line[i];
		if (!isspace(c))
			TruncatedSize = i + 1;
		}
	Line.resize(TruncatedSize);
	}

static void StripWhitespace(string &Str)
	{
	StripLeadingWhitespace(Str);
	StripTrailingWhitespace(Str);
	}

static void StripComment(string &Line)
	{
	size_t n = Line.find('#');
	if (n != string::npos)
		Line.resize(n);
	}

static void ParseLine(const string &Line, string &Param,
  string &Value)
	{
	size_t N = Line.size();
	size_t n = Line.find('=');
	if (n == string::npos)
		Die("Invalid parameter definition, missing '=' in line '%s'",
		  Line.c_str());
	if (n == 0 || n + 1 >= N)
		Die("Invalid parameter definition, missing name or value in line '%s'",
		  Line.c_str());

	Param = Line.substr(0, n - 1);
	Value = Line.substr(n + 1);

	StripWhitespace(Param);
	StripWhitespace(Value);
	}

bool ReadEvoParamsLine(FILE *f, string &Line)
	{
	const size_t MAX_LINE = 1023;
	char cLine[MAX_LINE+1];

	bool EndOfFile = ReadLineStdioFile(f, cLine, sizeof(cLine));
	if (EndOfFile)
		return true;

	Line = cLine;
	StripWhitespace(Line);
	StripComment(Line);

	return false;
	}

void LogEvoParams()
	{
	Log("\n");
	Log("# Model parameters\n");
	Log("\n");

#define POS(x, v, min, max)	Log(#x" = %"PF"\n", P_##x);
#define FLT(x, v, min, max)	Log(#x" = %g\n", P_##x);

#include "evoscalarparams.h"

	LogNucMx();
	LogAAMx();
	LogRateParams();
	Log("##\n");
	}

void ReadEvoParams(const char *FileName)
	{
	FILE *f = OpenStdioFile(FileName);

	for (;;)
		{
		string Line;
		bool EndOfFile = ReadEvoParamsLine(f, Line);
		if (EndOfFile)
			break;
		if (Line.empty())
			continue;
		if (Line[0] == '#' || Line[0] == ';')
			continue;

		if (Line == "NucleotideSubstProbs=")
			{
			ReadNucMx(f);
			continue;
			}
		else if (Line == "AminoAcidSubstProbs=")
			{
			ReadAAMx(f);
			continue;
			}
		else if (Line == "DeleteRates=")
			{
			ReadEventDist(f, "Delete");
			continue;
			}
		else if (Line == "InsertRates=")
			{
			ReadEventDist(f, "Insert");
			continue;
			}
		else if (Line == "TandemRates=")
			{
			ReadEventDist(f, "Tandem");
			continue;
			}
		else if (Line == "InvertRates=")
			{
			ReadEventDist(f, "Invert");
			continue;
			}
		else if (Line == "CopyRates=")
			{
			ReadEventDist(f, "Copy");
			continue;
			}
		else if (Line == "MoveRates=")
			{
			ReadEventDist(f, "Move");
			continue;
			}
		else if (Line == "InterChrCopyRates=")
			{
			ReadEventDist(f, "InterChrCopy");
			continue;
			}
		else if (Line == "InterChrMoveRates=")
			{
			ReadEventDist(f, "InterChrMove");
			continue;
			}
		else if (Line == "NCCELengthDist=")
			{
			ReadEventDist(f, "NCCE");
			continue;
			}

		string Param;
		string Value;
		ParseLine(Line, Param, Value);
		SetEvoParam(Param, Value);
		}
	CloseStdioFile(f);
	}
