#include "revfile.h"
#include "utils.h"
#include "progress.h"

#define VALIDATE	0
#define PROGRESS	0

#ifdef _MSC_VER
#pragma warning(disable : 4800)	// int->bool conversion
#endif

void RevComp(char *S, unsigned L);
void RevComp(StdStr &S);

// for getpid():
#ifdef	_MSC_VER
#include <process.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif

#define ReadData(v)		ReadStdioFile(m_f, &v, (unsigned) sizeof(v))
#define WriteData(v)	WriteStdioFile(m_f, &v, (unsigned) sizeof(v))

#define STATE(x, y)	if (m_State != x) Die("RevFile: function call in invalid state"); m_State = y;
#define STATEC(x)	if (m_State != x) Die("RevFile: function call in invalid state");

static FILE *CreateTmpFile(const char *Prefix, std::string &sFileName)
	{
// In case > 1 rev file being created at one time,
// file name must be unique within the process.
	static int TmpFileCounter = 0;
	int pid = getpid();
	char FileName[128];
	sprintf(FileName, "%s.%d.%d.tmp", Prefix, pid, ++TmpFileCounter);
	sFileName = FileName;
	return CreateStdioFile(FileName);
	}

// Hash function used in gawk
// Simple and fast, not especially good
// behavior but good enough for us.
static uint32 Hash(const byte *Key, unsigned Bytes)
	{
	uint32 h = 0;
	for (unsigned i = 0; i < Bytes; ++i)
		{
		byte c = Key[i];
		h = c + (h << 6) + (h << 16) - h;
		}
	return h;
    }

RevFile::RevFile()
	{
	m_FileName = "";
	m_State = Closed;
	m_AlnInfoCount = 0;
	m_AlnCount = 0;
	m_fAlnInfoTOC = 0;
	m_fAlnInfoSegs = 0;
	m_fAlnTOC = 0;
	m_fAlns = 0;
	m_fSeqs = 0;
	m_f = 0;
	m_Hash = 0;
	}

RevFile::~RevFile()
	{
	if (m_State != Closed)
		Close();
	}

void RevFile::Create(const std::string &FileName)
	{
	std::vector<GenInfo> Genomes; // empty vector
	Create(FileName, Genomes);
	}

void RevFile::Create(const std::string &FileName,
  const std::vector<GenInfo> &Genomes)
	{
	STATE(Closed, Writing)

	m_f = CreateStdioFile(FileName.c_str());
	m_FileName = FileName;

	m_Genomes = Genomes;
	m_AlnInfoCount = 0;
	m_AlnCount = 0;

	m_fAlnInfoTOC = CreateTmpFile("alninfotoc", m_AlnInfoTOCFileName);
	m_fAlnInfoSegs = CreateTmpFile("alninfosegs", m_AlnInfoSegsFileName);
	m_fAlnTOC = CreateTmpFile("alntoc", m_AlnTOCFileName);
	m_fAlns = CreateTmpFile("aln", m_AlnsFileName);
	}

void RevFile::Open(const std::string &FileName)
	{
	STATE(Closed, Reading)

	m_f = OpenStdioFile(FileName.c_str());
	m_FileName = FileName;

// Header:
	ReadData(m_Hdr);
	SeekStdioFile(m_f, (off64_t) m_Hdr.GenomesPos);

	if (m_Hdr.Magic != RevFileHdrMagic || m_Hdr.EndHdrMagic != RevFileEndHdrMagic)
		Die("Invalid rev file %s, bad magic [%x,%x] [%x,%x]",
		  m_FileName.c_str(), m_Hdr.Magic, RevFileHdrMagic,
		  m_Hdr.EndHdrMagic, RevFileEndHdrMagic);

	if (m_Hdr.FileFormatVersion != RevFileFormatVersion)
		Die("Rev file %s is file format v%u, this code requires v%u",
		  (unsigned) m_Hdr.FileFormatVersion, RevFileFormatVersion);

	m_AlnInfoCount = (unsigned) m_Hdr.AlnInfoCount;
	m_AlnCount = (unsigned) m_Hdr.AlnCount;

// Genomes:
	const unsigned GenomeCount = m_Hdr.GenomeCount;
	if (GenomeCount == 0)
		Die("Reading .rev file: zero genomes");

	if (GenomeCount > MaxSaneGenomeCount)
		Die("Reading .rev file, header claims and insane number of genomes %u",
		  GenomeCount);

	m_Genomes.resize(GenomeCount);
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		m_Genomes[GenomeIndex].FromFile(m_f);

// Hash table:
	if (m_Hdr.HashTableBytes != m_Hdr.HashTableEntryCount*sizeof(HashTableEntry))
		Die("Hash table size conflict");

	m_Hash = myalloc(HashTableEntry, m_Hdr.HashTableEntryCount);
	ReadStdioFile(m_f, (off64_t) m_Hdr.HashTablePos, m_Hash, m_Hdr.HashTableBytes);

	m_HashTableEntryCount = m_Hdr.HashTableEntryCount;

// Seqs:
	const unsigned SeqCount = m_Hdr.SeqCount;
	m_SeqPosLength.reserve(SeqCount);
	off64_t SeqTOCPos = (off64_t) m_Hdr.SeqTOCPos;
	SeekStdioFile(m_f, SeqTOCPos);

	for (unsigned SeqIndex = 0; SeqIndex < SeqCount; ++SeqIndex)
		{
		DiskSeqTOCEntry TE;
		ReadStdioFile(m_f, &TE, sizeof(TE));

		IntPair GenChr(TE.GenIndex, TE.ChrId);
		LongIntPair PosLength((uint64) TE.FilePos, (uint64) TE.Length);

		std::map<IntPair, unsigned>::value_type v(GenChr, SeqIndex);

		std::pair<std::map<IntPair, unsigned>::iterator, bool> Result = 
		  m_SeqTOC.insert(v);

		if (!Result.second)
			Die("RevFile::Open, dupe Gen,Chr seq TOC entry %u,%u",
			  TE.GenIndex, TE.ChrId);

		m_SeqPosLength.push_back(PosLength);
		}
	}

unsigned RevFile::AddAlnInfo(const AlnInfo &AI)
	{
	return AddAlnInfo(AI, UINT_MAX);
	}

unsigned RevFile::AddAlnInfo(const AlnInfo &AI, unsigned AlnIndex)
	{
	STATEC(Writing)

	unsigned SegCount = (unsigned) AI.Segs.size();
	unsigned AttrCount = (unsigned) AI.Attrs.size();

	DiskAlnInfoTOCEntry AITE;
	AITE.AlnIndex = AlnIndex;
	AITE.SegCount = (uint16) SegCount;
	AITE.AttrCount = (uint16) AttrCount;
	AITE.Score = (float32) AI.Score;
	AITE.FilePos = (uint64) GetStdioFilePos(m_fAlnInfoSegs);

	WriteStdioFile(m_fAlnInfoTOC, &AITE, sizeof(AITE));

	unsigned SegAttrBytes = AITE.SegCount*sizeof(DiskGCSeg) +
	  AITE.AttrCount*sizeof(DiskAlnAttr);

	byte *Buffer = myalloc(byte, SegAttrBytes);

	byte *BuffPos = Buffer;
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &S = AI.Segs[SegIndex];

		const unsigned GenomeCount = (unsigned) m_Genomes.size();
		const unsigned GenomeIndex = S.GenomeIndex;
		const unsigned ChrId = S.ChrId;
		if (GenomeIndex >= GenomeCount)
			Die("RevFile::AddAlnInfo, GenomeIndex[%u]=%u, count=%u",
			  SegIndex, GenomeIndex, GenomeCount);

		DiskGCSeg DS;
		DS.GenomeIndex = (uint16) GenomeIndex;
		DS.ChrId = ChrId;
		DS.lo = (uint32) S.lo;
		DS.hi = (uint32) S.hi;
		DS.Plus = (uint16) S.Plus;

		if ((coord_t) DS.lo != S.lo || (coord_t) DS.hi != S.hi)
			Die("coord_t overflowed uint32");

		memcpy(BuffPos, &DS, sizeof(DS));
		BuffPos += sizeof(DS);

		AddToIndexList((unsigned) S.GenomeIndex, (unsigned) S.ChrId, m_AlnInfoCount);
		}

	for (unsigned AttrIndex = 0; AttrIndex < AttrCount; ++AttrIndex)
		{
		DiskAlnAttr DA;
		unsigned Attr = AI.Attrs[AttrIndex];
		DA.Attr = (uint16) Attr;
		DA.Value = AI.Values[AttrIndex];

		if ((unsigned) DA.Attr != Attr)
			Die("Aln attr overflows uint16");

		memcpy(BuffPos, &DA, sizeof(DA));
		BuffPos += sizeof(DA);
		}

	assert((unsigned) (BuffPos - Buffer) == SegAttrBytes);

	WriteStdioFile(m_fAlnInfoSegs, Buffer, SegAttrBytes);

	myfree(Buffer);
	return m_AlnInfoCount++;
	}

unsigned RevFile::AddAln(const AlnInfo &AI, const StdStrVec &Rows)
	{
	STATEC(Writing)

	if (Rows.size() == 0)
		return UINT_MAX;

	if (AI.Segs.size() != Rows.size())
		Die("RevFile::AddAln, %u segs, %u rows",
		  (unsigned) AI.Segs.size(), (unsigned) Rows.size());

	ValidateSegLengths(AI, Rows);

	unsigned AlnIndex = m_AlnCount++;
	unsigned AlnInfoIndex = AddAlnInfo(AI, AlnIndex);

	unsigned Length0 = (unsigned) Rows[0].size();

	DiskAlnTOCEntry DH;
	DH.RowCount = (uint32) Rows.size();
	DH.ColCount = (uint32) Length0;
	DH.AlnInfoIndex = AlnInfoIndex;
	DH.FilePos = (uint64) GetStdioFilePos(m_fAlns);

	WriteStdioFile(m_fAlnTOC, &DH, sizeof(DH));

	unsigned Bytes = (unsigned) (DH.RowCount*DH.ColCount);
	char *Buffer = myalloc(char, Bytes);
	char *BuffPos = Buffer;
	for (unsigned RowIndex = 0; RowIndex < DH.RowCount; ++RowIndex)
		{
		const std::string &Row = Rows[RowIndex];
		unsigned Length = (unsigned) Row.size();
		if (Length != Length0)
			Die("AddAln, rows different lengths");

		memcpy(BuffPos, Row.c_str(), Length0);

		BuffPos += DH.ColCount;
		}
	assert(BuffPos == Buffer + Bytes);

	WriteStdioFile(m_fAlns, Buffer, Bytes);

	myfree(Buffer);

	return AlnInfoIndex;
	}

void RevFile::Close()
	{
	switch (m_State)
		{
	case Closed:
		return;
	case Reading:
		CloseReading();
		return;
	case Writing:
		CloseWriting();
		return;
	default:
		Die("RevFile::Close, invalid state");
		}
	}

void RevFile::CloseReading()
	{
	STATE(Reading, Closed)

	CloseStdioFile(m_f);
	myfree(m_Hash);
	m_Hash = 0;
	}

void RevFile::CloseWriting()
	{
	STATE(Writing, Closed)

// Write invalid header.
// Write correct values when all other data written.
// This is designed to cause Open to fail unless all writes are completed.
	memset(&m_Hdr, 0, sizeof(m_Hdr));
	WriteStdioFile(m_f, 0, &m_Hdr, sizeof(m_Hdr));

	WriteGenomes();
	WriteAlnInfoTOC();
	WriteAlnInfoSegs();
	WriteAlnTOC();
	WriteAlns();
	WriteHashTable();
	WriteSeqs();

// Update header with final values
	m_Hdr.GenomeCount = (uint32) m_Genomes.size();
	m_Hdr.AlnCount = (uint32) m_AlnCount;
	m_Hdr.AlnInfoCount = (uint32) m_AlnInfoCount;
	m_Hdr.GenomesPos = (uint64) sizeof(m_Hdr);

// Write updated header
	m_Hdr.Magic = RevFileHdrMagic;
	m_Hdr.EndHdrMagic = RevFileEndHdrMagic;
	m_Hdr.FileFormatVersion = RevFileFormatVersion;

	WriteStdioFile(m_f, 0, &m_Hdr, sizeof(m_Hdr));

	m_Genomes.clear();

	m_AlnInfoCount = 0;
	m_AlnCount = 0;

	CloseStdioFile(m_f);
	m_f = 0;

	myfree(m_Hash);
	m_Hash = 0;
	m_HashTableEntryCount = 0;
	}

HashTableEntry *RevFile::FindHashEntry(unsigned GenomeIndex, unsigned ChrId) const
	{
	unsigned AKey[2];
	AKey[0] = GenomeIndex;
	AKey[1] = ChrId;

	byte *Key = (byte *) AKey;

	unsigned TableIndex = Hash(Key, sizeof(AKey))%m_HashTableEntryCount;
	unsigned SafetyCounter = 0;
	HashTableEntry *H = 0;
	for (;;)
		{
		H = &m_Hash[TableIndex];
		if (H->GenomeIndex == NullGenomeIndex)
			return H;

		if (H->GenomeIndex == GenomeIndex && H->ChrId == ChrId)
			return H;

		++TableIndex;
		TableIndex %= m_HashTableEntryCount;

		++SafetyCounter;
		if (SafetyCounter >= m_HashTableEntryCount)
			Die("hash table full");
		}
	}

void RevFile::AddHashEntry(unsigned GenomeIndex, unsigned ChrId, uint64 Pos,
  unsigned AlnCount)
	{
	HashTableEntry *H = FindHashEntry(GenomeIndex, ChrId);
	if (H->GenomeIndex != NullGenomeIndex)
		Die("RevFile::AddHashEntry, already present");

	H->GenomeIndex = (uint16) GenomeIndex;
	H->ChrId = ChrId;
	H->FileOffset = Pos;
	H->AlnInfoCount = AlnCount;
	}

void RevFile::AddToIndexList(unsigned GenomeIndex, unsigned ChrId, unsigned AlnInfoIndex)
	{
	GenChr GC(GenomeIndex, ChrId);
	GenChrToIntListMap::iterator p = m_IndexLists.find(GC);
	if (p == m_IndexLists.end())
		{
		IntList IndexList;
		IndexList.push_back(AlnInfoIndex);
		m_IndexLists.insert(std::pair<GenChr, IntList>(GC, IndexList));
		}
	else
		{
		IntList &IndexList = p->second;
		IndexList.push_back(AlnInfoIndex);
		}
	}

unsigned RevFile::GetAlnCount() const
	{
	return m_AlnCount;
	}

unsigned RevFile::GetAlnInfoCount() const
	{
	return m_AlnInfoCount;
	}

unsigned RevFile::GetGenomeCount() const
	{
	return (unsigned) m_Genomes.size();
	}

const std::string &RevFile::GetGenomeName(unsigned GenomeIndex) const
	{
	if (GenomeIndex >= (unsigned) m_Genomes.size())
		Die("GetGenomeName(%u), invalid genome", GenomeIndex);

	return m_Genomes[GenomeIndex].GetName();
	}

unsigned RevFile::GetAlnInfoX(unsigned AlnInfoIndex, AlnInfoX &AIX) const
	{
	unsigned AlnIndex = GetAlnInfo(AlnInfoIndex, AIX.AI);
	AIX.AlnInfoIndex = AlnInfoIndex;
	AIX.AlnIndex = AlnIndex;
	return AlnIndex;
	}

unsigned RevFile::GetAlnInfo(unsigned AlnInfoIndex, AlnInfo &AI) const
	{
	if (AlnInfoIndex >= m_AlnInfoCount)
		Die("RevFile::GetAlnRows(%u) count=%u", AlnInfoIndex, m_AlnInfoCount);

	off64_t TOCPos = (off64_t)
	  m_Hdr.AlnInfoTOCPos + AlnInfoIndex*sizeof(DiskAlnTOCEntry);

	DiskAlnInfoTOCEntry DATE;
	ReadStdioFile(m_f, TOCPos, &DATE, sizeof(DATE));

	return GetAlnInfoDATE(DATE, AI);
	}

unsigned RevFile::GetAlnInfoDATE(const DiskAlnInfoTOCEntry &DATE, AlnInfo &AI) const
	{
	STATEC(Reading)

	AI.Segs.clear();
	AI.Values.clear();
	AI.Attrs.clear();

	AI.Score = (ai_score_t) DATE.Score;
	AI.AlnIndex = DATE.AlnIndex;

	unsigned Bytes = DATE.SegCount*sizeof(DiskGCSeg) +
	  DATE.AttrCount*sizeof(DiskAlnAttr);

	byte *Buffer = myalloc(byte, Bytes);
	ReadStdioFile(m_f, (off64_t) DATE.FilePos, Buffer, Bytes);

	DiskGCSeg *BuffSeg = (DiskGCSeg *) Buffer;
	for (unsigned SegIndex = 0; SegIndex < DATE.SegCount; ++SegIndex)
		{
		GCSeg S;
		S.GenomeIndex = (unsigned) BuffSeg->GenomeIndex;
		S.ChrId = (unsigned) BuffSeg->ChrId;
		S.hi = (coord_t) BuffSeg->hi;
		S.lo = (coord_t) BuffSeg->lo;
		S.Plus = static_cast<bool>(BuffSeg->Plus);

		AI.Segs.push_back(S);

		++BuffSeg;
		}

	DiskAlnAttr *BuffAA = (DiskAlnAttr *) (Buffer + DATE.SegCount*sizeof(DiskGCSeg));
	for (unsigned AttrIndex = 0; AttrIndex < DATE.AttrCount; ++AttrIndex)
		{
		AlnAttr Attr = (AlnAttr) BuffAA->Attr;
		unsigned Value = (unsigned) BuffAA->Value;

		AI.Attrs.push_back(Attr);
		AI.Values.push_back(Value);

		++BuffAA;
		}

	myfree(Buffer);

	return DATE.AlnIndex;
	}

unsigned RevFile::GetAlnRows(unsigned AlnIndex, StdStrVec &Rows) const
	{
	STATEC(Reading)

	if (AlnIndex >= m_AlnCount)
		Die("RevFile::GetAlnRows(%u) count=%u", AlnIndex, m_AlnCount);

	Rows.clear();

	off64_t TOCPos = (off64_t) m_Hdr.AlnTOCPos + AlnIndex*sizeof(DiskAlnTOCEntry);

	DiskAlnTOCEntry DATE;
	ReadStdioFile(m_f, TOCPos, &DATE, sizeof(DATE));

	unsigned Bytes = (unsigned) (DATE.RowCount*DATE.ColCount);
	char *Buffer = myalloc(char, Bytes);

	ReadStdioFile(m_f, (off64_t) DATE.FilePos, Buffer, Bytes);

	for (unsigned RowIndex = 0; RowIndex < DATE.RowCount; ++RowIndex)
		{
		char *ptrRow = Buffer + RowIndex*DATE.ColCount;
		std::string Row(ptrRow, DATE.ColCount);
		Rows.push_back(Row);
		}

	myfree(Buffer);
	return DATE.AlnInfoIndex;
	}

const std::string &RevFile::GetChrName(unsigned GenomeIndex, unsigned ChrId) const
	{
	if (GenomeIndex >= (unsigned) m_Genomes.size())
		Die("RevFile::GetChrName, genome=%u, count=%u",
		  GenomeIndex, (unsigned) m_Genomes.size());

	const GenInfo &GI = m_Genomes[GenomeIndex];
	const ChrInfo &CI = GI.GetChrInfo(ChrId);
	return CI.GetName();
	}

// Does NOT clear AlnInfoIndexes parameter.
// Allows repeated call to build a larger vector.
void RevFile::GetAlnInfoIndexes(const IntPairVec &GenChrPairs,
  IntVec &AIIndexes) const
	{
// Simply concatenates results from each (gen, chr) pair instead of doing AND.
	Die("Not correctly implemented");
	std::set<unsigned> AlnInfoIndexes;
	const unsigned PairCount = (unsigned) GenChrPairs.size();
#if	PROGRESS
	Log("GetAlnInfoIndexes, %u pairs", PairCount);
#endif
	for (unsigned PairIndex = 0; PairIndex < PairCount; ++PairIndex)
		{
		const std::pair<unsigned, unsigned> GenChr = GenChrPairs[PairIndex];
		unsigned GenIndex = GenChr.first;
		unsigned ChrId = GenChr.second;

		GetAppendAlnInfoIndexesChr(GenIndex, ChrId, AlnInfoIndexes);
		}

	for (std::set<unsigned>::const_iterator p = AlnInfoIndexes.begin();
	  p != AlnInfoIndexes.end(); ++p)
		{
		unsigned AlnInfoIndex = *p;
		AIIndexes.push_back(AlnInfoIndex);
		}
	}

// Does NOT clear AlnInfoIndexes parameter.
// Allows repeated call to build a larger set.
void RevFile::GetAppendAlnInfoIndexesChr(unsigned GenIndex, unsigned ChrId,
  IntSet &AlnInfoIndexes) const
	{
	STATEC(Reading)

	HashTableEntry *H = FindHashEntry(GenIndex, ChrId);
	if (H->GenomeIndex == NullGenomeIndex)
		return;
	
	off64_t Pos = (off64_t) H->FileOffset;
	SeekStdioFile(m_f, Pos);

	unsigned AICount = (unsigned) H->AlnInfoCount;
	uint32 *AIIndexes = myalloc(uint32, AICount);
	ReadStdioFile(m_f, AIIndexes, AICount*sizeof(uint32));

#if	PROGRESS
	Progress("GetAlnInfoIndexesPair, Gen=%u Chr=%u AICount=%u",
	  GenIndex, ChrId, AICount);
#endif
	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = (unsigned) AIIndexes[i];
#if	VALIDATE
		{
		AlnInfo AI;
		GetAlnInfo(AIIndex, AI);
		const GCSeg &Seg = AI.GetSeg(GenIndex);
		if (Seg.ChrId != ChrId)
			Die("RevFile::GetAlnInfoIndexesPair, chr index %u seg %u",
			  ChrId, Seg.ChrId);
		}
#endif
		AlnInfoIndexes.insert(AIIndex);
		}
#if	PROGRESS
	Progress("GetAlnInfoIndexesPair done.");
#endif

	myfree(AIIndexes);
	}

void RevFile::GetAppendAlnInfoIndexesChr(unsigned GenIndex, unsigned ChrId,
  IntVec &AlnInfoIndexes) const
	{
	IntSet setAlnInfoIndexes;
	GetAppendAlnInfoIndexesChr(GenIndex, ChrId, setAlnInfoIndexes);

	for_CIntSet(setAlnInfoIndexes, p)
		{
		AlnInfoIndexes.push_back(*p);
		}
	}

void RevFile::WriteGenomes()
	{
	const unsigned GenomeCount = (unsigned) m_Genomes.size();
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		m_Genomes[GenomeIndex].ToFile(m_f);
	}

// Build TOC (Table Of Contents) for AlnInfo's.
// Do this by appending AlnInfoTOC tmp file to Rev file.
// This requires updating the DiskAlnInfoTOCEntry.FilePos
// field by the offset of the TOC in the Rev file.
void RevFile::WriteAlnInfoTOC()
	{
	off64_t Pos = GetStdioFilePos(m_f);
	m_Hdr.AlnInfoTOCPos = (uint64) Pos;

	if (m_AlnInfoCount > 0)
		{
		unsigned AlnInfoTOCBytes = m_AlnInfoCount*sizeof(DiskAlnInfoTOCEntry);
		unsigned AlnInfoSegsBytes = (unsigned) GetStdioFileSize(m_fAlnInfoSegs);
		asserta(AlnInfoTOCBytes == (unsigned) GetStdioFileSize(m_fAlnInfoTOC));

		uint64 AlnInfoSegsBasePos = (uint64) (m_Hdr.AlnInfoTOCPos + AlnInfoTOCBytes);

	// Read entire AlnInfo TOC tmp file into memory
		DiskAlnInfoTOCEntry *TOCBuffer = myalloc(DiskAlnInfoTOCEntry, m_AlnInfoCount);

		ReadStdioFile(m_fAlnInfoTOC, 0, TOCBuffer, AlnInfoTOCBytes);

		DiskAlnInfoTOCEntry *TOC = TOCBuffer;
		for (unsigned AIIndex = 0; AIIndex < m_AlnInfoCount; ++AIIndex)
			{
			uint64 TmpFilePos = TOC->FilePos;
			asserta(TmpFilePos < AlnInfoSegsBytes);
			TOC->FilePos = TmpFilePos + AlnInfoSegsBasePos;
			TOC++;
			}

	// Append TOC to Rev file
		WriteStdioFile(m_f, TOCBuffer, AlnInfoTOCBytes);

	// Done with AlnInfoTOC
		myfree(TOCBuffer);
		TOCBuffer = 0;
		}

	CloseStdioFile(m_fAlnInfoTOC);
	m_fAlnInfoTOC = 0;

	DeleteStdioFile(m_AlnInfoTOCFileName.c_str());
	}

// Append AlnInfo segs tmp to Rev file.
// No relocation needed, can copy data as-is.
void RevFile::WriteAlnInfoSegs()
	{
	if (m_AlnInfoCount > 0)
		{
		unsigned AlnInfoSegsBytes = (unsigned) GetStdioFileSize(m_fAlnInfoSegs);
		byte *AlnInfoSegsBuffer = myalloc(byte, AlnInfoSegsBytes);
		ReadStdioFile(m_fAlnInfoSegs, 0, AlnInfoSegsBuffer, AlnInfoSegsBytes);
		WriteStdioFile(m_f, AlnInfoSegsBuffer, AlnInfoSegsBytes);
		myfree(AlnInfoSegsBuffer);
		}

	CloseStdioFile(m_fAlnInfoSegs);
	m_fAlnInfoSegs = 0;

	DeleteStdioFile(m_AlnInfoSegsFileName.c_str());
	}

void RevFile::WriteAlnTOC()
	{
	m_Hdr.AlnTOCPos = (uint64) GetStdioFilePos(m_f);
	if (m_AlnCount > 0)
		{
		unsigned AlnTOCBytes = m_AlnCount*sizeof(DiskAlnTOCEntry);
		uint64 AlnBasePos = (uint64) (m_Hdr.AlnTOCPos + AlnTOCBytes);

	// Read entire Aln TOC tmp file into memory
		DiskAlnTOCEntry *TOCBuffer = myalloc(DiskAlnTOCEntry, m_AlnInfoCount);

		ReadStdioFile(m_fAlnTOC, 0, TOCBuffer, AlnTOCBytes);

		DiskAlnTOCEntry *TOC = TOCBuffer;

		for (unsigned AlnIndex = 0; AlnIndex < m_AlnInfoCount; ++AlnIndex)
			{
			TOC->FilePos += AlnBasePos;
			++TOC;
			}

	// Append TOC to Rev file
		WriteStdioFile(m_f, TOCBuffer, AlnTOCBytes);

	// Done with AlnTOC
		myfree(TOCBuffer);
		TOCBuffer = 0;
		}

	CloseStdioFile(m_fAlnTOC);
	m_fAlnInfoTOC = 0;

	DeleteStdioFile(m_AlnTOCFileName.c_str());
	}

// Append Alns tmp to Rev file.
// No relocation needed, can copy data as-is.
void RevFile::WriteAlns()
	{
	if (m_AlnCount > 0)
		{
		//unsigned AlnsBytes = (unsigned) GetStdioFileSize(m_fAlns);
		//byte *AlnsBuffer = myalloc(byte, AlnsBytes);
		//ReadStdioFile(m_fAlns, 0, AlnsBuffer, AlnsBytes);
		//WriteStdioFile(m_f, AlnsBuffer, AlnsBytes);
		//myfree(AlnsBuffer);
		CopyStdioFile(m_fAlns, m_f);
		}

	CloseStdioFile(m_fAlns);
	m_fAlns = 0;

	DeleteStdioFile(m_AlnsFileName.c_str());
	}

void RevFile::WriteSeqs()
	{
	const unsigned SeqCount = (unsigned) m_SeqTOC.size();
	off64_t SeqTOCPos = GetStdioFilePos(m_f);

	m_Hdr.SeqTOCPos = (uint64) SeqTOCPos;
	m_Hdr.SeqCount = SeqCount;

	if (SeqCount == 0)
		return;

// Position of first entry in seq TOC
	DiskSeqTOCEntry ZeroTE;
	memset(&ZeroTE, 0, sizeof(ZeroTE));
	for (unsigned i = 0; i < SeqCount; ++i)
		WriteStdioFile(m_f, &ZeroTE, sizeof(ZeroTE));

// Compute base position for sequence data
	off64_t RevFileSeqPos = SeqTOCPos + SeqCount*sizeof(DiskSeqTOCEntry);
	if (RevFileSeqPos != GetStdioFilePos(m_f))
		Die("SeqBasePos failed");

	for (std::map<IntPair, unsigned>::const_iterator p = m_SeqTOC.begin();
	  p != m_SeqTOC.end(); ++p)
		{
		const std::map<IntPair, unsigned>::value_type &v = *p;

		const IntPair &GenChr = v.first;
		const unsigned &SeqIndex = v.second;

		if (SeqIndex >= SeqCount)
			Die("WriteSeqs, SeqIndex >= SeqCount");

		const LongIntPair &PosLength = m_SeqPosLength[SeqIndex];
		uint64 u64TmpFilePos = PosLength.first;

		off64_t TmpFilePos = (off64_t) u64TmpFilePos;

		if ((uint64) TmpFilePos != u64TmpFilePos)
			Die("RevFile::WriteSeqs, TmpFilePos overflows off64_t");

		uint64 u64Length = PosLength.second;
		unsigned Length = (unsigned) u64Length;
		if ((uint64) Length != u64Length)
			Die("RevFile::WriteSeqs, chr length overflows unsigned int");

	// Read seq from tmp file
		char *Seq = (char *) alloc(Length);
		ReadStdioFile(m_fSeqs, TmpFilePos, Seq, Length);

	// Write seq to rev file
		WriteStdioFile(m_f, RevFileSeqPos, Seq, Length);
		myfree(Seq);

        DiskSeqTOCEntry E;
		E.GenIndex = GenChr.first;
		E.ChrId = GenChr.second;
		E.FilePos = RevFileSeqPos;
		E.Length = PosLength.second;

		off64_t TOCEntryPos = SeqTOCPos + SeqIndex*sizeof(DiskSeqTOCEntry);
		WriteStdioFile(m_f, TOCEntryPos, &E, sizeof(E));

		RevFileSeqPos += Length;
		}

	CloseStdioFile(m_fSeqs);
	m_fSeqs = 0;
	DeleteStdioFile(m_SeqsFileName.c_str());
	}

// Build index lists and hash table, append to file.
void RevFile::WriteHashTable()
	{
	unsigned GenChrPairCount = (unsigned) m_IndexLists.size();
	m_HashTableEntryCount = (unsigned) ((GenChrPairCount + 1)*HashExpandFactor);
	m_Hash = myalloc(HashTableEntry, m_HashTableEntryCount);
	memset(m_Hash, 0xff, m_HashTableEntryCount*sizeof(HashTableEntry));

	for (GenChrToIntListMap::const_iterator p = m_IndexLists.begin();
	  p != m_IndexLists.end(); ++p)
		{
		const GenChr GC = p->first;
		const IntList &Indexes = p->second;

		uint32 IndexCount = (uint32) Indexes.size();
		unsigned AICount = (unsigned) IndexCount;

		uint64 Pos = GetStdioFilePos(m_f);
		AddHashEntry(GC.GenomeIndex, GC.ChrId, Pos, AICount);

	// Create list for one pair in memory buffer & write buffer.
	// Could speed this up with one buffer for all pairs,
	// but this is easier to code & more scalable.
		unsigned Count = (unsigned) Indexes.size();
		unsigned *Buffer = myalloc(unsigned, Count);
		unsigned *BuffPos = Buffer;
		for_CIntList(Indexes, q)
			{
			uint32 Index = (uint32) *q;
			*BuffPos++ = Index;
			}
		WriteStdioFile(m_f, Buffer, Count*sizeof(uint32));
		myfree(Buffer);
		}

	unsigned HashTableBytes = m_HashTableEntryCount*sizeof(HashTableEntry);

	m_Hdr.HashTablePos = GetStdioFilePos(m_f);
	m_Hdr.HashTableBytes = HashTableBytes;
	m_Hdr.HashTableEntryCount = m_HashTableEntryCount;

// Write hash table
	WriteStdioFile(m_f, m_Hash, HashTableBytes);
	}

unsigned RevFile::GetAlnIndex(unsigned AlnInfoIndex) const
	{
	STATEC(Reading)

	off64_t TOCPos = (off64_t)
	  m_Hdr.AlnInfoTOCPos + AlnInfoIndex*sizeof(DiskAlnTOCEntry);

	DiskAlnInfoTOCEntry DATE;
	ReadStdioFile(m_f, TOCPos, &DATE, sizeof(DATE));

	return DATE.AlnIndex;
	}

unsigned RevFile::GetAlnInfoIndex(unsigned AlnIndex) const
	{
	STATEC(Reading)

	off64_t TOCPos = (off64_t) m_Hdr.AlnTOCPos + AlnIndex*sizeof(DiskAlnTOCEntry);

	DiskAlnTOCEntry DATE;
	ReadStdioFile(m_f, TOCPos, &DATE, sizeof(DATE));

	return DATE.AlnInfoIndex;
	}

unsigned RevFile::GetChrId(unsigned GenomeIndex, const std::string &ChrName) const
	{
	const GenInfo &GI = GetGenomeInfo(GenomeIndex);
	return GI.GetChrId(ChrName);
	}

unsigned RevFile::GetChrId(unsigned GenomeIndex, unsigned ChrIndex) const
	{
	const GenInfo &GI = GetGenomeInfo(GenomeIndex);
	return GI.GetChrId(ChrIndex);
	}

unsigned RevFile::GetChrIndex(unsigned GenomeIndex, unsigned ChrId) const
	{
	const GenInfo &GI = GetGenomeInfo(GenomeIndex);
	return GI.GetChrIndex(ChrId);
	}

const GenInfo &RevFile::GetGenomeInfo(unsigned GenomeIndex) const
	{
	unsigned GenomeCount = (unsigned) m_Genomes.size();
	if (GenomeIndex >= GenomeCount)
		Die("RevFile::GetGenomeInfo(%u), count=%u", GenomeIndex, GenomeCount);
	return m_Genomes[GenomeIndex];
	}

GenInfo *RevFile::GetModifiableGenomeInfoPtr(const std::string &GenomeName,
  unsigned *ptrGenomeIndex)
	{
	unsigned GenomeCount = (unsigned) m_Genomes.size();
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		GenInfo &GI = m_Genomes[GenomeIndex];
		if (GenomeName == GI.GetName())
			{
			*ptrGenomeIndex = GenomeIndex;
			return &GI;
			}
		}
	return 0;
	}

const std::vector<GenInfo> &RevFile::GetGenomes() const
	{
	return m_Genomes;
	}

const ChrInfo &RevFile::GetChrInfo(unsigned GenomeIndex, unsigned ChrId) const
	{
	unsigned GenomeCount = (unsigned) m_Genomes.size();
	if (GenomeIndex >= GenomeCount)
		Die("RevFile::GetChrInfo(%u), count=%u", GenomeIndex, GenomeCount);

	const GenInfo &GI = m_Genomes[GenomeIndex];
	return GI.GetChrInfo(ChrId);
	}

// Determine active chr pairs by reading all segs in all AIs (slow!)
void RevFile::GetActiveChrPairs(unsigned GenomeIndexA, unsigned GenomeIndexB,
  std::set<IntPair> &ChrIdPairs) const
	{
	ChrIdPairs.clear();

	const unsigned AICount = GetAlnInfoCount();
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		AlnInfo AI;
		GetAlnInfo(AIIndex, AI);
		if (AI.Segs.size() != 2)
			continue;

		const GCSeg *SegA = AI.GetSegPtr(GenomeIndexA);
		const GCSeg *SegB = AI.GetSegPtr(GenomeIndexB);

		if (SegA != 0 && SegB != 0)
			{
			IntPair ChrIdPair(SegA->ChrId, SegB->ChrId);
			ChrIdPairs.insert(ChrIdPair);
			}
		}
	}

void RevFile::GetActiveChrTuples(const std::vector<unsigned> /*GenomeIndexes*/,
  std::set<IntVec> &/*ChrIdTuples*/) const
	{
	Die("Not implemented");
	}

void RevFile::GetAlnInfos(const IntVec &AlnInfoIndexes,
  std::vector<AlnInfo> &AlnInfos) const
	{
	const unsigned AICount = (unsigned) AlnInfoIndexes.size();
	AlnInfos.reserve(AICount);

	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = AlnInfoIndexes[i];

		AlnInfo AI;
		GetAlnInfo(AIIndex, AI);

		AlnInfos.push_back(AI);
		}
	}

void RevFile::GetAlnInfoXs(const IntVec &AlnInfoIndexes,
  std::vector<AlnInfoX> &AlnInfoXs) const
	{
	const unsigned AIXCount = (unsigned) AlnInfoIndexes.size();
	AlnInfoXs.reserve(AIXCount);

	for (unsigned i = 0; i < AIXCount; ++i)
		{
		unsigned AIIndex = AlnInfoIndexes[i];

		AlnInfoX AIX;
		GetAlnInfo(AIIndex, AIX.AI);

		AIX.AlnInfoIndex = AIIndex;
		AIX.AlnIndex = GetAlnIndex(AIIndex);

		AlnInfoXs.push_back(AIX);
		}
	}

void RevFile::GetAlnInfos(const IntSet &AIIndexes, AIVec &AIs) const
	{
	const unsigned AICount = (unsigned) AIIndexes.size();
	AIs.reserve(AICount);

	for (IntSet::const_iterator p = AIIndexes.begin(); p != AIIndexes.end(); ++p)
		{
		unsigned AIIndex = *p;

		AlnInfo AI;
		GetAlnInfo(AIIndex, AI);

		AIs.push_back(AI);
		}
	}

unsigned RevFile::GetGenomeIndex(const std::string &GenomeName) const
	{
	const unsigned GenomeCount = GetGenomeCount();
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		const std::string &Name = GetGenomeName(GenomeIndex);
		if (GenomeName == Name)
			return GenomeIndex;
		}
	Die("RevFile::GetGenomeIndex(%s), not found", GenomeName.c_str());
	return UINT_MAX;
	}

unsigned RevFile::AddSeq(const std::string &GenomeName,
  const std::string &ChrName, const char *Seq, coord_t Length)
	{
	STATEC(Writing)

	unsigned GenIndex;
	GenInfo *GI = GetModifiableGenomeInfoPtr(GenomeName, &GenIndex);
	if (GI == 0)
		{
		GenIndex = AddGenome(GenomeName);
		GI = &(m_Genomes[GenIndex]);
		}

	unsigned ChrId = UINT_MAX;
	ChrInfo *CI = GI->GetModifiableChrInfoPtr(ChrName);
	if (CI == 0)
		ChrId = GI->AddChr(ChrName, Length);
	else
		{
		ChrId = CI->GetId();
		CI->SetLength(Length);
		}

	if (m_fSeqs == 0)
		m_fSeqs = CreateTmpFile("seq", m_SeqsFileName);

	off64_t FilePos = GetStdioFilePos(m_fSeqs);

	IntPair GenChr(GenIndex, ChrId);
	LongIntPair PosLength((uint64) FilePos, (uint64) Length);

	unsigned SeqIndex = (unsigned) m_SeqPosLength.size();
	m_SeqPosLength.push_back(PosLength);

	std::map<IntPair, unsigned>::value_type v(GenChr, SeqIndex);

	std::pair<std::map<IntPair, unsigned>::iterator, bool> Result =
	  m_SeqTOC.insert(v);

	if (!Result.second)
		Die("RevFile::AddSeq(%u,%u), already added", GenIndex, ChrId);

	size_t stLength = (size_t) Length;
	if ((coord_t) stLength != Length)
		Die("Sequence length overflows size_t");

	WriteStdioFile(m_fSeqs, Seq, stLength);

	return SeqIndex;
	}

unsigned RevFile::AddGenome(const std::string &Name)
	{
	unsigned GenIndex = (unsigned) m_Genomes.size();

	GenInfo GI;
	GI.Create(Name);

	m_Genomes.push_back(GI);

	return GenIndex;
	}

const ChrInfo *RevFile::GetChrInfo(unsigned GenIndex,
  const std::string &ChrName) const
	{
	const GenInfo &GI = GetGenomeInfo(GenIndex);
	return GI.GetChrInfoPtr(ChrName);
	}

coord_t RevFile::GetSeqLengthByIndex(unsigned SeqIndex) const
	{
	unsigned SeqCount = GetSeqCount();
	if (SeqIndex >= SeqCount)
		Die("RevFile::GetSeqLengthByIndex(%u), SeqCount=%u", SeqIndex, SeqCount);

	const LongIntPair &PosLength = m_SeqPosLength[SeqIndex];
	return (coord_t) PosLength.second;
	}

bool RevFile::IsSeqPresent(unsigned GenomeIndex, unsigned ChrId) const
	{
	IntPair GenChr(GenomeIndex, ChrId);
	std::map<IntPair, unsigned>::const_iterator p = m_SeqTOC.find(GenChr);
	return p != m_SeqTOC.end();
	}

coord_t RevFile::GetSeqLength(unsigned GenomeIndex, unsigned ChrId) const
	{
	IntPair GenChr(GenomeIndex, ChrId);
	std::map<IntPair, unsigned>::const_iterator p = m_SeqTOC.find(GenChr);
	if (p == m_SeqTOC.end())
		Die("RevFile::GetSeqLength, not found");

	unsigned SeqIndex = p->second;
	unsigned SeqCount = GetSeqCount();
	if (SeqIndex >= SeqCount)
		Die("RevFile::GetSeqLength, SeqIndex >= SeqCount");

	const LongIntPair &PosLength = m_SeqPosLength[SeqIndex];
	return (coord_t) PosLength.second;
	}

void RevFile::GetSeq(unsigned GenomeIndex, unsigned ChrId, char *Seq, unsigned Bytes) const
	{
	STATEC(Reading)

	IntPair GenChr(GenomeIndex, ChrId);
	std::map<IntPair, unsigned>::const_iterator p = m_SeqTOC.find(GenChr);
	if (p == m_SeqTOC.end())
		Die("RevFile::GetSeq, not found");

	unsigned SeqIndex = p->second;
	unsigned SeqCount = GetSeqCount();
	if (SeqIndex >= SeqCount)
		Die("RevFile::GetSeq, SeqIndex >= SeqCount");

	const LongIntPair &PosLength = m_SeqPosLength[SeqIndex];
	off64_t Pos = (off64_t) PosLength.first;
	unsigned Length = (unsigned) PosLength.second;

	if (Bytes < Length)
		Die("RevFile::GetSeq, buffer bytes=%u shorter than seq length=%u",
		  Bytes, Length);

	ReadStdioFile(m_f, Pos, Seq, Length);
	}

void RevFile::GetSubSeq3(const GCSeg &Seg, StdStr &Seq) const
	{
	GetSubSeq2(Seg.GenomeIndex, Seg.ChrId, Seq, Seg.lo, Seg.GetLength());
	if (!Seg.Plus)
		RevComp(Seq);
	}

void RevFile::GetSubSeq2(unsigned GenomeIndex, unsigned ChrId, StdStr &ssSeq,
  unsigned From, unsigned Length, bool DoRevComp) const
	{
	char *Seq = (char *) malloc(Length+1);
	if (Seq == 0)
		Die("RevFile::GetSubSeq, Out of memory (requested %u bytes)", Length);
	GetSubSeq(GenomeIndex, ChrId, Seq, From, Length, DoRevComp);
	Seq[Length] = 0;
	ssSeq = Seq;
	delete[] Seq;
	}

void RevFile::GetSubSeq(unsigned GenomeIndex, unsigned ChrId, char *Seq,
  unsigned From, unsigned Length, bool DoRevComp) const
	{
	STATEC(Reading)

	IntPair GenChr(GenomeIndex, ChrId);
	std::map<IntPair, unsigned>::const_iterator p = m_SeqTOC.find(GenChr);
	if (p == m_SeqTOC.end())
		Die("RevFile::GetSubSeq, not found");

	unsigned SeqIndex = p->second;
	unsigned SeqCount = GetSeqCount();
	if (SeqIndex >= SeqCount)
		Die("RevFile::GetSeq, SeqIndex >= SeqCount");

	const LongIntPair &PosLength = m_SeqPosLength[SeqIndex];
	off64_t Pos = (off64_t) PosLength.first;
	unsigned SeqLength = (unsigned) PosLength.second;

	if (From + Length - 1 >= SeqLength)
		Die("RevFile::GetSubSeq, subseq past end of seq");

	ReadStdioFile(m_f, Pos + From, Seq, Length);

	if (DoRevComp)
		RevComp(Seq, Length);
	}

unsigned RevFile::GetSeqCount() const
	{
	return (unsigned) m_SeqPosLength.size();
	}

void RevFile::GetSeqByIndex(unsigned SeqIndex, char *Seq, unsigned Bytes,
  unsigned *ptrGenomeIndex, unsigned *ptrChrId) const
	{
	unsigned SeqCount = GetSeqCount();
	if (SeqIndex >= SeqCount)
		Die("RevFile::GetSeqByIndex(%u), out of range", SeqIndex);

	DiskSeqTOCEntry TE;
	off64_t TEPos = (off64_t) m_Hdr.SeqTOCPos + SeqIndex*sizeof(TE);
	ReadStdioFile(m_f, TEPos, &TE, sizeof(TE));

	*ptrGenomeIndex = TE.GenIndex;
	*ptrChrId = TE.ChrId;
	unsigned Length = (unsigned) TE.Length;
	off64_t SeqPos = (off64_t) TE.FilePos;

	if (Length > Bytes)
		Die("RevFile::GetSeqByIndex, buff size=%u < seq length=%u",
		  Bytes, Length);
	ReadStdioFile(m_f, SeqPos, Seq, Length);
	}

unsigned RevFile::GetChrCount(unsigned GenomeIndex) const
	{
	const GenInfo &GI = GetGenomeInfo(GenomeIndex);
	return GI.GetChrCount();
	}

// Special case for getting all AlnInfos in file.
// For now, dumb implementation; can optimize
// later if desired.
void RevFile::GetAllAlnInfos(AIVec &AIs, bool ShowProgress) const
	{
	AIs.clear();
	unsigned AICount = GetAlnInfoCount();
	AIs.reserve(AICount);
	if (ShowProgress)
		ProgressStart("Reading %u AIs", AICount);

	SetStdioFilePos(m_f, m_Hdr.AlnInfoTOCPos);

	DiskAlnInfoTOCEntry *TOCBuffer = myalloc(DiskAlnInfoTOCEntry, AICount);

	unsigned TOCBytes = m_Hdr.AlnInfoCount*sizeof(DiskAlnTOCEntry);
	ReadStdioFile(m_f, m_Hdr.AlnInfoTOCPos, TOCBuffer, TOCBytes);

	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0 && ShowProgress)
			ProgressStep(AIIndex, AICount);
		DiskAlnInfoTOCEntry *DATE = TOCBuffer + AIIndex;

		AlnInfo AI;
		GetAlnInfoDATE(*DATE, AI);
		AIs.push_back(AI);
		}
	if (ShowProgress)
		ProgressDone();

	myfree(TOCBuffer);
	}

void RevFile::GetAllAlnInfoXs(AIXVec &AIXs, bool ShowProgress) const
	{
	AIXs.clear();
	unsigned AICount = GetAlnInfoCount();
	AIXs.reserve(AICount);
	if (ShowProgress)
		ProgressStart("Reading %u AIs", AICount);

	SetStdioFilePos(m_f, m_Hdr.AlnInfoTOCPos);

	DiskAlnInfoTOCEntry *TOCBuffer = myalloc(DiskAlnInfoTOCEntry, AICount);

	unsigned TOCBytes = m_Hdr.AlnInfoCount*sizeof(DiskAlnTOCEntry);
	ReadStdioFile(m_f, m_Hdr.AlnInfoTOCPos, TOCBuffer, TOCBytes);

	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0 && ShowProgress)
			ProgressStep(AIIndex, AICount);
		DiskAlnInfoTOCEntry *DATE = TOCBuffer + AIIndex;

		AlnInfoX &AIX = AIXs[AIIndex];
		GetAlnInfoDATE(*DATE, AIX.AI);
		AIX.AlnIndex = AIX.AI.AlnIndex;
		AIX.AlnInfoIndex = AIIndex;
		AIXs.push_back(AIX);
		}
	if (ShowProgress)
		ProgressDone();

	myfree(TOCBuffer);
	}

coord_t RevFile::GetChrLength(unsigned GenomeIndex, unsigned ChrId) const
	{
	const ChrInfo &CI = GetChrInfo(GenomeIndex, ChrId);
	return CI.GetLength();
	}

void RevFile::CopyAln(const RevFile &rhs, unsigned rhsAlnInfoIndex)
	{
	STATEC(Writing)

	AlnInfo AI;
	rhs.GetAlnInfo(rhsAlnInfoIndex, AI);

	unsigned rhsAlnIndex = rhs.GetAlnIndex(rhsAlnInfoIndex);
	if (rhsAlnIndex == UINT_MAX)
		AddAlnInfo(AI);
	else
		{
		StdStrVec Rows;
		rhs.GetAlnRows(rhsAlnIndex, Rows);
		AddAln(AI, Rows);
		}
	}

void RevFile::ValidateSegLengths(const AlnInfo &AI, const StdStrVec &Rows)
	{
	const unsigned RowCount = (unsigned) Rows.size();
	const unsigned SegCount = (unsigned) AI.Segs.size();
	if (RowCount != SegCount)
		{
		LogAI(AI);
		Die("RevFile::ValidateSegLengths, RowCount=%u != SegCount=%u",
		  RowCount, SegCount);
		}
	if (RowCount == 0)
		return;

	const unsigned ColCount = (unsigned) Rows[0].size();
	for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
		{
		const std::string &Row = Rows[RowIndex];
		const GCSeg &Seg = AI.Segs[RowIndex];
		if (Seg.lo > Seg.hi)
			{
			LogAI(AI);
			Die("RevFile::ValidateSegLengths, Invalid Seg.lo=%u > Seg.hi=%u",
			  (unsigned) Seg.lo, (unsigned) Seg.hi);
			}

		unsigned SegLength = (Seg.hi - Seg.lo + 1);
		const unsigned N = (unsigned) Row.size();
		if (N != ColCount)
			{
			LogAI(AI);
			Die("RevFile::ValidateSegLengths, row %u size=%u, row 0 size=%u",
			  RowIndex, N, ColCount);
			}

		unsigned RowLength = 0;
		for (unsigned i = 0; i < ColCount; ++i)
			if (Row[i] != '-')
				++RowLength;
			
		if (SegLength != RowLength)
			{
			LogAIN(AI);
			Log("Seq=%.64s\n", Row.c_str());
			Die("RevFile::ValidateSegLengths, Row=%u SegLength %u != RowLength %u",
			  RowIndex, SegLength, RowLength);
			}
		}
	}

void RevFile::GetSeqGenChrByIndex(unsigned SeqIndex, unsigned *ptrGenomeIndex,
  unsigned *ptrChrId) const
	{
	unsigned SeqCount = GetSeqCount();
	if (SeqIndex >= SeqCount)
		Die("RevFile::GetSeqGenChrByIndex(%u), out of range", SeqIndex);

	DiskSeqTOCEntry TE;
	off64_t TEPos = (off64_t) m_Hdr.SeqTOCPos + SeqIndex*sizeof(TE);
	ReadStdioFile(m_f, TEPos, &TE, sizeof(TE));

	*ptrGenomeIndex = TE.GenIndex;
	*ptrChrId = TE.ChrId;
	}

const std::string &RevFile::GetFileName() const
	{
	return m_FileName;
	}
