#include "seqfile.h"
#include "utils.h"

#define STATE(x, y)	if (m_State != x) Die("SeqFile: function call in invalid state"); m_State = y;
#define STATEC(x)	if (m_State != x) Die("SeqFile: function call in invalid state");

#define ReadData(v)		ReadStdioFile(m_f, &v, (unsigned) sizeof(v))
#define WriteData(v)	WriteStdioFile(m_f, &v, (unsigned) sizeof(v))

void RevComp(char *S, unsigned L);

SeqFile::SeqFile()
	{
	m_FileName = 0;
	m_f = NULL;
	m_Creating = false;
	m_ChrCount = 0;
	}

SeqFile::~SeqFile()
	{
	Close();
	}

void SeqFile::Open(const char *FileName)
	{
	if (m_Creating)
		Die("SeqFile::Open after Create");
	if (m_f != NULL)
		Die("SeqFile::Open without Close");

	m_FileName = strsave(FileName);

	m_ChrPosIn.clear();
	m_ChrPosOut.clear();

	m_f = OpenStdioFile(FileName);
	ReadData(m_Header);
	if (m_Header.Magic != SeqFileMagic)
		Die("Invalid SeqFile %s (bad magic)", FileName);

// On platforms where sizeof(off_t) < sizeof(uint64), may not
// be able to read files created on Linux (> 2GB).
	if (m_Header.FileSize != (uint64) (off_t) m_Header.FileSize)
		Die("Seq file too big");

	SeekStdioFile(m_f, (off_t) m_Header.ChrNameIndexStartPos);
	m_ChrNameIndex.FromFile(m_f);

	if (m_ChrNameIndex.size() != m_Header.ChrCount)
		Die("Invalid SeqFile, chr name index count %u %u",
		  (unsigned) m_ChrNameIndex.size(), m_Header.ChrCount);

	m_ChrCount = m_Header.ChrCount;

	SeekStdioFile(m_f, (off_t) m_Header.ChrSeqDataIndexStartPos);

	m_ChrPosIn.clear();
	m_ChrLengthIn.clear();

	m_ChrPosIn.reserve(m_ChrCount);
	m_ChrLengthIn.reserve(m_ChrCount);

	for (unsigned ChrIndex = 0; ChrIndex < m_ChrCount; ++ChrIndex)
		{
		uint64 Pos;
		ReadData(Pos);
		m_ChrPosIn.push_back(Pos);
		}

	for (unsigned ChrIndex = 0; ChrIndex < m_ChrCount; ++ChrIndex)
		{
		coord_t Length;
		ReadData(Length);
		m_ChrLengthIn.push_back(Length);
		}

	m_Creating = false;
	}

const ChrNameIndex &SeqFile::GetChrNameIndex() const
	{
	return m_ChrNameIndex;
	}

unsigned SeqFile::GetChrCount() const
	{
	return m_Header.ChrCount;
	}

coord_t SeqFile::GetChrLength(unsigned ChrIndex) const
	{
	if (m_Creating)
		Die("SeqFile::GetChrLength not supported while creating");
	if (ChrIndex >= m_ChrCount)
		Die("SeqFile::GetChrLength(%u), index size=%u",
		  ChrIndex, m_ChrCount);
	return m_ChrLengthIn[ChrIndex];
	}

void SeqFile::GetSeqData(unsigned ChrIndex, coord_t Pos, char *Buffer,
  coord_t Length) const
	{
	if (ChrIndex >= m_ChrCount)
		Die("SeqFile::GetSeqData(%u), index size=%u",
		  ChrIndex, m_ChrCount);

	uint64 ChrStartPos = m_ChrPosIn[ChrIndex];
	coord_t ChrLength = GetChrLength(ChrIndex);
	if (Pos + Length - 1 >= ChrLength)
		Die("SeqFile::GetSeqData pos=%u length=%u end=%u past chr end %u",
		  (unsigned) Pos,
		  (unsigned) Length,
		  (unsigned) (Pos + Length - 1),
		  (unsigned) (ChrLength - 1));

	SeekStdioFile(m_f, (off_t) (ChrStartPos + Pos));
	ReadStdioFile(m_f, Buffer, (off_t) Length);
	}

void SeqFile::Create(const char *FileName, const char *GenomeName)
	{
	m_f = CreateStdioFile(FileName);
	memset(&m_Header, 0, sizeof(m_Header));
	safestrcpy(m_Header.GenomeName, GenomeName, sizeof(m_Header.GenomeName));
	WriteData(m_Header);
	m_Creating = true;
	}

void SeqFile::AppendChr(const char *ChrName)
	{
	unsigned ChrIndex = (unsigned) m_ChrPosOut.size();

	if (ChrIndex > 0)
		m_ChrLengthOut.push_back(m_CurrChrLength);
	m_CurrChrLength = 0;

	m_ChrNameIndex.Add(ChrIndex, ChrName);

	uint64 Pos = GetStdioFilePos(m_f);
	m_ChrPosOut.push_back(Pos);

	++m_ChrCount;
	assert(m_ChrCount == m_ChrPosOut.size());
	}

void SeqFile::AppendChrData(const char *SeqData, coord_t Length)
	{
	WriteStdioFile(m_f, SeqData, (off_t) Length);
	m_CurrChrLength += Length;
	}

void SeqFile::Close()
	{
	if (m_f == 0)
		return;

	if (m_FileName != 0)
		{
		myfree(m_FileName);
		m_FileName = 0;
		}

	if (!m_Creating)
		{
		CloseStdioFile(m_f);
		m_ChrPosIn.clear();
		m_ChrPosOut.clear();
		m_f = 0;
		return;
		}

	m_Header.ChrSeqDataIndexStartPos = GetStdioFilePos(m_f);

	if (m_ChrCount > 0)
		m_ChrLengthOut.push_back(m_CurrChrLength);

	assert(m_ChrPosOut.size() == m_ChrCount);
	assert(m_ChrLengthOut.size() == m_ChrCount);

	for (std::list<uint64>::const_iterator p = m_ChrPosOut.begin();
	  p != m_ChrPosOut.end(); ++p)
		{
		uint64 Pos = *p;
		WriteData(Pos);
		}

	for (std::list<coord_t>::const_iterator p = m_ChrLengthOut.begin();
	  p != m_ChrLengthOut.end(); ++p)
		{
		coord_t Length = *p;
		WriteData(Length);
		}

	m_Header.ChrCount = m_ChrCount;
	m_Header.ChrNameIndexStartPos = GetStdioFilePos(m_f);
	m_ChrNameIndex.ToFile(m_f);
	m_Header.Magic = SeqFileMagic;
	m_Header.FileSize = GetStdioFilePos(m_f);
	SeekStdioFile(m_f, 0);
	WriteData(m_Header);

	CloseStdioFile(m_f);
	m_f = 0;

	m_ChrPosIn.clear();
	m_ChrPosOut.clear();

	m_Creating = false;
	}

const char *SeqFile::GetGenomeName() const
	{
	return m_Header.GenomeName;
	}

const char *SeqFile::GetChrName(unsigned ChrIndex) const
	{
	return m_ChrNameIndex.GetName(ChrIndex);
	}

unsigned SeqFile::GetChrIndex(const char *ChrName) const
	{
	unsigned Index;
	bool Found = m_ChrNameIndex.GetIndex(ChrName, &Index);
	if (!Found)
		Die("Seq file %s: Chr %s not found", m_FileName, ChrName);
	return Index;
	}
