#include "stack.h"
#include "timing.h"

/***
Stacker functions that deal with the issue of whether an output
column can be the predecessor or successor of another column.
***/

/***
Given a segment, report whether a base-with-sign can be a successor
(Next=true) or predecessor (Next=false) of that segment. This may
require an inversion (change of sign) of the base, in which case
CONSEC_VALID_INVERT is returned.
***/
CONSEC Stacker::ValidConsecutiveSeg(const GCSeg &Seg, const BaseWithSign &BWS,
  bool Next)
	{
	if (Seg.GenomeIndex != BWS.Base.GenIndex)
		return CONSEC_INVALID;
	if (Seg.ChrId != BWS.Base.ChrId)
		return CONSEC_INVALID;
	if (Next)
		{
		if (Seg.Plus && BWS.Base.Pos == Seg.hi + 1)
			{
			if (BWS.Plus)
				return CONSEC_VALID;
			else
				return CONSEC_VALID_INVERT;
			}
		if (!Seg.Plus && BWS.Base.Pos + 1 == Seg.lo)
			{
			if (!BWS.Plus)
				return CONSEC_VALID;
			else
				return CONSEC_VALID_INVERT;
			}
		return CONSEC_INVALID;
		}
	else
		{
		assert(!Next);
		if (Seg.Plus && BWS.Base.Pos + 1 == Seg.lo)
			{
			if (BWS.Plus)
				return CONSEC_VALID;
			else
				return CONSEC_VALID_INVERT;
			}
		if (!Seg.Plus && BWS.Base.Pos == Seg.hi + 1)
			{
			if (!BWS.Plus)
				return CONSEC_VALID;
			else
				return CONSEC_VALID_INVERT;
			}
		return CONSEC_INVALID;
		}
	}

/***
Given a base-with-sign and a set of segments, identify a segment
such that the base could be a succecessor (Next=true) or predecessor
(Next=false) of the segment.
***/
unsigned Stacker::GetAdjacentSegIndex(const BaseWithSign &BWS,
  const GCSegVec &Segs, CONSEC &Consec, bool Next)
	{
	Consec = CONSEC_ILLEGAL_ENUM_VALUE;

	bool SameChr = false;
	const unsigned SegCount = SIZE(Segs);
	FORIX(Seg)
		{
		const GCSeg &Seg = Segs[SegIndex];
		Consec = ValidConsecutiveSeg(Seg, BWS, Next);
		switch (Consec)
			{
		case CONSEC_VALID:
			return SegIndex;
		case CONSEC_VALID_INVERT:
			return SegIndex;

		case CONSEC_INVALID:
			break;			

		case CONSEC_NEW_CHR:
		case CONSEC_ILLEGAL_ENUM_VALUE:
		default:
			asserta(false);
			}

		if (BWS.Base.GenIndex == Seg.GenomeIndex && BWS.Base.ChrId == Seg.ChrId)
			SameChr = true;
		}

	if (SameChr)
		Consec = CONSEC_INVALID;
	else
		Consec = CONSEC_NEW_CHR;
	return UINT_MAX;
	}

/***
Given a segments and a set of bases-with-sign, identify a base
such that the segment could be a succecessor (Next=true) or
predecessor (Next=false) of the base.
***/
unsigned Stacker::GetAdjacentBaseIndex(const GCSeg &Seg,
  const std::vector<BaseWithSign> &Bases, CONSEC &Consec, bool Next)
	{
	Consec = CONSEC_ILLEGAL_ENUM_VALUE;

	bool SameChr = false;
	const unsigned BaseCount = SIZE(Bases);
	FORIX(Base)
		{
		const BaseWithSign &BWS = Bases[BaseIndex];
		Consec = ValidConsecutiveSeg(Seg, BWS, !Next);
		switch (Consec)
			{
		case CONSEC_VALID:
			return BaseIndex;
		case CONSEC_VALID_INVERT:
			return BaseIndex;

		case CONSEC_INVALID:
			break;			

		case CONSEC_NEW_CHR:
		case CONSEC_ILLEGAL_ENUM_VALUE:
		default:
			asserta(false);
			}

		if (BWS.Base.GenIndex == Seg.GenomeIndex && BWS.Base.ChrId == Seg.ChrId)
			SameChr = true;
		}

	if (SameChr)
		Consec = CONSEC_INVALID;
	else
		Consec = CONSEC_NEW_CHR;
	return UINT_MAX;
	}

/***
Using the current state of the growing output alignment as stored
in OutputAI, search for an output column that could be appended
(Next=true) or prepended (Next=false) to the alignment. If the column
must be inverted in order to be pre/appended, Invert is set to true,
otherwise to false.
***/
unsigned Stacker::GetAdjacentOutputCol(bool Next, bool &Invert)
	{
	Invert = false;
	const GCSegVec &Segs = OutputAI.Segs;
	const unsigned SegCount = SIZE(Segs);
	assert(SegCount > 0);

	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &Seg = Segs[SegIndex];
		GenChrPos GCP;
		GCP.GenIndex = Seg.GenomeIndex;
		GCP.ChrId = Seg.ChrId;
		if (Next)
			{
			if (Seg.Plus)
				GCP.Pos = Seg.hi + 1;
			else
				{
				if (Seg.lo == 0)
					return UINT_MAX;
				GCP.Pos = Seg.lo - 1;
				}
			}
		else
			{
			if (Seg.Plus)
				{
				if (Seg.lo == 0)
					return UINT_MAX;
				GCP.Pos = Seg.lo - 1;
				}
			else
				GCP.Pos = Seg.hi + 1;
			}
		std::map<GenChrPos, unsigned>::const_iterator p = GCPToColIndex.find(GCP);
		if (p == GCPToColIndex.end())
			continue;
		unsigned ColIndex = p->second;
		const std::vector<BaseWithSign> &Bases = OutputColBases[ColIndex];
		bool VC = ValidConsecutive(Segs, Bases, Next, Invert);
		if (VC)
			return ColIndex;
		}
	return UINT_MAX;
	}

/***
Given a set of segments and set of bases-with-sign, test whether
the bases can succeed (Next=true) or precede (Next=false) the
segments. If the column must be inverted, Invert is set to
true, otherwise to false.

We require that at least one segment must match a base. Otherwise
we will get alignments like this:

       XY
	XXXX----
	XXXX----
	----XXXX
	----XXXX
	   XY

Here, column Y could follow column X without causing a conflict,
but we prefer to split such alignments into two blocks.

Hence the AnySegments flag in the code below.
***/

bool Stacker::ValidConsecutive(const GCSegVec &Segs,
  const std::vector<BaseWithSign> &Bases, bool Next, bool &Invert)
	{
	Invert = false;
	const unsigned BaseCount = SIZE(Bases);
	if (BaseCount == 0)
		{
		return false;
		}
	const unsigned SegCount = SIZE(Segs);

	asserta(!OutputCols.empty());
	const StdStr &AdjCol = Next ? OutputCols.back() : OutputCols.front();
	asserta(SIZE(AdjCol) == SegCount);

	bool AnyCommonSegs = false;
	FORIX(Base)
		{
		const BaseWithSign &BWS = Bases[BaseIndex];
		CONSEC Consec = CONSEC_ILLEGAL_ENUM_VALUE;
		unsigned AdjSegIndex = GetAdjacentSegIndex(BWS, Segs, Consec, Next);
		switch (Consec)
			{
		case CONSEC_VALID:
			if (Invert)
				{
				return false;
				}
			if (AdjCol[AdjSegIndex] != '-')
				AnyCommonSegs = true;
			continue;

		case CONSEC_VALID_INVERT:
			if (BaseIndex == 0)
				Invert = true;
			else if (!Invert)
				{
				return false;
				}
			if (AdjCol[AdjSegIndex] != '-')
				AnyCommonSegs = true;
			continue;

		case CONSEC_INVALID:
			{
			return false;
			}

		case CONSEC_NEW_CHR:
			continue;

		case CONSEC_ILLEGAL_ENUM_VALUE:
		default:
			asserta(false);
			}
		}

	if (!AnyCommonSegs)
		return false;

	return true;
	}
