#include "stack.h"

#define	TRACE		0
#define BOTH_EDGES	0

/***
Stacker functions for building a graph in which nodes are bases
and an edge indicates that its bases are aligned to each other.

We are only interested in connected components, so we don't need
all edges. We can save time and space as follows: in each column
of an input alignment, create edges only between adjacent ungapped
positions. This gives O(N) instead of O(N^2) in the number of rows.

Example. The following column (shown horizontally) has 5 bases.
There would be (5 x 4)/2 = 10 edges if we created an edge for
every pair. However, we only need 4 to induce the same connected
component in the final graph:

			 ---		 ---
			|	|		|	|
			|	v		|	v

	X	-	X	X	-	X	X

	|		^	|		^
	|       |	|		|
	 -------	 --------

***/

const BaseInfo &Stacker::GetBase(unsigned BaseIndex) const
	{
	assert(BaseIndex < SIZE(Bases));
	return Bases[BaseIndex];
	}

unsigned Stacker::AddBase(const BaseInfo &Base)
	{
#if	TRACE
	Log("AddBase ");
	LogBaseInfo(Base);
	Log("\n");
#endif
	std::map<BaseInfo, unsigned>::const_iterator p = BaseInfoToIndex.find(Base);
	if (p != BaseInfoToIndex.end())
		{
		const BaseInfo &ExistingBase = p->first;
		asserta(toupper(Base.Letter) == toupper(ExistingBase.Letter));
#if	TRACE
		Log(" Existing=%u\n", p->second);
#endif
		return p->second;
		}
	unsigned BaseIndex = SIZE(Bases);
	Bases.push_back(Base);

	std::map<BaseInfo, unsigned>::value_type v(Base, BaseIndex);
	std::pair<std::map<BaseInfo, unsigned>::iterator, bool> Result =
	  BaseInfoToIndex.insert(v);
	asserta(Result.second);

	IntVec EmptyIntVec;
	BoolVec EmptyBoolVec;

	Edges.push_back(EmptyIntVec);
	EdgePluses.push_back(EmptyBoolVec);

#if	TRACE
	Log(" New=%u\n", BaseIndex);
#endif
	return BaseIndex;
	}

void Stacker::AddToBaseGraph(const RevFile &In, const AlnInfoX &AIX)
	{
	unsigned AlnIndex = In.GetAlnIndex(AIX.AlnInfoIndex);
	StdStrVec Rows;
	In.GetAlnRows(AlnIndex, Rows);

#if	TRACE
	Log("AddToBaseGraph\n");
	LogAIRows(AIX.AI, Rows);
#endif

	const AlnInfo &AI = AIX.AI;
	const GCSegVec &Segs = AI.Segs;
	const unsigned SegCount = SIZE(Segs);
	if (SegCount == 0)
		return;

	double ConfScore = AIX.AI.GetFloatAttr(ALN_ATTR_ConfScore);
	if (ConfScore == DBL_MAX)
		ConfScore = 0;

	unsigned BaseCount = 0;
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &Seg = Segs[SegIndex];
		BaseCount += Seg.GetLength();
		}
	double ConfScorePerBase = ConfScore/BaseCount;

	std::vector<BaseInfo> Bases;
	Bases.resize(SegCount);
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &Seg = Segs[SegIndex];
		BaseInfo &Base = Bases[SegIndex];
		Base.GenIndex = Seg.GenomeIndex;
		Base.ChrId = Seg.ChrId;
		bool Plus = Seg.Plus;
		Base.Pos = (Plus ? Seg.lo : Seg.hi);
		Base.ConfScore = ConfScorePerBase;
		}

	const StdStr &Row0 = Rows[0];
	const unsigned ColCount = SIZE(Row0);

	for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
		{
#if	TRACE
		{
		Log("\n");
		Log("Col%u of %u ", ColIndex, ColCount);
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = Segs[SegIndex];
			const BaseInfo &Base = Bases[SegIndex];
			char Letter = Rows[SegIndex][ColIndex];
			if (!Seg.Plus)
				Letter = CompLetter(Letter);
			Log(" %u,%u[%u]%c%c",
			  Base.GenIndex,
			  Base.ChrId,
			  (unsigned) Base.Pos,
			  Letter,
			  Seg.Plus ? '+' : '-');
			}
		Log("\n");
		}
#endif
		unsigned PrevNonGappedBaseIndex = UINT_MAX;
		unsigned PrevNonGappedSegIndex = UINT_MAX;
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = Segs[SegIndex];
			char Letter = Rows[SegIndex][ColIndex];
			if (Letter == '-')
				continue;

			BaseInfo &Base = Bases[SegIndex];
			if (!Seg.Plus)
				Letter = CompLetter(Letter);
			Base.Letter = Letter;
			unsigned BaseIndex = AddBase(Base);
#if	TRACE
			Log("AddBase ");
			LogBaseInfo(Base);
			Log(" BaseIndex=%u\n", BaseIndex);
#endif
			if (PrevNonGappedBaseIndex != UINT_MAX)
				{
				asserta(PrevNonGappedSegIndex != UINT_MAX);
				const GCSeg &PrevSeg = Segs[PrevNonGappedSegIndex];

				bool EdgePlus = (Seg.Plus == PrevSeg.Plus);

				Edges[BaseIndex].push_back(PrevNonGappedBaseIndex);
				EdgePluses[BaseIndex].push_back(EdgePlus);
#if	BOTH_EDGES
				Edges[PrevNonGappedBaseIndex].push_back(BaseIndex);
				EdgePluses[PrevNonGappedBaseIndex].push_back(EdgePlus);
#endif
				}
			PrevNonGappedBaseIndex = BaseIndex;
			PrevNonGappedSegIndex = SegIndex;
			if (Seg.Plus)
				++Base.Pos;
			else
				--Base.Pos;
			}
		}
	}

void Stacker::BuildBaseGraph()
	{
	const unsigned ONE_MILLION = 1000000;
	Bases.reserve(ONE_MILLION);
	Edges.reserve(ONE_MILLION);
	EdgePluses.reserve(ONE_MILLION);

	const AIXVec &AIXs = *ptrAIXs;
	const unsigned AIXCount = SIZE(AIXs);
	for (unsigned AIXIndex = 0; AIXIndex < AIXCount; ++AIXIndex)
		{
		const AlnInfoX &AIX = AIXs[AIXIndex];
		AddToBaseGraph(*ptrIn, AIX);
		}
#if	TRACE
	LogGraph();
#endif
	}
