#include "stack.h"

#define TRACE			0
#define	TRACE_OUTPUT	0

// Delete an output column.
// This is called each time a column is been added to
// the growing output alignment, to avoid it being
// found again later.
void Stacker::DeleteOutputCol(unsigned ColIndex)
	{
#if	TRACE
	Log("DeleteOutputCol(%u)\n", ColIndex);
#endif
	std::vector<BaseWithSign> &Col = OutputColBases[ColIndex];
	unsigned BaseCount = SIZE(Col);
	FORIX(Base)
		{
		const BaseWithSign &BWS = Col[BaseIndex];
		GenChrPos GCP;
		GCP.GenIndex = BWS.Base.GenIndex;
		GCP.ChrId = BWS.Base.ChrId;
		GCP.Pos = BWS.Base.Pos;
		std::map<GenChrPos, unsigned>::const_iterator p = GCPToColIndex.find(GCP);
		asserta(p != GCPToColIndex.end());
		GCPToColIndex.erase(GCP);
		}
	Col.clear();
	}

// Convert the completed output alignment, in which alignment
// rows are stored as lists (in order to allow O(1) inserts
// both at the front and back) into the StdStrVec format required
// to write an alignment to a RevFile.
void Stacker::BuildOutputRows(StdStrVec &Rows) const
	{
	const unsigned ColCount = SIZE(OutputCols);
	const unsigned SegCount = SIZE(OutputAI.Segs);
	asserta(SegCount > 0);
	asserta(ColCount > 0);

	Rows.resize(SegCount);
	for (std::list<StdStr>::const_iterator p =
	  OutputCols.begin(); p != OutputCols.end(); ++p)
		{
		const StdStr &Col = *p;
		asserta(SIZE(Col) == SegCount);
		FORIX(Seg)
			{
			StdStr &Row = Rows[SegIndex];
			Row.push_back(Col[SegIndex]);
			}
		}
	}

void Stacker::WriteOutputBlock()
	{
#if	TRACE
	Log("\n");
	Log("WriteOutputBlock:\n");
	LogOutputBlock();
#endif

	const unsigned ColCount = SIZE(OutputCols);
	const unsigned SegCount = SIZE(OutputAI.Segs);
	asserta(SegCount > 0);
	asserta(ColCount > 0);

	if (ColCount < g_MinBlockCols)
		{
#if	TRACE || TRACE_OUTPUT
		Log("--- Short block deleted:\n");
		StdStrVec OutputRows;
		BuildOutputRows(OutputRows);
		LogAIRows(OutputAI, OutputRows);
#endif
		ShortColCount += ColCount;
		return;
		}

	StdStrVec OutputRows;
	BuildOutputRows(OutputRows);

	unsigned Scale = SegCount - 1;
	double ConfScore = (Scale == 0 ? 0 : OutputConfScore/Scale);
	OutputAI.SetFloatAttr(ALN_ATTR_ConfScore, ConfScore);

	ptrOut->AddAln(OutputAI, OutputRows);
	++OutputBlockCount;
	OutputAIs.push_back(OutputAI);
#if	TRACE || TRACE_OUTPUT
	Log("Final output block:\n");
	LogAIRows(OutputAI, OutputRows);
#endif
	}

void Stacker::InitOutputBlock(const std::vector<BaseWithSign> &Bases)
	{
#if	TRACE
	Log("\n");
	Log("InitOutputBlock:");
	LogCol(Bases);
	Log("\n");
#endif
	OutputAI.Clear();
	OutputCols.clear();
	OutputConfScore = 0;

	const unsigned BaseCount = SIZE(Bases);
	asserta(BaseCount > 0);
	StdStr Col;
	for (unsigned BaseIndex = 0; BaseIndex < BaseCount; ++BaseIndex)
		{
		const BaseWithSign &BWS = Bases[BaseIndex];

		GCSeg Seg;
		Seg.GenomeIndex = BWS.Base.GenIndex;
		Seg.ChrId = BWS.Base.ChrId;
		Seg.lo = BWS.Base.Pos;
		Seg.hi = BWS.Base.Pos;
		Seg.Plus = BWS.Plus;

		OutputAI.Segs.push_back(Seg);

		char Letter = BWS.GetLetter();
		Col.push_back(Letter);
		OutputConfScore += BWS.Base.ConfScore;
		}
	OutputCols.push_back(Col);

#if	TRACE
	Log("\n");
	Log("After InitOutputBlock, output block:\n");
	LogOutputBlock();
#endif
	}

static void IncSegCoord(GCSeg &Seg, bool Next)
	{
// Next means adding to right-hand end of alignment.
	if (Next)
		{
		if (Seg.Plus)
		// right, + seg
			++Seg.hi;
		else
			{
		// right, - seg
			asserta(Seg.lo > 0);
			--Seg.lo;
			}
		}
	else
		{
		if (Seg.Plus)
			{
		// left, + seg
			asserta(Seg.lo > 0);
			--Seg.lo;
			}
		else
		// left, - seg
			++Seg.hi;
		}
	}

void Stacker::AddOutputCol(unsigned ColIndex, bool Invert, bool Next)
	{
#if	TRACE
	Log("AddOutputCol(ColIndex=%u, Invert=%c, Next=%c)\n",
	  ColIndex,
	  Invert ? 'T' : 'F',
	  Next ? 'T' : 'F');
	LogCol(OutputColBases[ColIndex]);
	Log("\n");
#endif
	const std::vector<BaseWithSign> &Bases = OutputColBases[ColIndex];
	const unsigned BaseCount = SIZE(Bases);

	GCSegVec &Segs = OutputAI.Segs;
	const unsigned SegCount = SIZE(Segs);

	StdStr NewCol;
	NewCol.resize(SegCount, '-');

	GCSegVec NewSegs = Segs;
	StdStr NewLetters;
	FORIX(Base)
		{
		const BaseWithSign &BWS = Bases[BaseIndex];

		CONSEC Consec = CONSEC_ILLEGAL_ENUM_VALUE;
		unsigned SegIndex = GetAdjacentSegIndex(BWS, Segs, Consec, Next);

		switch (Consec)
			{
		case CONSEC_VALID:
			{
			asserta(!Invert);
			GCSeg &Seg = NewSegs[SegIndex];
			IncSegCoord(Seg, Next);

			char Letter = BWS.GetLetter();
			NewCol[SegIndex] = Letter;
			break;
			}

		case CONSEC_VALID_INVERT:
			{
			asserta(Invert);
			GCSeg &Seg = NewSegs[SegIndex];
			IncSegCoord(Seg, Next);

			char Letter = BWS.GetLetter();
			Letter = CompLetter(Letter);
			NewCol[SegIndex] = Letter;
			break;
			}

		case CONSEC_INVALID:
			asserta(false);

		case CONSEC_NEW_CHR:
			{
			GCSeg NewSeg;
			NewSeg.GenomeIndex = BWS.Base.GenIndex;
			NewSeg.ChrId = BWS.Base.ChrId;
			NewSeg.lo = BWS.Base.Pos;
			NewSeg.hi = BWS.Base.Pos;
			NewSeg.Plus = BWS.Plus;

			NewSegs.push_back(NewSeg);

			char Letter = BWS.GetLetter();
			NewCol.push_back(Letter);

			for (std::list<StdStr>::iterator p = OutputCols.begin();
			  p != OutputCols.end(); ++p)
				{
				StdStr &Col = *p;
				Col.push_back('-');
				}
			continue;
			}

		case CONSEC_ILLEGAL_ENUM_VALUE:
		default:
			asserta(false);
			}

		char Letter = BWS.GetLetter();
		if (Invert)
			Letter = CompLetter(Letter);
		NewCol[SegIndex] = Letter;
		}

	Segs = NewSegs;

	if (Next)
		OutputCols.push_back(NewCol);
	else
		OutputCols.push_front(NewCol);

	DeleteOutputCol(ColIndex);

#if	TRACE
	Log("After AddOutputCol:\n");
	LogOutputBlock();
#endif
	}
