#include "stack.h"

#define TRACE	0

RevFile *g_ptrDel = 0;
unsigned g_MaxAlnCount = 256;
unsigned g_MaxBlockRows = 256;
unsigned g_MinBlockCols = 8;
unsigned g_InconsistentRatio = 3;

unsigned g_SaveCCIndex = UINT_MAX;
unsigned g_SaveCCGenIndex = UINT_MAX;
unsigned g_SaveCCChrId = UINT_MAX;
coord_t g_SaveCCPos = COORD_T_MAX;

unsigned g_DeletedBlockCount = 0;
bool g_LogInconsistentAlns = false;

static void CopyToDelFile(const RevFile &In, unsigned CCIndex, const AIXVec &AIXs)
	{
	const unsigned AlnCount = SIZE(AIXs);
	g_DeletedBlockCount += AlnCount;
	if (g_ptrDel == 0)
		{
		Progress("CC %u has %u alns, skipping (no -del file specified)",
		  CCIndex, AlnCount);
		return;
		}

	Progress("CC %u has %u alns, copying to -del file", CCIndex, AlnCount);
	for (unsigned i = 0; i < AlnCount; ++i)
		{
		const AlnInfoX &AIX = AIXs[i];
		g_ptrDel->CopyAln(In, AIX.AlnInfoIndex);
		}
	}

void StackMulti()
	{
	const char *InFileName = RequiredValueOpt("multi");
	const char *OutFileName = RequiredValueOpt("out");
	const char *DelFileName = ValueOpt("del");

	IntOpt("maxalns", &g_MaxAlnCount);
	IntOpt("maxblockrows", &g_MaxBlockRows);
	IntOpt("minblockcols", &g_MinBlockCols);
	IntOpt("iratio", &g_InconsistentRatio);

	IntOpt("savecc", &g_SaveCCIndex);
	IntOpt("saveccgen", &g_SaveCCGenIndex);
	if (g_SaveCCGenIndex != UINT_MAX)
		{
		g_SaveCCChrId = RequiredIntOpt("saveccchr");
		g_SaveCCPos = (coord_t) RequiredIntOpt("saveccpos");
		}

	bool ShowCC = FlagOpt("showcc");
	g_LogInconsistentAlns = FlagOpt("loginc");

	bool Chunk = (ValueOpt("chunk") != 0 || ValueOpt("of") != 0);
	unsigned ChunkIndex = UINT_MAX;
	unsigned ChunkCount = UINT_MAX;
	if (Chunk)
		{
		ChunkIndex = RequiredIntOpt("chunk");
		ChunkCount = RequiredIntOpt("of");
		if (ChunkCount == 0)
			Die("-of must be > 0");
		if (ChunkIndex == 0 || ChunkIndex > ChunkCount)
			Die("-chunk must be in range 1 .. -of");

	// Convert to zero-based
		--ChunkIndex;
		}

	RevFile In;
	RevFile Out;

	In.Open(InFileName);
	const unsigned GenCount = In.GetGenomeCount();
	const std::vector<GenInfo> &Genomes = In.GetGenomes();
	Out.Create(OutFileName, Genomes);
	if (DelFileName != 0)
		{
		g_ptrDel = new RevFile;
		g_ptrDel->Create(DelFileName, Genomes);
		}

	const unsigned GlobalAICount = In.GetAlnInfoCount();

// Create overlap graph
	IntVecVec GlobalEdges;
	GlobalEdges.resize(GlobalAICount);

	unsigned TotalChrCount = 0;
	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		TotalChrCount += In.GetChrCount(GenIndex);

// For all genomes...
	unsigned TotalChrIndex = 0;
	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		{
		const GenInfo &GI = In.GetGenomeInfo(GenIndex);

	// For all chromosomes...
		const unsigned ChrCount = In.GetChrCount(GenIndex);
		for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
			{
			const ChrInfo &CI = GI.GetChrInfoByIndex(ChrIndex);
			unsigned ChrId = CI.GetId();
			const std::string &ChrName = CI.GetName();
			double Pct = (++TotalChrIndex)*100.0/TotalChrCount;
			Progress("%.16s[%3u] %u of %u, %.1f%%",
			  ChrName.c_str(), ChrId, TotalChrIndex, TotalChrCount, Pct);

			IntVec AIIndexes;
			In.GetAppendAlnInfoIndexesChr(GenIndex, ChrId, AIIndexes);

			AIXVec ChrAIXs;
			In.GetAlnInfoXs(AIIndexes, ChrAIXs);

			IntVecVec ChrEdges;
			MakeOverlapGraph(ChrAIXs, GenIndex, ChrId, ChrEdges);

		// Append edges for this chromosome to global set of edges
			const unsigned ChrAIXCount = (unsigned) ChrAIXs.size();
			for (unsigned i = 0; i < ChrAIXCount; ++i)
				{
				const AlnInfoX &AIX = ChrAIXs[i];
				const IntVec &Neighbors = ChrEdges[i];

				unsigned AIIndex = AIX.AlnInfoIndex;
				assert(AIIndex < GlobalAICount);

				IntVec &GlobalNeighbors = GlobalEdges[AIIndex];

				unsigned NeighborCount = (unsigned) Neighbors.size();
				for (unsigned j = 0; j < NeighborCount; ++j)
					{
					unsigned k = Neighbors[j];
					assert(k < ChrAIXCount);

					const AlnInfoX &NeighborAIX = ChrAIXs[k];
					unsigned NeighborAIIndex = NeighborAIX.AlnInfoIndex;
					assert(NeighborAIIndex < GlobalAICount);

					GlobalNeighbors.push_back(NeighborAIIndex);
					}
				}
			}
		}

	Progress("Compute connected components");
	IntVecVec CCs;
	GetConnectedComponents(GlobalEdges, CCs, true);
	const unsigned CCCount = (unsigned) CCs.size();

// Find number of CCs with > 1 alignment, purely for reporting purposes
#if	TRACE
	Log("\n");
	Log("ConnComp  Alignments\n");
	Log("========  ==========\n");
#endif
	unsigned GT1Count = 0;
	for (unsigned CCIndex = 0; CCIndex < CCCount; ++CCIndex)
		{
		const IntVec &CC = CCs[CCIndex];
		const unsigned CCSize = (unsigned) CC.size();
		if (CCSize > 1)
			++GT1Count;
#if	TRACE
		if (CCSize > 1)
			Log("%8u  %10u\n", CCIndex, CCSize);
#endif
		}
	Progress("%u CCs, %u with > 1 alignment", CCCount, GT1Count);

	if (g_SaveCCGenIndex != UINT_MAX)
		{
		for (unsigned CCIndex = 0; CCIndex < CCCount; ++CCIndex)
			{
			const IntVec &CC = CCs[CCIndex];
			const unsigned CCSize = (unsigned) CC.size();

			AIXVec AIXs;
			AIXs.reserve(CCSize);
			for (unsigned i = 0; i < CCSize; ++i)
				{
				unsigned AIIndex = CC[i];
				AlnInfoX AIX;
				In.GetAlnInfoX(AIIndex, AIX);
				const GCSegVec &Segs = AIX.AI.Segs;
				const unsigned SegCount = SIZE(Segs);
				FORIX(Seg)
					{
					const GCSeg &Seg = Segs[SegIndex];
					if (Seg.GenomeIndex == g_SaveCCGenIndex &&
					  Seg.ChrId == g_SaveCCChrId &&
					  g_SaveCCPos >= Seg.lo && g_SaveCCPos <= Seg.hi)
						{
						Progress("Saving CC %u", CCIndex);
						for (unsigned j = 0; j < CCSize; ++j)
							{
							unsigned AIIndex = CC[j];
							Out.CopyAln(In, AIIndex);
							}
						Out.Close();
						return;
						}
					}
				}
			}
		Die("Gen %u Chr %u Pos %u not found",
		  g_SaveCCGenIndex, g_SaveCCChrId, g_SaveCCPos);
		}

// CCs[i] is a vector of integers. Each integer in such
// a vector is a AIIndex for In.
	if (g_SaveCCIndex != UINT_MAX)
		{
		if (g_SaveCCIndex >= CCCount)
			Die("-savecc %u, there are only %u CCs", g_SaveCCIndex, CCCount);

		const IntVec &CC = CCs[g_SaveCCIndex];
		const unsigned CCSize = (unsigned) CC.size();

		AIXVec AIXs;
		AIXs.reserve(CCSize);
		for (unsigned i = 0; i < CCSize; ++i)
			{
			unsigned AIIndex = CC[i];
			Out.CopyAln(In, AIIndex);
			}
		Progress("CC %u saved", g_SaveCCIndex);
		Out.Close();
		return;
		}

	ProgressStart("Processing CCs");
	unsigned TotalDeletedBigColCount = 0;
	unsigned TotalInconsistentColCount = 0;
	unsigned TotalShortColCount = 0;
	unsigned InputAlnCount = 0;
	unsigned TotalOutputBlockCount = 0;
	unsigned FromCCIndex = 0;
	unsigned EndCCIndex = CCCount;
	if (Chunk && CCCount > 0)
		{
		unsigned CCsPerChunk = (CCCount + ChunkCount - 1)/ChunkCount;
		FromCCIndex = ChunkIndex*CCsPerChunk;
		EndCCIndex = FromCCIndex + CCsPerChunk;
		if (EndCCIndex > CCCount)
			EndCCIndex = CCCount;
		}

	for (unsigned CCIndex = FromCCIndex; CCIndex < EndCCIndex; ++CCIndex)
		{
		if (CCIndex%10 == 0 && !ShowCC)
			ProgressStep(CCIndex, CCCount);

		const IntVec &CC = CCs[CCIndex];
		const unsigned CCSize = (unsigned) CC.size();

		if (ShowCC)
			{
			double Pct = (double) CCIndex * 100.0 / CCCount;
			Progress("CC %u of %u %.1f%%", CCIndex, CCCount, Pct);
			}

		AIXVec AIXs;
		AIXs.reserve(CCSize);
		for (unsigned i = 0; i < CCSize; ++i)
			{
			unsigned AIIndex = CC[i];
			assert(AIIndex < GlobalAICount);

			AlnInfoX AIX;
			In.GetAlnInfoX(AIIndex, AIX);

			AIXs.push_back(AIX);
			}

		unsigned AlnCount = SIZE(AIXs);
		asserta(AlnCount == CCSize);
		InputAlnCount += AlnCount;
		if (AlnCount > g_MaxAlnCount && g_MaxAlnCount != 0)
			{
			CopyToDelFile(In, CCIndex, AIXs);
			continue;
			}

		Stacker T(In, AIXs, CCIndex, Out);

		if (ShowCC)
			ProgressNoRes(
			  "CC %u, %u AIs %u outblocks cols: %u short %u del %u incon",
			  CCIndex,
			  CCSize,
			  T.OutputBlockCount,
			  T.ShortColCount,
			  T.DeletedBigColCount,
			  T.InconsistentColCount);
		else
			{
			if (T.DeletedBigColCount > 0 || T.InconsistentColCount > 0)
				ProgressNoRes("CC %u, %u cols deleted, %u inconsistent",
				  CCIndex,
				  T.DeletedBigColCount,
				  T.InconsistentColCount);
			}

		TotalOutputBlockCount += T.OutputBlockCount;
		TotalDeletedBigColCount += T.DeletedBigColCount;
		TotalShortColCount += T.ShortColCount;
		TotalInconsistentColCount += T.InconsistentColCount;
		}
	ProgressDone();
	ProgressNoRes("");
	ProgressNoRes("Input alns     %8u", InputAlnCount);
	ProgressNoRes("Output alns    %8u", TotalOutputBlockCount);
	ProgressNoRes("Deleted blocks %8u  (too many overlaps)", g_DeletedBlockCount);
	ProgressNoRes("Deleted cols   %8u  (col too big)", TotalDeletedBigColCount);
	ProgressNoRes("Deleted cols   %8u  (block too small)", TotalShortColCount);
	ProgressNoRes("Incons. cols   %8u", TotalInconsistentColCount);
	ProgressNoRes("");
	Progress("Closing files");
	In.Close();
	Out.Close();
	if (g_ptrDel != 0)
		{
		g_ptrDel->Close();
		delete g_ptrDel;
		}
	}
