#include "stack.h"

#define TRACE	0

/***
Stack by selecting the alignment with the best confidence
score for each base in the genome.

The algorithm scans from left to right through the genome,
keeping track of the highest scoring alignment (H) at each
point, and the point P at which H became the winner. At each
point Q where a new highest-scoring alignment is discovered,
the range P to (Q-1) in H is output and a new H is assigned.

Key variables
-------------

	List L of alignments that overlap the current point
	in the scan.

	H is the highest-scoring alignment at the current
	position in the scan.

	Position P which is defined as the position in H
	where it most recently became the highest-scoring
	alignment. P is either the start of H or one base
	following the hi bendpoint of a higher-scoring
	alignment that overlaps H.

Algorithm
---------

For each chromosome
	L = empty
	H = empty
	For each bendpoint B in increasing coordinate order:
		Let A = alignment(B)
		If B is lo:
			If confidence(A) > confidence(H):
				Output block in H from P to B - 1
				H = A
				P = B
			Add A to L

		If B is hi:
			Delete A from L
			If H is A:
				Output block in H from P to B
				Delete all alignments in L with hi <= B
				H = alignment in L with highest confidence score
				P = B + 1
***/

const double LARGE_NEGATIVE = -9e99;

unsigned g_GenIndex;

// List L, sorted in increasing confidence score order
static std::multimap<double, unsigned> g_ConfScoreToBPIndex;

static const GCSeg &GetSeg(unsigned BPIndex, const AIBPVec &BPs, const AIVec &AIs)
	{
	const AIBP &BP = BPs[BPIndex];
	unsigned AIIndex = BP.AlnInfoIndex;
	unsigned SegIndex = BP.SegIndex;
	const AlnInfo &AI = AIs[AIIndex];
	return AI.Segs[SegIndex];
	}

void LogAlnList(const AIBPVec &BPs, const AIVec &AIs)
	{
	for (std::multimap<double, unsigned>::const_iterator p = g_ConfScoreToBPIndex.begin();
	  p != g_ConfScoreToBPIndex.end(); ++p)
		{
		const std::pair<double, unsigned> &v = *p;
		unsigned BPIndex = v.second;
		const AIBP &BP = BPs[BPIndex];
		Log(" Conf=%g, BP=%u %s AI=%u Seg[%u]=",
		  v.first,
		  BPIndex,
		  BP.IsLo ? "lo" : "hi",
		  BP.AlnInfoIndex,
		  BP.SegIndex);
		const GCSeg &Seg = GetSeg(BPIndex, BPs, AIs);
		LogGCSeg(Seg);
		Log("\n");
		}
	Log("\n");
	}

void StackTopOneGenome(const char *InFileName, const char *OutFileName, unsigned GenIndex)
	{
	g_GenIndex = GenIndex;

	RevFile In;
	RevFile Out;

	In.Open(InFileName);
	const std::vector<GenInfo> &Genomes = In.GetGenomes();
	Out.Create(OutFileName, Genomes);

	AIVec AIs;
	In.GetAllAlnInfos(AIs);

	Progress("Get bendpoints");
	AIBPVec BPs;
	GetBendpoints(AIs, g_GenIndex, BPs, 0);

	const unsigned BPCount = SIZE(BPs);
	unsigned PrevChrId = UINT_MAX;
	coord_t StartPos = COORD_T_MAX;
	unsigned BestBPIndex = UINT_MAX;
	double BestConfScore = LARGE_NEGATIVE;
	ProgressStart("Stacking");
	for (unsigned BPIndex = 0; BPIndex < BPCount; ++BPIndex)
		{
		if (BPIndex%100 == 0)
			ProgressStep(BPIndex, BPCount);

		const AIBP &BP = BPs[BPIndex];
		if (BP.ChrId != PrevChrId)
			{
			g_ConfScoreToBPIndex.clear();
			PrevChrId = BP.ChrId;
			BestConfScore = LARGE_NEGATIVE;
			BestBPIndex = UINT_MAX;
			StartPos = COORD_T_MAX;
			}

		unsigned AIIndex = BP.AlnInfoIndex;
		const AlnInfo &AI = AIs[AIIndex];
		if (!AI.HasAttr(ALN_ATTR_ConfScore))
			Die("AI %u does not have confidence score", AIIndex);
		double ConfScore = AI.GetFloatAttr(ALN_ATTR_ConfScore);

#if	TRACE
		{
		Log("\n");
		Log("BP=%u %s AI=%u Chr=%u Pos=%u ConfScore=%g\n",
		  BPIndex,
		  BP.IsLo ? "lo" : "hi",
		  BP.AlnInfoIndex,
		  BP.ChrId,
		  (unsigned) BP.Pos,
		  ConfScore);
		Log("StartPos=");
		if (StartPos == COORD_T_MAX)
			Log("*");
		else
			Log("%u", (unsigned) StartPos);
		Log(" BestBP=");
		if (BestBPIndex == UINT_MAX)
			Log("*");
		else
			Log("%u", BestBPIndex);
		Log(" BestConfScore=");
		if (BestConfScore == LARGE_NEGATIVE)
			Log("*");
		else
			Log("%g", BestConfScore);
		Log("\n");
		if (BestBPIndex != UINT_MAX)
			{
			const GCSeg &BestSeg = GetSeg(BestBPIndex, BPs, AIs);
			Log("BestSeg=");
			LogGCSeg(BestSeg);
			Log("\n");
			}
		Log("List:\n");
		LogAlnList(BPs, AIs);
		}
#endif

		if (BP.IsLo)
			{
			if (ConfScore > BestConfScore)
				{
				if (BestBPIndex != UINT_MAX && BP.Pos > 0)
					OutputBlockTop(In, BestBPIndex, AIs, BPs, StartPos, BP.Pos - 1, Out);
				BestBPIndex = BPIndex;
				if (StartPos == COORD_T_MAX || BP.Pos >= StartPos)
					StartPos = BP.Pos;
				BestConfScore = ConfScore;
				}
			std::pair<double, unsigned> ConfScoreAIIndex(ConfScore, BPIndex);
			g_ConfScoreToBPIndex.insert(ConfScoreAIIndex);
			}
		else
			{
			assert(!BP.IsLo);

		// Delete segment from list.
		// There may be more than one entry with the same confidence score.
		// It may alrady have been deleted.
			for (std::multimap<double, unsigned>::iterator p =
			  g_ConfScoreToBPIndex.find(ConfScore);
			  p != g_ConfScoreToBPIndex.end() && p->first == ConfScore; ++p)
				{
				unsigned ListBPIndex = p->second;
				const AIBP &ListBP = BPs[ListBPIndex];
				if (ListBP.AlnInfoIndex == BP.AlnInfoIndex &&
				  ListBP.SegIndex == BP.SegIndex)
					{
					g_ConfScoreToBPIndex.erase(p);
					break;
					}
				}

			if (BestBPIndex != UINT_MAX)
				{
				const AIBP &BestBP = BPs[BestBPIndex];
				if (BP.AlnInfoIndex == BestBP.AlnInfoIndex && BP.SegIndex == BestBP.SegIndex)
					{
					OutputBlockTop(In, BestBPIndex, AIs, BPs, StartPos, BP.Pos, Out);
					StartPos = BP.Pos + 1;
#if	TRACE
					Log("List before delete, BP.Pos=%u StartPos=%u:\n",
					  (unsigned) BP.Pos, (unsigned) StartPos);
					LogAlnList(BPs, AIs);
#endif
				// Delete all alignments with hi <= this bendpoint
				// Care needed with this loop as it dynamically deletes
					for (std::multimap<double, unsigned>::iterator p =
					  g_ConfScoreToBPIndex.begin(); p != g_ConfScoreToBPIndex.end(); )
						{
						const std::pair<double, unsigned> &v = *p;
						unsigned BPIndex = v.second;
						const GCSeg &Seg = GetSeg(BPIndex, BPs, AIs);
#if	TRACE
						Log("BP %u, hi=%u ", BPIndex, (unsigned) Seg.hi);
#endif
						if (Seg.hi < StartPos)
							{
#if	TRACE
							Log("erase\n");
#endif
							std::multimap<double, unsigned>::iterator Saved_p = p;
							++p;
							g_ConfScoreToBPIndex.erase(Saved_p);
							}
						else
							{
							++p;
#if	TRACE
							Log("keep\n");
#endif
							}
						}
#if	TRACE
					Log("List after delete, BP.Pos=%u:\n", (unsigned) BP.Pos);
					LogAlnList(BPs, AIs);
#endif
					if (g_ConfScoreToBPIndex.size() > 0)
						{
						std::multimap<double, unsigned>::iterator ptrEnd = g_ConfScoreToBPIndex.end();
						std::multimap<double, unsigned>::iterator ptrLast = --ptrEnd;
						const std::pair<double, unsigned> &BestConfScoreBPIndex = *ptrLast;
						BestConfScore = BestConfScoreBPIndex.first;
						BestBPIndex = BestConfScoreBPIndex.second;
						const GCSeg &BestSeg = GetSeg(BestBPIndex, BPs, AIs);
#if	TRACE
						Log("New top seg: BP=%u ", BestBPIndex);
						LogGCSeg(BestSeg);
						Log("\n");
#endif
						if (BestSeg.hi == BP.Pos)
							{
							BestConfScore = LARGE_NEGATIVE;
							BestBPIndex = UINT_MAX;
							StartPos = COORD_T_MAX;
							}
						if (StartPos != COORD_T_MAX &&
						  (StartPos < BestSeg.lo || StartPos > BestSeg.hi))
							Die("New top seg out of range StartPos=%u Seg=%u - %u",
							  (unsigned) StartPos,
							  (unsigned) BestSeg.lo,
							  (unsigned) BestSeg.hi);
						}
					else
						{
#if	TRACE
						Log("List empty, no best seg\n");
#endif
						BestConfScore = LARGE_NEGATIVE;
						BestBPIndex = UINT_MAX;
						}
					}
				}
			else
				{
				BestConfScore = LARGE_NEGATIVE;
				BestBPIndex = UINT_MAX;
				}
			}
		}
	ProgressDone();
	asserta(g_ConfScoreToBPIndex.empty());

// Save memory
	AIs.clear();

	AIXVec AIXs;
	In.GetAllAlnInfoXs(AIXs);
	const unsigned AIXCount = SIZE(AIXs);
	unsigned LackingCount = 0;
	ProgressStart("Copying alignments lacking genome %u", GenIndex);
	for (unsigned AIIndex = 0; AIIndex < AIXCount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AIXCount);
		const AlnInfoX &AIX = AIXs[AIIndex];
		const GCSeg *ptrSeg = AIX.AI.GetSegPtr(GenIndex);
		if (ptrSeg == 0)
			{
			++LackingCount;
			Out.CopyAln(In, AIX.AlnInfoIndex);
			}
		}
	ProgressDone();
	Progress("%u alignments lacking genome %u", LackingCount, GenIndex);

	Progress("Closing files and writing output");
	In.Close();
	Out.Close();
	}
	
void StackTop()
	{
	const char *InFileName = RequiredValueOpt("top");
	const char *OutFileName = RequiredValueOpt("out");

	if (ValueOpt("chunk"))
		Die("-chunk not supported with -top");

	if (ValueOpt("genindex"))
		{
		unsigned GenIndex = RequiredIntOpt("genindex");
		StackTopOneGenome(InFileName, OutFileName, GenIndex);
		return;
		}

	RevFile In;
	In.Open(InFileName);
	const unsigned GenCount = In.GetGenomeCount();

	std::string TmpFileName;
	MakeTmpFileName("stack", TmpFileName);

	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		{
		Progress("Stacking genome %u of %u %.16s",
		  GenIndex + 1,
		  GenCount,
		  In.GetGenomeName(GenIndex).c_str());

		const char *StepInFileName;
		if (GenIndex == 0)
			StepInFileName = InFileName;
		else
			StepInFileName = TmpFileName.c_str();

		StackTopOneGenome(StepInFileName, OutFileName, GenIndex);

		if (GenIndex == GenCount - 1)
			DeleteStdioFile(TmpFileName.c_str());
		else
			RenameStdioFile(OutFileName, TmpFileName.c_str());
		}
	}
