#include "evo.h"

void Unwind()
	{
	const char *InFileName = RequiredValueOpt("unwind");
	const char *OutFileName = RequiredValueOpt("out");

	FILE *fIn = OpenStdioFile(InFileName);
	FILE *fOut = CreateStdioFile(OutFileName);

	EvoFileHeader InHdr;
	vector<MULIndexEntryData> InIEs;
	GenInfo Genome;
	ReadEvoIndex(fIn, InHdr, InIEs, Genome);

	const unsigned ChrCount = Genome.GetChrCount();

	IntSet ChrIndexes;
	unsigned MULIndex = InHdr.FirstMULIndex;
	while (MULIndex != UINT_MAX)
		{
		asserta(MULIndex < SIZE(InIEs));
		const MULIndexEntryData &IE = InIEs[MULIndex];
		unsigned ChrIndex = IE.MI.ChrIndex;
		if (ChrIndex != UINT_MAX)
			ChrIndexes.insert(ChrIndex);
		MULIndex = IE.Next;
		}

	if (SIZE(ChrIndexes) != ChrCount)
		Die("Chr count mismatch");

// Validate chr indexes are consecutive
	unsigned PrevChrIndex = UINT_MAX;
	for (IntSet::const_iterator p = ChrIndexes.begin(); p != ChrIndexes.end(); ++p)
		{
		unsigned ChrIndex = *p;
		if (ChrIndex == UINT_MAX)
			continue;
		if (PrevChrIndex != UINT_MAX && PrevChrIndex + 1 != ChrIndex)
			Die("Chr indexes not consecutive %u %u", PrevChrIndex, ChrIndex);
		PrevChrIndex = ChrIndex;
		}

// One MUL per chromosome when unwound.
	unsigned UnwoundMULCount = ChrCount;
	InitEvoFile(fOut, UnwoundMULCount, Genome);

	BitFile BFIn;
	BitFile BFOut;

	BFIn.OpenReading(fIn);
	BFOut.OpenWriting(fOut);

	MULIndex = InHdr.FirstMULIndex;
	if (MULIndex >= InHdr.LeafMULCount)
		{
		DumpEvoHdr(fIn);
		Die("Invalid evo header, FirstMUL=%u >= LeafMULCount=%u",
		  MULIndex, InHdr.LeafMULCount);
		}

	unsigned OutputMULCount = 0;
	gpos_t OutputBaseCount = 0;
	unsigned ChrIndex = UINT_MAX;

	vector<MULIndexEntryData> OutIEs;
	OutIEs.resize(ChrCount);

	MULIndexEntryData *OutIE = 0;

	for (;;)
		{
		++OutputMULCount;
		double Pct = OutputMULCount*100.0/InHdr.LeafMULCount;
		Progress("MUL %u of %u %.1f%%", OutputMULCount, InHdr.LeafMULCount, Pct);
		MULIndexEntryData InIE;
		ReadIE(fIn, MULIndex, InIE, true);
		Log("MUL %u IE=", MULIndex);
		InIE.LogMe();
		Log("\n");

		if (InIE.MI.ChrIndex != ChrIndex)
			{
			if (InIE.MI.ChrIndex != UINT_MAX && ChrIndex + 1 != InIE.MI.ChrIndex)
				Die("Chr indexes not consecutive");

			ChrIndex = InIE.MI.ChrIndex;
			OutIE = &(OutIEs[ChrIndex]);

			OutIE = &(OutIEs[ChrIndex]);
			OutIE->Magic = MULIEMagic;
			OutIE->FileBytePos = BFOut.GetBytePos();
			OutIE->FileBitPos = BFOut.GetBitPos();
			OutIE->InMULIndex = ChrIndex;
			OutIE->Leaf = true;
			OutIE->Next = ChrIndex + 1;

			OutIE->MI.AncestralLength = 0;
			OutIE->MI.EvolvedLength = 0;
			OutIE->MI.ChrIndex = ChrIndex;
			OutIE->MI.MinPos = OutputBaseCount;
			OutIE->MI.MaxIndex = InIE.MI.MaxIndex;
			}

		OutIE->MI.AncestralLength += InIE.MI.AncestralLength;
		OutIE->MI.EvolvedLength += InIE.MI.EvolvedLength;
	
		MicroEvolver Micro;
		Micro.FromBitFile(BFIn, InIE.FileBytePos, InIE.FileBitPos, InIE.MI);
		OutputBaseCount += Micro.EvolvedToFile(BFOut, InIE.MI.Plus);

		if (InIE.Next == UINT_MAX)
			MicroEvolver::EndOfFileToFile(BFOut);

		OutIE->MI.MaxIndex = OutputBaseCount;

		if (InIE.Next == UINT_MAX)
			break;
		if (InIE.Next >= InHdr.LeafMULCount)
			{
			DumpEvoHdr(fIn);
			Die("Invalid evo link, Next[%u]=%u > MULCount=%u",
			  MULIndex, InIE.Next, InHdr.LeafMULCount);
			}
		MULIndex = InIE.Next;
		}

	OutIEs[ChrCount-1].Next = UINT_MAX;

	Progress("Total %u MULs, %"PF" bases", OutputMULCount, OutputBaseCount);

// Write final header and index entry
	EvoFileHeader EvoHdr;
	EvoHdr.StartMagic = EvoStartHdrMagic;
	EvoHdr.TotalMULCount = ChrCount;
	EvoHdr.LeafMULCount = ChrCount;
	EvoHdr.FirstMULIndex = 0;
	EvoHdr.TotalBaseCount = OutputBaseCount;
	EvoHdr.LeafBaseCount = OutputBaseCount;
	EvoHdr.EndMagic = EvoEndHdrMagic;

	WriteEvoIndex(fOut, EvoHdr, OutIEs, Genome);

	BFIn.Close();
	BFOut.Close();
	CloseStdioFile(fIn);
	CloseStdioFile(fOut);
	}
