#include <map>
#include <math.h>
#include "evolver.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Letters);
bool IsStartCodon(const string &Letters);

extern unsigned g_InsAccConstrained;
extern unsigned g_InsRejConstrained;
extern unsigned g_InsAccNeutral;
extern unsigned g_DelAccConstrained;
extern unsigned g_DelRejConstrained;
extern unsigned g_DelAccNeutral;
extern unsigned g_DelAccCDS;

bool Evolver::AcceptRange(gpos_t Start, gpos_t Length, bool *ptrAnyConstraint) const
	{
	gpos_t Pos = Start;
	if (ptrAnyConstraint != 0)
		*ptrAnyConstraint = false;
	for (gpos_t i = 0; i < Length; ++i)
		{
		double Prob = GetAcceptProb(Pos);
		if (Prob < 1.0 && ptrAnyConstraint != 0)
			*ptrAnyConstraint = true;
		bool Accept = RandTrue(Prob);
		if (!Accept)
			return false;

		Pos = GetNextPos(Pos);
		}
	return true;
	}

double Evolver::GetRangeAcceptProb(gpos_t Start, gpos_t Length) const
	{
	gpos_t Pos = Start;
	double RangeProb = 1.0;
	for (gpos_t i = 0; i < Length; ++i)
		{
		double Prob = GetAcceptProb(Pos);
		RangeProb *= Prob;

	// Optimization: if falls below minimum threshold,
	// zero is reasonable approximation.
		if (RangeProb < MIN_ACCEPT_PROB)
			return 0.0;

		Pos = GetNextPos(Pos);
		}
	return RangeProb;
	}

double Evolver::GetAcceptProb(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.GetAcceptProb();
	}

bool Evolver::AcceptInsertLow(gpos_t Pos, const string &Seq, int *ptrWhy,
  bool *ptrAnyConstraint) const
	{
	extern unsigned g_InsertRejectAccProb;
	extern unsigned g_InsertRejectInCDSMod3Count;
	extern unsigned g_InsertCreatesStopCodonCount;
	extern unsigned g_InsertInStartOrStopCodonCount;
	extern unsigned g_InsertAccConstrained;

	int NotUsedInt;
	if (ptrWhy == 0)
		ptrWhy = &NotUsedInt;
	*ptrWhy = 0;

	bool NotUsedBool;
	if (ptrAnyConstraint == 0)
		ptrAnyConstraint = &NotUsedBool;
	*ptrAnyConstraint = false;

	int Length = (int) Seq.size();
	double Prob1 = GetAcceptProb(Pos);
	gpos_t Pos2 = GetNextPos(Pos);
	if (Pos2 == GPOS_T_MAX)
		{
		*ptrWhy = 1;
		return false;
		}
	double Prob2 = GetAcceptProb(Pos2);

	if (Prob1 < 1.0 || Prob2 < 1.0)
		*ptrAnyConstraint = true;

	double p = sqrt(Prob1*Prob2);
	double Prob = pow(p, Length);
	if (!RandTrue(Prob))
		{
		++g_InsertRejectAccProb;
		*ptrWhy = 2;
		return false;
		}
	if (p < 1.0)
		++g_InsertAccConstrained;

	if (PosInCDS(Pos) || PosInCDS(Pos2))
		{
		if (Length%3 != 0)
			{
			++g_InsertRejectInCDSMod3Count;
			*ptrWhy = 3;
			return false;
			}

		if (InsertCreatesStopCodon(Pos, Seq))
			{
			++g_InsertCreatesStopCodonCount;
			*ptrWhy = 4;
			return false;
			}

		if (InsertPosInStartCodon(Pos) || InsertPosInStopCodon(Pos))
			{
			++g_InsertInStartOrStopCodonCount;
			*ptrWhy = 5;
			return false;
			}
		}
	return true;
	}

bool Evolver::AcceptDeleteLow(gpos_t Pos, gpos_t Length, int *ptrWhy,
  bool *ptrAnyConstraint) const
	{
	int NotUsedInt;
	if (ptrWhy == 0)
		ptrWhy = &NotUsedInt;

	bool NotUsedBool;
	if (ptrAnyConstraint == 0)
		ptrAnyConstraint = &NotUsedBool;

	*ptrWhy = 0;
	*ptrAnyConstraint = false;

	bool AllCDS = false;
	if (RangeHasCDS(Pos, Length, AllCDS))
		{
		if (Length%3 != 0)
			{
			*ptrWhy = 1;
			return false;
			}

		if (!AllCDS)
			{
			*ptrWhy = 2;
			return false;
			}

	// Codon delete P^2 instead of P^3 per codon
		double RangeProb = GetRangeAcceptProb(Pos, Length);
		if (RangeProb < 1.0)
			*ptrAnyConstraint = true;

		double Prob = pow(RangeProb, 2.0/3.0);
		if (!RandTrue(Prob))
			{
			*ptrWhy = 3;
			return false;
			}

		if (RangeHasStartCodon(Pos, Length))
			{
			*ptrWhy = 4;
			return false;
			}

		if (RangeHasStopCodon(Pos, Length))
			{
			*ptrWhy = 5;
			return false;
			}

		if (DeleteCreatesStopCodon(Pos, Length))
			{
			*ptrWhy = 6;
			return false;
			}
		++g_DelAccCDS;
		}
	else
		{
		if (!AcceptRange(Pos, Length, ptrAnyConstraint))
			{
			*ptrWhy = 7;
			return false;
			}
		}

	if (RangeHasIntron(Pos, Length))
		{
		if (!IntronLengthIsGE(Pos, P_MinIntronLength + Length))
			{
			*ptrWhy = 8;
			return false;
			}
		}

	gpos_t PrevPos = GetPrevPos(Pos);
	if (PrevPos == GPOS_T_MAX)
		PrevPos = Pos;
	gpos_t xLength = Length + 2;
	if (!RangeExists(PrevPos, xLength))
		xLength = Length + 1;

	if (RangeSpansTwoGenes(PrevPos, xLength))
		{
		*ptrWhy = 9;
		return false;
		}

	if (PrevPos == UINT_MAX)
		return false;
	const BaseInfo &PrevBase = GetBase(PrevPos);
	if (PrevBase.IsSplice())
		{
		*ptrWhy = 10;
		return false;
		}

	gpos_t EndPos = GetRangeEndPos(Pos, Length);
	gpos_t NextPos = GetNextPos(EndPos);
	if (NextPos == UINT_MAX)
		{
		*ptrWhy = 11;
		return false;
		}

	const BaseInfo &NextBase = GetBase(NextPos);
	if (NextBase.IsSplice())
		{
		*ptrWhy = 10;
		return false;
		}

	return true;
	}

bool Evolver::AcceptInvertLow(gpos_t Pos, gpos_t Length, int *ptrWhy) const
	{
	*ptrWhy = 0;

	gpos_t EndPos = GetRangeEndPos(Pos, Length);

	if (PosInCE(Pos) || PosInCE(EndPos))
		{
		*ptrWhy = 1;
		return false;
		}

	bool StartInGene = PosInGene(Pos);
	bool EndInGene = PosInGene(EndPos);
	if (StartInGene && EndInGene)
		{
		if (RangeHasCE(Pos, Length))
			{
			*ptrWhy = 2;
			return false;
			}
		}

	if ((StartInGene && !EndInGene) || (EndInGene && !StartInGene))
		{
		*ptrWhy = 3;
		return false;
		}

	return true;
	}

bool Evolver::InsertCreatesStopCodon(gpos_t Pos, const string &Seq_) const
	{
	const unsigned L = SIZE(Seq_);
	if (L%3 != 0)
		return false;

	CodonInfo Codon;
	bool Ok = GetCodon(Pos, Codon);
	if (!Ok)
		return false;

	string Prefix;
	string Suffix;
	const BaseInfo &Base = GetBase(Pos);
	switch (Base.Codex)
		{
	case 1:
		Prefix = Codon.Letters[0];
		Suffix = Codon.Letters[1];
		Suffix += Codon.Letters[2];
		break;

	case 2:
		Prefix = Codon.Letters[0];
		Prefix += Codon.Letters[1];
		Suffix = Codon.Letters[2];
		break;

	case 3:
		break;

	case -1:
		break;

	case -2:
		Prefix = Codon.Letters[0];
		Suffix = Codon.Letters[1];
		Suffix += Codon.Letters[2];
		break;

	case -3:
		Prefix = Codon.Letters[0];
		Prefix += Codon.Letters[1];
		Suffix = Codon.Letters[2];
		break;

	default:
		asserta(false);
		}

	string Seq(Seq_);
	if (!Codon.Strand)
		{
		void RevComp(string &);
		RevComp(Seq);
		}
	Seq = Prefix + Seq + Suffix;
	return HasStopCodon(Seq);
	}

// Assume both positions in same coding exon (splice sites, START
// and STOP prevent deletion from spanning two exons).
bool Evolver::DeleteCreatesStopCodon(gpos_t Pos, gpos_t Length) const
	{
	if (!PosInCDS(Pos))
		return false;
	gpos_t EndPos = GetRangeEndPos(Pos, Length);
	asserta(PosInCDS(EndPos));

	gpos_t PrevPos = GetAdjExonPos(Pos, false);
	const BaseInfo &PrevBase = GetBase(PrevPos);
	asserta(PrevBase.IsCDS());

	string Codon;
	switch (PrevBase.Codex)
		{
	case 1:
		{
		Codon.push_back(PrevBase.GetCodonLetter());

		gpos_t Pos2 = GetAdjExonPos(EndPos, true);
		if (Pos2 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base2 = GetBase(Pos2);
		if (Base2.Codex != 2)
			{
			Base2.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex 2");
			}
		Codon.push_back(Base2.GetCodonLetter());

		gpos_t Pos3 = GetAdjExonPos(Pos2, true);
		if (Pos3 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base3 = GetBase(Pos3);
		if (Base3.Codex != 3)
			{
			Base3.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex 3");
			}
		Codon.push_back(Base3.GetCodonLetter());
		break;
		}

	case 2:
		{
		gpos_t Pos1 = GetAdjExonPos(PrevPos, false);
		if (Pos1 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base1 = GetBase(Pos1);
		if (Base1.Codex != 1)
			{
			Base1.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex 1");
			}
		Codon.push_back(Base1.GetCodonLetter());
		Codon.push_back(PrevBase.GetCodonLetter());

		gpos_t Pos3 = GetAdjExonPos(EndPos, true);
		if (Pos3 == GPOS_T_MAX)
			return false;

		const BaseInfo &Base3 = GetBase(Pos3);
		if (Base3.Codex != 3)
			{
			Base3.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex 3");
			}
		Codon.push_back(Base3.GetCodonLetter());
		break;
		}

	case 3:
		return false;

	case -1:
		return false;

	case -2:
		{
		gpos_t Pos1 = GetAdjExonPos(EndPos, true);
		if (Pos1 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base1 = GetBase(Pos1);
		if (Base1.Codex != -1)
			{
			Base1.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex -1");
			}

		Codon.push_back(Base1.GetCodonLetter());
		Codon.push_back(PrevBase.GetCodonLetter());

		gpos_t Pos3 = GetAdjExonPos(PrevPos, false);
		if (Pos3 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base3 = GetBase(Pos3);
		if (Base3.Codex != -3)
			{
			Base1.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex -3");
			}
		Codon.push_back(Base3.GetCodonLetter());
		break;
		}

	case -3:
		{
		gpos_t Pos2 = GetAdjExonPos(EndPos, true);
		if (Pos2 == GPOS_T_MAX)
			return false;
		gpos_t Pos1 = GetAdjExonPos(Pos2, true);
		if (Pos1 == GPOS_T_MAX)
			return false;

		const BaseInfo &Base1 = GetBase(Pos1);
		const BaseInfo &Base2 = GetBase(Pos2);

		if (Base1.Codex != -1)
			{
			Base1.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex -1");
			}
		if (Base2.Codex != -2)
			{
			Base2.LogMe(true);
			Die("DeleteCreatesStopCodon, expected Codex -2");
			}

		Codon.push_back(Base1.GetCodonLetter());
		Codon.push_back(Base2.GetCodonLetter());
		Codon.push_back(PrevBase.GetCodonLetter());
		break;
		}

	default:
		asserta(false);
		}
	return IsStopCodon(Codon);
	}

bool Evolver::AcceptInvert(gpos_t Pos, gpos_t Length, int *ptrWhy) const
	{
	return AcceptInvertLow(Pos, Length, ptrWhy);
	}

bool Evolver::AcceptDelete(gpos_t Pos, gpos_t Length, int *ptrWhy) const
	{
	bool AnyConstraint = false;
	bool Accept = AcceptDeleteLow(Pos, Length, ptrWhy, &AnyConstraint);
	if (!Accept && AnyConstraint)
		++g_DelRejConstrained;
	if (Accept)
		{
		if (AnyConstraint)
			++g_DelAccConstrained;
		else
			++g_DelAccNeutral;
		}
	return Accept;
	}

bool Evolver::AcceptInsert(gpos_t Pos, const string &Seq, int *ptrWhy) const
	{
	bool AnyConstraint = false;
	bool Accept = AcceptInsertLow(Pos, Seq, ptrWhy, &AnyConstraint);
	if (!Accept && AnyConstraint)
		++g_InsRejConstrained;
	if (Accept)
		{
		if (AnyConstraint)
			++g_InsAccConstrained;
		else
			++g_InsAccNeutral;
		}
	return Accept;
	}
