#include "align4.h"

#define TRACE	0

static const AIVec *g_CmpHits;
static unsigned g_GenomeIndexForSort;

static bool ByLo(unsigned i, unsigned j)
	{
	assert(i < g_CmpHits->size());
	assert(j < g_CmpHits->size());
	const AlnInfo &AI1 = (*g_CmpHits)[i];
	const AlnInfo &AI2 = (*g_CmpHits)[j];
	return AI1.GetSeg(g_GenomeIndexForSort).lo <
	  AI2.GetSeg(g_GenomeIndexForSort).lo;
	}

static void AssertSortedByLo(const IntVec &SortOrder, unsigned GenomeIndex)
	{
	for (unsigned i = 0; i < SortOrder.size() - 1; ++i)
		{
		int HitIndex1 = SortOrder[i];
		int HitIndex2 = SortOrder[i+1];
		const AlnInfo &AI1 = (*g_CmpHits)[HitIndex1];
		const AlnInfo &AI2 = (*g_CmpHits)[HitIndex2];
		assert(AI1.GetSeg(GenomeIndex).lo <= AI2.GetSeg(GenomeIndex).lo);
		}
	}

static void AlignChain4(const IntVecVec &Chains, unsigned ChainIndex,
  const AIVec &InHits, const StdStrVec &InHitPaths, const AIVec &InHoles,
  const StdStrVec &InHolePaths, const IntVec &HitToLoHole, const IntVec &HitToHiHole,
  AIVec &OutHits, StdStrVec &OutPaths)
	{
#if	TRACE
	Log("AlignChain\n");
	Log("InHits:\n");
	LogAIs(InHits);
	Log("InHoles:\n");
	LogAIs(InHoles);
#endif

	const unsigned InHitCount = (unsigned) InHits.size();
	if (InHitCount == 0)
		return;

	const unsigned GenomeIndexA = GetFilterGenomeIndexA();
	const unsigned GenomeIndexB = GetFilterGenomeIndexB();

	const unsigned ChrIdA = GetFilterChrIdA();
	const unsigned ChrIdB = GetFilterChrIdB();

	const unsigned InHoleCount = (unsigned) InHoles.size();

	const IntVec &ChainHits = Chains[ChainIndex];
	const unsigned ChainHitCount = (unsigned) ChainHits.size();

// Sort by genome position in preparation
// for concatenating paths
	IntVec SortOrder;
	SortOrder.resize(ChainHitCount);

	const IntVec &HitIndexes = Chains[ChainIndex];
	unsigned i = 0;
	for_CIntVec(HitIndexes, p)
		{
		unsigned HitIndex = *p;
		SortOrder[i] = HitIndex;
		++i;
		}
	assert(i == ChainHitCount);

	g_CmpHits = &InHits;
	g_GenomeIndexForSort = GenomeIndexA;
	std::sort(SortOrder.begin(), SortOrder.end(), ByLo);

#if	TRACE
	{
	Log("\n");
	Log("Sorted chain:\n");
	Log("  Hit  PHole  NHole\n");
	Log("=====  =====  =====\n");
	for (unsigned i = 0; i < ChainHitCount; ++i)
		{
		unsigned HitIndex = SortOrder[i];
		Log("%5u", HitIndex);

		unsigned PHole = HitToLoHole[HitIndex];
		if (PHole == UINT_MAX)
			Log("      -");
		else
			Log("  %5u", PHole);

		unsigned NHole = HitToHiHole[HitIndex];
		if (NHole == UINT_MAX)
			Log("      -");
		else
			Log("  %5u", NHole);

		Log("\n");
		}
	Log("\n");
	}
#endif

	AssertSortedByLo(SortOrder, GenomeIndexA);
	AssertSortedByLo(SortOrder, GenomeIndexB);

// Build chain of hits & holes combined
// Note that first and last hits may be missing
// predecessor and successor hole respectively.
// All other hits are joined by a hole.
	AIVec ChainAIs;
	IntVec ChainAIIndexes;
	BoolVec IsHole;

	unsigned FirstHitIndex = SortOrder.front();
	unsigned FirstHoleIndex = HitToLoHole[FirstHitIndex];

#if	TRACE
	Log("FirstHitIndex=%u FirstHoleIndex=%u\n", FirstHitIndex, FirstHoleIndex);
#endif

	if (FirstHoleIndex != UINT_MAX)
		{
		assert(FirstHoleIndex < InHoleCount);
		const AlnInfo &FirstHole = InHoles[FirstHoleIndex];
		ChainAIs.push_back(FirstHole);

		ChainAIIndexes.push_back(FirstHoleIndex);
		IsHole.push_back(true);
		}

	for (unsigned i = 0; i < ChainHitCount; ++i)
		{
		unsigned HitIndex = SortOrder[i];
		assert(HitIndex < InHitCount);

		unsigned PrevHoleIndex = HitToLoHole[HitIndex];
		unsigned NextHoleIndex = HitToHiHole[HitIndex];

		if (PrevHoleIndex == UINT_MAX && i != 0)
			Die("Internal hit missing hole HitToLoHole[%u]=UINT_MAX", HitIndex);
		if (NextHoleIndex == UINT_MAX && i != ChainHitCount - 1)
			Die("Internal hit missing hole HitToHiHole[%u]=UINT_MAX", HitIndex);

		const AlnInfo &Hit = InHits[HitIndex];
		ChainAIs.push_back(Hit);

		ChainAIIndexes.push_back(HitIndex);
		IsHole.push_back(false);

		if (NextHoleIndex != UINT_MAX)
			{
			const AlnInfo &Hole = InHoles[NextHoleIndex];
			ChainAIs.push_back(Hole);

			ChainAIIndexes.push_back(NextHoleIndex);
			IsHole.push_back(true);
			}
		}

	const unsigned ChainAICount = (unsigned) ChainAIs.size();

#if	TRACE
	{
	Log("\n");
	Log("Chain with holes:\n");
	for (size_t i = 0; i < ChainAICount; ++i)
		{
		const AlnInfo &AI = *(ChainAIs[i]);
		if (AI.Type == BndFile::Hit)
			Log(" Hit  ");
		else if (AI.Type == BndFile::Hole)
			Log(" Hole ");
		else
			Log(" ???? ");
		Log(" [%3u]  ", ChainAIIndexes[i]);
		LogAI(AI);

		if (i > 0)
			{
			const AlnInfo &PrevAI = *(ChainAIs[i-1]);
			if (AI.Segs[0].lo != PrevAI.Segs[0].hi + 1)
				Log(" **NOT CONTIGUOUS (Genome 0)  ");
			if (AI.Segs[1].lo != PrevAI.Segs[1].hi + 1)
				Log(" **NOT CONTIGUOUS (Genome 1)  ");
			}

		Log("\n");
		}
	}
#endif

// Assert contiguous coverage
#ifndef NDEBUG
	{
	assert(ChainAIIndexes.size() == ChainAIs.size());
	assert(IsHole.size() == ChainAIs.size());

	for (size_t i = 0; i < ChainAIs.size(); ++i)
		{
		const AlnInfo &AI = ChainAIs[i];
		if (i > 0)
			{
			const AlnInfo &PrevAI = ChainAIs[i-1];
			assert(AI.GetSeg(GenomeIndexA).lo == PrevAI.GetSeg(GenomeIndexA).hi + 1);
			assert(AI.GetSeg(GenomeIndexB).lo == PrevAI.GetSeg(GenomeIndexB).hi + 1);
			}
		}
	}
#endif

	AlnInfo BoundRect;
	GetBoundRect(ChainAIs, GenomeIndexA, GenomeIndexB, BoundRect);

// Concatenate paths
	std::string GlobalPath;
	for (unsigned i = 0; i < ChainAICount; ++i)
		{
		unsigned AIIndex = ChainAIIndexes[i];
		if (IsHole[i])
			{
			const StdStr &Path = InHolePaths[AIIndex];
#ifndef NDEBUG
			assert(AIIndex < InHoleCount);
			const AlnInfo &ChainAI = ChainAIs[i];
			const std::string &InHolePath = InHolePaths[AIIndex];
			ValidateHitPath(ChainAI, GenomeIndexA, GenomeIndexB, InHolePath);
#endif
			GlobalPath += Path;
			}
		else
			{
			const StdStr &Path = InHitPaths[AIIndex];
#ifndef NDEBUG
			assert(AIIndex < InHitCount);
			const AlnInfo &ChainAI = ChainAIs[i];
			ValidateHitPath(ChainAI, GenomeIndexA, GenomeIndexB, Path);
#endif
			GlobalPath += Path;
			}
		}

// Get sequence in each genome chain for entire chain
	const GCSeg &BoundRectSegA = BoundRect.GetSeg(GenomeIndexA);
	const GCSeg &BoundRectSegB = BoundRect.GetSeg(GenomeIndexB);

	unsigned LengthA = (unsigned) BoundRectSegA.GetLength();
	unsigned LengthB = (unsigned) BoundRectSegB.GetLength();

	char *SeqA = myalloc(char, LengthA);
	char *SeqB = myalloc(char, LengthB);

	const RevFile &SeqRevFileA = GetSeqRevFileA();
	const RevFile &SeqRevFileB = GetSeqRevFileB();

	FilterAwareGetSeqData(SeqRevFileA, GenomeIndexA, ChrIdA, BoundRectSegA.lo,
	  LengthA, SeqA, true);

	FilterAwareGetSeqData(SeqRevFileB, GenomeIndexB, ChrIdB, BoundRectSegB.lo,
	  LengthB, SeqB, true);

// Write global alignment to trace file
	AlnInfo BoundRectAI;
	BoundRectAI.Score = 0;
	BoundRectAI.Segs.push_back(BoundRectSegA);
	BoundRectAI.Segs.push_back(BoundRectSegB);

	RevFile *ptrTraceRevFile = GetTraceRevFilePtr();
	WriteAln(ptrTraceRevFile, BoundRectAI, GlobalPath, ALN_ATTR_TraceType,
	  TraceType_ChainGlobalAln);

// Get optimal sub-alignments as in nwlocal
	AIVec LocalOutHits;
	StdStrVec LocalOutPaths;
	IntVec ColStarts;
	IntVec ColEnds;
	PostProcSlide(SeqA, LengthA, SeqB, LengthB, GlobalPath, LocalOutHits,
	  LocalOutPaths, ColStarts, ColEnds);

// Fix up coordinates because PostProcSlide() returns coords
// w.r.t. start of path.
	AddOffsetsVec(BoundRectAI, LocalOutHits);

	OutHits.insert(OutHits.end(), LocalOutHits.begin(), LocalOutHits.end());
	OutPaths.insert(OutPaths.end(), LocalOutPaths.begin(), LocalOutPaths.end());

#if	TRACE
	{
	const unsigned AlnCount = (unsigned) OutHits.size();
	Log("%u optimal subalns:\n", AlnCount);
	for (unsigned i = 0; i < AlnCount; ++i)
		{
		const AlnInfo &Hit = *(OutHits[i]);
		const char *Path = OutPaths[i];
		char *Row1;
		char *Row2;
		extern void HitToAln(const AlnInfo &Hit, const char *Path, char **ptrRow1, char **ptrRow2);
		HitToAln(Hit, Path, &Row1, &Row2);
		Log("\n");
		if (strlen(Path) < 100)
			{
			Log("%s\n", Row1);
			Log("%s\n", Row2);
			}
		else
			{
			Log("%100.100s...\n", Row1);
			Log("%100.100s...\n", Row2);
			}
		myfree(Row1);
		myfree(Row2);
		}
	Log("\n");
	}
#endif

	myfree(SeqA);
	myfree(SeqB);
	}

void AlignChains4(const IntVecVec &Chains, const AIVec &InHits,
  const StdStrVec &InHitPaths, const AIVec &InHoles,
  const StdStrVec &InHolePaths, const IntVec &HitToLoHole,
  const IntVec &HitToHiHole, AIVec &ChainHits, StdStrVec &ChainPaths)
	{
	const unsigned ChainCount = (unsigned) Chains.size();
	for (unsigned ChainIndex = 0; ChainIndex < ChainCount; ++ChainIndex)
		AlignChain4(Chains, ChainIndex, InHits, InHitPaths, InHoles, InHolePaths,
		  HitToLoHole, HitToHiHole, ChainHits, ChainPaths);
	}
