#ifndef ALNINFO_H
#define ALNINFO_H

#include "utils.h"
#include <float.h>
#include <set>

// Numerical values for AlnAttr and AIType_* should
// NEVER change as these are stored on disk.
enum AlnAttr
	{
	ALN_ATTR_BundleId = 0,
	ALN_ATTR_AlnLength = 1,
	ALN_ATTR_Placeholder = 2,
	ALN_ATTR_None = 3,
	ALN_ATTR_TraceType = 4,
	ALN_ATTR_ConfScore = 5,
	ALN_ATTR_BundleScore = 6,
	ALN_ATTR_ScaledBundleScore = 7,
	ALN_ATTR_MultixTruncatedEndBlock = 8,
	ALN_ATTR_UnalignedBlock = 9,
	ALN_ATTR_AncestralPos = 10,
	ALN_ATTR_Paralog = 11,
	ALN_ATTR_GeneIndex1 = 12,
	ALN_ATTR_GeneIndex2 = 13,
	ALN_ATTR_Frame = 14
	};

enum AlnAttrType
	{
	ALN_ATTR_TYPE_UInt32,
	ALN_ATTR_TYPE_Float32
	};

static AlnAttrType AATypes[] =
	{
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_BundleId = 0,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_AlnLength = 1,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_Placeholder = 2,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_None = 3,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_TraceType = 4,
	ALN_ATTR_TYPE_Float32,	//ALN_ATTR_ConfScore = 5,
	ALN_ATTR_TYPE_Float32,	//ALN_ATTR_BundleScore = 6,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_ScaledBundleScore = 7,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_MultixTruncatedEndBlock = 8,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_UnalignedBlock = 9,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_AncestralPos = 10,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_Paralog = 11,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_GeneIndex1 = 12,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_GeneIndex2 = 13,
	ALN_ATTR_TYPE_UInt32,	//ALN_ATTR_Frame = 14,
	};
const unsigned KnownAttrCount = sizeof(AATypes)/sizeof(AATypes[0]);

// Trace types are intended for use only internally
// in debugging and tracing code, and also in .rev
// trace files used for debugging and visualization,
// not in "real" output files.
const unsigned TraceType_Hit = 1;
const unsigned TraceType_RealignedHit = 2;
const unsigned TraceType_RealignedHole = 3;
const unsigned TraceType_ChainHit = 4;
const unsigned TraceType_HoleMinusHit = 5;
const unsigned TraceType_StartHole = 6;
const unsigned TraceType_EndHole = 7;
const unsigned TraceType_Hole = 8;
const unsigned TraceType_ChainGlobalAln = 9;
const unsigned TraceType_BoundRect = 10;
const unsigned TraceType_MultixLeftSubAln = 11;
const unsigned TraceType_MultixRightSubAln = 12;
const unsigned TraceType_MultixGuideAln = 13;
const unsigned TraceType_MultixLeftRemainder = 14;
const unsigned TraceType_MultixRightRemainder = 15;

const char *AlnAttrToStr(AlnAttr Attr);

typedef std::pair<coord_t, coord_t> Interval;

struct GCSeg
	{
	unsigned GenomeIndex;
	unsigned ChrId;
	coord_t lo;
	coord_t hi;
	bool Plus;

	void LogMe(bool NL = false) const
		{
		Log("%u,%u %"PF"-%"PF"%c(%u)",
		  GenomeIndex,
		  ChrId,
		  lo,
		  hi,
		  pom(Plus),
		  GetLength());
		if (NL)
			Log("\n");
		}

	unsigned GetLength() const
		{
		return hi - lo + 1;
		}

	bool operator<(const GCSeg &rhs) const
		{
		if (GenomeIndex != rhs.GenomeIndex)
			return GenomeIndex < rhs.GenomeIndex;
		if (ChrId != rhs.ChrId)
			return ChrId < rhs.ChrId;
		if (Plus != rhs.Plus)
			return (int) Plus < (int) rhs.Plus;
		if (lo != rhs.lo)
			return lo < rhs.lo;
		return hi < rhs.hi;
		}

	coord_t GetMidpoint() const
		{
		return (hi + lo)/2;
		}

	coord_t GetFrom() const
		{
		return Plus ? lo : hi;
		}

	coord_t GetTo() const
		{
		return Plus ? hi : lo;
		}

	coord_t GetOverlap(const GCSeg &s) const
		{
		if (GenomeIndex != s.GenomeIndex)
			return 0;

		if (ChrId != s.ChrId)
			return 0;

		coord_t maxlo = std::max(lo, s.lo);
		coord_t minhi = std::min(hi, s.hi);

		if (maxlo > minhi)
			return 0;

		return minhi - maxlo + 1;
		}

	bool operator==(const GCSeg &rhs) const
		{
		return GenomeIndex == rhs.GenomeIndex && ChrId == rhs.ChrId &&
		  lo == rhs.lo && hi == rhs.hi && Plus == rhs.Plus;
		}

	Interval operator()() const
		{
		return Interval(lo, hi);
		}

	void Init(coord_t p)
		{
		if (Plus)
			{
			lo = p;
			hi = lo - 1;
			}
		else
			{
			hi = p;
			lo = hi + 1;
			}
		}

	void Inc()
		{
		if (Plus)
			++hi;
		else
			{
			asserta(lo > 0);
			--lo;
			}
		}
	};

typedef std::vector<GCSeg> GCSegVec;

struct AlnInfo
	{
	float Score;
	GCSegVec Segs;
	std::vector <AlnAttr> Attrs;
	std::vector<unsigned> Values;
	unsigned AlnIndex;

	AlnInfo()
		{
		Clear();
		}

	void Clear()
		{
		Score = 0;
		Segs.clear();
		Attrs.clear();
		Values.clear();
		AlnIndex = UINT_MAX;
		}

	unsigned GetSegCount() const
		{
		return (unsigned) Segs.size();
		}

// The GetSeg family of functions return the FIRST segment for the
// given genome index. So they do not support multiple segments from
// the same genome.
	const GCSeg &GetSeg(unsigned GenomeIndex) const
		{
		for (std::vector<GCSeg>::const_iterator p = Segs.begin(); p != Segs.end(); ++p)
			{
			const GCSeg &Seg = *p;
			if (Seg.GenomeIndex == GenomeIndex)
				return Seg;
			}
		LogMe();
		Die("AlnInfo::GetSeg(%u), genome not present", GenomeIndex);
		return *((const GCSeg *) 0);
		}

	unsigned GetSegIndex(unsigned GenomeIndex) const
		{
		const unsigned SegCount = (unsigned) Segs.size();
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = Segs[SegIndex];
			if (Seg.GenomeIndex == GenomeIndex)
				return SegIndex;
			}
		LogMe();
		Die("AlnInfo::GetSeg(%u), genome not present", GenomeIndex);
		return UINT_MAX;
		}

	GCSeg *GetModifiableSegPtr(unsigned GenomeIndex)
		{
		for (std::vector<GCSeg>::iterator p = Segs.begin(); p != Segs.end(); ++p)
			{
			GCSeg &Seg = *p;
			if (Seg.GenomeIndex == GenomeIndex)
				return &Seg;
			}
		return 0;
		}

	const GCSeg *GetSegPtr(unsigned GenomeIndex) const
		{
		for (std::vector<GCSeg>::const_iterator p = Segs.begin(); p != Segs.end(); ++p)
			{
			const GCSeg &Seg = *p;
			if (Seg.GenomeIndex == GenomeIndex)
				return &Seg;
			}
		return 0;
		}

	const GCSeg *GetSegPtr(unsigned GenomeIndex, unsigned ChrId) const
		{
		for (std::vector<GCSeg>::const_iterator p = Segs.begin(); p != Segs.end(); ++p)
			{
			const GCSeg &Seg = *p;
			if (Seg.GenomeIndex == GenomeIndex && Seg.ChrId == ChrId)
				return &Seg;
			}
		return 0;
		}

	bool HasAttr(AlnAttr Attr) const
		{
		const unsigned AttrCount = (unsigned) Attrs.size();
		for (unsigned i = 0; i < AttrCount; ++i)
			{
			if (Attr == Attrs[i])
				return true;
			}
		return false;
		}

	unsigned GetAttr(AlnAttr Attr) const
		{
		if ((unsigned) Attr < KnownAttrCount && AATypes[Attr] != ALN_ATTR_TYPE_UInt32)
			Die("GetAttr(%s), type is not uint32", AlnAttrToStr(Attr));
		const unsigned AttrCount = (unsigned) Attrs.size();
		for (unsigned i = 0; i < AttrCount; ++i)
			{
			if (Attr == Attrs[i])
				return Values[i];
			}
		return UINT_MAX;
		}

	double GetFloatAttr(AlnAttr Attr) const
		{
		if ((unsigned) Attr < KnownAttrCount && AATypes[Attr] != ALN_ATTR_TYPE_Float32)
			Die("GetFloatAttr(%s), type is not float32", AlnAttrToStr(Attr));
		const unsigned AttrCount = (unsigned) Attrs.size();
		for (unsigned i = 0; i < AttrCount; ++i)
			{
			if (Attr == Attrs[i])
				{
				unsigned uValue = Values[i];
				double dValue = UIntToFloat32(uValue);
				return dValue;
				}
			}
		return DBL_MAX;
		}

	unsigned GetBundleId() const
		{
		return GetAttr(ALN_ATTR_BundleId);
		}

	unsigned GetAlnLength() const
		{
		unsigned AlnLength = GetAttr(ALN_ATTR_AlnLength);
		if (AlnLength == UINT_MAX)
			return 0;
		return AlnLength;
		}

	void SetAttr(AlnAttr Attr, unsigned Value)
		{
		if ((unsigned) Attr < KnownAttrCount && AATypes[Attr] != ALN_ATTR_TYPE_UInt32)
			Die("SetAttr(%s), type is not uint32", AlnAttrToStr(Attr));
		const unsigned AttrCount = (unsigned) Attrs.size();
		for (unsigned AttrIndex = 0; AttrIndex < AttrCount; ++AttrIndex)
			{
			if (Attrs[AttrIndex] == Attr)
				{
				Values[AttrIndex] = Value;
				return;
				}
			}

		Attrs.push_back(Attr);
		Values.push_back(Value);
		}

	void SetFloatAttr(AlnAttr Attr, double Value)
		{
		if ((unsigned) Attr < KnownAttrCount && AATypes[Attr] != ALN_ATTR_TYPE_Float32)
			Die("SetFloatAttr(%s), type is not float32", AlnAttrToStr(Attr));

		const unsigned AttrCount = (unsigned) Attrs.size();
		for (unsigned AttrIndex = 0; AttrIndex < AttrCount; ++AttrIndex)
			{
			if (Attrs[AttrIndex] == Attr)
				{
				unsigned uValue = Float32ToUInt((float32) Value);
				Values[AttrIndex] = uValue;
				return;
				}
			}

		Attrs.push_back(Attr);
		unsigned uValue = Float32ToUInt((float) Value);
		Values.push_back(uValue);
		}

	void SetBundleIndex(unsigned BundleId)
		{
		SetAttr(ALN_ATTR_BundleId, BundleId);
		}

	void SetAlnLength(unsigned AlnLength)
		{
		SetAttr(ALN_ATTR_AlnLength, AlnLength);
		}

	bool operator<(const AlnInfo &rhs) const
		{
		if (Score != rhs.Score)
			return Score < rhs.Score;

		const unsigned SegCount = (unsigned) Segs.size();
		const unsigned rhsSegCount = (unsigned) rhs.Segs.size();

		if (SegCount != rhsSegCount)
			return SegCount < rhsSegCount;

		const unsigned AttrCount = (unsigned) Attrs.size();
		const unsigned rhsAttrCount = (unsigned) rhs.Attrs.size();

		if (AttrCount != rhsAttrCount)
			return AttrCount < rhsAttrCount;

		for (unsigned i = 0; i < SegCount; ++i)
			{
			const GCSeg &Seg = Segs[i];
			const GCSeg &rhsSeg = rhs.Segs[i];
			if (!(Seg == rhsSeg))
				return Seg < rhsSeg;
			}

		for (unsigned i = 0; i < AttrCount; ++i)
			{
			if (Attrs[i] != rhs.Attrs[i])
				return Attrs[i] < rhs.Attrs[i];

			if (Values[i] != rhs.Values[i])
				return Values[i] != rhs.Values[i];
			}

		return false;
		}

	void LogMe(bool NL = true) const
		{
		const unsigned SegCount = (unsigned) Segs.size();
		for (unsigned i = 0; i < SegCount; ++i)
			{
			const GCSeg &Seg = Segs[i];
			Log(" %u,%u %u-%u(%u) %c",
			  Seg.GenomeIndex,
			  Seg.ChrId,
			  (unsigned) Seg.lo,
			  (unsigned) Seg.hi,
			  (unsigned) Seg.GetLength(),
			  Seg.Plus ? '+' : '-');
			}

	for (unsigned i = 0; i < SIZE(Attrs); ++i)
		{
		AlnAttr Attr = Attrs[i];
		unsigned Value = Values[i];

		switch (AATypes[Attr])
			{
		case ALN_ATTR_TYPE_UInt32:
			Log(" %s=%u", AlnAttrToStr(Attr), Value);
			break;
		case ALN_ATTR_TYPE_Float32:
			{
			double dValue = UIntToDouble(Value);
			Log(" %s=%g", AlnAttrToStr(Attr), dValue);
			break;
			}
		default:
			Log(" (Unknown AAType %u!)", AATypes[Attr]);
			}
		}

		if (NL)
			Log("\n");
		}

	bool operator==(const AlnInfo &rhs) const
		{
		if (Score != rhs.Score)
			return false;

		const unsigned SegCount = (unsigned) Segs.size();
		const unsigned rhsSegCount = (unsigned) rhs.Segs.size();

		if (SegCount != rhsSegCount)
			return false;

		const unsigned AttrCount = (unsigned) Attrs.size();
		const unsigned rhsAttrCount = (unsigned) rhs.Attrs.size();

		if (AttrCount != rhsAttrCount)
			return false;

		for (unsigned i = 0; i < SegCount; ++i)
			{
			const GCSeg &Seg = Segs[i];
			const GCSeg &rhsSeg = rhs.Segs[i];
			if (!(Seg == rhsSeg))
				return false;
			}

		for (unsigned i = 0; i < AttrCount; ++i)
			{
			if (Attrs[i] != rhs.Attrs[i])
				return false;

			if (Values[i] != rhs.Values[i])
				return false;
			}

		return true;
		}
	};

struct GenChr
	{
	unsigned GenomeIndex;
	unsigned ChrId;

	GenChr()
		{
		GenomeIndex = UINT_MAX;
		ChrId = UINT_MAX;
		}
	GenChr(unsigned GI, unsigned CI)
		{
		GenomeIndex = GI;
		ChrId = CI;
		}

	bool operator<(const GenChr &rhs) const
		{
		if (GenomeIndex < rhs.GenomeIndex)
			return true;
		if (GenomeIndex > rhs.GenomeIndex)
			return false;
		if (ChrId < rhs.ChrId)
			return true;
		return false;
		}
	};

struct AlnInfoX
	{
	AlnInfo AI;
	unsigned AlnInfoIndex;
	unsigned AlnIndex;
	};

struct AIBP
	{
	unsigned AlnInfoIndex;

// Need SegIndex and / or ChrId in different situations;
// they are in principle redundant (because ChrId =
// AIs[AlnInfoIndex].Segs[SegIndex].ChrId), but in
// practice it is very convenient to have both.
	unsigned SegIndex;
	unsigned ChrId;

	coord_t Pos;
	bool IsLo;

/***
There does not appear to be a sort order that allows overlap detection
by accumulating a context-free copy count.

Consider this example:

Pos		Lo/Hi	AIIndex	CopyCount
===		=====	=======	=========
1		lo		1		1
2		hi		1		0
2		lo		2		1
2		hi		2		0

Same segments with different AIIndex assignments:

Pos		Lo/Hi	AIIndex	CopyCount
===		=====	=======	=========
1		lo		2		1
2		lo		1		2 <<<< exceeds 1
2		hi		1		1
2		hi		2		0
***/
	bool operator<(const AIBP &rhs) const
		{
		if (ChrId != rhs.ChrId)
			return ChrId < rhs.ChrId;

		if (Pos != rhs.Pos)
			return Pos < rhs.Pos;

		if (IsLo != rhs.IsLo)
			return IsLo;

		if (AlnInfoIndex != rhs.AlnInfoIndex)
			return AlnInfoIndex < rhs.AlnInfoIndex;

		return SegIndex < rhs.SegIndex;
		}
	};

typedef std::vector<AIBP> AIBPVec;
typedef std::list<AIBP> AIBPList;

typedef std::vector<AlnInfo> AIVec;
typedef std::set<AlnInfo> AISet;

typedef std::vector<AlnInfoX> AIXVec;

void LogGCSeg(const GCSeg &Seg);
void LogAIs(const AIVec &AIs);
void LogAIXs(const AIXVec &AIXs);
void LogAI(const AlnInfo &AI);
void LogAIX(const AlnInfoX &AIX);
void LogAIN(const AlnInfo &AI);
void LogAIXN(const AlnInfoX &AIX);

void LogBP(const AIBP &BP);
void LogBPs(const AIBPVec &BPs);

void AIXsToAIs(const AIXVec &AIXs, AIVec &AIs);

const char *AlnAttrToStr(AlnAttr Attr);

#endif // ALNINFO_H
