#define TRACE	0

#include "revfile.h"
#include "getopts.h"
#include "progress.h"
#include "lhist.h"
#include "ivtree.h"
#include "lohi.h"
#include "gff2.h"
#include "base.h"

void GetLoHiGaps(const vector <LoHi> &LHs, vector<LoHi> &Gaps,
  unsigned Index);
void GetAnnotsByCol(const AlnInfo &AI, const StdStrVec &Rows,
  unsigned GenomeIndex, const vector<LoHi> &AnnotsByPos,
  vector<LoHi> &AnnotsByCol, vector<unsigned> &ColToOffset);
void GetOvAnnots(const vector<LoHi> &Annots1, const vector<LoHi> &Annots2,
  vector<LoHi> &Annots);

static LHistData g_IndelDist;
static LHistData g_PctIdDist;
static LHistData g_PctIdDist1;

static vector<LHistData> g_IndelDists;
static vector<LHistData> g_PctIdDists;
static vector<LHistData> g_PctIdDists1;

static vector<IvTree<LoHi> > g_AnnotIxs1;
static vector<IvTree<LoHi> > g_AnnotIxs2;
static unsigned g_GenomeIndex1;
static unsigned g_GenomeIndex2;
static FILE *g_fOut1;
static FILE *g_fOut2;
static vector<vector<unsigned> > g_Venn;

static bool IsParalog(const AlnInfo &AI)
	{
	const unsigned AttrCount = SIZE(AI.Attrs);
	for (unsigned i = 0; i < AttrCount; ++i)
		{
		if (AI.Attrs[i] == ALN_ATTR_Paralog && AI.Values[i] != 0)
			return true;
		}
	return false;
	}

static void LogLoHiHdr()
	{
	Log("        lo          hi         Len  Type\n");
	Log("==========  ==========  ==========  =====\n");
	}

static void LogLoHiCE(const LoHi &LH)
	{
	Log("%10"PF"  %10"PF"  %10"PF"  %s\n",
	  LH.Lo, LH.Hi, LH.Hi - LH.Lo + 1, CETypeToStr((CE_TYPE) LH.Index));
	}

void LogLoHiCEs(const string &Msg, const vector<LoHi> &LHs)
	{
	Log("\n");
	if (!Msg.empty())
		Log("%s:\n", Msg.c_str());
	LogLoHiHdr();
	const unsigned N = SIZE(LHs);
	for (unsigned i = 0; i < N; ++i)
		{
		const LoHi &LH = LHs[i];
		LogLoHiCE(LH);
		}
	}

static void GetGapLengths(unsigned FromCol, unsigned ToCol,
  const StdStr &Row1, const StdStr &Row2, vector<unsigned> &Lengths)
	{
	Lengths.clear();

	unsigned GapLength1 = 0;
	unsigned GapLength2 = 0;
	unsigned ColCount = SIZE(Row1);
	asserta(SIZE(Row2) == ColCount);
	asserta(FromCol < ColCount);
	asserta(ToCol < ColCount);
	asserta(FromCol <= ToCol);
	for (unsigned ColIndex = FromCol; ColIndex < ToCol; ++ColIndex)
		{
		char c1 = Row1[ColIndex];
		char c2 = Row2[ColIndex];
		bool Gap1 = (c1 == '-');
		bool Gap2 = (c2 == '-');
		if (Gap1 && Gap2)
			continue;

		if (Gap1)
			++GapLength1;
		else
			{
			if (GapLength1 > 0)
				{
				Lengths.push_back(GapLength1);
				GapLength1 = 0;
				}
			}

		if (Gap2)
			++GapLength2;
		else
			{
			if (GapLength2 > 0)
				{
				Lengths.push_back(GapLength2);
				GapLength2 = 0;
				}
			}
		}
	if (GapLength1 > 0)
		Lengths.push_back(GapLength1);
	if (GapLength2 > 0)
		Lengths.push_back(GapLength2);
	}

// Add neutral ranges at beginning and end segment
static void AddNeutralMargins(const GCSeg &Seg, vector<LoHi> &Ranges)
	{
	sort(Ranges.begin(), Ranges.end());

	LoHi Neutral;
	Neutral.Index = CE_Neutral;

	if (Ranges.empty())
		{
		Neutral.Lo = Seg.lo;
		Neutral.Hi = Seg.hi;
		Ranges.push_back(Neutral);
		return;
		}

	gpos_t CELo = Ranges.front().Lo;
	gpos_t CEHi = Ranges.back().Hi;

	if (CELo > Seg.lo)
		{
		Neutral.Lo = Seg.lo;
		Neutral.Hi = CELo - 1;
		Ranges.push_back(Neutral);
		}

	if (CEHi < Seg.hi)
		{
		Neutral.Lo = CEHi + 1;
		Neutral.Hi = Seg.hi;
		Ranges.push_back(Neutral);
		}
	}

static void GetAnnots(const AlnInfo &AI, unsigned GenomeIndex,
  const vector<IvTree<LoHi> > &AnnotIxs, vector<LoHi> &Annots)
	{
	const GCSeg &Seg = AI.GetSeg(GenomeIndex);

	unsigned ChrId = Seg.ChrId;
	asserta(ChrId < SIZE(AnnotIxs));
	const IvTree<LoHi> &AnnotIx = AnnotIxs[ChrId];

// Get CE annotations
	pair<gpos_t, gpos_t> Iv(Seg.lo, Seg.hi);
	AnnotIx.GetOverlaps(Iv, Annots);

// Add neutral ranges before first CE and after last CE
	AddNeutralMargins(Seg, Annots);

// Get internal neutral ranges (gaps between CE ranges)
	vector<LoHi> NeutralRanges;
	GetLoHiGaps(Annots, NeutralRanges, CE_Neutral);

// Append neutral ranges to CE ranges
	Annots.insert(Annots.end(), NeutralRanges.begin(), NeutralRanges.end());

	sort(Annots.begin(), Annots.end());
	}

static void OutputRecPair(const RevFile &Rev, CE_TYPE CEType1, CE_TYPE CEType2,
  unsigned FromCol, unsigned ToCol,
  const GCSeg &Seg1, const GCSeg &Seg2,
  const vector<unsigned> &ColToOffset1, const vector<unsigned> &ColToOffset2,
  const string &Attrs)
	{
	asserta(FromCol < SIZE(ColToOffset1));
	asserta(FromCol < SIZE(ColToOffset2));
	asserta(ToCol < SIZE(ColToOffset1));
	asserta(ToCol < SIZE(ColToOffset2));

	unsigned FromOffset1 = ColToOffset1[FromCol];
	unsigned FromOffset2 = ColToOffset2[FromCol];

	unsigned ToOffset1 = ColToOffset1[ToCol];
	unsigned ToOffset2 = ColToOffset2[ToCol];

	unsigned LoOffset1 = min(FromOffset1, ToOffset1);
	unsigned LoOffset2 = min(FromOffset2, ToOffset2);

	unsigned HiOffset1 = max(FromOffset1, ToOffset1);
	unsigned HiOffset2 = max(FromOffset2, ToOffset2);

	coord_t Start1 = Seg1.lo + LoOffset1 + 1;
	coord_t End1 = Seg1.lo + HiOffset1 + 1;

	coord_t Start2 = Seg2.lo + LoOffset2 + 1;
	coord_t End2 = Seg2.lo + HiOffset2 + 1;

	const string &ChrName1 = Rev.GetChrName(Seg1.GenomeIndex, Seg1.ChrId);
	const string &ChrName2 = Rev.GetChrName(Seg2.GenomeIndex, Seg2.ChrId);

	GFFRecord2 Rec;
	Rec.SeqName = ChrName1;
	Rec.Source = "evo";
	Rec.Score = 0;
	Rec.Strand = Seg1.Plus == Seg2.Plus ? '+' : '-';
	Rec.Frame = INT_MAX;
	Rec.Attrs = Attrs;
	GFFAppendTargetAttrs(Rec.Attrs, ChrName2, Start2, End2);
	if (CEType1 != CEType2)
		GFFAppendStrAttr(Rec.Attrs, "cetype2", CETypeToStr(CEType2));

	Rec.Feature = CETypeToStr(CEType1);
	Rec.Start = Start1;
	Rec.End = End1;
	if (g_fOut1 != 0)
		WriteGFFRec(g_fOut1, Rec);

	Rec.Feature = CETypeToStr(CEType2);
	Rec.Start = Start2;
	Rec.End = End2;

	Rec.Attrs = Attrs;
	GFFAppendTargetAttrs(Rec.Attrs, ChrName1, Start1, End1);
	if (CEType1 != CEType2)
		GFFAppendStrAttr(Rec.Attrs, "cetype2", CETypeToStr(CEType1));
	if (g_fOut2 != 0)
		WriteGFFRec(g_fOut2, Rec);
	}

static double GetPctId(unsigned FromCol, unsigned ToCol,
  const StdStr &Row1, const StdStr &Row2, unsigned &AlignedColCount)
	{
	AlignedColCount = 0;
	const unsigned N = SIZE(Row1);
	asserta(SIZE(Row2) == N);
	unsigned SameCount = 0;
	asserta(FromCol < N);
	asserta(ToCol < N);
	asserta(FromCol <= ToCol);
	for (unsigned i = FromCol; i <= ToCol; ++i)
		{
		char c1 = Row1[i];
		if (c1 == '-')
			continue;
		char c2 = Row2[i];
		if (c2 == '-')
			continue;
		++AlignedColCount;
		if (toupper(c1) == toupper(c2))
			++SameCount;
		}
	if (AlignedColCount == 0)
		return -1;
	return (double) SameCount*100.0 / (double) AlignedColCount;
	}

static void MakeGapLengthAttr(const vector<unsigned> &GapLengths,
  string &Attr)
	{
	Attr.clear();

	map<unsigned, unsigned> LengthToCount;
	for (vector<unsigned>::const_iterator p = GapLengths.begin();
	  p != GapLengths.end(); ++p)
		{
		unsigned Length = *p;
		if (LengthToCount.find(Length) == LengthToCount.end())
			LengthToCount[Length] = 1;
		else
			++LengthToCount[Length];
		}

	for (map<unsigned, unsigned>::const_iterator p = LengthToCount.begin();
	  p != LengthToCount.end(); ++p)
		{
		unsigned Length = p->first;
		unsigned Count = p->second;
		char Str[32];
		if (Count > 1)
			sprintf(Str, "%u*%u", Count, Length);
		else
			sprintf(Str, "%u", Length);
		if (!Attr.empty())
			Attr += ";";
		Attr += Str;
		}
	}

static void UpdateStats(const RevFile &Rev, unsigned FromCol, unsigned ToCol,
  CE_TYPE CEType, const GCSeg &Seg1, const GCSeg &Seg2, const StdStr &Row1,
  const StdStr &Row2, const vector<unsigned> &ColToOffset1,
  const vector<unsigned> &ColToOffset2)
	{
	asserta((unsigned) CEType < CETypeCount);

	string Attrs;

	vector<unsigned> GapLengths;
	GetGapLengths(FromCol, ToCol, Row1, Row2, GapLengths);
	if (!GapLengths.empty())
		{
		string GapLengthAttr;
		MakeGapLengthAttr(GapLengths, GapLengthAttr);
		GFFAppendStrAttr(Attrs, "indels", GapLengthAttr);
		}

	unsigned AlignedColCount = UINT_MAX;
	double PctId = GetPctId(FromCol, ToCol, Row1, Row2, AlignedColCount);
	if (PctId == -1)
		GFFAppendStrAttr(Attrs, "pctid", "allgaps");
	else
		{
		if (PctId == 100)
			{
			g_PctIdDist.Add(100, AlignedColCount);
			g_PctIdDist1.Add(100, AlignedColCount);
			g_PctIdDists[CEType].Add(100, AlignedColCount);
			g_PctIdDists1[CEType].Add(100, AlignedColCount);
			}
		else
			{
			unsigned i = (unsigned) (PctId + 0.5);
			g_PctIdDist.Add(i, AlignedColCount);
			g_PctIdDist1.Add(i, AlignedColCount);
			g_PctIdDists[CEType].Add(i, AlignedColCount);
			g_PctIdDists1[CEType].Add(i, AlignedColCount);
			}
		char strPctId[16];
		sprintf(strPctId, "%.1f", PctId);
		GFFAppendStrAttr(Attrs, "pctid", strPctId);
		}

	OutputRecPair(Rev, CEType, CEType, FromCol, ToCol, Seg1, Seg2, ColToOffset1,
	  ColToOffset2, Attrs);
	}

static void UpdateIndelDist(LHistData &Dist, const vector<unsigned> &GapLengths)
	{
	Dist.Add(GapLengths);
	}

static void DoAln(const RevFile &Rev, const AlnInfo &AI, const StdStrVec &Rows)
	{
	vector<LoHi> Annots1;
	vector<LoHi> Annots2;
	GetAnnots(AI, g_GenomeIndex1, g_AnnotIxs1, Annots1);
	GetAnnots(AI, g_GenomeIndex2, g_AnnotIxs2, Annots2);

	vector<LoHi> AnnotsByCol1;
	vector<LoHi> AnnotsByCol2;

	vector<unsigned> ColToOffset1;
	vector<unsigned> ColToOffset2;

	GetAnnotsByCol(AI, Rows, g_GenomeIndex1, Annots1, AnnotsByCol1, ColToOffset1);
	GetAnnotsByCol(AI, Rows, g_GenomeIndex2, Annots2, AnnotsByCol2, ColToOffset2);

	vector<LoHi> AnnotsByCol;
	GetOvAnnots(AnnotsByCol1, AnnotsByCol2, AnnotsByCol);

	unsigned SegIndex1 = AI.GetSegIndex(g_GenomeIndex1);
	unsigned SegIndex2 = AI.GetSegIndex(g_GenomeIndex2);

	const GCSeg &Seg1 = AI.Segs[SegIndex1];
	const GCSeg &Seg2 = AI.Segs[SegIndex2];

	const unsigned RowCount = SIZE(Rows);
	asserta(SegIndex1 < RowCount);
	asserta(SegIndex2 < RowCount);

	const StdStr &Row1 = Rows[SegIndex1];
	const StdStr &Row2 = Rows[SegIndex2];

	const unsigned ColCount = SIZE(Row1);
	asserta(SIZE(Row2) == ColCount);

	vector<unsigned> GapLengths;
	if (ColCount > 0)
		{
		GetGapLengths(0, ColCount - 1, Row1, Row2, GapLengths);
		UpdateIndelDist(g_IndelDist, GapLengths);
		}

	const unsigned AnnotCount = SIZE(AnnotsByCol);
	for (unsigned i = 0; i < AnnotCount; ++i)
		{
		const LoHi &LH = AnnotsByCol[i];
		CE_TYPE CEType1 = (CE_TYPE) (LH.Index & 0xffff);
		CE_TYPE CEType2 = (CE_TYPE) (LH.Index >> 16);

		unsigned FromCol = LH.Lo;
		unsigned ToCol = LH.Hi;

		asserta(FromCol <= ToCol);
		unsigned ColCount = ToCol - FromCol + 1;
		g_Venn[CEType1][CEType2] += ColCount;

		if (CEType1 == CE_Invalid || CEType2 == CE_Invalid || CEType1 != CEType2)
			{
			OutputRecPair(Rev, CEType1, CEType2, FromCol, ToCol, Seg1, Seg2,
			  ColToOffset1, ColToOffset2, "");
			continue;
			}
		UpdateStats(Rev, FromCol, ToCol, CEType1, Seg1, Seg2,  Row1, Row2, ColToOffset1,
		  ColToOffset2);

		vector<unsigned> GapLengths;
		GetGapLengths(FromCol, ToCol, Row1, Row2, GapLengths);

		asserta((unsigned) CEType1 < CETypeCount);
		UpdateIndelDist(g_IndelDists[CEType1], GapLengths);
		}

#if	TRACE
	{
	Log("\n");
	Log("DoAln");
	AI.LogMe();
	LogLoHiCEs("Annots1", Annots1);
	LogLoHiCEs("Annots2", Annots2);

	LogLoHiCEs("AnnotsByCol1", AnnotsByCol1);
	LogLoHiCEs("AnnotsByCol2", AnnotsByCol2);

	LogLoHiCEs("AnnotsByCol", AnnotsByCol);
	}
#endif
	}

CE_TYPE FeatureToCEType(const string &Feature)
	{
	if (Feature == "CDS")
		return CE_CDS;
	else if (Feature == "UTR")
		return CE_UTR;
	else if (Feature == "NXE")
		return CE_NXE;
	else if (Feature == "NGE")
		return CE_NGE;
	else
		return CE_Invalid;
	}

static void ReadAnnotIxs(const RevFile &Rev, unsigned GenomeIndex,
  const char *FileName, vector<IvTree<LoHi> > &AnnotIxs)
	{
	FILE *f = OpenStdioFile(FileName);

	IvTree<LoHi> *AnnotIx = 0;
	string ChrName;
	unsigned ChrId = UINT_MAX;
	GFFRecord2 Rec;
	ProgressStartFile(f, "Reading %s", FileName);
	unsigned RecordCount = 0;
	unsigned SkippedRecordCount = 0;
	while (GetNextGFFRecord(f, Rec))
		{
		if (RecordCount++%100 == 0)
			ProgressStepFile();

		if (ChrId == UINT_MAX || Rec.SeqName != ChrName)
			{
			ChrName = Rec.SeqName;
			ChrId = Rev.GetChrId(GenomeIndex, ChrName);
			unsigned ChrIndex = Rev.GetChrIndex(GenomeIndex, ChrId);
			if (ChrId != ChrIndex)
				Die("ChrId != ChrIndex, sparse chr ids not supported");
			asserta(ChrId < SIZE(AnnotIxs));
			AnnotIx = &AnnotIxs[ChrId];
			}

		CE_TYPE CET = FeatureToCEType(Rec.Feature);
		if (CET == CE_Invalid)
			{
			++SkippedRecordCount;
			continue;
			}

		asserta(Rec.Start > 0);
		asserta(Rec.End >= Rec.Start);
		LoHi LH(Rec.Start - 1, Rec.End - 1, CET);

		AnnotIx->Add(LH);
		}
	ProgressDone();
	Progress("%u GFF records, %u skipped", RecordCount, SkippedRecordCount);
	CloseStdioFile(f);
	}

static void OutputVenn(const char *FileName1, const char *FileName2)
	{
	Log("\n");
	Log("Agreement matrix for annotations.\n");
	Log("Values are numbers of alignment columns with annotations of given type.\n");
	Log("Rows %s\n", FileName1);
	Log("Columns %s\n", FileName2);
	Log("\n");

// Column headings
	Log("             ");
	for (unsigned Type1 = 0; Type1 < CETypeCount; ++Type1)
		{
		if (Type1 == CE_Splice || Type1 == CE_Unknown || Type1 == CE_Invalid)
			continue;
		Log(" %10.10s", CETypeToStr((CE_TYPE) Type1));
		}
	Log("\n");
	Log("             ");
	for (unsigned Type1 = 0; Type1 < CETypeCount; ++Type1)
		{
		if (Type1 == CE_Splice || Type1 == CE_Unknown || Type1 == CE_Invalid)
			continue;
		Log("  ---------");
		}
	Log("\n");

// Matrix
	vector<unsigned> ColTotals(CETypeCount, 0);
	for (unsigned Type1 = 0; Type1 < CETypeCount; ++Type1)
		{
		if (Type1 == CE_Splice || Type1 == CE_Unknown || Type1 == CE_Invalid)
			continue;
		Log("%10.10s | ", CETypeToStr((CE_TYPE) Type1));
		unsigned RowTotal = 0;
		for (unsigned Type2 = 0; Type2 < CETypeCount; ++Type2)
			{
			if (Type2 == CE_Splice || Type2 == CE_Unknown || Type2 == CE_Invalid)
				continue;
			unsigned Count = g_Venn[Type1][Type2];
			Log(" %10u", Count);
			RowTotal += Count;
			ColTotals[Type2] += Count;
			}
		Log("  | %10u", RowTotal);
		Log("\n");
		}

// Row totals
	Log("             ");
	for (unsigned Type1 = 0; Type1 < CETypeCount; ++Type1)
		{
		if (Type1 == CE_Splice || Type1 == CE_Unknown || Type1 == CE_Invalid)
			continue;
		Log("  ---------");
		}
	Log("\n");
	Log("             ");
	for (unsigned Type1 = 0; Type1 < CETypeCount; ++Type1)
		{
		if (Type1 == CE_Splice || Type1 == CE_Unknown || Type1 == CE_Invalid)
			continue;
		Log(" %10u", ColTotals[Type1]);
		}
	Log("\n");

	for (unsigned Type1 = 0; Type1 < CETypeCount; ++Type1)
		{
		if (Type1 == CE_Unknown || Type1 == CE_Invalid)
			{
			for (unsigned Type2 = 0; Type2 < CETypeCount; ++Type2)
				{
				if (Type2 != CE_Unknown && Type2 != CE_Invalid)
					continue;
				unsigned Count = g_Venn[Type2][Type1];
				if (Count > 0)
					Warning("Unexpcted annotation alignment %s<->%s in %u columns",
					  CETypeToStr((CE_TYPE) Type1),
					  CETypeToStr((CE_TYPE) Type2),
					  Count);
				}
			}
		}
	}

void AlnStats()
	{
	const char *AlnFileName = RequiredValueOpt("alnstats");

	const char *InFileName1 = ValueOpt("annots1");
	const char *InFileName2 = ValueOpt("annots2");

	const char *OutFileName1 = ValueOpt("outannots1");
	const char *OutFileName2 = ValueOpt("outannots2");

//  -----------------------------------------------------
//  Common code with ProtStats
	g_GenomeIndex1 = 0;//TODO
	g_GenomeIndex2 = 1;//TODO

	RevFile Rev;
	Rev.Open(AlnFileName);

	const unsigned ChrCount1 = Rev.GetChrCount(g_GenomeIndex1);
	const unsigned ChrCount2 = Rev.GetChrCount(g_GenomeIndex2);

	g_AnnotIxs1.resize(ChrCount1);
	g_AnnotIxs2.resize(ChrCount2);

	map<unsigned, vector<LoHi> > GeneIndexToCDSs1;
	map<unsigned, vector<LoHi> > GeneIndexToCDSs2;

	if (InFileName1 != 0)
		ReadAnnotIxs(Rev, g_GenomeIndex1, InFileName1, g_AnnotIxs1);

	if (InFileName2 != 0)
		ReadAnnotIxs(Rev, g_GenomeIndex2, InFileName2, g_AnnotIxs2);

	if (OutFileName1 != 0)
		g_fOut1 = CreateStdioFile(OutFileName1);

	if (OutFileName2 != 0)
		g_fOut2 = CreateStdioFile(OutFileName2);

#if	0//TRACE
	Log("\n");
	Log("Annots1\n");
	Log("        lo          hi         Len  Type\n");
	Log("==========  ==========  ==========  =====\n");
	g_AnnotIx1.Traverse(LogLoHiCE);
	Log("\n");
	Log("Annots2\n");
	Log("        lo          hi         Len  Type\n");
	Log("==========  ==========  ==========  =====\n");
	g_AnnotIx2.Traverse(LogLoHiCE);
#endif

	AIVec AIs;
	Rev.GetAllAlnInfos(AIs);
	const unsigned AICount = (unsigned) AIs.size();

	g_IndelDist.Init("Total indel length distribution)");
	g_PctIdDist.InitPctId("Total %id distribution (bin width 10)");
	g_PctIdDist1.Init("Total %id distribution (bin width 1)", 0, 100, 1);

	g_IndelDists.resize(CETypeCount);
	g_PctIdDists.resize(CETypeCount);
	g_PctIdDists1.resize(CETypeCount);
	g_Venn.resize(CETypeCount);
	for (unsigned i = 0; i < CETypeCount; ++i)
		{
		g_Venn[i].resize(CETypeCount);

		string Name = CETypeToStr((CE_TYPE) i);
		g_IndelDists[i].Init("Indel dist. " + Name);
		g_PctIdDists[i].InitPctId("Pct id. dist. " + Name + " (bin width 10)");
		g_PctIdDists1[i].Init("Pct id. dist. " + Name + " (bin width 1)", 0, 100, 1);
		}

	unsigned ParalogCount = 0;
	ProgressStart("Reading alignments");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%100 == 0)
			ProgressStep(AIIndex, AICount);

//		unsigned AlnIndex = Rev.GetAlnIndex(AIIndex);
		unsigned AlnIndex = AIs[AIIndex].AlnIndex;
		if (AlnIndex == UINT_MAX)
			continue;

		StdStrVec Rows;
		Rev.GetAlnRows(AlnIndex, Rows);

		const AlnInfo &AI = AIs[AIIndex];
		if (IsParalog(AI) == 1)
			{
			++ParalogCount;
			continue;
			}
		DoAln(Rev, AI, Rows);
#if	TRACE
		if (AIIndex > 10)
			Die("Premature (TRACE)");
#endif
		}
	ProgressDone();
	Progress("%u paralagous alignments skipped", ParalogCount);

	for (unsigned i = 0; i < CETypeCount; ++i)
		{
		const LHistData &LH = g_IndelDists[i];
		if (LH.GetCount() > 0)
			LH.LogHist(true);
		}

	for (unsigned i = 0; i < CETypeCount; ++i)
		{
		const LHistData &LH = g_PctIdDists[i];
		if (LH.GetCount() > 0)
			LH.LogHist(true);
		}

	for (unsigned i = 0; i < CETypeCount; ++i)
		{
		const LHistData &LH = g_PctIdDists1[i];
		if (LH.GetCount() > 0)
			LH.LogHist(true);
		}

	g_IndelDist.LogHist(true);
	g_PctIdDist.LogHist();
	g_PctIdDist1.LogHist();

	OutputVenn(InFileName1, InFileName2);

	Rev.Close();
	}
