#include "ivtree.h"
#include "gff2.h"
#include "ratemgr.h"
#include "progress.h"
#include "myrand.h"
#include "getopts.h"
#include "evomodel.h"
#include "genncces.h"

void SetModel(const char *FileName);

static gpos_t g_SeqLength;
static string g_SeqName;
static map<unsigned, GeneData> g_Genes;

double Normal(double x, double Mean, double StdDev)
	{
	double C = sqrt(2*M_PI);
	double a = (x - Mean);
	double D = (a*a)/(2.0*StdDev*StdDev);
	return (1.0/(StdDev*C))*exp(-D);
	}

static vector<pair<unsigned, double> > g_FractThresh;

static void InitRandomFract()
	{
	const unsigned BinCount = 100;
	vector<double> Norm(100, FLT_MAX);
	double Max_x = 0.5;
	double Total = 0;
	for (unsigned i = 0; i < BinCount; ++i)
		{
		double x = (Max_x*i)/BinCount;
		double y = Normal(x, P_MeanInterGeneFract, P_StdDevInterGeneFract);
		Total += y;
//		Log("Normal(%.3f) = %.3f\n", x, y);
		}

	double Accum = 0;
	for (unsigned i = 0; i < BinCount; ++i)
		{
		double x = (Max_x*i)/BinCount;
		double y = Normal(x, P_MeanInterGeneFract, P_StdDevInterGeneFract);
		Accum += y;
		unsigned Thresh = (unsigned) (UINT_MAX*Accum/Total + 0.5);
		g_FractThresh.push_back(pair<unsigned, double>(Thresh, x));
		}
	}
 
static double GetRandomFract()
	{
	unsigned r = RandInt(UINT_MAX);
	unsigned N = SIZE(g_FractThresh);
	for (unsigned i = 0; i < N; ++i)
		if (r <= g_FractThresh[i].first)
			return g_FractThresh[i].second;

	Die("GetRandomFract failed");
	UNREACHABLE_RETURN(0);
	}

static void AppendGeneCE(unsigned GeneIndex, gpos_t lo, gpos_t hi)
	{
	map<unsigned, GeneData>::iterator p = g_Genes.find(GeneIndex);
	if (p == g_Genes.end())
		{
		GeneData Gene;
		Gene.Index = GeneIndex;
		Gene.lo = lo;
		Gene.hi = hi;
		g_Genes[GeneIndex] = Gene;
		}
	else
		{
		GeneData &Gene = p->second;
		Gene.lo = min(lo, Gene.lo);
		Gene.hi = max(hi, Gene.hi);
		}
	}

static void LogGenes()
	{
	Log("\n");
	Log("Genes:\n");
	Log("Index          lo          hi\n");
	Log("=====  ==========  ==========\n");
	for (map<unsigned, GeneData>::const_iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		const GeneData &Gene = p->second;
		Log("%5u  %10"PF"  %10"PF"\n",
		  p->first, Gene.lo, Gene.hi);
		}
	Log("\n");
	}

void LogLoHi(const LoHi &LH)
	{
	Log("%10"PF"  %10"PF"  %10u\n", LH.Lo, LH.Hi, LH.Index);
	}

void AssignNCCEs()
	{
	const char *NCCEFileName = RequiredValueOpt("assncces");
	const char *GeneFileName = RequiredValueOpt("genes");
	const char *OutFileName = RequiredValueOpt("out");
	const char *ModelFileName = ValueOpt("model");
	g_SeqLength = RequiredIntOpt("length");
	bool Dump = FlagOpt("dump");

	SetModel(ModelFileName);
	InitRandomFract();

	g_SeqName = "";
	ProgressStart("Reading genes");
	unsigned RecordCount = 0;
	FILE *fGenes = OpenStdioFile(GeneFileName);
	GFFRecord2 Rec;
	unsigned FileSize = (unsigned) GetStdioFileSize(fGenes);
	while (GetNextGFFRecord(fGenes, Rec))
		{
		if ((RecordCount++)%1000 == 0)
			{
			unsigned Pos = (unsigned) GetStdioFilePos(fGenes);
			ProgressStep(Pos, FileSize);
			}
		const string Name(Rec.SeqName);
		if (g_SeqName == "")
			g_SeqName = Name;
		else if (Name != g_SeqName)
			Die("Gene file has more than one sequence name '%s', '%s'",
			  g_SeqName.c_str(), Name.c_str());

		gpos_t lo = Rec.Start - 1;
		gpos_t hi = Rec.End - 1;

		string strGeneIndex;
		bool Found = GFFGetAttr(Rec.Attrs, "gene_index", strGeneIndex);
		if (Found)
			{
			unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
			AppendGeneCE(GeneIndex, lo, hi);
			}
		}
	ProgressDone();
	CloseStdioFile(fGenes);
	fGenes = 0;

	vector<GFFRecord2> NCCEs;
	IvTree<LoHi> IntervalTree;
	ProgressStart("Reading NCCEs");
	RecordCount = 0;
	FILE *fNCCEs = OpenStdioFile(NCCEFileName);
	FileSize = (unsigned) GetStdioFileSize(fNCCEs);
	while (GetNextGFFRecord(fNCCEs, Rec))
		{
		if ((RecordCount++)%1000 == 0)
			{
			unsigned Pos = (unsigned) GetStdioFilePos(fNCCEs);
			ProgressStep(Pos, FileSize);
			}
		const string Name(Rec.SeqName);
		if (g_SeqName == "")
			g_SeqName = Name;
		else if (Name != g_SeqName)
			Die("NCCE file has more than one sequence name '%s', '%s'",
			  g_SeqName.c_str(), Name.c_str());
		if (Rec.Feature != "NCCE")
			Die("NCCE file has record with feature '%s'", Rec.Feature.c_str());

		gpos_t lo = Rec.Start - 1;
		gpos_t hi = Rec.End - 1;
		unsigned RecIndex = SIZE(NCCEs);
		Rec.Feature = "NGE";
		NCCEs.push_back(Rec);

		LoHi LH;
		LH.Lo = lo;
		LH.Hi = hi;
		LH.Index = RecIndex;
		IntervalTree.Add(LH);
		}
	ProgressDone();
	CloseStdioFile(fNCCEs);
	fNCCEs = 0;

	vector<GeneData> GeneVec;
	const unsigned GeneCount = SIZE(g_Genes);
	GeneVec.reserve(GeneCount);
	for (map<unsigned, GeneData>::const_iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		const GeneData &Gene = p->second;
		GeneVec.push_back(Gene);
		}

	if (Dump)
		{
		Log("\n");
		Log("Inter-gene regions:\n");
		Log("Gene1  Gene2  InterGeneLo  InterGeneHi      Length   Fract      Margin\n");
		Log("=====  =====  ===========  ===========  ==========  ======  ==========\n");
		}

	sort(GeneVec.begin(), GeneVec.end());
	gpos_t TotalMargin = 0;
	double TotalFract = 0;
	double TotalFract2 = 0;
	for (unsigned i = 1; i < GeneCount; ++i)
		{
		GeneData &Gene1 = GeneVec[i-1];
		GeneData &Gene2 = GeneVec[i];
		if (Gene1.hi + 1 == Gene2.lo)
			{
			Gene1.ExtendedHi = Gene1.hi;
			Gene2.ExtendedLo = Gene2.lo;
			continue;
			}
		gpos_t InterGeneLo = Gene1.hi + 1;
		gpos_t InterGeneHi = Gene2.lo - 1;
		if (InterGeneLo > InterGeneHi)
			{
			Log("\n");
			Log("Genes %u %u\n", i-1, i);
			Log("Gene1: ");
			Gene1.LogMe(true);
			Log("Gene2: ");
			Gene2.LogMe(true);
			Die("InterGene Lo > Hi");
			}

		gpos_t InterGeneLength = InterGeneHi - InterGeneLo + 1;
		if (InterGeneLength > P_MaxGeneMargin)
			InterGeneLength = P_MaxGeneMargin;

		double Fract = GetRandomFract();
		TotalFract += Fract;
		TotalFract2 += Fract*Fract;

		gpos_t Margin = (gpos_t) (InterGeneLength*Fract/2.0);
		Gene1.ExtendedHi = Gene1.hi + Margin;
		Gene2.ExtendedLo = Gene2.lo - Margin;
		asserta(Gene1.hi < Gene2.lo);
		TotalMargin += Margin;
		if (Dump)
			{
			Log("%5u  %5u  %11"PF"  %11"PF"  %10"PF"  %6.4f  %10"PF"\n",
			  Gene1.Index,
			  Gene2.Index,
			  InterGeneLo,
			  InterGeneHi,
			  InterGeneLength,
			  Fract,
			  Margin);
			}
		}
	if (Dump)
		Log("\n");

	if (GeneCount > 1)
		{
		double MeanFract = TotalFract/(GeneCount - 1);
		double MeanFract2 = TotalFract2/(GeneCount - 1);
		double StdDev = sqrt(MeanFract2 - MeanFract*MeanFract);
		Log("\n");
		Log("Mean inter-gene distance fract %.3f (target %.3f), std dev %.3f (target %.3f)\n",
		  MeanFract,
		  P_MeanInterGeneFract,
		  StdDev,
		  P_StdDevInterGeneFract);
		Log("\n");

		gpos_t AvgMargin = TotalMargin/(GeneCount - 1);
		GeneData &FirstGene = GeneVec.front();
		GeneData &LastGene = GeneVec.back();

		FirstGene.ExtendedLo = FirstGene.lo;
		LastGene.ExtendedHi = LastGene.hi;

		if (FirstGene.lo > AvgMargin)
			FirstGene.ExtendedLo -= AvgMargin;

		LastGene.ExtendedHi = LastGene.hi + AvgMargin;
		if (LastGene.ExtendedHi >= g_SeqLength)
			LastGene.ExtendedHi = g_SeqLength - 1;
		}

	if (Dump)
		{
		Log("\n");
		Log("Interval index:\n");
		Log("        lo          hi          ix\n");
		Log("==========  ==========  ==========\n");
		IntervalTree.Traverse(LogLoHi);
		Log("\n");
		}

	Progress("Assigning NCCEs to genes");
	for (unsigned i = 0; i < GeneCount; ++i)
		{
		const GeneData &Gene = GeneVec[i];

		gpos_t lo = Gene.ExtendedLo;
		gpos_t hi = Gene.ExtendedHi;

		vector<LoHi> Results;
		IntervalTree.GetOverlaps(pair<gpos_t, gpos_t>(lo, hi), Results);

		const unsigned N = SIZE(Results);
		for (unsigned j = 0; j < N; ++j)
			{
			const LoHi &LH = Results[j];
			unsigned RecIndex = LH.Index;
			asserta(RecIndex < SIZE(NCCEs));
			GFFRecord2 &Rec = NCCEs[RecIndex];
			string &Attrs = Rec.Attrs;
			string Value;
			if (GFFGetAttr(Attrs, "gene_index", Value))
				continue;

			Rec.Feature = "NXE";
			if (!Rec.Attrs.empty() && Rec.Attrs[SIZE(Rec.Attrs) - 1] != ';')
				Rec.Attrs += "; ";

			char Str[32];
			sprintf(Str, "gene_index %u;", Gene.Index);
			Rec.Attrs += Str;
			}
		}

	Progress("Writing to %s", OutFileName);
	FILE *fOut = CreateStdioFile(OutFileName);
	unsigned RecCount = SIZE(NCCEs);
	for (unsigned i = 0; i < RecCount; ++i)
		WriteGFFRec(fOut, NCCEs[i]);
	CloseStdioFile(fOut);

	LogGenes();
	}
