#include "utils.h"

#define	TEST	0

/***
ComputeBasesAffected computes the number of bases affected
by a rate range. Rates for intermediate lengths are computed
by linear interpolation. 

"Bases affected" means the average number of affected bases
per genome base for a branch length of 1.0. For example, a value
of 0.1 means that on average 10% of bases will be affected by
the event. The estimate ignores multiple mutations per site,
so values close to or larger than one are less meaningful.

To derive the obscure formula, start with the sum:

	BasesAffected = sum(i=0..R) L(i) f(i)

where L(i) is the i'th length and f(i) is the i'th frequency.
Define d = f_hi - f_lo, R  L_hi - Llo and substitute:

	L(i) = L_lo + i
	f(i) = f_lo + id/R

Use sum(i=1..N) i = N(N + 1)/2 and sum(i=1..N) i^2 = N(N + 1)(2N + 1))/6.
***/

double ComputeBasesAffected(gpos_t Llo, gpos_t Lhi, double flo, double fhi)
	{
	asserta(Lhi >= Llo);
	double d = fhi - flo;
	gpos_t R = Lhi - Llo;
	return (R + 1)*(Llo*(flo + d/2) + R*flo/2 + (d/6)*(2*R + 1));
	}

#if	TEST
static double ComputeBasesAffected_BruteForce(gpos_t Llo, gpos_t Lhi,
  double flo, double fhi)
	{
	double Sum = Llo*flo;
	for (gpos_t L = Llo + 1; L <= Lhi; ++L)
		{
		double r = flo + (fhi - flo)*(L - Llo)/(Lhi - Llo);
		Sum += L*r;
		}
	return Sum;
	}

static void TestCBA(gpos_t Llo, gpos_t Lhi, double flo, double fhi)
	{
	printf("%10.4g  %10.4g\n",
	  ComputeBasesAffected(Llo, Lhi, flo, fhi),
	  ComputeBasesAffected_BruteForce(Llo, Lhi, flo, fhi));
	}

void TestCBA()
	{
	printf("       CBA       Brute\n");
	printf("==========  ==========\n");
	TestCBA(10, 20, 0.1, 0.01);
	TestCBA(10, 100, 0.01, 0.1);
	TestCBA(10, 1000, 1e-4, 1e-6);
	}
#endif // TEST
