#include "revfile.h"
#include "utils.h"
#include "getopts.h"

#define TRACE	0

typedef bool (*ptrLinkFunction)(const AlnInfo &AI1, const AlnInfo &AI2);
typedef unsigned (*ptrDistanceFunction)(const AlnInfo &AI1, const AlnInfo &AI2);

void MakeEdges(const AIVec &AIs, ptrLinkFunction Link, unsigned GenomeIndex,
  unsigned MaxDist, IntVecVec &Edges, bool ShowProgress);

void Transpose(const IntVecVec &Edges, IntVecVec &EdgesT);

void GetConnectedComponents(const IntVecVec &AdjMx_, IntVecVec &CCs,
  bool ShowProgress);

void Chain(const AIVec &InAIs, unsigned GenomeIndex, ptrLinkFunction Link,
  unsigned MaxWidth, ptrDistanceFunction Distance, IntVecVec &Chains)
	{
#if	TRACE
	Log("\n");
	Log("Chain, AIs=\n");
	LogAIs(InAIs);
#endif
	const unsigned AICount = (unsigned) InAIs.size();

	IntVecVec Edges;
	MakeEdges(InAIs, Link, GenomeIndex, MaxWidth, Edges, false);

#if	TRACE
	{
	Log("\n");
	Log("After MakeEdges, Edges=\n");
	LogIntVecVec(Edges);
	}
#endif

// Create graph in ChainEdges which is first attempt to join up microchains.
// For each node, create an edge to its closest acceptable target (there may
// be none). 
// This graph has all out-degrees <= 1, but may have in-degrees > 1.
#if	TRACE
	Log("\n");
	Log("Make ChainEdges\n");
#endif
	IntVecVec ChainEdges;
	ChainEdges.resize(AICount);
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const IntVec &Targets = Edges[AIIndex];
#if	TRACE
		Log("AI%u, %u targets\n", AIIndex, (unsigned) Targets.size());
#endif
		if (Targets.size() == 0)
			continue;

	// Keep shortest edge, delete all others
		const unsigned BIG = 1000000;
		unsigned ShortestIndex = BIG;
		unsigned ShortestDistance = BIG;
		const AlnInfo &AI = InAIs[AIIndex];
		for_CIntVec(Targets, p)
			{
			unsigned TargetIndex = *p;
			const AlnInfo &Target = InAIs[TargetIndex];

			unsigned d = Distance(AI, Target);
			if (d < ShortestDistance)
				{
				ShortestIndex = TargetIndex;
				ShortestDistance = d;
				}
			}
		if (ShortestIndex != BIG)
			{
#if	TRACE
			Log("  Add edge %u->%u (shortest distance=%u)\n",
			  AIIndex, ShortestIndex, ShortestDistance);
#endif
			ChainEdges[AIIndex].push_back(ShortestIndex);
			}
#if	TRACE
		else
			{
			Log("  Shortest distance = BIG\n");
			}
#endif
		}

// Create transpose. Out-edges in transpose are in-edges in ChainEdges.
// For each node that has out-degree > 1 in ChainEdgesT (i.e., in-degree
// > 1 in ChainEdges), choose neighbor with shortest distance.
	IntVecVec ChainEdgesT;
	Transpose(ChainEdges, ChainEdgesT);
	assert(ChainEdgesT.size() == AICount);
#if	TRACE
	Log("\n");
	Log("ChainEdges Transpose=\n");
	LogIntVecVec(ChainEdgesT);
#endif

// ChainEdges2 is the final microchain graph.
// A microchain is a connected component of this graph.
#if	TRACE
	Log("\n");
	Log("Build ChainEdges2\n");
#endif
	IntVecVec ChainEdges2;
	ChainEdges2.resize(AICount);
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const IntVec &L = ChainEdgesT[AIIndex];
		unsigned ListLength = (unsigned) L.size();
		if (ListLength == 0)
			continue;
		else if (ListLength == 1)
			{
			unsigned SourceIndex = L.front();
#if	TRACE
			Log("List length one, add edge %u->%u\n", SourceIndex, AIIndex);
#endif
			ChainEdges2[SourceIndex].push_back(AIIndex);
			continue;
			}
		const unsigned BIG = 1000000;
		unsigned ShortestIndex = BIG;
		unsigned ShortestDistance = BIG;
		const AlnInfo &AI = InAIs[AIIndex];
		const IntVec&Sources = ChainEdgesT[AIIndex];
		for_CIntVec(Sources, p)
			{
			unsigned SourceIndex = *p;
			const AlnInfo &Source = InAIs[SourceIndex];

			unsigned d = Distance(Source, AI);
			if (d < ShortestDistance)
				{
				ShortestIndex = SourceIndex;
				ShortestDistance = d;
				}
			}
		assert(ShortestIndex != BIG);
		assert(ShortestIndex < AICount);
#if	TRACE
			Log("List length >1 , add edge %u->%u (shortest distance = %u)\n",
			  ShortestIndex, AIIndex, ShortestDistance);
#endif
		ChainEdges2[ShortestIndex].push_back(AIIndex);
		}
#if	TRACE
	Log("\n");
	Log("ChainEdges2=\n");
	LogIntVecVec(ChainEdges2);
#endif

	GetConnectedComponents(ChainEdges2, Chains, false);

#if	TRACE
	Log("\n");
	Log("Chains=\n");
	LogIntVecVec(Chains);
#endif
	}
