#include "utils.h"
#include "fastastr.h"
#include "getopts.h"
#include "progress.h"
#include "xlat.h"
#include <ctype.h>

static const unsigned CTDim = CHAR_TYPE_Max+1;

static uint64 g_CharTypeCounts[CTDim][CTDim];
static uint64 g_CharTypeCounts1[CTDim];
static uint64 g_CharTypeCounts2[CTDim];

static bool g_Dump = false;
const unsigned g_Context = 8;

static const char *h(uint64 i)
	{
	return atoh(uint64tostr(i));
	}

static void CompareSeqs(const char *Name, std::vector<unsigned char> Seq1,
  std::vector<unsigned char> Seq2, ALPHA Alpha)
	{
	unsigned SeqLength = (unsigned) Seq1.size();
	if (SeqLength != Seq2.size())
		Die("Seq lengths differ %u, %s, %s",
		  SeqLength, (unsigned) Seq2.size(), Name);

	for (unsigned Pos = 0; Pos < SeqLength; ++Pos)
		{
		unsigned char c1 = Seq1[Pos];
		unsigned char c2 = Seq2[Pos];

		if (isalpha(c1) && isalpha(c2) && tolower(c1) != tolower(c2))
			Die("Difference letter at position %u '%c' '%c' %s", Pos, c1, c2, Name);

		CHAR_TYPE t1 = CharType(c1, Alpha);
		CHAR_TYPE t2 = CharType(c2, Alpha);

		if (g_Dump && (t1 != t2))
			{
			Log("%10u ", Pos);
			int iPos = (int) Pos;
			int iStart = iPos - g_Context/2;
			int iEnd = iPos + g_Context/2;
			if (iStart < 0)
				iStart = 0;
			if (iEnd >= (int) SeqLength)
				iEnd = SeqLength  - 1;
			for (int i = iStart; i <= iEnd; ++i)
				{
				if (i - iStart == (int) g_Context/2)
					Log(" ");
				Log("%c", Seq1[i]);
				if (i - iStart == (int) g_Context/2)
					Log(" ");
				}
			Log("  ");
			for (int i = iStart; i <= iEnd; ++i)
				{
				if (i - iStart == (int) g_Context/2)
					Log(" ");
				Log("%c", Seq2[i]);
				if (i - iStart == (int) g_Context/2)
					Log(" ");
				}
			Log("\n");
			}

		++(g_CharTypeCounts[t1][t2]);
		++(g_CharTypeCounts1[t1]);
		++(g_CharTypeCounts2[t2]);
		}
	}

void CmpMask()
	{
	const char *InFileName1 = RequiredValueOpt("in1");
	const char *InFileName2 = RequiredValueOpt("in2");
	g_Dump = FlagOpt("dump");
	bool Amino = FlagOpt("aminoseq");

	ALPHA Alpha = (Amino ? ALPHA_Amino : ALPHA_Nucleo);

	FastaStream FS1;
	FastaStream FS2;
	FS1.Open(InFileName1);
	FS2.Open(InFileName2);

	std::vector<char> Label1;
	std::vector<char> Label2;
	std::vector<unsigned char> Seq1;
	std::vector<unsigned char> Seq2;

	ProgressStart("Comparing sequence masking");
	unsigned Steps = (unsigned) (FS1.GetFileSize()/1000000 + 1);
	unsigned PrevStep = UINT_MAX;
	for (;;)
		{
		unsigned Step = (unsigned) (FS1.GetFilePos()/1000000);
		if (Step != PrevStep)
			{
			ProgressStep(Step, Steps);
			PrevStep = Step;
			}

		bool Ok1 = FS1.GetNextSeq(Label1, Seq1);
		bool Ok2 = FS2.GetNextSeq(Label2, Seq2);
		if (Ok1 != Ok2)
			Die("Different number of sequences");
		if (!Ok1 && !Ok2)
			break;
		assert(Ok1 && Ok2);

		const char *Name1 = (const char *) &(Label1.front());
		const char *Name2 = (const char *) &(Label2.front());
		unsigned Len1 = (unsigned) Label1.size();
		unsigned Len2 = (unsigned) Label2.size();
		if (Label1 != Label2)
			Die("Labels differ: '%.*s', '%.*s'", Len1, Name1, Len2, Name2);

		CompareSeqs(Name1, Seq1, Seq2, Alpha);
		}
	ProgressDone();

	int Len1 = (int) strlen(InFileName1);
	int Len2 = (int) strlen(InFileName2);
	Len1 = (Len1 > 8 ? Len1 : 8);
	Len2 = (Len2 > 8 ? Len2 : 8);
	int MaxLen = (Len1 > Len2 ? Len1 : Len2);

	ProgressNoRes("");
	ProgressNoRes("%*.*s", MaxLen, MaxLen, InFileName1);
	uint64 Total1 = 0;
	for (int t = 0; t <= CHAR_TYPE_Max; ++t)
		{
		if (g_CharTypeCounts1[t] == 0)
			continue;
		ProgressNoRes("%*.*s  %16s",
		  MaxLen, MaxLen, CharTypeToStr((CHAR_TYPE) t),
		  h(g_CharTypeCounts1[t]));
		Total1 += g_CharTypeCounts1[t];
		}
	ProgressNoRes("%*.*s  %16s", MaxLen, MaxLen, "TOTAL", h(Total1));

	ProgressNoRes("");
	ProgressNoRes("");
	ProgressNoRes("%*.*s", MaxLen, MaxLen, InFileName2);
	uint64 Total2 = 0;
	for (int t = 0; t <= CHAR_TYPE_Max; ++t)
		{
		if (g_CharTypeCounts2[t] == 0)
			continue;
		ProgressNoRes("%*.*s  %16s",
		  MaxLen, MaxLen, CharTypeToStr((CHAR_TYPE) t),
		  h(g_CharTypeCounts2[t]));
		Total2 += g_CharTypeCounts2[t];
		}
	ProgressNoRes("%*.*s  %16s", MaxLen, MaxLen, "TOTAL", h(Total2));

	ProgressNoRes("");
	ProgressNoRes("");
	uint64 AgreeCount = 0;

	ProgressNoRes("");
	ProgressNoRes("%.*s  %.*s", Len1, InFileName1, Len2, InFileName2);
	for (int t1 = 0; t1 <= CHAR_TYPE_Max; ++t1)
		{
		if (g_CharTypeCounts[t1][t1] == 0)
			continue;
		ProgressNoRes("%*.*s  %*.*s  %s",
		  Len1, Len1, CharTypeToStr((CHAR_TYPE) t1),
		  Len2, Len2, CharTypeToStr((CHAR_TYPE) t1),
		  h(g_CharTypeCounts[t1][t1]));
		AgreeCount += g_CharTypeCounts[t1][t1];
		}

	uint64 DisagreeCount = 0;
	for (int t1 = 0; t1 <= CHAR_TYPE_Max; ++t1)
		{
		for (int t2 = 0; t2 <= CHAR_TYPE_Max; ++t2)
			{
			if (t1 == t2)
				continue;
			if (g_CharTypeCounts[t1][t2] == 0)
				continue;
			ProgressNoRes("%*.*s  %*.*s  %s",
			  Len1, Len1, CharTypeToStr((CHAR_TYPE) t1),
			  Len2, Len2, CharTypeToStr((CHAR_TYPE) t2),
			  h(g_CharTypeCounts[t1][t2]));
			DisagreeCount += g_CharTypeCounts[t1][t2];
			}
		}
	ProgressNoRes("");
	ProgressNoRes("");
	ProgressNoRes("Agree    %s", h(AgreeCount));
	ProgressNoRes("Disagree %s", h(DisagreeCount));
	}
