#include "revfile.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"

char CompLetter(char c);

static map<pair<unsigned, unsigned>, pair<unsigned, unsigned> > g_GenChrToLoHiA;
static map<pair<unsigned, unsigned>, pair<unsigned, unsigned> > g_GenChrToLoHiB;
static map<pair<unsigned, unsigned>, string> g_GenChrToSeqA;
static map<pair<unsigned, unsigned>, string> g_GenChrToSeqB;
static unsigned g_ErrCount;

static void InitSeqs(const RevFile &Rev,
  const map<pair<unsigned, unsigned>, pair<unsigned, unsigned> > &GenChrToLoHi,
  map<pair<unsigned, unsigned>, string> &GenChrToSeq)
	{
	Log("\n");
	Log("  Gen    Chr          Lo          Hi         Len  Label\n");
	Log("-----  -----  ----------  ----------  ----------  -----\n");
	GenChrToSeq.clear();
	for (map<pair<unsigned, unsigned>, pair<unsigned, unsigned> >::const_iterator p = GenChrToLoHi.begin();
	  p != GenChrToLoHi.end(); ++p)
		{
		const pair<unsigned, unsigned> &GenChr = p->first;
		const pair<unsigned, unsigned> &LoHi = p->second;
		unsigned L = LoHi.second - LoHi.first + 1;

		Log("%5u  %5u  %10u  %10u  %10u  %s.%s\n",
		  GenChr.first,
		  GenChr.second,
		  LoHi.first,
		  LoHi.second,
		  L,
		  Rev.GetGenomeName(GenChr.first).c_str(),
		  Rev.GetChrName(GenChr.first, GenChr.second).c_str());

		GenChrToSeq[GenChr] = string("");
		GenChrToSeq[GenChr].resize(L, '?');
		}
	}

static void UpdateLoHis(unsigned GenomeIndex, unsigned ChrId, unsigned Lo, unsigned Hi,
  map<pair<unsigned, unsigned>, pair<unsigned, unsigned> > &GenChrToLoHi)
	{
	pair<unsigned, unsigned> GenChr(GenomeIndex, ChrId);
	pair<unsigned, unsigned> LoHi(Lo, Hi);

	map<pair<unsigned, unsigned>, pair<unsigned, unsigned> >::iterator p =
	  GenChrToLoHi.find(GenChr);
	if (p == GenChrToLoHi.end())
		GenChrToLoHi[GenChr] = LoHi;
	else
		{
		pair<unsigned, unsigned> &LoHi2 = p->second;
		LoHi2.first = min(LoHi2.first, Lo);
		LoHi2.second = max(LoHi2.second, Hi);
		}
	}

static void SetLoHis(const vector<AlnInfo> &AIs,
  map<pair<unsigned, unsigned>, pair<unsigned, unsigned> > &GenChrToLoHi)
	{
	GenChrToLoHi.clear();
	const unsigned AICount = SIZE(AIs);
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const AlnInfo &AI = AIs[AIIndex];
		const vector<GCSeg> &Segs = AI.Segs;
		const unsigned SegCount = SIZE(Segs);
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = Segs[SegIndex];
			UpdateLoHis(Seg.GenomeIndex, Seg.ChrId, Seg.lo, Seg.hi, GenChrToLoHi);
			}
		}
	}

static void UpdateSeq(const GCSeg &Seg, string &Seq, unsigned Lo,
  const string &Row)
	{
	bool Plus = Seg.Plus;
	unsigned Start = Seg.Plus ? Seg.lo : Seg.hi;
	unsigned Pos = Start - Lo;
	const unsigned L = SIZE(Row);
	for (unsigned i = 0; i < L; ++i)
		{
		char c = toupper(Row[i]);
		if (c == '-')
			continue;
		if (!Plus)
			c = CompLetter(c);
		Log("Gen=%u Chr=%u Lo=%u Start=%u Pos=%u c=%c Plus=%c Seq=%c\n",
		  Seg.GenomeIndex, Seg.ChrId, Lo, Start, Pos, c, pom(Plus), Seq[Pos]);//@@
		asserta(Pos < SIZE(Seq));
		if (Seq[Pos] == '?')
			Seq[Pos] = c;
		else
			{
			if (Seq[Pos] != c)
				{
				++g_ErrCount;
				Log("** Gen=%u Chr=%u Lo=%u Start=%u Pos=%u c=%c Plus=%c Seq=%c\n",
				  Seg.GenomeIndex, Seg.ChrId, Lo, Start, Pos, c, pom(Plus), Seq[Pos]);
				}
			}
		if (Plus)
			++Pos;
		else
			--Pos;
		}
	}

static void SetSeqs(const RevFile &Rev,
  const map<pair<unsigned, unsigned>, pair<unsigned, unsigned> > &GenChrToLoHi,
  map<pair<unsigned, unsigned>, string> &GenChrToSeq)
	{
	const unsigned AICount = Rev.GetAlnInfoCount();
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		AlnInfo AI;
		Rev.GetAlnInfo(AIIndex, AI);

		unsigned AlnIndex = Rev.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			continue;

		StdStrVec Rows;
		Rev.GetAlnRows(AlnIndex, Rows);

		const unsigned SegCount = SIZE(AI.Segs);
		asserta(SIZE(Rows) == SegCount);
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = AI.Segs[SegIndex];
			const string &Row = Rows[SegIndex];
			pair<unsigned, unsigned> GenChr(Seg.GenomeIndex, Seg.ChrId);
			map<pair<unsigned, unsigned>, string>::iterator p = 
			  GenChrToSeq.find(GenChr);
			map<pair<unsigned, unsigned>, pair<unsigned, unsigned> >::const_iterator q = 
			  GenChrToLoHi.find(GenChr);
			asserta(q != GenChrToLoHi.end());
			string &Seq = p->second;
			unsigned Lo = q->second.first;
			unsigned Hi = q->second.second;
			asserta(Seg.lo >= Lo);
			asserta(Lo + Seg.GetLength() - 1 <= Hi);
			UpdateSeq(Seg, Seq, Lo, Row);
			}
		}
	}

void CmpRows()
	{
	const char *InFileNameA = RequiredValueOpt("in1");
	const char *InFileNameB = RequiredValueOpt("in2");

	RevFile InA;
	RevFile InB;

	InA.Open(InFileNameA);
	InB.Open(InFileNameB);

	vector<AlnInfo> AIs;
	InA.GetAllAlnInfos(AIs);
	SetLoHis(AIs, g_GenChrToLoHiA);

	InB.GetAllAlnInfos(AIs);
	SetLoHis(AIs, g_GenChrToLoHiB);

	InitSeqs(InA, g_GenChrToLoHiA, g_GenChrToSeqA);
	InitSeqs(InB, g_GenChrToLoHiB, g_GenChrToSeqB);

	SetSeqs(InA, g_GenChrToLoHiA, g_GenChrToSeqA);
	SetSeqs(InB, g_GenChrToLoHiB, g_GenChrToSeqB);
	}
