#include "utils.h"
#include "getopts.h"
#include "revfile.h"
#include "progress.h"

void CvtRev2MAF();
void CvtRev2MFA();
void CvtRev2GFF();
void CvtRev2FASTA();
void CvtRevGFF2FASTA();
void CvtRev2ASA();

void CvtBlast2Rev(bool UseLastZFormat);
void CvtGFF2Rev();
void CvtFASTA2Rev();
void CvtMAF2Rev();
void CvtBMA2Rev();
void CvtMFA2Rev();

void MergeRev();
void MergeRev2();
void SplitRev();
void CmpRev();
void CmpRows();
void MergeRevSeqs();
void StripParalogs();
void Project();
void MStats();

void XRange();
void XChrs();
void XAIs();
void XGFFAlns();
void XGFFSeqs();
void XSubSeq();

void ValidateAIs();
void ValidateAlns();
void ValidateSegLengths();

void DumpRev();
void DumpGenIds();
void DumpChrIds();
void DumpActiveChrPairs();
void DumpAI();
void DumpRevHdr();

void DiffChrSeq();
void GFFCmp();
void GFFCover();

void BundleScores();
void ShowGenomeSizes();
void ShowChrSizes();
void GenStats();

static void ShowCommands()
	{
	printf("\n");
	printf("File format conversions from .rev:\n");
	printf("   -fromrev <revfile> -toasa <asafile> -genome <name> [-minlength <L>]\n");
	printf("   -fromrev <revfile> -togff <gfffile> [-useids] [-minlength L]\n");
	printf("   -fromrev <revfile> -tofasta <fastafile> [-gff <gfffile> -genome <genome>]\n");
	printf("   -fromrev <revfile> -tomfa <fastafile> [-gendotid] [-decrsize|-incrsize]\n");
	printf("   -fromrev <revfile> -tomaf <maffile> [-nocmt] [-useids] [-trunclabel] [-usechrids] [-nopairwise]\n"
	"               [-maf_cmdline <x>] [-maf_scoring <x>] [-maf_program <x>]\n");
	printf("\n");
	printf("File format conversions to .rev:\n");
	printf("   -fromblast <blastfile> -torev <revfile> -seq1 <refile> -seq2 <revfile> [-minpct <p>] [-minlength <n>] [-gendotid]\n");
	printf("   -fromlastz <blastfile> -torev <revfile> -seq1 <refile> -seq2 <revfile> [-minpct <p>] [-minlength <n>] [-gendotid]\n");
	printf("   -frombma <bmafile> -torev <revfile>\n");
	printf("   -fromfasta <fastafile> -torev <revfile>\n");
	printf("   -fromfastafiles <textfile> -torev <revfile> -genome <genomename>\n");
	printf("       <textfile> has one fasta file name per line.\n");
	printf("   -fromgff <gfffile> -torev <revfile> [-gendotid]\n");
	printf("   -frommaf <maffile> -torev <revfile>\n");
	printf("\n");
	printf("Verify alignment coordinates are within chromosome sizes:\n");
	printf("    -valais <revfile>\n");
	printf("\n");
	printf("Verify alignment letters agree with sequence file(s), default maxerrs=10:\n");
	printf("    -valalns <revfile> -seqfiles <seq1>,<seq2>... [-maxerrs <n>]\n");
	printf("\n");
	printf("Verify AlnInfo segment lengths agree with alignment row lengths, default maxerrs=10:\n");
	printf("    -valseglengths <revfile> [-maxerrs <n>]\n");
	printf("\n");
	printf("Dump .rev file information to log file:\n");
	printf("    -dumprev <revfile>               Disk file data structures\n");
	printf("    -dumprevhdr <revfile>            File header\n");
	printf("    -dumpgenids <revfile>            Genome indexes and names\n");
	printf("    -dumpchrids <revfile>            Chromosome indexes and names\n");
	printf("    -dumpacp <revfile>               Active chromosome pairs\n");
	printf("    -dumpai <revfile> -aiindex <i>   AlnInfo (and rows, if present)\n");
	printf("\n");
	printf("Merge / split .rev files containing alignments:\n");
	printf("    -mergerev <n> -torev <revfile> [-prefix <p>] [-suffix <s>]\n");
	printf("    -splitrev <revfile> -n <n> [-prefix <p>] [-suffix <s>]\n");
	printf("        Files are named <p><i><n>, <i> = 1, 2 ... <n>, must have same genomes\n");
	printf("    -mergerev2 <revfile1>,<revfile2>,...,<revfileN> -out <revfile>\n");
	printf("        Files may have different sets of genomes\n");
	printf("\n");
	printf("Merge .rev files containing sequences:\n");
	printf("    -mergerevseqs <revfile1>,<revfile2>,...,<revfileN> -out <revfile>\n");
	printf("\n");
	printf("Extract subset of .rev file, if -trunc then truncate alns to given range:\n");
	printf("    -xrange <revfile> -out <revfile> [-trunc] [-logaiindexes] [-and]\n");
	printf("        [-genid <g> -chrid <c> -lo <l> -hi <h>] | [-rangefile <r>]\n");
	printf("        Lines in range file are: genid chrid lo hi.\n");
	printf("    -xchrs <revfile> -out <revfile> -chrfile <c>\n");
	printf("        Lines in chr file are: genix0 chrid0 genix1 chrid1.\n");
	printf("    -xais <revfile> -out <revfile> -indexes i,j,k... \n");
	printf("        Copy AIs+alignments (if present) with given AI indexes\n");
	printf("    -xgffalns <revfile> -out <revfile> -gff <gfffile>\n");
	printf("        Copy alignment segments that overlap GFF records to output file (always truncates).\n");
	printf("    -xsubseq <revfile> -out <fastafile> -genome <gen> -chr <chr> [-lo <lo>] [-hi <hi>] [-length <len>] [-revcomp] [-label <lbl>]\n");
	printf("        Copy subsequence to FASTA file. Coordinates are 0-indexed.\n");
	printf("    -xgffseqs <revfile> -gff <gfffile> -genome <name> -out <fastafile>\n");
	printf("        Copy subsequences given by GFF records to FASTA file\n");
	printf("    -stripparalogs <revfile> -out <revfile>\n");
	printf("        Copy rev file alignments, discarding those with paralogy attribute != 0.\n");
	printf("\n");
	printf("Misc:\n");
	printf("    -diffchrseq <chrname> -genome <gen> -in1 <revfile> -in2 <revfile> [-ignorecase] [-nwild]\n");
	printf("        Report differences in stored sequence found in both rev files.\n");
	printf("    -gffcmp -in1 <gfffile> -in2 <gfffile> -out <gfffile>\n");
	printf("        Find union and symmetric difference of two GFF files\n");
	printf("    -gffcover <gfffile> -out <gfffile>\n");
	printf("        Output regions covered (eliminates overlaps).\n");
	printf("    -bundlescores <revfile> -out <revfile>\n");
	printf("    -showgenomesizes <revfile>\n");
	printf("    -showchrsizes <revfile> [-genid <i>]\n");
	}

extern bool g_ShowProgress;

int main(int argc, char *argv[])
	{
	if (argc == 1)
		{
		ShowCommands();
		return 0;
		}

	ProcessArgVect(argc, argv);

	g_ShowProgress = !FlagOpt("quiet");

	if (ValueOpt("fromrev") && ValueOpt("tomaf"))
		CvtRev2MAF();
	else if (ValueOpt("fromrev") && ValueOpt("tomfa"))
		CvtRev2MFA();
	else if (ValueOpt("fromrev") && ValueOpt("togff"))
		CvtRev2GFF();
	else if (ValueOpt("fromblast") && ValueOpt("torev"))
		CvtBlast2Rev(false);
	else if (ValueOpt("fromlastz") && ValueOpt("torev"))
		CvtBlast2Rev(true);
	else if (ValueOpt("fromgff") && ValueOpt("torev"))
		CvtGFF2Rev();
	else if (ValueOpt("frommaf") && ValueOpt("torev"))
		CvtMAF2Rev();
	else if (ValueOpt("frommfa") && ValueOpt("torev"))
		CvtMFA2Rev();
	else if (ValueOpt("frombma") && ValueOpt("torev"))
		CvtBMA2Rev();
	else if (ValueOpt("fromrev") && ValueOpt("toasa"))
		CvtRev2ASA();
	else if (ValueOpt("fromrev") && ValueOpt("tofasta"))
		{
		if (ValueOpt("gff"))
			CvtRevGFF2FASTA();
		else
			CvtRev2FASTA();
		}
	else if ((ValueOpt("fromfasta") || ValueOpt("fromfastafiles")) && ValueOpt("torev"))
		CvtFASTA2Rev();
	else if (ValueOpt("xrange"))
		XRange();
	else if (ValueOpt("xchrs"))
		XChrs();
	else if (ValueOpt("xais"))
		XAIs();
	else if (ValueOpt("valais"))
		ValidateAIs();
	else if (ValueOpt("valalns"))
		ValidateAlns();
	else if (ValueOpt("valseglens"))
		ValidateSegLengths();
	else if (ValueOpt("mergerev"))
		MergeRev();
	else if (ValueOpt("mergerev2"))
		MergeRev2();
	else if (ValueOpt("splitrev"))
		SplitRev();
	else if (ValueOpt("mergerevseqs"))
		MergeRevSeqs();
	else if (ValueOpt("dumprev"))
		DumpRev();
	else if (ValueOpt("dumprevhdr"))
		DumpRevHdr();
	else if (ValueOpt("dumpgenids"))
		DumpGenIds();
	else if (ValueOpt("dumpchrids"))
		DumpChrIds();
	else if (ValueOpt("dumpacp"))
		DumpActiveChrPairs();
	else if (ValueOpt("dumpai"))
		DumpAI();
	else if (ValueOpt("bundlescores"))
		BundleScores();
	else if (ValueOpt("showgenomesizes"))
		ShowGenomeSizes();
	else if (ValueOpt("showchrsizes"))
		ShowChrSizes();
	else if (ValueOpt("xgffalns"))
		XGFFAlns();
	else if (ValueOpt("xgffseqs"))
		XGFFSeqs();
	else if (ValueOpt("xsubseq"))
		XSubSeq();
	else if (FlagOpt("gffcmp"))
		GFFCmp();
	else if (ValueOpt("gffcover"))
		GFFCover();
	else if (ValueOpt("diffchrseq"))
		DiffChrSeq();
	else if (ValueOpt("genstats"))
		GenStats();
	else if (ValueOpt("stripparalogs"))
		StripParalogs();
	else if (ValueOpt("project"))
		Project();
	else if (FlagOpt("cmprev"))
		CmpRev();
	else if (FlagOpt("cmprows"))
		CmpRows();
	else if (ValueOpt("mstats"))
		MStats();
	else
		Die("Invalid or unsupported command line");

	ProgressExit();
	return 0;
	}
