#include "revfile.h"
#include "utils.h"
#include "getopts.h"

/***
BMA is Bob's Multiple Alignment file format.

It is designed to be convenient to write by hand for testing
revolver software.

BMA files are ASCII text.

There are four types of lines:

1. Blank, separate multiple alignments from each other.

2. Chromosome definition:

	GenomeName ChromosomeName Size

3. One segment from a multiple aligment:

	GenomeName ChromosomeName Lo Plus Row

4. Comments. Lines beginning with # are ignored.

Chromosome lines must come before multiple alignments.

Example.

DMel Chr1 22000000
DMel Chr2 21000000
DPse Chr1 25000000
DPse Chr2 23000000

DMel Chr1 123 + AGCTT-GGGT
DPse Chr1 789 + AGT-TTGGGT

DMel Chr2 123 + AAATTCCTCT
DPse Chr1 789 - ATATTCGTCT
***/

void Split(const std::string &Str, std::vector<std::string> &Fields);
void StripGaps(const std::string &In, std::string &Out);

static std::vector<GenInfo> g_Genomes;

static GenInfo &GetGenome(const StdStr &Name)
	{
	const unsigned GenCount = SIZE(g_Genomes);
	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		{
		GenInfo &GI = g_Genomes[GenIndex];
		if (GI.GetName() == Name)
			return GI;
		}
	GenInfo GI;
	GI.SetName(Name);
	g_Genomes.push_back(GI);
	return g_Genomes.back();
	}

static unsigned GetGenomeIndex(const StdStr &Name)
	{
	const unsigned GenCount = SIZE(g_Genomes);
	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		{
		GenInfo &GI = g_Genomes[GenIndex];
		if (GI.GetName() == Name)
			return GenIndex;
		}
	Die("Genome %s not found", Name.c_str());
	return UINT_MAX;
	}

unsigned GetChrId(const StdStr &GenName, const StdStr &ChrName)
	{
	const GenInfo &GI = GetGenome(GenName);
	return GI.GetChrId(ChrName);
	}

static void AddChr(const StdStr &GenName, const StdStr &ChrName, coord_t Length)
	{
	GenInfo &GI = GetGenome(GenName);
	GI.AddChr(ChrName, Length);
	}

static AlnInfo g_OutAI;
static StdStrVec g_OutRows;

static void AddRow(const StdStr &GenName, const StdStr &ChrName, coord_t lo,
  bool Plus, const StdStr &Row)
	{
	StdStr UngappedRow;
	StripGaps(Row, UngappedRow);

	GCSeg Seg;
	Seg.GenomeIndex = GetGenomeIndex(GenName);
	Seg.ChrId = GetChrId(GenName, ChrName);
	Seg.lo = lo;
	Seg.hi = lo + SIZE(UngappedRow) - 1;
	Seg.Plus = Plus;
	
	g_OutAI.Segs.push_back(Seg);
	g_OutRows.push_back(Row);
	}

static void EndAln(RevFile &Out)
	{
	Out.AddAln(g_OutAI, g_OutRows);

	g_OutAI.Clear();
	g_OutRows.clear();
	}

void CvtBMA2Rev()
	{
	const char *BMAFileName = RequiredValueOpt("frombma");
	const char *RevFileName = RequiredValueOpt("torev");

	FILE *fIn = OpenStdioFile(BMAFileName);

	const unsigned BUFF_SIZE = 64*1024;
	char Buffer[BUFF_SIZE];
	for (;;)
		{
		bool EndOfFile = ReadLineStdioFile(fIn, Buffer, sizeof(Buffer));
		if (EndOfFile)
			Die("No alignments in .bma file");

		if (Buffer[0] == '#')
			continue;

		StdStrVec Fields;
		Split(Buffer, Fields);

		if (SIZE(Fields) == 0)
			continue;

		if (SIZE(Fields) == 3)
			{
			const StdStr &GenName = Fields[0];
			const StdStr &ChrName = Fields[1];
			coord_t Length = StrToCoord(Fields[2].c_str());
			AddChr(GenName, ChrName, Length);
			}

		if (SIZE(Fields) == 5)
			break;
		}

	RevFile Out;
	Out.Create(RevFileName, g_Genomes);

	do
		{
		if (Buffer[0] == '#')
			continue;

		StdStrVec Fields;
		Split(Buffer, Fields);

		if (SIZE(Fields) == 0)
			{
			EndAln(Out);
			continue;
			}

		if (SIZE(Fields) != 5)
			Die("Expected 5 fields, got '%s'", Buffer);

		const StdStr &GenName = Fields[0];
		const StdStr &ChrName = Fields[1];
		coord_t lo = StrToCoord(Fields[2].c_str());
		bool Plus = false;
		if (Fields[3] == "+")
			Plus = true;
		else if (Fields[3] == "-")
			Plus = false;
		else
			Die("Invalid plus '%s' in line '%s'", Fields[3].c_str(), Buffer);
		const StdStr &Row = Fields[4];
		AddRow(GenName, ChrName, lo, Plus, Row);
		}
	while (!ReadLineStdioFile(fIn, Buffer, sizeof(Buffer)));
	EndAln(Out);
	}
