#include "revfile.h"
#include "utils.h"
#include "gff.h"
#include "getopts.h"
#include <string.h>

#define	TRACE	0

bool GetNextGFFRecord(FILE *f, GFFRecord &Rec);
bool HasTargetAttrs(const char *Attrs);

#if	TRACE
static void LogMap(const std::map<std::string, ChrInfo> &Chrs)
	{
	Log("             Key     Id      Length              Name\n");
	Log("================  =====  ==========  ================\n");
	for (std::map<std::string, ChrInfo>::const_iterator p = Chrs.begin();
	  p != Chrs.end(); ++p)
		{
		const std::string &Name = p->first;
		const ChrInfo &CI = p->second;
		Log("%16.16s  %5u  %10u  %16.16s\n",
		  Name.c_str(),
		  CI.GetId(),
		  (unsigned) CI.GetLength(),
		  CI.GetName().c_str());
		}
	}
#endif

static void AddToMap(std::map<std::string, ChrInfo> &Chrs, const char *ChrName,
  coord_t hi)
	{
	std::map<std::string, ChrInfo>::iterator p = Chrs.find(ChrName);
	if (p == Chrs.end())
		{
		ChrInfo CI;
		unsigned Id = (unsigned) Chrs.size();
		CI.Create(Id, ChrName, hi+1);
		std::map<std::string, ChrInfo>::value_type v(ChrName, CI);
		Chrs.insert(v);
		return;
		}
	ChrInfo &CI = p->second;
	assert(CI.GetName() == ChrName);
	if (hi > CI.GetLength())
		CI.SetLength(hi);
	}

static unsigned GetChrId(const std::map<std::string, ChrInfo> &Chrs, const char *Name)
	{
	std::map<std::string, ChrInfo>::const_iterator p = Chrs.find(Name);
	if (p == Chrs.end())
		Die("GetChrId: not found");
	const ChrInfo &CI = p->second;
	return CI.GetId();
	}

static void MapToVec(std::map<std::string, ChrInfo> &Chrs,
  std::vector<ChrInfo> &vChrs)
	{
	vChrs.clear();
	for (std::map<std::string, ChrInfo>::iterator p = Chrs.begin(); p != Chrs.end(); ++p)
		{
		const ChrInfo &CI = p->second;
		vChrs.push_back(CI);
		}
	}

const GenInfo &RetrieveGenInfo(RevFile &In, const char *GenomeName)
{
	if (GenomeName != 0)
		return In.GetGenomeInfo(In.GetGenomeIndex(GenomeName));
	else
	{
		if (In.GetGenomeCount() != 1)
			Die("Expecting revfile with one genome (otherwise use -gen1)");
		return In.GetGenomeInfo(0);
	}
}

void CvtGFF2Rev1D()
{
	const char *GFFFileName = RequiredValueOpt("fromgff");
	const char *RevFileName = RequiredValueOpt("torev");
	const char *Seq1FileName = RequiredValueOpt("seq1");

	/* Retrieve Genome Info from input revfile */
	RevFile Seq1Rev;
	Seq1Rev.Open(Seq1FileName);
	const GenInfo &Genome1 = RetrieveGenInfo(Seq1Rev, ValueOpt("gen1"));

	/* Open GFF File */
	FILE *fGFF = OpenStdioFile(GFFFileName);

	/* Create output Revfile */
	RevFile Out;
	std::vector<GenInfo> GV;
	GV.push_back(Genome1);
	Out.Create(RevFileName, GV);

	bool DoRelabel = FlagOpt("gendotid");

	GFFRecord Rec;
	while (GetNextGFFRecord(fGFF, Rec))
	{
		AlnInfo AI;
		AI.Segs.resize(1);
		AI.Segs[0].GenomeIndex = 0;
		if (DoRelabel)
		{
			char *ptrDot = (char *) strrchr(Rec.SeqName, '.');
			if ((ptrDot == 0) || (*++ptrDot == 0))
				Die("Invalid name %s", Rec.SeqName);
			AI.Segs[0].ChrId = strtoul(ptrDot, NULL, 10);
			Genome1.GetChrIndex(AI.Segs[0].ChrId); //Just make sure it's there!
		}
		else
			AI.Segs[0].ChrId = Genome1.GetChrId(Rec.SeqName);
		AI.Segs[0].lo = Rec.Start - 1;
		AI.Segs[0].hi = Rec.End - 1;
		AI.Segs[0].Plus = (Rec.Strand == '+');
		AI.Score = (ai_score_t) Rec.Score;
		Out.AddAlnInfo(AI);
	}

	Out.Close();
	Seq1Rev.Close();
}

void CvtGFF2Rev2D()
	{
	const char *GFFFileName = RequiredValueOpt("fromgff");
	const char *RevFileName = RequiredValueOpt("torev");
	const char *GenomeNameA = RequiredValueOpt("gen1");
	const char *GenomeNameB = RequiredValueOpt("gen2");

	FILE *fGFF = OpenStdioFile(GFFFileName);

// Pass 1: collect chromosome names & sizes
	std::map<std::string, ChrInfo> Chrs;
	std::map<std::string, ChrInfo> TargetChrs;

	GFFRecord Rec;
	while (GetNextGFFRecord(fGFF, Rec))
		{
		if (!HasTargetAttrs(Rec.Attrs))
			continue;

		const char *ChrName = Rec.SeqName;

		char TargetChrName[4096];
		coord_t TargetStart;
		coord_t TargetEnd;
		ParseTargetAttrs(Rec.Attrs, TargetChrName, sizeof(TargetChrName),
		  &TargetStart, &TargetEnd);

		AddToMap(Chrs, ChrName, Rec.End);
		AddToMap(TargetChrs, TargetChrName, TargetEnd);

#if	TRACE
		Log("\n");
		Log("After AddToMap, Chrs=\n");
		Log("\n");
		LogMap(Chrs);
		Log("TargetChrs=\n");
		LogMap(TargetChrs);
#endif
		}

// Build genomes vector from chromosome info collected in pass 1
	std::vector<ChrInfo> vChrs;
	std::vector<ChrInfo> vTargetChrs;

	MapToVec(Chrs, vChrs);
	MapToVec(TargetChrs, vTargetChrs);

	std::vector<GenInfo> Genomes;
	Genomes.resize(2);
	Genomes[0].Create(GenomeNameA, vChrs);
	Genomes[1].Create(GenomeNameB, vTargetChrs);

// Create output file
	RevFile Out;
	Out.Create(RevFileName, Genomes);

// Pass 2: create AlnInfo's from GFF records
	SeekStdioFile(fGFF, 0);
	while (GetNextGFFRecord(fGFF, Rec))
		{
		if (!HasTargetAttrs(Rec.Attrs))
			continue;

		char TargetChrName[4096];
		coord_t TargetStart;
		coord_t TargetEnd;
		ParseTargetAttrs(Rec.Attrs, TargetChrName, sizeof(TargetChrName),
		  &TargetStart, &TargetEnd);

		if (Rec.Start == 0 || Rec.End == 0 || Rec.Start > Rec.End)
			Die("Invalid coords in GFF record");

		if (TargetStart == 0 || TargetEnd == 0 || TargetStart > TargetEnd)
			Die("Invalid coords in GFF record");

		AlnInfo AI;
		AI.Segs.resize(2);

		AI.Score = (ai_score_t) Rec.Score;

		GCSeg &Seg = AI.Segs[0];
		GCSeg &TargetSeg = AI.Segs[1];

		unsigned ChrId = GetChrId(Chrs, Rec.SeqName);
		unsigned TargetChrId = GetChrId(TargetChrs, TargetChrName);

		Seg.GenomeIndex = 0;
		Seg.ChrId = ChrId;
		Seg.lo = Rec.Start - 1;
		Seg.hi = Rec.End - 1;
		Seg.Plus = true;

		TargetSeg.GenomeIndex = 1;
		TargetSeg.ChrId = TargetChrId;
		TargetSeg.lo = TargetStart - 1;
		TargetSeg.hi = TargetEnd - 1;
		TargetSeg.Plus = (Rec.Strand == '+');

		unsigned BundleId;
		bool BundleFound = ParseBundleAttrs(Rec.Attrs, &BundleId);
		if (BundleFound)
			AI.SetBundleIndex(BundleId);

		unsigned AlnLen;
		bool AlnLenFound = ParseBundleAttrs(Rec.Attrs, &AlnLen);
		if (AlnLenFound)
			AI.SetAttr(ALN_ATTR_AlnLength, AlnLen);

		Out.AddAlnInfo(AI);
		}

	Out.Close();
	}

void CvtGFF2Rev()
{
	if (ValueOpt("seq1"))
		CvtGFF2Rev1D();
	else
		CvtGFF2Rev2D();
}
