#include "utils.h"
#include "maffile.h"
#include "revfile.h"
#include "getopts.h"

const char *TruncLabel(const char *Label);

// Name is assumed to be either genome.chr, or species.
// If species, then species is used both as genome and
// chr name.
static void SplitSeqName(const StdStr &SeqName, StdStr &GenName, StdStr &ChrName)
	{
	std::string::size_type n = SeqName.find_last_of('.');
	if (n == std::string::npos || n == 0 || n + 1 == SIZE(SeqName))
		{
		GenName = SeqName;
		ChrName = SeqName;
		}
	else
		{
		ChrName = SeqName.substr(n+1);
		GenName = SeqName.substr(0, n);
		}
	}

static void CreateSeg(const std::vector<GenInfo> &Genomes,
  const std::map<StdStr, unsigned> &GenNameToGenIndex,
  const StdStr &SeqName, coord_t lo, coord_t hi, bool Plus, GCSeg &Seg)
	{
	StdStr GenName;
	StdStr ChrName;
	SplitSeqName(SeqName, GenName, ChrName);

	std::map<StdStr, unsigned>::const_iterator p = GenNameToGenIndex.find(GenName);
	asserta(p != GenNameToGenIndex.end());
	unsigned GenIndex = p->second;
	const GenInfo &GI = Genomes[GenIndex];
	unsigned ChrId = GI.GetChrId(ChrName);
	Seg.GenomeIndex = GenIndex;
	Seg.ChrId = ChrId;
	Seg.lo = lo;
	Seg.hi = hi;
	Seg.Plus = Plus;
	}

void CvtMAF2Rev()
	{
	const char *MAFFileName = RequiredValueOpt("frommaf");
	const char *RevFileName = RequiredValueOpt("torev");

	MAFFile MAF;
	MAF.Open(MAFFileName);

	std::map<StdStr, unsigned> GenNameToGenIndex;

// First pass through MAF file to build Genomes
	std::vector<GenInfo> Genomes;

	PosVec Los;
	PosVec His;
	BoolVec Pluses;
	StdStrVec Rows;
	StdStrVec SeqNames;
	PosVec SeqLengths;
	double Score;
	while (MAF.GetNextAln(Los, His, Pluses, Rows, SeqNames, SeqLengths, Score))
		{
		const unsigned RowCount = SIZE(SeqNames);
		FORIX(Row)
			{
			const StdStr &SeqName = SeqNames[RowIndex];

			StdStr GenName;
			StdStr ChrName;
			SplitSeqName(SeqName, GenName, ChrName);

			coord_t ChrLength = SeqLengths[RowIndex];

			std::map<StdStr, unsigned>::const_iterator p =
			  GenNameToGenIndex.find(GenName);

			unsigned GenIndex = UINT_MAX;
			if (p == GenNameToGenIndex.end())
				{
			// New genome
				GenInfo GI;
				GI.Create(GenName);
				GenIndex = SIZE(Genomes);
				Genomes.push_back(GI);
				GenNameToGenIndex[GenName] = GenIndex;
				}
			else
				GenIndex = p->second;

			GenInfo &GI = Genomes[GenIndex];
			const ChrInfo *ptrCI = GI.GetChrInfoPtr(ChrName);
			if (ptrCI == 0)
			// New chromosome
				GI.AddChr(ChrName, ChrLength);
			else
				{
				if (ptrCI->GetLength() != ChrLength)
					Die("Chr %s length differs %u, %u",
					  ChrName.c_str(),
					  (unsigned) ChrLength,
					  (unsigned) ptrCI->GetLength());
				}
			}
		}

	RevFile Rev;
	Rev.Create(RevFileName, Genomes);

// Second pass to copy alignments from MAF to Rev
	MAF.Rewind();
	while (MAF.GetNextAln(Los, His, Pluses, Rows, SeqNames, SeqLengths, Score))
		{
		AlnInfo AI;

		const unsigned SegCount = SIZE(Los);
		FORIX(Seg)
			{
			const StdStr &SeqName = SeqNames[SegIndex];
			coord_t lo = Los[SegIndex];
			coord_t hi = His[SegIndex];
			bool Plus = Pluses[SegIndex];

			GCSeg Seg;
			CreateSeg(Genomes, GenNameToGenIndex, SeqName, lo, hi, Plus, Seg);
			AI.Segs.push_back(Seg);
			}

		AI.Score = (float) Score;
		Rev.AddAln(AI, Rows);
		}

	Rev.Close();
	}
