#include "utils.h"
#include "revfile.h"
#include "getopts.h"

void TruncLabel(std::string &Label);
void WriteGFFRecord(FILE *f, const char *Label, const char *RecType,
  coord_t lo, coord_t hi, ai_score_t Score, char Strand, const char *Attrs);

static bool g_AIIndexes = false;
static bool g_UseIds = false;
static bool g_PctId = false;
static unsigned g_MinLength = 0;

static void Output(const RevFile &Rev, unsigned AlnInfoIndex, FILE *fGFF)
	{
	AlnInfo AI;
	Rev.GetAlnInfo(AlnInfoIndex, AI);

	const unsigned SegCount = (unsigned) AI.Segs.size();
	if (SegCount < 2)
		Die("SegCount < 2");
	if (g_PctId && SegCount != 2)
		Die("-pctid requires pairwise alignments");

	double PctId = -1;
	if (g_PctId)
		{
		unsigned AlnIndex = Rev.GetAlnIndex(AlnInfoIndex);
		if (AlnIndex != UINT_MAX)
			{
			StdStrVec Rows;
			Rev.GetAlnRows(AlnIndex, Rows);
			asserta(SIZE(Rows) == 2);
			const string &Row1 = Rows[0];
			const string &Row2 = Rows[1];
			const unsigned L = SIZE(Row1);
			asserta(SIZE(Row2) == L);
			unsigned Ids = 0;
			unsigned Pairs = 0;
			for (unsigned i = 0; i < L; ++i)
				{
				char c1 = Row1[i];
				char c2 = Row2[i];
				if (c1 == '-')
					continue;
				if (c2 == '-')
					continue;
				++Pairs;
				if (toupper(c1) == toupper(c2))
					++Ids;
				}
			if (Pairs > 0)
				PctId = Ids*100.0/Pairs;
			}
		}

	coord_t lo = COORD_T_MAX;
	coord_t hi = COORD_T_MAX;
	std::string strAttrs;
	std::string Label = "?";

	bool Invert = !AI.Segs[0].Plus;
	char Strand = (AI.Segs[0].Plus == AI.Segs[1].Plus) ? '+' : '-';

	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &S = AI.Segs[SegIndex];
		if (S.GetLength() < g_MinLength)
			return;

		std::string SeqName;

		if (g_UseIds)
			{
			char cstrGenDotChr[32];
			sprintf(cstrGenDotChr, "%u.%u", S.GenomeIndex, S.ChrId);
			SeqName = cstrGenDotChr;
			}
		else
			{
			std::string GenName = Rev.GetGenomeName(S.GenomeIndex);
			std::string ChrName = Rev.GetChrName(S.GenomeIndex, S.ChrId);

			TruncLabel(GenName);
			TruncLabel(ChrName);

			SeqName = GenName + "." + ChrName;
			}

		if (SegIndex == 0)
			{
			Label = SeqName;
			lo = S.lo;
			hi = S.hi;
			}
		else
			{
			coord_t Target_lo = S.lo;
			coord_t Target_hi = S.hi;
			char Target_Strand = '?';
			if (Invert)
				Target_Strand = (S.Plus ? '-' : '+');
			else
				Target_Strand = (S.Plus ? '+' : '-');

			char szTargetN[32];
			char szCoords[128];
			if (SegIndex == 1)
				{
				if (strAttrs.size() != 0)
					strAttrs += " ";
				strcpy(szTargetN, "Target ");
				sprintf(szCoords, " %u %u %c;",
				  (unsigned) (Target_lo + 1),
				  (unsigned) (Target_hi + 1),
				  Target_Strand);
				}
			else
				{
				sprintf(szTargetN, " Target%u ", SegIndex);
				sprintf(szCoords, " %u %u %c;",
				  (unsigned) (Target_lo + 1),
				  (unsigned) (Target_hi + 1),
				  Target_Strand);
				}
			strAttrs += szTargetN;
			strAttrs += SeqName;
			strAttrs += szCoords;
			}
		}

	if (PctId >= 0)
		{
		char Str[64];
		sprintf(Str, " pctid \"%.1f\";", PctId);
		strAttrs += Str;
		}

	if (g_AIIndexes)
		{
		char Str[64];
		sprintf(Str, " AIIndex[%u];", AlnInfoIndex);
		strAttrs += Str;
		}

	const unsigned AttrCount = (unsigned) AI.Attrs.size();
	if (AttrCount != (unsigned) AI.Values.size())
		Die("Attr/Value sizes disagree");

	for (unsigned i = 0; i < AttrCount; ++i)
		{
		AlnAttr Attr = AI.Attrs[i];
		unsigned Value = AI.Values[i];

		char Str[64];
		switch (AATypes[Attr])
			{
		case ALN_ATTR_TYPE_UInt32:
			sprintf(Str, " %s %u;", AlnAttrToStr(Attr), Value);
			break;
		case ALN_ATTR_TYPE_Float32:
			{
			double dValue = UIntToDouble(Value);
			sprintf(Str, " %s %g;", AlnAttrToStr(Attr), dValue);
			break;
			}
		default:
			Die("Unknown AAType");
			}

		strAttrs += Str;
		}

	WriteGFFRecord(fGFF, Label.c_str(), "hit", lo, hi, AI.Score, Strand, strAttrs.c_str());
	}

void CvtRev2GFF()
	{
	const char *RevFileName = RequiredValueOpt("fromrev");
	const char *GFFFileName = RequiredValueOpt("togff");
	g_AIIndexes = FlagOpt("aiindexes");
	g_UseIds = FlagOpt("useids");
	IntOpt("minlength", &g_MinLength);
	g_PctId = FlagOpt("pctid");

	RevFile Rev;
	Rev.Open(RevFileName);

	FILE *fGFF = CreateStdioFile(GFFFileName);

	const unsigned AlnInfoCount = Rev.GetAlnInfoCount();
	for (unsigned AIIndex = 0; AIIndex < AlnInfoCount; ++AIIndex)
		Output(Rev, AIIndex, fGFF);

	Rev.Close();
	CloseStdioFile(fGFF);
	}
