#include <memory.h>
#include <stdio.h>
#include <string>
#include <sstream>
#include <algorithm>
#include <vector>
#include <utility>
#include "utils.h"
#include "revfile.h"
#include "progress.h"
#include "getopts.h"

const unsigned FASTA_BLOCK = 60;

void WriteFasta(FILE *fOut, const char *seq, unsigned Length, const char *Label)
{
	fprintf(fOut, ">%s\n", Label);

	unsigned i = 0;

	while (i < Length)
	{
		unsigned size = Length - i;
		if (size > FASTA_BLOCK)
			size = FASTA_BLOCK;

		WriteStdioFile(fOut, seq + i, size);
		fprintf(fOut, "\n");
		i += size;
	}
}

bool ChrIncSizeCompare(std::pair<unsigned,unsigned> a, std::pair<unsigned,unsigned> b)
{
	return (a.second < b.second) || ((a.second == b.second) && (a.first < b.first));
}

bool ChrDecSizeCompare(std::pair<unsigned,unsigned> a, std::pair<unsigned,unsigned> b)
{
	return (a.second > b.second) || ((a.second == b.second) && (a.first < b.first));
}

static void Rev2Mfa(const char *RevFileName, const char *MfaFileName, bool relabel, bool decrsize, bool incrsize)
{
	FILE *fout = CreateStdioFile(MfaFileName);

	RevFile InFile;
	InFile.Open(RevFileName);

	std::vector<std::pair<unsigned, unsigned> > ChrIndexes;

	uint64 TotalChars = 0;
	unsigned SeqCount = InFile.GetSeqCount();
	ChrIndexes.reserve(SeqCount);
	for (unsigned index = 0; index != SeqCount; index++)
	{
		unsigned ChrSize = InFile.GetSeqLengthByIndex(index);
		TotalChars += ChrSize;
		ChrIndexes.push_back(std::make_pair(index, ChrSize));
	}

	if (incrsize)
		std::sort(ChrIndexes.begin(), ChrIndexes.end(), ChrIncSizeCompare);
	else if (decrsize)
		std::sort(ChrIndexes.begin(), ChrIndexes.end(), ChrDecSizeCompare);
	
	/* Unfortunately ProgressBar supports ints, so... */
	uint64 StepCount = TotalChars;
	float StepScale = 1.0;
	if (StepCount > INT_MAX)
	{
		StepScale = ((float)INT_MAX) / StepCount;
		StepCount = INT_MAX;
	}

	Progress("Converting %s to %s", RevFileName, MfaFileName);
	ProgressStart("Processing...");

	uint64 CharsOut = 0;
	for (std::vector<std::pair<unsigned,unsigned> >::const_iterator ci = ChrIndexes.begin(); ci != ChrIndexes.end(); ++ci)
	{
		unsigned GenomeIndex = 0, ChrID = 0;
		char *seq = new char[ci->second];
		InFile.GetSeqByIndex(ci->first, seq, ci->second, &GenomeIndex, &ChrID);
		std::stringstream Label;
		if (relabel)
			Label << InFile.GetGenomeName(GenomeIndex) << "." << ChrID;
		else
			Label << InFile.GetChrName(GenomeIndex, ChrID);
		WriteFasta(fout, seq, ci->second, Label.str().c_str());
		CharsOut += ci->second;
		ProgressStep((int)(CharsOut*StepScale), (int)StepCount);
		delete seq;
	}

	ProgressDone();
	InFile.Close();
	CloseStdioFile(fout);
}

void CvtRev2MFA()
{
	const char *RevFileName = RequiredValueOpt("fromrev");
	const char *MfaFileName = RequiredValueOpt("tomfa");
	bool relabel = FlagOpt("gendotid");
	bool decrsize = FlagOpt("decrsize");
	bool incrsize = FlagOpt("incrsize");

	Rev2Mfa(RevFileName, MfaFileName, relabel, decrsize, incrsize);
}
