#include "evo.h"
#include "gff2.h"

static map<unsigned, double> g_GeneIndexToSumProbs;
static map<unsigned, unsigned> g_GeneIndexToBaseCount;

static double GetSumProbs(const string &Probs)
	{
	double SumProb = 0.0;
	unsigned N = SIZE(Probs);
	if (N%2 != 0)
		Die("Invalid probs attribute, length %u is odd", N);

	for (unsigned i = 0; i < N; i += 2)
		{
		char c1 = Probs[i];
		char c2 = Probs[i+1];
		if (!isxdigit(c1) || !isxdigit(c2))
			{
			if (isprint(c1) && isprint(c2))
				Die("Invalid hex digits '%c%c' in probs", c1, c2);
			else
				Die("Invalid hex digits %02x%02x in probs", (byte) c1, (byte) c2);
			}
		static char s[3];
		s[0] = c1;
		s[1] = c2;
		char *NotUsed;
		unsigned ProbIndex = strtoul(s, &NotUsed, 16);
		SumProb += ProbIndex/255.0;
		}
	return SumProb;
	}

void DeleteRandGenes(const char *GeneFileName, const char *OutFileName, int N)
	{
	ProgressStart("Delete genes, reading input");
	unsigned RecordCount = 0;
	FILE *fGenes = OpenStdioFile(GeneFileName);
	FILE *fOut = CreateStdioFile(OutFileName);

	GFFRecord2 Rec;
	unsigned FileSize = (unsigned) GetStdioFileSize(fGenes);

// First pass to compute MAC
	while (GetNextGFFRecord(fGenes, Rec))
		{
		if ((RecordCount++)%1000 == 0)
			{
			unsigned Pos = (unsigned) GetStdioFilePos(fGenes);
			ProgressStep(Pos, FileSize);
			}

		string strGeneIndex;
		bool Found = GFFGetAttr(Rec.Attrs, "gene_index", strGeneIndex);
		if (!Found)
			continue;
		unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
		string Probs;
		GFFGetRequiredAttr(Rec.Attrs, "probs", Probs);

		double SumProbs = GetSumProbs(Probs);
		unsigned Length = Rec.GetLength();
		map<unsigned, double>::const_iterator p = g_GeneIndexToSumProbs.find(GeneIndex);
		if (p == g_GeneIndexToSumProbs.end())
			{
			g_GeneIndexToSumProbs[GeneIndex] = SumProbs;
			g_GeneIndexToBaseCount[GeneIndex] = Length;
			}
		else
			{
			g_GeneIndexToSumProbs[GeneIndex] += SumProbs;
			g_GeneIndexToBaseCount[GeneIndex] += Length;
			}
		}
	ProgressDone();

	vector<unsigned> GeneIndexes;
	vector<double> MACs;
	for (map<unsigned, double>::const_iterator p = g_GeneIndexToSumProbs.begin();
	  p != g_GeneIndexToSumProbs.end(); ++p)
		{
		unsigned GeneIndex = p->first;
		double SumProb = p->second;
		unsigned BaseCount = g_GeneIndexToBaseCount[GeneIndex];
		asserta(BaseCount > 0);
		double MAC = SumProb/BaseCount;

		GeneIndexes.push_back(GeneIndex);
		MACs.push_back(MAC);
		}

	set<unsigned> GeneIndexesToDelete;

	unsigned GeneCount = SIZE(GeneIndexes);
	unsigned DeleteCount = 0;
	if (N < 0)
		DeleteCount = (unsigned) -N;
	else if ((unsigned) N > GeneCount)
		Warning("Target count > gene count, none deleted");
	else
		DeleteCount = GeneCount - (unsigned) N;

	if (DeleteCount == GeneCount)
		{
		Warning("Deleting all genes");
		DeleteCount = GeneCount;
		}

	Log("\n");
	Log("Delete: ");
	for (unsigned i = 0; i < DeleteCount; ++i)
		{
		unsigned Index = RandWeightDist(MACs);
		asserta(Index < SIZE(GeneIndexes));
		unsigned GeneIndexToDelete = GeneIndexes[Index];
		Log(" %u(%g)", GeneIndexToDelete, MACs[Index]);
		GeneIndexesToDelete.insert(GeneIndexToDelete);

		unsigned N = SIZE(MACs);

		GeneIndexes[Index] = GeneIndexes[N-1];
		MACs[Index] = MACs[N-1];

		GeneIndexes.resize(N-1);
		MACs.resize(N-1);
		}
	Log("\n");

	ProgressStart("Writing output minus deleted genes");
	SetStdioFilePos(fGenes, 0);
// Second pass copies input to output, skipping genes to delete.
	while (GetNextGFFRecord(fGenes, Rec))
		{
		if ((RecordCount++)%1000 == 0)
			{
			unsigned Pos = (unsigned) GetStdioFilePos(fGenes);
			ProgressStep(Pos, FileSize);
			}

		string strGeneIndex;
		bool Found = GFFGetAttr(Rec.Attrs, "gene_index", strGeneIndex);
		if (Found)
			{
			unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
			if (GeneIndexesToDelete.find(GeneIndex) != GeneIndexesToDelete.end())
				continue;
			}

		WriteGFFRec(fOut, Rec);
		}
	ProgressDone();

	CloseStdioFile(fGenes);
	CloseStdioFile(fOut);
	g_GeneIndexToSumProbs.clear();
	g_GeneIndexToBaseCount.clear();
	}

void DeleteRandGenes()
	{
	const char *GeneFileName = RequiredValueOpt("delrandgenes");
	const char *OutFileName = RequiredValueOpt("out");
	const char *strCount = RequiredValueOpt("n");
	int N = atoi(strCount);
	DeleteRandGenes(GeneFileName, OutFileName, N);
	}
