#include "evolver.h"
#include "gff2.h"
#include "progress.h"
#include "evomodel.h"

// TODO: adjacent genes messes up gene boundary finding
// (1) must validate and fix before starting
// (2) how to maintain?

static set<string> g_UnknownFeatures;

int8 FrameToFirstCodex(int8 Frame, bool Strand, gpos_t Length);

static unsigned g_FFCount = 0;

bool IsGeneFeature(const string &Feature)
	{
	return Feature == "CDS" || Feature == "UTR" || Feature == "NXE";
	}

bool IsConservedFeature(const string &Feature)
	{
	return IsGeneFeature(Feature) || Feature == "NGE";
	}

bool IsStartCodon(const string &Codon)
	{
	return Codon == "ATG";
	}

bool IsStopCodon(const string &Codon)
	{
	return Codon == "TAG" || Codon == "TAA" || Codon == "TGA";
	}

void Evolver::SetIsland(gpos_t Start, gpos_t Length, bool On)
	{
	gpos_t Pos = Start;
	for (gpos_t i = 0; i < Length; ++i)
		{
		BaseInfo &Base = GetBase(Pos);
		Base.Island = On;
		Pos = GetNextPos(Pos);
		}
	}

void Evolver::SetAnnot(const GFFRecord2 &Rec)
	{
	const string &Feature = Rec.Feature;

	if (Rec.End == 0)
		Die("GFF record has end=0, coords are 1-based");

	gpos_t Start = Rec.Start - 1;
	gpos_t End = Rec.End - 1;
	asserta(End >= Start);
	gpos_t Length = End - Start + 1;

	if (!PosExists(Start) || !PosExists(End))
		Die("GFF record start=%"PF", end=%"PF", deleted / out of range",
		  Rec.Start, Rec.End);

	if (Start > End)
		Die("GFF record has start=%"PF" > end=%"PF, Rec.Start, Rec.End);

	if (Feature == "island")
		{
		SetIsland(Start, Length, true);
		return;
		}

	else if (Feature == "tandem")
		{
		gpos_t Length = End - Start + 1;
		if (Length < P_MinTandemAnnot || Length > P_MaxTandemAnnot)
			return;
		gpos_t MotifLength = GFFGetRequiredIntAttr(Rec.Attrs, "replen");
		SetTandemMotifLength(Start, Length, MotifLength);
		return;
		}

	else if (IsConservedFeature(Feature))
		{
		string Probs;
		GFFGetRequiredAttr(Rec.Attrs, "probs", Probs);
		gpos_t Length = Rec.GetLength();
		unsigned N = SIZE(Probs);
		if (N%2 != 0)
			{
			Rec.LogMe(true);
			Die("Invalid probs attribute (%s), length %u is odd",
			  Rec.Feature.c_str(), N);
			}
		if (N/2 != Length)
			{
			Rec.LogMe(true);
			Die("Length mismatch (%s), %u probs, %u bases",
			  Rec.Feature.c_str(), N/2, Length);
			}
		SetBaseProbs(Start, End, Probs);
		}
	else
		g_UnknownFeatures.insert(Feature);

	SetGeneAnnot(Rec);
	}

void Evolver::SetGeneAnnot(const GFFRecord2 &Rec)
	{
	if (!IsGeneFeature(Rec.Feature))
		return;

	unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
 
	asserta(Rec.Start > 0);
	asserta(Rec.End >= Rec.Start);
	gpos_t Start = Rec.Start - 1;
	gpos_t Length = Rec.End - Rec.Start + 1;

	if (GeneIndex >= m_NextGeneIndex)
		m_NextGeneIndex = GeneIndex + 1;

	if (Rec.Feature == "CDS")
		{
		bool Plus = Rec.GetPlus();
		int Frame = Rec.Frame;
		if (Frame < 0 || Frame > 2)
			Die("Invalid frame %d in CDS record Start=%"PF, Frame, Rec.Start);
		SetCodexes(Start, Length, (int8) Frame, Plus);
		}
	else if (Rec.Feature == "UTR")
		SetUTR(Start, Length);
	}

// Get codon containing Pos. Positions that are not exonic,
// don't exist or are deleted are indicated by '?'.
// Returns true iff a full codon is successfully retrieved.
bool Evolver::GetCodon(gpos_t Pos, CodonInfo &Codon) const
	{
	Codon.Clear();

	if (!PosExists(Pos))
		return false;
	const BaseInfo &Base = GetBase(Pos);
	if (!Base.IsCoding())
		return false;

	int Codex = Base.Codex;
	int absCodex = abs(Codex);
	asserta(absCodex >= 1 && absCodex <= 3);
	Codon.Letters[absCodex - 1] = Base.GetCodonLetter();
	Codon.Pos[absCodex - 1] = Pos;
	Codon.Strand = (Base.Codex > 0);

// Rewind to codon start
	gpos_t CurrPos = Pos;
	for (int aC = absCodex - 1; aC >= 1 ; --aC)
		{
		CurrPos = GetAdjExonPos(CurrPos, !Codon.Strand);
		if (!PosExists(CurrPos))
			return false;
		const BaseInfo &Base = GetBase(CurrPos);
		if (!Base.IsCoding())
			return false;
		if (aC != abs(Base.Codex))
			{
			LogNeighborsAsTable(Pos);
			Log("\naC=%d Codex=%d Base=", aC, Base.Codex);
			Base.LogMe();
			Log("\n");
			ValidateCDSs();
			Die("Codex error getting codon");
			}
		Codon.Letters[aC - 1] = Base.GetCodonLetter();
		Codon.Pos[aC - 1] = CurrPos;
		}

// Move forward to codon end
	CurrPos = Pos;
	for (int aC = absCodex + 1; aC <= 3 ; ++aC)
		{
		CurrPos = GetAdjExonPos(CurrPos, Codon.Strand);
		if (!PosExists(CurrPos))
			return false;
		const BaseInfo &Base = GetBase(CurrPos);
		if (!Base.IsCoding())
			return false;
		if (aC != abs(Base.Codex))
			{
			LogNeighborsAsTable(Pos);
			Log("\naC=%d Codex=%d Base=", aC, Base.Codex);
			Base.LogMe();
			Log("\n");
			ValidateCDSs();
			Die("Codex error getting codon");
			}
		Codon.Letters[aC - 1] = Base.GetCodonLetter();
		Codon.Pos[aC - 1] = CurrPos;
		}
	return true;
	}

void Evolver::SetBaseProbs(gpos_t Start, gpos_t End, const string &Probs)
	{
	gpos_t Length = End - Start + 1;
	unsigned N = SIZE(Probs);
	if (N%2 != 0)
		Die("Invalid probs attribute, length %u is odd", N);
	if (N/2 != Length)
		Die("Length mismatch, %u probs, %u bases", N/2, Length);
	unsigned i = 0;
	for (gpos_t Pos = Start; Pos <= End; ++Pos)
		{
		char c1 = Probs[i++];
		char c2 = Probs[i++];
		if (!isxdigit(c1) || !isxdigit(c2))
			{
			if (isprint(c1) && isprint(c2))
				Die("Invalid hex digits '%c%c' in probs", c1, c2);
			else
				Die("Invalid hex digits %02x%02x in probs", (byte) c1, (byte) c2);
			}
		BaseInfo &Base = GetBase(Pos);
		static char s[3];
		s[0] = c1;
		s[1] = c2;
		char *NotUsed;
		unsigned ProbIndex = strtoul(s, &NotUsed, 16);
		if (ProbIndex == NEUTRAL_PROBIX)
			{
			++g_FFCount;
			ProbIndex = 1;
			}
		Base.ProbIx = (byte) ProbIndex;
		}
	}

void Evolver::GetAnnots(const char *FileName, const string &SeqName)
	{
	FILE *f = OpenStdioFile(FileName);

	unsigned ErrCount = 0;
	GFFRecord2 Rec;
	ProgressStartFile(f, "Reading %s", FileName);

	map<unsigned, gpos_t> GeneLos;
	map<unsigned, gpos_t> GeneHis;

	map<unsigned, gpos_t> CDSLos;
	map<unsigned, gpos_t> CDSHis;

	unsigned RecordCount = 0;
	unsigned SkippedCount = 0;
	while (GetNextGFFRecord(f, Rec))
		{
		if (RecordCount++%10 == 0)
			ProgressStepFile();
	
		if (Rec.SeqName != SeqName)
			{
			++SkippedCount;
			continue;
			}

		unsigned GeneIndex = UINT_MAX;
		if (GFFGetIntAttr(Rec.Attrs, "gene_index", GeneIndex))
			{
			asserta(Rec.Start > 0);
			asserta(Rec.End >= Rec.Start);
			gpos_t lo = Rec.Start - 1;
			gpos_t hi = Rec.End - 1;
			map<unsigned, gpos_t>::const_iterator p = GeneLos.find(GeneIndex);
			if (p == GeneLos.end())
				{
				GeneLos[GeneIndex] = lo;
				GeneHis[GeneIndex] = hi;
				}
			else
				{
				GeneLos[GeneIndex] = min(lo, GeneLos[GeneIndex]);
				GeneHis[GeneIndex] = max(hi, GeneHis[GeneIndex]);
				}

			if (Rec.Feature == "CDS")
				{
				map<unsigned, gpos_t>::const_iterator p = CDSLos.find(GeneIndex);
				if (p == CDSLos.end())
					{
					CDSLos[GeneIndex] = lo;
					CDSHis[GeneIndex] = hi;
					}
				else
					{
					CDSLos[GeneIndex] = min(lo, CDSLos[GeneIndex]);
					CDSHis[GeneIndex] = max(hi, CDSHis[GeneIndex]);
					}
				}
			}
		SetAnnot(Rec);
		}
	ProgressDone();

	const unsigned UnknownFeatureCount = SIZE(g_UnknownFeatures);
	if (UnknownFeatureCount != 0)
		{
		string s;
		for (set<string>::const_iterator p = g_UnknownFeatures.begin();
		  p != g_UnknownFeatures.end(); ++p)
			s += " " + *p;
		Warning("Unknown feature types in GFF file: %s", s.c_str());
		}

	if (g_FFCount > 0)
		Warning("Accept probabilty 0xff=1.0 found for %u base(s), set to 0xfe=254/255", g_FFCount);

	ProgressStart("Initialize genes");
	asserta(SIZE(GeneHis) == SIZE(GeneLos));

	map<unsigned, gpos_t>::const_iterator pLos = GeneLos.begin();
	map<unsigned, gpos_t>::const_iterator pHis = GeneHis.begin();
	for (;;)
		{
		if (pLos == GeneLos.end())
			break;

		gpos_t lo = pLos->second;
		gpos_t hi = pHis->second;
		gpos_t Length = hi - lo + 1;

		SetIntronsAndUntranscribed(lo, Length);
		SetSplices(lo, Length);

		++pLos;
		++pHis;
		}
	ProgressDone();

	unsigned CDSCount = SIZE(CDSLos);
	asserta(SIZE(CDSHis) == CDSCount);

	pLos = CDSLos.begin();
	pHis = CDSHis.begin();
	for (;;)
		{
		if (pLos == CDSLos.end())
			break;

		unsigned GeneIndex = pLos->first;
		asserta(pHis->first == GeneIndex);

		gpos_t lo = pLos->second;
		gpos_t hi = pHis->second;

		const BaseInfo &LoBase = GetBase(lo);

		gpos_t StartCodonFirstBasePos = GPOS_T_MAX;
		gpos_t StopCodonLastBasePos = GPOS_T_MAX;
		bool Strand = LoBase.GetStrand();

		if (Strand)
			{
			StartCodonFirstBasePos = lo;
			StopCodonLastBasePos = hi;
			}
		else
			{
			StartCodonFirstBasePos = hi;
			StopCodonLastBasePos = lo;
			}

		gpos_t Pos2 = GetAdjExonPos(StopCodonLastBasePos, !Strand);
		gpos_t StopCodonFirstBasePos = GetAdjExonPos(Pos2, !Strand);

		unsigned n = SetStartCodon(StartCodonFirstBasePos, Strand);
		if (n > 0)
			Log("GeneIndex=%u lo=%u hi=%u\n", GeneIndex, lo, hi);
		ErrCount += n;
		n = SetStopCodon(StopCodonFirstBasePos, Strand);
		if (n > 0)
			Log("GeneIndex=%u lo=%u hi=%u\n", GeneIndex, lo, hi);
		ErrCount += n;

		m_GeneIndexes[StartCodonFirstBasePos] = GeneIndex;

		++pLos;
		++pHis;
		}

	CloseStdioFile(f);
	}

void Evolver::SetCodexes(gpos_t Start, gpos_t Length, int8 Frame, bool Plus)
	{
	int8 Codex = FrameToFirstCodex(Frame, Plus, Length);
	gpos_t Pos = Start;
	for (gpos_t i = 0; i < Length; ++i)
		{
		BaseInfo &Base = GetBase(Pos);
		Base.Type = BT_CDS;
		Base.Codex = Codex;
		Codex = IncCodex(Codex);
		Pos = GetNextPos(Pos);
		}
	}

void Evolver::SetUTR(gpos_t Start, gpos_t Length)
	{
	gpos_t Pos = Start;
	for (gpos_t i = 0; i < Length; ++i)
		{
		BaseInfo &Base = GetBase(Pos);
		Base.SetUTR();
		Pos = GetNextPos(Pos);
		}
	}

void Evolver::SetSplices(gpos_t Start, gpos_t Length)
	{
	gpos_t Pos = Start;
	bool InIntron = false;
	for (gpos_t i = 0; i < Length; ++i)
		{
		BaseInfo &Base = GetBase(Pos);
		if (Base.IsIntron())
			{
			if (!InIntron)
				{
				Base.SetSplice();
				gpos_t NextPos = GetNextPos(Pos);
				if (NextPos != GPOS_T_MAX)
					{
					BaseInfo &Base = GetBase(NextPos);
					if (Base.IsIntron())
						Base.SetSplice();
					}
				InIntron = true;
				}
			}
		else
			{
			if (InIntron)
				{
				gpos_t PrevPos = GetPrevPos(Pos);
				asserta(PrevPos != GPOS_T_MAX);
				BaseInfo &Base = GetBase(PrevPos);
				Base.SetSplice();
				PrevPos = GetPrevPos(PrevPos);
				if (PrevPos != GPOS_T_MAX)
					{
					BaseInfo &Base = GetBase(PrevPos);
					if (Base.IsIntron())
						Base.SetSplice();
					}
				InIntron = false;
				}
			}
		Pos = GetNextPos(Pos);
		}
	}

void Evolver::SetIntronsAndUntranscribed(gpos_t Start, gpos_t Length)
	{
	gpos_t Pos = Start;
	gpos_t FirstExonPos = GPOS_T_MAX;
	gpos_t LastExonPos = GPOS_T_MAX;
	for (gpos_t i = 0; i < Length; ++i)
		{
		BaseInfo &Base = GetBase(Pos);
		if (Base.IsExon())
			{
			if (FirstExonPos == GPOS_T_MAX)
				FirstExonPos = Pos;
			LastExonPos = Pos;
			}
		Pos = GetNextPos(Pos);
		}

	if (FirstExonPos == GPOS_T_MAX)
		{
		asserta(LastExonPos == GPOS_T_MAX);
		Warning("SetIntronsAndUntranscribed(%"PF", %"PF"), no exons",
		  Start, Length);
		return;
		}

	Pos = FirstExonPos;
	for (gpos_t i = 0; i < Length; ++i)
		{
		if (Pos == LastExonPos)
			break;
		BaseInfo &Base = GetBase(Pos);
		if (!Base.IsExon())
			Base.SetIntron();
		Pos = GetNextPos(Pos);
		}

	Pos = Start;
	bool BeforeFirstExon = true;
	bool AfterLastExon = false;
	for (gpos_t i = 0; i < Length; ++i)
		{
		if (Pos == FirstExonPos)
			BeforeFirstExon = false;
		if (BeforeFirstExon || AfterLastExon)
			{
			BaseInfo &Base = GetBase(Pos);
			Base.SetUntranscribedGene();
			}
		if (Pos == LastExonPos)
			AfterLastExon = true;
		Pos = GetNextPos(Pos);
		}
	}

unsigned Evolver::SetStartCodon(gpos_t Pos, bool Strand)
	{
	gpos_t Pos2 = GetAdjExonPos(Pos, Strand);
	gpos_t Pos3 = GetAdjExonPos(Pos2, Strand);

	BaseInfo &Base1 = GetBase(Pos);
	BaseInfo &Base2 = GetBase(Pos2);
	BaseInfo &Base3 = GetBase(Pos3);

	string Letters;
	Letters.push_back(Base1.GetStrandLetter(Strand));
	Letters.push_back(Base2.GetStrandLetter(Strand));
	Letters.push_back(Base3.GetStrandLetter(Strand));

	if (!IsStartCodon(Letters))
		{
		Warning("Pos %"PF": codon '%s' at start of CDS, expected ATG",
		  Pos, Letters.c_str());
		return 1;
		}

	Base1.SetStartCodon();
	Base2.SetStartCodon();
	Base3.SetStartCodon();

	Base1.Codex = (Strand ? 1 : -1);
	Base2.Codex = (Strand ? 2 : -2);
	Base3.Codex = (Strand ? 3 : -3);

	return 0;
	}

unsigned Evolver::SetStopCodon(gpos_t Pos, bool Strand)
	{
	gpos_t Pos2 = GetAdjExonPos(Pos, Strand);
	gpos_t Pos3 = GetAdjExonPos(Pos2, Strand);

	BaseInfo &Base1 = GetBase(Pos);
	BaseInfo &Base2 = GetBase(Pos2);
	BaseInfo &Base3 = GetBase(Pos3);

	string Letters;
	Letters.push_back(Base1.GetStrandLetter(Strand));
	Letters.push_back(Base2.GetStrandLetter(Strand));
	Letters.push_back(Base3.GetStrandLetter(Strand));

	if (!IsStopCodon(Letters))
		{
		Warning("Pos %"PF": codon '%s' at end of CDS, expected STOP",
		  Pos, Letters.c_str());
		return 1;
		}

	Base1.SetStopCodon();
	Base2.SetStopCodon();
	Base3.SetStopCodon();

	Base1.Codex = (Strand ? 1 : -1);
	Base2.Codex = (Strand ? 2 : -2);
	Base3.Codex = (Strand ? 3 : -3);

	return 0;
	}
