#include "evolver.h"
#include "evofile.h"
#include "progress.h"
#include "ratemgr.h"
#include "getopts.h"
#include "iovecs.h"

void LogNucMx();
extern double g_SubstProbs[4][4];
extern double g_CGFailProb;
extern double g_ATFailProb;

void Evolver::MEsToFile(FILE *f) const
	{
	unsigned MECount = SIZE(m_MEs);
	WRITEV(MECount);
	for (unsigned MEIndex = 0; MEIndex < MECount; ++MEIndex)
		{
		const MEInfo &ME = m_MEs[MEIndex];
		WRITEV(ME.Rate);
		WRITEV(ME.FractWithDeletion);
		WRITEV(ME.MeanDeleteLength);
		WRITEV(ME.StdDevDeleteLength);
		WRITEV(ME.AcceptedInsertCount);

		unsigned RejectedInsertCount = 0; // Placeholder for now-defunct member
		WRITEV(RejectedInsertCount);
		WRITEV(ME.DeleteCount);
		WRITEV(ME.RandThresh);

		unsigned NameLength = SIZE(ME.Name);
		WRITEV(NameLength);
		for (unsigned i = 0; i < NameLength; ++i)
			WRITEV(ME.Name.c_str()[i]);

		unsigned SeqLength = SIZE(ME.Seq);
		WRITEV(SeqLength);
		for (unsigned i = 0; i < SeqLength; ++i)
			WRITEV(ME.Seq.c_str()[i]);
		}
	}

void Evolver::MEsFromFile(FILE *f)
	{
	m_MEs.clear();
	unsigned MECount;
	READV(MECount);
	for (unsigned MEIndex = 0; MEIndex < MECount; ++MEIndex)
		{
		MEInfo ME;
		READV(ME.Rate);
		READV(ME.FractWithDeletion);
		READV(ME.MeanDeleteLength);
		READV(ME.StdDevDeleteLength);
		READV(ME.AcceptedInsertCount);

		unsigned RejectedInsertCount; // Placeholder for now-defunct member
		READV(RejectedInsertCount);
		READV(ME.DeleteCount);
		READV(ME.RandThresh);

		unsigned NameLength;
		READV(NameLength);
		for (unsigned i = 0; i < NameLength; ++i)
			{
			char c;
			READV(c);
			ME.Name.push_back(c);
			}

		unsigned SeqLength;
		READV(SeqLength);
		for (unsigned i = 0; i < SeqLength; ++i)
			{
			char c;
			READV(c);
			ME.Seq.push_back(c);
			}
		m_MEs.push_back(ME);
		}
	}

void Evolver::WriteEventGenState(FILE *f) const
	{
	WRITEV(m_BranchLength);
	WRITEV(m_EventIndex);
	WRITEV(m_EventCount);
	WRITEV(m_LastAcceptedEventIndex);

	WRITEV(m_AverageWaitingTime);
	WRITEV(m_TandemRateRecomputeCounter);

	WriteVec(f, m_IntegratedEventRates);
	WriteVec(f, m_EventRandDistHis);

#define R(e) extern RateMgr *g_##e##RateMgr; g_##e##RateMgr->ToFile(f);
// Only subset of events have a RateMgr, can't #include "events.h" here.
	R(Delete)
	R(Invert)
	R(Move)
	R(Copy)
	R(Tandem)
	R(Insert)
	R(MEInsert)
	R(NCCE)
#undef R
	}

void Evolver::ReadEventGenState(FILE *f)
	{
	READV(m_BranchLength);
	READV(m_EventIndex);
	READV(m_EventCount);
	READV(m_LastAcceptedEventIndex);

	READV(m_AverageWaitingTime);
	READV(m_TandemRateRecomputeCounter);

	ReadVec(f, m_IntegratedEventRates);
	ReadVec(f, m_EventRandDistHis);

#define R(e) extern RateMgr *g_##e##RateMgr; g_##e##RateMgr->FromFile(f);
// Only subset of events have a RateMgr, can't #include "events.h" here.
	R(Delete)
	R(Invert)
	R(Move)
	R(Copy)
	R(Tandem)
	R(Insert)
	R(MEInsert)
	R(NCCE)
#undef R
	}

void Evolver::ToFile(const char *FileName, bool ShowProgress) const
	{
	PushShowProgress(ShowProgress);
	Progress("Event %u, writing state to %.32s", m_EventIndex, FileName);
	FILE *f = CreateStdioFile(FileName);

	EvoFileHdr Hdr;
	memset(&Hdr, 0xff, sizeof(Hdr));

// Bad magics until file finalized to provoke error on reading
	Hdr.StartMagic = 0;
	Hdr.EndMagic = 0;

	Hdr.AncestralLength = m_AncestralLength;
	Hdr.EvolvedLength = m_EvolvedLength;

	Hdr.BasesLength = SIZE(m_Bases);
	Hdr.AncPositionsLength = SIZE(m_AncPositions);
	Hdr.TandemCount = SIZE(m_Tandems);
	Hdr.GeneCount = SIZE(m_GeneIndexes);
	Hdr.ParalogCount = SIZE(m_Paralogs);

	Hdr.NextGeneIndex = m_NextGeneIndex;
	Hdr.FirstEvolvedPos = m_FirstEvolvedPos;
	Hdr.FirstDeletedPos = m_FirstDeletedPos;
	Hdr.LastEvolvedPos = m_LastEvolvedPos;
	Hdr.LastDeletedPos = m_LastDeletedPos;
	Hdr.AncChrId = m_AncChrId;

	WriteStdioFile(f, &Hdr, sizeof(Hdr));

	Progress("Write %u bases", SIZE(m_Bases));
	WriteVec(f, m_Bases);

	Progress("Write %u fwd/bwd pointers", SIZE(m_BaseListFwds));
	WriteVec(f, m_BaseListFwds);
	WriteVec(f, m_BaseListBwds);

	Progress("Write %u ext. anc. positions", SIZE(m_AncPositions));
	WriteVec(f, m_AncPositions);

	Progress("Write %u tandems", SIZE(m_Tandems));
	WriteIntMap(f, m_Tandems);

	Progress("Write %u MEs", SIZE(m_MEs));
	MEsToFile(f);

	Progress("Write %u gene indexes, %u paralogs",
	  SIZE(m_GeneIndexes), SIZE(m_Paralogs));
	WriteIntMap(f, m_GeneIndexes);
	WriteIntMap(f, m_Paralogs);

	Progress("Write event generation state");
	WriteEventGenState(f);

	Progress("Write random number state");
	WriteRandState(f);

	Progress("Write nuc subst mx");
	WRITEV(g_SubstProbs);
	WRITEV(g_CGFailProb);
	WRITEV(g_ATFailProb);
	WRITEV(m_DoCpGSweeps);
	unsigned Count = m_CpGSweepEventCounter + 1;
	WRITEV(Count);
	WRITEV(m_CpGSweepInterval);

	Progress("Write ancestral genome info");
	m_AncGenInfo.ToFile(f);

// Write header with correct magics
	SetStdioFilePos(f, 0);
	Hdr.StartMagic = EvoFileHdrStartMagic;
	Hdr.EndMagic = EvoFileHdrEndMagic;
	WriteStdioFile(f, &Hdr, sizeof(Hdr));

	PopShowProgress();
	CloseStdioFile(f);
	}

void Evolver::FromFile(const char *FileName, bool ShowProgress)
	{
	Clear();

	PushShowProgress(ShowProgress);
	Progress("Read %.32s", FileName);

	FILE *f = OpenStdioFile(FileName);

	EvoFileHdr Hdr;
	ReadStdioFile(f, &Hdr, sizeof(Hdr));

	asserta(Hdr.StartMagic == EvoFileHdrStartMagic);
	asserta(Hdr.EndMagic == EvoFileHdrEndMagic);

	m_AncestralLength = Hdr.AncestralLength;
	m_EvolvedLength = Hdr.EvolvedLength;
	m_NextGeneIndex = Hdr.NextGeneIndex;
	m_FirstEvolvedPos = Hdr.FirstEvolvedPos;
	m_FirstDeletedPos = Hdr.FirstDeletedPos;
	m_LastEvolvedPos = Hdr.LastEvolvedPos;
	m_LastDeletedPos = Hdr.LastDeletedPos;
	m_AncChrId = Hdr.AncChrId;

	double ExtendFactor = 2.0;
	FloatOpt("ext", &ExtendFactor);
	size_t ExtendedSize = (size_t) (ExtendFactor*m_AncestralLength);

	m_Bases.reserve(ExtendedSize);
	m_BaseListFwds.reserve(ExtendedSize);
	m_BaseListBwds.reserve(ExtendedSize);

	ReadVec(f, m_Bases);
	Progress("%u bases", SIZE(m_Bases));
	asserta(SIZE(m_Bases) == Hdr.BasesLength);

	ReadVec(f, m_BaseListFwds);
	ReadVec(f, m_BaseListBwds);
	asserta(SIZE(m_BaseListFwds) == Hdr.BasesLength);
	asserta(SIZE(m_BaseListBwds) == Hdr.BasesLength);
	Progress("%u fwd/bwd pointers", SIZE(m_BaseListFwds));

	ReadVec(f, m_AncPositions);
	Progress("%u ext. anc. positions", SIZE(m_AncPositions));
	asserta(SIZE(m_AncPositions) == Hdr.AncPositionsLength);

	ReadIntMap(f, m_Tandems);
	Progress("%u tandems", SIZE(m_Tandems));
	asserta(SIZE(m_Tandems) == Hdr.TandemCount);

	MEsFromFile(f);
	Progress("%u MEs", SIZE(m_MEs));

	ReadIntMap(f, m_GeneIndexes);
	ReadIntMap(f, m_Paralogs);
	Progress("%u gene indexes, %u paralogs",
	  SIZE(m_GeneIndexes), SIZE(m_Paralogs));
	asserta(SIZE(m_GeneIndexes) == Hdr.GeneCount);
	asserta(SIZE(m_Paralogs) == Hdr.ParalogCount);

	ReadEventGenState(f);
	ReadRandState(f);

	READV(g_SubstProbs);
	READV(g_CGFailProb);
	READV(g_ATFailProb);
	READV(m_DoCpGSweeps);
	READV(m_CpGSweepEventCounter);
	READV(m_CpGSweepInterval);
	LogNucMx();

	m_AncGenInfo.FromFile(f);

	CloseStdioFile(f);
	PopShowProgress();
	}
