#define TRACE	0

#include "utils.h"
#include "evolver.h"
#include "myrand.h"
#include "progress.h"
#include "evomodel.h"

char CompLetter(char c);

static bool g_IsACGT[256];

static bool InitACGT()
	{
#define x(c)	g_IsACGT[(unsigned char) c] = true;
	x('A')
	x('C')
	x('G')
	x('T')
#undef x
	return true;
	}
static bool AGCTInitialized = InitACGT();

bool isacgt(char c)
	{
	return g_IsACGT[(unsigned char) c];
	}

const char *BaseType2ToStr(BaseType2 Type)
	{
	switch (Type)
		{
#define c(x)	case BT2_##x: return #x;
	c(NGE)
	c(NXE)
	c(NGNeutral)
	c(UGNeutral)
	c(UTR)
	c(CDSPlus)
	c(CDSMinus)
	c(START)
	c(STOP)
	c(Splice)
	c(IntronNeutral)
	c(IntronCE)
	default:
		break;
		}
	return "BT2_?";
	}

const char *BaseTypeToStr(BaseType Type)
	{
	switch (Type)
		{
	case BT_NonGene:
		return "NonGene";
	case BT_UntranscribedGene:
		return "UntrGene";
	case BT_UTR:
		return "UTR";
	case BT_CDS:
		return "CDS";
	case BT_START:
		return "START";
	case BT_STOP:
		return "STOP";
	case BT_Splice:
		return "Splice";
	case BT_Intron:
		return "Intron";
		}
	return "?Type?";
	}

const char *BaseCategoryToStr(BaseCategory c)
	{
	switch (c)
		{
	case BC_Ancestral:
		return "Anc";
	case BC_Copy:
		return "Cpy";
	case BC_Novel:
		return "Nov";
	default:
		static char Str[16];
		sprintf(Str, "?%d", c);
		return Str;
		}
	}

// If a base has codex n, returns the codex of the
// following base on the plus strand.
int8 IncCodex(int8 Codex)
	{
	switch (Codex)
		{
	case 1:
		return 2;
	case 2:
		return 3;
	case 3:
		return 1;
	case -1:
		return -3;
	case -2:
		return -1;
	case -3:
		return -2;
		}
	asserta(false);
	UNREACHABLE_RETURN(SCHAR_MAX);
	}

// If a base has codex n, returns the codex of the
// following base in translation order
int8 IncCodex2(int8 Codex)
	{
	switch (Codex)
		{
	case 1:
		return 2;
	case 2:
		return 3;
	case 3:
		return 1;
	case -1:
		return -2;
	case -2:
		return -3;
	case -3:
		return -1;
		}
	asserta(false);
	UNREACHABLE_RETURN(SCHAR_MAX);
	}

// If a base has codex n, returns the codex of the
// preceding base on the plus strand.
int8 DecCodex(int8 Codex)
	{
	switch (Codex)
		{
	case 1:
		return 3;
	case 2:
		return 1;
	case 3:
		return 2;
	case -1:
		return -2;
	case -2:
		return -3;
	case -3:
		return -1;
		}
	asserta(false);
	UNREACHABLE_RETURN(SCHAR_MAX);
	}

int8 IncCodexStrand(int8 Codex, bool Strand)
	{
	return Strand ? IncCodex(Codex) : DecCodex(Codex);
	}

int8 DecCodexStrand(int8 Codex, bool Strand)
	{
	return Strand ? DecCodex(Codex) : IncCodex(Codex);
	}

int8 IncCodexL(int8 Codex, gpos_t L)
	{
	for (gpos_t i = 0; i < L; ++i)
		Codex = IncCodex(Codex);
	return Codex;
	}

int8 IncCodex2L(int8 Codex, gpos_t L)
	{
	for (gpos_t i = 0; i < L; ++i)
		Codex = IncCodex2(Codex);
	return Codex;
	}

Evolver::Evolver()
	{
	Clear();
	}

void Evolver::Clear()
	{
	m_AncGenInfo.Clear();
	m_AncChrId = UINT_MAX;
	m_LogEvents = false;
	m_AverageWaitingTime = FLT_MAX;
	m_IntegratedEventRates.clear();
	m_CurrentEvent = EVENT_TYPE_Invalid;
	m_LastAcceptedEvent = EVENT_TYPE_Invalid;
	m_LastAcceptedEventPos = GPOS_T_MAX;
	m_CurrentEventPos = GPOS_T_MAX;
	m_NextGeneIndex = 0;
	m_EventIndex = 0;
	m_EventCount = 0;
	m_LastAcceptedEventIndex = 0;
	m_LogEventFrom = UINT_MAX;
	m_LogEventTo = UINT_MAX;

	m_CpGSweepInterval = 0;
	m_CpGSweepEventCounter = 0;
	m_AncestralLength = 0;
	m_EvolvedLength = 0;
	m_Bases.clear();
	m_BaseListFwds.clear();
	m_BaseListBwds.clear();
	m_FirstEvolvedPos = UINT32_MAX;
	m_FirstDeletedPos = UINT32_MAX;
	m_LastEvolvedPos = UINT32_MAX;
	m_LastDeletedPos = UINT32_MAX;
	m_AncPositions.clear();
	m_Tandems.clear();
	m_TandemRateRecomputeCounter = 0;
	m_MEs.clear();
	}

void Evolver::FromRandomString(gpos_t Length)
	{
	string Seq;
	uint32 Length32 = uint64_to_unsigned(Length);
	RandSeq(Seq, Length32);
	FromString(Seq);
	}

void Evolver::FromString(const string &Seq, bool ShowProgress)
	{
	FromCString(Seq.c_str(), SIZE(Seq), ShowProgress);
	}

void Evolver::FromCString(const char *Seq, unsigned Length, bool ShowProgress)
	{
	Clear();

	BaseInfo Base;
	Base.SetLetter('?');
	Base.Plus = true;
	Base.Island = false;
	Base.ProbIx = NEUTRAL_PROBIX;
	Base.SetNonGene();
	BaseList_Init(Seq, Length, Base, ShowProgress);
	}

void Evolver::ToString(string &Seq) const
	{
	Seq.clear();
	gpos_t BaseCount64 = GetEvolvedLength();
	uint32 BaseCount32 = coord64_t_to_uint32(BaseCount64);
	Seq.reserve(BaseCount32);
	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];
		Seq.push_back(Base.GetOrientedLetter());
		}
	}

void Evolver::GetSubSeq(gpos_t Pos, gpos_t Length, string &Seq,
  bool Plus) const
	{
	Seq.clear();
	if (Length == 0)
		return;
	uint32 p = Pos;
	for (unsigned i = 0; i < Length; ++i)
		{
		asserta(p != UINT32_MAX);
		const BaseInfo &Base = m_Bases[p];
		Seq.push_back(Base.GetOrientedLetter());
		p = m_BaseListFwds[p];
		}
	if (!Plus)
		{
		void RevComp(string &Seq);
		RevComp(Seq);
		}
	}

// Insert base following target pos
gpos_t Evolver::InsertNovelBase(gpos_t TargetPos, char Letter)
	{
	BaseInfo Base;
	Base.SetLetter(Letter);
	Base.SetAncLetter(NovelLetter);
	Base.Plus = true;
	Base.Type = BT_NonGene;
	Base.Codex = 0;
	Base.Deconstrain();

	return BaseList_Insert(Base, TargetPos, UINT32_MAX);
	}

bool Evolver::PosExists(gpos_t Pos) const
	{
	if (Pos == GPOS_T_MAX || Pos >= SIZE(m_Bases))
		return false;
	const BaseInfo &Base = m_Bases[Pos];
	return Base.GetLetter() != '-';
	}

bool Evolver::RangeExists(gpos_t Pos, gpos_t Length) const
	{
	if (Length == 0)
		return false;

	if (!PosExists(Pos))
		return false;

	for (gpos_t i = 0; i < Length; ++i)
		{
		if (Pos == GPOS_T_MAX)
			return false;
		Pos = GetNextPos(Pos);
		}
	return true;
	}

void Evolver::MoveBase(gpos_t FromPos, gpos_t ToPos)
	{
	if (FromPos == ToPos)
		return;
	const BaseInfo Base = GetBase(FromPos);
	unsigned GeneIndex = UINT_MAX;
	bool IsStart = Base.IsStartCodon() && (Base.Codex == 1 || Base.Codex == -1);
	if (IsStart)
		{
		map<gpos_t, unsigned>::iterator p = m_GeneIndexes.find(FromPos);
		if (p == m_GeneIndexes.end())
			Die("MoveBase, START at pos %"PF" not in m_GeneIndexes", FromPos);
		GeneIndex = p->second;
		m_GeneIndexes.erase(p);
		}
	BaseList_Delete(FromPos);
	gpos_t NewPos = BaseList_Insert(Base, ToPos, FromPos);
	if (IsStart)
		m_GeneIndexes[NewPos] = GeneIndex;
	}

// Get pos at end of ramge.
gpos_t Evolver::GetRangeEndPos(gpos_t Pos, gpos_t Length, bool *ptrExists) const
	{
	if (!PosExists(Pos))
		{
		if (ptrExists != 0)
			*ptrExists = false;
		return GPOS_T_MAX;
		}

	gpos_t p = Pos;
	for (gpos_t i = 0; i + 1 < Length; ++i)
		{
		p = GetNextPos(p);
		if (Pos == GPOS_T_MAX)
			{
			if (ptrExists != 0)
				*ptrExists = false;
			return GPOS_T_MAX;
			}
		}

	if (ptrExists != 0)
		*ptrExists = true;
	return p;
	}

gpos_t Evolver::GetRangeStartPos(gpos_t Pos, gpos_t Length) const
	{
	for (gpos_t i = 0; i < Length; ++i)
		Pos = GetPrevPos(Pos);
	return Pos;
	}

bool Evolver::PosInRange(gpos_t Pos, gpos_t StartPos, gpos_t Length) const
	{
	if (Length == 0)
		Die("Evolver::PosInRange, Length=0");

	gpos_t p = StartPos;
	for (uint32 i = 0; i < Length; ++i)
		{
		if (p == Pos)
			return true;
		if (p == UINT32_MAX)
			Die("PosInRange: range overflows sequence");
		p = GetNextPos(p);
		}
	return false;
	}

void Evolver::DeleteBase(gpos_t Pos)
	{
	m_Tandems.erase(Pos);
	BaseList_Delete(Pos);
	}

// Return random pos
gpos_t Evolver::GetRandomPos() const
	{
	const gpos_t BaseCount = GetExtendedLength();
	if (BaseCount == 0)
		return GPOS_T_MAX;

	const unsigned MaxTries = 10;
	for (unsigned Try = 0; Try < MaxTries; ++Try)
		{
		gpos_t Pos = RandInt(BaseCount);
		if (!PosIsDeleted(Pos))
			return Pos;
		}
	return GPOS_T_MAX;
	}

gpos_t Evolver::GetHomologousPos(gpos_t Pos) const
	{
	if (Pos < m_AncestralLength)
		return Pos;
	uint32 Index = Pos - m_AncestralLength;
	asserta(Index < SIZE(m_AncPositions));
	gpos_t HomPos = m_AncPositions[Index];
	if (HomPos < m_AncestralLength)
		return HomPos;
	return GPOS_T_MAX;
	}

gpos_t Evolver::GetExtendedBaseCount() const
	{
	return SIZE(m_AncPositions);
	}

bool Evolver::PosIsDeleted(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsDeleted();
	}

gpos_t Evolver::GetRandomTandemPos() const
	{
	unsigned N = SIZE(m_Tandems);
	if (N == 0)
		return GPOS_T_MAX;

	unsigned i = RandInt(N);
	map<gpos_t, gpos_t>::const_iterator p = m_Tandems.begin();
	for (unsigned j = 0; j < i; ++j)
		++p;
	return p->first;
	}

gpos_t Evolver::GetTandemMotifLength(gpos_t Pos) const
	{
	map<gpos_t, gpos_t>::const_iterator p = m_Tandems.find(Pos);
	if (p == m_Tandems.end())
		return GPOS_T_MAX;
	return p->second;
	}

void Evolver::SetTandemMotifLength(gpos_t Pos_, gpos_t Length, gpos_t MotifLength)
	{
	gpos_t Pos = Pos_;
	for (gpos_t i = 1; ; ++i)
		{
		if (Pos == GPOS_T_MAX || PosIsDeleted(Pos))
			return;
		m_Tandems[Pos] = MotifLength;
		if (i == Length)
			break;
		Pos = GetNextPos(Pos);
		}
	}

void Evolver::RecomputeTandemRates()
	{
	if (m_TandemRateRecomputeCounter++%1000 == 0)
		ComputeEventRates(DBL_MAX);
	}

void Evolver::SetInsertConstraints(gpos_t InsertPos, gpos_t Length)
	{
	if (Length == 0)
		return;

	const BaseInfo &InsBase = GetBase(InsertPos);
	asserta(!InsBase.IsSplice());
	int8 Codex = InsBase.Codex;
	bool CDS = InsBase.IsCoding();
	bool Island = InsBase.Island;
	if (Length > P_MaxCpGInsert)
		Island = false;
	uint8 Type = InsBase.Type;
	if (Type == BT_START)
		Type = BT_CDS;

// Nasty special case: insert pos is first base of start codon
// on reverse strand.
	if (CDS && InsBase.Codex < 0 && InsBase.IsStartCodon())
		{
		if (InsBase.Codex != -1)
			{
			Log("\n");
			if (Length > 16)
				Length = 16;
			LogNeighborsAsTable(InsertPos, Length*3);
			Log("Insert pos %"PF" length %"PF"\n", InsertPos, Length);
			Die("SetInsertConstraints: insert into start codon");
			}
		CDS = false;
		Type = BT_UTR;
		}

	gpos_t Pos = InsertPos;
	for (gpos_t i = 0; i < Length; ++i)
		{
		Pos = GetNextPos(Pos);
		BaseInfo &Base = GetBase(Pos);
		Base.Type = Type;
		Base.Island = Island;
		if (CDS)
			Codex = IncCodex(Codex);
		Base.Codex = Codex;
		Base.ProbIx = InsBase.ProbIx;
		}
	}

void Evolver::GetEvolvedSeq(string &Seq) const
	{
	gpos_t Length = GetEvolvedLength();
	Seq.reserve(Length);

	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];
		char Letter = Base.GetOrientedLetter();
		asserta(isacgt(Letter));
		Seq.push_back(Letter);
		}
	}

void Evolver::GetCatStats(CatStats &Stats) const
	{
	memset(&Stats, 0, sizeof(Stats));

	Stats.TandemBaseCount = SIZE(m_Tandems);
	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];
		++(Stats.BaseCount);
		if (Base.IsConstrained())
			++(Stats.ConstrainedBaseCount);
		else
			++(Stats.NeutralBaseCount);
		if (Base.IsGene())
			++(Stats.GeneBaseCount);
		if (Base.IsCDS())
			++(Stats.CDSBaseCount);
		if (Base.IsUTR())
			++(Stats.UTRBaseCount);
		if (Base.IsSplice())
			++(Stats.SpliceBaseCount);
		if (Base.IsNXE())
			++(Stats.NXEBaseCount);
		if (Base.IsNGE())
			++(Stats.NGEBaseCount);
		if (Base.IsIntron())
			++(Stats.IntronBaseCount);
		if (Base.IsIsland())
			++(Stats.CpGBaseCount);

		switch (GetCategory(p))
			{
		case BC_Ancestral:
			++(Stats.AncestralBaseCount);
			break;
		case BC_Copy:
			++(Stats.CopyBaseCount);
			break;
		case BC_Novel:
			++(Stats.NovelBaseCount);
			break;
		default:
			asserta(false);
			}
		if (!Base.Plus && IsNovel(p))
			++(Stats.InvertedBaseCount);
		}
	}

// 0: 0
// 1: 1-9
// 1: 10-99
// 2: 100-999
// 3: 1000-9999
// 4: 10000-99999
// 5: 100000+
// 6: min
// 7: max
static void IncLengthDist(vector<gpos_t> &Dist, unsigned Length)
	{
	if (Length == 0)
		++(Dist[0]);
	if (Length < 10)
		++(Dist[1]);
	else if (Length < 100)
		++(Dist[2]);
	else if (Length < 1000)
		++(Dist[3]);
	else if (Length < 10000)
		++(Dist[4]);
	else if (Length < 100000)
		++(Dist[5]);
	else
		++(Dist[6]);

	Dist[7] = min(Length, Dist[7]);
	Dist[8] = max(Length, Dist[8]);
	}

void Evolver::GetGeneModelLengthDist(vector<gpos_t> &Introns,
  vector<gpos_t> &Exons) const
	{
	Introns.clear();
	Exons.clear();

	Introns.resize(9, 0);
	Exons.resize(9, 0);

	Introns[7] = GPOS_T_MAX;
	Exons[7] = GPOS_T_MAX;

	vector<GeneInfo> Genes;
	GetGenes(Genes);
	const unsigned GeneCount = SIZE(Genes);
	for (unsigned i = 0; i < GeneCount; ++i)
		{
		const GeneInfo &Gene = Genes[i];
		const unsigned ExonCount = SIZE(Gene.Exons);
		for (unsigned ExonIndex = 0; ExonIndex < ExonCount; ++ExonIndex)
			{
			const ExonInfo &Exon = Gene.Exons[ExonIndex];
			IncLengthDist(Exons, Exon.Length);
			}

		const unsigned IntronCount = SIZE(Gene.Introns);
		for (unsigned IntronIndex = 0; IntronIndex < IntronCount; ++IntronIndex)
			{
			const IntronInfo &Intron = Gene.Introns[IntronIndex];
			IncLengthDist(Introns, Intron.Length);
			}
		}
	}

static bool BaseIsIsland(const BaseInfo &Base)
	{
	return Base.IsUTR();
	}

static bool BaseIsUTR(const BaseInfo &Base)
	{
	return Base.IsUTR();
	}

static bool BaseIsCDS(const BaseInfo &Base)
	{
	return Base.IsCDS();
	}

static bool BaseIsNXE(const BaseInfo &Base)
	{
	return Base.IsNXE();
	}

static bool BaseIsNGE(const BaseInfo &Base)
	{
	return Base.IsNGE();
	}

static bool BaseIsExon(const BaseInfo &Base)
	{
	return Base.IsExon();
	}

unsigned Evolver::GetUTRCount() const
	{
	return GetCECount(BaseIsUTR);
	}

unsigned Evolver::GetIslandCount() const
	{
	return GetCECount(BaseIsIsland);
	}

unsigned Evolver::GetExonCount() const
	{
	return GetCECount(BaseIsExon);
	}

unsigned Evolver::GetCDSCount() const
	{
	return GetCECount(BaseIsCDS);
	}

unsigned Evolver::GetNXECount() const
	{
	return GetCECount(BaseIsNXE);
	}

unsigned Evolver::GetNGECount() const
	{
	return GetCECount(BaseIsNGE);
	}

unsigned Evolver::GetCECount(PTR_IS_CE_FUNC IsCE) const
	{
	unsigned CECount = 0;
	bool InCE = false;
	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];
		if (IsCE(Base))
			{
			if (!InCE)
				{
				InCE = true;
				++CECount;
				}
			}
		else
			InCE = false;
		}
	return CECount;
	}

unsigned Evolver::GetGeneCount() const
	{
	return SIZE(m_GeneIndexes);
	}

gpos_t Evolver::GetNextTriplet(gpos_t Pos, string &Letters, bool Strand) const
	{
	gpos_t Pos2 = GetAdjPos(Pos, Strand);
	if (Pos2 == GPOS_T_MAX)
		return GPOS_T_MAX;
	gpos_t Pos3 = GetAdjPos(Pos2, Strand);
	if (Pos3 == GPOS_T_MAX)
		return GPOS_T_MAX;

	const BaseInfo &Base1 = GetBase(Pos);
	const BaseInfo &Base2 = GetBase(Pos2);
	const BaseInfo &Base3 = GetBase(Pos3);

	Letters.resize(3);
	Letters[0] = Base1.GetStrandLetter(Strand);
	Letters[1] = Base2.GetStrandLetter(Strand);
	Letters[2] = Base3.GetStrandLetter(Strand);
	return GetAdjPos(Pos3, Strand);
	}

gpos_t Evolver::GetIntronLength(gpos_t Pos_) const
	{
	const BaseInfo &Base = GetBase(Pos_);
	asserta(Base.IsIntron());

	unsigned Length = 1;
	gpos_t Pos = Pos_;
	for (;;)
		{
		Pos = GetPrevPos(Pos);
		const BaseInfo &Base = GetBase(Pos);
		asserta(Base.IsGene());
		if (Base.IsExon())
			break;
		++Length;
		}

	Pos = Pos_;
	for (;;)
		{
		Pos = GetNextPos(Pos);
		const BaseInfo &Base = GetBase(Pos);
		asserta(Base.IsGene());
		if (Base.IsExon())
			break;
		++Length;
		}
	return Length;
	}

bool Evolver::IntronLengthIsGE(gpos_t Pos, gpos_t Length) const
	{
	gpos_t p = Pos;
	gpos_t Count = 0;
	for (;;)
		{
		if (Count >= Length)
			return true;
		const BaseInfo &Base = GetBase(p);
		if (!Base.IsIntron())
			break;
		p = GetNextPos(p);
		if (p == GPOS_T_MAX)
			break;
		++Count;
		}

	p = Pos;
	for (;;)
		{
		p = GetPrevPos(p);
		if (p == GPOS_T_MAX)
			return false;
		const BaseInfo &Base = GetBase(p);
		if (!Base.IsIntron())
			return false;
		if (++Count >= Length)
			return true;
		}
	}

BaseCategory Evolver::GetCategory(gpos_t Pos) const
	{
	if (Pos < m_AncestralLength)
		return BC_Ancestral;

	unsigned Index = (Pos - m_AncestralLength);
	asserta(Index < SIZE(m_AncPositions));
	uint32 AncPos = m_AncPositions[Index];
	if (AncPos >= m_AncestralLength)
		return BC_Novel;
	return BC_Copy;
	}

bool Evolver::IsCopy(gpos_t Pos) const
	{
	return GetCategory(Pos) == BC_Copy;
	}

bool Evolver::IsNovel(gpos_t Pos) const
	{
	return GetCategory(Pos) == BC_Novel;
	}

char Evolver::GetAncLetter(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.GetAncLetter();
	}

bool Evolver::PosInStartCodon(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsStartCodon();
	}

bool Evolver::PosInStopCodon(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsStopCodon();
	}

bool Evolver::InsertPosInStartCodon(gpos_t Pos) const
	{
	gpos_t NextPos = GetNextPos(Pos);
	if (NextPos == GPOS_T_MAX)
		return false;
	return PosInStartCodon(Pos) || PosInStartCodon(NextPos);
	}

bool Evolver::InsertPosInStopCodon(gpos_t Pos) const
	{
	gpos_t NextPos = GetNextPos(Pos);
	if (NextPos == GPOS_T_MAX)
		return false;
	return PosInStopCodon(Pos) || PosInStopCodon(NextPos);
	}

bool Evolver::PosInExon(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsExon();
	}

bool Evolver::PosInIntron(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsIntron();
	}

bool Evolver::RangeHasStartCodon(gpos_t Pos, gpos_t Length) const
	{
	gpos_t p = Pos;
	for (gpos_t i = 0; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsStartCodon())
			return true;
		p = GetNextPos(p);
		}
	return false;
	}

bool Evolver::RangeHasGene(gpos_t Pos, gpos_t Length) const
	{
	gpos_t p = Pos;
	for (gpos_t i = 0; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsGene())
			return true;
		p = GetNextPos(p);
		}
	return false;
	}

bool Evolver::RangeHasStopCodon(gpos_t Pos, gpos_t Length) const
	{
	gpos_t p = Pos;
	for (gpos_t i = 0; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsStopCodon())
			return true;
		p = GetNextPos(p);
		}
	return false;
	}

bool Evolver::RangeHasCDS(gpos_t Pos, gpos_t Length, bool &AllCDS) const
	{
	gpos_t p = Pos;
	bool AnyCDS = false;
	AllCDS = true;
	for (gpos_t i = 0; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsCDS())
			{
			if (!AllCDS)
				return true;
			AnyCDS = true;
			}
		else
			AllCDS = false;
		p = GetNextPos(p);
		}
	return AnyCDS;
	}

bool Evolver::PosInGene(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsGene();
	}

bool Evolver::PosInCDS(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsCDS();
	}

bool Evolver::PosInCE(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsConstrained();
	}

bool Evolver::PosInNGE(gpos_t Pos) const
	{
	const BaseInfo &Base = GetBase(Pos);
	return Base.IsNGE();
	}

bool Evolver::RangeHasCE(gpos_t Pos, gpos_t Length) const
	{
	gpos_t p = Pos;
	for (gpos_t i = 0; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsConstrained())
			return true;
		p = GetNextPos(p);
		}
	return false;
	}

bool Evolver::RangeSpansTwoGenes(gpos_t Pos, gpos_t Length) const
	{
	gpos_t p = Pos;
	bool FoundGene = false;
	gpos_t i = 0;
	for (; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsGene())
			{
			FoundGene = true;
			break;
			}
		p = GetNextPos(p);
		}
	if (!FoundGene)
		return false;

	bool FoundGeneEnd = false;
	for (; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (!Base.IsGene())
			{
			FoundGeneEnd = true;
			break;
			}
		p = GetNextPos(p);
		}

	for (; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsGene())
			return true;
		p = GetNextPos(p);
		}
	return false;
	}

bool Evolver::RangeHasIntron(gpos_t Pos, gpos_t Length) const
	{
	gpos_t p = Pos;
	for (gpos_t i = 0; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(p);
		if (Base.IsIntron())
			return true;
		p = GetNextPos(p);
		}
	return false;
	}

bool Evolver::ContainsCEFragment(gpos_t Pos, gpos_t Length) const
	{
	if (PosInGene(Pos) || PosInNGE(Pos))
		return true;

	gpos_t EndPos = GetRangeEndPos(Pos, Length);
	return PosInGene(EndPos) || PosInNGE(EndPos);
	}

void Evolver::SetLogEventFromTo(unsigned From, unsigned To)
	{
	m_LogEventFrom = From;
	m_LogEventTo = To;
	}

bool Evolver::PosInCEOrGene(gpos_t Pos) const
	{
	return PosInGene(Pos) || PosInCE(Pos);
	}

