#ifndef evolver_h
#define evolver_h

#include <string>
#include <set>
#include <algorithm>
#include "utils.h"
#include "base.h"
#include "gene.h"
#include "muttypes.h"
#include "codon.h"
#include "eventresult.h"
#include "lohi.h"
#include "ivtree.h"
#include "evoassert.h"
#include "geninfo.h"

struct GFFRecord2;
class RevFile;
struct AlnInfo;

using namespace std;

class Evolver;
typedef void (*GotAlignmentCallback)(const Evolver &Evo, const AlnInfo &AI,
  const StdStrVec &Rows);
typedef EVENT_RESULT (*PTR_TRY_EVENT_FUNC)(Evolver &Evo);
typedef bool (*PTR_IS_CE_FUNC)(const BaseInfo &Base);

struct CatStats
	{
	gpos_t BaseCount;
	gpos_t GeneBaseCount;
	gpos_t IntronBaseCount;
	gpos_t CDSBaseCount;
	gpos_t UTRBaseCount;
	gpos_t SpliceBaseCount;
	gpos_t NXEBaseCount;
	gpos_t NGEBaseCount;
	gpos_t ConstrainedBaseCount;
	gpos_t NeutralBaseCount;
	gpos_t AncestralBaseCount;
	gpos_t NovelBaseCount;
	gpos_t CopyBaseCount;
	gpos_t InvertedBaseCount;
	gpos_t TandemBaseCount;
	gpos_t CpGBaseCount;
	};

struct MEInfo
	{
	string Name;
	bool RateSpecified;
	bool RatePerBaseSpecified;
	double Rate;
	double RateParam;
	double RatePerBaseParam;
	double FractWithDeletion;
	double MeanDeleteLength;
	double StdDevDeleteLength;
	string Seq;
	unsigned AcceptedInsertCount;
	unsigned DeleteCount;
	unsigned RandThresh;
	unsigned InsertedBaseCount;

	MEInfo()
		{
		Name = "?";
		Rate = 0;
		Seq = "";
		RateSpecified = false;
		RatePerBaseSpecified = false;
		RateParam = -1;
		RatePerBaseParam = -1;
		AcceptedInsertCount = 0;
		DeleteCount = 0;
		RandThresh = UINT_MAX;
		FractWithDeletion = 0;
		MeanDeleteLength = 0;
		StdDevDeleteLength = 0;
		InsertedBaseCount = 0;
		}
	};

enum GENE_DUPE_RESULT
	{
	GDR_Slower,
	GDR_Same,
	GDR_Faster,
	GDR_Lost
	};

class Evolver
	{
private:
// Disable unimplimented copy c'tor and some important operators.
	Evolver(Evolver &);
	Evolver &operator=(const Evolver &rhs);
	bool operator==(const Evolver &rhs);
	bool operator<(const Evolver &rhs);

	void GetContigRanges(gpos_t Pos, gpos_t Length, vector<gpos_t> &Starts,
	  vector<gpos_t> &Lengths, bool *ptrExists = 0) const;
	gpos_t GetRandomPos() const;
	gpos_t GetRandomTandemPos() const;
	bool PosExists(gpos_t Pos) const;

public:
	Evolver();
	void Clear();

	void Validate(bool ShowProgress = false) const;
	void ValidateCDSs() const;
	void PrintAndDie(const char *Format, ...) const;

	void LogMe() const;
	void DumpMe() const;
	void LogEvolvedAsTable() const;
	void LogRegions(gpos_t Pos, gpos_t Length) const;
	void LogNeighborRegions(gpos_t Pos, unsigned Margin = 8) const;
	void LogRangeAsTable(gpos_t Pos, gpos_t Length) const;
	void LogStats() const;
	void LogNeighborsAsTable(gpos_t Pos, unsigned Margin = 8) const;
	void LogTandems() const;
	void LogGeneIndexes() const;
	void LogSizeStats() const;

	void GetCatStats(CatStats &Stats) const;

	void Run();

	gpos_t GetTandemBaseCount() const
		{
		return SIZE(m_Tandems);
		}

// The "real" number of ancestral bases.
	gpos_t GetAncestralLength() const
		{
		return m_AncestralLength;
		}

	gpos_t GetMaxPos() const;
	gpos_t GetMaxIndex() const;
	gpos_t GetMaxIndexSlow() const;
	gpos_t GetExtendedBaseCount() const;
	//void Defrag();

	gpos_t GetRelativePos(gpos_t AbsolutePos) const;
	gpos_t GetAbsolutePos(gpos_t RelativePos) const;
	gpos_t GetHomologousPos(gpos_t Pos) const;
	BaseCategory GetCategory(gpos_t Pos) const;

	bool IsValidPos(gpos_t Pos) const
		{
		return Pos < SIZE(m_Bases);
		}

	gpos_t GetEvolvedLength() const
		{
		return m_EvolvedLength;
		}

	gpos_t GetExtendedLength() const
		{
		return SIZE(m_Bases);
		}

// File I/O
	void FromRevFile(const RevFile &Rev, unsigned GenomeIndex,
	  unsigned ChrId, bool ShowProgress = true);
	void FromFile(const char *FileName, bool ShowProgress = true);
	void ToFile(const char *FileName, bool ShowProgress = true) const;
	void WriteEventGenState(FILE *f) const;
	void ReadEventGenState(FILE *f);
	void MEsToFile(FILE *f) const;
	void MEsFromFile(FILE *f);

	const GenInfo &GetAncGenInfo() const { return m_AncGenInfo; }
	unsigned GetAncChrId() const { return m_AncChrId; }
	void GetEvolvedSeq(string &Seq) const;
	void SetLogEvents(bool On) { m_LogEvents = On; }
	void SetLogEventFromTo(unsigned From, unsigned To);
	void InitBranchLength(double BranchLength);
	void InitEventCount(unsigned EventCount);

// Creation
	void FromRandomString(gpos_t Length);
	void FromString(const string &Seq, bool Progress = false);
	void FromCString(const char *Seq, unsigned Length, bool Progress = false);
	void ToString(string &Seq) const;

#define E(e)	EVENT_RESULT Try_##e();
#include "events.h"
#undef E

	void Delete(gpos_t Pos, gpos_t Length);
	void Invert(gpos_t Pos, gpos_t Length);
	void Copy(gpos_t FromPos, gpos_t ToPos, gpos_t Length,
	  vector<gpos_t> &GeneStarts, vector<gpos_t> &CopiedGeneStarts,
	  bool DoInvert = false);
	void Move(gpos_t FromPos, gpos_t ToPos, gpos_t Length);
	void Insert(gpos_t Pos, const string &Seq);

	void InitCpG_Sweeps();
	void CpG_Sweep(double Duration);
	void CpGBase(gpos_t Pos, double Prob);
	void GetCpGCounts(gpos_t &CpGBaseCount, gpos_t &IslandBaseCount) const;
	void GetComposition(gpos_t LetterCounts[4], gpos_t DinucCounts[4][4]) const;

	void GenerateInsertSeq(gpos_t Length);

	bool PosInCDS(gpos_t Pos) const;
	double GetAcceptProb(gpos_t Pos) const;
	double GetRangeAcceptProb(gpos_t Start, gpos_t Length) const;
	bool AcceptRange(gpos_t Pos, gpos_t Length, bool *ptrAnyConstraint = 0) const;

	bool AcceptInvert(gpos_t Pos, gpos_t Length, int *ptrWhy) const;
	bool AcceptInsert(gpos_t Pos, const string &Seq, int *ptrWhy = 0) const;
	bool AcceptDelete(gpos_t Pos, gpos_t Length, int *ptrWhy = 0) const;

	bool AcceptDeleteLow(gpos_t Pos, gpos_t Length, int *ptrWhy = 0,
	  bool *ptrAnyConstraint = 0) const;
	bool AcceptInvertLow(gpos_t Pos, gpos_t Length, int *ptrWhy) const;
	bool AcceptInsertLow(gpos_t Pos, const string &Seq, int *ptrWhy = 0,
	  bool *ptrAnyConstraint = 0) const;

	bool AcceptExonSubst(gpos_t Pos, char NewLetter) const;
	bool AcceptCodonSubst(double AcceptProb, const string &CodonBefore,
	  const string &CodonAfter) const;
	bool InsertCreatesStopCodon(gpos_t Pos, const string &Seq) const;
	bool DeleteCreatesStopCodon(gpos_t Pos, gpos_t Length) const;
	void GetExonRange(gpos_t Pos, gpos_t &ExonStart, gpos_t &ExonLength) const;
	bool AcceptDelete(uint8 CC);
	void GetSubSeq(gpos_t Pos, gpos_t Length, string &Seq,
	  bool Plus = true) const;
	void GetAnnots(const char *GFFFileName, const string &SeqName);
	void SetAnnot(const GFFRecord2 &Rec);
	void SetGeneAnnot(const GFFRecord2 &Rec);
	void SetBaseProbs(gpos_t Start, gpos_t Length, const string &Probs);
	void SetIsland(gpos_t Start, gpos_t Length, bool On);
	void SetTandemMotifLength(gpos_t Pos, gpos_t Length, gpos_t MotifLength);
	bool GetCodon(gpos_t Pos, CodonInfo &Codon) const;
	void SetInsertConstraints(gpos_t InsertPos, gpos_t Length);
	void SetCodexes(gpos_t Pos, gpos_t Length, int8 Frame, bool Plus);
	void SetIntronsAndUntranscribed(gpos_t Pos, gpos_t Length);
	void SetSplices(gpos_t Pos, gpos_t Length);
	void SetCDS(gpos_t Pos, gpos_t Length, bool Strand);
	void SetUTR(gpos_t Pos, gpos_t Length);
	unsigned SetStartCodon(gpos_t Pos, bool Strand);
	unsigned SetStopCodon(gpos_t Pos, bool Strand);
	bool TryMoveUTRTermGene(const GeneInfo &Gene);
	bool IntronLengthIsGE(gpos_t Pos, gpos_t Length) const;
	bool PosInStartCodon(gpos_t Pos) const;
	bool PosInStopCodon(gpos_t Pos) const;
	bool InsertPosInStartCodon(gpos_t Pos) const;
	bool InsertPosInStopCodon(gpos_t Pos) const;
	bool PosInExon(gpos_t Pos) const;
	bool PosInIntron(gpos_t Pos) const;
	bool RangeHasStartCodon(gpos_t Pos, gpos_t Length) const;
	bool RangeHasStopCodon(gpos_t Pos, gpos_t Length) const;
	bool RangeHasCDS(gpos_t Pos, gpos_t Length, bool &AllCDS) const;
	bool RangeHasGene(gpos_t Pos, gpos_t Length) const;
	bool PosInCE(gpos_t Pos) const;
	bool PosInGene(gpos_t Pos) const;
	bool PosInNGE(gpos_t Pos) const;
	bool RangeHasCE(gpos_t Pos, gpos_t Length) const;
	bool RangeHasIntron(gpos_t Pos, gpos_t Length) const;
	bool RangeSpansTwoGenes(gpos_t Pos, gpos_t Length) const;
	bool ContainsCEFragment(gpos_t Pos, gpos_t Length) const;
	void UpdateStartCodonMap(gpos_t OldPos, gpos_t NewPos);

// Exon delete candidates
	unsigned GetCandDeleteCDSExonIndex(const GeneInfo &Gene) const;
	unsigned GetCandDeleteUTRExonIndex(const GeneInfo &Gene) const;

// Move START/STOP candidates
	gpos_t GetMoveStartCodonIntoUTRCand(const GeneInfo &Gene) const;
	gpos_t GetMoveStartCodonIntoCDSCand(const GeneInfo &Gene) const;
	gpos_t GetMoveStopCodonIntoUTRCand(const GeneInfo &Gene) const;
	gpos_t GetMoveStopCodonIntoCDSCand(const GeneInfo &Gene) const;

// Splice site move candidates (per gene)
// UTR:
	bool GetMoveUTRAcceptorIntoIntronCand(const GeneInfo &Gene,
	  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const;
	bool GetMoveUTRDonorIntronCand(const GeneInfo &Gene, gpos_t &OldDonorPos,
	  gpos_t &NewDonorPos) const;
	bool GetMoveAcceptorIntoUTRCand(const GeneInfo &Gene, gpos_t &OldAcceptorPos,
	  gpos_t &NewAcceptorPos) const;
	bool GetMoveUTRAcceptorIntronCand(const GeneInfo &Gene, gpos_t &OldAcceptorPos,
	  gpos_t &NewAcceptorPos) const;
	bool GetMoveUTRDonorIntoIntronCand(const GeneInfo &Gene, gpos_t &OldDonorPos,
	  gpos_t &NewDonorPos) const;
	bool GetMoveDonorIntoUTRCand(const GeneInfo &Gene, gpos_t &OldDonorPos,
	  gpos_t &NewDonorPos) const;

// CDS:
	bool GetMoveCDSDonorIntronCand(const GeneInfo &Gene, gpos_t &OldDonorPos,
	  gpos_t &NewDonorPos) const;
	bool GetMoveDonorIntoCDSCand(const GeneInfo &Gene, gpos_t &OldDonorPos,
	  gpos_t &NewDonorPos) const;
	bool GetMoveCDSAcceptorIntronCand(const GeneInfo &Gene, gpos_t &OldAcceptorPos,
	  gpos_t &NewAcceptorPos) const;
	bool GetMoveAcceptorIntoCDSCand(const GeneInfo &Gene, gpos_t &OldAcceptorPos,
	  gpos_t &NewAcceptorPos) const;

// Splice site move candidates (per exon)
// UTR:
	bool GetMoveUTRDonorIntronCandExon(const GeneInfo &Gene, unsigned ExonIndex,
	  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const;
	bool GetMoveUTRAcceptorIntoIntronCandExon(const GeneInfo &Gene, unsigned ExonIndexX,
	  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const;
	bool GetMoveUTRDonorIntoIntronCandExon(const GeneInfo &Gene,
	  unsigned ExonIndexX, gpos_t &OldDonorPos, gpos_t &NewDonorPos) const;

// CDS:
	bool GetMoveCDSDonorIntronCandExon(const GeneInfo &Gene, unsigned ExonIndex,
	  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const;
	bool GetMoveCDSAcceptorIntronCandExon(const GeneInfo &Gene, unsigned ExonIndex,
	  gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const;

	void GetRandomizedExonIndexes(const GeneInfo &Gene,
	  vector<unsigned> &ExonIndexes) const;
	bool IsFirstUTR(const GeneInfo &Gene, unsigned ExonIndex) const;

	bool FindNearbyCpGIsland(gpos_t Pos, bool Strand, gpos_t MaxDist,
	  gpos_t &StartPos, gpos_t &EndPos) const;

	void GetGeneWeightVec(const vector<GeneInfo> &Genes,
	  vector<double> &Weights) const;
	void InitTryGenes(vector<GeneInfo> &Genes, vector<double> &MACs) const;
	bool NextTryGene(vector<GeneInfo> &Genes, vector<double> &MACs,
	  GeneInfo &Gene) const;

	gpos_t GetTandemMotifLength(gpos_t Pos) const;
	void RecomputeTandemRates();
	gpos_t GetAdjExonPos(gpos_t Pos, bool Strand) const;
	gpos_t GetPrevExonPos(gpos_t Pos, bool Strand) const;
	void GetNextPair(gpos_t Pos, string &Letters, bool Strand) const;
	gpos_t GetNextExonTriplet(gpos_t Pos, string &Letters, bool Strand) const;
	gpos_t GetPrevExonTriplet(gpos_t Pos, string &Letters, bool Strand) const;
	gpos_t GetNextIntronTriplet(gpos_t Pos, string &Letters, bool Strand) const;
	bool IsCodon(gpos_t Pos, bool Strand) const;
	void LogExonTriplet(gpos_t Pos, bool Strand) const;
	void SetTriplet(gpos_t Pos, const string &Letters, bool Strand);
	bool GetFractCodonBeforeIntron(gpos_t Pos, string &Letters, bool Strand,
	  bool &StartStop) const;
	bool GetFractCodonAfterIntron(gpos_t Pos, string &Letters, bool Strand,
	  bool &StartStop) const;
	bool GetFractCodonExonEnd(const ExonInfo &Exon, string &FractCodon,
	  bool Strand, bool &StartStop) const;
	bool GetFractCodonExonStart(const ExonInfo &Exon, string &FractCodon,
	  bool Strand, bool &StartStop) const;
	gpos_t GetNextTriplet(gpos_t Pos, string &Letters, bool Strand) const;
	gpos_t GetIntronLength(gpos_t Pos) const;

// Mobile elements
	void ReadMEs(const char *FileName);
	void AddME(const MEInfo &ME);
	void LogMEs() const;
	double GetMERate() const;
	void SetMERandThreshs();
	MEInfo &GetRandomME();

// Coordinate stuff
	bool PosIsDeleted(gpos_t Pos) const;
	bool RangeExists(gpos_t Pos, gpos_t Length) const;
	bool PosInRange(gpos_t Pos, gpos_t StartPos, gpos_t Length) const;

	void FindAlignments(GotAlignmentCallback GotAlignment) const;
	void GenerateAnnots(FILE *f, const char *ChrName) const;
	void GenerateAnnot(FILE *f, const char *ChrName, CE_TYPE CEType,
	  gpos_t Pos, gpos_t HomPos, gpos_t EvStartPos, gpos_t EndEvPos,
	  const IvTree<LoHi> &EvGeneIx, const map<unsigned, bool> &GeneIndexToStrand) const;
	void GenerateIslandAnnot(FILE *f, const char *ChrName, gpos_t EvStartPos,
	  gpos_t EvEndPos, gpos_t AncStartPos) const;
	void GenerateProbsAttr(FILE *f, gpos_t Start, gpos_t Length) const;
	void DupeGenes(const vector<gpos_t> &GeneStarts,
	  const vector<gpos_t> &CopiedGeneStarts, bool DoInvert);
	void DupeGene(gpos_t FromPos, gpos_t ToPos, bool DoInvert);
	int GetCDSFrame(gpos_t StartPos, gpos_t Length) const;

	void GetGeneIndexes(set<unsigned> &Indexes) const;
	void GetGene(gpos_t Pos, GeneInfo &Gene) const;
	void GetGeneRanges(vector<gpos_t> &Starts, vector<gpos_t> &Lengths) const;
	void GetGeneRangesEv(IvTree<LoHi> &Genes) const;
	gpos_t GetGeneRange(gpos_t Pos, gpos_t &Start, gpos_t &Length) const;
	unsigned GetGeneIndex(gpos_t StartCodonFirstBasePos) const;
	void GetGenes(vector<GeneInfo> &Genes) const;
	void ScanGenes(vector<GeneInfo> &Genes) const;
	void GetNeutralRanges(gpos_t Pos, gpos_t Length, bool InGene,
	  vector<gpos_t> &Starts, vector<gpos_t> &Lengths) const;
	void GetNGEs(vector<gpos_t> &Starts, vector<gpos_t> &Lengths,
	  vector<double> &MACs) const;
	void GetNXEs(gpos_t RangeStart, gpos_t RangeLength, vector<gpos_t> &Starts,
	  vector<gpos_t> &Lengths, vector<double> &MACs) const;
	void GetIslands(gpos_t RangeStart, gpos_t RangeLength, vector<gpos_t> &Starts,
	  vector<gpos_t> &Lengths) const;
	void GetOceans(gpos_t RangeStart, gpos_t RangeLength, vector<gpos_t> &Starts,
	  vector<gpos_t> &Lengths) const;

	unsigned ValidateGenes(bool ShowProgress = false) const;
	unsigned ValidateGene(const GeneInfo &Gene) const;

	bool GetCandNontermUTR(const GeneInfo &Gene, gpos_t TargetUTRLength,
	  gpos_t &UTRLength, gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const;
	bool GetCandNontermUTRIntron(const GeneInfo &Gene, unsigned IntronIndex,
	  gpos_t TargetUTRLength, gpos_t &UTRLength, gpos_t &NewDonorPos,
	  gpos_t &NewAcceptorPos) const;
	bool GetCandFirstUTR(const GeneInfo &Gene, gpos_t Length,
	  gpos_t &UTRStartPos, gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const;
	bool GetCandLastUTR(const GeneInfo &Gene, gpos_t Length,
	  gpos_t &UTRStartPos, gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const;
	bool GetCandCDS(const GeneInfo &Gene, gpos_t CDSLength, gpos_t &NewDonorPos,
	  gpos_t &NewAcceptorSecondPos) const;
	bool GetCandCDSIntron(const GeneInfo &Gene, gpos_t CDSLength,
	  unsigned IntronIndex, gpos_t &NewDonorPos, gpos_t &NewAcceptorSecondPos) const;
	bool GetMoveDonorIntoCDSCandExon(const GeneInfo &Gene, unsigned ExonIndex,
	  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const;
	bool GetMoveDonorIntoUTRCandExon(const GeneInfo &Gene, unsigned ExonIndex,
	  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const;
	bool GetMoveAcceptorIntoUTRCandExon(const GeneInfo &Gene, unsigned ExonIndex,
	  gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const;

	bool HasStopCodon(const string &Seq) const;
	void GetUTRBaseCounts(const GeneInfo &Gene, gpos_t &FirstUTRBaseCount,
	  gpos_t &LastUTRBaseCount) const;
	gpos_t GetUTRBaseCount(const GeneInfo &Gene, unsigned ExonIndex) const;
	void ChangeSpeed(gpos_t Start, gpos_t Length, double Accel);
	void Deconstrain(gpos_t Start, gpos_t Length, bool DisableGene);

	void GetIntegratedRates(vector<double> &Rates) const;

	unsigned GetGeneCount() const;
	unsigned GetUTRCount() const;
	unsigned GetCDSCount() const;
	unsigned GetNXECount() const;
	unsigned GetNGECount() const;
	unsigned GetExonCount() const;
	unsigned GetIslandCount() const;
	unsigned GetCECount(PTR_IS_CE_FUNC IsCE) const;

	void ComputeEventRates(double ATFract);
	void LogEventRates() const;

	void GetGeneModelLengthDist(vector<gpos_t> &Introns, vector<gpos_t> &Exons) const;
	gpos_t GetGeneLength(gpos_t Pos) const;

	void LogGeneHeading() const;
	void LogGene(const GeneInfo &Gene, bool Heading = true) const;
	void LogGenes(const vector<GeneInfo> &Genes) const;

private:
// Per-base events
	gpos_t InsertNovelBase(gpos_t TargetPos, char Letter);
	void DeleteBase(gpos_t Pos);
	void MoveBase(gpos_t FromPos, gpos_t InsertPos);

	gpos_t GetRangeEndPos(gpos_t Pos, gpos_t Length, bool *ptrExists = 0) const;
	gpos_t GetRangeStartPos(gpos_t Pos, gpos_t Length) const;

	gpos_t GetNextPos(gpos_t Pos) const
		{
		gpos_t L = SIZE(m_BaseListFwds);
		if (Pos >= L)
			Die("Evolver::GetNextPos(%"PF"), L=%"PF, Pos, L);
		return m_BaseListFwds[Pos];
		}

	gpos_t GetPrevPos(gpos_t Pos) const
		{
		gpos_t L = SIZE(m_BaseListFwds);
		if (Pos >= L)
			Die("Evolver::GetNextPos(%"PF"), L=%"PF, Pos, L);
		return m_BaseListBwds[Pos];
		}

	gpos_t GetAdjPos(gpos_t Pos, bool Plus) const { return Plus ? GetNextPos(Pos) : GetPrevPos(Pos); }
	BaseInfo &GetBase(gpos_t Pos) { asserta(Pos < SIZE(m_Bases)); return m_Bases[Pos]; }
	const BaseInfo &GetBase(gpos_t Pos) const { asserta(Pos < SIZE(m_Bases)); return m_Bases[Pos]; }

	void AppendBaseToAln(GotAlignmentCallback GotAlignment, gpos_t EvPos, 
	  gpos_t PrevHomPos, gpos_t Pos, AlnInfo &AI, StdStrVec &Rows) const;
	bool ValidAlnSuccessor(gpos_t PrevHomPos, gpos_t Pos) const;
	void EndAln(GotAlignmentCallback GotAlignment, AlnInfo &AI,
	  StdStrVec &Rows) const;
	void StartAln(gpos_t EvPos, gpos_t AncPos, AlnInfo &AI, StdStrVec &Rows) const;
	bool PosInCEOrGene(gpos_t Pos) const;
	bool InCpG(gpos_t Pos) const;
	char GetAncLetter(gpos_t Pos) const;
	void DeleteGeneIndex(unsigned GeneIndex);

	bool IsCopy(gpos_t Pos) const;
	bool IsNovel(gpos_t Pos) const;

	void BaseList_Init(const char *Seq, unsigned L, const BaseInfo &BaseTemplate,
	  bool ShowProgress = true);
	uint32 BaseList_Insert(const BaseInfo &Base, uint32 AfterPos, uint32 AncPos);
	void BaseList_Delete(uint32 Pos);
	void BaseList_Invert(gpos_t Pos, gpos_t Length);
	void BaseList_Validate() const;
	void BaseList_Log() const;
	void BaseList_LogSummary() const;
	void BaseList_LogPos(uint32 Pos) const;

// Parameters
	EVENT_TYPE GetRandomEventType();

	EVENT_RESULT TryEvent();
	double GetRandomWaitingTime() const;

private:
	GenInfo m_AncGenInfo;
	unsigned m_AncChrId;
	bool m_LogEvents;
	double m_AverageWaitingTime;
	vector<double> m_IntegratedEventRates;
	vector<unsigned> m_EventRandDistHis;
	EVENT_TYPE m_CurrentEvent;
	EVENT_TYPE m_LastAcceptedEvent;
	gpos_t m_LastAcceptedEventPos;
	gpos_t m_CurrentEventPos;
	unsigned m_NextGeneIndex;

	double m_BranchLength;
	unsigned m_EventCount;
	unsigned m_EventIndex;
	unsigned m_LastAcceptedEventIndex;

	unsigned m_LogEventFrom;
	unsigned m_LogEventTo;

	bool m_DoCpGSweeps;
	unsigned m_CpGSweepEventCounter;
	unsigned m_CpGSweepInterval;
	
	gpos_t m_AncestralLength;
	gpos_t m_EvolvedLength;

// Bases in ancestral order
	vector<BaseInfo> m_Bases;

// Double-linked list of bases, values are ancestral
// positions = indexes into m_Bases[].
	vector<uint32> m_BaseListFwds;
	vector<uint32> m_BaseListBwds;

// List heads and tails for evolved and deleted bases
	uint32 m_FirstEvolvedPos;
	uint32 m_FirstDeletedPos;
	uint32 m_LastEvolvedPos;
	uint32 m_LastDeletedPos;

// Ancestral positions of bases in the extended genome.
// Index into m_AncPositions[] is Pos - m_AncestralLength.
	vector<uint32> m_AncPositions;

// Bases in tandem arrays. Used to select positions for tandem
// expand / contract events. The key is the ancestral position,
// the value is the motif length.
	map<gpos_t, gpos_t> m_Tandems;

// Hack to reduce cost of recomputing rates
	unsigned m_TandemRateRecomputeCounter;

// Mobile elements
	vector<MEInfo> m_MEs;

// Map ancestral position of START codon to gene index.
// The START codon rarely moves.
	map<gpos_t, unsigned> m_GeneIndexes;

// If a gene is duplicated, maps the new gene index onto
// to original gene.
	map<unsigned, unsigned> m_Paralogs;
	};

const char *BaseCategoryToStr(BaseCategory c);
const char *GeneDupeResultToStr(GENE_DUPE_RESULT c);
void GetGeneMACs(const vector<GeneInfo> &Genes, vector<double> &MACs);
void GetGeneDupeResults(GENE_DUPE_RESULT &Old, GENE_DUPE_RESULT &New);

#define FOR_EVBASES(p)	for (uint32 p = m_FirstEvolvedPos; p != UINT_MAX; p = m_BaseListFwds[p])

#define ACCEPT_INVERT(EventName, Pos, Length)			\
	{													\
	int Why = 0;										\
	bool Accept = AcceptInvert(Pos, Length, &Why);		\
	if (!Accept)										\
		{												\
		switch (Why)									\
			{											\
		case 1:	EV_REJECT(EventName, Length, PartialCE);\
		case 2:	EV_REJECT(EventName, Length, GeneCE);	\
		case 3:	EV_REJECT(EventName, Length, PartialGene);\
		default: asserta(false);						\
			}											\
		}												\
	}

#define ACCEPT_INSERT(EventName, ToPos, Seq)			\
	{													\
	const unsigned Length = SIZE(Seq);					\
	int Why = 0;										\
	bool Accept = AcceptInsert(ToPos, Seq, &Why);		\
	if (!Accept)										\
		{												\
		switch (Why)									\
			{											\
		case 1: EV_REJECT(EventName, Length, InsEndOfSeq);\
		case 2:	EV_REJECT(EventName, Length, InsAccProb);\
		case 3:	EV_REJECT(EventName, Length, InsCDSMod3);\
		case 4:	EV_REJECT(EventName, Length, InsCreatesSTOP);\
		case 5:	EV_REJECT(EventName, Length, InsInSTARTorSTOP);\
		default: asserta(false);						\
			}											\
		}												\
	}

#define ACCEPT_INSERTN(EventName, ToPos, Seq)			\
	{													\
	int Why = 0;										\
	bool Accept = AcceptInsert(ToPos, Seq, &Why);		\
	if (!Accept)										\
		{												\
		switch (Why)									\
			{											\
		case 1: EV_REJECTN(EventName, InsEndOfSeq);		\
		case 2:	EV_REJECTN(EventName, InsAccProb);		\
		case 3:	EV_REJECTN(EventName, InsCDSMod3);		\
		case 4:	EV_REJECTN(EventName, InsCreatesSTOP);	\
		case 5:	EV_REJECTN(EventName, InsInSTARTorSTOP);\
		default: asserta(false);						\
			}											\
		}												\
	}

#define ACCEPT_DELETE(EventName, Pos, Length)			\
	{													\
	int Why = 0;										\
	bool Accept = AcceptDelete(Pos, Length, &Why);		\
	if (!Accept)										\
		{												\
		switch (Why)									\
			{											\
		case 1: EV_REJECT(EventName, Length, DelCDSMod3);	\
		case 2:	EV_REJECT(EventName, Length, DelNotAllCDS);\
		case 3:	EV_REJECT(EventName, Length, DelCodonDel);	\
		case 4:	EV_REJECT(EventName, Length, DelHasSTART);	\
		case 5:	EV_REJECT(EventName, Length, DelHasSTOP);	\
		case 6:	EV_REJECT(EventName, Length, DelCreatesSTOP);\
		case 7:	EV_REJECT(EventName, Length, DelAccProb);	\
		case 8:	EV_REJECT(EventName, Length, DelIntronLength);	\
		case 9:	EV_REJECT(EventName, Length, DelTwoGenes);	\
		case 10:EV_REJECT(EventName, Length, DelSplice);	\
		case 11:EV_REJECT(EventName, Length, DelEndOfSeq);	\
		default: asserta(false);						\
			}											\
		}												\
	}

#endif // evolver_h
