#ifndef FILTER4_H
#define FILTER4_H

#include "myassert.h"
#include "revfile.h"

/***
Support for "filters".
A filter is a program that takes a rev file as input
and processes one bundle at a time. The program supplies
two functions, one for initialization (InitFilter4)
and one for processing a bundle (Filter4). The code for
processing standard command-line arguments, opening
the input file, collecting bundles into an AIVec etc.
is standardized and shared between all filters. This
shared code includes main().
***/

typedef void (*PTR_VOID_FN_VOID)(void);

// Undefined references to be supplied by each
// filter program.
void InitFilter4();
void Filter4(const AIVec &AIs, const IntVec &AIIndexes);

// Optional call-back functions to allow progress
// messages, counter updates etc.
void SetBeforeChrPairCallback(PTR_VOID_FN_VOID ptrFn);
void SetAfterChrPairCallback(PTR_VOID_FN_VOID ptrFn);
void SetAfterAllChrPairsCallback(PTR_VOID_FN_VOID ptrFn);

// Functions to retrieve current filter state
// for use in Filter4() and in call-backs:
const RevFile &GetFilterInputRevFile();

// Genome indexes, fixed by command line,
// do not vary while processing.
unsigned GetFilterGenomeIndexA();
unsigned GetFilterGenomeIndexB();

// Chr ids currently being processed.
unsigned GetFilterChrIdA();
unsigned GetFilterChrIdB();

const ChrInfo &GetFilterChrInfoA();
const ChrInfo &GetFilterChrInfoB();

// Bundle currently being processed.
unsigned GetFilterBundleId();

// Value of -bundle <b>, if specified
// (requests that a single bundle be processed).
unsigned GetCmdLineBundleId();

// Request (or not) inversion of minus bundles
void SetInvertMinusBundles(bool Invert);
bool GetInvertMinusBundles();

// Filter-aware input and output functions.
void FilterAwareGetSeqData(const RevFile &Rev, unsigned GenomeIndex,
  unsigned ChrId, coord_t Pos, unsigned Length, char *Seq, bool Plus);

void FilterAwareOutputAlnInfo(RevFile &Out, const AlnInfo &UserAI);
void FilterAwareOutputAlnRows(RevFile &Out, const AlnInfo &UserAI,
  const StdStrVec &UserRows);

// Filter-aware alignment functions
enum AlignStyle
	{
	ALIGN_STYLE_PlusOnly,
	ALIGN_STYLE_MinusOnly,
	ALIGN_STYLE_PlusAndMinus
	};

// Does not clear Hits or Paths to allow cumulative calls.
void FilterAwareAppendAlignAILocal(const AlnInfo &AI, AIVec &Hits, StdStrVec &Paths,
  AlignStyle Style);

void FilterAwareAlignAIsLocal(const AIVec &AIs, AIVec &Hits, StdStrVec &Paths,
  AlignStyle Style);

void FilterAwareAppendAlignAIGlobal(const AlnInfo &AI, StdStr &Path);

void FilterAwareAlignAIsGlobal(const AIVec &AIs, StdStrVec &Paths);

void FilterAwareMakeAln(const AlnInfo &AI, const StdStr &Path, StdStrVec &Rows);

void FilterMakePairwiseAlnInfo(coord_t Alo, coord_t Ahi, coord_t Blo, coord_t Bhi,
  bool Plus, AlnInfo &AI, AlnAttr Attr = ALN_ATTR_None, unsigned AlnAttrValue = 0);

// Private functions used by filter framework:
void FilterWrapper4(const AIVec &AIs, const IntVec &AIIndexes);
void FilterChrPair4(const RevFile &In);
void FilterAllChrPairs4(const RevFile &In);
bool GetRevChrPairOpts(const RevFile &Rev, unsigned *ptrGenomeIndexA,
  unsigned *ptrGenomeIndexB, unsigned *ptrChrIdA, unsigned *ptrChrIdB);
void SetFilterBundleId(unsigned BundleId);
void SetFilterChrIdA(unsigned ChrId);
void SetFilterChrIdB(unsigned ChrId);
void BeforeChrPair_();
void AfterChrPair_();
bool GetInvertMinusBundles();
void RevComp(const std::string &Seq, std::string &RevCompSeq);
void FilterMakeTrueAI(const AlnInfo &UserAI, AlnInfo &TrueAI);
bool FilterIsInverting();
unsigned GetFilterInvGenomeIndex();
unsigned GetFilterInvChrId();
void RevComp(char *Seq, unsigned Length);
//void FilterInvertRows(const AlnInfo &AI, const StdStrVec &Rows, StdStrVec &InvRows);
void FilterInvertAI(const AlnInfo &AI, AlnInfo &InvAI);
void FilterInvertAIs(const AIVec &AIs, AIVec &InvAIs);

#endif	// FILTER4_H
