#include "filter4.h"
#include "progress.h"

struct BundleInfo
	{
	unsigned ChrIdA;
	unsigned ChrIdB;
	unsigned BundleId;

	bool operator<(const BundleInfo &rhs) const
		{
		if (ChrIdA != rhs.ChrIdA)
			return ChrIdA < rhs.ChrIdA;
		if (ChrIdB != rhs.ChrIdB)
			return ChrIdB < rhs.ChrIdB;
		return BundleId < rhs.BundleId;
		}
	};

static std::vector<BundleInfo> *g_ptrBundleInfos;

static void AddToBundle(const BundleInfo &BI, const AlnInfo &AI, unsigned AIIndex,
  std::vector<AIVec> &Bundles, std::vector<IntVec> &BundleAIIndexes,
  std::vector<BundleInfo> &BundleInfos,
  std::map<BundleInfo, unsigned> &BundleInfoToIndex)
	{
	std::map<BundleInfo, unsigned>::const_iterator p =
	  BundleInfoToIndex.find(BI);

	if (p == BundleInfoToIndex.end())
		{
		assert(Bundles.size() == BundleInfoToIndex.size());
		assert(Bundles.size() == BundleInfos.size());

		unsigned BundleIndex = (unsigned) BundleInfoToIndex.size();
		std::map<BundleInfo, unsigned>::value_type v(BI, BundleIndex);
		std::pair<std::map<BundleInfo, unsigned>::iterator, bool> Result =
		  BundleInfoToIndex.insert(v);
		if (!Result.second)
			Die("AddToBundle: insertion failed");

		AIVec AV;
		AV.push_back(AI);
		Bundles.push_back(AV);

		IntVec IV;
		IV.push_back(AIIndex);
		BundleAIIndexes.push_back(IV);

		BundleInfos.push_back(BI);
		}
	else
		{
		unsigned BundleIndex = p->second;
		assert(BundleIndex < (unsigned) Bundles.size());

		Bundles[BundleIndex].push_back(AI);
		BundleAIIndexes[BundleIndex].push_back(AIIndex);
		}
	}

static bool LTSortOrder(unsigned i, unsigned j)
	{
	const BundleInfo &BIi = (*g_ptrBundleInfos)[i];
	const BundleInfo &BIj = (*g_ptrBundleInfos)[j];
	return BIi < BIj;
	}

/***
Faster but more memory-intensive algorithm for
filtering all bundles in given genome pair.
***/
void FilterAllChrPairs4(const RevFile &In)
	{
	unsigned GenomeIndexA = GetFilterGenomeIndexA();
	unsigned GenomeIndexB = GetFilterGenomeIndexB();

	Progress("Get AIs");
	AIVec AIs;
	In.GetAllAlnInfos(AIs);

	std::vector<AIVec> Bundles;
	std::vector<IntVec> BundleAIIndexes;
	std::vector<BundleInfo> BundleInfos;
	std::map<BundleInfo, unsigned> BundleInfoToIndex;

	g_ptrBundleInfos = &BundleInfos;

	unsigned AICount = (unsigned) AIs.size();

	ProgressStart("Build bundles");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AICount);

		const AlnInfo &AI = AIs[AIIndex];

		const GCSeg *SegA = AI.GetSegPtr(GenomeIndexA);
		const GCSeg *SegB = AI.GetSegPtr(GenomeIndexB);
		if (SegA == 0 || SegB == 0)
			continue;

		unsigned BundleId = AI.GetBundleId();

		BundleInfo BI;
		BI.ChrIdA = SegA->ChrId;
		BI.ChrIdB = SegB->ChrId;
		BI.BundleId = BundleId;

		AddToBundle(BI, AI, AIIndex, Bundles, BundleAIIndexes, BundleInfos,
		  BundleInfoToIndex);
		}
	ProgressDone();

	const unsigned BundleCount = (unsigned) Bundles.size();

	Progress("%u bundles", BundleCount);

	unsigned LastChrIdA = UINT_MAX;
	unsigned LastChrIdB = UINT_MAX;

	IntVec SortOrder;
	SortOrder.resize(BundleCount);
	for (unsigned BundleIndex = 0; BundleIndex < BundleCount; ++BundleIndex)
		SortOrder[BundleIndex] = BundleIndex;

	std::sort(SortOrder.begin(), SortOrder.end(), LTSortOrder);

	ProgressStart("Processing");
	for (unsigned i = 0; i < BundleCount; ++i)
		{
		if (i%100 == 0)
			ProgressStep(i, BundleCount);

		unsigned BundleIndex = SortOrder[i];
		assert(BundleIndex < BundleCount);
		const BundleInfo &BI = BundleInfos[BundleIndex];

		unsigned ChrIdA = BI.ChrIdA;
		unsigned ChrIdB = BI.ChrIdB;

		if (ChrIdA != LastChrIdA || ChrIdB != LastChrIdB)
			{
			if (LastChrIdA != UINT_MAX)
				{
				assert(LastChrIdB != UINT_MAX);
				AfterChrPair_();
				}
			SetFilterChrIdA(ChrIdA);
			SetFilterChrIdB(ChrIdB);
			BeforeChrPair_();
			}

		SetFilterBundleId(BI.BundleId);

		const AIVec &AIs = Bundles[BundleIndex];
		const IntVec &AIIndexes = BundleAIIndexes[BundleIndex];

		FilterWrapper4(AIs, AIIndexes);

		LastChrIdA = ChrIdA;
		LastChrIdB = ChrIdB;
		}
	ProgressDone();

	if (LastChrIdA != UINT_MAX)
		{
		assert(LastChrIdB != UINT_MAX);
		AfterChrPair_();
		}
	}
