#include "filter4.h"
#include "progress.h"

/***
Wrapper functions for calling user's filter.
If requested, minus strand bundles will be inverted
so that the user's code need only consider plus
bundles. This can be a very useful simplification.

"Iversion" means that the coordinates are reflected
around an axis at L_B/2, where L_B is the length
of the B chromosome.


  |        
B |     \True
  |      \        
e |       \  /
m |________\/_____ L_B/2
o |        /\
n |       /  \
e |      /     
G |     /User   
   ----------------->
         Genome A

The result of this transformation is that a minus bundle
becomes a plus bundle, and vice versa. Inversion is its
own inverse.

To invoke this feature, SetInvertMinusBundles(true)
should be called from within InitFilter4().

The user's filter code must call FilterAware*() functions 
to get and output AlnInfo's, sequences and alignments;
segments and sequences will then be un-inverted as needed,
making the invert / uninvert process transparent to
user code.

The implementation of the FilterAware*() functions must be
careful to track what are "User" AlnInfo's and sequences,
and what are "True" objects. A "User" object is an object
that passes between the user's filter and the FilterAware*()
library functions, either as an input or an output parameter.
When inversion is enabled, User objects always appear to be
part of a plus bundle.

User objects may be inverted, providing that a negative bundle
is being processed and SetInvertMinusBundles(true) was called.
Then:

(a) A User sequence will be inverted iff it is a segment of
the inverted chromosome. Here, inversion means that the
sequence is reverse-complemented.

(b) A User GCSeg object will be inverted iff it has the
coordinates of a segment of the inverted chromosome.
In this case, the transformation is a reflection
around :

	UserGCSeg.lo = ChrLength - TrueGCSeg.hi - 1
	UserGCSeg.hi = ChrLength - TrueGCSeg.lo - 1
	UserGCSeg.Plus = !TrueGCSeg.Plus

(c) A User AlnInfo will be inverted iff it has a segment
of the inverted chromosome, in which case that segment
is inverted as in (b) above.
***/

static bool g_Inverting;
static unsigned g_InvGenomeIndex;
static unsigned g_InvChrId;
static coord_t g_InvChrLength;

bool FilterIsInverting()
	{
	return g_Inverting;
	}

unsigned GetFilterInvGenomeIndex()
	{
	if (!g_Inverting)
		return UINT_MAX;
	return g_InvGenomeIndex;
	}

unsigned GetFilterInvChrId()
	{
	if (!g_Inverting)
		return UINT_MAX;
	return g_InvChrId;
	}

#if	0
void FilterInvertRows(const AlnInfo &AI, const StdStrVec &Rows, StdStrVec &InvRows)
	{
	InvRows.clear();

	if (!FilterIsInverting())
		{
		InvRows = Rows;
		return;
		}

	const unsigned RowCount = (unsigned) Rows.size();
	for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
		{
// Doesn't matter whether AI is inverted or not, is only
// used to determine the row, if any, that needs to be
// reverse-complemented.
		const GCSeg &Seg = AI.Segs[RowIndex];
		const std::string &Row = Rows[RowIndex];
		if (Seg.GenomeIndex == g_InvGenomeIndex)
			{
			if (Seg.ChrId != g_InvChrId)
				Die("FilterInvertRows: Seg.ChrId != g_InvChrId");

			std::string RevCompRow;
			RevComp(Row, RevCompRow);
			InvRows.push_back(RevCompRow);
			}
		else
			InvRows.push_back(Row);
		}
	}
#endif

void FilterInvertAI(const AlnInfo &AI, AlnInfo &InvAI)
	{
	InvAI = AI;
	if (!FilterIsInverting())
		return;

	const GCSeg *Seg = InvAI.GetModifiableSegPtr(g_InvGenomeIndex);
	if (Seg != 0)
		{
		if (Seg->ChrId != g_InvChrId)
			{
			Log("\n");
			LogAI(AI);
			Log("\n");
			Die("FilterInvertAI: ChrId=%u != g_InvChrId=%u",
			  Seg->ChrId, g_InvChrId);
			}

		coord_t lo = Seg->lo;
		coord_t hi = Seg->hi;
		if (hi >= g_InvChrLength)
			{
			Log("\n");
			LogAI(AI);
			Log("\n");
			Die("FilterInvertAI: hi=%u >= ChrLength=%u",
			  (unsigned) hi, (unsigned) g_InvChrLength);
			}

		GCSeg *InvSeg = InvAI.GetModifiableSegPtr(g_InvGenomeIndex);
		InvSeg->lo = g_InvChrLength - hi - 1;
		InvSeg->hi = g_InvChrLength - lo - 1;
		InvSeg->Plus = !Seg->Plus;
		}
	}

void FilterInvertAIs(const AIVec &AIs, AIVec &InvAIs)
	{
	InvAIs.clear();
	const unsigned AICount = (unsigned) AIs.size();
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const AlnInfo &AI = AIs[AIIndex];
		AlnInfo InvAI = AIs[AIIndex];
		FilterInvertAI(AI, InvAI);
		InvAIs.push_back(InvAI);
		}
	}

// Check first 8 AIs to determine whether bundle is
// on negative strand using simple majority vote.
// Must of course be true, not faked (i.e., possibly
// inverted) AIs.
bool AIsArePlus(const AIVec &TrueAIs, unsigned GenomeIndexA,
  unsigned GenomeIndexB)
	{
	unsigned N = (unsigned) TrueAIs.size();
	unsigned PlusCount = 0;
	unsigned MinusCount = 0;
	if (N > 8)
		N = 8;
	for (unsigned AIIndex = 0; AIIndex < N; ++AIIndex)
		{
		const AlnInfo &TrueAI = TrueAIs[AIIndex];
		const GCSeg &TrueSegA = TrueAI.GetSeg(GenomeIndexA);
		const GCSeg &TrueSegB = TrueAI.GetSeg(GenomeIndexB);
		if (TrueSegA.Plus == TrueSegB.Plus)
			++PlusCount;
		else
			++MinusCount;
		}
	assert(PlusCount + MinusCount == N);
	return PlusCount >= MinusCount;
	}

void FilterWrapper4(const AIVec &TrueAIs, const IntVec &AIIndexes)
	{
	bool Invert = GetInvertMinusBundles();
	if (!Invert ||
	  AIsArePlus(TrueAIs, GetFilterGenomeIndexA(), GetFilterGenomeIndexB()))
		{
		g_Inverting = false;
		Filter4(TrueAIs, AIIndexes);
		return;
		}

	g_Inverting = true;
	g_InvGenomeIndex = GetFilterGenomeIndexB();
	g_InvChrId = GetFilterChrIdB();

	const RevFile &In = GetFilterInputRevFile();
	g_InvChrLength = In.GetChrLength(g_InvGenomeIndex, g_InvChrId);

	AIVec UserAIs;
	FilterInvertAIs(TrueAIs, UserAIs);
	Filter4(UserAIs, AIIndexes);
	}
