#define TRACE		0

#include "evolver.h"
#include "revfile.h"
#include "progress.h"

char CompLetter(char c);

extern bool g_ValAlnSeqs;
extern unsigned g_ParalogousAlnCount;

/***
Each BaseInfo in m_EvolvedBases[] is a trivial alignment of a base in
the ancestral genome to a base in the evolved genome. To assemble maximal
alignments as a sequence of BaseInfo's, we traverse the evolved genome
from start to finish. Alignments are terminated where the ancestral
genome changes strand or where the ancestral genome coordinates are
not consecutive. However, the latter condition is not quite that simple.
Consider the following, where a base is represented by an aligned pair
of coordinates:

	An	1245
	Ev	1234

Note that ancestral base 3 is missing. This could be due to a deletion
or a rearrangement. If base 3 was deleted, this becomes a single alignment:

	An	12345
	Ev	12-34

But if base 3 has been relocated:

	An	12453
	Ev	12345

then the alignment must be broken up:

	An	12
	Ev	12

	An	45
	Ev	34

	An	3
	Ev	5

Novel insertions also add a complication. Let 'n' represent a novel base,
and consider this scenario (base pairs again in evolved order):

	An	12nn34
	Ev	123456

This is a short insertion, and we would want to represent it as follows:

	An	12--34
	Ev	123456

However, consider the case where the insertion is long, and / or long
compared with the aligned flanking regions:

	An	12nnnnn34
	Ev	123456789

We might prefer to represent this as two short alignments:

	An	12
	Ev	12

	An	34
	Ev	89

However, for simplicity we choose to always maximize length.
***/

void ValAln(const AlnInfo &AI, const StdStrVec &Rows, const char *Seq0,
  coord_t Seq0Len, const char *Seq1, coord_t Seq1Len);
extern char *g_AncSeq;
extern const char *g_EvSeq;
extern gpos_t g_AncSeqLength;
extern gpos_t g_EvSeqLength;

static unsigned g_AlnCount = 0;

void DeleteTrailingGaps(AlnInfo &AI, StdStrVec &Rows)
	{
	const unsigned L = SIZE(Rows[0]);
	if (L == 0)
		return;

	GCSegVec &Segs = AI.Segs;
	asserta(SIZE(Rows) == 2);
	asserta(SIZE(Segs) == 2);

	StdStr &Row0 = Rows[0];
	StdStr &Row1 = Rows[1];
	GCSeg &Seg0 = Segs[0];
	GCSeg &Seg1 = Segs[1];
	for (unsigned i = L - 1; i > 0; --i)
		{
		char c0 = Row0[i];
		char c1 = Row1[i];
		if (c0 == '-')
			{
			asserta(c1 != '-');
			if (Seg1.Plus)
				--Seg1.hi;
			else
				{
				asserta(Seg1.lo > 0);
				--Seg1.lo;
				}
			Row0.resize(i);
			Row1.resize(i);
			}
		else if (c1 == '-')
			{
			asserta(c0 != '-');
			if (Seg0.Plus)
				--Seg0.hi;
			else
				{
				asserta(Seg0.lo > 0);
				--Seg0.lo;
				}
			Row0.resize(i);
			Row1.resize(i);
			}
		else
			break;
		}
	}

void Evolver::EndAln(GotAlignmentCallback GotAlignment, AlnInfo &AI,
  StdStrVec &Rows) const
	{
	if (Rows[0].empty())
		return;
	++g_AlnCount;
TRC
	Log("\n");
	Log("EndAln AI=");
	AI.LogMe();
	Log("\n");
ENDTRC

	if (g_ValAlnSeqs)
		ValAln(AI, Rows, g_AncSeq, g_AncSeqLength, g_EvSeq, g_EvSeqLength);

	DeleteTrailingGaps(AI, Rows);
	GotAlignment(*this, AI, Rows);
	}

void Evolver::StartAln(gpos_t EvPos, gpos_t AncPos, AlnInfo &AI,
  StdStrVec &Rows) const
	{
	AI.Clear();
	AI.Segs.resize(2);

	Rows.clear();
	Rows.resize(2);

	if (IsCopy(AncPos))
		{
		++g_ParalogousAlnCount;
		AI.SetAttr(ALN_ATTR_Paralog, 1);
		}
	else
		AI.SetAttr(ALN_ATTR_Paralog, 0);

	const BaseInfo &Base = GetBase(AncPos);
	gpos_t HomPos = GetHomologousPos(AncPos);

	AI.Segs[0].GenomeIndex = 0;
	AI.Segs[1].GenomeIndex = 1;

	AI.Segs[0].ChrId = 0;
	AI.Segs[0].ChrId = m_AncChrId;

	AI.Segs[0].lo = HomPos;
	AI.Segs[0].hi = HomPos;

	AI.Segs[1].lo = EvPos;
	AI.Segs[1].hi = EvPos;

/***
Orientation
-----------
Alignment columns are added in evolved sequence order.

The genome index of the ancestral genome is 0, the evolved
genome is 1. The evolved strand is always +, the ancestral
strand will be - if there has been an inversion.

Because the evolved strand is always +, i.e. always is a
substring of the evolved sequence, Base.GetOrientedLetter()
must be used to obtain the letter for the evolved row of an
alignment.

The ancestral letter must be Base.AncLetter if the ancestral
strand is +, otherwise CompLetter(Base.AncLetter); i.e.
use Base.GetOrientedAncLetter() assuming that Base.Plus
matches the ancestral strand.
***/

	AI.Segs[0].Plus = Base.Plus;
	AI.Segs[1].Plus = true;

	char AncLetter = Base.GetOrientedAncLetter();
	char EvLetter = Base.GetOrientedLetter();

	Rows[0].push_back(AncLetter);
	Rows[1].push_back(EvLetter);

TRC
	{
	Log("\n");
	Log("StartAln, Base=");
	Base.LogMe();
	Log("\n");
	Log("     EvPos      AncPos      HomPos  PrevNonNov  Cat  AE  +  Succ      Anc lo      Anc hi  +       Ev lo       Ev hi  +        Cols\n");
	Log("==========  ==========  ==========  ==========  ===  ==  =  ====  ==========  ==========  =  ==========  ==========  =  ==========\n");
	Log("%10"PF, EvPos);
	Log("  %10"PF, AncPos);
	if (HomPos == GPOS_T_MAX)
		Log("           *");
	else
		Log("  %10"PF, HomPos);
	Log("           *");
	Log("  %3.3s",  BaseCategoryToStr(GetCategory(AncPos)));
	Log("  %c%c", AncLetter, Base.GetLetter());
	Log("  %c", pom(Base.Plus));
	Log("  %4c", tof(true));
	Log("  %10u", AI.Segs[0].lo);
	Log("  %10u", AI.Segs[0].hi);
	Log("  %c", pom(AI.Segs[0].Plus));
	Log("  %10u", AI.Segs[1].lo);
	Log("  %10u", AI.Segs[1].hi);
	Log("  %c", pom(AI.Segs[1].Plus));
	Log("  %10u", SIZE(Rows[0]));
	Log("\n");
	}
ENDTRC
	}

bool Evolver::ValidAlnSuccessor(gpos_t PrevNonNovelPos, gpos_t Pos) const
	{
	if (PrevNonNovelPos == GPOS_T_MAX || Pos == GPOS_T_MAX)
		return true;

	const BaseInfo *PrevHomBase = &GetBase(PrevNonNovelPos);
	const BaseInfo *Base = &GetBase(Pos);

	if (IsNovel(Pos))
		return true;

	if (Base->Plus != PrevHomBase->Plus)
		return false;

	if (GetCategory(Pos) != GetCategory(PrevNonNovelPos))
		return false;

	gpos_t PrevNonNovelHomPos = GetHomologousPos(PrevNonNovelPos);
	gpos_t HomPos = GetHomologousPos(Pos);

	if (Base->Plus)
		{
		if (PrevNonNovelHomPos >= HomPos)
			return false;
		}
	else
		{
		if (PrevNonNovelHomPos <= HomPos)
			return false;
		}

	gpos_t LoPos = min(Pos, PrevNonNovelPos);
	gpos_t HiPos = max(Pos, PrevNonNovelPos);

	for (gpos_t Pos2 = LoPos + 1; Pos2 < HiPos; ++Pos2)
		{
		bool Del = PosIsDeleted(Pos2);
		if (!Del)
			return false;
		}
	return true;
	}

void Evolver::AppendBaseToAln(GotAlignmentCallback GotAlignment,
  gpos_t EvPos,  gpos_t PrevNonNovelPos, gpos_t AncPos, AlnInfo &AI,
  StdStrVec &Rows) const
	{
	bool ValidSucc = ValidAlnSuccessor(PrevNonNovelPos, AncPos);
	if (!ValidSucc)
		{
		EndAln(GotAlignment, AI, Rows);
		AI.Clear();
		Rows[0].clear();
		Rows[1].clear();
		}

	if (Rows[0].empty())
		{
	// Don't start with insertion.
	// (Can't start with deletion because we're scanning evolved sequence).
		if (IsNovel(AncPos))
			return;

		StartAln(EvPos, AncPos, AI, Rows);
		return;
		}

	GCSeg &AncSeg = AI.Segs[0];
	GCSeg &EvSeg = AI.Segs[1];

	const BaseInfo &Base = GetBase(AncPos);
	char EvLetter = Base.GetOrientedLetter();
	char AncLetter = Base.GetOrientedAncLetter();

	bool Novel = IsNovel(AncPos);
	if (Novel)
		AncLetter = '-';

	if (!Novel)
		{
		gpos_t HomPos = GetHomologousPos(AncPos);
		asserta(HomPos != GPOS_T_MAX);
		if (AncSeg.Plus)
			{
			if (AncSeg.hi >= HomPos)
				{
				Log("\n");
				Log("AppendBaseToAln EvPos=%"PF" PrevNonNovelPos=%"PF" AncPos=%"PF"\n",
				  EvPos, PrevNonNovelPos, AncPos);
				Log("AI=");
				AI.LogMe();
				Log("ValidAlnSucc(PrevNonNovelPos=%"PF", AncPos=%"PF")=%c\n",
				  PrevNonNovelPos, AncPos, tof(ValidSucc));
				Die("FindAlignments, plus del: AncSeg.lo=%"PF" >= HomPos(%"PF")=%"PF,
				  AncSeg.lo, AncPos, HomPos);
				}

		// Deletions
			while (AncSeg.hi + 1 < HomPos)
				{
				++(AncSeg.hi);
				char Letter = GetAncLetter(AncSeg.hi);
TRC
				Log("          ");
				Log("  %10"PF, AncPos);
				gpos_t HomPos = GetHomologousPos(AncSeg.hi);
				if (HomPos == GPOS_T_MAX)
					Log("           *");
				else
					Log("  %10"PF, HomPos);
				if (PrevNonNovelPos == GPOS_T_MAX)
					Log("           *");
				else
					Log("  %10"PF, PrevNonNovelPos);
				Log("  DEL");
				Log("  %c-", Letter);
				Log("  +");
				Log("      ");
				Log("  %10u", AI.Segs[0].lo);
				Log("  %10u", AI.Segs[0].hi);
				Log("  %c", pom(AI.Segs[0].Plus));
				Log("  %10u", AI.Segs[1].lo);
				Log("  %10u", AI.Segs[1].hi);
				Log("  %c", pom(AI.Segs[1].Plus));
				Log("  %10u", SIZE(Rows[0]));
				Log("\n");
ENDTRC
				Rows[0].push_back(Letter);
				Rows[1].push_back('-');
				}
			}
		else
			{
			if (AncSeg.lo <= HomPos)
				{
				Log("\n");
				Log("AppendBaseToAln EvPos=%"PF" PrevNonNovelPos=%"PF" AncPos=%"PF"\n",
				  EvPos, PrevNonNovelPos, AncPos);
				Log("AI=");
				AI.LogMe();
				Log("ValidAlnSucc(PrevNonNovelPos=%"PF", AncPos=%"PF")=%c\n",
				  PrevNonNovelPos, AncPos, tof(ValidSucc));
				Die("FindAlignments, minus del: AncSeg.lo=%"PF" <= HomPos(%"PF")=%"PF,
				  AncSeg.lo, AncPos, HomPos);
				}

		// Deletions
			while (AncSeg.lo - 1 > HomPos)
				{
				asserta(AncSeg.lo > 0);
				--(AncSeg.lo);
				char AncLetter = CompLetter(GetAncLetter(AncSeg.lo));
TRC
				Log("          ");
				Log("  %10"PF, AncPos);
				gpos_t HomPos = GetHomologousPos(AncSeg.lo);
				if (HomPos == GPOS_T_MAX)
					Log("           *");
				else
					Log("  %10"PF, HomPos);
				if (PrevNonNovelPos == GPOS_T_MAX)
					Log("           *");
				else
					Log("  %10"PF, PrevNonNovelPos);
				Log("  DEL");
				Log("  %c-", AncLetter);
				Log("  -");
				Log("      ");
				Log("  %10u", AI.Segs[0].lo);
				Log("  %10u", AI.Segs[0].hi);
				Log("  %c", pom(AI.Segs[0].Plus));
				Log("  %10u", AI.Segs[1].lo);
				Log("  %10u", AI.Segs[1].hi);
				Log("  %c", pom(AI.Segs[1].Plus));
				Log("  %10u", SIZE(Rows[0]));
				Log("\n");
ENDTRC
				Rows[0].push_back(AncLetter);
				Rows[1].push_back('-');
				}
			}
		}

	Rows[0].push_back(AncLetter);
	Rows[1].push_back(EvLetter);

	if (AncLetter != '-')
		AncSeg.Inc();

	asserta(EvLetter != '-');
	EvSeg.Inc();

TRC
	{
	const BaseInfo &Base = GetBase(AncPos);
	gpos_t HomPos = GetHomologousPos(AncPos);
	Log("%10"PF, EvPos);
	Log("  %10"PF, AncPos);
	if (HomPos == GPOS_T_MAX)
		Log("           *");
	else
		Log("  %10"PF, HomPos);
	if (PrevNonNovelPos == GPOS_T_MAX)
		Log("           *");
	else
		Log("  %10"PF, PrevNonNovelPos);
	Log("  %3.3s", BaseCategoryToStr(GetCategory(AncPos)));
	Log("  %c%c", Base.GetAncLetter(), Base.GetLetter());
	Log("  %c", pom(Base.Plus));
	Log("     %c", tof(ValidSucc));
	Log("  %10u", AI.Segs[0].lo);
	Log("  %10u", AI.Segs[0].hi);
	Log("  %c", pom(AI.Segs[0].Plus));
	Log("  %10u", AI.Segs[1].lo);
	Log("  %10u", AI.Segs[1].hi);
	Log("  %c", pom(AI.Segs[1].Plus));
	Log("  %10u", SIZE(Rows[0]));
	Log("\n");
	}
ENDTRC
	asserta(Novel || Base.Plus == AncSeg.Plus);
	}

void Evolver::FindAlignments(GotAlignmentCallback GotAlignment) const
	{
	gpos_t EvolvedLength = GetEvolvedLength();
TRC
	Log("\n");
	Log("FindAlignments\n");
ENDTRC
	AlnInfo AI;
	StdStrVec Rows(2);

	gpos_t EvPos = 0;
	gpos_t PrevNonNovelPos = GPOS_T_MAX;
	ProgressStart("Finding alignments");
	FOR_EVBASES(AncPos)
		{
		if (EvPos%10000 == 0)
			{
			SetProgressMsg("Finding Alignments (%u)", g_AlnCount);
			ProgressStep(EvPos, EvolvedLength);
			}
		AppendBaseToAln(GotAlignment, EvPos, PrevNonNovelPos, AncPos, AI, Rows);
		if (!IsNovel(AncPos))
			PrevNonNovelPos = AncPos;
		++EvPos;
		}
	SetProgressMsg("Finding Alignments (%u)", g_AlnCount);
	ProgressDone();
	EndAln(GotAlignment, AI, Rows);
	}
