#include <string>
#include "utils.h"
#include "myrand.h"
#include "getopts.h"
#include "readfasta.h"
#include "gff2.h"
#include "progress.h"

using namespace std;

bool isacgt(char c);
void WriteFasta(FILE *f, const char *SeqData, unsigned Length);
double Pct(double, double);

static unsigned g_Counts[UCHAR_MAX+1];
static unsigned g_SeqLength;
static unsigned g_RandomizedCount;
static unsigned g_ExcizedCount;
static unsigned g_MinNs = 32;
static unsigned g_RunCount;
static string g_SeqName;

static FILE *g_fOutRandNs;
static FILE *g_fOutX;
static FILE *g_fGFF;

static const char *g_LabelRandNs;
static const char *g_LabelX;

static void OnInvalidLetter(unsigned char c)
	{
	++(g_Counts[c]);
	}

static void OnNRun(unsigned Pos, unsigned Length, string &SeqR, string &SeqX)
	{
	++g_RunCount;

	bool DoExcize = (Length >= g_MinNs);
	if (g_fGFF != 0 && DoExcize)
		{
		GFFRecord2 Rec;
		Rec.SeqName = g_SeqName;
		Rec.Source = "evo";
		Rec.Feature = "nblock";
		Rec.Start = Pos + 1;
		Rec.End = Rec.Start + Length - 1;
		Rec.Strand = '.';
		Rec.Score = 0;
		WriteGFFRec(g_fGFF, Rec);
		}

	bool DoR = (g_fOutRandNs != 0);
	bool DoX = (g_fOutX != 0);

	if (DoExcize)
		g_ExcizedCount += Length;

	g_RandomizedCount += Length;
	for (unsigned i = 0; i < Length; ++i)
		{
		char c = RandLetter();
		if (DoX && !DoExcize)
			SeqX.push_back(c);
		if (DoR)
			SeqR.push_back(c);
		}
	}

static void OnSeq(const string &Label, const string &Seq)
	{
	g_SeqName = Label;
	g_SeqLength = SIZE(Seq);

	bool DoR = (g_fOutRandNs != 0);
	bool DoX = (g_fOutX != 0);

	string SeqR;
	string SeqX;

	if (DoR)
		SeqR.reserve(g_SeqLength);

	if (DoX)
		SeqX.reserve(g_SeqLength);

	unsigned NRunLength = 0;
	unsigned Pos = 0;
	unsigned NRunStartPos = UINT_MAX;
	for (string::const_iterator p = Seq.begin(); p != Seq.end(); ++p)
		{
		char c = (char) toupper(*p);
		if (isacgt(c))
			{
			if (NRunLength > 0)
				{
				OnNRun(NRunStartPos, NRunLength, SeqR, SeqX);
				NRunLength = 0;
				}

			if (DoX)
				SeqX.push_back(c);
			if (DoR)
				SeqR.push_back(c);
			}
		else
			{
			OnInvalidLetter(c);
			if (NRunLength == 0)
				{
				NRunStartPos = Pos;
				NRunLength = 1;
				}
			else
				++NRunLength;
			}
		++Pos;
		}

	if (NRunLength > 0)
		OnNRun(NRunStartPos, NRunLength, SeqR, SeqX);

	if (g_fOutRandNs != 0)
		{
		string LabelRandNs;
		if (g_LabelRandNs != 0)
			LabelRandNs = g_LabelRandNs;
		else
			LabelRandNs = Label + ".randns";
		Progress("Writing sequence with randomized Ns");
		fprintf(g_fOutRandNs, ">%s\n", LabelRandNs.c_str());
		WriteFasta(g_fOutRandNs, SeqR.c_str(), SIZE(SeqR));
		SeqR.clear();
		}

	if (g_fOutX != 0)
		{
		string LabelX;
		if (g_LabelX != 0)
			LabelX = g_LabelX;
		else
			LabelX = Label + ".xns";
		Progress("Writing sequence with excized Ns");
		fprintf(g_fOutX, ">%s\n", LabelX.c_str());
		WriteFasta(g_fOutX, SeqX.c_str(), SIZE(SeqX));
		}
	}

static void ReportCounts(unsigned Length)
	{
	unsigned Total = 0;
	Dump("\n");
	Dump("%u runs of >= %u Ns\n", g_RunCount, g_MinNs);
	for (unsigned i = 0; i <= UCHAR_MAX; ++i)
		{
		if (g_Counts[i] == 0)
			continue;
		if (Total == 0)
			{
			Dump("\n");
			Dump("Letter      Number\n");
			Dump("======  ==========\n");
			}
		if (isprint(i))
			Dump("%6c", (char) i);
		else
			Dump("    %02x", i);
		Dump("  %10u", g_Counts[i]);
		Dump("\n");
		Total += g_Counts[i];
		}
	Dump("\n");
	Dump("Invalid     %10u  (%5.1f%%)\n",
	  Total, Pct(Total, Length));

	Dump("Randomized  %10u  (%5.1f%%)\n",
	  g_RandomizedCount, Pct(g_RandomizedCount, Length));

	Dump("Excized     %10u  (%5.1f%%)\n",
	  g_ExcizedCount, Pct(g_ExcizedCount, Length));
	}

// Find and (optionally) randomize Ns (and other invalid characters)
// in FASTA sequence.
void FindNs()
	{
	const char *InFileName = RequiredValueOpt("findns");
	const char *OutRandNsFileName = ValueOpt("out_randns");
	const char *OutExciseFileName = ValueOpt("out_x");
	const char *GFFFileName = ValueOpt("out_gff");
	g_LabelRandNs = ValueOpt("label_randns");
	g_LabelX = ValueOpt("label_x");
	IntOpt("minns", &g_MinNs);

	if (OutRandNsFileName != 0)
		g_fOutRandNs = CreateStdioFile(OutRandNsFileName);

	if (OutExciseFileName != 0)
		g_fOutX = CreateStdioFile(OutExciseFileName);

	if (GFFFileName != 0)
		g_fGFF = CreateStdioFile(GFFFileName);

	memset(g_Counts, 0, sizeof(g_Counts));
	ReadFasta(InFileName, OnSeq);
	ReportCounts(g_SeqLength);

	if (g_fOutRandNs != 0)
		CloseStdioFile(g_fOutRandNs);

	if (g_fOutX != 0)
		CloseStdioFile(g_fOutX);

	if (g_fGFF != 0)
		CloseStdioFile(g_fGFF);
	}
