#include "evolver.h"
#include "myrand.h"
#include "evomodel.h"

double GetSlowerAccel();
double GetFasterAccel();

extern unsigned g_GeneDupeCount;
extern unsigned g_ActiveGeneDupeCount;
extern unsigned g_GeneInGeneInsertCount;
extern unsigned g_GeneDupeWeights[4][4];

bool IsActiveResult(GENE_DUPE_RESULT Result)
	{
	return Result != GDR_Lost;
	}

void GetGeneDupeResults(GENE_DUPE_RESULT &Old, GENE_DUPE_RESULT &New)
	{
	vector<double> Weights(16);

	for (unsigned i = 0; i < 4; ++i)
		for (unsigned j = 0; j < 4; ++j)
			Weights[4*i + j] = g_GeneDupeWeights[i][j];

	unsigned k = RandWeightDist(Weights);
	unsigned uOld = k%4;
	unsigned uNew = k/4;

	Old = (GENE_DUPE_RESULT) uOld;
	New = (GENE_DUPE_RESULT) uNew;
	}

void Evolver::DupeGenes(const vector<gpos_t> &GeneStarts,
  const vector<gpos_t> &CopiedGeneStarts, bool DoInvert)
	{
	const unsigned N = SIZE(GeneStarts);
	asserta(SIZE(CopiedGeneStarts) == N);
	if (N == 0)
		return;

	for (unsigned i = 0; i < N; ++i)
		{
		++g_GeneDupeCount;
		gpos_t Pos = CopiedGeneStarts[i];
		const BaseInfo &Base = GetBase(Pos);
		if (Base.IsGene())
			{
			++g_GeneInGeneInsertCount;
			continue;
			}

		GENE_DUPE_RESULT OldResult;
		GENE_DUPE_RESULT NewResult;
		GetGeneDupeResults(OldResult, NewResult);

		bool NewActive = IsActiveResult(NewResult);
		gpos_t OldStart = GeneStarts[i];
		gpos_t NewStart = CopiedGeneStarts[i];
		const BaseInfo &OldStartBase = GetBase(OldStart);
		Log("@GENE_DUPE Old %s New %s (Event=%u, OldStart=%u, NewStart=%u)\n",
		  GeneDupeResultToStr(OldResult),
		  GeneDupeResultToStr(NewResult),
		  m_EventIndex,
		  OldStart,
		  NewStart);
		if (!OldStartBase.IsGene())
			{
			Log("\n");
			OldStartBase.LogMe();
			Die("Old start base at pos = %"PF" not gene", OldStart);
			}

	// Must do new one first to copy current constraints
	// in DupeGene.
		if (NewActive)
			{
			DupeGene(OldStart, NewStart, DoInvert);

			gpos_t RangeStart;
			gpos_t RangeLength;
			GetGeneRange(NewStart, RangeStart, RangeLength);

			switch (NewResult)
				{
			case GDR_Same:
				break;

			case GDR_Slower:
				{
				ChangeSpeed(RangeStart, RangeLength, GetSlowerAccel());
				GeneInfo NewGene;
				GetGene(RangeStart, NewGene);
				unsigned ErrCount = ValidateGene(NewGene);
				if (ErrCount > 0)
					Die("DupeGene: errors in copied gene after decel");
				break;
				}

			case GDR_Faster:
				{
				ChangeSpeed(RangeStart, RangeLength, GetFasterAccel());
				GeneInfo NewGene;
				GetGene(RangeStart, NewGene);
				unsigned ErrCount = ValidateGene(NewGene);
				if (ErrCount > 0)
					Die("DupeGene: errors in copied gene after accel");
				break;
				}

		// Cannot be Lost if NewActive.
		// Copied bases are initially neutral, so nothing
		// to do if new gene is lost.
			case GDR_Lost:
				asserta(false);
				break;

			default:
				asserta(false);
				}
			}

		gpos_t RangeStart;
		gpos_t RangeLength;
		GetGeneRange(OldStart, RangeStart, RangeLength);

		switch (OldResult)
			{
		case GDR_Same:
			break;

		case GDR_Slower:
			{
			ChangeSpeed(RangeStart, RangeLength, GetSlowerAccel());
			GeneInfo NewGene;
			GetGene(RangeStart, NewGene);
			unsigned ErrCount = ValidateGene(NewGene);
			if (ErrCount > 0)
				Die("DupeGene: errors in old gene after decel");
			break;
			}

		case GDR_Faster:
			{
			ChangeSpeed(RangeStart, RangeLength, GetFasterAccel());
			GeneInfo NewGene;
			GetGene(RangeStart, NewGene);
			unsigned ErrCount = ValidateGene(NewGene);
			if (ErrCount > 0)
				Die("DupeGene: errors in old gene after accel");
			break;
			}

		case GDR_Lost:
			{
			GeneInfo Gene;
			GetGene(RangeStart, Gene);
			unsigned GeneIndex = Gene.Index;
			DeleteGeneIndex(GeneIndex);
			Deconstrain(RangeStart, RangeLength, true);
			break;
			}

		default:
			asserta(false);
			}
		}
	}

void Evolver::DupeGene(gpos_t FromPos, gpos_t ToPos, bool DoInvert)
	{
	++g_ActiveGeneDupeCount;

	gpos_t SourcePos = FromPos;
	gpos_t TargetPos = ToPos;

	unsigned TargetGeneIndex = m_NextGeneIndex++;

	gpos_t NewStartPos = GPOS_T_MAX;
	bool StartFound = false;
	for (;;)
		{
		const BaseInfo &FromBase = GetBase(SourcePos);
		if (!FromBase.IsGene())
			break;

#if	0 // Not sure if this assertion is correct
		gpos_t SourceHomPos = GetHomologousPos(SourcePos);
		gpos_t TargetHomPos = GetHomologousPos(TargetPos);
		if (SourceHomPos != TargetHomPos)
            Die(
"DupeGene(%u,%u,%c), GetHomologousPos(SourcePos=%u)=%u != GetHomologousPos(TargetPos=%u)=%u",
              FromPos, 
              ToPos,
			  tof(DoInvert),
              SourcePos,
              GetHomologousPos(SourcePos),
              TargetPos,
              GetHomologousPos(TargetPos));
#endif

		BaseInfo &ToBase = GetBase(TargetPos);

		ToBase.Type = FromBase.Type;
		ToBase.ProbIx = FromBase.ProbIx;
		ToBase.Codex = (DoInvert ? -FromBase.Codex : FromBase.Codex);

		if (ToBase.Type == BT_START && (FromBase.Codex == 1 || FromBase.Codex == -1))
			{
			StartFound = true;
			map<gpos_t, unsigned>::const_iterator p = m_GeneIndexes.find(SourcePos);
			asserta(p != m_GeneIndexes.end());
			unsigned SourceGeneIndex = p->second;
			NewStartPos = TargetPos;
			m_GeneIndexes[NewStartPos] = TargetGeneIndex;
			m_Paralogs[TargetGeneIndex] = SourceGeneIndex;
			}

		SourcePos = GetNextPos(SourcePos);
		TargetPos = GetAdjPos(TargetPos, !DoInvert);
		}
	asserta(StartFound);

	GeneInfo NewGene;
	GetGene(ToPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("DupeGene: errors in copied gene");
	}

void Evolver::DeleteGeneIndex(unsigned GeneIndex)
	{
	for (map<gpos_t, unsigned>::iterator p = m_GeneIndexes.begin();
	  p != m_GeneIndexes.end(); ++p)
		{
		if (p->second == GeneIndex)
			{
			m_GeneIndexes.erase(p);
			break;
			}
		}

	for (map<gpos_t, unsigned>::iterator p = m_Paralogs.begin();
	  p != m_Paralogs.end(); ++p)
		{
		if (p->second == GeneIndex)
			{
			m_Paralogs.erase(p);
			return;
			}
		}
	}
