#include "utils.h"
#include "gff.h"
#include "getopts.h"
#include "progress.h"
#include <algorithm>
#include <vector>
#include <set>

#define TRACE	0

const unsigned MaxDepth = 9;
static unsigned BandWidth = 50000;
static unsigned TopBands = 1000;

static unsigned g_LoLength[] =
	{
	10,
	100,
	1000,
	10000,
	100000,
	1000000,
	};

const unsigned LengthBinCount = sizeof(g_LoLength)/sizeof(g_LoLength[0]);

using namespace std;

bool HasTargetAttrs(const char *Attrs);

struct BPData
	{
	unsigned Chr;
	unsigned Pos;
	bool lo;

	bool operator<(const BPData &rhs) const
		{
		if (Chr != rhs.Chr)
			return Chr < rhs.Chr;

		if (Pos == rhs.Pos)
			return lo && !rhs.lo;

		return Pos < rhs.Pos;
		}
	};

struct DiagData
	{
	unsigned Chr1;
	unsigned Chr2;
	bool Strand;
	int Diag;
	unsigned BaseCount;

	bool operator<(const DiagData &rhs) const
		{
		if (Chr1 != rhs.Chr1)
			return Chr1 < rhs.Chr1;
		if (Chr2 != rhs.Chr2)
			return Chr2 < rhs.Chr2;
		if (Strand != rhs.Strand)
			return Strand;
		return Diag < rhs.Diag;
		}
	};

static void IncCoverageCounts(vector<BPData> &BPs, vector<double> &Counts)
	{
	if (BPs.empty())
		return;

	sort(BPs.begin(), BPs.end());

	const BPData &BP0 = BPs[0];
	asserta(BP0.lo);
	unsigned Start = BP0.Pos;
	int Depth = 1;
	unsigned Count = 0;
	unsigned Chr = BP0.Chr;
#if	TRACE
	Log("\n");
	Log("ComputeCoverageCounts %u BPs\n", SIZE(BPs));
	Log("   BP    Chr         Pos       Start  lo  Depth       Count\n");
	Log("-----  -----  ----------  ----------  --  -----  ----------\n");
	Log("%5u  %5u  %10u  %10u  %2s  %5u  %10u\n",
	  0, BPs[0].Chr, BPs[0].Pos, Start, BPs[0].lo ? "lo" : "HI", Depth, Count);
#endif
	for (unsigned i = 1; i < SIZE(BPs); ++i)
		{
		const BPData &BP = BPs[i];
		if (BP.Chr != Chr)
			{
			if (!BP.lo)
				Die("!BP.lo");
			Start = BP.Pos;
			Depth = 1;
			}
		else
			{
			if (BP.lo)
				{
				Count = BP.Pos - Start;
				Start = BP.Pos;
				if (Count > 0)
					Counts[min((unsigned) Depth, MaxDepth)] += Count;
				++Depth;
				}
			else
				{
				Count = BP.Pos - Start + 1;
				Start = BP.Pos + 1;

				if (Count != 0 && Depth <= 0)
					{
					Log("Depth = %d\n", Depth);
					Log("Chr = %u\n", Chr);
					Log("Start = %u\n", Start);
					Log("Pos = %u\n", BP.Pos);
					Log("Count = %u\n", Count);
					asserta(false);
					}

				if (Count > 0)
					Counts[min((unsigned) Depth, MaxDepth)] += Count;

				--Depth;
				}
			}
		Chr = BP.Chr;
#if	TRACE
		Log("%5u  %5u  %10u  %10u  %2s  %5u  %10u\n",
		  i, BP.Chr, BP.Pos, Start, BP.lo ? "lo" : "HI", Depth, Count);
#endif
		}
	}

static double LogCoverageCounts(const vector<double> &Counts1,
  vector<double> &Counts2)
	{
	Log("\n");
	Log("Depth     Genome1     Genome2       Total     Pct\n");
	Log("-----  ----------  ----------  ----------  ------\n");
	double Total1 = 0;
	double Total2 = 0;
	double Total12 = 0;
	for (unsigned i = 0; i <= MaxDepth; ++i)
		{
		double Count1 = Counts1[i];
		double Count2 = Counts2[i];
		Total12 += Count1 + Count2;
		}

	for (unsigned i = 0; i <= MaxDepth; ++i)
		{
		double Count1 = Counts1[i];
		double Count2 = Counts2[i];
		double Total = Count1 + Count2;
		if (Total == 0)
			continue;

		Total1 += Count1;
		Total2 += Count2;

		double Pct = 0;
		if (Total12 > 0)
			Pct = 100.0*Total/Total12;
		Log("%5u%c %10.0f  %10.0f  %10.0f  %6.2f%%\n",
		  i, i == MaxDepth ? '+' : ' ', Count1, Count2, Total, Pct);
		}
	Log("=====  ==========  ==========  ==========\n");
	Log("%5.5s  %10.0f  %10.0f  %10.0f\n", "Sum", Total1, Total2, Total12);

	double Count0 = Counts1[0] + Counts2[0];
	
	if (Total12 > Count0)
		return 100.0*double(Total12 - Count0)/double(Total12);
	return 0;
	}

static map<string, unsigned> g_LabelToChr;
static map<string, unsigned> g_TargetLabelToChr;

static unsigned GetChr(const string &Label)
	{
	if (g_LabelToChr.find(Label) != g_LabelToChr.end())
		return g_LabelToChr[Label];
	unsigned i = g_LabelToChr.size();
	g_LabelToChr[Label] = i;
	return i;
	}

static unsigned GetTargetChr(const string &Label)
	{
	if (g_TargetLabelToChr.find(Label) != g_TargetLabelToChr.end())
		return g_TargetLabelToChr[Label];
	unsigned i = g_TargetLabelToChr.size();
	g_TargetLabelToChr[Label] = i;
	return i;
	}

//static unsigned LoLength(unsigned Bin)
//	{
//	asserta(Bin < LengthBinCount);
//	return g_LoLength[Bin];
//	}

static unsigned LengthToBin(unsigned Length)
	{
	for (unsigned i = 0; i < LengthBinCount; ++i)
		if (Length < g_LoLength[i])
			return i;
	return LengthBinCount - 1;
	}

static void DoDiags(vector<DiagData> &Diags, unsigned &TopBand,
  unsigned &BottomBand)
	{
	TopBand = 0;
	BottomBand = 0;
	if (Diags.empty())
		return;

	sort(Diags.begin(), Diags.end());
	vector<unsigned> BaseCounts;

	//Log("\n");
	//Log(" Chr1   Chr2         Diag       Bases\n");
	//Log("-----  -----  -----------  ----------\n");
	const DiagData *LastDD = 0;
	unsigned BaseCount = 0;
	for (unsigned i = 0; i < SIZE(Diags); ++i)
		{
		const DiagData &DD = Diags[i];
		//Log("%5u  %5u  %11d  %10u", DD.Chr1, DD.Chr2, DD.Diag, DD.BaseCount);
		if (i > 0 && DD.Diag/BandWidth == LastDD->Diag/BandWidth &&
		  DD.Strand == LastDD->Strand && DD.Chr1 == LastDD->Chr1 &&
		  DD.Chr2 == LastDD->Chr2)
			{
			//Log(" =");
			}
		else
			{
			BaseCounts.push_back(BaseCount);
			BaseCount = 0;
			}
		BaseCount += DD.BaseCount;			
		//Log("\n");
		LastDD = &DD;
		}
	BaseCounts.push_back(BaseCount);

	sort(BaseCounts.begin(), BaseCounts.end());

	unsigned M = SIZE(BaseCounts);
	unsigned N = M;
	if (N > TopBands)
		N = TopBands;

	unsigned C = BaseCounts[M-1];
	Log("\n");
	Log("Top %4u bands\n", N);
	Log("--------------\n");
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned c = BaseCounts[M-i-1];
		Log("%14u", c);

		unsigned k = (unsigned) (32*float(c)/float(C));
		Log("  ");
		for (unsigned j = 0; j < k; ++j)
			Log("*");
		Log("\n");
		}

	TopBand = M-1 < SIZE(BaseCounts) ? BaseCounts[M-1] : 0;
	BottomBand = M-N-1 < SIZE(BaseCounts) ? BaseCounts[M-N-1] : 0;
	}

void GenStats()
	{
	const char *FileName = RequiredValueOpt("genstats");
	FILE *fGFF = OpenStdioFile(FileName);

	string Label;
	GFFRecord Rec;	ProgressStartFile(fGFF, "Reading");
	unsigned RecordCount = 0;
	unsigned HitCount = 0;

	vector<unsigned> LengthBinCounts(LengthBinCount, 0);
	vector<double> PctIdCounts(10, 0);

	vector<double> CovCounts1(MaxDepth+1, 0);
	vector<double> CovCounts2(MaxDepth+1, 0);

	vector<DiagData> Diags;
	vector<BPData> BPs1;
	vector<BPData> BPs2;
	double TotalBases = 0;
	double InvertedBases = 0;
	double TotalPctId = 0;
	double TotalPctIdBases = 0;
	while (GetNextGFFRecord(fGFF, Rec))
		{
		if (RecordCount++%100000 == 0)
			ProgressStepFile();

		if (!HasTargetAttrs(Rec.Attrs))
			continue;

		++HitCount;

		char TargetChrName[4096];
		coord_t TargetStart;
		coord_t TargetEnd;
		ParseTargetAttrs(Rec.Attrs, TargetChrName, sizeof(TargetChrName),
		  &TargetStart, &TargetEnd);

		unsigned uLength = Rec.GetLength();
		unsigned Bin = LengthToBin(uLength);
		++(LengthBinCounts[Bin]);
		double Length = (double) uLength;
		TotalBases += Length;
		if (Rec.Strand == '-')
			InvertedBases += Length;

		const char *p = strstr(Rec.Attrs, "pctid");
		if (p != 0)
			{
			double dPctId = atof(p+7);
			unsigned PctId = (unsigned) dPctId;
			if (PctId <= 100)
				{
				if (PctId == 100)
					PctId = 99;
				unsigned PctIdBin = PctId/10;
				asserta(PctIdBin < 10);
				PctIdCounts[PctIdBin] += Length;

				TotalPctId += Length*PctId;
				TotalPctIdBases += Length;
				}
			}

		if (Rec.Start == 0 || Rec.End == 0 || Rec.Start > Rec.End)
			{
			Rec.LogMe();
			Die("Invalid source coords in GFF record");
			}

		if (TargetStart == 0 || TargetEnd == 0 || TargetStart > TargetEnd)
			{
			Rec.LogMe();
			Die("Invalid target coords in GFF record");
			}

		unsigned Chr1 = GetChr(Rec.SeqName);
		unsigned Chr2 = GetTargetChr(TargetChrName);

		DiagData DD;
		double Mid = ((double) Rec.Start + (double) Rec.End)/2.0;
		double TargetMid = ((double) TargetStart + (double) TargetEnd)/2.0;

		DD.Chr1 = Chr1;
		DD.Chr2 = Chr2;
		DD.Diag = (int) (Mid - TargetMid);
		DD.BaseCount = Rec.GetLength();
		Diags.push_back(DD);

		BPData BP;
		BP.Chr = Chr1;
		BP.Pos = Rec.Start;
		BP.lo = true;

		BPs1.push_back(BP);

		BP.Pos = Rec.End;
		BP.lo = false;

		BPs1.push_back(BP);

		BP.Chr = Chr2;
		BP.Pos = TargetStart;
		BP.lo = true;
		BPs2.push_back(BP);

		BP.Pos = TargetEnd;
		BP.lo = false;
		BPs2.push_back(BP);
		}
	ProgressDone();

	Log("\n");
	Log("Records        %10u\n", RecordCount);
	Log("Hits           %10u\n", HitCount);

	Progress("Computing counts 1");
	IncCoverageCounts(BPs1, CovCounts1);
	Progress("Computing counts 2");
	IncCoverageCounts(BPs2, CovCounts2);

	unsigned TopBand;
	unsigned BottomBand;
	DoDiags(Diags, TopBand, BottomBand);

	double Coverage = LogCoverageCounts(CovCounts1, CovCounts2);

	Log("\n");
	Log("      Length        Hits\n");
	Log("------------  ----------\n");
	unsigned h = 0;
	for (unsigned i = 0; i < LengthBinCount; ++i)
		{
		unsigned Count = LengthBinCounts[i];
		if (i + 1 == LengthBinCount)
			Log(">=%10u", g_LoLength[i-1]);
		else
			Log("< %10u", g_LoLength[i]);
		Log("  %10u", Count);
		h += Count;
		Log("\n");
		}
	Log("============  ==========\n");
	Log("  %10.10s  %10u\n", "", h);
	Log("\n");

	Log("\n");
	Log("   PctId       Bases     Pct\n");
	Log("--------  ----------  ------\n");
	double TotalCount = 0;
	for (unsigned i = 0; i < 10; ++i)
		{
		double Count = PctIdCounts[i];
		TotalCount += Count;
		}

	for (unsigned i = 0; i < 10; ++i)
		{
		double Count = PctIdCounts[i];
		if (Count == 0)
			continue;
		Log("%2u - %3u  %10.0f  %6.1f%%\n", i*10, (i+1)*10, Count, 100*Count/TotalCount);
		}

	Log("\n");
	Log("Coverage       %12.1f%%\n", Coverage);
	if (TotalPctIdBases > 0)
		Log("Avg pct id     %12.1f%%\n", TotalPctId/TotalPctIdBases);
	Log("Hit bases      %12.0f\n", TotalBases);
	double InvPct = (TotalBases == 0.0f ? 0.0f : 100.0*InvertedBases/TotalBases);
	Log("Inverted bases %12.0f (%.1f%%)\n", InvertedBases, InvPct);
	Log("Top band       %12u\n", TopBand);
	Log("%4u'th band   %12u (%.2g%% of top)\n",
	  TopBands, BottomBand, 100.0*float(BottomBand)/float(TopBand));
	}
