#include <string.h>
#include "evolver.h"
#include "evomodel.h"
#include "progress.h"

void RevComp(string &Seq);
uint8 IncFrame(uint8 Frame, gpos_t Length);
bool IsStartCodon(const string &Codon);
bool IsStopCodon(const string &Codon);
int8 IncCodexL(int8 Codex, gpos_t L);

// Return position of first base in each STOP codon.
// Only considers one frame, the obvious one in which first letter
// of sequence is the first base in frame.
void GetStopCodons(const string &Seq, vector<gpos_t> &Starts)
	{
	Starts.clear();
	if (Seq.size() < 3)
		return;

	char Codon[4];
	Codon[3] = 0;
	const size_t N = Seq.size();
	for (size_t i = 0; i + 2 < N; i += 3)
		{
		Codon[0] = Seq[i];
		Codon[1] = Seq[i+1];
		Codon[2] = Seq[i+2];
		if (IsStopCodon(Codon))
			Starts.push_back((gpos_t) i);
		}
	}

void Evolver::GetGeneIndexes(set<unsigned> &Indexes) const
	{
	Indexes.clear();
	for (map<gpos_t, unsigned>::const_iterator p = m_GeneIndexes.begin();
	  p != m_GeneIndexes.end(); ++p)
		Indexes.insert(p->second);
	}

void Evolver::GetGeneRanges(vector<gpos_t> &Starts, vector<gpos_t> &Lengths) const
	{
	Starts.clear();
	Lengths.clear();

	bool InGene = false;
	gpos_t Length = 0;
	gpos_t EvPos = 0;
	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];
		if (Base.IsGene())
			{
			if (!InGene)
				{
				Starts.push_back(p);
				InGene = true;
				Length = 1;
				}
			else
				++Length;
			}
		else
			{
			if (InGene)
				Lengths.push_back(Length);
			InGene = false;
			}
		++EvPos;
		}
	if (InGene)
		Lengths.push_back(Length);

	asserta(Starts.size() == Lengths.size());
	}

void Evolver::GetGeneRangesEv(IvTree<LoHi> &Genes) const
	{
	bool InGene = false;
	LoHi LH;
	LH.Lo = GPOS_T_MAX;
	LH.Hi = 0;
	gpos_t EvPos = 0;
	FOR_EVBASES(AncPos)
		{
		const BaseInfo &Base = m_Bases[AncPos];
		if (Base.IsGene())
			{
			if (!InGene)
				{
				LH.Lo = EvPos;
				LH.Hi = EvPos;
				LH.Index = UINT_MAX;
				InGene = true;
				}
			else
				++LH.Hi;
			}
		else
			{
			if (InGene)
				{
				if (LH.Index == UINT_MAX)
					Die("GetGeneRangesEv, Start codon not found");
				Genes.Add(LH);
				}
			InGene = false;
			}

		if (Base.IsStartCodon())
			{
			asserta(InGene);
			map<gpos_t, unsigned>::const_iterator p = m_GeneIndexes.find(AncPos);
			if (p != m_GeneIndexes.end())
				{
				if (LH.Index != UINT_MAX)
					Die("GetGeneRangesEv, two start codons found");
				LH.Index = p->second;
				}
			}

		++EvPos;
		}
	if (InGene)
		Genes.Add(LH);
	}

gpos_t Evolver::GetGeneRange(gpos_t Pos, gpos_t &Start, gpos_t &Length) const
	{
	const BaseInfo &Base = GetBase(Pos);
	asserta(Base.IsGene());
	Length = 1;

	Start = Pos;
	for (;;)
		{
		gpos_t PrevPos = GetPrevPos(Start);
		if (PrevPos == GPOS_T_MAX)
			break;
		const BaseInfo &Base = GetBase(PrevPos);
		if (!Base.IsGene())
			break;
		Start = PrevPos;
		++Length;
		}

	gpos_t End = Pos;
	for (;;)
		{
		gpos_t NextPos = GetNextPos(End);
		if (NextPos == GPOS_T_MAX)
			break;
		const BaseInfo &Base = GetBase(NextPos);
		if (!Base.IsGene())
			break;
		End = NextPos;
		++Length;
		}
	return End;
	}

void Evolver::ScanGenes(vector<GeneInfo> &Genes) const
	{
	Genes.clear();
	gpos_t Pos = m_FirstEvolvedPos;
	for (;;)
		{
		if (Pos == GPOS_T_MAX)
			return;
		if (PosInGene(Pos))
			{
			gpos_t Start = GPOS_T_MAX;
			gpos_t Length = GPOS_T_MAX;
			gpos_t End = GetGeneRange(Pos, Start, Length);

			GeneInfo Gene;
			GetGene(Start, Gene);
			Genes.push_back(Gene);

			Pos = End;
			}
		Pos = GetNextPos(Pos);
		}
	}

void Evolver::GetGenes(vector<GeneInfo> &Genes) const
	{
	for (map<gpos_t, unsigned>::const_iterator p = m_GeneIndexes.begin();
	  p != m_GeneIndexes.end(); ++p)
		{
		gpos_t StartPos = p->first;
		GeneInfo Gene;
		GetGene(StartPos, Gene);
		Genes.push_back(Gene);
		}
	}

static bool SafeIsExon(const BaseInfo *Base)
	{
	if (Base == 0)
		return false;
	return Base->IsExon();
	}

static bool SafeIsIntron(const BaseInfo *Base)
	{
	if (Base == 0)
		return false;
	return Base->IsIntron();
	}

static bool SafeIsCDS(const BaseInfo *Base)
	{
	if (Base == 0)
		return false;
	return Base->IsCDS();
	}

static bool SafeIsUTR(const BaseInfo *Base)
	{
	if (Base == 0)
		return false;
	return Base->IsUTR();
	}

static bool SafeIsUntranscribed(const BaseInfo *Base)
	{
	if (Base == 0)
		return false;
	return Base->IsUntranscribedGene();
	}

static bool SafeIsTranscribed(const BaseInfo *Base)
	{
	if (Base == 0)
		return false;
	return Base->IsExon() || Base->IsIntron();
	}

static bool CDSStart(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return !SafeIsCDS(Base1) && SafeIsCDS(Base2);
	}

static bool CDSEnd(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return SafeIsCDS(Base1) && !SafeIsCDS(Base2);
	}

static bool ExonStart(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return !SafeIsExon(Base1) && SafeIsExon(Base2);
	}

static bool ExonEnd(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return SafeIsExon(Base1) && !SafeIsExon(Base2);
	}

static bool UTRStart(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return !SafeIsUTR(Base1) && SafeIsUTR(Base2);
	}

static bool UTREnd(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return SafeIsUTR(Base1) && !SafeIsUTR(Base2);
	}

static bool TranscriptionStart(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return !SafeIsTranscribed(Base1) && SafeIsTranscribed(Base2);
	}

static bool IntronStart(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return !SafeIsIntron(Base1) && SafeIsIntron(Base2);
	}

static bool IntronEnd(const BaseInfo *Base1, const BaseInfo *Base2)
	{
	return SafeIsIntron(Base1) && !SafeIsIntron(Base2);
	}

void Evolver::GetGene(gpos_t Pos, GeneInfo &Gene) const
	{
	Gene.Clear();

	Gene.EndPos = GetGeneRange(Pos, Gene.Pos, Gene.Length);

	const BaseInfo &FirstBase = GetBase(Gene.Pos);
	if (!FirstBase.IsGene())
		Die("GetGene: pos not in gene");

	ExonInfo Exon;
	IntronInfo Intron;
	CDSInfo CDS;
	UTRInfo UTR;
	gpos_t CurrPos = Gene.Pos;
	const BaseInfo *PrevBase = 0;
	bool StrandKnown = false;
	gpos_t TotalConstrainedBaseCount = 0;
	double TotalAccProb = 0;

	unsigned ExonStartCount = 0;
	unsigned CDSStartCount = 0;
	unsigned UTRStartCount = 0;
	unsigned IntronStartCount = 0;

	unsigned ExonEndCount = 0;
	unsigned CDSEndCount = 0;
	unsigned UTREndCount = 0;
	unsigned IntronEndCount = 0;

	gpos_t TranscriptionStartPos = GPOS_T_MAX;

	for (gpos_t i = 0; i <= Gene.Length; ++i)
		{
		const BaseInfo *Base = (i == Gene.Length ? 0 : &(GetBase(CurrPos)));

		bool Transcribed = (SafeIsExon(Base) || SafeIsIntron(Base));
		if (!Transcribed && Base != 0 && !SafeIsUntranscribed(Base))
			{
			Log("\n");
			Base->LogMe(true);
			Die("Gene base must be exon, intron or untranscribed");
			}

		if (TranscriptionStart(PrevBase, Base))
			{
			if (TranscriptionStartPos == GPOS_T_MAX)
				TranscriptionStartPos = CurrPos;
			else
				{
				if (Gene.Length < 10000)
					LogRangeAsTable(Gene.Pos, Gene.Length);
				Die("Gene has two transcribed regions %"PF", %"PF,
				  TranscriptionStartPos, CurrPos);
				}
			}

		if (Base != 0 && Base->IsConstrained())
			{
			++TotalConstrainedBaseCount;
			TotalAccProb += Base->GetAcceptProb();
			}

		if (CDSStart(PrevBase, Base))
			{
			++CDSStartCount;
			CDS.Clear();
			CDS.Pos = CurrPos;
			CDS.Length = 0;
			CDS.MeanAccProb = 0;
			CDS.ExonIndex = SIZE(Gene.Exons);
			}
		else if (UTRStart(PrevBase, Base))
			{
			++UTRStartCount;
			UTR.Clear();
			UTR.Pos = CurrPos;
			UTR.Length = 0;
			UTR.MeanAccProb = 0;
			UTR.ExonIndex = SIZE(Gene.Exons);
			}

		if (ExonStart(PrevBase, Base))
			{
			assert(CDSStart(PrevBase, Base) || UTRStart(PrevBase, Base));
			++ExonStartCount;
			Exon.Clear();
			Exon.Pos = CurrPos;
			Exon.Length = 0;
			Exon.MeanAccProb = 0;
			Exon.AllCoding = true;
			Exon.AllUTR = true;
			}
		else if (IntronStart(PrevBase, Base))
			{
			++IntronStartCount;
			Intron.Clear();
			Intron.Pos = CurrPos;
			Intron.Length = 0;
			}

		if (SafeIsCDS(Base))
			{
			Exon.AllUTR = false;
			Exon.CDSIndex = SIZE(Gene.CDSs);
			if (!StrandKnown)
				Gene.Plus = Base->PlusFrame();
			else
				{
				if (Gene.Plus != Base->PlusFrame())
					{
					Log("\n");
					Log("First base: ");
					FirstBase.LogMe();
					Log("\n");
					Log("Curr. base: ");
					Base->LogMe();
					Log("\n");
					Die("Gene has mixed strands");
					}
				}
			++CDS.Length;
			CDS.MeanAccProb += Base->GetAcceptProb();
			}
		else if (SafeIsUTR(Base))
			{
			Exon.AllCoding = false;
			Exon.UTRIndex = SIZE(Gene.UTRs);
			UTR.MeanAccProb += Base->GetAcceptProb();
			++UTR.Length;
			}

		if (SafeIsExon(Base))
			{
			Exon.MeanAccProb += Base->GetAcceptProb();
			++Exon.Length;
			}

		if (SafeIsIntron(Base))
			++Intron.Length;

		if (ExonEnd(PrevBase, Base))
			{
			++ExonEndCount;
			Gene.Exons.push_back(Exon);
			}
	
		if (IntronEnd(PrevBase, Base))
			{
			++IntronEndCount;
			Gene.Introns.push_back(Intron);
			}

		if (CDSEnd(PrevBase, Base))
			{
			++CDSEndCount;
			Gene.CDSs.push_back(CDS);
			}
		else if (UTREnd(PrevBase, Base))
			{
			++UTREndCount;
			Gene.UTRs.push_back(UTR);
			}

		if (i == Gene.Length || CurrPos == GPOS_T_MAX)
			break;
		CurrPos = GetNextPos(CurrPos);
		PrevBase = Base;
		}

	asserta(ExonStartCount == ExonEndCount);
	asserta(IntronStartCount == IntronEndCount);
	asserta(CDSStartCount == CDSEndCount);
	asserta(UTRStartCount == UTREndCount);

	if (TotalConstrainedBaseCount == 0)
		{
		if (Gene.Length < 10000)
			LogRangeAsTable(Gene.Pos, Gene.Length);
		Die("GetGene: no constrained bases");
		}
	Gene.MeanAccProb = TotalAccProb/TotalConstrainedBaseCount;

	const unsigned ExonCount = SIZE(Gene.Exons);
	const unsigned IntronCount = SIZE(Gene.Introns);
	const unsigned CDSCount = SIZE(Gene.CDSs);
	const unsigned UTRCount = SIZE(Gene.UTRs);

	for (unsigned i = 0; i < ExonCount; ++i)
		{
		ExonInfo &Exon = Gene.Exons[i];
		asserta(Exon.Length > 0);
		Exon.MeanAccProb /= Exon.Length;
		}

	for (unsigned i = 0; i < CDSCount; ++i)
		{
		CDSInfo &CDS = Gene.CDSs[i];
		asserta(CDS.Length > 0);
		CDS.MeanAccProb /= CDS.Length;
		}

	for (unsigned i = 0; i < UTRCount; ++i)
		{
		UTRInfo &UTR = Gene.UTRs[i];
		asserta(UTR.Length > 0);
		UTR.MeanAccProb /= UTR.Length;
		asserta(UTR.ExonIndex < ExonCount);
		asserta(!Gene.Exons[UTR.ExonIndex].AllCoding);
		}

	uint8 Frame = 0;
	for (unsigned i = 0; i < CDSCount; ++i)
		{
		unsigned CDSIndex = (Gene.Plus ? i : CDSCount - i - 1);
		CDSInfo &CDS = Gene.CDSs[CDSIndex];
		CDS.Frame = Frame;
		gpos_t FirstPos =
		  Gene.Plus ? CDS.Pos : GetRangeEndPos(CDS.Pos, CDS.Length);
		if (!IsValidPos(FirstPos))
			{
			Gene.LogMe();
			Die("CDS %u invalid first pos=%"PF, i, FirstPos);
			}
		const BaseInfo &Base = GetBase(FirstPos);
		CDS.Codex = Base.Codex;
		Frame = IncFrame(Frame, CDS.Length);
		asserta(CDS.ExonIndex < ExonCount);
		asserta(!Gene.Exons[CDS.ExonIndex].AllUTR);
		}

	string Codon;
	if (Gene.CDSs.empty())
		{
		Gene.LogMe();
		Die("No CDSs in gene");
		}
	const CDSInfo &FirstCDS = Gene.CDSs.front();
	const CDSInfo &LastCDS = Gene.CDSs.back();
	bool Strand = Gene.Plus;
	if (Strand)
		{
		Gene.StartCodonFirstBasePos = FirstCDS.Pos;
		Gene.StopCodonLastBasePos = GetRangeEndPos(LastCDS.Pos, LastCDS.Length);
		}
	else
		{
		Gene.StartCodonFirstBasePos = GetRangeEndPos(LastCDS.Pos, LastCDS.Length);
		Gene.StopCodonLastBasePos = FirstCDS.Pos;
		}

	gpos_t StartPos2 = GetAdjExonPos(Gene.StartCodonFirstBasePos, Strand);
	asserta(StartPos2 != GPOS_T_MAX);

	Gene.StartCodonLastBasePos = GetAdjExonPos(StartPos2, Strand);
	asserta(Gene.StartCodonLastBasePos != GPOS_T_MAX);

	gpos_t StopPos2 = GetAdjExonPos(Gene.StopCodonLastBasePos, !Strand);
	asserta(StopPos2 != GPOS_T_MAX);

	Gene.StopCodonFirstBasePos = GetAdjExonPos(StopPos2, !Strand);
	asserta(Gene.StopCodonFirstBasePos != GPOS_T_MAX);

	if (ExonCount != IntronCount + 1)
		{
		LogGene(Gene);
		Gene.LogMe();
		Die("GetGene: ExonCount %u != IntronCount %u + 1",
		  ExonCount, IntronCount);
		}

	Gene.Index = UINT_MAX;
	map<gpos_t, unsigned>::const_iterator p =
	  m_GeneIndexes.find(Gene.StartCodonFirstBasePos);
	if (p == m_GeneIndexes.end())
		{
		// LogGeneIndexes();
		LogGene(Gene);
		Log("Closest entries: ");
		p = m_GeneIndexes.upper_bound(Gene.StartCodonFirstBasePos);
		if (p == m_GeneIndexes.end())
			Log("upper_bound=end()\n");
		else
			{
			Log("StartPos %"PF" GeneIndex %u\n", p->first, p->second);
			if (p != m_GeneIndexes.begin())
				{
				--p;
				Log("StartPos %"PF" GeneIndex %u\n", p->first, p->second);
				}
			}
		Die("GetGene, index %"PF", not in map", Gene.StartCodonFirstBasePos);
		}
	Gene.Index = p->second;

	asserta(ExonCount > 0);
	asserta(ExonCount == IntronCount + 1);
	asserta(CDSCount > 0);
	}

void Evolver::LogGeneHeading() const
	{
	Log("\n");
	Log(" Gene  Feature         Pos      Length         End    MAC  St.  Stp  Fr  Ex  +\n");
	Log("=====  =======  ==========  ==========  ==========  =====  ===  ===  ==  ==  =\n");
	}

void Evolver::LogGenes(const vector<GeneInfo> &Genes) const
	{
	const unsigned N = SIZE(Genes);
	LogGeneHeading();
	for (unsigned i = 0; i < N; ++i)
		{
		const GeneInfo &Gene = Genes[i];
		LogGene(Gene, false);
		}
	}

void Evolver::LogGene(const GeneInfo &Gene, bool Heading) const
	{
	if (Heading)
		{
		Log("\n");
		Log("Gene");
		if (Gene.Index != UINT_MAX)
			Log(" %u", Gene.Index);
		Log(" Start1 %"PF, Gene.StartCodonFirstBasePos);
		Log(" Start3 %"PF, Gene.StartCodonLastBasePos);
		Log(" Stop1 %"PF, Gene.StopCodonFirstBasePos);
		Log(" Stop3 %"PF, Gene.StopCodonLastBasePos);
		Log("\n");
		LogGeneHeading();
		}

	bool Strand = Gene.Plus;

	string StartCodon;
	string StopCodon;

	GetNextExonTriplet(Gene.StartCodonFirstBasePos,
	  StartCodon, Strand);

	GetNextExonTriplet(Gene.StopCodonFirstBasePos,
	  StopCodon, Strand);

	const unsigned ExonCount = SIZE(Gene.Exons);
	const unsigned IntronCount = SIZE(Gene.Introns);
	const unsigned CDSCount = SIZE(Gene.CDSs);
	const unsigned UTRCount = SIZE(Gene.UTRs);

	if (Gene.Index == UINT_MAX)
		Log("    *");
	else
		Log("%5u", Gene.Index);
	Log("  %7.7s", "Gene");
	Log("  %10"PF, Gene.Pos);
	Log("  %10"PF, Gene.Length);
	Log("  %10"PF, GetRangeEndPos(Gene.Pos, Gene.Length));
	Log("  %5.3f", Gene.MeanAccProb);
	Log("  %3.5s", StartCodon.c_str());
	Log("  %3.3s", StopCodon.c_str());
	Log("        ");
	Log("  %c", pom(Gene.Plus));
	Log("\n");
	Log("\n");

	for (unsigned i = 0; i < ExonCount; ++i)
		{
		const ExonInfo &Exon = Gene.Exons[i];
		gpos_t ExonEndPos = GetRangeEndPos(Exon.Pos, Exon.Length);
		gpos_t StartPos = Strand ? Exon.Pos : ExonEndPos;
		gpos_t EndPos = Strand ? ExonEndPos : Exon.Pos;
		EndPos = GetAdjPos(EndPos, !Strand);

		string Start;
		string End;

		GetNextPair(StartPos, Start, Strand);
		GetNextPair(EndPos, End, Strand);

		Log("     ");
		Log("  %7.7s", "Exon");
		Log("  %10"PF, Exon.Pos);
		Log("  %10"PF, Exon.Length);
		Log("  %10"PF, ExonEndPos);
		Log("  %5.3f", Exon.MeanAccProb);
		Log("  %3.3s", Start.c_str());
		Log("  %3.3s", End.c_str());
		Log("    ");
		Log("  %2u", i);
		Log("  %c", pom(Gene.Plus));
		Log("\n");
		}
	Log("\n");

	for (unsigned i = 0; i < IntronCount; ++i)
		{
		const IntronInfo &Intron = Gene.Introns[i];

		gpos_t IntronEndPos = GetRangeEndPos(Intron.Pos, Intron.Length);
		gpos_t StartPos = Strand ? Intron.Pos : IntronEndPos;
		gpos_t EndPos = Strand ? IntronEndPos : Intron.Pos;
		EndPos = GetAdjPos(EndPos, !Strand);

		string Start;
		string End;

		GetNextPair(StartPos, Start, Strand);
		GetNextPair(EndPos, End, Strand);

		Log("     ");
		Log("  %7.7s", "Intron");
		Log("  %10"PF, Intron.Pos);
		Log("  %10"PF, Intron.Length);
		Log("  %10"PF, GetRangeEndPos(Intron.Pos, Intron.Length));
		Log("       ");
		Log("  %3.3s", Start.c_str());
		Log("  %3.3s", End.c_str());
		Log("        ");
		Log("  %c", pom(Gene.Plus));
		Log("\n");
		}
	if (IntronCount > 0)
		Log("\n");

	for (unsigned i = 0; i < UTRCount; ++i)
		{
		const UTRInfo &UTR = Gene.UTRs[i];
		Log("     ");
		Log("  %7.7s", "UTR");
		Log("  %10"PF, UTR.Pos);
		Log("  %10"PF, UTR.Length);
		Log("  %10"PF, GetRangeEndPos(UTR.Pos, UTR.Length));
		Log("  %5.3f", UTR.MeanAccProb);
		Log("          ");
		Log("    ");
		Log("  %2u", UTR.ExonIndex);
		Log("  %c", pom(Gene.Plus));
		Log("\n");
		}
	if (UTRCount > 0)
		Log("\n");

	for (unsigned i = 0; i < CDSCount; ++i)
		{
		const CDSInfo &CDS = Gene.CDSs[i];
		Log("     ");
		Log("  %7.7s", "CDS");
		Log("  %10"PF, CDS.Pos);
		Log("  %10"PF, CDS.Length);
		Log("  %10"PF, GetRangeEndPos(CDS.Pos, CDS.Length));
		Log("  %5.3f", CDS.MeanAccProb);
		Log("          ");
		Log("  %2d", CDS.Frame);
		Log("  %2u", CDS.ExonIndex);
		Log("  %c", pom(Gene.Plus));
		Log("  cx%d", CDS.Codex);
		Log("\n");
		}
	Log("\n");
	}

unsigned Evolver::ValidateGenes(bool ShowProgress) const
	{
	if (ShowProgress)
		Progress("Validate genes");

	unsigned ErrCount = 0;
	vector<GeneInfo> Genes;
	vector<GeneInfo> Genes2;
	ScanGenes(Genes);
	GetGenes(Genes2);
	const unsigned N = SIZE(Genes);
	const unsigned N2 = SIZE(Genes2);
	if (N != N2)
		{
		++ErrCount;
		Warning("ScanGenes returns %u genes, GetGenes %u", N, N2);
		}

	if (N != SIZE(m_GeneIndexes))
		{
		++ErrCount;
		Warning("ScanGenes returns %u genes, m_GeneIndexes has %u",
		  N, SIZE(m_GeneIndexes));
		}
		
	for (unsigned i = 0; i < N; ++i)
		ErrCount += ValidateGene(Genes[i]);
	if (ErrCount > 0)
		Die("Gene structure errors");
	return ErrCount;
	}

unsigned Evolver::ValidateGene(const GeneInfo &Gene) const
	{
	unsigned ErrCount = 0;

	bool Strand = Gene.Plus;
	const unsigned ExonCount = SIZE(Gene.Exons);
	const unsigned IntronCount = SIZE(Gene.Introns);
	const unsigned CDSCount = SIZE(Gene.CDSs);

	asserta(ExonCount > 0);
	asserta(ExonCount == IntronCount + 1);
	asserta(CDSCount <= ExonCount);

	CodonInfo StartCI;
	CodonInfo StopCI;

	if (!GetCodon(Gene.StartCodonFirstBasePos, StartCI))
		{
		++ErrCount;
		Warning("GetCodon at StartCodonFirstBasePos=%"PF" failed",
		  Gene.StartCodonFirstBasePos);
		}

	if (!GetCodon(Gene.StopCodonFirstBasePos, StopCI))
		{
		++ErrCount;
		Warning("GetCodon at StopCodonFirstBasePos=%"PF" failed",
		  Gene.StopCodonFirstBasePos);
		}

	for (unsigned i = 0; i < 3; ++i)
		{
		gpos_t Pos = StartCI.Pos[i];
		if (Pos == GPOS_T_MAX)
			{
			++ErrCount;
			StartCI.LogMe(true);
			Warning("Get start codon failed");
			continue;
			}
		const BaseInfo &Base = GetBase(Pos);
		if (!Base.IsStartCodon())
			{
			++ErrCount;
			StartCI.LogMe(true);
			Base.LogMe(true);
			Warning("StartCodon[%u] at pos %"PF" not start codon", i, Pos);
			}
		}

	for (unsigned i = 0; i < 3; ++i)
		{
		gpos_t Pos = StopCI.Pos[i];
		if (Pos == GPOS_T_MAX)
			{
			++ErrCount;
			StopCI.LogMe(true);
			Warning("Get stop codon failed");
			continue;
			}
		const BaseInfo &Base = GetBase(Pos);
		if (!Base.IsStopCodon())
			{
			++ErrCount;
			StopCI.LogMe(true);
			Base.LogMe(true);
			Warning("StopCodon[%u] at pos %"PF" not stop codon", i, Pos);
			}
		}

	string StartCodon;
	string StopCodon;

	GetNextExonTriplet(Gene.StartCodonFirstBasePos, StartCodon, Strand);
	GetNextExonTriplet(Gene.StopCodonFirstBasePos, StopCodon, Strand);

	if (!IsStartCodon(StartCodon))
		{
		++ErrCount;
		Warning("First codon at pos %"PF" is %s, should be ATG",
		  Gene.StartCodonFirstBasePos, StartCodon.c_str());
		}

	if (!IsStopCodon(StopCodon))
		{
		++ErrCount;
		Warning("Last codon at pos %"PF" is %s, should be STOP (TAG,TAA,TGA)",
		  Gene.StopCodonFirstBasePos, StopCodon.c_str());
		}

	string StopCodon2;
	GetPrevExonTriplet(Gene.StopCodonLastBasePos, StopCodon2, Strand);

	if (StopCodon2 != StopCodon)
		Die("Stop codon (2) bug gene %u %s %s",
		  Gene.Index, StopCodon.c_str(), StopCodon2.c_str());

	string Codon;
	gpos_t Pos = Gene.StartCodonFirstBasePos;
	while (Pos != GPOS_T_MAX)
		{
		if (!IsCodon(Pos, Strand))
			{
			++ErrCount;
			{
			Log("\n");
			Log("Expecting codexes 1,2,3 or -1,-2,-3 but got:\n");
			gpos_t Pos2 = Gene.StartCodonFirstBasePos;
			while (Pos2 != GPOS_T_MAX)
				{
				const BaseInfo &Base = GetBase(Pos2);
				Log("%"PF" ", Pos2);
				Base.LogMe(true);
				Pos2 = GetAdjExonPos(Pos2, Strand);
				}
			Log("\n");
			}
			Log("\n");
			LogExonTriplet(Pos, Strand);
			Warning("Invalid codon or missing STOP in CDS %"PF"%c", Pos, pom(Strand));
			break;
			}
		gpos_t NextPos = GetNextExonTriplet(Pos, Codon, Strand);
		if (IsStopCodon(Codon))
			{
			++ErrCount;
			Warning("Stop codon %s in CDS at pos %"PF, Codon.c_str(), Pos);
			break;
			}
		if (NextPos == Gene.StopCodonFirstBasePos)
			break;
		Pos = NextPos;
		}

	Pos = Gene.Pos;
	unsigned SpliceBaseCount = 0;
	unsigned StartBaseCount = 0;
	unsigned StopBaseCount = 0;
	unsigned CDSBaseCount = 0;
	for (gpos_t i = 0; i < Gene.Length; ++i)
		{
		const BaseInfo &Base = GetBase(Pos);
		if (Base.IsSplice())
			{
			if (!Base.IsFullyConstrained())
				{
				++ErrCount;
				Log("\n");
				Base.LogMe(true);
				Warning("SpliceBase not fully constrained");
				}
			++SpliceBaseCount;
			}
		else if (Base.IsStartCodon())
			++StartBaseCount;
		else if (Base.IsStopCodon())
			++StopBaseCount;
		else if (Base.IsCDS())
			++CDSBaseCount;

		Pos = GetNextPos(Pos);
		}

	if (CDSBaseCount%3 != 0)
		{
		++ErrCount;
		Warning("CDS base count = %u, not multiple of 3", CDSBaseCount);
		}

	if (SpliceBaseCount != 4*IntronCount)
		{
		++ErrCount;
		Warning("SpliceBaseCount=%u != 4*IntronCount=%u = %u",
		  SpliceBaseCount, IntronCount, 4*IntronCount);
		}

	if (StartBaseCount != 3 || StopBaseCount != 3)
		{
		++ErrCount;
		Warning("StartBaseCount = %u, StopBaseCount = %u",
		  StartBaseCount, StopBaseCount);
		}

	for (unsigned i = 0; i < IntronCount; ++i)
		{
		const IntronInfo &Intron = Gene.Introns[i];
		if (Intron.Length < 4)
			{
			++ErrCount;
			Log("\n");
			Intron.LogMe(true);
			Warning("Intron length < 4");
			continue;
			}
		Pos = Intron.Pos;
		const BaseInfo &Base1 = GetBase(Pos);
		if (!Base1.IsSplice())
			{
			++ErrCount;
			Log("\n");
			Gene.LogMe();
			Base1.LogMe(true);
			Warning("Expected splice(1) at pos %"PF, Pos);
			continue;
			}
		Pos = GetNextPos(Pos);
		const BaseInfo &Base2 = GetBase(Pos);
		if (!Base2.IsSplice())
			{
			++ErrCount;
			Log("\n");
			Gene.LogMe();
			Base2.LogMe(true);
			Warning("Expected splice(2) at pos %"PF, Pos);
			continue;
			}

		Pos = GetRangeEndPos(Intron.Pos, Intron.Length);
		const BaseInfo &Base3 = GetBase(Pos);
		if (!Base3.IsSplice())
			{
			++ErrCount;
			Log("\n");
			Gene.LogMe();
			Base3.LogMe(true);
			Warning("Expected splice(3) at pos %"PF, Pos);
			continue;
			}
		Pos = GetPrevPos(Pos);
		const BaseInfo &Base4 = GetBase(Pos);
		asserta(Base4.IsSplice());
		}

	if (ErrCount > 0)
		LogGene(Gene);
	return ErrCount;
	}

void GeneInfo::CDSOffsetToPos(gpos_t Offset, unsigned &CDSIndex,
  unsigned &CDSPos, gpos_t &Pos) const
	{
	gpos_t Remaining = Offset;
	const unsigned CDSCount = SIZE(CDSs);
	for (CDSIndex = 0; CDSIndex < CDSCount; ++CDSIndex)
		{
		const CDSInfo &CDS = GetCDSX(CDSIndex);
		if (Remaining < CDS.Length)
			{
			CDSPos = Remaining;
			Pos = Plus ? CDS.Pos + Remaining : CDS.Pos + CDS.Length - Remaining - 1;
			return;
			}
		Remaining -= CDS.Length;
		}
	Die("CDSOffsetToPos: overflowed CDSs");
	}

void Evolver::GetNeutralRanges(gpos_t Start, gpos_t Length, bool InGene,
  vector<gpos_t> &Starts, vector<gpos_t> &Lengths) const
	{
	Starts.clear();
	bool InNeutral  = false;
	gpos_t RangeLength = GPOS_T_MAX;
	gpos_t Pos = Start;
	for (gpos_t i = 0; i < Length; ++i)
		{
		if (Pos == UINT32_MAX)
			break;
		const BaseInfo &Base = GetBase(Pos);
		if ((InGene == Base.IsGene()) && Base.IsUnconstrained())
			{
			if (!InNeutral)
				{
				Starts.push_back(Pos);
				RangeLength = 0;
				InNeutral = true;
				}
			}
		else
			{
			if (InNeutral)
				{
				asserta(Length > 0);
				Lengths.push_back(RangeLength);
				}
			InNeutral = false;
			}
		if (InNeutral)
			++RangeLength;

		Pos = GetNextPos(Pos);
		}
	if (InNeutral)
		{
		asserta(RangeLength > 0);
		Lengths.push_back(RangeLength);
		}
	asserta(SIZE(Starts) == SIZE(Lengths));
	}

void Evolver::GetNGEs(vector<gpos_t> &Starts, vector<gpos_t> &Lengths,
  vector<double> &MACs) const
	{
	Starts.clear();
	bool InNGE  = false;
	gpos_t Length = GPOS_T_MAX;
	double SumMACs = FLT_MAX;
	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];
		if (Base.IsNGE())
			{
			if (!InNGE)
				{
				Starts.push_back(p);
				Length = 0;
				SumMACs = 0;
				InNGE = true;
				}
			}
		else
			{
			if (InNGE)
				{
				asserta(Length > 0);
				Lengths.push_back(Length);
				MACs.push_back(SumMACs/Length);
				}
			InNGE = false;
			}
		if (InNGE)
			{
			++Length;
			SumMACs += Base.GetAcceptProb();
			}
		}
	if (InNGE)
		{
		asserta(Length > 0);
		Lengths.push_back(Length);
		MACs.push_back(SumMACs/Length);
		}

	asserta(SIZE(Starts) == SIZE(Lengths));
	asserta(SIZE(Starts) == SIZE(MACs));
	}

void Evolver::GetNXEs(gpos_t RangeStart, gpos_t RangeLength, vector<gpos_t> &Starts,
  vector<gpos_t> &Lengths, vector<double> &MACs) const
	{
	Starts.clear();
	bool InNXE  = false;
	gpos_t Length = GPOS_T_MAX;
	double SumMACs = FLT_MAX;

	gpos_t Pos = RangeStart;
	for (gpos_t i = 0; i < RangeLength; ++i)
		{
		if (Pos == UINT32_MAX)
			break;
		const BaseInfo &Base = GetBase(Pos);
		if (Base.IsNXE())
			{
			if (!InNXE)
				{
				Starts.push_back(Pos);
				Length = 0;
				SumMACs = 0;
				InNXE = true;
				}
			}
		else
			{
			if (InNXE)
				{
				asserta(Length > 0);
				Lengths.push_back(Length);
				MACs.push_back(SumMACs/Length);
				}
			InNXE = false;
			}
		if (InNXE)
			{
			++Length;
			SumMACs += Base.GetAcceptProb();
			}
		Pos = GetNextPos(Pos);
		}
	if (InNXE)
		{
		asserta(Length > 0);
		Lengths.push_back(Length);
		MACs.push_back(SumMACs/Length);
		}

	asserta(SIZE(Starts) == SIZE(Lengths));
	asserta(SIZE(Starts) == SIZE(MACs));
	}

void Evolver::GetIslands(gpos_t RangeStart, gpos_t RangeLength, vector<gpos_t> &Starts,
  vector<gpos_t> &Lengths) const
	{
	Starts.clear();
	bool InIsland  = false;
	gpos_t Length = GPOS_T_MAX;

	gpos_t Pos = RangeStart;
	for (gpos_t i = 0; i < RangeLength; ++i)
		{
		if (Pos == UINT32_MAX)
			break;
		const BaseInfo &Base = GetBase(Pos);
		if (Base.IsIsland())
			{
			if (!InIsland)
				{
				Starts.push_back(Pos);
				Length = 0;
				InIsland = true;
				}
			}
		else
			{
			if (InIsland)
				{
				asserta(Length > 0);
				Lengths.push_back(Length);
				}
			InIsland = false;
			}
		if (InIsland)
			++Length;
		Pos = GetNextPos(Pos);
		}
	if (InIsland)
		{
		asserta(Length > 0);
		Lengths.push_back(Length);
		}

	asserta(SIZE(Starts) == SIZE(Lengths));
	}

void Evolver::GetOceans(gpos_t RangeStart, gpos_t RangeLength, vector<gpos_t> &Starts,
  vector<gpos_t> &Lengths) const
	{
	Starts.clear();
	bool InOcean  = false;
	gpos_t Length = GPOS_T_MAX;

	gpos_t Pos = RangeStart;
	for (gpos_t i = 0; i < RangeLength; ++i)
		{
		if (Pos == UINT32_MAX)
			break;
		const BaseInfo &Base = GetBase(Pos);
		if (Base.IsOcean())
			{
			if (!InOcean)
				{
				Starts.push_back(Pos);
				Length = 0;
				InOcean = true;
				}
			}
		else
			{
			if (InOcean)
				{
				asserta(Length > 0);
				Lengths.push_back(Length);
				}
			InOcean = false;
			}
		if (InOcean)
			++Length;
		Pos = GetNextPos(Pos);
		}
	if (InOcean)
		{
		asserta(Length > 0);
		Lengths.push_back(Length);
		}

	asserta(SIZE(Starts) == SIZE(Lengths));
	}

unsigned Evolver::GetGeneIndex(gpos_t StartCodonFirstBasePos) const
	{
	map<gpos_t, unsigned>::const_iterator p =
	  m_GeneIndexes.find(StartCodonFirstBasePos);
	if (p == m_GeneIndexes.end())
		Die("GetGeneIndex %"PF", not in map", StartCodonFirstBasePos);
	return p->second;
	}

void Evolver::UpdateStartCodonMap(gpos_t OldPos, gpos_t NewPos)
	{
	map<gpos_t, unsigned>::iterator p = m_GeneIndexes.find(OldPos);
	if (p == m_GeneIndexes.end())
		Die("UpdateStartCodonMap %"PF" not found", OldPos);
	unsigned GeneIndex = p->second;
	m_GeneIndexes.erase(p);
	m_GeneIndexes[NewPos] = GeneIndex;
	}
