#include "revfile.h"
#include "gff2.h"
#include "getopts.h"
#include "progress.h"

void RevComp(string &Seq);
void WriteFasta(FILE *f, const char *SeqData, unsigned Length);
bool IsStartCodon(const string &Codon);
bool IsStopCodon(const string &Codon);
char LettersToCodonChar(char c1, char c2, char c3);

static map<unsigned, string> GeneIndexToSeq;
static map<unsigned, string> GeneIndexToGenChr;
static map<unsigned, bool> GeneIndexToStrand;

void Xlat(const string &DNASeq, string &AASeq)
	{
	AASeq.clear();
	const unsigned L = SIZE(DNASeq);
	if (L%3 != 0)
		Die("Xlat: seq length %u, fractional codon", L);

	for (unsigned i = 0; i < L; )
		{
#define U(c)	((char) toupper((c)))
		char c1 = U(DNASeq[i++]);
		char c2 = U(DNASeq[i++]);
		char c3 = U(DNASeq[i++]);

		char AA = LettersToCodonChar(c1, c2, c3);
		if (i == L && AA == '*')
			break;
		AASeq.push_back(AA);
		}
	}

void GetProteins()
	{
	const char *SeqFileName = RequiredValueOpt("getprots");
	const char *GFFFileName = RequiredValueOpt("annots");
	const char *DNAFileName = ValueOpt("dna");
	const char *AAFileName = ValueOpt("aa");
	const char *GenomeName = ValueOpt("genome");
	unsigned GenomeIndex = 0;
	IntOpt("genix", &GenomeIndex);

	if (DNAFileName == 0 && AAFileName == 0)
		Die("Must specify -dna and/or -aa");

	RevFile Rev;
	Rev.Open(SeqFileName);

	if (GenomeName != 0)
		GenomeIndex = Rev.GetGenomeIndex(GenomeName);
	else
		GenomeName = Rev.GetGenomeName(GenomeIndex).c_str();

	FILE *fGFF = OpenStdioFile(GFFFileName);
	FILE *fDNA = 0;
	FILE *fAA = 0;

	if (DNAFileName != 0)
		fDNA = CreateStdioFile(DNAFileName);
	if (AAFileName != 0)
		fAA = CreateStdioFile(AAFileName);

	GFFRecord2 Rec;
	ProgressStartFile(fGFF, "Reading");
	unsigned Count = 0;
	while (GetNextGFFRecord(fGFF, Rec))
		{
		if (Rec.Feature != "CDS")
			continue;
		if (Count++%20 == 0)
			ProgressStepFile();

		unsigned ChrId = Rev.GetChrId(GenomeIndex, Rec.SeqName);

		asserta(Rec.Start > 0);
		coord_t lo = Rec.Start - 1;
		coord_t Length = Rec.GetLength();

		char *Seq = new char[Length+1];
		Rev.GetSubSeq(GenomeIndex, ChrId, Seq, lo, Length);
		Seq[Length] = 0;

		const string GenChr = string(GenomeName) + "." + Rec.SeqName;
		bool Strand = (Rec.Strand == '+');

		unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
		map<unsigned, string>::const_iterator p = GeneIndexToSeq.find(GeneIndex);
		if (p == GeneIndexToSeq.end())
			{
			GeneIndexToSeq[GeneIndex] = "";
			GeneIndexToGenChr[GeneIndex] = GenChr;
			GeneIndexToStrand[GeneIndex] = Strand;
			}
		else
			{
			map<unsigned, string>::const_iterator q = GeneIndexToGenChr.find(GeneIndex);
			asserta(q != GeneIndexToGenChr.end());
			if (q->second != GenChr)
				Die("Gene %u found both in two chrs, %s and %s",
				  q->second.c_str(), GenChr.c_str());

			map<unsigned, bool>::const_iterator r = GeneIndexToStrand.find(GeneIndex);
			asserta(r != GeneIndexToStrand.end());
			if (r->second != Strand)
				Die("Gene %u has CDS records on both + and - strand", GeneIndex);
			}

		GeneIndexToSeq[GeneIndex] += Seq;
		delete[] Seq;
		}
	ProgressDone();

	const unsigned GeneCount = SIZE(GeneIndexToSeq);
	ProgressStart("Writing %u genes", GeneCount);
	Count = 0;
	for (map<unsigned, string>::const_iterator p = GeneIndexToSeq.begin();
	  p != GeneIndexToSeq.end(); ++p)
		{
		if (Count++%100 == 0)
			ProgressStep(Count, GeneCount);

		unsigned GeneIndex = p->first;
		string Seq = p->second;

		map<unsigned, string>::const_iterator q = GeneIndexToGenChr.find(GeneIndex);
		asserta(q != GeneIndexToGenChr.end());

		map<unsigned, bool>::const_iterator r = GeneIndexToStrand.find(GeneIndex);
		asserta(r != GeneIndexToStrand.end());

		asserta(r->first == GeneIndex);
		bool Strand = r->second;
		const string &GenChr = q->second;

		unsigned L = SIZE(Seq);
		if (L < 3)
			{
			Warning("Gene %s.%u length %u bases, skipping",
			  GenChr.c_str(), GeneIndex, L);
			continue;
			}
		if (L%3 != 0)
			Warning("Gene %s.%u length %u bases, fractional codon",
			  GenChr.c_str(), GeneIndex, L);

		if (!Strand)
			RevComp(Seq);
	
		string Start;
		string Stop;
		Start.push_back((char) toupper(Seq[0]));
		Start.push_back((char) toupper(Seq[1]));
		Start.push_back((char) toupper(Seq[2]));
		Stop.push_back((char) toupper(Seq[L-3]));
		Stop.push_back((char) toupper(Seq[L-2]));
		Stop.push_back((char) toupper(Seq[L-1]));

		if (!IsStartCodon(Start))
			Warning("Gene %s.%u invalid start codon %s",
			  GenChr.c_str(), GeneIndex, Start.c_str());
		if (!IsStopCodon(Stop))
			Warning("Gene %s.%u invalid stop codon %s",
			  GenChr.c_str(), GeneIndex, Stop.c_str());
	
		if (fDNA != 0)
			{
			fprintf(fDNA, ">%s.%u\n", GenChr.c_str(), GeneIndex);
			WriteFasta(fDNA, Seq.c_str(), L);
			}

		if (fAA != 0)
			{
			if (L%3 != 0)
				Seq = Seq.substr(0, L - L%3);
			string AASeq;
			Xlat(Seq, AASeq);
			fprintf(fAA, ">%s.%u\n", GenChr.c_str(), GeneIndex);
			if (AASeq.find('*') != string::npos)
				Warning("In-frame stop codon in gene %s.%u", GenChr.c_str(), GeneIndex);
			WriteFasta(fAA, AASeq.c_str(), SIZE(AASeq));
			}
		}
	ProgressDone();

	Rev.Close();
	CloseStdioFile(fGFF);
	if (fDNA != 0)
		CloseStdioFile(fDNA);
	if (fAA != 0)
		CloseStdioFile(fAA);
	}
