#include <math.h>
#include "ratemgr.h"
#include "myrand.h"

#define TEST	0

/***
GetRandLengthRR() returns a random length sampled from the
"trapezoidal" distribution defined by the RateRange.

To see how this works, consider the rate histogram.

rlo   0
      0   1
      0   1
      0   1    i r(i)
      0   1    i       
      0   1    i       N  rhi
	  0   1    i       N
     Llo  ...  L  ... Lhi

To sample from this distribution, we conceptually lay out the
histogram along the real line 0..U, where U is the sum of rates
sum(L=Llo..Lhi) r(L):

Llo            L       Lhi
000000111111...iiiii...NN
0.0         ...   S     U

We generate a real random number S in the range 0 .. U, and
determine which histogram bar we are in, this gives us the
desired random L.

To determine the histogram bar, we need to solve the equation:

	s(L) = S

for L, where s(L) is the sum of histogram bars up to length L,
i.e. s(L) = sum(k=Llo .. L) r(k), where r(k) is the rate at
length k. By linear interpolation,

	r(L) = rlo + g(L - Llo)

where g is the gradient,

	g	= (rhi - rlo)/(Lhi - Llo).

Now we have:

	S	= sum(k=Llo..L) r(k)

		= sum(k=Llo..L) (rlo + g(k - Llo))

		= (L - Llo + 1) rlo + g sum(k=Llo..L) (k - Llo)

		= (L - Llo + 1) rlo + g sum(d=0..L - Llo) d

		= (d + 1) rlo + g d (d + 1)/2

where d = L - Lo,

		= d rlo + rlo + gd^2/2 + gd/2

giving the quadratic:

		g/2 d^2 + (rlo + g/2) d + rlo - S = 0.

Sampling S in the range 0 .. U gives d's in the range -1 .. (Lhi - Llo)
because the way we defined variables above, the sum for d = 0 is already
rlo, so d = 0 is the /end/ of the first bar, d = -1 is the start of the
first bar, and the random length we want is L = Llo + floor(d + 1).

Note that we must actually sample from 0 .. U-epsilon to avoid the special
case d=Lhi-Llo+1 when S=U.
***/

gpos_t GetRandLengthRR(const RateRange &RR)
	{
	if (RR.HiLength == RR.LoLength)
		return RR.HiLength;

// Length range
	double R = RR.HiLength - RR.LoLength;

// Random sum-of-lengths in range 0 .. SumRates-epsilon
	double SumRates = (R + 1)*(RR.LoRate + RR.HiRate)/2;
	double S = SumRates * (double) RandInt(UINT_MAX) /
	  ((double) UINT_MAX);

// Gradient
	double g = (RR.HiRate - RR.LoRate)/R;

// Have sum of rates S, solve for i.
// Quadratic a*i^2 + b*i + c = 0
	double a = g/2.0;
	double b = RR.LoRate + g/2.0;
	double c = RR.LoRate - S;

// Solve quadratic
	double F = b*b - 4*a*c;
	double d = (-b + sqrt(F))/(2*a);

// Round to integer length
	gpos_t L = RR.LoLength + (gpos_t) floor(d + 1);
	if (L < RR.LoLength)
		L = RR.LoLength;
	if (L > RR.HiLength)
		L = RR.HiLength;
	return L;
	}

#if	TEST
void TestLHF(gpos_t LoLength, gpos_t HiLength, double LoRate, double HiRate)
	{
	BigLHF RR;
	RR.LoLength = LoLength;
	RR.HiLength = HiLength;
	RR.LoRate = LoRate;
	RR.HiRate = HiRate;

	gpos_t R = RR.HiLength - RR.LoLength;

	vector<double> Rates(R+1);
	double SumRates = 0;
	for (gpos_t L = RR.LoLength; L <= RR.HiLength; ++L)
		{
		double r = RR.LoRate + (L - RR.LoLength)*(RR.HiRate - RR.LoRate)/R;
		SumRates += r;
		Rates[L - RR.LoLength] = r;
		}

	vector<double> ExpectedWeights(R+1);
	for (gpos_t L = RR.LoLength; L <= RR.HiLength; ++L)
		ExpectedWeights[L - RR.LoLength] = Rates[L - RR.LoLength]/SumRates;

	IntVec Counts(R+1);
	const unsigned N = 1000000;
	for (unsigned i = 0; i < N; ++i)
		{
		gpos_t L = GetRandLengthLHF(RR);
		asserta(L >= RR.LoLength && L <= RR.HiLength);
		++(Counts[L - RR.LoLength]);
		}

	vector<double> Weights(R+1);
	for (gpos_t L = RR.LoLength; L <= RR.HiLength; ++L)
		Weights[L - RR.LoLength] = double(Counts[L - RR.LoLength])/double(N);

	printf("\n");
	printf("\n");
	printf("Len %"PF" - %"PF", Rate %.3g - %.3g\n",
	  LoLength, HiLength, LoRate, HiRate);
	Log("\n");
	printf("         L        Rate  ExpectedWt      Weight     Acc\n");
	printf("==========  ==========  ==========  ==========  ======\n");
	for (gpos_t L = RR.LoLength; L <= RR.HiLength; ++L)
		{
		gpos_t i = L - RR.LoLength;
		double r = RR.LoRate + (L - RR.LoLength)*(RR.HiRate - RR.LoRate)/R;
		double e = ExpectedWeights[i];
		double w = Weights[i];
		double a = 100.0*w/e;
		printf("%10"PF"  %10.4g  %10.4g  %10.4g  %6.1f %%\n", L, r, e, w, a);
		}
	}

void TestLHF()
	{
	TestLHF(10, 100, 1.0, 10.0);
	}

#endif // TEST
