#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "types.h"
#include "utils.h"
#include "gff2.h"
#include <string>

#define	TRACE	0

void ParseStrVec(const std::string &In, std::vector<std::string> &Out, char c);

// GFF fields are:
// <seqname> <source> <feature> <start> <end> <score> <strand> <frame> [attributes] [comments]
//     0         1         2        3      4      5        6       7         8           9

static unsigned GFFLineNr;

unsigned GFFGetRequiredIntAttr(const string &Attrs,
  const string &Name)
	{
	unsigned Value = UINT_MAX;
	bool Found = GFFGetIntAttr(Attrs, Name, Value);
	if (!Found)
		Die("Required attr '%s' not found in '%.32s'",
		  Name.c_str(), Attrs.c_str());
	return Value;
	}

void GFFGetRequiredAttr(const GFFRecord2 &Rec, const string &Name,
  string &Value)
	{
	bool Found = GFFGetAttr(Rec.Attrs, Name, Value);
	if (!Found)
		{
		Rec.LogMe();
		Die("Required attr '%s' not found in GFF record", Name.c_str());
		}
	}

void GFFGetRequiredAttr(const string &Attrs, const string &Name,
  string &Value)
	{
	bool Found = GFFGetAttr(Attrs, Name, Value);
	if (!Found)
		Die("Required attr '%s' not found in '%.32s'",
		  Name.c_str(), Attrs.c_str());
	}

bool GFFGetIntAttr(const string &Attrs, const string &Name, unsigned &Value)
	{
	string strValue;
	bool Found = GFFGetAttr(Attrs, Name, strValue);
	if (!Found)
		return false;
	for (string::const_iterator p = strValue.begin();
	  p != strValue.end(); ++p)
		if (!isdigit(*p))
			Die("Invalid integer attribute %s=%s",
			  Name.c_str(), strValue.c_str());
	Value = (unsigned) atoi(strValue.c_str());
	return true;
	}

// This is more tolerant that UCSC GTF format, which is accepted as a subset.
// UCSC requires attributes to end in a semicolon which must then be separated
// from the start of any subsequent attribute by exactly one space character.
// I allow abitrary white space (including none) before and after the semi-colon,
// and also allow end-of-line to terminate the last attribute.
// <Idle><InName><AfterName>("<InStringValue>"|<InNumericValue>)<AfterValue>(;|EndOfLine)<Idle>
void GFFGetAttrs(const string &Attrs, vector<string> &Names, vector<string> &Values)
	{
	enum
		{
		Idle,
		InName,
		AfterName,
		InStringValue,
		InNumericValue,
		} State = Idle;

	string CurrentName;
	string CurrentValue;
#if	TRACE
	Log("State  c       CurrentName      CurrentValue\n");
	Log("=====  =  ================  ================\n");
#endif
	for (string::const_iterator p = Attrs.begin(); p != Attrs.end(); ++p)
		{
		char c = *p;
#if	TRACE
		Log("%5d  %c  %16.16s  %16.16s\n",
		  (int) State, c, CurrentName.c_str(), CurrentValue.c_str());
#endif
		if (State == Idle && !CurrentName.empty())
			{
			Names.push_back(CurrentName);
			Values.push_back(CurrentValue);
			CurrentName.clear();
			CurrentValue.clear();
			}
		switch (State)
			{
		case Idle:
			if (isspace(c))
				continue;
			else if (c == ';')
				continue;
			else if (isalpha(c))
				{
				CurrentName = c;
				CurrentValue.clear();
				State = InName;
				}
			else
				{
				Log("\n%.32s\n", Attrs.c_str());
				Die("Unexpected '%c' in attrs '%32.32s", c, Attrs.c_str());
				}
			break;

		case InName:
			if (isspace(c))
				State = AfterName;
			else if (c == '"')
				{
				CurrentValue = '"';
				State = InStringValue;
				}
			else
				CurrentName += c;
			break;

		case AfterName:
			if (isspace(c))
				continue;
			else if (c == '"')
				{
				CurrentValue = '"';
				State = InStringValue;
				}
			else if (isdigit(c) || c == '.')
				{
				CurrentValue = c;
				State = InNumericValue;
				continue;
				}
			else
				{
				Log("\n%.32s\n", Attrs.c_str());
				Die("Unexpected '%c' after '%s' in attrs '%.32s'",
				  c, CurrentName.c_str(), Attrs.c_str());
				}
			break;

		case InStringValue:
			if (c == '"')
				{
				CurrentValue += '"';
				State = Idle;
				continue;
				}
			CurrentValue += c;
			break;

		case InNumericValue:
			if (c == ';' || isspace(c))
				{
				State = Idle;
				continue;
				}
			else if (isdigit(c) || c == '.')
				CurrentValue += c;
			else
				Die("'%c' invalid in numeric attribute '%s'",
				  c, CurrentName.c_str());
			break;

		default:
			asserta(false);
			}
		}
	if (State != Idle && State != InNumericValue)
		Die("Unexpected end-of-line in attributes '%.32s'", Attrs.c_str());

	if (!CurrentName.empty())
		{
		Names.push_back(CurrentName);
		Values.push_back(CurrentValue);
		}
	}

bool GFFGetAttr(const string &Attrs, const string &Name, string &Value)
	{
	vector<string> Names;
	vector<string> Values;
	GFFGetAttrs(Attrs, Names, Values);
	const unsigned N = SIZE(Names);
	asserta(SIZE(Values) == N);
	for (unsigned i = 0; i < N; ++i)
		if (Names[i] == Name)
			{
			Value = Values[i];
			const unsigned L = SIZE(Value);
			if (L >= 2 && Value[0] == '"')
				Value = Value.substr(1, L-2);
			return true;
			}

	return false;
	}

bool GetNextGFFRecord(FILE *f, GFFRecord2 &Rec)
	{
	for (;;)
		{
		++GFFLineNr;
		string Line;
		bool EndOfFile = ReadLineStdioFile(f, Line);
		if (EndOfFile)
			return false;

		if ('#' == Line[0])
			continue;

		StdStrVec Fields;
		ParseStrVec(Line, Fields, '\t');
		unsigned FieldCount = SIZE(Fields);
		if (FieldCount != 8 && FieldCount != 9)
			Die("GFF record must have 8 or 9 fields, has %u in line %u",
			  FieldCount, GFFLineNr);

		char *NotUsed;

		Rec.SeqName = Fields[0];
		Rec.Source = Fields[1];
		Rec.Feature = Fields[2];
		Rec.Start = strtoul(Fields[3].c_str(), &NotUsed, 10);
		Rec.End = strtoul(Fields[4].c_str(), &NotUsed, 10);;
		Rec.Score = (float) atof(Fields[5].c_str());
		Rec.Strand = Fields[6][0];
		Rec.Frame = Fields[7][0] == '.' ? INT_MAX : atoi(Fields[7].c_str());

		if (FieldCount > 8)
			Rec.Attrs = Fields[8];
		else
			Rec.Attrs.clear();

		return true;
		}
	}

// GFF fields are:
// <seqname> <source> <feature> <start> <end> <score> <strand> <frame> [attributes] [comments]
//     0         1         2        3      4      5        6       7         8           9
void WriteGFFRec(FILE *f, const GFFRecord2 &Rec)
	{
	fprintf(f, "%s\t%s\t%s",
	  Rec.SeqName.c_str(),
	  Rec.Source.c_str(),
	  Rec.Feature.c_str());
	fprintf(f, "\t%"COORD_T_PRINTF, Rec.Start);
	fprintf(f, "\t%"COORD_T_PRINTF, Rec.End);
	fprintf(f, "\t%g\t%c", Rec.Score, Rec.Strand);
	if (Rec.Frame == INT_MAX)
		fprintf(f, "\t.");
	else
		fprintf(f, "\t%d", Rec.Frame);
	fprintf(f, "\t%s\n", Rec.Attrs.c_str());
	}

void LogGFFRec(const GFFRecord2 &Rec)
	{
	Log("%s %s %s", Rec.SeqName.c_str(), Rec.Source.c_str(), Rec.Feature.c_str());
	Log(" Start=%"COORD_T_PRINTF, Rec.Start);
	Log(" End=%"COORD_T_PRINTF, Rec.End);
	Log(" Score=%g %c", Rec.Score, Rec.Strand);
	if (Rec.Frame == -1)
		Log(" .");
	else
		Log(" %d", Rec.Frame);
	Log(" %s\n", Rec.Attrs.c_str());
	}

void GFFAppendIntAttr(string &Attrs, const string &Name, unsigned Value)
	{
	char strValue[32];
	sprintf(strValue, "%u", Value);
	if (Attrs.empty())
		Attrs = Name + " " + strValue + ";";
	else
		Attrs += " " + Name + " " + strValue + ";";
	}

void GFFPrependIntAttr(string &Attrs, const string &Name, unsigned Value)
	{
	char strValue[32];
	sprintf(strValue, "%u", Value);
	if (Attrs.empty())
		Attrs = Name + " " + strValue + ";";
	else
		Attrs = Name + " " + strValue + ";" + " " + Attrs;
	}

void GFFAppendStrAttr(string &Attrs, const string &Name, const string &Value)
	{
	if (!Attrs.empty())
		{
		if (Attrs[SIZE(Attrs) - 1] == ';')
			Attrs += " ";
		else
			Attrs += "; ";
		}
	Attrs += Name + " \"" + Value + "\";";
	}

void GFFPrependStrAttr(GFFRecord2 &Rec, const string &Name, const string &Value)
	{
	if (Rec.Attrs.empty())
		Rec.Attrs = Name + " \"" + Value + "\";";
	else
		Rec.Attrs = Name + " \"" + Value + "\";" + " " + Rec.Attrs;
	}

bool GFFDeleteAttr(string &Attrs, const string &Name)
	{
	vector<string> Names;
	vector<string> Values;
	GFFGetAttrs(Attrs, Names, Values);
	const unsigned N = SIZE(Names);
	asserta(SIZE(Values) == N);
	bool Found = false;
	Attrs = "";
	for (unsigned i = 0; i < N; ++i)
		{
		if (Names[i] == Name)
			Found = true;
		else
			{
			if (!Attrs.empty())
				Attrs += " ";
			Attrs += Names[i] + " " + Values[i] + ";";
			}
		}

	return Found;
	}

void GFFAppendTargetAttrs(string &Attrs, const string &SeqName, coord_t Lo,
  coord_t Hi)
	{
	GFFAppendStrAttr(Attrs, "target", SeqName);
	GFFAppendIntAttr(Attrs, "tstart", Lo + 1);
	GFFAppendIntAttr(Attrs, "tend", Hi + 1);
	}
