#include "getopts.h"
#include "utils.h"
#include "progress.h"
#include "gff.h"
#include "bp.h"
#include <list>
#include <vector>
#include <set>

#define TRACE	0

typedef std::pair<unsigned, unsigned> IntPair;
typedef std::set<std::pair<unsigned, unsigned> > IntPairSet;

#if	TRACE
static void LogBPs(const std::list<BendPoint> &BPs)
	{
	Log("\n");
	int Counts[2];
	Counts[0] = 0;
	Counts[1] = 0;
	Log("       Pos  lh   File    Rec  Count0  Count1\n");
	Log("==========  ==  =====  =====  ======  ======\n");
	for (std::list<BendPoint>::const_iterator p = BPs.begin();
	  p != BPs.end(); ++p)
		{
		const BendPoint &BP = *p;
		if (BP.FileIndex == 0)
			BP.lo ? ++(Counts[0]) : --(Counts[0]);
		else if (BP.FileIndex == 1)
			BP.lo ? ++(Counts[1]) : --(Counts[1]);
		Log("%10"COORD_T_PRINTF"  %s  %5u  %5u  %6d  %6d",
		  BP.Pos,
		  BP.lo ? "lo" : "HI",
		  BP.FileIndex,
		  BP.RecIndex,
		  Counts[0],
		  Counts[1]);
		if (BP.FileIndex > 1)
			Log(" **ERROR**");
		Log("\n");
		}
	}
#endif

static void ReadGFFRecs(const char *FileName, std::vector<GFFRecord> &Recs)
	{
	Recs.clear();
	FILE *f = OpenStdioFile(FileName);
	GFFRecord Rec;
	while (GetNextGFFRecord(f, Rec))
		{
		SaveGFFStrings(Rec);
		Recs.push_back(Rec);
		}
	CloseStdioFile(f);
	}

static void AppendBPs(const std::vector<GFFRecord> &Recs,
  std::list<BendPoint> &BPs, unsigned FileIndex)
	{
	unsigned RecCount = SIZE(Recs);
	FORIX(Rec)
		{
		const GFFRecord &Rec = Recs[RecIndex];
		BendPoint BP;
		BP.Pos = Rec.Start;
		BP.lo = true;
		BP.FileIndex = FileIndex;
		BP.RecIndex = RecIndex;
		BPs.push_back(BP);

		BP.Pos = Rec.End;
		BP.lo = false;
		BPs.push_back(BP);
		}
	}

// Start cover from one file, so other file goes from itself only to both.
// Should be no active recs from file we're starting to cover.
static void StartCover(FILE *fOut, std::vector<std::vector<GFFRecord> > &Recs,
  unsigned FileIndex, coord_t Pos, unsigned * /*Counts*/, const IntPairSet &ActiveRecs)
	{
	for (IntPairSet::const_iterator p = ActiveRecs.begin();
	  p != ActiveRecs.end(); ++p)
		{
		const IntPair &RecId = *p;
		unsigned ActiveFileIndex = RecId.first;
		unsigned ActiveRecIndex = RecId.second;
		asserta(ActiveFileIndex != FileIndex);

		GFFRecord &Rec = Recs[ActiveFileIndex][ActiveRecIndex];
		if (Rec.Start == Pos)
			continue;
		size_t n = strlen(Rec.Attrs);
		const char *OldAttrs = Rec.Attrs;
		char *NewAttrs = new char[n + 64];
		sprintf(NewAttrs, "%s ; only%u", OldAttrs, ActiveFileIndex + 1);
		Rec.Attrs = NewAttrs;
		coord_t OldEnd = Rec.End;
		Rec.End = Pos - 1;
		WriteGFFRec(fOut, Rec);
		Rec.End = OldEnd;

		Rec.Attrs = OldAttrs;
		Rec.Start = Pos;
		Rec.Attrs = OldAttrs;

		free((void *) NewAttrs);
		}
	}

// End cover from one file so other file goes from both to itself only,
// Should be no active recs from file for which we're ending coverage.
static void EndCover(FILE *fOut, std::vector<std::vector<GFFRecord> > &Recs,
  unsigned FileIndex, coord_t Pos, unsigned * /*Counts*/, const IntPairSet &ActiveRecs)
	{
	for (IntPairSet::const_iterator p = ActiveRecs.begin();
	  p != ActiveRecs.end(); ++p)
		{
		const IntPair &RecId = *p;
		unsigned ActiveFileIndex = RecId.first;
		unsigned ActiveRecIndex = RecId.second;
		asserta(ActiveFileIndex != FileIndex);

		GFFRecord &Rec = Recs[ActiveFileIndex][ActiveRecIndex];
		if (Rec.Start == Pos)
			continue;
		size_t n = strlen(Rec.Attrs);
		const char *OldAttrs = Rec.Attrs;
		char *NewAttrs = new char[n + 64];
		sprintf(NewAttrs, "%s ; both", OldAttrs);
		Rec.Attrs = NewAttrs;
		coord_t OldEnd = Rec.End;
		Rec.End = Pos;
		WriteGFFRec(fOut, Rec);

		Rec.End = OldEnd;
		Rec.Attrs = OldAttrs;
		Rec.Start = Pos + 1;

		free((void *) NewAttrs);
		}
	}

static void OutRec(FILE *fOut, GFFRecord &Rec, unsigned FileIndex,
  const unsigned *Counts)
	{
	if (Rec.Start > Rec.End)
		return;

	unsigned OtherFileIndex = (FileIndex == 0 ? 1 : 0);
	char Attr[16];
	if (Counts[OtherFileIndex] == 0)
		sprintf(Attr, "only%u", FileIndex + 1);
	else
		strcpy(Attr, "both");

	const char *OldAttrs = Rec.Attrs;
	size_t n = strlen(Rec.Attrs);
	char *NewAttrs = new char[n + 64];
	sprintf(NewAttrs, "%s ; %s", OldAttrs, Attr);
	Rec.Attrs = NewAttrs;
	WriteGFFRec(fOut, Rec);
	free(NewAttrs);
	Rec.Attrs = OldAttrs;
	FreeGFFStrings(Rec);
	}

void GFFCmp()
	{
	const char *InFileName1 = RequiredValueOpt("in1");
	const char *InFileName2 = RequiredValueOpt("in2");
	const char *OutFileName = RequiredValueOpt("out");

	FILE *fOut = CreateStdioFile(OutFileName);

	std::vector<std::vector<GFFRecord> > Recs(2);

	ReadGFFRecs(InFileName1, Recs[0]);
	ReadGFFRecs(InFileName2, Recs[1]);

	std::list<BendPoint> BPs;
	AppendBPs(Recs[0], BPs, 0);
	AppendBPs(Recs[1], BPs, 1);
	BPs.sort();

#if	TRACE
	LogBPs(BPs);
#endif

	IntPairSet ActiveRecs;
	unsigned Counts[2];
	Counts[0] = 0;
	Counts[1] = 0;
#if	TRACE
	Log("\n");
	Log("       Pos  lh   File    Rec  Count0  Count1  Active\n");
	Log("==========  ==  =====  =====  ======  ======  ======\n");
#endif
	for (std::list<BendPoint>::const_iterator p = BPs.begin();
	  p != BPs.end(); ++p)
		{
		const BendPoint &BP = *p;
#if	TRACE
		{
		Log("%10"COORD_T_PRINTF"  %s  %5u  %5u  %6d  %6d",
		  BP.Pos,
		  BP.lo ? "lo" : "HI",
		  BP.FileIndex,
		  BP.RecIndex,
		  Counts[0],
		  Counts[1]);
		for (IntPairSet::const_iterator q = ActiveRecs.begin();
		  q != ActiveRecs.end(); ++q)
			{
			const IntPair RecId = *q;
			Log("  (%u,%u)", RecId.first, RecId.second);
			}
		Log("\n");
		}
#endif
		unsigned FileIndex = BP.FileIndex;
		unsigned RecIndex = BP.RecIndex;
		std::pair<unsigned, unsigned> RecId(FileIndex, RecIndex);
		asserta(FileIndex < 2);
		unsigned &Count = Counts[FileIndex];
		if (BP.lo)
			{
			if (Count == 0)
				StartCover(fOut, Recs, FileIndex, BP.Pos, Counts, ActiveRecs);
			ActiveRecs.insert(RecId);
			++Count;
			}
		else
			{
			if (Count == 0)
				Die("Count=0");
			size_t n = ActiveRecs.erase(RecId);
			asserta(n == 1);
			--Count;
			if (Count == 0)
				EndCover(fOut, Recs, FileIndex, BP.Pos, Counts, ActiveRecs);
			OutRec(fOut, Recs[FileIndex][RecIndex], FileIndex, Counts);
			}
		}
	}
