#include "utils.h"
#include "revfile.h"
#include "progress.h"
#include "getopts.h"
#include "uniquebid.h"

void GetBendpoints(const AIVec &AIs, unsigned GenomeIndex, AIBPVec &BPs,
  unsigned Margin);
void Output(const RevFile &In, RevFile &Out);
void TruncLabelGenomes(std::vector<GenInfo> &Genomes);
void GetBundleLoHi(const UniqueBundleId &UB,
  const std::map<UniqueBundleId, coord_t> &los,
  const std::map<UniqueBundleId, coord_t> &his,
  coord_t &lo, coord_t &hi);

extern AIVec g_AIs;
extern BoolVec g_Del;	// true if aln is to be deleted
extern std::map<UniqueBundleId, score_t> g_BundleScores;
extern RevFile *g_RevDel;

static unsigned g_DeletedHitCount;
static unsigned g_MaxCount = 4;

static IntList g_HCC_AIIndexes;
static coord_t g_HCC_StartPos;
static std::map<UniqueBundleId, coord_t> g_BundleLos;
static std::map<UniqueBundleId, coord_t> g_BundleHis;
static unsigned g_RegionCount;
static unsigned g_NoCoveringBundleCount;
static unsigned g_GenomeIndex;

static std::set<UniqueBundleId> g_UBs;

static void GetUB(const AlnInfo &AI, UniqueBundleId &UB)
	{
	UB.BundleId = AI.GetBundleId();
	if (UB.BundleId == UINT_MAX)
		Die("AI has no bundle id");

	const GCSeg &Seg0 = AI.GetSeg(0);
	const GCSeg &Seg1 = AI.GetSeg(1);
	UB.ChrId0 = Seg0.ChrId;
	UB.ChrId1 = Seg1.ChrId;
	}

static void AddToHighCopyCountRegion(unsigned AIIndex)
	{
	assert(AIIndex < (unsigned) g_AIs.size());

	g_HCC_AIIndexes.push_back(AIIndex);

	const AlnInfo &AI = g_AIs[AIIndex];

	UniqueBundleId UB;
	GetUB(AI, UB);
	g_UBs.insert(UB);
	}

static void StartHighCopyCountRegion(coord_t Pos, const IntList &Window)
	{
	g_HCC_AIIndexes.clear();
	g_UBs.clear();

	g_HCC_StartPos = Pos;
	++g_RegionCount;

	for_CIntList(Window, p)
		{
		unsigned AIIndex = *p;
		AddToHighCopyCountRegion(AIIndex);
		}
	}

static void EndHighCopyCountRegion(coord_t EndPos,
  const std::map<UniqueBundleId, coord_t> &/* los */,
  const std::map<UniqueBundleId, coord_t> &/* hi */)
	{
// Find highest-scoring bundle that covers region
	UniqueBundleId HighestScoringUB;
	HighestScoringUB.BundleId = UINT_MAX;
	HighestScoringUB.ChrId0 = UINT_MAX;
	HighestScoringUB.ChrId1 = UINT_MAX;
	score_t HighestScore = MINUS_INF;
	for (std::set<UniqueBundleId>::const_iterator p = g_UBs.begin();
	  p != g_UBs.end(); ++p)
		{
		const UniqueBundleId &UB = *p;

		coord_t lo;
		coord_t hi;
		GetBundleLoHi(UB, g_BundleLos, g_BundleHis, lo, hi);

		if (lo <= g_HCC_StartPos && hi >= EndPos)
			{
			score_t BundleScore = GetBundleScore(UB);
			if (BundleScore > HighestScore)
				{
				HighestScore = BundleScore;
				HighestScoringUB = UB;
				}
			}
		}

// Found any such bundle?
	if (HighestScoringUB.BundleId == UINT_MAX)
		{
		++g_NoCoveringBundleCount;
		return;
		}

	for_CIntList(g_HCC_AIIndexes, p)
		{
		unsigned AIIndex = *p;
		assert(AIIndex < g_Del.size());
		const AlnInfo &AI = g_AIs[AIIndex];
		const GCSeg &Seg = AI.GetSeg(g_GenomeIndex);

	// Keep hit unless it is 100% within HCC region
		if (Seg.lo < g_HCC_StartPos || Seg.hi > EndPos)
			continue;

		UniqueBundleId UB;
		GetUB(AI, UB);

	// Delete hit unless it is in highest scoring bundle
		if (!(UB == HighestScoringUB))
			{
			if (!g_Del[AIIndex])
				{
				++g_DeletedHitCount;
				g_Del[AIIndex] = true;
				}
			}
		}
	}

// This is O(N), could be O(log N) by using set instead of list.
static void DeleteFromWindow(IntList &Window, const AIBP &BP)
	{
	for_IntList(Window, p)
		{
		unsigned AIIndex = *p;
		if (AIIndex == BP.AlnInfoIndex)
			{
			Window.erase(p);
			return;
			}
		}
	Die("Hit not found in window");
	}

// TODO: Similar to several other subroutines, including ReduceOverlaps(),
// should consolidate.
static void HighCopyCountFilterGenome(unsigned GenomeIndex)
	{
	g_GenomeIndex = GenomeIndex;

	ComputeBundleLoHis(g_AIs, GenomeIndex, g_BundleLos, g_BundleHis);

// Get hit bendpoints.
	std::vector<AIBP> BPs;
	GetBendpoints(g_AIs, GenomeIndex, BPs, 0);

// Sliding window to compute copy counts.
// A hit enters a list when its lo point enters the window,
// and exits the list when its hi point exits the window.
	IntList Window;
	const unsigned BPCount = (unsigned) BPs.size();
	const unsigned AICount = (unsigned) g_AIs.size();
	unsigned PrevChrId = UINT_MAX;
	unsigned CopyCount = 0;
	ProgressStart("Filter genome %u", GenomeIndex);
	for (unsigned BPIndex = 0; BPIndex < BPCount; ++BPIndex)
		{
		if (BPIndex%10000 == 0)
			ProgressStep(BPIndex, BPCount);
		const AIBP &BP = BPs[BPIndex];
		unsigned AIIndex = BP.AlnInfoIndex;
		assert(AIIndex < AICount);

#if	TRACE
		{
		const GCSeg &Seg = g_AIs[AIIndex].GetSeg(GenomeIndex);
		unsigned lo = (unsigned) Seg.lo;
		unsigned hi = (unsigned) Seg.hi;
		Log("BP: Pos=%u %s AI=%u lo=%u hi=%u del=%c\n",
		  BP.Pos,
		  BP.IsLo ? "lo" : "hi",
		  BP.AlnInfoIndex,
		  lo,
		  hi,
		  g_Del[AIIndex] ? 'Y' : 'N');
		Log("Window: ");
		LogList(Window);
		Log("\n");
		if (g_MaxBPsTraced > 0 && ++g_TraceCounter >= g_MaxBPsTraced)
			Die("Traced %u BPs", g_TraceCounter);
		}
#endif
		if (BP.ChrId != PrevChrId)
			{
			PrevChrId = BP.ChrId;
			asserta(CopyCount == 0);
			Window.clear();
			}

		if (BP.IsLo)
			{
		// lo: add to window
			Window.push_back(AIIndex);
			++CopyCount;
			if (CopyCount == g_MaxCount)
				StartHighCopyCountRegion(BP.Pos, Window);
			else if (CopyCount > g_MaxCount)
				AddToHighCopyCountRegion(BP.AlnInfoIndex);
			}
		else
			{
			assert(!BP.IsLo);
			asserta(CopyCount > 0);

			DeleteFromWindow(Window, BP);
			--CopyCount;
			if (CopyCount == g_MaxCount)
				EndHighCopyCountRegion(BP.Pos, g_BundleLos, g_BundleHis);
			}
		}
	ProgressDone();
	}

void HighCopyCountFilter()
	{
	const char *InFileName = RequiredValueOpt("hcc");
	const char *OutFileName = RequiredValueOpt("out");
	const char *DelFileName = ValueOpt("del");
	IntOpt("maxcount", &g_MaxCount);

	RevFile In;
	RevFile Out;

	In.Open(InFileName);
	if (In.GetGenomeCount() != 2)
		Die("glover requires pairwise alignment, genome count > 2");

	if (DelFileName != 0)
		{
		g_RevDel = new RevFile;
		const std::vector<GenInfo> &Genomes = In.GetGenomes();
		g_RevDel->Create(DelFileName, Genomes);
		}

	std::vector<GenInfo> Genomes = In.GetGenomes();

	TruncLabelGenomes(Genomes);

	Out.Create(OutFileName, Genomes);

	Progress("Get AIs");
	In.GetAllAlnInfos(g_AIs);
	unsigned HitCount = (unsigned) g_AIs.size();
	Progress("%u AIs", HitCount);

	const unsigned AICount = (unsigned) g_AIs.size();
	g_Del.resize(AICount);
	for (unsigned i = 0; i < AICount; ++i)
		g_Del[i] = false;

	Progress("Compute bundle scores");
	ComputeBundleScores(g_AIs, g_BundleScores);
	HighCopyCountFilterGenome(0);
	HighCopyCountFilterGenome(1);

	if (HitCount > 0)
		{
		double Pct = (double) g_DeletedHitCount * 100.0 / (double) HitCount;
		Progress("%u HCC regions, %u not covered",
		  g_RegionCount, g_NoCoveringBundleCount);
		Progress("%u of %u hits deleted, %.1f%%", g_DeletedHitCount, HitCount, Pct);
		}

	Output(In, Out);

	Progress("Closing files and writing output");
	In.Close();
	Out.Close();

	if (g_RevDel != 0)
		{
		g_RevDel->Close();
		delete g_RevDel;
		g_RevDel = 0;
		}
	}

