#include "ice.h"
#include "myrand.h"
#include "evomodel.h"
#include "progress.h"

/***
Inter-chromosome evolver.

A chromosome is represented as a sequence of 'contigs'.

A contig is either a conserved element (CE) or a neutral
region.

Each contig has properties including its ancestral and
evolved coordinates (chr, position), whether it is the
beginning or end of a gene, and the number of times it
has been rearranged (moved or copied).

In InterChrEvolver, ranges are specified by evolved
coordinates, in contrast to [intra-chr]Evolver where they
are specified by ancestral coordinates.

Contigs are maintained sorted by (evolved chr, evolved pos),
this is implemented using std::set. Specifically, a
chromosome is represented as a std::set<ContigData>,
where a "contig" is a maximal set of bases of the same
type (Neutral, CDS, UTR, NXE or NGE) that are contiguous
in both the evolved and ancestral sequence.

Partial CEs and partial genes may not be moved.

If a complete gene is copied, gene duplication is
triggered using the same outcome weight table as in the
intra-chr evolver.

Ancestral annotations are saved. They are not used during
evolution but are needed to generate evolved annotations,
(e.g the probs attribute).

Events are implemented using naive routines that traverse
chromosomes from start to finish.

Much more efficient algorithms are possible, but we expect
a relatively small number of inter-chr events so we assume
efficiency is irrelevant here and write code for ease of
implementation and correctness.
[NOTE added later -- this was a bad assumption due to not
thinking through the biology very clearly. Ideally this code
would be re-written with better efficiency in mind.]
***/

InterChrEvolver::InterChrEvolver()
	{
	Clear();
	}

void InterChrEvolver::Clear()
	{
	m_LogEvents = false;
	m_LastAcceptedEventIndex = UINT_MAX;
	m_LastAcceptedEvent = EVENT_TYPE_Invalid;
	m_CurrentEvent = EVENT_TYPE_Invalid;
	m_LastAcceptedEventPos = GPOS_T_MAX;
	m_NextNewGeneIndex = 0;

	m_EventIndex = 0;
	m_LastAcceptedEventIndex = 0;

	m_IntegratedEventRates.clear();
	m_ChrLengths.clear();
	m_AncChrLengths.clear();
	m_ChrNames.clear();
	m_AncChrNames.clear();
	m_Contigs.clear();
	}

void InterChrEvolver::InitChrs(const vector<string> &ChrNames,
  const vector<gpos_t> &ChrLengths)
	{
	asserta(SIZE(ChrNames) == SIZE(ChrLengths));

	Clear();

	m_ChrNames = ChrNames;
	m_ChrLengths = ChrLengths;
	m_AncChrNames = ChrNames;
	m_AncChrLengths = ChrLengths;
	}

void InterChrEvolver::GetRandomPos(unsigned &ChrIndex, gpos_t &Pos) const
	{
// Pick a chromosome, weighted by length
	vector<double> Weights;
	const unsigned ChrCount = GetChrCount();
	for (unsigned ChrIndex2 = 0; ChrIndex2 < ChrCount; ++ChrIndex2)
		{
		gpos_t ChrLength = GetChrLength(ChrIndex2);
		Weights.push_back(ChrLength);
		}
	ChrIndex = RandWeightDist(Weights);

// Pick uniformly distributed random position within chromosome
	asserta(sizeof(gpos_t) == sizeof(unsigned));
	gpos_t Length = GetChrLength(ChrIndex);
	Pos = RandInt(Length);
	}

void InterChrEvolver::GetRandomPosDifferentChr(unsigned ChrIndex,
  unsigned &DiffChrIndex, gpos_t &Pos) const
	{
	const unsigned ChrCount = GetChrCount();
	asserta(ChrCount > 1);
	vector<double> Weights;
	for (unsigned ChrIndex2 = 0; ChrIndex2 < ChrCount; ++ChrIndex2)
		{
		if (ChrIndex2 == ChrIndex)
			continue;
		gpos_t ChrLength = GetChrLength(ChrIndex2);
		Weights.push_back(ChrLength);
		}
	DiffChrIndex = RandWeightDist(Weights);
	if (DiffChrIndex >= ChrIndex)
		++DiffChrIndex;
	asserta(sizeof(gpos_t) == sizeof(unsigned));
	Pos = RandInt(GetChrLength(DiffChrIndex));
	}

bool InterChrEvolver::AcceptInsert(unsigned ChrIndex, gpos_t Pos,
  gpos_t /*Length*/) const
	{
	ContigSet Contigs1;
	ContigSet Contigs2;
	GetContigs(ChrIndex, Pos, 1, Contigs1);
	GetContigs(ChrIndex, Pos + 1, 1, Contigs2);
	asserta(Contigs1.size() == 1 && Contigs2.size() == 1);
	const ContigData &Contig1 = *(Contigs1.begin());
	const ContigData &Contig2 = *(Contigs2.begin());
	if (Contig1.Type != CT_Neutral || Contig2.Type != CT_Neutral)
		return false;
	return true;
	}

gpos_t InterChrEvolver::GetLength() const
	{
	gpos_t Length = 0;
	const unsigned ChrCount = SIZE(m_ChrLengths);
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		Length += GetChrLength(ChrIndex);
	return Length;
	}

void InterChrEvolver::GetChrParts(unsigned ChrIndex, gpos_t Pos, gpos_t &Length1,
  gpos_t &Length2) const
	{
	gpos_t ChrLength = GetChrLength(ChrIndex);
	asserta(Pos < ChrLength);
	Length1 = Pos;
	Length2 = ChrLength - Pos;
	}

gpos_t InterChrEvolver::GetChrLength(unsigned ChrIndex) const
	{
	asserta(ChrIndex < SIZE(m_ChrLengths));
	return m_ChrLengths[ChrIndex];
	}

bool InterChrEvolver::RangeExists(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const
	{
	gpos_t ChrLength = GetChrLength(ChrIndex);
	return Pos + Length < ChrLength;
	}

void InterChrEvolver::GetContigs(unsigned ChrIndex, gpos_t Pos, gpos_t Length,
  ContigSet &Contigs) const
	{
	GetOverlappingContigs(ChrIndex, Pos, Length, Contigs);
	if (Contigs.empty())
		return;

	ContigData FirstContig = *(Contigs.begin());
	asserta(FirstContig.EvPos <= Pos);
	size_t n = Contigs.erase(FirstContig);
	asserta(n == 1);

	gpos_t Diff = Pos - FirstContig.EvPos;
	FirstContig.EvPos += Diff;
	if (FirstContig.Plus)
		FirstContig.AncPos += Diff;

	asserta(Diff < FirstContig.Length);
	FirstContig.Length -= Diff;

	Contigs.insert(FirstContig);

	ContigSetCIter p = Contigs.end();
	--p;
	ContigData LastContig = *p;
	gpos_t Hi = Pos + Length - 1;
	gpos_t ContigHi = LastContig.GetEvHi();
	if (ContigHi > Hi)
		{
		size_t n = Contigs.erase(LastContig);
		asserta(n == 1);
		Diff = ContigHi - Hi;
		asserta(Diff < LastContig.Length);
		LastContig.Length -= Diff;
		if (!LastContig.Plus)
			{
			asserta(LastContig.AncPos >= Diff);
			LastContig.AncPos -= Diff;
			}

		Contigs.insert(LastContig);
		}

#if	0 // DEBUG
	AssertContiguous(Contigs);
#endif
	}

void InterChrEvolver::GetOverlappingContigs(unsigned ChrIndex, gpos_t Pos,
  gpos_t Length, ContigSet &Contigs) const
	{
	Contigs.clear();
	gpos_t EvHi = Pos + Length - 1;

	ContigSetCIter p = GetContigCIter(ChrIndex, Pos);
	for (;;)
		{
		ContigData Contig = *p;
		if (Contig.EvChrIndex != ChrIndex || Contig.EvPos > EvHi)
			break;
		Contigs.insert(Contig);
		++p;
		if (p == m_Contigs.end())
			break;
		}
#if	0 // DEBUG
	AssertContiguous(Contigs);
#endif
	}

void InterChrEvolver::GetContigsPosGE(unsigned ChrIndex, gpos_t Pos,
  ContigSet &Contigs) const
	{
	gpos_t ChrLength = GetChrLength(ChrIndex);
	ContigSetCIter p = GetContigCIter(ChrIndex, Pos);
	const ContigData &Contig = *p;
	asserta(Contig.EvPos == Pos);
	gpos_t Length = ChrLength - Pos;
	GetOverlappingContigs(ChrIndex, Pos, Length, Contigs);

	ContigData Range;
	GetContigsRange(Contigs, Range);
	asserta(Range.EvPos == Pos);
	asserta(Range.GetEvHi() + 1 == ChrLength);
	}

void InterChrEvolver::AppendContigs(unsigned ChrIndex, const ContigSet &Contigs)
	{
	ContigSet NewContigs;
	gpos_t Pos = GetChrLength(ChrIndex);
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		Contig.EvChrIndex = ChrIndex;
		Contig.EvPos = Pos;
		Pos += Contig.Length;
		NewContigs.insert(Contig);
		}
	m_Contigs.insert(NewContigs.begin(), NewContigs.end());
	}

void InterChrEvolver::InsertContigs(unsigned ChrIndex, gpos_t Pos,
  const ContigSet &Contigs)
	{
#if	0 // DEBUG
	AssertContiguous(Contigs);
#endif

	Split(ChrIndex, Pos + 1);

	ContigData Range;
	GetContigsRange(Contigs, Range);

	ContigSetCIter p = m_Contigs.find(ContigData(ChrIndex, Pos + 1));
	asserta(p != m_Contigs.end());

	ContigSet FollowingContigs;
	GetContigsPosGE(ChrIndex, Pos + 1, FollowingContigs);

	ContigSet NewContigs1;
	for (ContigSetCIter p = FollowingContigs.begin();
	  p != FollowingContigs.end(); ++p)
		{
		ContigData Contig = *p;
		size_t n = m_Contigs.erase(Contig);
		asserta(n == 1);

		Contig.EvPos += Range.Length;
		NewContigs1.insert(Contig);
		}
#if	0 // DEBUG
	AssertContiguous(NewContigs1);
#endif
	ContigSet NewContigs2;
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		Contig.EvChrIndex = ChrIndex;
		asserta(Contig.EvPos >= Range.EvPos);
		Contig.EvPos = Pos + (Contig.EvPos - Range.EvPos) + 1;
		NewContigs2.insert(Contig);
		}
#if	0 // DEBUG
	AssertContiguous(NewContigs2);
#endif

	m_Contigs.insert(NewContigs1.begin(), NewContigs1.end());
	m_Contigs.insert(NewContigs2.begin(), NewContigs2.end());

#if	0 // DEBUG
	AssertContiguous(m_Contigs);
#endif

	asserta(ChrIndex < SIZE(m_ChrLengths));
	m_ChrLengths[ChrIndex] += Range.Length;
	}

void InterChrEvolver::InvertContigs(ContigSet &Contigs)
	{
#if	0 // DEBUG
	AssertContiguous(Contigs);
#endif
	ContigData Range;
	GetContigsRange(Contigs, Range);

	gpos_t StartPos = Range.EvPos;
	gpos_t EndPos = Range.GetEvHi();
	ContigSet NewContigs;
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		asserta(Contig.EvPos >= StartPos);
		gpos_t RelPos = Contig.EvPos - StartPos;
		if (RelPos + Contig.Length >= EndPos)
			{
			Log("\n");
			Log("InvertContigs\n");
			LogContigs(Contigs);
			Contig.LogMe(true);
			Log("RelPos %"PF" + Contig.Length %"PF" = %"PF" >= EndPos %"PF,
			  RelPos, Contig.Length, RelPos + Contig.Length, EndPos);
			}
		gpos_t NewEvPos = EndPos - (RelPos + Contig.Length - 1);
		if (NewEvPos < StartPos)
			{
			Log("\n");
			Log("InvertContigs\n");
			LogContigs(Contigs);
			Contig.LogMe(true);
			Die("NewEvPos %"PF" < StartPos %"PF, NewEvPos, StartPos);
			}
		Contig.EvPos = NewEvPos;
		Contig.Plus = !Contig.Plus;
		NewContigs.insert(Contig);
		}
	Contigs = NewContigs;
#if	0 // DEBUG
	AssertContiguous(Contigs);
#endif
	}

void InterChrEvolver::InitNeutrals()
	{
	set<unsigned> NonEmptyChrIndexes;

	const ContigData *LastContig = 0;
	list<ContigData> NeutralContigs;
	unsigned LastGeneIndex = UINT_MAX;
	const unsigned ContigCount = SIZE(m_Contigs);
	unsigned Counter = 0;
	ProgressStart("Init neutrals");
	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		if (Counter++%1000 == 0)
			ProgressStep(Counter, ContigCount);
		const ContigData *Contig = &(*p);

		unsigned ThisGeneIndex = Contig->GeneIndex;
		unsigned GeneIndex = LastGeneIndex;
		if (ThisGeneIndex != LastGeneIndex)
			GeneIndex = UINT_MAX;

		NonEmptyChrIndexes.insert(Contig->EvChrIndex);
		if (Contig->EvPos == 0)
			{
			LastContig = Contig;
			LastGeneIndex = ThisGeneIndex;
			continue;
			}

		gpos_t StartPos = 0;
		if (LastContig != 0 && LastContig->EvChrIndex == Contig->EvChrIndex)
			StartPos = LastContig->GetEvHi() + 1;
		gpos_t EndPos = Contig->EvPos - 1;
		if (StartPos > EndPos)
			{
			LastContig = Contig;
			LastGeneIndex = ThisGeneIndex;
			continue;
			}

		ContigData NeutralContig;
		NeutralContig.Type = CT_Neutral;
		NeutralContig.EvChrIndex = Contig->EvChrIndex;
		NeutralContig.AncChrIndex = Contig->AncChrIndex;
		NeutralContig.EvPos = StartPos;
		NeutralContig.AncPos = StartPos;
		NeutralContig.Length = EndPos - StartPos + 1;
		NeutralContig.RelocCount = 0;
		NeutralContig.Plus = true;
		NeutralContig.Copy = false;
		NeutralContig.GeneIndex = GeneIndex;
		if (NeutralContig.Length > 0)
			NeutralContigs.push_back(NeutralContig);

		LastContig = Contig;
		LastGeneIndex = ThisGeneIndex;
		}
	ProgressDone();

	const unsigned ChrCount = GetChrCount();
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		gpos_t ChrLength = GetChrLength(ChrIndex);
		asserta(ChrLength > 0);
		gpos_t StartPos = 0;
		if (NonEmptyChrIndexes.find(ChrIndex) != NonEmptyChrIndexes.end())
			{
			const ContigData &LastContig = GetLastContig(ChrIndex);
			StartPos = LastContig.GetEvHi() + 1;
			}
		if (StartPos == ChrLength)
			continue;
		asserta(StartPos < ChrLength);
		gpos_t EndPos = ChrLength - 1;

		ContigData NeutralContig;
		NeutralContig.Type = CT_Neutral;
		NeutralContig.EvPos = StartPos;
		NeutralContig.AncPos = StartPos;
		NeutralContig.EvChrIndex = ChrIndex;
		NeutralContig.AncChrIndex = ChrIndex;
		NeutralContig.GeneIndex = UINT_MAX;
		NeutralContig.Length = EndPos - StartPos + 1;
		if (NeutralContig.Length == 0)
			continue;

		NeutralContigs.push_back(NeutralContig);
		}

	Progress("Inserting neutrals");
	m_Contigs.insert(NeutralContigs.begin(), NeutralContigs.end());
	}

const ContigData &InterChrEvolver::GetLastContig(unsigned ChrIndex) const
	{
	asserta(ChrIndex < GetChrCount());
	ContigSetCIter p = m_Contigs.upper_bound(ContigData(ChrIndex, GPOS_T_MAX));
	asserta(p != m_Contigs.begin());
	--p;
	const ContigData &Contig = *p;
	if (Contig.EvChrIndex != ChrIndex)
		{
		LogMe();
		Log("\n");
		Contig.LogMe();
		Die("GetLastContig, Contig.EvChrIndex != ChrIndex=%u", ChrIndex);
		}
	return Contig;
	}

void InterChrEvolver::GetContigsRange(const ContigSet &Contigs,
  ContigData &Range) const
	{
	asserta(!Contigs.empty());
	const ContigData &FirstContig = *(Contigs.begin());
	ContigSetCIter p = Contigs.end();
	--p;
	const ContigData &LastContig = *p;
	asserta(FirstContig.EvChrIndex == LastContig.EvChrIndex);
	Range.Type = CT_Range;
	Range.EvChrIndex = FirstContig.EvChrIndex;
	Range.EvPos = FirstContig.EvPos;
	Range.Length = LastContig.GetEvHi() - FirstContig.EvPos + 1;
	Range.AncPos = GPOS_T_MAX;
	Range.AncChrIndex = UINT_MAX;
	}

void InterChrEvolver::TruncateChr(unsigned ChrIndex, gpos_t Pos, ContigSet &DeletedContigs)
	{
	Split(ChrIndex, Pos);
	gpos_t Length = GetChrLength(ChrIndex) - Pos - 1;
	DeleteContigs(ChrIndex, Pos, Length, DeletedContigs);
	}

void InterChrEvolver::DeleteContigs(unsigned ChrIndex, gpos_t Pos, gpos_t Length,
  ContigSet &DeletedContigs)
	{
	gpos_t EndPos = Pos + Length - 1;

	Split(ChrIndex, Pos);
	Split(ChrIndex, EndPos + 1);

	GetContigs(ChrIndex, Pos, Length, DeletedContigs);

	ContigData RangeContig;
	GetContigsRange(DeletedContigs, RangeContig);
	asserta(RangeContig.EvPos == Pos);
	asserta(RangeContig.Length == Length);

	for (ContigSetCIter p = DeletedContigs.begin(); p != DeletedContigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		size_t n = m_Contigs.erase(Contig);
		asserta(n == 1);
		}

	ContigSet FollowingContigs;
	GetContigsPosGE(ChrIndex, EndPos + 1, FollowingContigs);

	ContigSet NewContigs;
	for (ContigSetCIter p = FollowingContigs.begin();
	  p != FollowingContigs.end(); ++p)
		{
		ContigData Contig = *p;
		size_t n = m_Contigs.erase(Contig);
		asserta(n == 1);
		Contig.EvPos -= Length;
		NewContigs.insert(Contig);
		}

	m_Contigs.insert(NewContigs.begin(), NewContigs.end());

	asserta(ChrIndex < SIZE(m_ChrLengths));
	m_ChrLengths[ChrIndex] -= Length;
	}

// Ensure that there is a contig with Contig.EvPos=Pos.
// Usually this means splitting an existing contig.
void InterChrEvolver::Split(unsigned ChrIndex, gpos_t Pos)
	{
	ContigData Contig;
	GetContig(ChrIndex, Pos, Contig);

// Is Pos already the first pos of containing contig?
	if (Contig.EvPos == Pos)
		return;

// Should never split a constrained contig.
	if (Contig.Type != CT_Neutral)
		{
		// LogMe();
		Log("Split Chr %u Pos %"PF"\n", ChrIndex, Pos);
		Log("Contig: ");
		Contig.LogMe(true);
		Die("Split in CE");
		}

	size_t n = m_Contigs.erase(Contig);
	asserta(n == 1);

	gpos_t Hi = Contig.GetEvHi();
	asserta(Pos >= Contig.EvPos);
	gpos_t Diff = Pos - Contig.EvPos;
	gpos_t OldLength = Contig.Length;
	gpos_t OldAncPos = Contig.AncPos;
	Contig.Length = Diff;
	if (!Contig.Plus)
		Contig.AncPos += (OldLength - Diff);
	m_Contigs.insert(Contig);

	Contig.EvPos = Pos;
	if (Contig.Plus)
		Contig.AncPos += Diff;
	else
		Contig.AncPos = OldAncPos;
	Contig.Length = Hi - Pos + 1;
	m_Contigs.insert(Contig);
#if	0 // DEBUG
	AssertContiguous(m_Contigs);
#endif
	}

void InterChrEvolver::GetContig(unsigned ChrIndex, gpos_t Pos,
  ContigData &Contig) const
	{
	ContigSetCIter p = GetContigCIter(ChrIndex, Pos);
	Contig = *p;
	asserta(Pos >= Contig.EvPos && Pos <= Contig.GetEvHi());
	}

ContigSetCIter InterChrEvolver::GetContigCIter(unsigned ChrIndex,
  gpos_t Pos) const
	{
	ContigSetCIter p = m_Contigs.upper_bound(ContigData(ChrIndex, Pos));
	if (p != m_Contigs.begin())
		--p;

	const ContigData &Contig = *p;
	if (Pos < Contig.EvPos || Pos > Contig.GetEvHi())
		{
		Log("\n");
		Log("\n");
		ContigSetCIter q = m_Contigs.upper_bound(ContigData(ChrIndex, Pos));
		if (q == m_Contigs.begin())
			Log("upper_bound=begin()\n");
		Log("upper_bound   = ");
		(*q).LogMe(true);
		Log("upper_bound-- = ");
		Contig.LogMe(true);
		Die("InterChrEvolver::GetContigCIter(%u, %"PF") failed", ChrIndex, Pos);
		}
	return p;
	}

void InterChrEvolver::Deconstrain(ContigSet &Contigs, bool DeGene) const
	{
	ContigSet NewContigs;
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		DeconstrainContig(Contig, DeGene);
		NewContigs.insert(Contig);
		}
	Contigs = NewContigs;
	}

void InterChrEvolver::DeconstrainContig(ContigData &Contig,
  bool DeGene) const
	{
	Contig.Type = CT_Neutral;
	if (DeGene)
		Contig.GeneIndex = UINT_MAX;
	}

void InterChrEvolver::GetCompleteGeneRanges(unsigned ChrIndex, gpos_t Pos,
  gpos_t Length, map<unsigned, pair<gpos_t, gpos_t> > &LoHis) const
	{
	LoHis.clear();

	bool Strand_NotUsed;
	gpos_t EndPos = GetRangeEndPos(ChrIndex, Pos, Length);
	unsigned GeneIndexStart = GetGeneIndex(ChrIndex, Pos, Strand_NotUsed);
	unsigned GeneIndexEnd = GetGeneIndex(ChrIndex, EndPos, Strand_NotUsed);

// Check an easy common special case as an optimization
	if (GeneIndexStart != UINT_MAX && GeneIndexStart == GeneIndexEnd)
		return;

	ContigSet Contigs;
	GetContigs(ChrIndex, Pos, Length, Contigs);

	unsigned PrevGeneIndex = UINT_MAX;
	gpos_t CurrGeneStart = GPOS_T_MAX;
	for (ContigSet::const_iterator p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		bool FirstContig = (p == Contigs.begin());

		ContigSet::const_iterator q = p;
		++q;
		bool LastContig = (q == Contigs.end());

		const ContigData &Contig = *p;
		unsigned GeneIndex = Contig.GeneIndex;

		bool GeneEnd = (!LastContig && PrevGeneIndex != UINT_MAX &&
		  PrevGeneIndex != GeneIndex);

		bool GeneStart = (!FirstContig && GeneIndex != UINT_MAX &&
		  GeneIndex != PrevGeneIndex);

		if (GeneEnd)
			{
			if (CurrGeneStart != GPOS_T_MAX)
				{
			// End of /previous/ contig
				gpos_t GeneEnd = Contig.EvPos - 1;
				LoHis[PrevGeneIndex] = pair<gpos_t, gpos_t>(CurrGeneStart, GeneEnd);
				}
			CurrGeneStart = GPOS_T_MAX;
			}

	// Must do this /after/ check for GeneEnd to avoid
	// overwriting CurrGeneStart before use.
		if (GeneStart)
			CurrGeneStart = Contig.EvPos;

		PrevGeneIndex = GeneIndex;
		}
	}

unsigned InterChrEvolver::GetGeneIndex(unsigned ChrIndex, gpos_t Pos,
  bool &Strand) const
	{
	ContigData Contig;
	GetContig(ChrIndex, Pos, Contig);
	Strand = Contig.Plus;
	return Contig.GeneIndex;
	}

bool InterChrEvolver::PosInGene(unsigned ChrIndex, gpos_t Pos) const
	{
	bool Strand_NotUsed;
	return GetGeneIndex(ChrIndex, Pos, Strand_NotUsed) != UINT_MAX;
	}

bool InterChrEvolver::RangeHasCE(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const
	{
	ContigSet Contigs;
	GetContigs(ChrIndex, Pos, Length, Contigs);
	for (ContigSet::const_iterator p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.Type != CT_Neutral)
			return true;
		}
	return false;
	}

bool InterChrEvolver::PosIsConstrained(unsigned ChrIndex, gpos_t Pos) const
	{
	ContigSet Contigs;
	GetContigs(ChrIndex, Pos, 1, Contigs);
	asserta(SIZE(Contigs) == 1);
	const ContigData &Contig = *Contigs.begin();
	return !(Contig.Type == CT_Neutral);
	}

// Compress by mergeing contiguous contigs, i.e. adjacent
// contigs that are contiguous in both the ancestral and
// evolved genomes.
void InterChrEvolver::MergeContiguousNeutrals()
	{
	gpos_t OldLength = GetLength();
	ContigSet NewContigs;
	ContigData CurrContig;
	CurrContig.EvChrIndex = UINT_MAX;
	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (CurrContig.EvChrIndex == UINT_MAX)
			{
			CurrContig = Contig;
			continue;
			}

		if (AreContiguousNeutrals(CurrContig, Contig))
			MergeContiguousNeutrals(CurrContig, Contig);
		else
			{
			NewContigs.insert(CurrContig);
			CurrContig = Contig;
			}
		}
	if (CurrContig.EvChrIndex != UINT_MAX)
		NewContigs.insert(CurrContig);

	m_Contigs = NewContigs;
	gpos_t NewLength = GetLength();
	asserta(NewLength == OldLength);
	Validate("End of MergeContiguousNeutrals");
	}

// Merge contiguous contigs, allowing merge of neutrals and CEs.
void InterChrEvolver::GetMaximalContigs(ContigSet &Contigs) const
	{
	Contigs.clear();

	ContigData CurrContig;
	CurrContig.EvChrIndex = UINT_MAX;
	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (CurrContig.EvChrIndex == UINT_MAX)
			{
			CurrContig = Contig;
			continue;
			}

		if (AreContiguous(CurrContig, Contig))
			MergeContiguous(CurrContig, Contig);
		else
			{
			Contigs.insert(CurrContig);
			CurrContig = Contig;
			}
		}
	if (CurrContig.EvChrIndex != UINT_MAX)
		Contigs.insert(CurrContig);

	AssertContiguous(Contigs);
	}

// True iff contigs are neutrals and are contiguous in BOTH
// ancestral and evolved genomes.
bool InterChrEvolver::AreContiguousNeutrals(const ContigData &Contig1,
  const ContigData &Contig2) const
	{
	if (Contig1.Type != CT_Neutral)
		return false;
	if (Contig2.Type != CT_Neutral)
		return false;
	if (Contig1.EvChrIndex != Contig2.EvChrIndex)
		return false;
	if (Contig1.AncChrIndex != Contig2.AncChrIndex)
		return false;
	if (Contig1.Plus != Contig2.Plus)
		return false;
	if (Contig1.Copy != Contig2.Copy)
		return false;
	if (Contig1.Plus)
		{
		if (Contig1.AncPos + Contig1.Length != Contig2.AncPos)
			return false;
		if (Contig1.EvPos + Contig1.Length != Contig2.EvPos)
			return false;
		}
	else
		{
		if (Contig2.AncPos + Contig2.Length != Contig1.AncPos)
			return false;
		if (Contig2.EvPos + Contig2.Length != Contig1.EvPos)
			return false;
		}
	return true;
	}

// Contigs are contiguous in both anc and ev genomes?
bool InterChrEvolver::AreContiguous(const ContigData &Contig1,
  const ContigData &Contig2) const
	{
	if (Contig1.AncChrIndex != Contig2.AncChrIndex)
		return false;
	if (Contig1.Plus != Contig2.Plus)
		return false;
	if (Contig1.Copy != Contig2.Copy)
		return false;
	if (Contig1.Plus)
		{
		if (Contig1.AncPos + Contig1.Length != Contig2.AncPos)
			return false;
		if (Contig1.EvPos + Contig1.Length != Contig2.EvPos)
			return false;
		}
	else
		{
		if (Contig2.AncPos + Contig2.Length != Contig1.AncPos)
			return false;
		if (Contig2.EvPos + Contig2.Length != Contig1.EvPos)
			return false;
		}
	return true;
	}

void InterChrEvolver::MergeContiguousNeutrals(ContigData &Contig1,
  const ContigData &Contig2) const
	{
	asserta(Contig1.Type == CT_Neutral && Contig2.Type == CT_Neutral);
	if (!Contig1.Plus)
		{
		asserta(Contig2.AncPos + Contig2.Length == Contig1.AncPos);
		Contig1.AncPos = Contig2.AncPos;
		}
	Contig1.Length += Contig2.Length;
	}

void InterChrEvolver::MergeContiguous(ContigData &Contig1,
  const ContigData &Contig2) const
	{
	ContigData NewContig1(Contig1);
	NewContig1.Length += Contig2.Length;
	NewContig1.Type = CT_Merged;
	if (!Contig1.Plus)
		{
		asserta(Contig2.AncPos + Contig2.Length == Contig1.AncPos);
		NewContig1.AncPos = Contig2.AncPos;
		}
	asserta(NewContig1.AncChrIndex < SIZE(m_AncChrLengths));
	if (NewContig1.GetAncHi() >= m_AncChrLengths[Contig1.AncChrIndex])
		{
		Log("Contig1=");
		Contig1.LogMe(true);
		Log("Contig2=");
		Contig2.LogMe(true);
		Log("Merged=");
		NewContig1.LogMe(true);
		Die("AncHi %"PF " > chr length %"PF,
		  Contig1.GetAncHi(), m_AncChrLengths[Contig1.AncChrIndex]);
		}
	Contig1 = NewContig1;
	}

void InterChrEvolver::GetGene(unsigned GeneIndex, const ContigSet &Contigs,
  ContigSet &Gene) const
	{
	Gene.clear();
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.GeneIndex == GeneIndex)
			Gene.insert(Contig);
		}
	}

void InterChrEvolver::GetAncPos(unsigned EvChrIndex, gpos_t EvPos,
  unsigned &AncChrIndex, gpos_t &AncPos) const
	{
	ContigData Contig;
	GetContig(EvChrIndex, EvPos, Contig);
	asserta(EvPos >= Contig.EvPos && EvPos <= Contig.GetEvHi());
	AncChrIndex = Contig.AncChrIndex;
	if (Contig.Plus)
		AncPos = Contig.AncPos + (EvPos - Contig.EvPos);
	else
		AncPos = Contig.GetAncHi() - (EvPos - Contig.EvPos);
	}

void InterChrEvolver::SetCopy(ContigSet &Contigs) const
	{
	ContigSet NewContigs;
	for (ContigSet::iterator p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		Contig.Copy = true;
		NewContigs.insert(Contig);
		}
	Contigs = NewContigs;
	}

void InterChrEvolver::IncRelocCount(ContigSet &Contigs) const
	{
	ContigSet NewContigs;
	for (ContigSet::iterator p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		++(Contig.RelocCount);
		NewContigs.insert(Contig);
		}
	Contigs = NewContigs;
	}

void InterChrEvolver::SetGeneIndex(ContigSet &Contigs,
  unsigned GeneIndex, bool Strand) const
	{
	ContigSet NewContigs;
	for (ContigSet::iterator p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		Contig.GeneIndex = GeneIndex;
		if (GeneIndex != UINT_MAX)
			Contig.Plus = Strand;
		NewContigs.insert(Contig);
		}
	Contigs = NewContigs;
	}

void InterChrEvolver::GetAncChrSizes(const ContigSet &Contigs,
  vector<gpos_t> &Sizes) const
	{
	Sizes.clear();
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.Copy)
			continue;
		if (Contig.AncChrIndex + 1 > SIZE(Sizes))
			Sizes.resize(Contig.AncChrIndex + 1, 0);
		Sizes[Contig.AncChrIndex] += Contig.Length;
		}

	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.Copy)
			continue;
		if (Contig.GetAncHi() >= Sizes[Contig.AncChrIndex])
			{
			Contig.LogMe(true);
			Die("AncHi=%"PF" > chr length %"PF,
			  Contig.GetAncHi(), Sizes[Contig.AncChrIndex]);
			}
		}
	}

unsigned InterChrEvolver::GetAncChrIndex(const string &Name) const
	{
	const unsigned ChrCount = SIZE(m_AncChrNames);
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		if (Name == m_AncChrNames[ChrIndex])
			return ChrIndex;
	Die("GetAncChrIndex(%s), not found", Name.c_str());
	UNREACHABLE_RETURN(UINT_MAX);
	}

void InterChrEvolver::GetGeneIndexes(IntSet &Indexes) const
	{
	Indexes.clear();
	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.GeneIndex != UINT_MAX)
			Indexes.insert(Contig.GeneIndex);
		}
	}

gpos_t InterChrEvolver::GetConstrainedLength() const
	{
	gpos_t Length = 0;
	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.IsCE())
			Length += Contig.Length;
		}
	return Length;
	}

void InterChrEvolver::RecomputeChrLengths()
	{
	const unsigned ChrCount = GetChrCount();
	asserta(SIZE(m_ChrLengths) == ChrCount);

	m_ChrLengths.clear();
	m_ChrLengths.resize(ChrCount, 0);
	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		asserta(Contig.EvChrIndex < ChrCount);
		m_ChrLengths[Contig.EvChrIndex] += Contig.Length;
		}
	}

gpos_t InterChrEvolver::GetRangeEndPos(unsigned /*ChrIndex*/, gpos_t Pos,
  gpos_t Length) const
	{
	return Pos + Length - 1;
	}

// Any gene bases, may be one or more partial or full
bool InterChrEvolver::RangeHasGene(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const
	{
	ContigSet Contigs;
	GetContigs(ChrIndex, Pos, Length, Contigs);

	for (ContigSet::const_iterator p = Contigs.begin();
	  p != Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.GeneIndex != UINT_MAX)
			return true;
		}
	return false;
	}
