#ifndef icevo_h
#define icevo_h

#include <set>
#include "intlist.h"
#include "utils.h"
#include "event.h"
#include "eventresult.h"
#include "gff2.h"
#include "evolver.h"

using namespace std;

class RevFile;
struct ChrInfo;

const unsigned IceFileMagic = 0x1cef11e;

enum ContigType
	{
	CT_Neutral,
	CT_CDS,
	CT_UTR,
	CT_NXE,
	CT_NGE,
	CT_Merged,
	CT_Range,
	};

const char *ContigTypeToStr(ContigType Type);

struct ChrPosData
	{
	unsigned ChrIndex;
	gpos_t Pos;

	ChrPosData()
		{
		ChrIndex = UINT_MAX;
		Pos = GPOS_T_MAX;
		}

	bool operator<(const ChrPosData &rhs) const
		{
		if (ChrIndex != rhs.ChrIndex)
			return ChrIndex < rhs.ChrIndex;
		return Pos < rhs.Pos;
		}
	};

/***
In InterChrEvolver, ranges are specified by evolved coordinates, in
contrast to [intra-chr]Evolver where they are specified by ancestral
coordinates.

A chromosome is represented as a std::set<ContigData>, where a "contig"
is a maximal set of bases of the same type (Neutral, CDS, UTR, NXE or
NGE) that are contiguous in both the evolved and ancestral sequence.
***/
struct ContigData
	{
	ContigData()
		{
		Type = (ContigType) INT_MIN;
		AncChrIndex = UINT_MAX;
		AncPos = GPOS_T_MAX;
		EvChrIndex = UINT_MAX;
		EvPos = GPOS_T_MAX;
		Length = GPOS_T_MAX;
		GeneIndex = UINT_MAX;
		RelocCount = 0;
		Plus = true;
		Copy = false;
		}

	ContigData(unsigned EvChrIndex_, gpos_t EvPos_)
		{
		EvChrIndex = EvChrIndex_;
		EvPos = EvPos_;
		AncChrIndex = UINT_MAX;
		AncPos = GPOS_T_MAX;
		Length = GPOS_T_MAX;
		GeneIndex = UINT_MAX;
		RelocCount = 0;
		Plus = true;
		Copy = false;
		}

	ContigType Type;
	unsigned EvChrIndex;
	gpos_t EvPos;
	unsigned AncChrIndex;
	gpos_t AncPos;
	gpos_t Length;
	unsigned GeneIndex;
	unsigned RelocCount;
	bool Plus;
	bool Copy;

	bool operator<(const ContigData &rhs) const
		{
		if (EvChrIndex != rhs.EvChrIndex)
			return EvChrIndex < rhs.EvChrIndex;
		return EvPos < rhs.EvPos;
		}

	gpos_t GetEvHi() const
		{
		return EvPos + Length - 1;
		}

	gpos_t GetAncHi() const
		{
		return AncPos + Length - 1;
		}

	bool IsCE() const
		{
		return Type == CT_CDS || Type == CT_UTR ||
		  Type == CT_NXE || Type == CT_NGE;
		}

	bool IsGeneCE() const
		{
		return Type == CT_CDS || Type == CT_UTR || Type == CT_NXE;
		}

	void LogMe(bool NL = false) const
		{
		Log("%s", ContigTypeToStr(Type));
		Log(" EvChr=%u", EvChrIndex);
		Log(" EvPos=%"PF, EvPos);
		Log(" EvHi=%"PF, GetEvHi());
		Log(" AncChr=%u", AncChrIndex);
		Log(" AncPos=%"PF, AncPos);
		Log(" AncHi=%"PF, GetAncHi());
		Log(" L=%"PF, Length);
		Log(" %c", pom(Plus));
		Log(" Cpy=%c", tof(Copy));
		if (GeneIndex != UINT_MAX)
			Log("  Gene=%u", GeneIndex);
		if (NL)
			Log("\n");
		}
	};

typedef set<ContigData> ContigSet;
typedef ContigSet::const_iterator ContigSetCIter;

class InterChrEvolver
	{
public:
	InterChrEvolver();
	void Clear();
	void FromRev(const RevFile &Rev, unsigned GenomeIndex);
	void FromFile(const char *FileName);
	void ToFile(const char *FileName) const;
	void ReadChrLengths(const char *FileName);
	void ReadAnnots(const char *FileName, vector<GFFRecord2> &IslandRecs);
	void InitChrs(const vector<string> &ChrNames, const vector<gpos_t> &ChrLengths);
	void InitNeutrals();

	void SetLogEvents(bool On)
		{
		m_LogEvents = On;
		}

	void PrintAndDie(const char *Format, ...) const;

	void Run();
	void ToGFF(const char *FileName) const;
	void AlnToRev(const char *AlnFileName, const RevFile &AnSeqRev,
	  unsigned AnGenIndex, const char *EvGenomeName, bool Dump);
	void SeqToRev(const char *FileName, const RevFile &AncSeqRev,
	  unsigned AncGenIndex, const char *EvGenomeName) const;
	void ContigToGFF(FILE *f, const ContigData &Contig) const;
	void GetChrInfo(vector<ChrInfo> &CIs) const;
	void AnnotsToGFF(const char *FileName) const;
	void WriteChrNames(const char *OutChrNamesFileName) const;

	gpos_t GetLength() const;
// Need synonym for EV_ACCEPT*() macros
	gpos_t GetEvolvedLength() const { return GetLength(); }

	gpos_t GetChrCount() const
		{
		return SIZE(m_ChrLengths);
		}

	unsigned GetContigCount() const
		{
		return SIZE(m_Contigs);
		}

	const vector<gpos_t> &GetChrLengths() const
		{
		return m_ChrLengths;
		}

	const vector<string> &GetChrNames() const
		{
		return m_ChrNames;
		}

	void GetRandomPos(unsigned &ChrIndex, gpos_t &Pos) const;
	void GetRandomPosDifferentChr(unsigned ChrIndex, unsigned &DiffChrIndex,
	  gpos_t &Pos) const;
	gpos_t GetChrLength(unsigned ChrIndex) const;
	bool RangeExists(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const;
	void GetAncPos(unsigned EvChrIndex, gpos_t EvPos, unsigned &AncChrIndex,
	  gpos_t &AncPos) const;
	unsigned GetAncChrIndex(const string &Name) const;

	void GetContigsRange(const ContigSet &Contigs, ContigData &Range) const;

	void ComputeEventRates();
	void GetEventRates(vector<double> &Rates) const
		{
		Rates = m_IntegratedEventRates;
		}
	double GetAvgWaitTime() const;
	double GetTotalIntegratedRate() const;
	void SetEventCount();
	void SetEventCount2(unsigned EventCount);

	EVENT_TYPE GetRandomEventType() const;

	EVENT_RESULT TryEvent();

	EVENT_RESULT Try_InterChrCopy();
	EVENT_RESULT Try_InterChrMove();
	EVENT_RESULT Try_ChrFuse();
	EVENT_RESULT Try_ChrSplit();
	EVENT_RESULT Try_RecipTransloc();
	EVENT_RESULT Try_NonRecipTransloc();

	gpos_t GetRangeEndPos(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const;
	
	void Copy(unsigned FromChrIndex, gpos_t FromPos, unsigned ToChrIndex,
	  gpos_t ToPos, gpos_t Length, bool Invert, unsigned ToGeneIndex, bool Strand);
	void Move(unsigned FromChrIndex, gpos_t FromPos, unsigned ToChrIndex,
	  gpos_t ToPos, gpos_t Length, bool Invert, unsigned ToGeneIndex, bool Strand);
	void FuseChrs(unsigned ChrIndex1, unsigned ChrIndex2);
	void SplitChr(unsigned ChrIndex, gpos_t Pos);

	void Split(unsigned ChrIndex, gpos_t Pos);
	void GetContigs(unsigned ChrIndex, gpos_t Pos, gpos_t Length,
	  ContigSet &Contigs) const;
	void GetOverlappingContigs(unsigned ChrIndex, gpos_t Pos, gpos_t Length,
	  ContigSet &Contigs) const;
	void GetContigsPosGE(unsigned ChrIndex, gpos_t Pos,
	  ContigSet &Contigs) const;
	void InsertContigs(unsigned ChrIndex, gpos_t Pos, const ContigSet &Contigs);
	void AppendContigs(unsigned ChrIndex, const ContigSet &Contigs);
	void DeleteContigs(unsigned ChrIndex, gpos_t Pos, gpos_t Length,
	  ContigSet &DeletedContigs);
	void TruncateChr(unsigned ChrIndex, gpos_t Pos, ContigSet &DeletedContigs);
	void InvertContigs(ContigSet &Contigs);
	const ContigData &GetLastContig(unsigned ChrIndex) const;
	void GetContig(unsigned ChrIndex, gpos_t Pos, ContigData &Contig) const;
	ContigSetCIter GetContigCIter(unsigned ChrIndex, gpos_t Pos) const;
	void Deconstrain(ContigSet &Contigs, bool DeGene) const;
	void DeconstrainContig(ContigData &Contig, bool DeGene) const;
	void MergeContiguousNeutrals();
	void GetMaximalContigs(ContigSet &Contigs) const;
	bool AreContiguous(const ContigData &Contig1,
	  const ContigData &Contig2) const;
	bool AreContiguousNeutrals(const ContigData &Contig1,
	  const ContigData &Contig2) const;
	void MergeContiguousNeutrals(ContigData &Contig1,
	  const ContigData &Contig2) const;
	void MergeContiguous(ContigData &Contig1, const ContigData &Contig2) const;
	void SetCopy(ContigSet &Contigs) const;
	void IncRelocCount(ContigSet &Contigs) const;
	void SetGeneIndex(ContigSet &Contigs, unsigned GeneIndex, bool Strand) const;
	void GetAncChrSizes(const ContigSet &Contigs, vector<gpos_t> &Sizes) const;
	const GFFRecord2 &GetAncAnnot(unsigned ChrIndex, gpos_t Pos) const;

	void GeneDupe(unsigned FromChrIndex, gpos_t FromPos, unsigned ToChrIndex,
	  gpos_t ToPos, gpos_t CopyLength, bool DoInvert);
	void DeleteGene(const ContigSet &Gene);

	bool AcceptInsert(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const;
	bool PosInGene(unsigned ChrIndex, gpos_t Pos) const;
	bool PosIsConstrained(unsigned ChrIndex, gpos_t Pos) const;
	bool RangeHasCE(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const;
	bool RangeHasGene(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const;
	void GetCompleteGeneRanges(unsigned ChrIndex, gpos_t Pos, gpos_t Length,
	  map<unsigned, pair<gpos_t, gpos_t> > &LoHis) const;
	unsigned GetGeneIndex(unsigned ChrIndex, gpos_t Pos, bool &Strand) const;
	void GetGene(unsigned GeneIndex, const ContigSet &Contigs,
	  ContigSet &Gene) const;
	unsigned ActivateCopiedGene(const ContigSet &Gene, unsigned FromChrIndex,
	  gpos_t FromPos, unsigned ToChrIndex, gpos_t ToPos, gpos_t CopyLength,
	  bool DoInvert);
	void GetGeneIndexes(IntSet &Indexes) const;
	gpos_t GetConstrainedLength() const;
	void RecomputeChrLengths();

	void GetChrParts(unsigned ChrIndex, gpos_t Pos, gpos_t &Length1,
	  gpos_t &Length2) const;

	void LogMe() const;
	void LogRange(unsigned ChrIndex, gpos_t Pos, gpos_t Length) const;
	void LogContigs(const ContigSet &Contigs) const;

	void Validate(const string &Why) const;
	void ValidateIntact() const;
	void ValidateChrLengths() const;
	void AssertContiguous(const ContigSet &Contigs) const;
	void ValidateGenes() const;

public:
	double m_Duration;

private:
	bool m_LogEvents;
	mutable string m_ValidateWhy;
	unsigned m_LastAcceptedEventIndex;
	unsigned m_EventCount;
	unsigned m_EventIndex;
	gpos_t m_LastAcceptedEventPos;
	gpos_t m_CurrentEventPos;
	EVENT_TYPE m_LastAcceptedEvent;
	EVENT_TYPE m_CurrentEvent;
	vector<double> m_IntegratedEventRates;

// Saved info for ancestral genome
	vector<gpos_t> m_AncChrLengths;
	vector<string> m_AncChrNames;
	map<ChrPosData, GFFRecord2> m_AncAnnots;

// Evolved genome
	vector<unsigned> m_ChrLengths;
	vector<string> m_ChrNames;
	ContigSet m_Contigs;
	unsigned m_NextNewGeneIndex;

	map<unsigned, double> m_GeneAccels;
	};

#endif // icevo_h
