#include "ice.h"
#include "evomodel.h"

void InterChrEvolver::Copy(unsigned FromChrIndex, gpos_t FromPos, 
  unsigned ToChrIndex, gpos_t ToPos, gpos_t Length, bool Invert,
  unsigned ToGeneIndex, bool Strand)
	{
	asserta(FromChrIndex != ToChrIndex);

	set<ContigData> Contigs;
	GetContigs(FromChrIndex, FromPos, Length, Contigs);
	if (Invert)
		InvertContigs(Contigs);
	Deconstrain(Contigs, true);
	SetCopy(Contigs);
	IncRelocCount(Contigs);
	SetGeneIndex(Contigs, ToGeneIndex, Strand);
	InsertContigs(ToChrIndex, ToPos, Contigs);
	}

void InterChrEvolver::Move(unsigned FromChrIndex, gpos_t FromPos, 
  unsigned ToChrIndex, gpos_t ToPos, gpos_t Length, bool Invert,
  unsigned ToGeneIndex, bool Strand)
	{
	asserta(FromChrIndex != ToChrIndex);

	set<ContigData> Contigs;
	DeleteContigs(FromChrIndex, FromPos, Length, Contigs);
	if (Invert)
		InvertContigs(Contigs);
	IncRelocCount(Contigs);
	if (ToGeneIndex != UINT_MAX)
		SetGeneIndex(Contigs, ToGeneIndex, Strand);
	InsertContigs(ToChrIndex, ToPos, Contigs);
	}

// Append chr2 to chr1.
void InterChrEvolver::FuseChrs(unsigned ChrIndex1, unsigned ChrIndex2)
	{
	Validate("Start of FuseChrs");

	asserta(ChrIndex1 != ChrIndex2);

	string NewName = m_ChrNames[ChrIndex1] + "+" + m_ChrNames[ChrIndex2];
	if (NewName.size() > P_MaxChrName)
		NewName = NewName.substr(0, P_MaxChrName) + "..";

// ChrIndexLo becomes Chr1+Chr2
// ChrIndexHi deleted.
	unsigned ChrIndexLo = min(ChrIndex1, ChrIndex2);
	unsigned ChrIndexHi = max(ChrIndex1, ChrIndex2);

	gpos_t ChrLength1 = GetChrLength(ChrIndex1);
	gpos_t ChrLength2 = GetChrLength(ChrIndex2);

	ContigSet Contigs1;
	ContigSet Contigs2;
	GetContigs(ChrIndex1, 0, ChrLength1, Contigs1);
	GetContigs(ChrIndex2, 0, ChrLength2, Contigs2);

// Delete all contigs for chr1
	for (ContigSetCIter p = Contigs1.begin(); p != Contigs1.end(); ++p)
		{
		const ContigData &Contig = *p;
		size_t n = m_Contigs.erase(Contig);
		asserta(n == 1);
		}

// Delete all contigs for chr2
	for (ContigSetCIter p = Contigs2.begin(); p != Contigs2.end(); ++p)
		{
		const ContigData Contig = *p;
		size_t n = m_Contigs.erase(Contig);
		asserta(n == 1);
		}

// Adjust coords in chr1 contigs
	for (ContigSetCIter p = Contigs1.begin(); p != Contigs1.end(); ++p)
		{
		ContigData Contig = *p;
		Contig.EvChrIndex = ChrIndexLo;
		m_Contigs.insert(Contig);
		}

// Adjust coords in chr2 contigs
	for (ContigSetCIter p = Contigs2.begin(); p != Contigs2.end(); ++p)
		{
		ContigData Contig = *p;
		Contig.EvChrIndex = ChrIndexLo;
		Contig.EvPos += ChrLength1;
		m_Contigs.insert(Contig);
		}

// Adjust chr indexes in later chrs
	ContigSet NewContigs;
	const unsigned ChrCount = GetChrCount();
	for (unsigned ChrIndex = ChrIndexHi + 1; ChrIndex < ChrCount; ++ChrIndex)
		{
		ContigSet Contigs;
		gpos_t ChrLength = GetChrLength(ChrIndex);
		GetContigs(ChrIndex, 0, ChrLength, Contigs);

		for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
			{
			ContigData Contig = *p;
			size_t n = m_Contigs.erase(Contig);
			asserta(n == 1);
			--(Contig.EvChrIndex);
			NewContigs.insert(Contig);
			}
		}
	m_Contigs.insert(NewContigs.begin(), NewContigs.end());

	for (unsigned ChrIndex = ChrIndexHi; ChrIndex + 1 < ChrCount; ++ChrIndex)
		{
		m_ChrLengths[ChrIndex] = m_ChrLengths[ChrIndex+1];
		m_ChrNames[ChrIndex] = m_ChrNames[ChrIndex+1];
		}

	m_ChrLengths[ChrIndexLo] = ChrLength1 + ChrLength2;
	m_ChrNames[ChrIndexLo] = NewName;

	m_ChrLengths.resize(ChrCount-1);
	m_ChrNames.resize(ChrCount-1);

	Validate("End of FuseChrs");
	}

// Pos is first pos in new chr
void InterChrEvolver::SplitChr(unsigned ChrIndex, gpos_t Pos)
	{
	Validate("Start of SplitChr");

	gpos_t ChrLength = GetChrLength(ChrIndex);
	asserta(Pos > 0 && Pos + 1 < ChrLength);

	const unsigned ChrCount = GetChrCount();

	Split(ChrIndex, Pos);

	ContigSet Contigs;
	GetContigsPosGE(ChrIndex, Pos, Contigs);

	ContigSet NewContigs;
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		ContigData Contig = *p;
		size_t n = m_Contigs.erase(Contig);
		asserta(n == 1);

		asserta(Contig.EvPos >= Pos);
		Contig.EvChrIndex = ChrCount;
		Contig.EvPos -= Pos;
		NewContigs.insert(Contig);
		}

	AssertContiguous(NewContigs);
	m_Contigs.insert(NewContigs.begin(), NewContigs.end());

	asserta(Pos < ChrLength);
	gpos_t NewChrLength = ChrLength - Pos;

	string OldName = m_ChrNames[ChrIndex];
	m_ChrNames[ChrIndex] = OldName + ".1";
	m_ChrNames.push_back(OldName + ".2");

	m_ChrLengths[ChrIndex] = Pos;
	m_ChrLengths.push_back(NewChrLength);

	Validate("End of SplitChr");
	}
