#ifndef _IVTREE_H
#define _IVTREE_H

#include <cassert>
#include <utility>
#include <vector>

template <typename T, typename Coord = unsigned int>
class defGetIvFunc
{
	public:
	
	std::pair<Coord, Coord> operator()(const T &t) const
	{
		return t();
	}
};

/* Interval Tree
 * -------------
 *
 * Usage: IvTree <T, Coord, G>
 * 
 * The IvTree is an interval tree. Intervals are of type pair<Coord,Coord>;
 * Coord defaults to unsigned int unless explicitly given. The tree stores
 * data of type T. In order to convert those objects of type T to an interval,
 * functor G is called upon T. By default G is the () operator of T.
 *
 * Constructors: IvTree()
 *                 [empty constructor]
 *               IvTree(g)
 *                 [gives a specific instance of G. G must be a function
 *                 that takes const T& and returns pair<Coord,Coord>]
 *
 * Methods: Add(const T &t)
 *            [adds a new object to the tree]
 *          Delete(const T &t)
 *            [removes an existing object from the tree]
 *          GetOverlaps(Interval iv, vector<T> &results)
 *            [iv is a pair<Coord,Coord>; this will find all overlapping
 *            objects T and append them in results]
 *          Traverse(void (*tfunc)(const T &t))
 *            [traverses the whole tree in order, calling tfunc on every
 *            node -- useful if you want for example to print the tree]
 *          Clear()
 *            [clears the tree]
 *
 * Example of usage: T is a SegPair, Coord is unsigned int, G is a function that
 * returns either the X-projection or the Y-projection of T. See sptree.cpp for
 * such an implementation.
 */

template <typename T, typename Coord = unsigned int, typename GetIvFunc = defGetIvFunc<T, Coord> >
class IvTree
{
	public:
	
	typedef std::pair<Coord, Coord> Interval;

	IvTree(GetIvFunc g) : GetIv(g)
		{
		Init();
		}

	IvTree()
		{
		Init();
		}

	~IvTree();

	void operator=(const IvTree<T, Coord, GetIvFunc> &rhs)
		{
		asserta(rhs.root == rhs.nil);
		}

	IvTree(const IvTree<T, Coord, GetIvFunc> &/*rhs*/)
		{
		Init();
		}

	void Add(const T &t);
	bool Delete(const T &t) { return Delete(t, root); }
	void Traverse(void (*tfunc)(const T &t)) const;
	void GetOverlaps(Interval iv, std::vector<T> &results) const;
	void Clear();
		
	private:
	
	class IvRBNode
	{
		public:

		T key;
		Coord maxc;
		IvRBNode *left;
		IvRBNode *right;
		IvRBNode *p;
		bool isRed;
		
		IvRBNode() {}
		IvRBNode(const T &t) : key(t) {}
	};

	GetIvFunc GetIv;
	IvRBNode *root;
	IvRBNode *nil;

	IvRBNode *NewNode(const T &t)
	{
		IvRBNode *x = new IvRBNode(t);
		x->left = x->right = x->p = nil;
		x->isRed = false;
		Fix(x);
		return x;
	}

	void Init()
	{
		root = nil = new IvRBNode;
		nil->left = nil->right = nil->p = nil;
		nil->isRed = false;
	}

	void Fix(IvRBNode *x)
	{
		if (x == nil)
			return;

		x->maxc = GetIv(x->key).second;
		if ((x->left != nil) && (x->maxc < x->left->maxc))
			x->maxc = x->left->maxc;
		if ((x->right != nil) && (x->maxc < x->right->maxc))
			x->maxc = x->right->maxc;
	}

	void LeftRotate(IvRBNode *x);
	void RightRotate(IvRBNode *y);
	void TraverseNode(void (*tfunc)(const T &t), IvRBNode *x) const;
	void DeleteTree(IvRBNode *x);
	void DeleteNode(IvRBNode *z);
	void DeleteFixup(IvRBNode *z);
	bool Delete(const T &t, IvRBNode *root);

	bool Overlaps(Interval &iv1, Interval &iv2) const
	{
		return (!((iv1.second < iv2.first) || (iv2.second < iv1.first)));
	}

	void RecurseGetOverlaps(Interval iv, std::vector<T> &results, IvRBNode *r) const;
};

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::Traverse(void (*tfunc)(const T &t)) const
{
	TraverseNode(tfunc, root);
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::TraverseNode(void (*tfunc)(const T &t), IvRBNode *x) const
{
	if (x != nil)
	{
		TraverseNode(tfunc, x->left);
		tfunc(x->key);
		TraverseNode(tfunc, x->right);
	}
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::DeleteNode(IvRBNode *z)
{
	IvRBNode *y = z, *w, *x;

	if ((z->left == nil) || (z->right == nil))
		y = z;
	else
	{
		x = z;
		if (x->right != nil)
		{
			x = x->right;
			while (x->left != nil)
				x = x->left;
			y = x;
		}
		else
		{
			IvRBNode *t = x->p;
			while ((t != nil) && (x == t->right))
			{
				x = t;
				t = t->p;
			}
			y = t;
		}
	}

	if (y->left != nil)
		x = y->left;
	else
		x = y->right;
	x->p = y->p;
	if (y->p == nil)
		root = x;
	else
	{
		if (y == y->p->left)
			y->p->left = x;
		else
			y->p->right = x;

		w = x;
		while (w != root)
		{
			w = w->p;
			Fix(w);
		}
	}

	if (y != z)
	{
		z->key = y->key;
		w = z;
		Fix(w);
		while (w != root)
		{
			w = w->p;
			Fix(w);
		}
	}

	if (!(y->isRed))
		DeleteFixup(x);

	if (y != nil)
		delete y;
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::DeleteFixup(IvRBNode *x)
{
	while ((x != root) && (!(x->isRed)))
	{
		if (x == x->p->left)
		{
			IvRBNode *w;

			w = x->p->right;
			if (w->isRed)
			{
				w->isRed = false;
				x->p->isRed = true;
				LeftRotate(x->p);
				w = x->p->right;
			}
			if ((!(w->left->isRed)) && (!(w->right->isRed)))
			{
				w->isRed = true;
				x = x->p;
			}
			else
			{
				if (!(w->right->isRed))
				{
					w->left->isRed = false;
					w->isRed = true;
					RightRotate(w);
					w = x->p->right;
				}
				w->isRed = x->p->isRed;
				x->p->isRed = false;
				w->right->isRed = false;
				LeftRotate(x->p);
				x = root;
			}
		}
		else
		{
			IvRBNode *w;

			w = x->p->left;
			if (w->isRed)
			{
				w->isRed = false;
				x->p->isRed = true;
				RightRotate(x->p);
				w = x->p->left;
			}
			if ((!(w->right->isRed)) && (!(w->left->isRed)))
			{
				w->isRed = true;
				x = x->p;
			}
			else
			{
				if (!(w->left->isRed))
				{
					w->right->isRed = false;
					w->isRed = true;
					LeftRotate(w);
					w = x->p->left;
				}
				w->isRed = x->p->isRed;
				x->p->isRed = false;
				w->left->isRed = false;
				RightRotate(x->p);
				x = root;
			}
		}
	}
	x->isRed = false;
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::LeftRotate(IvRBNode *x)
{
	assert(x->right != nil);

	IvRBNode *y;

	y = x->right;
	x->right = y->left;
	if (y->left != nil)
		y->left->p = x;

	y->p = x->p;
	if (x->p == nil)
		root = y;
	else if (x == x->p->left)
		x->p->left = y;
	else
		x->p->right = y;

	y->left = x;
	x->p = y;

	Fix(x);
	Fix(y);
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::RightRotate(IvRBNode *y)
{
	assert(y->left != nil);

	IvRBNode *x;

	x = y->left;
	y->left = x->right;
	if (x->right != nil)
		x->right->p = y;

	x->p = y->p;
	if (y->p == nil)
		root = x;
	else if (y == y->p->right)
		y->p->right = x;
	else
		y->p->left = x;

	x->right = y;
	y->p = x;

	Fix(y);
	Fix(x);
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::Add(const T &t)
{
	IvRBNode *z = NewNode(t), *y = nil, *x = root;

	while (x != nil)
	{
		y = x;
		if (GetIv(z->key).first < GetIv(x->key).first)
			x = x->left;
		else
			x = x->right;
	}
	
	z->p = y;

	if (y == nil)
		root = z;
	else if (GetIv(z->key).first < GetIv(y->key).first)
		y->left = z;
	else
		y->right = z;

	x = z;

	while (z != root)
	{
		z = z->p;
		Fix(z);
	}

	x->isRed = true;

	while (x != root && x->p->isRed)
	{
		if (x->p == x->p->p->left)
		{
			y = x->p->p->right;
			if (y->isRed)
			{
				x->p->isRed = false;
				y->isRed = false;
				x->p->p->isRed = true;
				x = x->p->p;
			}
			else
			{
				if (x == x->p->right)
				{
					x = x->p;
					LeftRotate(x);
				}
				x->p->isRed = false;
				x->p->p->isRed = true;
				RightRotate(x->p->p);
			}
		}
		else
		{
			y = x->p->p->left;
			if (y->isRed)
			{
				x->p->isRed = false;
				y->isRed = false;
				x->p->p->isRed = true;
				x = x->p->p;
			}
			else
			{
				if (x == x->p->left)
				{
					x = x->p;
					RightRotate(x);
				}
				x->p->isRed = false;
				x->p->p->isRed = true;
				LeftRotate(x->p->p);
			}
		}
	}

	root->isRed = false;
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::GetOverlaps(Interval iv, std::vector<T> &results) const
{
	results.clear();
	RecurseGetOverlaps(iv, results, root);
}

template <typename T, typename Coord, typename GetIvFunc>
bool IvTree<T, Coord, GetIvFunc>::Delete(const T &t, IvRBNode *root)
{
	IvRBNode *x = root;

	Interval ivt = GetIv(t);
	while (x != nil)
	{
		Interval ivx = GetIv(x->key);
		if (ivt.first < ivx.first)
			x = x->left;
		else if (ivt.first > ivx.first)
			x = x->right;
		else
		{
			if (t == x->key)
			{
				DeleteNode(x);
				return true;
			}
			else
			{
				return (Delete(t, x->left) || Delete(t, x->right));
			}
		}
	}
	return false;
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::RecurseGetOverlaps(Interval iv, std::vector<T> &results, IvRBNode *root) const
{
	if (root == nil)
		return;
	
	if ((root->left != nil) && (!(root->left->maxc < iv.first)))
		RecurseGetOverlaps(iv, results, root->left);

	Interval riv = GetIv(root->key);
	if (Overlaps(iv, riv))
		results.push_back(root->key);

	if ((root->right != nil) && (!(iv.second < riv.first)))
		RecurseGetOverlaps(iv, results, root->right);
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::DeleteTree(IvRBNode *root)
{
	if (root != nil)
	{
		DeleteTree(root->left);
		DeleteTree(root->right);
		delete root;
	}
}

template <typename T, typename Coord, typename GetIvFunc>
void IvTree<T, Coord, GetIvFunc>::Clear()
{
	DeleteTree(root);
	root = nil;
}

template <typename T, typename Coord, typename GetIvFunc>
IvTree<T, Coord, GetIvFunc>::~IvTree()
{
	DeleteTree(root);
	delete nil;
}

#endif
