#include <errno.h>
#include <algorithm>
#include <string>
#include "utils.h"
#include "revfile.h"
#include "progress.h"
#include "getopts.h"
#include "kfile.h"
#include "kvfile.h"
#include "tids.h"
#include "bmgraph.h"
#include "minheap.h"

#define TIDLEN 10
#define PENALTY_MIN 10

VALUE_OPT ValueOpts[] =
{
	{ "alignan",            0 },
	{ "andanR",             0 },
	{ "andkpL",             0 },
	{ "andkpR",             0 },
	{ "andscoredkpR",       0 },
	{ "dumpan",             0 },
	{ "dumpdecomp",         0 },
	{ "dumpkp",             0 },
	{ "findlimits",         0 },
	{ "findlimits2",        0 },
	{ "flist",              0 },
	{ "froman",             0 },
	{ "fromanL",            0 },
	{ "fromkp",             0 },
	{ "fromrev",            0 },
	{ "fromscoredkpL",      0 },
	{ "fromsortedkp",       0 },
	{ "fromsorteduniqkp",   0 },
	{ "genomeindex",        0 },
	{ "log",                0 },
	{ "maxchrom",           0 },
	{ "maxentries",         0 },
	{ "mergesortedkp",      0 },
	{ "minsize",            0 },
	{ "out",                0 },
	{ "scoremergesortedkp", 0 },
	{ "sortdecomp",         0 },
	{ "sortkp",             0 },
	{ "toalldecomp",        0 },
	{ "toan",               0 },
	{ "todecomp",           0 },
	{ "todummyan",          0 },
	{ "tokp",               0 },
	{ "tomincovkp",         0 },
	{ "tosortedan",         0 },
	{ "tosortedkp",         0 },
	{ "touniqkp",           0 },
	{ "verifygks",          0 },
	{ "withsorteddecomp",   0 }
};
int ValueOptCount = sizeof(ValueOpts)/sizeof(ValueOpts[0]);

FLAG_OPT FlagOpts[] =
{
	{ "flipgenomes",  0 },
	{ "bysecond",     0 }
};
int FlagOptCount = sizeof(FlagOpts)/sizeof(FlagOpts[0]);

void Usage()
{
	fprintf(stderr,
"Usage:\n"
"    kf\n"
"      -fromrev <revfile> -tokp <kfile>\n"
"      -fromrev <revfile> -todummyan <kvfile> [-genomeindex <i>] [-maxchrom <n>]\n"
"      -dumpkp <kfile>\n"
"      -dumpdecomp <kfile>\n"
"      -dumpan <kfile> [-minsize <count>]\n"
"      -sortkp <kfile> [-bysecond]\n"
"      -sortdecomp <dfile>\n"
"      -fromkp <kfile> -tosortedkp <kfile> [-bysecond] [-maxentries <n>]\n"
"      -froman <kfile> -tosortedan <kfile> [-maxentries <n>]\n"
"      -fromsortedkp <kfile> -touniqkp <kfile>\n"
"      -fromsorteduniqkp <kfile> -tomincovkp <kfile>\n"
"      -froman <kvfile> -todecomp <dfile> [-genomeindex <i>]\n"
"      -froman <kvfile> -toalldecomp <dfile_prefix>\n"
"      -mergesortedkp <kfile> -withsorteddecomp <dfile> -tokp <kfile>\n"
"      -scoremergesortedkp <kfile> -tokp <kfile>\n"
"      -fromscoredkpL <kfile> -andscoredkpR <kfile> -tomincovkp <kfile>\n"
"      -fromanL -andanR -andkpL -andkpR -toan\n"
"      -verifygks <kvfile>\n"
"      -findlimits <kvfile>\n"
"      -findlimits2 <kvfile>\n"
"\n"
	);
}

void ParseStrVec(const std::string &In, std::vector<std::string> &Out, char Sep);

template <typename T>
int VectorSortedLocate(const std::vector<T> &V, const T &Element)
{
	std::pair<typename std::vector<T>::const_iterator, typename std::vector<T>::const_iterator> ip = std::equal_range(V.begin(), V.end(), Element);

	if (ip.first == ip.second)
		Die("Unable to locate element in vector");

	return ip.first - V.begin();
}

template <typename T>
int VectorUnsortedLocate(const std::vector<T> &V, const T &Element)
{
	for (typename std::vector<T>::const_iterator i = V.begin(); i != V.end(); ++i)
		if (Element == *i)
			return (i - V.begin());

	Die("Unable to locate element in vector");

	return 0; // Unreachable
}

template <typename T, typename CompareFunc = std::less<T> >
class VectorIndexCmp
{
	const std::vector<T> &V;
	CompareFunc C;

	public:
	VectorIndexCmp(const std::vector<T> &Vec, CompareFunc Cmp = CompareFunc()) : V(Vec), C(Cmp) { }
	bool operator()(const unsigned &a, const unsigned &b) const
	{
		return C(V[a], V[b]);
	}
};

typedef std::set<unsigned> UnsignedSet;
typedef std::list<KPair> KPairList;

void RevToKFPair(const char *RevFileName, const char *KFileName, bool FlipGenomes)
{
	RevFile InFile;
	KFilePair KF;

	InFile.Open(RevFileName);
	if (InFile.GetGenomeCount() != 2)
		Die("This conversion only works with two genomes");

	GenInfoVec Genomes = InFile.GetGenomes();
	if (FlipGenomes)
		std::reverse(Genomes.begin(), Genomes.end());

	KF.Create(KFileName, Genomes);

	RevToKF(InFile, KF, FlipGenomes ? 1 : 0, TIDLEN);

	KF.Close();
	InFile.Close();
	Progress("Done");
}

//TODO deprecated
void RevToKFPair2(const char *RevFileName, const char *KFileName)
{
	RevFile InFile;
	KFilePair KF;

	InFile.Open(RevFileName);
	if (InFile.GetGenomeCount() != 2)
		Die("This conversion only works with two genomes");
	KF.Create(KFileName, InFile.GetGenomes());

	OffsetsVec Offsets;
	TMap Map;

	FillMap(Offsets, Map, InFile, 0, TIDLEN);

	Progress("Converting...");
	for (TMap::const_iterator i = Map.begin(); i != Map.end(); ++i)
	{
		for (std::map<int, int>::const_iterator j = i->second.begin(); j != i->second.end(); ++j)
		{
			KPair KP;
			KP.k[0] = i->first;
			KP.k[1] = j->first;
			KP.score = j->second;
			KF.AddEntry(KP);
		}
	}

	KF.Close();
	InFile.Close();
	Progress("Done");
}

void DumpKPair(const char *KFileName)
{
	KFilePair KF;
	KF.Open(KFileName);

	unsigned EntryCount = KF.GetEntryCount();
	for (unsigned i = 0; i != EntryCount; ++i)
	{
		KPair KP;
		KF.GetNextEntry(KP);
		printf("%u/%u\t%d\t%d\t%u\n", i+1, EntryCount, KP.k[0], KP.k[1], KP.score);
	}
}

void DumpKDecomp(const char *KFileName)
{
	KFileDecomp KF;
	KF.Open(KFileName);

	unsigned EntryCount = KF.GetEntryCount();
	for (unsigned i = 0; i != EntryCount; ++i)
	{
		KDecomp KD;
		KF.GetNextEntry(KD);
		printf("%u/%u\t%d\t%u\n", i+1, EntryCount, KD.k, KD.idx);
	}
}

void DumpANFile(const char *ANFileName, unsigned MinSize)
{
	AlnNodeFile InFile;
	InFile.Open(ANFileName);

	unsigned EntryCount = InFile.GetEntryCount();
	unsigned EntriesOut = 0;
	printf("File contains %u genomes, %u entries\n", InFile.GetGenomeCount(), EntryCount);
	for (unsigned i = 0; i != EntryCount; ++i)
	{
		AlnNode AN;
		InFile.GetEntry(AN, i);
		if (AN.GKmers.size() >= MinSize)
		{
			++EntriesOut;
			printf("Entry %u:\n", i);
			for (unsigned j = 0; j != AN.GKmers.size(); ++j)
			{
				printf("\t%u\t%d\n", AN.GKmers[j].G, AN.GKmers[j].k);
			}
		}
	}
	printf("Dumped %u entries to the output\n", EntriesOut);
}

void VerifyGKS(const char *ANFileName)
{
	AlnNodeFile InFile;
	InFile.Open(ANFileName);

	ProgressStart("Verifying");
	for (unsigned i = 0; i != InFile.GetEntryCount(); ++i)
	{
		if ((i & 0xffff) == 0)
			ProgressStep(i, InFile.GetEntryCount());
		AlnNode AN;
		InFile.GetEntry(AN, i);
		for (unsigned j = 1; j < AN.GKmers.size(); ++j)
		{
			if (AN.GKmers[j].G <= AN.GKmers[j-1].G)
				Die("Invalid GKmer (%u,%d) in entry #%u in file %s", AN.GKmers[j].G, AN.GKmers[j].k, i, ANFileName);
		}
	}
	ProgressDone();
	InFile.Close();
}

void FindLimits(const char *ANFileName)
{
	AlnNodeFile InFile;
	InFile.Open(ANFileName);

	AlnNode InitAN, PrevAN;
	//FIXME
	InFile.GetEntry(PrevAN, 0);
	InitAN = PrevAN;
	for (unsigned i = 1; i < InFile.GetEntryCount(); ++i)
	{
		AlnNode AN;
		InFile.GetEntry(AN, i);
		if (!PrevAN.IsAdjacentTo(AN))
		{
			for (unsigned j = 0; j < InitAN.GKmers.size(); ++j)
			{
				printf("Genome %u: %d - %d\n", InitAN.GKmers[j].G, InitAN.GKmers[j].k, PrevAN.GKmers[j].k);
			}
			printf("\n");
			InitAN = AN;
		}
		PrevAN = AN;
	}
	InFile.Close();
}

//TODO
void FindLimits2(const char *ANFileName)
{
	AlnNodeFile InFile;
	InFile.Open(ANFileName);

	unsigned EntryCount = InFile.GetEntryCount();
	std::vector<bool> Used(EntryCount, false);

	for (unsigned i = 0; i != EntryCount; ++i)
	{
		if (Used[i])
			continue;

		AlnNode AN;
		Used[i] = true;

		InFile.GetEntry(AN, i);
		printf("DEBUG: %u\n", i);
		for (unsigned k = 0; k < AN.GetGKmerCount(); ++k)
		{
			printf("\t%u\t%d\n", AN.GetGKmer(k).G, AN.GetGKmer(k).k);
		}
		unsigned DEBUGcount = 0;
		for (unsigned j = i + 1; j < EntryCount; ++j)
		{
			AlnNode AN2;
			InFile.GetEntry(AN2, j);
			if (AN2.GetGKmer(0).IsAdjacentTo(AN.GetGKmer(0)))
			{
				if (AN2.IsAdjacentTo(AN))
				{
					++DEBUGcount;
					printf("\tDEBUGIN: %u [COUNT: %u]\n", j, DEBUGcount);
					for (unsigned k = 0; k < AN2.GetGKmerCount(); ++k)
					{
						printf("\t\t%u\t%d\n", AN2.GetGKmer(k).G, AN2.GetGKmer(k).k);
					}
				}
			}
			else
				break;
		}
	}


}

void ToSortedAN(const char *ANFileName, const char *SortedANFileName, unsigned MaxEntries)
{
	AlnNodeFile InFile, OutFile;

	InFile.Open(ANFileName);

	unsigned EntryCount = InFile.GetEntryCount();
	if (EntryCount <= MaxEntries)
	{
		Progress("Reading input file");
		std::vector<AlnNode> InNodes;
		InFile.GetAllEntries(InNodes);

		Progress("Sorting");
		std::sort(InNodes.begin(), InNodes.end());

		Progress("Writing output file");
		AlnNodeFile OutFile;
		OutFile.Create(SortedANFileName, InFile.GetGenomes());
		OutFile.AddEntries(InNodes);
		OutFile.Close();
		InFile.Close();
	}
	else
	{
		unsigned n = EntryCount / MaxEntries;

		if ((EntryCount % MaxEntries) > 0)
			++n;

		std::vector<AlnNodeFile> ANFiles(n);
		std::vector<FILE *> fp(n);

		unsigned EntriesLeft = EntryCount;

		for (unsigned i = 0; i < n; ++i)
		{
			fp[i] = CreateStdioTmpFile();
			ANFiles[i].CreateFP(fp[i], InFile.GetGenomes());
			unsigned Count = (EntriesLeft > MaxEntries) ? MaxEntries : EntriesLeft;

			Progress("Reading chunk #%u out of %u", i+1, n);
			std::vector<AlnNode> ANV;
			InFile.GetEntries(ANV, EntryCount - EntriesLeft, EntryCount - EntriesLeft + Count - 1);
			EntriesLeft -= Count;

			Progress("Sorting chunk #%u out of %u", i+1, n);
			std::sort(ANV.begin(), ANV.end());

			Progress("Writing chunk #%u out of %u", i+1, n);
			ANFiles[i].AddEntries(ANV);
			ANFiles[i].Close();
		}

		std::vector<unsigned> Counts(n);
		std::vector<AlnNode> ANodes(n);
		VectorIndexCmp<AlnNode> vcmp(ANodes);
		MinHeap<unsigned, VectorIndexCmp<AlnNode> > heap(vcmp);

		for (unsigned i = 0; i < n; ++i)
		{
			ANFiles[i].OpenFP(fp[i]);
			Counts[i] = ANFiles[i].GetEntryCount();
			assert(Counts[i] > 0);
			ANFiles[i].GetEntry(ANodes[i], 0);
			heap.insert(i);
		}

		AlnNodeFile OutFile;
		OutFile.Create(SortedANFileName, InFile.GetGenomes());

		ProgressStart("Merging");
		unsigned EntriesProcessed = 0;
		while (!heap.empty())
		{
			if ((EntriesProcessed & 0xffff) == 0)
			{
				ProgressStep(EntriesProcessed, EntryCount);
			}
			unsigned i = heap.min();
			heap.remove_min();
			OutFile.AddEntry(ANodes[i]);
			--Counts[i];
			++EntriesProcessed;
			if (Counts[i] != 0)
			{
				ANFiles[i].GetEntry(ANodes[i], ANFiles[i].GetEntryCount() - Counts[i]);
				heap.insert(i);
			}
		}
		ProgressDone();

		for (unsigned i = 0; i < n; ++i)
		{
			ANFiles[i].Close();
		}

		InFile.Close();
		OutFile.Close();
	}

	
}

void ToSortedKP(const char *KFileName, const char *SortedKFileName, bool BySecond, unsigned MaxEntries)
{
	KFilePair KF;
	KF.Open(KFileName);

	KPairCompare KPC(BySecond ? 1 : 0);

	unsigned EntryCount = KF.GetEntryCount();
	if (EntryCount <= MaxEntries)
	{
		Progress("Reading input file");
		std::vector<KPair> KPairs;
		KF.GetAllEntries(KPairs);
		KF.Close();

		Progress("Sorting");
		std::sort(KPairs.begin(), KPairs.end(), KPC);

		Progress("Writing output file");
		KFilePair SortedKF;
		SortedKF.Create(SortedKFileName, KF.GetGenomes());
		SortedKF.AddEntries(KPairs);
		SortedKF.Close();
	}
	else
	{
		unsigned n = EntryCount / MaxEntries;

		if ((EntryCount % MaxEntries) > 0)
			++n;

		std::vector<FILE *> fp(n);
		std::vector<unsigned> EntriesLeft(n);
		std::vector<KPair> NextPair(n);
		std::vector<unsigned> BestFile(n);
		for (unsigned i = 0; i != n; ++i)
		{
			EntriesLeft[i] = (EntryCount > MaxEntries) ? MaxEntries : EntryCount;
			EntryCount -= EntriesLeft[i];
			fp[i] = CreateStdioTmpFile();
			BestFile[i] = i;
			std::vector<KPair> KPairs(EntriesLeft[i]);

			Progress("Reading chunk %u of %u", i + 1, n);
			for (std::vector<KPair>::iterator j = KPairs.begin(); j != KPairs.end(); ++j)
				KF.GetNextEntry(*j);

			Progress("Sorting chunk %u of %u", i + 1, n);
			std::sort(KPairs.begin(), KPairs.end(), KPC);
			NextPair[i] = KPairs[0];

			Progress("Writing chunk %u of %u", i + 1, n);
			for (std::vector<KPair>::const_iterator j = KPairs.begin(); j != KPairs.end(); ++j)
				WriteStdioFile(fp[i], &*j, sizeof(KPair));

			SeekStdioFile(fp[i], sizeof(KPair));
		}
		assert(EntryCount == 0);

		KFilePair SortedKF;
		SortedKF.Create(SortedKFileName, KF.GetGenomes());
		
		VectorIndexCmp<KPair, KPairCompare> MySortFunc(NextPair, KPC);

		std::sort(BestFile.begin(), BestFile.end(), MySortFunc);
		Progress("Merging chunks");
		while (BestFile.size() > 1)
		{
			unsigned idx = BestFile[0];
			SortedKF.AddEntry(NextPair[idx]);
			--EntriesLeft[idx];
			if (EntriesLeft[idx] == 0)
				BestFile.erase(BestFile.begin());
			else
			{
				ReadStdioFile(fp[idx], &NextPair[idx], sizeof(KPair));
				std::sort(BestFile.begin(), BestFile.end(), MySortFunc);
			}
		}
		assert(BestFile.size() == 1);

		unsigned idx = BestFile[0];
		for (;;)
		{
			SortedKF.AddEntry(NextPair[idx]);
			--EntriesLeft[idx];
			if (EntriesLeft[idx] == 0)
				break;
			ReadStdioFile(fp[idx], &NextPair[idx], sizeof(KPair));
		}

		Progress("Finalizing");
		for (unsigned i = 0; i != n; ++i)
			CloseStdioFile(fp[i]);

		SortedKF.Close();
	}
}

void SortKPair(const char *KFileName, bool BySecond)
{
	KFilePair KF;
	Progress("Sorting...");
	KF.Sort<KPairCompare>(KFileName, KPairCompare(BySecond ? 1 : 0));
	Progress("Done");
}

void SortKDecomp(const char *KFileName)
{
	KFileDecomp KF;
	Progress("Sorting...");
	KF.Sort<KDecompCompareByKmer>(KFileName, KDecompCompareByKmer());
	Progress("Done");
}

void SortedToUniq(const char *SortedKFileName, const char *UniqKFileName)
{
	KFilePair InKF, OutKF;

	InKF.Open(SortedKFileName);
	OutKF.Create(UniqKFileName, InKF.GetGenomes());

	ProgressStart("Removing duplicate entries...");
	unsigned EntryCount = InKF.GetEntryCount();
	if (EntryCount > 0)
	{
		KPair PrevKP;
		InKF.GetNextEntry(PrevKP);
		for (unsigned i = 1; i < EntryCount; ++i)
		{
			if ((i & 0xfffff) == 0)
				ProgressStep(i, EntryCount);
			KPair KP;
			InKF.GetNextEntry(KP);
			if ((abs(KP.k[0]) == abs(PrevKP.k[0])) && (abs(KP.k[1]) == abs(PrevKP.k[1])))
			{
				if (KP.score > PrevKP.score)
					PrevKP = KP;
			}
			else
			{
				OutKF.AddEntry(PrevKP);
				PrevKP = KP;
			}
		}
		OutKF.AddEntry(PrevKP);
	}

	ProgressDone();
	OutKF.Close();
	InKF.Close();
}

void RevToDummyAN(const char *RevFileName, const char *ANFileName, unsigned GenomeIndex)
{
	RevFile InFile;
	AlnNodeFile OutFile;

	InFile.Open(RevFileName);
	const GenInfo &GI = InFile.GetGenomeInfo(GenomeIndex);

	std::vector<GenInfo> Genomes;
	Genomes.push_back(GI);

	OutFile.Create(ANFileName, Genomes);

	unsigned MaxChrom = UINT32_MAX;
	IntOpt("maxchrom", &MaxChrom);

	AlnNode AN;
	AN.GKmers.resize(1);
	AN.GKmers[0].G = 0;

	unsigned MaxTid = 0;
	unsigned CurOffset = 0;
	unsigned ChrCount = GI.GetChrCount();
	/* This is broken unless chrid = chrindex, which is always the case for things generated by revolver */
	for (unsigned ChrIndex = 0; ChrIndex != ChrCount; ++ChrIndex)
	{
		unsigned ChrId = GI.GetChrId(ChrIndex);
		unsigned ChrLength = GI.GetChrInfo(ChrId).GetLength();
		if (ChrId >= ChrCount)
			Die("Sparse ChrId index");
		unsigned Diff = ChrLength / TIDLEN;
		if (Diff * TIDLEN < ChrLength)
			++Diff;
		++Diff;
		CurOffset += Diff;
		
		if (ChrId <= MaxChrom)
			MaxTid = CurOffset;
	}

	for (kmer k = 0; k <= (kmer)MaxTid; ++k)
	{
		AN.GKmers[0].k = k;
		OutFile.AddEntry(AN);
	}

	OutFile.Close();
	InFile.Close();
}

void MinCovKP(const char *UniqKPFileName, const char *MinCovKPFileName)
{
	KFilePair InFile, OutFile;

	InFile.Open(UniqKPFileName);
	OutFile.Create(MinCovKPFileName, InFile.GetGenomes());

	ProgressStart("Extracting nodes");
	unsigned EntryCount = InFile.GetEntryCount();
	uint32 MaxScore = 0;
	KPair KP;

	std::vector<unsigned> Nodes[2];
	Nodes[0].reserve(1024);
	Nodes[1].reserve(1024);
	for (unsigned i = 0; i != EntryCount; ++i)
	{
		if ((i & 0xfffff) == 0)
			ProgressStep(i, EntryCount);
		InFile.GetNextEntry(KP);

		Nodes[0].push_back(abs(KP.k[0]));
		Nodes[1].push_back(abs(KP.k[1]));
		if (MaxScore < KP.score)
			MaxScore = KP.score;
	}
	ProgressDone();

	Progress("Allocating graph...");
	std::sort(Nodes[0].begin(), Nodes[0].end());
	std::sort(Nodes[1].begin(), Nodes[1].end());
	Nodes[0].resize(std::unique(Nodes[0].begin(), Nodes[0].end()) - Nodes[0].begin());
	Nodes[1].resize(std::unique(Nodes[1].begin(), Nodes[1].end()) - Nodes[1].begin());

	unsigned LeftNodes = Nodes[0].size();
	unsigned RightNodes = Nodes[1].size();

	BMGraph B(LeftNodes + RightNodes + 2, EntryCount + LeftNodes + RightNodes + 1);

	ProgressStart("Filling edges");
	InFile.Rewind();
	for (unsigned i = 0; i != EntryCount; ++i)
	{
		if ((i & 0xfffff) == 0)
			ProgressStep(i, EntryCount);
		InFile.GetNextEntry(KP);

		int from = VectorSortedLocate<unsigned>(Nodes[0], abs(KP.k[0]));
		int to = VectorSortedLocate<unsigned>(Nodes[1], abs(KP.k[1]));

		B.FillEdge(from + 1, to + LeftNodes + 1, 0, 1, PENALTY_MIN + MaxScore - KP.score);
	}
	ProgressDone();

	B.FillEdge(LeftNodes + RightNodes + 1, 0, 0, LeftNodes + RightNodes, 0);

	for (unsigned i = 1; i <= LeftNodes; ++i)
		B.FillEdge(0, i, 1, RightNodes, 0);

	for (unsigned i = LeftNodes + 1; i <= LeftNodes + RightNodes; ++i)
		B.FillEdge(i, LeftNodes + RightNodes + 1, 1, LeftNodes, 0);

	Progress("Solving...");
	B.Solve();

	node *NodeArray = B.GetNodeArray();
	arc *ArcArray = B.GetArcArray();
	int *CapArray = B.GetCapArray();

	ProgressStart("Writing output");
	InFile.Rewind();
	InFile.GetNextEntry(KP);
	for (unsigned i = 1; i <= LeftNodes; ++i)
	{
		std::vector<unsigned> GoingTo;
		GoingTo.reserve(1000);
		for (arc *a = NodeArray[i].suspended; a != NodeArray[i+1].suspended; ++a)
		{
			unsigned ArcIndex = a - ArcArray;
			if (CapArray[ArcIndex] - a->r_cap > 0)
			{
				GoingTo.push_back(Nodes[1][a->head - NodeArray - LeftNodes - 1]);
				//printf("%u %u or %u %u\n", i, a->head - NodeArray, Nodes[0][i-1], Nodes[1][a->head - NodeArray - LeftNodes - 1]);
			}
		}
		unsigned GoingFrom = Nodes[0][i-1];
		if (GoingTo.size() != 0)
		{
			std::sort(GoingTo.begin(), GoingTo.end());
			for (std::vector<unsigned>::const_iterator j = GoingTo.begin(); j != GoingTo.end(); ++j)
			{
				while ((GoingFrom > (unsigned)abs(KP.k[0])) || ((GoingFrom == (unsigned)abs(KP.k[0])) && (*j > (unsigned)abs(KP.k[1]))))
					InFile.GetNextEntry(KP);
				OutFile.AddEntry(KP);
			}
		}
		if ((i & 0xfffff) == 0)
			ProgressStep(i, LeftNodes);
	}
	ProgressDone();

	InFile.Close();
	OutFile.Close();
}

//TODO confirm
void MinCov(KPairList &Edges, UnsignedSet Nodes_[2])
{
	unsigned EdgeCount = Edges.size();
	if (EdgeCount > 99000000)
	{
		Warning("Ignoring large graph (%u edges)", EdgeCount);
		Nodes_[0].clear();
		Nodes_[1].clear();
		Edges.clear();
		return;
	}
	if (EdgeCount > 1000000)
	{
		Progress("Processing large component of %u edges", EdgeCount);
	}
	// Copy the nodes to the a vector
	std::vector<unsigned> Nodes[2];
	Nodes[0].insert(Nodes[0].begin(), Nodes_[0].begin(), Nodes_[0].end());
	Nodes_[0].clear();
	Nodes[1].insert(Nodes[1].begin(), Nodes_[1].begin(), Nodes_[1].end());
	Nodes_[1].clear();

	// Quickly return for trivial problems
	unsigned LeftNodes = Nodes[0].size();
	unsigned RightNodes = Nodes[1].size();
	if ((LeftNodes == 1) || (RightNodes == 1))
		return;

	if (EdgeCount == 1)
		return;

	//Progress("Solving a problem of (%u,%u) nodes, %u edges\n", LeftNodes, RightNodes, EdgeCount);

	//Progress("DEBUG 1: Allocating graph");
	// Allocate the graph
	BMGraph B(LeftNodes + RightNodes + 2, EdgeCount + LeftNodes + RightNodes + 1);

	//Progress("DEBUG 2: Sorting edges");
	// Sort the edges -- needed later to merge the result
	Edges.sort(KPairCompare(0));

	//Progress("DEBUG 3: Filling edges");
	// Edges
	KPairList::const_iterator j = Edges.begin();
	for (unsigned i = 0; i != EdgeCount; ++i, ++j)
	{
		int from = VectorSortedLocate<unsigned>(Nodes[0], abs(j->k[0]));
		int to = VectorSortedLocate<unsigned>(Nodes[1], abs(j->k[1]));

		B.FillEdge(from + 1, to + LeftNodes + 1, 0, 1, j->score);
	}
	B.FillEdge(LeftNodes + RightNodes + 1, 0, 0, LeftNodes + RightNodes, 0);
	for (unsigned i = 1; i <= LeftNodes; ++i)
		B.FillEdge(0, i, 1, RightNodes, 0);
	for (unsigned i = LeftNodes + 1; i <= LeftNodes + RightNodes; ++i)
		B.FillEdge(i, LeftNodes + RightNodes + 1, 1, LeftNodes, 0);

	//Progress("DEBUG 4: Solving");
	// Solve
	B.Solve();
	//Progress("DEBUG 5: Filtering");

	/*
	 * Return results by filtering the "Edges" list object
	 */
	arc *ArcArray = B.GetArcArray();
	int *CapArray = B.GetCapArray();
	node *NodeArray = B.GetNodeArray();

	// To do that, for each left node the list of right nodes connected to it
	// is kept in a "to" vector
	std::vector<unsigned> to;
	KPairList::iterator li = Edges.begin();
	for (unsigned i = 1; i <= LeftNodes; ++i)
	{
		unsigned from = Nodes[0][i-1];
		to.clear();
		to.reserve(100);
		//Populate "to" vector
		for (arc *a = NodeArray[i].suspended; a != NodeArray[i+1].suspended; ++a)
		{
			unsigned ArcIndex = a - ArcArray;
			if (CapArray[ArcIndex] > a->r_cap)
				to.push_back(Nodes[1][a->head - NodeArray - LeftNodes - 1]);
		}
		//Filter the "Edges" list object according to "to"
		std::sort(to.begin(), to.end());
		for (unsigned j = 0; j != to.size(); ++j)
		{
			while ((from != (unsigned)abs(li->k[0])) || (to[j] != (unsigned)abs(li->k[1])))
				li = Edges.erase(li);
			++li;
		}
	}

	//Finish filtering
	while (li != Edges.end())
		li = Edges.erase(li);
	//Progress("DEBUG 6: Done");
	if (EdgeCount > 1000000)
	{
		Progress("Processing done, edges reduced to %u", Edges.size());
	}
}

//TODO confirm
void GetConnectedEdges(UnsignedSet &Pending, int Side, KPairList &ConnectedEdges, std::vector<bool> &UsedEdges, const KPair *Edges[2], unsigned EdgeCount, UnsignedSet Used[2])
{
	KPair KP;
	KP.score = 0;
	KP.k[0] = 0;
	KP.k[1] = 0;

	KPairPartialCompare KPPC(Side);

	for (UnsignedSet::const_iterator p = Pending.begin(); p != Pending.end(); ++p)
	{
		KP.k[Side] = *p;
		Used[Side].insert(*p);
		std::pair<const KPair *, const KPair *> ip = std::equal_range(Edges[Side], Edges[Side] + EdgeCount, KP, KPPC);

		while (ip.first != ip.second)
		{
			ConnectedEdges.push_back(*(ip.first));
			if (Side == 0)
				UsedEdges[ip.first - Edges[Side]] = true;
			++ip.first;
		}
	}
	Pending.clear();
}

void SuperMinCov(const char *InputFileNameL, const char *InputFileNameR, const char *OutputFileName)
{
	KFilePair InFiles[2];
	KFilePair OutFile;
	
	InFiles[0].Open(InputFileNameL);
	InFiles[1].Open(InputFileNameR);
	OutFile.Create(OutputFileName, InFiles[0].GetGenomes());

	const KPair *Edges[2];

	Edges[0] = InFiles[0].MemoryMapRO();
	Edges[1] = InFiles[1].MemoryMapRO();

	std::vector<bool> UsedEdges;

	unsigned EdgeCount = InFiles[0].GetEntryCount();

	Progress("Allocating used edge vector for %u entries", EdgeCount);
	UsedEdges.resize(EdgeCount, false);

	unsigned NextUnusedEdge = 0;
	unsigned MaxEdgeCount = 0;

	ProgressStart("Processing graph");
	while (NextUnusedEdge < EdgeCount)
	{
		if ((NextUnusedEdge & 0xfffff) == 0)
			ProgressStep(NextUnusedEdge, EdgeCount);

		if (UsedEdges[NextUnusedEdge])
		{
			++NextUnusedEdge;
			continue;
		}

		int Side = 0;
		int OtherSide = 1 - Side;
		UnsignedSet Used[2], Pending[2];
		KPairList ComponentEdges;

		Pending[Side].insert(abs(Edges[Side][NextUnusedEdge].k[Side]));

		while (!Pending[Side].empty())
		{
			KPairList ConnectedEdges;
			GetConnectedEdges(Pending[Side], Side, ConnectedEdges, UsedEdges, Edges, EdgeCount, Used);
			
			KPairList::iterator i = ConnectedEdges.begin();
			while (i != ConnectedEdges.end())
			{
				unsigned OtherNode = abs(i->k[OtherSide]);
				KPairList::iterator j = i;
				++i;
				if (Used[OtherSide].find(OtherNode) == Used[OtherSide].end())
				{
					Pending[OtherSide].insert(OtherNode);
					ComponentEdges.splice(ComponentEdges.end(), ConnectedEdges, j);
				}
				else
					ConnectedEdges.erase(j);
			}
			Side = 1 - Side;
			OtherSide = 1 - OtherSide;
		}
		MinCov(ComponentEdges, Used);
		if (MaxEdgeCount < ComponentEdges.size())
			MaxEdgeCount = ComponentEdges.size();
		KPairList::iterator i = ComponentEdges.begin();
		while (i != ComponentEdges.end())
		{
			OutFile.AddEntry(*i);
			i = ComponentEdges.erase(i);
		}
	}
	ProgressDone();
	Progress("Largest component had %u edges", MaxEdgeCount);

	InFiles[0].Close();
	InFiles[1].Close();
	OutFile.Close();
}

void ANToAllDecomp(const char *ANFileName, const char *DecompFileNamePrefix)
{
	AlnNodeFile InFile;
	InFile.Open(ANFileName);

	std::vector<KFileDecomp> OutFiles(InFile.GetGenomeCount());
	for (unsigned GenomeIndex = 0; GenomeIndex != InFile.GetGenomeCount(); ++GenomeIndex)
	{
		const GenInfo &GI = InFile.GetGenomeInfo(GenomeIndex);

		std::string DecompFileName = DecompFileNamePrefix;

		DecompFileName += ":";
		DecompFileName += GI.GetName();
		DecompFileName += ".kd";

		GenInfoVec Genomes(1, GI);
		OutFiles[GenomeIndex].Create(DecompFileName, Genomes);
	}


	KDecomp KD;
	ProgressStart("Building decompositions...");
	unsigned EntryCount = InFile.GetEntryCount();
	for (KD.idx = 0; KD.idx != EntryCount; ++KD.idx)
	{
		AlnNode AN;
		InFile.GetEntry(AN, KD.idx);
		for (unsigned i = 0; i != AN.GKmers.size(); ++i)
		{
			KD.k = AN.GKmers[i].k;
			OutFiles[AN.GKmers[i].G].AddEntry(KD);
		}
		if ((KD.idx & 0xffff) == 0)
			ProgressStep(KD.idx, EntryCount);
	}

	ProgressDone();
	InFile.Close();

	for (unsigned GenomeIndex = 0; GenomeIndex != InFile.GetGenomeCount(); ++GenomeIndex)
		OutFiles[GenomeIndex].Close();
}


void ANToDecomp(const char *ANFileName, const char *DecompFileName, unsigned GenomeIndex)
{
	AlnNodeFile InFile;
	KFileDecomp OutFile;

	InFile.Open(ANFileName);
	const GenInfo &GI = InFile.GetGenomeInfo(GenomeIndex);

	std::vector<GenInfo> Genomes;
	Genomes.push_back(GI);

	OutFile.Create(DecompFileName, Genomes);

	KDecomp KD;
	ProgressStart("Building decomposition...");
	unsigned EntryCount = InFile.GetEntryCount();
	for (KD.idx = 0; KD.idx != EntryCount; ++KD.idx)
	{
		AlnNode AN;
		InFile.GetEntry(AN, KD.idx);
		if (AN.GetKmerByGenome(GenomeIndex, KD.k))
			OutFile.AddEntry(KD);
		if ((KD.idx & 0xffff) == 0)
			ProgressStep(KD.idx, EntryCount);
	}

	ProgressDone();
	InFile.Close();
	OutFile.Close();
}

void MergeKPDecomp(const char *InputFileName, const char *InputFileName2, const char *OutputFileName)
{
	KFilePair InFile;
	KFileDecomp KFD;
	KFilePair OutFile;

	InFile.Open(InputFileName);
	KFD.Open(InputFileName2);

	if (KFD.GetGenomes().size() != 1)
		Die("Too many genomes in decomp. file");

	const GenInfoVec &Genomes = InFile.GetGenomes();
	unsigned GenomeIndex = VectorUnsortedLocate<GenInfo>(Genomes, KFD.GetGenomeInfo(0));
	OutFile.Create(OutputFileName, Genomes);

	ProgressStart("Merging...");

	unsigned KDCount = KFD.GetEntryCount();
	unsigned KPCount = InFile.GetEntryCount();
	unsigned KDidx = 0, KPidx = 0;

	KPair KP;
	KDecomp KD;
	std::vector<KPair> KPVec;
	std::vector<KDecomp> KDVec;

	if (KPCount > 0)
		InFile.GetNextEntry(KP);

	if (KDCount > 0)
		KFD.GetNextEntry(KD);

	while ((KDidx != KDCount) && (KPidx != KPCount))
	{
		assert(abs(KP.k[GenomeIndex]) >= abs(KD.k));
		KPVec.clear();
		KDVec.clear();

		while (abs(KP.k[GenomeIndex]) > abs(KD.k))
		{
			/* An EOF here is a mistake, so we'll let it die */
			++KDidx;
			KFD.GetNextEntry(KD);
		}

		do
		{
			KPVec.push_back(KP);
			++KPidx;
			if (KPidx == KPCount)
				break;
			InFile.GetNextEntry(KP);
		} while (abs(KP.k[GenomeIndex]) == abs(KPVec[0].k[GenomeIndex]));

		do
		{
			KDVec.push_back(KD);
			++KDidx;
			if (KDidx == KDCount)
				break;
			KFD.GetNextEntry(KD);
		} while (abs(KD.k) == abs(KDVec[0].k));

		for (unsigned i = 0; i != KPVec.size(); ++i)
		{
			KPair NewKP = KPVec[i];
			for (unsigned j = 0; j != KDVec.size(); ++j)
			{
				if ((KPVec[i].k[GenomeIndex] ^ KDVec[j].k) >= 0)
					NewKP.k[GenomeIndex] = KDVec[j].idx;
				else
					NewKP.k[GenomeIndex] = -(signed int)KDVec[j].idx;
				OutFile.AddEntry(NewKP);
			}
		}

		if ((KPidx & 0xffff) == 0)
			ProgressStep(KPidx, KPCount);
	}

	if (KPidx != KPCount)
		Die("Missing entries in KDecomp");

	ProgressDone();

	InFile.Close();
	KFD.Close();
	OutFile.Close();
}

class FileMergeInfo {
	public:

	KPair KP;
	int idx;

	bool operator<(const FileMergeInfo &other) const
	{
		if (abs(KP.k[0]) == abs(other.KP.k[0]))
			return abs(KP.k[1]) < abs(other.KP.k[1]);
		else
			return abs(KP.k[0]) < abs(other.KP.k[0]);
	}
};

void ScoreMergeKP(std::vector<std::string> &InFiles, const char *OutFileName)
{
	std::vector<KFilePair> KFiles(InFiles.size());
	std::vector<uint32> MaxScores(InFiles.size(), 0);
	std::vector<unsigned> Counts(InFiles.size(), 0);

	MinHeap<FileMergeInfo> heap;

	for (unsigned i = 0; i != KFiles.size(); ++i)
	{
		KFiles[i].Open(InFiles[i].c_str());
		Counts[i] = KFiles[i].GetEntryCount();
		if (Counts[i] == 0)
			KFiles[i].Close();
		else
		{
			ProgressStart("Determining maximum score for file #%u", i);
			for (unsigned j = 0; j != Counts[i]; ++j)
			{
				KPair KP;
				KFiles[i].GetNextEntry(KP);
				if (MaxScores[i] < KP.score)
					MaxScores[i] = KP.score;
				if ((j & 0xfffff) == 0)
					ProgressStep(j, Counts[i]);
			}
			ProgressDone();
			KFiles[i].Rewind();
			FileMergeInfo FMI;
			KFiles[i].GetNextEntry(FMI.KP);
			FMI.idx = i;
			heap.insert(FMI);
			--Counts[i];
		}
	}

	uint32 MaxScore = 0;
	for (unsigned i = 0; i != MaxScores.size(); ++i)
		MaxScore += MaxScores[i];

	KFilePair OutFile;
	OutFile.Create(OutFileName, std::vector<GenInfo>());

	Progress("Merging");
	std::vector<KPair> EquivPairs;
	while (!heap.empty())
	{
		EquivPairs.clear();

		do
		{
			FileMergeInfo FMI = heap.min();
			EquivPairs.push_back(FMI.KP);
			if (Counts[FMI.idx] > 0)
			{
				--Counts[FMI.idx];
				KFiles[FMI.idx].GetNextEntry(FMI.KP);
				heap.update_min(FMI);
			}
			else
			{
				KFiles[FMI.idx].Close();
				heap.remove_min();
			}
		} while ((!heap.empty()) && (EquivPairs[0].EqualAbsTo(heap.min().KP)));

		uint32 SignScore[2] = {0,0};

		for (unsigned i = 0; i != EquivPairs.size(); ++i)
			SignScore[((EquivPairs[i].k[0] ^ EquivPairs[i].k[1]) >= 0)] += EquivPairs[i].score;

		bool TargetSign = (SignScore[1] > SignScore[0]);

		if ((SignScore[0] > 0) && (SignScore[1] > 0))
		{
			Warning("Conflicting orientations for (%d,%d); will discard the weakest of scores (%u vs %u)\n", EquivPairs[0].k[0], EquivPairs[0].k[1], SignScore[0], SignScore[1]);

			unsigned i = EquivPairs.size();
			while (i--)
			{
				if (((EquivPairs[i].k[0] ^ EquivPairs[i].k[1]) >= 0) != TargetSign)
					EquivPairs.erase(EquivPairs.begin() + i);
			}
		}

		EquivPairs[0].score = PENALTY_MIN + MaxScore - SignScore[TargetSign];

		OutFile.AddEntry(EquivPairs[0]);
	}

	OutFile.Close();
}

//TODO Confirm
void MergeAN(const char *ANLeftFileName, const char *ANRightFileName, const char *KPLeftFileName, const char *KPRightFileName, const char *ANOutFileName)
{
	AlnNodeFile InFiles[2], OutFile;
	KFilePair KF[2];

	InFiles[0].Open(ANLeftFileName);
	InFiles[1].Open(ANRightFileName);
	KF[0].Open(KPLeftFileName);
	KF[1].Open(KPRightFileName);

	GenInfoVec Genomes(InFiles[0].GetGenomes());
	Genomes.insert(Genomes.end(), InFiles[1].GetGenomes().begin(), InFiles[1].GetGenomes().end());

	OutFile.Create(ANOutFileName, Genomes);

	unsigned LeftNodeCount = InFiles[0].GetEntryCount();
	unsigned RightNodeCount = InFiles[1].GetEntryCount();
	unsigned EdgeCount = KF[0].GetEntryCount();
	unsigned GenomeOffset = InFiles[0].GetGenomes().size();

	ProgressStart("Processing left file...");
	unsigned NextNode = 0;
	if (EdgeCount > 0)
	{
		KPair KP;
		KF[0].GetNextEntry(KP);
		--EdgeCount;
		bool havemore = true;
		do
		{
			while (NextNode != (unsigned)abs(KP.k[0]))
			{
				AlnNode AN;
				InFiles[0].GetEntry(AN, NextNode);
				OutFile.AddEntry(AN);
				++NextNode;
				if ((NextNode & 0xffff) == 0)
					ProgressStep(NextNode, LeftNodeCount);
			}

			AlnNode AN[2];
			InFiles[0].GetEntry(AN[0], NextNode);
			while (NextNode == (unsigned)abs(KP.k[0]))
			{
				InFiles[1].GetEntry(AN[1], abs(KP.k[1]));
				AN[0].GKmers.reserve(AN[0].GKmers.size() + AN[1].GKmers.size());

				if ((KP.k[0] ^ KP.k[1]) >= 0)
				{
					for (unsigned i = 0; i != AN[1].GetGKmerCount(); ++i)
						AN[0].GKmers.push_back(GKmer(AN[1].GKmers[i].G + GenomeOffset, AN[1].GKmers[i].k));
				}
				else
				{
					for (unsigned i = 0; i != AN[1].GetGKmerCount(); ++i)
						AN[0].GKmers.push_back(GKmer(AN[1].GKmers[i].G + GenomeOffset, -(AN[1].GKmers[i].k)));
				}
				OutFile.AddEntry(AN[0]);
				AN[0].GKmers.resize(AN[0].GKmers.size() - AN[1].GKmers.size());
				
				if (EdgeCount > 0)
				{
					KF[0].GetNextEntry(KP);
					--EdgeCount;
				}
				else
				{
					havemore = false;
					break;
				}
			}
			++NextNode;
		} while (havemore);
	}
	while (NextNode != LeftNodeCount)
	{
		AlnNode AN;
		InFiles[0].GetEntry(AN, NextNode);
		OutFile.AddEntry(AN);
		++NextNode;
	}
	ProgressDone();

	ProgressStart("Processing right file...");
	EdgeCount = KF[1].GetEntryCount();
	NextNode = 0;
	if (EdgeCount > 0)
	{
		KPair KP;
		KF[1].GetNextEntry(KP);
		--EdgeCount;
		bool havemore = true;
		do
		{
			while (NextNode != (unsigned)abs(KP.k[1]))
			{
				AlnNode AN;
				InFiles[1].GetEntry(AN, NextNode);
				for (unsigned i = 0; i != AN.GetGKmerCount(); ++i)
					AN.GKmers[i].G += GenomeOffset;
				OutFile.AddEntry(AN);
				++NextNode;
				if ((NextNode & 0xffff) == 0)
					ProgressStep(NextNode, RightNodeCount);
			}

			while (NextNode == (unsigned)abs(KP.k[1]))
			{
				if (EdgeCount > 0)
				{
					KF[1].GetNextEntry(KP);
					--EdgeCount;
				}
				else
				{
					havemore = false;
					break;
				}
			}
			++NextNode;
		} while (havemore);
	}
	while (NextNode != RightNodeCount)
	{
		AlnNode AN;
		InFiles[1].GetEntry(AN, NextNode);
		for (unsigned i = 0; i != AN.GetGKmerCount(); ++i)
			AN.GKmers[i].G += GenomeOffset;
		OutFile.AddEntry(AN);
		++NextNode;
	}
	ProgressDone();
}

void LoadFlist(const char *FlistFileName, std::vector<std::string> &List)
{
	FILE *fp;

	fp = OpenStdioFile(FlistFileName);
	std::string s;

	while(ReadLineStdioFile(fp, s))
		List.push_back(s);

	CloseStdioFile(fp);
}

void AlignAN(const char *ANFileName, const char *FlistFileName, const char *OutFileName)
{
	AlnNodeFile ANF;
	ANF.Open(ANFileName);
	unsigned n = ANF.GetGenomeCount();

	std::vector<std::string> FileList;
	LoadFlist(FlistFileName, FileList);
	if (FileList.size() != n)
		Die("File list requires as many genomes as the alignment file");

	std::vector<RevFile> RF(n);

	for (unsigned i = 0; i < n; ++i)
	{
		RF[i].Open(FileList[i]);
		if (RF[i].GetGenomeInfo(0) != ANF.GetGenomeInfo(i))
			Die("Genome mismatch between files for #%u", i);
	}

	RevFile Out;
	Out.Create(OutFileName, ANF.GetGenomes());
	
	//TODO
	
	/*
	 * Algorithm goes like this:
	 *
	 *
	 *
	 */
	
	for (unsigned i = 0; i < n; ++i)
		RF[i].Close();
	Out.Close();
}

int main(int argc, char *argv[])
{
	ProcessArgVect(argc, argv);
	const char *InputFileName;
	const char *InputFileName2;
	const char *InputFileName3;
	const char *InputFileName4;
	const char *OutputFileName;

	if (ValueOpt("fromrev") && ValueOpt("tokp"))
	{
		InputFileName = ValueOpt("fromrev");
		OutputFileName = ValueOpt("tokp");
		bool FlipGenomes = FlagOpt("flipgenomes");
		RevToKFPair(InputFileName, OutputFileName, FlipGenomes);
	}
	else if (ValueOpt("dumpkp"))
	{
		InputFileName = ValueOpt("dumpkp");
		DumpKPair(InputFileName);
	}
	else if (ValueOpt("dumpdecomp"))
	{
		InputFileName = ValueOpt("dumpdecomp");
		DumpKDecomp(InputFileName);
	}
	else if (ValueOpt("dumpan"))
	{
		unsigned MinSize = 0;
		InputFileName = ValueOpt("dumpan");
		IntOpt("minsize", &MinSize);
		DumpANFile(InputFileName, MinSize);
	}
	else if (ValueOpt("sortkp"))
	{
		InputFileName = ValueOpt("sortkp");
		bool BySecond = FlagOpt("bysecond");
		SortKPair(InputFileName, BySecond);
	}
	else if (ValueOpt("sortdecomp"))
	{
		InputFileName = ValueOpt("sortdecomp");
		SortKDecomp(InputFileName);
	}
	else if (ValueOpt("fromkp") && ValueOpt("tosortedkp"))
	{
		InputFileName = ValueOpt("fromkp");
		OutputFileName = ValueOpt("tosortedkp");
		bool BySecond = FlagOpt("bysecond");
		unsigned MaxEntries = 90000000;
		IntOpt("maxentries", &MaxEntries);
		ToSortedKP(InputFileName, OutputFileName, BySecond, MaxEntries);
	}
	else if (ValueOpt("froman") && ValueOpt("tosortedan"))
	{
		InputFileName = ValueOpt("froman");
		OutputFileName = ValueOpt("tosortedan");
		unsigned MaxEntries = 90000000;
		IntOpt("maxentries", &MaxEntries);
		ToSortedAN(InputFileName, OutputFileName, MaxEntries);
	}
	else if (ValueOpt("fromsortedkp") && ValueOpt("touniqkp"))
	{
		InputFileName = ValueOpt("fromsortedkp");
		OutputFileName = ValueOpt("touniqkp");
		SortedToUniq(InputFileName, OutputFileName);
	}
	else if (ValueOpt("fromrev") && ValueOpt("todummyan"))
	{
		InputFileName = ValueOpt("fromrev");
		OutputFileName = ValueOpt("todummyan");
		unsigned GenomeIndex = 0;
		IntOpt("genomeindex", &GenomeIndex);
		RevToDummyAN(InputFileName, OutputFileName, GenomeIndex);
	}
	else if (ValueOpt("fromsorteduniqkp") && ValueOpt("tomincovkp"))
	{
		InputFileName = ValueOpt("fromsorteduniqkp");
		OutputFileName = ValueOpt("tomincovkp");
		MinCovKP(InputFileName, OutputFileName);
	}
	else if (ValueOpt("froman") && ValueOpt("todecomp"))
	{
		InputFileName = ValueOpt("froman");
		OutputFileName = ValueOpt("todecomp");
		unsigned GenomeIndex = 0;
		IntOpt("genomeindex", &GenomeIndex);
		ANToDecomp(InputFileName, OutputFileName, GenomeIndex);
	}
	else if (ValueOpt("froman") && ValueOpt("toalldecomp"))
	{
		InputFileName = ValueOpt("froman");
		OutputFileName = ValueOpt("toalldecomp");
		ANToAllDecomp(InputFileName, OutputFileName);
	}
	else if (ValueOpt("mergesortedkp") && ValueOpt("withsorteddecomp") && ValueOpt("tokp"))
	{
		InputFileName = ValueOpt("mergesortedkp");
		InputFileName2 = ValueOpt("withsorteddecomp");
		OutputFileName = ValueOpt("tokp");
		MergeKPDecomp(InputFileName, InputFileName2, OutputFileName);
	}
	else if (ValueOpt("scoremergesortedkp") && ValueOpt("tokp"))
	{
		OutputFileName = ValueOpt("tokp");
		std::vector<std::string> InFiles;
		ParseStrVec(ValueOpt("scoremergesortedkp"), InFiles, ',');
		if (InFiles.size() == 0)
			Die("Invalid file list");
		ScoreMergeKP(InFiles, OutputFileName);
	}
	else if (ValueOpt("fromscoredkpL") && ValueOpt("andscoredkpR") && ValueOpt("tomincovkp"))
	{
		InputFileName = ValueOpt("fromscoredkpL");
		InputFileName2 = ValueOpt("andscoredkpR");
		OutputFileName = ValueOpt("tomincovkp");
		SuperMinCov(InputFileName, InputFileName2, OutputFileName);
	}
	else if (ValueOpt("fromanL") && ValueOpt("andanR") && ValueOpt("andkpL") && ValueOpt("andkpR") && ValueOpt("toan"))
	{
		InputFileName = ValueOpt("fromanL");
		InputFileName2 = ValueOpt("andanR");
		InputFileName3 = ValueOpt("andkpL");
		InputFileName4 = ValueOpt("andkpR");
		OutputFileName = ValueOpt("toan");
		MergeAN(InputFileName, InputFileName2, InputFileName3, InputFileName4, OutputFileName);
	}
	else if (ValueOpt("verifygks"))
	{
		InputFileName = ValueOpt("verifygks");
		VerifyGKS(InputFileName);
	}
	else if (ValueOpt("findlimits"))
	{
		InputFileName = ValueOpt("findlimits");
		FindLimits(InputFileName);
	}
	else if (ValueOpt("findlimits2"))
	{
		InputFileName = ValueOpt("findlimits2");
		FindLimits2(InputFileName);
	}
	else if (ValueOpt("alignan") && ValueOpt("flist") && ValueOpt("out"))
	{
		InputFileName = ValueOpt("alignan");
		InputFileName2 = ValueOpt("flist");
		OutputFileName = ValueOpt("out");
		AlignAN(InputFileName, InputFileName2, OutputFileName);
	}
	else
	{
		Usage();
	}

	return 0;
}

