#include "multix.h"

#define TRACE					0
#define TRACE_START_END_COL		0

/***
Given an alignment and an interval, return the subalignment
defined by the interval, which overlaps one of the segments in
the alignment. The interval is defined by (genome, chromosome,
lo, hi).

    ----------------------- 
    -----------------------  Input alignment (AIX.AI + Rows)
    ----------------------- 

       ^       ^ 
       |       |
        -------              Interval (Genome,Chr,lo,hi)

           |
           |
           v

       ---------            
       ---------             Sub-alignment (SubAIX.AI + SubRows)
       ---------
***/

static void GetStartEndColIndexes(coord_t lo, coord_t hi, const GCSeg &Seg,
  const StdStr &Row, unsigned &StartColIndex, unsigned &EndColIndex)
	{
	StartColIndex = UINT_MAX;
	EndColIndex = UINT_MAX;
	coord_t Pos = COORD_T_MAX;
	const unsigned ColCount = SIZE(Row);
	for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
		{
		char c = Row[ColIndex];
		if (c != '-')
			{
			if (Pos == COORD_T_MAX)
				Pos = Seg.lo;
			else
				++Pos;
			}
#if	TRACE_START_END_COL
		Log("Col %u Char %c Pos %u lo %u hi %u\n",
		  ColIndex, c, (unsigned) Pos, (unsigned) lo, (unsigned) hi);
#endif
		if (Pos == lo && StartColIndex == UINT_MAX)
			StartColIndex = ColIndex;

		if (Pos == hi)
			{
			EndColIndex = ColIndex;
			break;
			}
		}

	if (StartColIndex == UINT_MAX || EndColIndex == UINT_MAX)
		Die("GetStartEndColIndexes, start / end col not found");
	}

void MakeSubAln(const AlnInfoX &AIX, unsigned GenomeIndex, unsigned ChrId, coord_t lo,
  coord_t hi, const StdStrVec &Rows, AlnInfoX &SubAIX, StdStrVec &SubRows)
	{
#if	TRACE
	Log("MakeSubAln(Genome=%u Chr=%u lo=%u hi=%u)\n",
	  GenomeIndex, ChrId, (unsigned) lo, (unsigned) hi);
	Log("Alignment:\n");
	LogAIRows(AIX.AI, Rows);
#endif
	asserta(hi >= lo);

	const unsigned SegCount = AIX.AI.GetSegCount();
	const unsigned RowCount = (unsigned) Rows.size();

	asserta(SegCount == RowCount);
	asserta(SegCount > 0);

	SubAIX = AIX;
	SubAIX.AI.Segs.clear();
	SubRows.clear();

	unsigned AILength = AIX.AI.Segs[0].GetLength();
	asserta(AILength > 0);

	unsigned IntervalLength = (unsigned) (hi - lo);
	float LengthRatio = (float) IntervalLength / (float) AILength;

	SubAIX.AI.Score = AIX.AI.Score*LengthRatio;

	unsigned OverlappingSegIndex = GetOverlappingSegIndex(AIX.AI, GenomeIndex,
	  ChrId, lo, hi);
	asserta(OverlappingSegIndex != UINT_MAX);

	const GCSegVec &Segs = AIX.AI.Segs;
	const GCSeg &OverlappingSeg = Segs[OverlappingSegIndex];

	coord_t SubSegLo = std::max(lo, OverlappingSeg.lo);
	coord_t SubSegHi = std::min(hi, OverlappingSeg.hi);
	asserta(SubSegLo <= SubSegHi);

#if	TRACE
	Log("Overlapping seg %u SubSegLo=%u SubSegHi=%u\n",
	  OverlappingSegIndex,
	  (unsigned) SubSegLo,
	  (unsigned) SubSegHi);
#endif

	//const unsigned ColCount = (unsigned) Rows[0].size();
	//coord_t Pos = COORD_T_MAX;
	const StdStr &OverlappingRow = Rows[OverlappingSegIndex];

	unsigned StartColIndex = UINT_MAX;
	unsigned EndColIndex = UINT_MAX;
	GetStartEndColIndexes(SubSegLo, SubSegHi, OverlappingSeg, OverlappingRow,
	  StartColIndex, EndColIndex);

#if	TRACE
	Log("StartCol %u EndCol %u\n", StartColIndex, EndColIndex);
#endif

	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &Seg = Segs[SegIndex];
		const StdStr &Row = Rows[SegIndex];

#if TRACE
		Log("Seg %u", SegIndex);
		LogGCSeg(Seg);
		Log(" Row %.64s", Row.c_str());
		if (SIZE(Row) > 64)
			Log("...");
		Log("\n");
		Log("StartCol %u EndCol %u\n", StartColIndex, EndColIndex);
#endif

		GCSeg SubSeg;
		StdStr SubRow;
		bool NonEmpty = MakeSubSeg(Seg, Row, SubSeg, SubRow,
		  StartColIndex, EndColIndex);
		if (!NonEmpty)
			continue;

		SubAIX.AI.Segs.push_back(SubSeg);
		SubRows.push_back(SubRow);
		}

#if	TRACE
	Log("\n");
	Log("End of MakeSubAln, sub-alignment:\n");
	LogAIRows(SubAIX.AI, SubRows);
#endif
	}
