#include "utils.h"
#include "xlat.h"
#include "masktab.h"
#include <ctype.h>

#define TRACE	0

extern bool g_HardMask;
extern bool g_Overwrite;
extern std::vector<bool> g_Pattern;

static void MaskWordNucleo(std::vector<unsigned char> &Seq, unsigned Pos,
  unsigned MerSize)
	{
	assert(Pos + MerSize - 1 < (unsigned) Seq.size());
	assert(MerSize == (unsigned) g_Pattern.size());

	for (unsigned i = 0; i < MerSize; ++i)
		{
		if (g_Pattern[i])
			{
			unsigned SeqPos = Pos + i;
			if (g_HardMask)
				Seq[SeqPos] = 'N';
			else
				{
				unsigned char c = Seq[SeqPos];
				Seq[SeqPos] = tolower(c);
				}
			}
		}
	}

static void MaskWordAmino(std::vector<unsigned char> &Seq, unsigned Pos,
  unsigned MerSize)
	{
	assert(Pos + MerSize - 1 < (unsigned) Seq.size());
	assert(MerSize == (unsigned) g_Pattern.size());

	for (unsigned i = 0; i < MerSize; ++i)
		{
		if (g_Pattern[i])
			{
			unsigned SeqPos = Pos + i;
			if (g_HardMask)
				Seq[SeqPos] = 'X';
			else
				{
				unsigned char c = Seq[SeqPos];
				Seq[SeqPos] = tolower(c);
				}
			}
		}
	}

static bool DoMaskWordNucleoNucleo(const std::vector<unsigned char> &Seq, unsigned Pos,
  unsigned MerSize, const MaskTable &MT)
	{
	assert(MT.GetAlpha() == ALPHA_Nucleo);
	assert(Pos + MerSize - 1 < (unsigned) Seq.size());

	char StrWord[128];
	for (unsigned i = 0; i < MerSize; ++i)
		{
		unsigned char c = Seq[Pos+i];
		if (!isacgt(c))
			return false;
		StrWord[i] = c;
		}
	StrWord[MerSize] = 0;

	uint64 Word = NucStrToWord(StrWord, MerSize);
	bool Bit = MT.GetBit(Word);
	return Bit;
	}

static bool DoMaskWordNucleoAmino(const std::vector<unsigned char> &Seq, unsigned Pos,
  unsigned MerSize, const MaskTable &MT)
	{
	assert(Pos + MerSize - 1 < (unsigned) Seq.size());
	assert(MT.GetAlpha() == ALPHA_Amino);
	assert(MerSize%3 == 0);

	char StrWord[128];
	for (unsigned i = 0; i < MerSize; ++i)
		{
		unsigned char c = Seq[Pos+i];
		if (!isacgt(c))
			return false;
		StrWord[i] = c;
		}
	StrWord[MerSize] = 0;

	uint64 Word = NucStrToWord(StrWord, MerSize);
	uint64 RCWord = RevCompNucWord(Word, MerSize);

	uint64 AminoWord;
	uint64 RCAminoWord;
	bool Ok = NucWordToAminoWord(Word, MerSize/3, &AminoWord);
	bool OkRC = NucWordToAminoWord(RCWord, MerSize/3, &RCAminoWord);
#if	TRACE
	Log("Pos=%u", (unsigned) Pos);
	Log(" Word=%x=%s", Word, NucWordToStr(Word, MerSize));
	Log(" RCWord=%x=%s", RCWord, NucWordToStr(RCWord, MerSize));
	Log(" Amino=%x=%s", AminoWord, AminoWordToStr(AminoWord, MerSize/3));
	Log(" RCAmino=%x=%s", RCAminoWord, AminoWordToStr(RCAminoWord, MerSize/3));
	Log(" Mask=%c", (Ok && MT.GetBit(AminoWord)) ? 'T' : 'F');
	Log(" %c\n", (OkRC && MT.GetBit(RCAminoWord)) ? 'T' : 'F');
#endif
	bool DoMask = (Ok && MT.GetBit(AminoWord)) || (OkRC && MT.GetBit(RCAminoWord));
	return DoMask;
	}

static bool DoMaskWordAmino(const std::vector<unsigned char> &Seq, unsigned Pos,
  unsigned MerSize, const MaskTable &MT)
	{
	assert(Pos + MerSize - 1 < (unsigned) Seq.size());
	assert(MT.GetAlpha() == ALPHA_Amino);

	char StrWord[128];
	for (unsigned i = 0; i < MerSize; ++i)
		{
		unsigned char c = Seq[Pos+i];
		if (!isamino(c))
			return false;
		StrWord[i] = c;
		}

	StrWord[MerSize] = 0;
	uint64 Word = AminoStrToWord(StrWord, MerSize);
	bool Bit = MT.GetBit(Word);
	return Bit;
	}

static void MaskSeqNucleoNucleo(std::vector<unsigned char> &Seq, const MaskTable &MT)
	{
	const unsigned MerSize = MT.GetWordSize();
	const unsigned SeqLength = (unsigned) Seq.size();
	if (SeqLength != Seq.size())
		Die("MakeSeq, seq size overflow");
	if (SeqLength < MerSize)
		return;

	const unsigned MerCount = SeqLength - MerSize + 1;
	std::vector<bool> DoMask;
	DoMask.resize(MerCount);

	for (unsigned i = 0; i < MerCount; ++i)
		DoMask[i] = DoMaskWordNucleoNucleo(Seq, i, MerSize, MT);

	for (unsigned i = 0; i < MerCount; ++i)
		if (DoMask[i])
			MaskWordNucleo(Seq, i, MerSize);
	}

static void MaskSeqNucleoAmino(std::vector<unsigned char> &Seq, const MaskTable &MT)
	{
	const unsigned MerSize = MT.GetWordSize()*3;
	const unsigned SeqLength = (unsigned) Seq.size();
	if (SeqLength != Seq.size())
		Die("MakeSeq, seq size overflow");
	if (SeqLength < MerSize)
		return;

	const unsigned MerCount = SeqLength - MerSize + 1;
	std::vector<bool> DoMask;
	DoMask.resize(MerCount);

	for (unsigned i = 0; i < MerCount; ++i)
		DoMask[i] = DoMaskWordNucleoAmino(Seq, i, MerSize, MT);

	for (unsigned i = 0; i < MerCount; ++i)
		if (DoMask[i])
			MaskWordNucleo(Seq, i, MerSize);
	}

static void MaskSeqAmino(std::vector<unsigned char> &Seq, const MaskTable &MT)
	{
	ALPHA TabAlpha = MT.GetAlpha();
	if (TabAlpha == ALPHA_Nucleo)
		Die("Can't mask amino sequence with nucleotide table");
	const unsigned MerSize = MT.GetWordSize();

	const unsigned SeqLength = (unsigned) Seq.size();
	if (SeqLength != Seq.size())
		Die("MakeSeq, seq size overflow");
	if (SeqLength < MerSize)
		return;

	const unsigned MerCount = SeqLength - MerSize + 1;
	std::vector<bool> DoMask;
	DoMask.resize(MerCount);

	for (unsigned i = 0; i < MerCount; ++i)
		DoMask[i] = DoMaskWordAmino(Seq, i, MerSize, MT);

	for (unsigned i = 0; i < MerCount; ++i)
		if (DoMask[i])
			MaskWordAmino(Seq, i, MerSize);
	}

static void MaskSeqNucleo(std::vector<unsigned char> &Seq, const MaskTable &MT)
	{
	ALPHA TabAlpha = MT.GetAlpha();
	if (TabAlpha == ALPHA_Nucleo)
		MaskSeqNucleoNucleo(Seq, MT);
	else if (TabAlpha == ALPHA_Amino)
		MaskSeqNucleoAmino(Seq, MT);
	else
		Die("Cannot mask nucleo seq with tab alpha %d", TabAlpha);
	}

void MaskSeq(std::vector<unsigned char> &Seq, ALPHA SeqAlpha, const MaskTable &MT)
	{
	unsigned SeqLength = (unsigned) Seq.size();
	if (SeqLength != Seq.size())
		Die("MakeSeq, seq size overflow");

	if (g_Overwrite)
		{
		const unsigned SeqLength = (unsigned) Seq.size();
		for (unsigned i = 0; i < SeqLength; ++i)
			{
			unsigned char c = Seq[i];
			Seq[i] = toupper(c);
			}
		}

	ALPHA TabAlpha = MT.GetAlpha();

	if (SeqAlpha == ALPHA_Nucleo && TabAlpha == ALPHA_Nucleo)
		MaskSeqNucleo(Seq, MT);
	else if (SeqAlpha == ALPHA_Amino)
		MaskSeqAmino(Seq, MT);
	else
		Die("Invalid SeqAlpha %d", SeqAlpha);
	}
