#ifndef MINHEAP_H
#define MINHEAP_H

template <typename T, typename CompareFunc = std::less<T> >
class MinHeap
{

public:
	MinHeap() {}
	MinHeap(CompareFunc cf) : g_Cmp(cf) {}

	bool empty() const { return g_Heap.empty(); }
	size_t size() const { return g_Heap.size(); }

	void insert(const T &value);

	void reserve(size_t count) { g_Heap.reserve(count); };

	void update(int index, const T &value)
	{
		if (g_Cmp(value, g_Heap[index]))
			decrease(index, value);
		else
			increase(index, value);
	}

	void update_min(const T &value)
	{
		update(0, value);
	}

	const T &min() const { return g_Heap[0]; }
	T &min() { return g_Heap[0]; }

	void remove(int index);

	void remove_min() { remove(0); }

	void clear() { g_Heap.clear(); }

private:
	CompareFunc g_Cmp;

	std::vector<T> g_Heap;

	void decrease(int index, const T &new_value);
	void increase(int index, const T &new_value);
	int fix_up(int pos, const T &value);
	int fix_down(int pos, int HeapSize, const T &value);
};

template <typename T, typename CompareFunc>
int MinHeap<T, CompareFunc>
::fix_up(int pos, const T &value)
{
	while (pos > 0)
	{
		int parent = (pos - 1) >> 1;
		if (g_Cmp(value, g_Heap[parent]))
		{
			g_Heap[pos] = g_Heap[parent];
		}
		else
			break;
		pos = parent;
	}
	return pos;
}

template <typename T, typename CompareFunc>
int MinHeap<T, CompareFunc>
::fix_down(int pos, int HeapSize, const T &value)
{
	while (true)
	{
		int child = (pos << 1) + 1;
		if ((child + 1) < HeapSize)
		{
			if (!g_Cmp(g_Heap[child], value) &&
					!g_Cmp(g_Heap[child + 1], value))
				break;
			if (g_Cmp(g_Heap[child + 1], g_Heap[child]))
				child = child + 1;
		}
		else if (child < HeapSize)
		{
			if (!g_Cmp(g_Heap[child], value))
				break;
		}
		else
			break;
		g_Heap[pos] = g_Heap[child];
		pos = child;
	}
	return pos;
}

template <typename T, typename CompareFunc>
void MinHeap<T, CompareFunc>
::insert(const T &value)
{
	int pos = g_Heap.size();
	g_Heap.resize(pos + 1);

	pos = fix_up(pos, value);
	g_Heap[pos] = value;
}

template <typename T, typename CompareFunc>
void MinHeap<T, CompareFunc>
::decrease(int index, const T &new_value)
{
	int pos = fix_up(index, new_value);
	g_Heap[pos] = new_value;
	index = pos;
}

template <typename T, typename CompareFunc>
void MinHeap<T, CompareFunc>
::increase(int index, const T &new_value)
{
	int pos = fix_down(index, g_Heap.size(), new_value);
	g_Heap[pos] = new_value;
	index = pos;
}

template <typename T, typename CompareFunc>
void MinHeap<T, CompareFunc>
::remove(int index)
{
	int new_size = g_Heap.size() - 1;
	if (new_size > 0)
	{
		int pos = fix_down(index, new_size, g_Heap.back());
		g_Heap[pos] = g_Heap.back();
	}

	g_Heap.pop_back();
}

#endif
