#include "progress.h"
#include <algorithm>
#include <vector>
#include <list>
#include "mkedgesai.h"

#define	TRACE	0

/***
MakeEdges is an O(N log N) function that creates a graph from a set
of AlnInfo's (AIs) given a genome index and a Link function that
determines whether two AIs should be joined by an edge. The Link
function must have a maximum distance D such that if two AIs are
separated by more than D as projected onto the genome, then there can
be no edge between the AIs.

Link determines directed edges, i.e. Link(AI1, AI2) may be different
from Link(AI2, AI1). MakeEdges will avoid trivial cycles, i.e. if
Link(AI1, AI2) is true then no edge will be added for AI2->AI1 even
if Link(AI2, AI1) is true.

The algorithm sorts bendpoints (beginning and endpoints) of the input
AIs w.r.t. the genome and then slides a window of length D along that
genome. Each time a the left-hand edge of the window crosses a bendpoint,
the AI for that bendpoint is compared with all bendpoints in the window
using the Link function. This ensures that all pairs of AIs separated
by distances <= D are compared using Link.
***/

static IntVecVec *g_Edges;
static const AIVec *g_AIs;
static unsigned g_AICount;
static unsigned g_EdgesFound;

static inline const AlnInfo &GetAI(unsigned AIIndex)
	{
	assert(AIIndex < g_AICount);
	return (*g_AIs)[AIIndex];
	}

static void AddEdge(unsigned AIIndex1, unsigned AIIndex2)
	{
	assert(AIIndex1 < g_AICount);
	assert(AIIndex2 < g_AICount);
	(*g_Edges)[AIIndex1].push_back(AIIndex2);
	++g_EdgesFound;
	}

static void MakeEdgesFromWindow(const AIBPList &BPs, ptrLinkFunction Link)
	{
	unsigned BPCount = (unsigned) BPs.size();
	if (BPCount < 2)
		return;

// Get first entry
	const AIBP &FirstBendPoint = BPs.front();
	const unsigned FirstBendPointAIIndex = FirstBendPoint.AlnInfoIndex;
	assert(FirstBendPointAIIndex < g_AICount);
	const AlnInfo &FirstAI = GetAI(FirstBendPointAIIndex);

// Iterate through list, starting with 2nd entry
	AIBPList::const_iterator ptrBP = BPs.begin();
	for (++ptrBP; ptrBP != BPs.end(); ++ptrBP)
		{
		const AIBP &BP = *ptrBP;
		const unsigned AIIndex = BP.AlnInfoIndex;
		const AlnInfo &AI = GetAI(AIIndex);

		if (AIIndex == FirstBendPointAIIndex)
			continue;

		if (Link(FirstAI, AI))
			AddEdge(FirstBendPointAIIndex, AIIndex);
		else if (Link(AI, FirstAI))
			AddEdge(AIIndex, FirstBendPointAIIndex);
		}
	}

// Return width (in genome coordinates) of BendPoints.
static unsigned GetWindowWidth(const AIBPList &BPs)
	{
	const unsigned BPCount = (unsigned) BPs.size();
	if (0 == BPCount)
		return 0;

	const AIBP &FirstBP = BPs.front();
	const AIBP &LastBP = BPs.back();

	coord_t FirstPos = FirstBP.Pos;
	coord_t LastPos = LastBP.Pos;

	if (FirstPos > LastPos)
		Die("GetWindowWidth: BPs not sorted");

	return (unsigned) (LastPos - FirstPos);
	}

void MakeEdges(const AIVec &AIs, ptrLinkFunction Link, unsigned GenomeIndex,
  unsigned MaxDist, IntVecVec &Edges, bool ShowProgress)
	{
#if	TRACE
	Log("MakeEdges, AIs=\n");
	LogAIs(AIs);
#endif

	g_Edges = &Edges;
	g_AIs = &AIs;
	g_AICount = (unsigned) AIs.size();
	g_EdgesFound = 0;

	AIBPVec BPs;
	GetBendpoints(AIs, GenomeIndex, BPs, 0);

	std::sort(BPs.begin(), BPs.end());

#if	TRACE
	Log("\n");
	Log("Sorted BPs:\n");
	void LogAIBPs(const AIBPVec &BPs);
	LogAIBPs(BPs);
#endif

	const unsigned BPCount = (unsigned) BPs.size();
	assert(BPCount == g_AICount*2);

	Edges.clear();
	Edges.resize(g_AICount);

	if (ShowProgress)
		ProgressStart("Making edges for %u hits", g_AICount);

	AIBPList BPList;
	unsigned Counter = 0;
	for (AIBPVec::const_iterator p = BPs.begin(); p != BPs.end(); ++p)
		{
		if (ShowProgress && ++Counter%1000 == 0)
			ProgressStep(Counter, BPCount);

		const AIBP &BP = *p;
		BPList.push_back(BP);
		while (BPList.size() > 0 && GetWindowWidth(BPList) >= MaxDist)
			{
			MakeEdgesFromWindow(BPList, Link);
			BPList.pop_front();
			}
		}
	
// Flush at end of loop
	while (BPList.size() > 0)
		{
		MakeEdgesFromWindow(BPList, Link);
		BPList.pop_front();
		}

	if (ShowProgress)
		{
		ProgressDone();
		Progress("%u edges found", g_EdgesFound);
		}
	}

void MakeEdges(const AIXVec &AIXs, ptrLinkFunction Link, unsigned GenomeIndex,
  unsigned MaxDist, IntVecVec &Edges, bool ShowProgress)
	{
	AIVec AIs;
	AIXsToAIs(AIXs, AIs);
	MakeEdges(AIs, Link, GenomeIndex, MaxDist, Edges, ShowProgress);
	}
