#include "revfile.h"
#include "utils.h"
#include "filter4.h"
#include "segbagai.h"
#include <algorithm>
#include <math.h>

#define TRACE		0
#define FIG			0

/***
Seek an "optimal coverage". This is a set of AIs that attempt to
cover both genomes in a given region, minimizing overlaps (repeats).

The algorithm is greedy. In each pass, AIs are considered in order
of decreasing "strength". Strength is a relative measure of how good
the contribution of this AI will be to the coverage; this varies in
each pass as explained below.

The relation A covers B means that all the bases in B are found in A,
allowing for some small slop. Examples:

A covers B:
	________________ A
	  __________B


A covers B with allowable slop:
	________________ A
	       __________B


A does not cover B:
	________________ A
	             __________B


A set of gaps is maintained as AIs are added to the set. Initially
the gap is the entire region. Typically, when an AI is added it
will have the effect of deleting a gap and inserting two new gaps.

Before:
	Covered           Gap             Covered
	++++++++++---------------------++++++++++


After:
                      AI
                  ___________
	Covered                           Covered
	++++++++++----+++++++++++------++++++++++
	          Gap               Gap

Pass 1. An AI is added to the set iff it is covered by gaps
when projected onto both genomes. Note that if an AI is covered
by a gap, it cannot overlap another AI. Strength is the length
of the AI.

Pass 2. An AI is added iff it is covered by gaps when projected
onto one genome. This minimizes overlaps with respect to one
genome. Strength is the amount of gap covered minus the overlap
with both genomes. The strength changes dynamically as AIs are
added during the pass -- this is not taken into account; for
efficiency, AIs are sorted once and considered in that order.

Pass 3. An AI is added iff it covers a gap. The overlap must not
exceed a resonable fraction of the AI, say 50%, otherwise you
get long AIs covering short gaps.
***/

unsigned g_MinGapLength = 32;
double g_MinCovFract = 0.8;
double g_MinCovFract2 = 0.5;

static unsigned g_GenomeIndexA;
static unsigned g_GenomeIndexB;

void LogIntVec(const IntVec &v)
	{
	unsigned n = (unsigned) v.size();
	for (unsigned i = 0; i < n; ++i)
		Log(" %u", v[i]);
	}

typedef std::vector<int> SignedIntVec;

// Globals needed by Strength<n> sort routines.
static const AIVec *g_AIs;
static SignedIntVec *g_Strength;
static SegBag *g_GapsA;
static SegBag *g_GapsB;

coord_t GetAILength(const AlnInfo &AI)
	{
	const GCSeg &SegA = AI.GetSeg(g_GenomeIndexA);
	const GCSeg &SegB = AI.GetSeg(g_GenomeIndexB);

	double LenA = (double) SegA.GetLength();
	double LenB = (double) SegB.GetLength();

	return (coord_t) sqrt(LenA*LenA + LenB*LenB);
	}

static int Strength1(const AlnInfo &AI)
	{
	return (int) GetAILength(AI);
	}

//	Strength2 =
//		(Amount of gap covered) - (overlap with both genomes)
static int Strength2(const AlnInfo &AI)
	{
	const GCSeg &SegA = AI.GetSeg(g_GenomeIndexA);
	const GCSeg &SegB = AI.GetSeg(g_GenomeIndexB);

	coord_t GapCovA = g_GapsA->Coverage(SegA);
	coord_t GapCovB = g_GapsB->Coverage(SegB);

	coord_t LenA = SegA.GetLength();
	coord_t LenB = SegB.GetLength();

	assert(LenA >= GapCovA);
	assert(LenB >= GapCovB);

	coord_t OvA = LenA - GapCovA;
	coord_t OvB = LenB - GapCovB;

	int SA = (int) GapCovA - (int) OvA - (int) OvB;
	int SB = (int) GapCovB - (int) OvA - (int) OvB;
	return std::max(SA, SB);
	}

// Sort in DECREASING order
static bool Comp(unsigned i, unsigned j)
	{
	return (*g_Strength)[i] > (*g_Strength)[j];
	}

bool Covers(const GCSeg &SA, const GCSeg &SB, double MinFract)
	{
	double SBLen = SB.GetLength();
	double Ov = SA.GetOverlap(SB);
	double Ratio = Ov/SBLen;
	bool c = Ratio >= MinFract;
#if	TRACE
	Log("Covers(%u,%u %u-%u %c, %u,%u %u-%u %c, MinFract=%.2f)\n",
		SA.GenomeIndex, SA.ChrId, (unsigned) SA.lo, (unsigned) SA.hi, SA.Plus ? '+' : '-',
		SB.GenomeIndex, SB.ChrId, (unsigned) SB.lo, (unsigned) SB.hi, SB.Plus ? '+' : '-',
		MinFract);
	Log("  SBLen=%g Ov=%g Ratio=%g Covers=%d\n", SBLen, Ov, Ratio, c);
#endif
	return c;
	}

bool Covers(const GCSeg &SA, const GCSeg &SB, double MinAFract, double MinBFract)
	{
	double SALen = SA.GetLength();
	double SBLen = SB.GetLength();
	double Ov = SA.GetOverlap(SB);
	bool ret = Ov/SALen >= MinAFract && Ov/SBLen >= MinBFract;
#if	TRACE
	Log("Covers(%u,%u %u-%u %c, %u,%u %u-%u, %.2f, %.2f)\n",
		SA.GenomeIndex, SA.ChrId, (unsigned) SA.lo, (unsigned) SA.hi, SA.Plus ? '+' : '-',
		SB.GenomeIndex, SB.ChrId, (unsigned) SB.lo, (unsigned) SB.hi, SB.Plus ? '+' : '-',
		MinAFract, MinBFract);
	Log("Ov=%g   Ov/SALen=%g >= %g   Ov/SBLen=%g >= %g ret=%d\n",
	  Ov, Ov/SALen, MinAFract, Ov/SBLen, MinBFract, ret);
#endif
	return ret;
	}

static bool CoveredByOneGap(const GCSeg &s, const SegBag &Gaps, double MinFract)
	{
	coord_t GapPos;
	bool NearestSegFound = Gaps.GetNearestSeg(s.GetMidpoint(), &GapPos);
#if	TRACE
	Log("CoveredByOneGap(s=%u,%u %u-%u, MinFract=%g) NearestSegFound=%d\n",
	  s.GenomeIndex, s.ChrId, (unsigned) s.lo, (unsigned) s.hi, MinFract, NearestSegFound);
#endif
	if (!NearestSegFound)
		return false;
	GCSeg Gap;
	bool GapFound = Gaps.GetSeg(GapPos, &Gap);
#if	TRACE
	Log("  GapFound=%d Gap=%u,%u %u-%u\n",
	  GapFound, Gap.GenomeIndex, Gap.ChrId, (unsigned) Gap.lo, (unsigned) Gap.hi);
#endif
	assert(GapFound);
	if (!GapFound)
		return false;
	bool c = Covers(Gap, s, MinFract);
#if	TRACE
	Log("  Covers=%d\n", c);
#endif
	return c;
	}

static bool CoversOneGap(const GCSeg &s, const SegBag &Gaps, double MinGapFract,
  double MinAIFract)
	{
	coord_t GapPos;
	bool GapFound = Gaps.GetNearestSeg(s.GetMidpoint(), &GapPos);
	if (!GapFound)
		return false;
	GCSeg Gap;
	GapFound = Gaps.GetSeg(GapPos, &Gap);
	assert(GapFound);
	if (!GapFound)
		return false;
	return Covers(s, Gap, MinGapFract, MinAIFract);
	}

static void UpdateGaps(const GCSeg &s, SegBag &Gaps)
	{
#if	TRACE
	Log("Gaps before update: ");
	Gaps.LogMe();
	Log("\n");
#endif

	Gaps.Subtract(s, g_MinGapLength, g_MinCovFract);

#if	TRACE
	Log("Gaps after update: ");
	Gaps.LogMe();
	Log("\n\n");
#endif
	}

#if	FIG

static void GapsToFig(FILE *f, const AlnInfo &Region, const SegBag &GapsA,
  const SegBag &GapsB)
	{
	const SegMap &GapMap0 = GapsA.m_Segs;
	const SegMap &GapMap1 = GapsB.m_Segs;

	for (SegMapCIter p = GapMap0.begin(); p != GapMap0.end(); ++p)
		{
		const GCSeg &Gap = p->second;
//  rect left top bottom right pensize linestyle colorline
//    fillstyle colorfill
		fprintf(f, "rect %u %u %u %u 1 solid blue solid gray\n",
		  Gap.lo, Region.GetSeg(g_GenomeIndexA).hi, Gap.hi, Region.GetSeg(g_GenomeIndexA).lo);
		}

	for (SegMapCIter p = GapMap1.begin(); p != GapMap1.end(); ++p)
		{
		const GCSeg &Gap = p->second;
//  rect left top bottom right pensize linestyle colorline
//    fillstyle colorfill
		fprintf(f, "rect %u %u %u %u 1 solid green solid gray\n",
		  Region.GetSeg(g_GenomeIndexB).lo, Gap.hi, Region.GetSeg(g_GenomeIndexB).hi, Gap.lo);
		}
	}

static void ToFig(FILE *f, const AlnInfo &Region, const AIVec &AIs,
  BoolVec &Used, const SegBag &GapsA, const SegBag &GapsB)
	{
	const unsigned AICount = (unsigned) AIs.size();

	GapsToFig(f, Region, GapsA, GapsB);

	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const AlnInfo &AI = *(AIs[AIIndex]);
// line xfrom yfrom xto yto pensize linestyle colorline
		const char *Color = Used[AIIndex] ? "black" : "red";
		if (!AI.Reverse)
			fprintf(f, "line %u %u %u %u 1 solid %s\n",
			  AI.GetSeg(g_GenomeIndexA).lo, AI.GetSeg(g_GenomeIndexB).lo, AI.GetSeg(g_GenomeIndexA).hi, AI.GetSeg(g_GenomeIndexB).hi, Color);
		else
			fprintf(f, "line %u %u %u %u 1 solid %s\n",
			  AI.GetSeg(g_GenomeIndexA).lo, AI.GetSeg(g_GenomeIndexB).hi, AI.GetSeg(g_GenomeIndexA).hi, AI.GetSeg(g_GenomeIndexB).lo, Color);
		}
	}

#endif	// FIG

void OptimalCover(const AlnInfo &Region, const AIVec &AIs, BoolVec &Used)
	{
#if	TRACE
	Log("Region=");
	LogAIN(Region);
#endif

	g_GenomeIndexA = GetFilterGenomeIndexA();
	g_GenomeIndexB = GetFilterGenomeIndexB();

	const unsigned AICount = (unsigned) AIs.size();

	Used.clear();
	Used.reserve(AICount);
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		Used.push_back(false);

	SegBag GapsA;
	SegBag GapsB;

	GapsA.InsertSeg(Region.GetSeg(g_GenomeIndexA));
	GapsB.InsertSeg(Region.GetSeg(g_GenomeIndexB));

	IntVec SortOrder;
	SignedIntVec Strength;

	SortOrder.reserve(AICount);
	Strength.reserve(AICount);

	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const AlnInfo &AI = AIs[AIIndex];
		int Stren = Strength1(AI);
		SortOrder.push_back(AIIndex);
		Strength.push_back(Stren);
		}

	g_AIs = &AIs;
	g_GapsA = &GapsA;
	g_GapsB = &GapsB;
	g_Strength = &Strength;

	std::sort(SortOrder.begin(), SortOrder.end(), Comp);

#if	TRACE
	Log("\n\n*** PASS 1\n\n");
	Log("Sorted AIs:\n\n");
	Log("Index  AIIndex  Strength  AI\n");
	Log("-----  -------  --------  --\n");

	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = SortOrder[i];
		const AlnInfo &AI = AIs[AIIndex];
		Log("%5u  %7u  %8u  ", i, AIIndex, Strength[AIIndex]);
		LogAI(AI);
		Log("\n");
		}
	Log("\n");
#endif

	for (unsigned i = 0; i < AICount; ++i)
		{
#if	FIG
		{
		char FigFileName[128];
		sprintf(FigFileName, "p1_%03u_before.gfg", i);
		FILE *f = OpenStdioFile(FigFileName, true);
		ToFig(f, Region, AIs, Used, GapsA, GapsB);
		CloseStdioFile(f);
		}
#endif

		unsigned AIIndex = SortOrder[i];
		const AlnInfo &AI = AIs[AIIndex];

#if	TRACE
		{
		Log("i=%u AI%u=", i, AIIndex);
		LogAI(AI);
		Log(" Length=%u\n", GetAILength(AI));
		}
#endif

		bool CovA = CoveredByOneGap(AI.GetSeg(g_GenomeIndexA), GapsA, g_MinCovFract);
		bool CovB = CoveredByOneGap(AI.GetSeg(g_GenomeIndexB), GapsB, g_MinCovFract);
#if	TRACE
		{
		bool f = (!CovA || !CovB);
		Log("CovA=%d CovB=%d Used[%u]=%s\n", CovA, CovB, i, f ? "false" : "true");
		}
#endif
		if (!CovA || !CovB)
			continue;

		Used[AIIndex] = true;
#if	TRACE
		Log("Update gaps 0:\n");
#endif
		UpdateGaps(AI.GetSeg(g_GenomeIndexA), GapsA);
#if	TRACE
		Log("Update gaps 1:\n");
#endif
		UpdateGaps(AI.GetSeg(g_GenomeIndexB), GapsB);
#if	FIG
		{
		char FigFileName[128];
		sprintf(FigFileName, "p1_%03u_after.gfg", i);
		FILE *f = OpenStdioFile(FigFileName, true);
		ToFig(f, Region, AIs, Used, GapsA, GapsB);
		CloseStdioFile(f);
		}
#endif
		}

	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const AlnInfo &AI = AIs[AIIndex];
		int Stren = Strength2(AI);
		Strength.push_back(Stren);
		}

	std::sort(SortOrder.begin(), SortOrder.end(), Comp);

#if	TRACE
	Log("\n\n*** PASS 2\n\n");
	Log("Sorted AIs:\n\n");
	Log("Index  AIIndex  Strength  AI\n");
	Log("-----  -------  --------  --\n");

	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = SortOrder[i];
		const AlnInfo &AI = AIs[AIIndex];
		Log("%5u  %7u  %8u  ", i, AIIndex, Strength[AIIndex]);
		LogAI(AI);
		Log("\n");
		}
#endif

	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = SortOrder[i];
#if	FIG
		{
		char FigFileName[128];
		sprintf(FigFileName, "p2_%03u_before.gfg", i);
		FILE *f = OpenStdioFile(FigFileName, true);
		ToFig(f, Region, AIs, Used, GapsA, GapsB);
		CloseStdioFile(f);
		}
#endif
		const AlnInfo &AI = AIs[AIIndex];
#if	TRACE
		Log("i=%u AI%u=", i, AIIndex);
		LogAI(AI);
		Log(" Length=%u ", GetAILength(AI));
		if (Used[AIIndex])
			Log("USED\n");
#endif
		if (Used[AIIndex])
			continue;

		bool CovA = CoveredByOneGap(AI.GetSeg(g_GenomeIndexA), GapsA, g_MinCovFract);
		bool CovB = CoveredByOneGap(AI.GetSeg(g_GenomeIndexB), GapsB, g_MinCovFract);
#if	TRACE
		Log("CovA=%d CovB=%d\n", CovA, CovB);
#endif
		if (!CovA && !CovB)
			continue;

		Used[AIIndex] = true;
#if	TRACE
		Log("---\n");
		Log("Update gaps 0:\n");
#endif
		UpdateGaps(AI.GetSeg(g_GenomeIndexA), GapsA);
#if	TRACE
		Log("---\n");
		Log("Update gaps 1:\n");
#endif
		UpdateGaps(AI.GetSeg(g_GenomeIndexB), GapsB);
#if	FIG
		{
		char FigFileName[128];
		sprintf(FigFileName, "p2_%03u_after.gfg", i);
		FILE *f = OpenStdioFile(FigFileName, true);
		ToFig(f, Region, AIs, Used, GapsA, GapsB);
		CloseStdioFile(f);
		}
#endif
		}

// Pass 3. An AI is added iff it covers a gap.
#if	TRACE
	Log("\n\n*** PASS 3\n\n");
	Log("Sorted AIs:\n\n");
	Log("Index  AIIndex  Strength  AI\n");
	Log("-----  -------  --------  --\n");

	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = SortOrder[i];
		const AlnInfo &AI = AIs[AIIndex];
		Log("%5u  %7u  %8u  ", i, AIIndex, Strength[AIIndex]);
		LogAI(AI);
		Log("\n");
		}
#endif
	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = SortOrder[i];
#if	FIG
		{
		char FigFileName[128];
		sprintf(FigFileName, "p3_%03u_before.gfg", i);
		FILE *f = OpenStdioFile(FigFileName, true);
		ToFig(f, Region, AIs, Used, GapsA, GapsB);
		CloseStdioFile(f);
		}
#endif
		const AlnInfo &AI = AIs[AIIndex];
#if	TRACE
		Log("i=%u AI%u=", i, AIIndex);
		LogAI(AI);
		Log(" Length=%u\n", GetAILength(AI));
		if (Used[AIIndex])
			Log("USED\n");
#endif
		if (Used[AIIndex])
			continue;

		bool CovA = CoversOneGap(AI.GetSeg(g_GenomeIndexA), GapsA, g_MinCovFract2, g_MinCovFract2);
		bool CovB = CoversOneGap(AI.GetSeg(g_GenomeIndexB), GapsB, g_MinCovFract2, g_MinCovFract2);
#if	TRACE
		Log("CovA=%d CovB=%d\n", CovA, CovB);
#endif
		if (!CovA && !CovB)
			continue;

		Used[AIIndex] = true;
#if	TRACE
		Log("---\n");
		Log("Update gaps 0:\n");
#endif
		UpdateGaps(AI.GetSeg(g_GenomeIndexA), GapsA);
#if	TRACE
		Log("---\n");
		Log("Update gaps 1:\n");
#endif
		UpdateGaps(AI.GetSeg(g_GenomeIndexB), GapsB);
#if	FIG
		{
		char FigFileName[128];
		sprintf(FigFileName, "p3_%03u_after.gfg", i);
		FILE *f = OpenStdioFile(FigFileName, true);
		ToFig(f, Region, AIs, Used, GapsA, GapsB);
		CloseStdioFile(f);
		}
#endif
		}

#if	TRACE
	Log("\n");
	Log("OptimalCover:\n\n");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const AlnInfo &AI = AIs[AIIndex];
		Log("%s  ", Used[AIIndex] ? "Y" : "N");
		LogAI(AI);
		Log("\n");
		}
#endif
	}
