#define TRACE	0

#include <algorithm>
#include "alninfo.h"
#include "lohi.h"
#include "base.h"

void GetRowOffsetMap(const StdStr &Row, vector<unsigned> &OffsetToCol,
  vector<unsigned> &ColToOffset, bool Plus);
void LogLoHiCE(const LoHi &LH);
void LogLoHiHdr();

// Convert chromosome positions to columns
void GetAnnotsByCol(const AlnInfo &AI, const StdStrVec &Rows,
  unsigned GenomeIndex, const vector<LoHi> &AnnotsByPos,
  vector<LoHi> &AnnotsByCol, vector<unsigned> &ColToOffset)
	{
	AnnotsByCol.clear();

	unsigned SegIndex = AI.GetSegIndex(GenomeIndex);
	const GCSeg &Seg = AI.Segs[SegIndex];
	const StdStr &Row = Rows[SegIndex];

	vector<unsigned> OffsetToCol;
	GetRowOffsetMap(Row, OffsetToCol, ColToOffset, Seg.Plus);

	unsigned SegLength = (unsigned) Seg.GetLength();
	asserta(SIZE(OffsetToCol) == SegLength);

	const unsigned N = SIZE(AnnotsByPos);
	for (unsigned i = 0; i < N; ++i)
		{
		LoHi LHByPos = AnnotsByPos[i];

	// Annotations can begin (end) before (after)
	// the beginning (end) of the segment.
		if (LHByPos.Lo < Seg.lo)
			LHByPos.Lo = Seg.lo;
		if (LHByPos.Hi > Seg.hi)
			LHByPos.Hi = Seg.hi;
		asserta(LHByPos.Hi >= LHByPos.Lo);

		unsigned OffsetLo = LHByPos.Lo - Seg.lo;
		unsigned OffsetHi = LHByPos.Hi - Seg.lo;

		asserta(OffsetLo < SegLength);
		unsigned ColLo = OffsetToCol[OffsetLo];
		unsigned ColHi = OffsetToCol[OffsetHi];
		LoHi LHByCol;
		LHByCol.Index = LHByPos.Index;
		if (Seg.Plus)
			{
			asserta(ColLo <= ColHi);
			LHByCol.Lo = ColLo;
			LHByCol.Hi = ColHi;
			}
		else
			{
			asserta(ColHi <= ColLo);
			LHByCol.Lo = ColHi;
			LHByCol.Hi = ColLo;
			}
		AnnotsByCol.push_back(LHByCol);
		}
	}

/***
Given an alignment and a set of annotations for each genome,
determine a set of intervals where the annotations do and do
not agree. For example:

Genome A annots. :  <-Neutral-><----NGE---><----Neutral----->
Genome A sequence:  XXX--XXXXXXXX--XXXXXXXX--XXXXXXXXXXXXXXXX
Genome B sequence:  XXXXXXX-XXXXXXXXXX-XXXXXXXXXXXXXXXXXXXXXX
Genome B annots. :  <--Neutral-----><----NGE---><--Neutral-->

           Output:  <-Neutral-><-?-><-NGE-><-?-><--Neutral-->

"?" incidates a region where the annotations disagree.

AnnotsByCol[i].Lo/Hi are column indexes, AnnotsByPos.Lo/Hi are
chromosome positions.
***/
void GetOvAnnots(const vector<LoHi> &Annots1, const vector<LoHi> &Annots2,
  vector<LoHi> &Annots)
	{
#if	TRACE
	Log("\n");
	Log("GetOvAnnots\n");
#endif
	Annots.clear();

	const unsigned N1 = SIZE(Annots1);
	const unsigned N2 = SIZE(Annots2);
	const unsigned N = N1 +  N2;
	if (N == 0)
		return;

	vector<LoHiBendpoint> BPs;
	for (unsigned i = 0; i < N1 + N2; ++i)
		{
		const LoHi &LH = (i < N1 ? Annots1[i] : Annots2[i-N1]);

		LoHiBendpoint BP;
		BP.Pos = LH.Lo;
		BP.IsLo = true;
		BP.Index = i;
		BPs.push_back(BP);

		BP.Pos = LH.Hi;
		BP.IsLo = false;
		BPs.push_back(BP);
		}
	asserta(SIZE(BPs) == 2*N);

	sort(BPs.begin(), BPs.end());

	unsigned Lo = 0;
	CE_TYPE CEType1 = CE_Neutral;
	CE_TYPE CEType2 = CE_Neutral;
#if	TRACE
	Log("       Pos  lh  1       LH.lo       LH.Hi    LHType     Type1     Type2\n");
	Log("==========  ==  =  ==========  ==========  ========  ========  ========\n");
#endif
	for (unsigned i = 0; i < 2*N; ++i)
		{
		const LoHiBendpoint &BP = BPs[i];
		unsigned j = BP.Index;
		bool Is1 = (j < N1);
		const LoHi &LHIn = (Is1 ? Annots1[j] : Annots2[j-N1]);
		asserta((BP.IsLo ? LHIn.Lo : LHIn.Hi) == BP.Pos);
#if	TRACE
		{
		Log("%10u", BP.Pos);
		Log("  %s", BP.IsLo ? "Lo" : "Hi");
		Log("  %u", j < N1 ? 1 : 2);
		Log("  %10u", LHIn.Lo);
		Log("  %10u", LHIn.Hi);
		Log("  %8.8s", CETypeToStr((CE_TYPE) LHIn.Index));
		Log("  %8.8s", CETypeToStr(CEType1));
		Log("  %8.8s", CETypeToStr(CEType2));
		}
#endif
		if (Lo < BP.Pos)
			{
			LoHi LHOut;
			LHOut.Lo = Lo;
			LHOut.Hi = BP.IsLo ? BP.Pos - 1 : BP.Pos;
			LHOut.Index = CEType1 | (CEType2 << 16); 
			Annots.push_back(LHOut);
#if	TRACE
			{
			CE_TYPE t1 = (CE_TYPE) (LHOut.Index & 0xffff);
			CE_TYPE t2 = (CE_TYPE) (LHOut.Index >> 16);
			Log(" Out: %u-%u %s,%s",
			  LHOut.Lo, LHOut.Hi, CETypeToStr(t1), CETypeToStr(t2));
			}
#endif
			}

		if (BP.IsLo)
			{
			Lo = BP.Pos;
			(Is1 ? CEType1 : CEType2) = (CE_TYPE) LHIn.Index;
			}
		else
			{
			Lo = BP.Pos + 1;
			(Is1 ? CEType1 : CEType2) = (CE_TYPE) CE_Neutral;
			}
#if	TRACE
		Log("\n");
#endif
		}
	}
