#include <math.h>
#include "getopts.h"
#include "gff2.h"
#include "progress.h"
#include "hist.h"
#include "lhist.h"

double GetCorrel(const vector<pair<double, double> > &XYPairs);

static unsigned g_CurrentGeneBaseCount;
static double g_SumCurrentGeneBaseProbs;
static bool g_DoDump;

enum HIST_TYPE
	{
	HT_NGE,
	HT_GeneNCE,
	HT_CDS,
	HT_Gene,
	};
static const unsigned g_HTCount = 4;
static const unsigned g_MaxCorrelDistLog2 = 8; // 2^8 = 256

static HistData g_Hists[g_HTCount];
static LHistData g_LHists[g_HTCount];
static vector<vector<double> > g_SumDistCorrels;
static vector<unsigned> g_SumDistCorrelNs;

void GetProbsVec(const string &Attr, vector<double> &ProbsVec)
	{
	ProbsVec.clear();

	const char *Probs = Attr.c_str();

	unsigned Length = SIZE(Attr)/2;
	unsigned i = 0;
	for (unsigned k = 0; k < Length; ++k)
		{
		char c1 = Probs[i++];
		char c2 = Probs[i++];
		if (!isxdigit(c1) || !isxdigit(c2))
			{
			if (isprint(c1) && isprint(c2))
				Die("Invalid hex digits '%c%c' in probs", c1, c2);
			else
				Die("Invalid hex digits %02x%02x in probs", (byte) c1, (byte) c2);
			}
		static char s[3];
		s[0] = c1;
		s[1] = c2;
		char *NotUsed;
		unsigned ProbIndex = strtoul(s, &NotUsed, 16);
		double Prob = ProbIndex/256.0;
		ProbsVec.push_back(Prob);
		}
	}

double GetSum(const vector<double> &v)
	{
	double Sum = 0.0;

	unsigned N = SIZE(v);
	for (unsigned k = 0; k < N; ++k)
		Sum += v[k];
	return Sum;
	}

void GetProbsVec(const GFFRecord2 &Rec, vector<double> &Probs)
	{
	string ProbsAttr;
	bool Found = GFFGetAttr(Rec.Attrs, "probs", ProbsAttr);
	if (!Found)
		return;
	GetProbsVec(ProbsAttr, Probs);
	}

static void GetDistCorrels(const vector<double> &Probs,
  vector<double> &Correls)
	{
	Correls.clear();
	Correls.resize(g_MaxCorrelDistLog2);
	const unsigned L = SIZE(Probs);
	for (unsigned DistLog2 = 0; DistLog2 < g_MaxCorrelDistLog2; ++DistLog2)
		{
		unsigned Dist = (1 << DistLog2);
		vector<pair<double, double> > XYPairs;
		for (unsigned i = 0; i + Dist < L; ++i)
			{
			double p1 = Probs[i];
			double p2 = Probs[i + Dist];
			pair<double, double> XYPair(p1, p2);
			XYPairs.push_back(XYPair);
			}
		Correls[DistLog2] = GetCorrel(XYPairs);
		}
	}

static void UpdateCorrels(unsigned HType, const vector<double> &Probs)
	{
	asserta(HType < SIZE(g_SumDistCorrels));
	vector<double> Correls;
	GetDistCorrels(Probs, Correls);
	for (unsigned DistLog2 = 0; DistLog2 < g_MaxCorrelDistLog2; ++DistLog2)
		g_SumDistCorrels[HType][DistLog2] += Correls[DistLog2];
	++g_SumDistCorrelNs[HType];
	}

static void DoNonGeneCE(const GFFRecord2 &Rec)
	{
	unsigned Length = Rec.GetLength();
	vector<double> Probs;
	GetProbsVec(Rec, Probs);
	double SumProb = GetSum(Probs);
	double AvgProb = SumProb/Length;
	g_Hists[HT_NGE].Add(AvgProb);
	g_LHists[HT_NGE].Add(Length);
	UpdateCorrels(HT_NGE, Probs);
	}

static void DoGene(unsigned GeneIndex, const vector<GFFRecord2> &Recs)
	{
	const unsigned N = SIZE(Recs);
	double TotalCDSSumProb = 0;
	double TotalNCESumProb = 0;
	unsigned TotalCDSLength = 0;
	unsigned TotalNCELength = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		const GFFRecord2 &Rec = Recs[i];
		vector<double> Probs;
		GetProbsVec(Rec, Probs);
		double SumProb = GetSum(Probs);
		unsigned Length = Rec.GetLength();

		if (g_DoDump)
			Log("%5u  %7.7s  %7.4f\n",
			  GeneIndex, Rec.Feature.c_str(), SumProb/Length);

		g_CurrentGeneBaseCount += Length;
		g_SumCurrentGeneBaseProbs += SumProb;

		if (Rec.Feature == "CDS")
			{
			TotalCDSLength += Length;
			TotalCDSSumProb += SumProb;

			double AvgProb = SumProb/Length;
			g_Hists[HT_CDS].Add(AvgProb);
			g_LHists[HT_CDS].Add(Length);

			UpdateCorrels(HT_CDS, Probs);
			}
		else
			{
			TotalNCELength += Length;
			TotalNCESumProb += SumProb;

			double AvgProb = SumProb/Length;
			g_Hists[HT_GeneNCE].Add(AvgProb);
			g_LHists[HT_GeneNCE].Add(Length);

			UpdateCorrels(HT_GeneNCE, Probs);
			}
		}

	asserta(g_CurrentGeneBaseCount > 0);

	double GeneAvgProb = g_SumCurrentGeneBaseProbs/g_CurrentGeneBaseCount;
	double GeneCDSAvgProb = TotalCDSSumProb/TotalCDSLength;
	double GeneNCEAvgProb = TotalNCESumProb/TotalNCELength;

	g_Hists[HT_Gene].Add(GeneAvgProb);

	if (g_DoDump)
		{
		Log("%5u  %7.7s  %7.4f\n", GeneIndex, "GeneCDS", GeneCDSAvgProb);
		Log("%5u  %7.7s  %7.4f\n", GeneIndex, "GeneNCE", GeneNCEAvgProb);
		Log("%5u  %7.7s  %7.4f\n", GeneIndex, "Gene", GeneAvgProb);
		Log("\n");
		}

	g_SumCurrentGeneBaseProbs = 0;
	g_CurrentGeneBaseCount = 0;
	}

void ProbStats()
	{
	const char *GFFFileName = RequiredValueOpt("probstats");
	g_DoDump = FlagOpt("dump");

	g_Hists[HT_Gene].Init("Gene");
	g_Hists[HT_CDS].Init("CDS");
	g_Hists[HT_GeneNCE].Init("GeneNCE");
	g_Hists[HT_NGE].Init("NGE");

	g_LHists[HT_Gene].Init("Gene");
	g_LHists[HT_CDS].Init("CDS");
	g_LHists[HT_GeneNCE].Init("GeneNCE");
	g_LHists[HT_NGE].Init("NGE");

	g_SumDistCorrels.resize(g_HTCount);
	g_SumDistCorrelNs.resize(g_HTCount);
	for (unsigned i = 0; i < g_HTCount; ++i)
		g_SumDistCorrels[i].resize(g_MaxCorrelDistLog2);

	FILE *f = OpenStdioFile(GFFFileName);

	map<unsigned, vector<GFFRecord2> > Genes;
	vector<GFFRecord2> NonGeneCEs;

	GFFRecord2 Rec;
	ProgressStartFile(f, "Reading");
	unsigned Count = 0;
	while (GetNextGFFRecord(f, Rec))
		{
		if (Count++%100 == 0)
			ProgressStepFile();
		if (Rec.Feature == "tandem")
			continue;
		unsigned GeneIndex = UINT_MAX;
		if (GFFGetIntAttr(Rec.Attrs, "gene_index", GeneIndex))
			{
			map<unsigned, vector<GFFRecord2> >::const_iterator p =
			  Genes.find(GeneIndex);
			if (p == Genes.end())
				{
				vector<GFFRecord2> CEs;
				CEs.push_back(Rec);
				Genes[GeneIndex] = CEs;
				}
			else
				Genes[GeneIndex].push_back(Rec);
			}
		else
			NonGeneCEs.push_back(Rec);
		}
	ProgressDone();

	if (g_DoDump)
		{
		Log("\n");
		Log("Genes\n");
		Log("Index  Feature  AvgProb\n");
		Log("=====  =======  =======\n");
		}

	const unsigned GeneCount = SIZE(Genes);
	ProgressStart("Genes");
	Count = 0;
	for (map<unsigned, vector<GFFRecord2> >::const_iterator p = Genes.begin();
	  p != Genes.end(); ++p)
		{
		if (Count++%10 == 0)
			ProgressStep(Count, GeneCount);
		DoGene(p->first, p->second);
		}
	ProgressDone();

	const unsigned NGCount = SIZE(NonGeneCEs);
	ProgressStart("Non-gene CEs");
	Count = 0;
	for (vector<GFFRecord2>::const_iterator p = NonGeneCEs.begin();
	  p != NonGeneCEs.end(); ++p)
		{
		if (Count++%100 == 0)
			ProgressStep(Count, NGCount);
		DoNonGeneCE(*p);
		}
	ProgressDone();

	for (unsigned i = 0; i < g_HTCount; ++i)
		{
		if (i == HT_Gene)
			continue;
		unsigned N = g_SumDistCorrelNs[i];
		if (N == 0)
			continue;
		for (unsigned Dist = 0; Dist < g_MaxCorrelDistLog2; ++Dist)
			g_SumDistCorrels[i][Dist] /= N;
		}

	for (unsigned i = 0; i < g_HTCount; ++i)
		g_Hists[i].LogHist();

	for (unsigned i = 0; i < g_HTCount; ++i)
		if (i != HT_Gene)
			g_LHists[i].LogHist();

	Log("\n");
	Log("        Type          Nr    Mean  StdDev");
	for (unsigned i = 0; i < g_MaxCorrelDistLog2; ++i)
		Log("  %5u", 1 << i);
	Log("\n");

	Log("============  ==========  ======  ======");
	for (unsigned i = 0; i < g_MaxCorrelDistLog2; ++i)
		Log("  =====");
	Log("\n");

#define	X(x)									\
	{											\
	const HistData &Hist = g_Hists[HT_##x];		\
	unsigned N = Hist.GetCount();				\
	if (N > 0)									\
		{										\
		double Mean = Hist.GetMean();			\
		double StdDev = Hist.GetStdDev();		\
		Log("%12.12s", #x);						\
		Log("  %10u", N);						\
		Log("  %6.4f", Mean);					\
		Log("  %6.4f", StdDev);					\
		if (HT_##x != HT_Gene)					\
		for (unsigned Dist = 0; Dist < g_MaxCorrelDistLog2; ++Dist) \
			Log("  %5.2f", g_SumDistCorrels[HT_##x][Dist]); \
		Log("\n");								\
		}										\
	}

	X(Gene)
	X(CDS)
	X(GeneNCE)
	X(NGE)

#undef X

	CloseStdioFile(f);
	}
