#define TRACE	0

#include "revfile.h"
#include "progress.h"
#include "lhist.h"
#include "ivtree.h"
#include "lohi.h"
#include "gff2.h"
#include "base.h"
#include "getopts.h"
#include "codon.h"

void GetAlnRange(const AlnInfo &AI, const StdStrVec &Rows, unsigned GenomeIndex,
  unsigned ChrId, coord_t lo, coord_t hi, AlnInfo &RangeAI, StdStrVec &RangeRows);
void GetAlnsRange(const RevFile &Rev, unsigned GenomeIndex,
  unsigned ChrId, coord_t lo, coord_t hi,
  AIVec &AIs, vector<StdStrVec> &RowsVec);
void LogAIRows(const AlnInfo &AI, const StdStrVec &Rows, const RevFile *Rev);
void RevComp(string &Seq);
coord_t GetOverlapLength(coord_t loA, coord_t hiA, coord_t loB, coord_t hiB);
unsigned GetOverlappingSegIndex(const AlnInfo &AI, unsigned GenomeIndex,
  unsigned ChrId, coord_t lo, coord_t hi);
uint8 IncFrame(uint8 Frame, gpos_t Length);
char LettersToCodonChar(char c1, char c2, char c3);
double Pct(double, double);

// Hack to avoid redefining CDSInfo (MS linker can't cope).
// Suffix "UniqueSuffix" is arbitrary, should be unique.
#define CDSInfo CDSInfo_UniqueSuffix

struct CDSInfo
	{
	unsigned GeneIndex;
	unsigned ChrId;
	gpos_t Lo;
	gpos_t Hi;
	bool FirstCDSInGene;
	bool LastCDSInGene;
	unsigned Frame;
	bool Plus;

	bool operator<(const CDSInfo &rhs) const
		{
		if (Lo != rhs.Lo)
			return Lo < rhs.Lo;
		return Hi < rhs.Hi;
		}

	pair<gpos_t, gpos_t> GetIv() const
		{
		return pair<gpos_t, gpos_t>(Lo, Hi);
		}

	pair<gpos_t, gpos_t> operator()() const
		{
		return pair<gpos_t, gpos_t>(Lo, Hi);
		}

	void LogMe(bool NL = false) const
		{
		Log("Gene=%u chr=%u lo=%"PF" hi=%"PF" L=%"PF" 1st=%c Last=%c Frame=%c%d",
		  GeneIndex, ChrId, Lo, Hi, Hi-Lo+1, tof(FirstCDSInGene), tof(LastCDSInGene),
		  pom(Plus), Frame);
		if (NL)
			Log("\n");
		}
	};

static unsigned g_GenomeIndex1;
static unsigned g_GenomeIndex2;
static string g_GenomeName1;
static string g_GenomeName2;
static vector<IvTree<CDSInfo> > g_CDSIxs1;
static vector<IvTree<CDSInfo> > g_CDSIxs2;
static unsigned g_StrandMismatchCount;
static unsigned g_AlignedCDSBaseCount1 = 0;
static unsigned g_AlignedCDSBaseCount2 = 0;
static unsigned g_FrameSlipCount = 0;

static unsigned GetCDSIndex(const RevFile &Rev, unsigned GenomeIndex,
  const char *FileName, vector<CDSInfo> &CDSs, vector<IvTree<CDSInfo> > &Ixs)
	{
	CDSs.clear();
	FILE *f = OpenStdioFile(FileName);

	GFFRecord2 Rec;
	unsigned SkippedCount = 0;
	unsigned Count = 0;
	unsigned CDSBaseCount = 0;
	map<unsigned, unsigned> GeneIndexToCDSCount;
	ProgressStartFile(f, "Reading %.32s", FileName);
	unsigned ChrId = UINT_MAX;
	string ChrName;
	while (GetNextGFFRecord(f, Rec))
		{
		if (Count++%5000 == 0)
			ProgressStepFile();

		if (Rec.Feature != "CDS")
			{
			++SkippedCount;
			continue;
			}

		CDSBaseCount += Rec.GetLength();

		if (Rec.SeqName != ChrName)
			{
			ChrName = Rec.SeqName;
			ChrId = Rev.GetChrId(GenomeIndex, ChrName);
			unsigned ChrIndex = Rev.GetChrIndex(GenomeIndex, ChrId);
			if (ChrIndex != ChrId)
				Die("ChrIndex != ChrId, sparse chrids not supported");
			asserta(ChrIndex < SIZE(Ixs));
			}

		asserta(Rec.Start > 0);
		asserta(Rec.Start <= Rec.End);

		unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");

		CDSInfo CDS;
		CDS.ChrId = ChrId;
		CDS.Lo = Rec.Start - 1;
		CDS.Hi = Rec.End - 1;
		CDS.Frame = Rec.Frame;
		CDS.Plus = (Rec.Strand == '+');
		CDS.GeneIndex = GeneIndex;
		CDS.FirstCDSInGene = false;
		CDS.LastCDSInGene = false;

		map<unsigned, unsigned>::const_iterator p = GeneIndexToCDSCount.find(GeneIndex);
		if (p == GeneIndexToCDSCount.end())
			GeneIndexToCDSCount[GeneIndex] = 1;
		else
			++GeneIndexToCDSCount[GeneIndex];

		CDSs.push_back(CDS);
		}
	ProgressDone();
	Progress("%.32s: %u recs, skipped %u non-CDS", FileName, Count, SkippedCount);

	Progress("Sorting");
	sort(CDSs.begin(), CDSs.end());

	unsigned CDSIndex = 0;
	unsigned CurrGeneIndex = UINT_MAX;
	for (vector<CDSInfo>::iterator p = CDSs.begin(); p != CDSs.end(); ++p)
		{
		CDSInfo &CDS = *p;
		unsigned GeneIndex = CDS.GeneIndex;

		map<unsigned, unsigned>::const_iterator q = GeneIndexToCDSCount.find(GeneIndex);
		asserta(q != GeneIndexToCDSCount.end());
		asserta(q->first == GeneIndex);
		unsigned CDSCount = q->second;

		if (GeneIndex != CurrGeneIndex)
			CDSIndex = 0;
		else
			++CDSIndex;
		CurrGeneIndex = CDS.GeneIndex;

		if (CDS.Plus)
			{
			if (CDSIndex == 0)
				CDS.FirstCDSInGene = true;
			if (CDSIndex + 1 == CDSCount)
				CDS.LastCDSInGene = true;
			}
		else
			{
			if (CDSIndex == 0)
				CDS.LastCDSInGene = true;
			if (CDSIndex + 1 == CDSCount)
				CDS.FirstCDSInGene = true;
			}

		asserta(CDS.ChrId < SIZE(Ixs));
		Ixs[CDS.ChrId].Add(CDS);
		}

	CloseStdioFile(f);
	return CDSBaseCount;
	}

static void DoAln(const RevFile &InRev, const CDSInfo &CDS1, const CDSInfo &CDS2,
  const AlnInfo &AI, const StdStrVec &Rows, RevFile &OutRev)
	{
#if	TRACE
	Log("\n");
	Log("DoAln\n");
	Log("CDS1: ");
	CDS1.LogMe(true);
	Log("CDS2: ");
	CDS2.LogMe(true);
	LogAIRows(AI, Rows, &InRev);
#endif

	const GCSegVec &Segs = AI.Segs;
	if (SIZE(Segs) == 0)
		return;
	asserta(SIZE(Segs) >= 2);
	unsigned SegIndex1 = (Segs[0].GenomeIndex == g_GenomeIndex1 ? 0 : 1);
	unsigned SegIndex2 = (SegIndex1 == 0 ? 1 : 0);

	const GCSeg &Seg1 = Segs[SegIndex1];
	const GCSeg &Seg2 = Segs[SegIndex2];

	g_AlignedCDSBaseCount1 += Seg1.GetLength();
	g_AlignedCDSBaseCount2 += Seg2.GetLength();

	bool Rev1 = (CDS1.Plus != Seg1.Plus);
	bool Rev2 = (CDS2.Plus != Seg2.Plus);
	if (Rev1 != Rev2)
		{
		++g_StrandMismatchCount;
#if 0
		LogAIRows(AI, Rows, &InRev);
		Log("\n");
		Log("CDS1: ");
		CDS1.LogMe(true);
		Log("CDS2: ");
		CDS2.LogMe(true);
		Warning("Gene strands disagree");
#endif
		return;
		}

	AlnInfo OutAI;
	OutAI.Segs.push_back(Seg1);
	OutAI.Segs.push_back(Seg2);
	OutAI.SetAttr(ALN_ATTR_GeneIndex1, CDS1.GeneIndex);
	OutAI.SetAttr(ALN_ATTR_GeneIndex2, CDS2.GeneIndex);

	StdStrVec OutRows(2);
	StdStr &Row1 = OutRows[0];
	StdStr &Row2 = OutRows[1];
	Row1 = Rows[SegIndex1];
	Row2 = Rows[SegIndex2];

	if (Rev1)
		{
		OutAI.Segs[0].Plus = !OutAI.Segs[0].Plus;
		OutAI.Segs[1].Plus = !OutAI.Segs[1].Plus;
		RevComp(Row1);
		RevComp(Row2);
		}

	unsigned MissingLo1 = (unsigned) (Seg1.lo - CDS1.Lo);
	unsigned MissingLo2 = (unsigned) (Seg2.lo - CDS2.Lo);
	unsigned MissingHi1 = (unsigned) (CDS1.Hi - Seg1.hi);
	unsigned MissingHi2 = (unsigned) (CDS2.Hi - Seg2.hi);
#if TRACE
	if (MissingLo1 > 0)
		{
		CDS1.LogMe();
		Log(" %u bases missing from start of CDS1\n", MissingLo1);
		}
	if (MissingLo2 > 0)
		{
		CDS2.LogMe();
		Log(" %u bases missing from start of CDS2\n", MissingLo2);
		}
	if (MissingHi1)
		{
		CDS1.LogMe();
		Log(" %u bases missing from end of CDS1\n", MissingHi1);
		}
	if (MissingHi2)
		{
		CDS2.LogMe();
		Log(" %u bases missing from end of CDS2\n", MissingHi2);
		}
#endif

// 0 indicates that the feature begins with a whole codon at the 5' most base.
// 1 means that there is one extra base (the third base of a codon) before the
// first whole codon and 2 means that there are two extra bases (the second
// and third bases of the codon) before the first codon.
	unsigned FrameSkip1 = (CDS1.Plus ? MissingLo1 : MissingHi1);
	unsigned FrameSkip2 = (CDS2.Plus ? MissingLo2 : MissingHi2);
	unsigned Frame = IncFrame((uint8) CDS1.Frame, FrameSkip1);
	unsigned Frame2 = IncFrame((uint8) CDS2.Frame, FrameSkip2);
	if (Frame != Frame2)
		{
#if TRACE
		Log("\n");
		Log("FRAME MISMATCH\n");
		LogAIRows(AI, Rows, &InRev);
		Log("CDS1=");
		CDS1.LogMe(true);
		Log("CDS2=");
		CDS2.LogMe(true);
		Log("Skip1=%u Skip2=%u Strand1=%c Strand2=%c Frame=%u Frame2=%u\n",
		  FrameSkip1, FrameSkip2, pom(CDS1.Plus), pom(CDS2.Plus), Frame, Frame2);
		Log("Missing start1 %u, end1 %u, start2 %u, end2 %u\n",
		  MissingLo1, MissingHi1, MissingLo2, MissingHi2);
		Log("IncFrame(%u, %u) = %u\n", (uint8) CDS1.Frame, FrameSkip1, Frame);
		Log("IncFrame(%u, %u) = %u\n", (uint8) CDS2.Frame, FrameSkip2, Frame2);
		Warning("Frame mismatch skip1 %u, skip2 %u", FrameSkip1, FrameSkip2);
#endif
		++g_FrameSlipCount;
		return;
		}

	OutAI.SetAttr(ALN_ATTR_Frame, Frame);

#if	TRACE
	Log("\n");
	Log("OutRev.AddAln:\n");
	LogAIRows(OutAI, OutRows, &InRev);
#endif
	if (!OutRev.ValidateSegLengths(OutAI, OutRows, false))
		return;
	OutRev.AddAln(OutAI, OutRows);
	}

void CDSAlns()
	{
	const char *InFileName = RequiredValueOpt("cdsalns");
	const char *AnnotsFileName1 = ValueOpt("annots1");
	const char *AnnotsFileName2 = ValueOpt("annots2");
	const char *OutFileName = RequiredValueOpt("alns");

	RevFile InRev;
	InRev.Open(InFileName);

	RevFile OutRev;
	OutRev.Create(OutFileName, InRev.GetGenomes());

//  -----------------------------------------------------
//  Common code with AlnStats
	g_GenomeIndex1 = 0;//TODO
	g_GenomeIndex2 = 1;//TODO

	const unsigned ChrCount1 = InRev.GetChrCount(g_GenomeIndex1);
	const unsigned ChrCount2 = InRev.GetChrCount(g_GenomeIndex2);

	g_CDSIxs1.resize(ChrCount1);
	g_CDSIxs2.resize(ChrCount2);

	vector<CDSInfo> CDSs1;
	vector<CDSInfo> CDSs2;
	unsigned CDSBaseCount1 = 
	  GetCDSIndex(InRev, g_GenomeIndex1, AnnotsFileName1, CDSs1, g_CDSIxs1);
	unsigned CDSBaseCount2 = 
	  GetCDSIndex(InRev, g_GenomeIndex2, AnnotsFileName2, CDSs2, g_CDSIxs2);

	const unsigned CDSCount1 = SIZE(CDSs1);
	ProgressStart("Processing");
	for (unsigned CDSIndex1 = 0; CDSIndex1 < CDSCount1; ++CDSIndex1)
		{
		if (CDSIndex1%10 == 0)
			ProgressStep(CDSIndex1, CDSCount1);
		const CDSInfo &CDS1 = CDSs1[CDSIndex1];
#if	TRACE
		Log("\n");
		Log(">>> MAIN LOOP CDS1=");
		CDS1.LogMe(true);
#endif
		AIVec AIs;
		vector<StdStrVec> RowsVec;
		GetAlnsRange(InRev, g_GenomeIndex1, CDS1.ChrId, CDS1.Lo, CDS1.Hi, AIs, RowsVec);
		const unsigned AICount = SIZE(AIs);
		asserta(SIZE(RowsVec) == AICount);
		if (AICount == 0)
			{
#if	TRACE
			Log("CDS1 has no alignments (1)\n");
#endif
			continue;
			}

		for (unsigned i = 0; i < AICount; ++i)
			{
			const AlnInfo &AI = AIs[i];
			const StdStrVec &Rows = RowsVec[i];
#if	TRACE
			Log("CDS1 range AI\n");
			LogAIRows(AI, Rows, &InRev);
#endif
			const GCSegVec &Segs = AI.Segs;
			const unsigned SegCount = SIZE(Segs);
			unsigned SegIndex1 = GetOverlappingSegIndex(AI, g_GenomeIndex1, CDS1.ChrId,
			  CDS1.Lo, CDS1.Hi);
			if (SegIndex1 == UINT_MAX)
				{
				Log("\n");
				Log("AI=");
				AI.LogMe(true);
				Log("CDS1=");
				CDS1.LogMe();
				Warning("Overlap seg (1) failed");
				continue;
				}
			const GCSeg &Seg1 = Segs[SegIndex1];

			for (unsigned SegIndex2 = 0; SegIndex2 < SegCount; ++SegIndex2)
				{
				const GCSeg &Seg2 = Segs[SegIndex2];
				if (Seg2.GenomeIndex != g_GenomeIndex2)
					continue;

				pair<gpos_t, gpos_t> Iv(Seg2.lo, Seg2.hi);
				unsigned ChrId2 = Seg2.ChrId;
				asserta(ChrId2 < SIZE(g_CDSIxs2));
				const IvTree<CDSInfo> &Ix2 = g_CDSIxs2[ChrId2];

				vector<CDSInfo> CDSs2;
				Ix2.GetOverlaps(Iv, CDSs2);
				
				const unsigned CDSCount2 = SIZE(CDSs2);
				if (CDSCount2 == 0)
					{
#if	TRACE
					Log("CDS1 has no alignments (2)\n");
#endif
					continue;
					}
				
				for (unsigned CDSIndex2 = 0; CDSIndex2 < CDSCount2; ++CDSIndex2)
					{
					const CDSInfo &CDS2 = CDSs2[CDSIndex2];
#if	TRACE
					Log("CDS2=");
					CDS2.LogMe(true);
#endif
					unsigned SegIndex2 = GetOverlappingSegIndex(AI, g_GenomeIndex2,
					  CDS2.ChrId, CDS2.Lo, CDS2.Hi);
					if (SegIndex2 == UINT_MAX)
						{
						Warning("Seg overlap 2 failed");
						continue;
						}

					AlnInfo PairAI;
					PairAI.Segs.push_back(Seg1);
					PairAI.Segs.push_back(Seg2);

					StdStrVec PairRows;
					PairRows.push_back(Rows[SegIndex1]);
					PairRows.push_back(Rows[SegIndex2]);

					AlnInfo RangeAI;
					StdStrVec RangeRows;
					GetAlnRange(PairAI, PairRows, g_GenomeIndex2, CDS2.ChrId,
					  CDS2.Lo, CDS2.Hi, RangeAI, RangeRows);
					if (RangeRows.empty())
						continue;
#if	TRACE
					Log("GetAlnRange, PairAI=");
					PairAI.LogMe();
					Log("GetAlnRange, RangeAI=");
					RangeAI.LogMe();
					Log("RangeAln:\n");
					LogAIRows(RangeAI, RangeRows, &InRev);
#endif
					DoAln(InRev, CDS1, CDS2, RangeAI, RangeRows, OutRev);
					}
				}
			}
		}
	ProgressDone();

	if (g_StrandMismatchCount > 0)
		Log("%u strand mismatches skipped\n", g_StrandMismatchCount);
	if (g_FrameSlipCount > 0)
		Log("%u frame slips skipped (indel of 3n bases not at codon boundary)\n", g_FrameSlipCount);

	const string &GenName1 = InRev.GetGenomeName(g_GenomeIndex1);
	const string &GenName2 = InRev.GetGenomeName(g_GenomeIndex2);

	Log("\n");
	Log("%10.10s  First genome\n", GenName1.c_str());
	Log("%10.10s  Second genome\n", GenName2.c_str());
	Log("%10u  CDS bases first genome\n", CDSBaseCount1);
	Log("%10u  CDS bases second genome\n", CDSBaseCount2);
	Log("%10u  aligned CDS bases first genome\n", g_AlignedCDSBaseCount1);
	Log("%10u  aligned CDS bases second genome\n", g_AlignedCDSBaseCount2);
	}
