#include <math.h>
#include "ratemgr.h"
#include "myrand.h"

double ComputeBasesAffected(gpos_t Llo, gpos_t Lhi, double flo, double fhi);
gpos_t GetRandLengthRR(const RateRange &RR);

RateMgr::RateMgr()
	{
	Clear();
	}

void RateMgr::Clear()
	{
	m_Type = (RATE_TYPE) INT_MAX;
	m_Rates.clear();
	m_TrimmedRates.clear();
	m_Ranges.clear();
	m_TotalAcceptedBases = 0;
	m_TotalRejectedBases = 0;

	memset(m_ResultCounts, 0, sizeof(m_ResultCounts));
	}

static bool LT(const pair<gpos_t, double> &r1, const pair<gpos_t, double> &r2)
	{
	return r1.first < r2.first;
	}

void RateMgr::Init(RATE_TYPE Type, const RateVec &Rates)
	{
	Clear();

	m_Type = Type;
	m_Rates = Rates;

	sort(m_Rates.begin(), m_Rates.end(), LT);

// Verify sorted correctly in increasing length order
	const unsigned N = SIZE(m_Rates);
	for (unsigned i = 1; i < N; ++i)
		{
		gpos_t L1 = m_Rates[i-1].first;
		gpos_t L = m_Rates[i].first;
		if (L == L1)
			Die("RateMgr::Init(%s), length %"PF" specified twice", TypeStr(), L1);
		asserta(L > L1);
		}

	MakeTrimmedRates();
	MakeRanges();
	SetBinRates();
	}

static double Interpolate(gpos_t L, gpos_t L1, gpos_t L2, double R1, double R2)
	{
	if (L == L1)
		return R1;
	else if (L == L2)
		return R2;
	asserta(L2 != L1);
	double Gradient = (R2 - R1)/(double) (L2 - L1);
	return R1 + (L - L1)*Gradient;
	}

// If Length is within range but missing, insert interpolated value.
void RateMgr::InsertInterpolatedRate(RateList &Rates, gpos_t Length)
	{
	if (Rates.empty())
		return;

	const LRPair *PrevRate = &Rates.front();
	RateList::iterator p = Rates.begin();
	for (++p; p != Rates.end(); ++p)
		{
		const LRPair &ThisRate = *p;
		gpos_t PrevL = PrevRate->first;
		gpos_t ThisL = ThisRate.first;
		if (PrevL < Length && ThisL > Length)
			{
			gpos_t ThisL = ThisRate.first;
			double ThisR = ThisRate.second;
			double PrevR = PrevRate->second;
			double R = Interpolate(Length, PrevL, ThisL, PrevR, ThisR);
			Rates.insert(p, LRPair(Length, R));
			return;
			}
		PrevRate = &ThisRate;
		}
	}

// Compute modified rates convenient for constructing RateRange's:
//     (1) Replace consecutive zero with one zero where appropriate.
//     (2) Insert interpolated rates for m_MicroThreshold and
//         m_MicroThreshold+1 if missing.
void RateMgr::MakeTrimmedRates()
	{
	const unsigned N = SIZE(m_Rates);
	m_TrimmedRates.clear();

// Can delete a zero rate if it is the second of three consecutive zeroes, where
// the beginning and end of the list are considered to be zero.
	for (unsigned i = 0; i < N; ++i)
		{
		bool PrevZero = (i == 0 ? true : m_Rates[i-1].second == 0);
		bool ThisZero = (m_Rates[i].second == 0);
		bool NextZero = (i + 1 == N ? true : m_Rates[i+1].second == 0);
		bool Delete = (PrevZero && ThisZero && NextZero);
		if (!Delete)
			m_TrimmedRates.push_back(m_Rates[i]);
		}
	}

void RateMgr::MakeRanges()
	{
	m_Ranges.clear();

	if (m_TrimmedRates.empty())
		return;

	if (m_TrimmedRates.size() == 1)
		{
		const LRPair &LR = m_TrimmedRates.front();
		RateRange RR;
		RR.LoLength = LR.first;
		RR.HiLength = LR.first;
		RR.LoRate = LR.second;
		RR.HiRate = LR.second;
		RR.TotalRate = LR.second;
		m_Ranges.push_back(RR);
		return;
		}

	const LRPair *PrevLR = &m_TrimmedRates.front();
	RateList::const_iterator p = m_TrimmedRates.begin();
	RateList::const_iterator Lastp = m_TrimmedRates.end();
	--Lastp;
	for (++p; p != m_TrimmedRates.end(); ++p)
		{
		const LRPair &LR = *p;
		gpos_t PrevL = PrevLR->first;
		gpos_t L = LR.first;
		double PrevR = PrevLR->second;
		double R = LR.second;

		RateRange RR;
		RR.LoLength = PrevL;
		if (p == Lastp)
			RR.HiLength = L;
		else
			RR.HiLength = L - 1;

		RR.LoRate = PrevR;
		RR.HiRate = Interpolate(RR.HiLength, PrevL, L, PrevR, R);

		gpos_t Range = RR.HiLength - RR.LoLength + 1;
		double MinRate = min(RR.HiRate, RR.LoRate);
		double MaxRate = max(RR.HiRate, RR.LoRate);
		RR.TotalRate = Range*(MinRate + (MaxRate - MinRate)/2);

		m_Ranges.push_back(RR);

		PrevLR = &LR;
		}
	}

void RateMgr::LogMe() const
	{
	double IR = GetTotalRate();

	Log("\n");
	Log("%s\n", TypeStr());
	Log("Integ.Rate    %.3g\n", IR);

	Log("Rates:\n");
	Log("    Length        Rate\n");
	Log("==========  ==========\n");
	unsigned N = SIZE(m_Rates);
	for (unsigned i = 0; i < N; ++i)
		Log("%10"PF"  %10.3g\n", m_Rates[i].first, m_Rates[i].second);

	Log("\n");
	Log("Trimmed rates:\n");
	Log("   Length        Rate\n");
	Log("=========  ==========\n");
	for (RateList::const_iterator p = m_TrimmedRates.begin();
	  p != m_TrimmedRates.end(); ++p)
		Log("%10"PF"  %10.3g\n", (*p).first, (*p).second);

	Log("\n");
	Log("Ranges:\n");
	Log(" Lo Length   Hi Length     Lo Rate     Hi Rate    Int.Rate     BinRate   BinErr    RandMark\n");
	Log("==========  ==========  ==========  ==========  ==========  ==========  =======  ==========\n");

	N = SIZE(m_Ranges);
	for (unsigned i = 0; i < N; ++i)
		{
		const RateRange &RR = m_Ranges[i];

		Log("%10"PF, RR.LoLength);
		Log("  %10"PF, RR.HiLength);
		Log("  %10.3g", RR.LoRate);
		Log("  %10.3g", RR.HiRate);
		Log("  %10.3g", RR.TotalRate);
		Log("  %10.3g", RR.BinRate);
		if (fabs(RR.BinErr) >= 0.01)
			Log("  %+6.1f%%", RR.BinErr*100.0);
		else
			Log("         ");
		if (RR.RandMark == UINT_MAX && i + 1 != N)
			Log("           *");
		else
			Log("  %10u", RR.RandMark);

		Log("\n");
		}
	}

//void RateMgr::GetBinRates(vector<double> &Rates, bool DoMicro) const
//	{
//	const unsigned N = SIZE(m_Ranges);
//	double SumRate = 0;
//	for (unsigned i = 0; i < N; ++i)
//		{
//		const RateRange &RR = m_Ranges[i];
//		if (DoMicro == RR.IsMicro())
//			SumRate += RR.TotalRate;
//		}
//
//	if (SumRate == 0)
//		Rates.resize(N, 0);
//	else
//		{
//		const vector<unsigned> &WeightVec =
//		  (DoMicro ? m_MicroWeightVec : m_BigWeightVec);
//		GetWeightsFromWeightVec(WeightVec, Rates);
//		if (SIZE(Rates) < N)
//			Rates.resize(N, 0);
//		Normalize(Rates, SumRate);
//		}
//	}
//
//void RateMgr::SetRangeWeightVec(bool DoMicro)
//	{
//	vector<unsigned> &WeightVec = (DoMicro ? m_MicroWeightVec : m_BigWeightVec);
//	WeightVec.clear();
//
//	const unsigned N = SIZE(m_Ranges);
//	if (N == 0)
//		return;
//
//	vector<double> Weights(N);
//	double TotalRate = 0;
//	double MinRate = DBL_MAX;
//	for (unsigned i = 0; i < N; ++i)
//		{
//		const RateRange &RR = m_Ranges[i];
//		if (RR.IsMicro() == DoMicro)
//			{
//			double R = RR.TotalRate;
//			TotalRate += R;
//			if (R > 0 && R < MinRate)
//				MinRate = R;
//			Weights[i] = R;
//			}
//		else
//			Weights[i] = 0;
//		}
//	double BinsNeeded = (TotalRate/MinRate)*5;
//	if (BinsNeeded > P_MaxBins)
//		Die("RateMgr::SetRangeWeightVec(%s), too many bins needed", TypeStr());
//	if (BinsNeeded < P_MinBins)
//		BinsNeeded = P_MinBins;
//
//	SetWeightVec(Weights, WeightVec, (unsigned) BinsNeeded);
//	SetBinRates(DoMicro);
//	}

void RateMgr::SetBinRates()
	{
	double TotalRate = GetTotalRate();

	const unsigned N = SIZE(m_Ranges);
	double AccumRate = 0;
	unsigned LastMark = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		RateRange &RR = m_Ranges[i];
		if (RR.TotalRate == 0)
			{
			RR.RandMark = UINT_MAX;
			RR.BinRate = 0;
			RR.BinErr = 0;
			}
		else
			{
			AccumRate += RR.TotalRate;
			RR.RandMark = (unsigned) (AccumRate*UINT_MAX/TotalRate);
			RR.BinRate = (double) TotalRate*(RR.RandMark - LastMark)/(double) UINT_MAX;
			RR.BinErr = (RR.BinRate - RR.TotalRate)/RR.TotalRate;
			LastMark = RR.RandMark;
			}
		}
	}

void RateMgr::LogParams() const
	{
	Log("\n");
	if (m_Type == RT_NCCE)
		{
		Log("NCCELengthDist=\n", TypeStr());
		Log("#     Length      Weight\n");
		Log("# ==========  ==========\n");
		}
	else
		{
		Log("%sRates=\n", TypeStr());
		Log("#     Length        Rate\n");
		Log("# ==========  ==========\n");
		}
	const unsigned N = SIZE(m_Rates); 
	for (unsigned i = 0; i < N; ++i)
		{
		//bool PrevZero = (i == 0 ? true : m_Rates[i-1].second == 0);
		//bool ThisZero = (m_Rates[i].second == 0);
		//bool NextZero = (i + 1 == N ? true : m_Rates[i+1].second == 0);
		//bool Delete = (PrevZero && ThisZero && NextZero);
		//if (!Delete)
		//	{
		//	const LRPair &LR = m_Rates[i];
		//	Log("  %10"PF", %10.3g;\n", LR.first, LR.second);
		//	}
		const LRPair &LR = m_Rates[i];
		Log("  %10"PF", %10.3g;\n", LR.first, LR.second);
		}
	Log("//\n");
	}

double RateMgr::GetTotalRate() const
	{
	double Total = 0;
	const unsigned N = SIZE(m_Ranges);
	for (unsigned i = 0; i < N; ++i)
		{
		const RateRange &RR = m_Ranges[i];
		Total += RR.TotalRate;
		}
	return Total;
	}

void RateMgr::GetStats(RateStats &Stats) const
	{
	Stats.TotalRate = GetTotalRate();
	Stats.Rate1to10 = 0;
	Stats.Rate1to99 = 0;
	Stats.Rate100plus = 0;
	Stats.Bases1to10 = 0;
	Stats.Bases1to99 = 0;
	Stats.Bases100plus = 0;
	for (gpos_t Length = 1; Length <= 10; ++Length)
		{
		double Rate = GetRate(Length);
		Stats.Rate1to10 += Rate;
		Stats.Bases1to10 += Rate*Length;
		}
	for (gpos_t Length = 1; Length <= 99; ++Length)
		{
		double Rate = GetRate(Length);
		Stats.Rate1to99 += Rate;
		Stats.Bases1to99 += Rate*Length;
		}

	Stats.Rate100plus = 0;
	Stats.Bases100plus = 0;
	const unsigned N = SIZE(m_Ranges);
	for (unsigned i = 0; i < N; ++i)
		{
		const RateRange &RR = m_Ranges[i];
		if (RR.LoLength >= 100)
			{
			double Bases = 
			  ComputeBasesAffected(RR.LoLength, RR.HiLength, RR.LoRate, RR.HiRate);
			Stats.Rate100plus += RR.TotalRate;
			Stats.Bases100plus += Bases;
			}
		else
			{
			for (gpos_t L = 100; L <= RR.HiLength; ++L)
				{
				double Rate = GetRate(L);
				Stats.Rate100plus += Rate;
				Stats.Bases100plus += Rate*L;
				}
			}
		}
	}

double RateMgr::GetRate(gpos_t Length) const
	{
	const unsigned N = SIZE(m_Ranges);
	for (unsigned i = 0; i < N; ++i)
		{
		const RateRange &RR = m_Ranges[i];
		if (Length >= RR.LoLength && Length <= RR.HiLength)
			return Interpolate(Length, RR.LoLength, RR.HiLength, RR.LoRate, RR.HiRate);
		}
	return 0;
	}

RATE_TYPE RateMgr::GetType() const
	{
	return m_Type;
	}

const char *RateMgr::TypeStr() const
	{
	return RateTypeToStr(m_Type);
	}

gpos_t RateMgr::GetLength() const
	{
	unsigned r = RandInt(UINT_MAX-1);
	const unsigned N = SIZE(m_Ranges);
	for (unsigned i = 0; i < N; ++i)
		{
		const RateRange &RR = m_Ranges[i];
		if (r <= RR.RandMark)
			return GetRandLengthRR(RR);
		}
	LogMe();
	Die("RateMgr::GetLength(), rand mark test failed");
	UNREACHABLE_RETURN(GPOS_T_MAX);
	}

void RateMgr::StoreResult(EVENT_RESULT Result, gpos_t Length)
	{
	unsigned iResult = (unsigned) Result;
	asserta(iResult < 3);

	if (Result == ER_Accept)
		m_TotalAcceptedBases += Length;
	else if (Result == ER_Reject)
		m_TotalRejectedBases += Length;

	double Log10_Length = log10((double) Length);
	unsigned Bin = Length == 0 ? 0 : (unsigned) floor(Log10_Length) + 1;
	asserta(Bin < 16);
	++(m_ResultCounts[iResult][Bin]);
	}

void RateMgr::ToFile(FILE *f) const
	{
	WRITEV(m_Type);
	const unsigned RateCount = SIZE(m_Rates);
	WRITEV(RateCount);
	for (unsigned i = 0; i < RateCount; ++i)
		{
		gpos_t L = m_Rates[i].first;
		double R = m_Rates[i].second;
		WRITEV(L);
		WRITEV(R);
		}

	const unsigned TrimmedRateCount = SIZE(m_TrimmedRates);
	WRITEV(TrimmedRateCount);
	for (RateList::const_iterator p = m_TrimmedRates.begin();
	  p != m_TrimmedRates.end(); ++p)
		{
		const LRPair &LR = *p;
		WRITEV(LR.first);
		WRITEV(LR.second);
		}

	const unsigned RangeCount = SIZE(m_Ranges);
	WRITEV(RangeCount);
	for (unsigned i = 0; i < RangeCount; ++i)
		{
		const RateRange &RR = m_Ranges[i];
		WRITEV(RR);
		}

	WRITEV(m_TotalAcceptedBases);
	WRITEV(m_TotalRejectedBases);
	for (unsigned i = 0; i < 3; ++i)
		for (unsigned j = 0; j < 16; ++j)
			WRITEV(m_ResultCounts[i][j]);
	}

void RateMgr::FromFile(FILE *f)
	{
	READV(m_Type);
	unsigned RateCount;
	READV(RateCount);
	m_Rates.resize(RateCount);
	for (unsigned i = 0; i < RateCount; ++i)
		{
		gpos_t L;
		double R;
		READV(L);
		READV(R);
		m_Rates[i].first = L;
		m_Rates[i].second = R;
		}

	unsigned TrimmedRateCount;
	READV(TrimmedRateCount);
	m_TrimmedRates.resize(TrimmedRateCount);
	for (unsigned i = 0; i < TrimmedRateCount; ++i)
		{
		gpos_t L;
		double R;
		READV(L);
		READV(R);
		m_TrimmedRates.push_back(LRPair(L, R));
		}

	unsigned RangeCount;
	READV(RangeCount);
	m_Ranges.resize(RangeCount);
	for (unsigned i = 0; i < RangeCount; ++i)
		{
		RateRange &RR = m_Ranges[i];
		READV(RR);
		}

	READV(m_TotalAcceptedBases);
	READV(m_TotalRejectedBases);
	for (unsigned i = 0; i < 3; ++i)
		for (unsigned j = 0; j < 16; ++j)
			READV(m_ResultCounts[i][j]);
	}
