#include <string>
#include <vector>
#include "utils.h"
#include "intlist.h"
#include "evomodel.h"
#include "codon.h"

using namespace std;

bool ReadEvoParamsLine(FILE *f, string &Line);
void ParseStrVecWS(const std::string &In, std::vector<std::string> &Out);
char *CodonToStr(CODON c);

#define TRACE	0

// [21] needed for trailing nul, there are 20 chars.
static char g_AAChars[21] = "ACDEFGHIKLMNPQRSTVWY";

static const int MAX_LINE = 4096;
static char g_Heading[20];
static double g_AAMx[20][20] =
	{
#define _ (-1)	// invalid value
#include "aamx.h"
#undef _
	};

double GetAASubstProb(unsigned AminoIndexBefore, unsigned AminoIndexAfter)
	{
	asserta(AminoIndexBefore < 20);
	asserta(AminoIndexAfter < 20);
	return g_AAMx[AminoIndexBefore][AminoIndexAfter];
	}

static char LetterToChar(unsigned i)
	{
	asserta(i < 20);
	return g_AAChars[i];
	}

static unsigned CharToLetter(char c)
	{
	for (unsigned i = 0; i < 20; ++i)
		if (g_AAChars[i] == c)
			return i;
	Die("Invalid amino acid '%c'", c);
	UNREACHABLE_RETURN(UINT_MAX);
	}

static bool IsResidueChar(char c)
	{
	for (unsigned i = 0; i < 20; ++i)
		if (g_AAChars[i] == c)
			return true;
	return false;
	}

void LogAAMx()
	{
	Log("\n");
	Log("AminoAcidSubstProbs=\n");
	Log("     ");
	for (int i = 0; i < 20; ++i)
		Log(" %7c", LetterToChar(i));
	Log("\n");

	for (int i = 0; i < 20; ++i)
		{
		double SumRow = 0;
		Log("%c    ", LetterToChar(i));
		for (int j = 0; j < 20; ++j)
			{
			double p = g_AAMx[i][j];
			if (p == -1)
				Log("       .");
			else
				{
				Log(" %7.5f", p);
				SumRow += p;
				}
			}
		Log("  # %c = %7.5f\n", LetterToChar(i), SumRow);
		}
	Log("\n");
	}

void ReadAAMx(FILE *f)
	{
// Find column headers
	string strLine;
	const char *Line = 0;
	for (;;)
		{
		bool EndOfFile = ReadEvoParamsLine(f, strLine);
		if (EndOfFile)
			Die("Premature EOF in amino acid matrix");

		if (strLine.empty())
			continue;
		Line = strLine.c_str();
		if (Line[0] == '#')
			continue;
		else if (Line[0] == 'A') // leading white space stripped by ReadEvoParamsLine
			break;
		else
			Die("Invalid line in amino acid matrix: '%s'", Line);
		}

	BoolVec AADone(20);

// Read column headers
	unsigned HeadingCount = 0;
	for (const char *p = Line; *p; ++p)
		{
		char c = *p;
		if (!isspace(c))
			{
			if (!IsResidueChar(c))
				Die("Invalid column header '%c' in amino acid matrix", c);
			if (HeadingCount >= 20)
				Die("Too many column headings in amino acid matrix '%s'", Line);
			g_Heading[HeadingCount++] = c;
			}
		}

	if (HeadingCount != 20)
		Die("Error in matrix file: expected 20 column headers, line='%s'", Line);

// Zero out matrix
	for (int i = 0; i < 20; ++i)
		for (int j = 0; j < 20; ++j)
			g_AAMx[i][j] = 0.0;

// Read data lines
	for (unsigned RowIndex = 0; RowIndex < 20; ++RowIndex)
		{
		bool EndOfFile = ReadEvoParamsLine(f, strLine);
		if (EndOfFile)
			Die("Premature EOF in amino acid matrix");

		if (strLine.empty())
			continue;
		Line = strLine.c_str();
		if (Line[0] == '#')
			continue;

		vector<string> Fields;
		ParseStrVecWS(Line, Fields);

		if (Fields.size() != 21)
			Die("Expected 21 fields, got %u, in amino acid matrix row '%s'",
			  SIZE(Fields), Line);

		if (Fields[0].size() != 1)
			Die("Invalid amino acid letter '%s' in row '%s",
			  Fields[0].c_str(), Line);

		char c = Fields[0][0];
		if (!IsResidueChar(c))
			Die("Invalid amio acid letter '%c' in row '%s'", c, Line);

		unsigned RowLetter = CharToLetter(c);
		if (AADone[RowLetter])
			Die("Amino acid '%c' row specified twice", c);
		AADone[RowLetter] = true;

		for (unsigned Col = 0; Col < 20; ++Col)
			{
			char HeaderChar = g_Heading[Col];
			unsigned ColLetter = CharToLetter(HeaderChar);
			const char *Value = Fields[Col+1].c_str();
			if (strcmp(Value, ".") == 0)
				{
				g_AAMx[RowLetter][ColLetter] = -1;
				continue;
				}
			double v = atof(Value);
			if (v < 0.0 || v > 1.0)
				Die("Invalid probability '%g' in amino acid matrix row '%s'",
				  v, Line);
			if (RowLetter == ColLetter && v != 0)
				Die("Invalid amino acid matrix, %c->%c=%g, diagonal must be zero",
				  c, c, v);
			g_AAMx[RowLetter][ColLetter] = v;
			}
		}
	}

#if	0
void InitAAMx()
	{
	for (int i = 0; i < 20; ++i)
		{
		for (int j = 0; j < 20; ++j)
			{
			double p = g_AAMx[i][j];
			if (p < 0)
				continue;
			p *= P_AminoAcidProbMultiplier;
			if (p < 0 || p > 1)
				Die("Invalid amino acid prob %s-%s=%.3g (AminoAcidProbMultiplier=%.3g)",
				  CodonToStr((CODON) i),
				  CodonToStr((CODON) j),
				  p,
				  P_AminoAcidProbMultiplier);
			g_AAMx[i][j] = p;
			}
		}
	}
#endif // 0
