#include "revfile.h"
#include "utils.h"
#include "getopts.h"
#include "filter4.h"

#define	TRACE	0

/***
TODO: Doesn't consider the case where micro-chain A is
covered by a hole in another micro-chain B, in which case
A should be kept (it is a translocation in B).

Heuristic filter for deleting redundant HAIs due to repeats.
For example, here is the result of a tandem repeat found in both
genomes:

      B
A    /
/   /    
   /   C
  /   /
 /

Here we want to delete HAIs A and C, keeping B. The basic
idea is to discard an HSP that is covered by another HSP w.r.t.
both genomes.

In practice, using coverage based on single HAIs works
quite well. The most obvious mistakes are where the longer
alignment (B above) is broken up into two or more HAIs
separated by short gaps. In such cases, "outliers" like
A and C sometimes cover one of the HAIs in B, giving
a sub-optimal result. To fix this, we use "micro-chaining".
The goal is to make chains of HAIs that belong to a single
alignment like B above. Coverage can then be considered
with respect to micro-chains rather than individual HAIs.
The algorithm is similar to bundling. The main difference
is that the in- and out-degree of each node (HSP) is required
to be 1. If two or more target nodes are close enough to
a given node to be candidates for edges, the node that is
the best fit to a micro-chain is chosen. Goodness of fit
is a weighted combination of factors including being close
to the same diagonal (most important), separated by a
short distance and having little or no overlap (overlaps
are heavily penalized).
***/

extern RevFile *g_ptrOut;
extern RevFile *g_ptrDel;

extern unsigned g_ChrPairDeletedCount;

extern unsigned g_GenomeIndexA;
extern unsigned g_GenomeIndexB;

unsigned g_WindowLength = 1024;
unsigned g_MinusWindow = 64;
unsigned g_DiagFactor = 10;
unsigned g_LenFactor = 5;
unsigned g_OvFactor = 15;

typedef bool (*ptrLinkFunction)(const AlnInfo &AI1, const AlnInfo &AI2);
typedef unsigned (*ptrDistanceFunction)(const AlnInfo &AI1, const AlnInfo &AI2);

void Chain(const AIVec &InAIs, unsigned GenomeIndex, ptrLinkFunction Link,
  unsigned MaxWidth, ptrDistanceFunction Distance, IntVecVec &Chains);

// AIA covers AIB w.r.t. either genomes?
bool Covers(const AlnInfo &AIA, const AlnInfo &AIB, unsigned GenomeIndexA,
  unsigned GenomeIndexB);

void OptimalCover(const AlnInfo &Region, const AIVec &AIs, BoolVec &Used);

unsigned MicroChainDistance(const AlnInfo &AI1, const AlnInfo &AI2);

bool MicroChainLinkAIs(const AlnInfo &AI1, const AlnInfo &AI2);

void LogAIN(const GCSeg &AI);

struct MCInfo
	{
	AlnInfo AI;
	IntVec AIIndexes;
	std::list<GCSeg> Gaps;
	};

#if	TRACE
static void LogMCs(const std::vector<MCInfo> &MCs, const BoolVec Used,
  const AIVec &InAIs)
	{
	assert(MCs.size() == Used.size());
	const unsigned MCCount = (unsigned) MCs.size();
	for (unsigned MCIndex = 0; MCIndex < MCCount; ++MCIndex)
		{
		const MCInfo &MC = MCs[MCIndex];
		Log("\n");
		Log("MC%u Used=%d AI=", MCIndex, Used[MCIndex]);
		LogAI(MC.AI);
		Log("\n");
		unsigned AICount = (unsigned) MC.AIIndexes.size();
		for (unsigned i = 0; i < AICount; ++i)
			{
			Log("  ");
			unsigned AIIndex = MC.AIIndexes[i];
			LogAIN(InAIs[AIIndex]);
			}
		}
	Log("\n");
	}
#endif

// Repfix filter
void RepfixFilter4MicroChain(const AIVec &InAIs, const IntVec &AIIndexes)
	{
	g_GenomeIndexA = GetFilterGenomeIndexA();
	g_GenomeIndexB = GetFilterGenomeIndexB();

	const unsigned AICount = (unsigned) InAIs.size();
	if (AICount == 0)
		return;

	unsigned MaxWidth = g_WindowLength + g_MinusWindow;

	IntVecVec MicroChains;
	Chain(InAIs, g_GenomeIndexA, MicroChainLinkAIs, MaxWidth, MicroChainDistance,
	  MicroChains);

	const AlnInfo &FirstAI = InAIs.front();
	AlnInfo Region = FirstAI;

	const GCSeg &FirstAI_SegA = FirstAI.GetSeg(g_GenomeIndexA);
	const GCSeg &FirstAI_SegB = FirstAI.GetSeg(g_GenomeIndexB);

	unsigned MicroChainCount = (unsigned) MicroChains.size();
	std::vector<MCInfo> MCs;
	MCs.resize(MicroChainCount);
	for (unsigned MCIndex = 0; MCIndex < MicroChainCount; ++MCIndex)
		{
		MCInfo &MC = MCs[MCIndex];

		MC.AI.Score = 0;

		GCSeg Seg;
		Seg.GenomeIndex = g_GenomeIndexA;
		Seg.ChrId = FirstAI_SegA.ChrId;
		Seg.hi = 0;
		Seg.lo = 0;
		Seg.Plus = true;

		MC.AI.Segs.push_back(Seg);

		Seg.GenomeIndex = g_GenomeIndexB;
		Seg.ChrId = FirstAI_SegB.ChrId;

		MC.AI.Segs.push_back(Seg);

		coord_t minA = COORD_T_MAX;
		coord_t minB = COORD_T_MAX;
		coord_t maxA = 0;
		coord_t maxB = 0;
		const IntVec LocalAIIndexes = MicroChains[MCIndex];
		for_CIntVec(LocalAIIndexes, p)
			{
			unsigned LocalAIIndex = *p;
			assert(LocalAIIndex < AICount);
			const AlnInfo &AI = InAIs[LocalAIIndex];

			const GCSeg &SegA = AI.GetSeg(g_GenomeIndexA);
			const GCSeg &SegB = AI.GetSeg(g_GenomeIndexB);

			bool Plus = (SegA.Plus == SegB.Plus);

			MC.AI.GetModifiableSegPtr(g_GenomeIndexA)->Plus = true;
			MC.AI.GetModifiableSegPtr(g_GenomeIndexB)->Plus = Plus;

			MC.AIIndexes.push_back(LocalAIIndex);

			if (SegA.lo < minA)
				minA = SegA.lo;
			if (SegB.lo < minB)
				minB = SegB.lo;
			if (SegA.hi > maxA)
				maxA = SegA.hi;
			if (SegB.hi > maxB)
				maxB = SegB.hi;
			}
		MC.AI.Score = 0;
		MC.AI.GetModifiableSegPtr(g_GenomeIndexA)->lo = minA;
		MC.AI.GetModifiableSegPtr(g_GenomeIndexA)->hi = maxA;
		MC.AI.GetModifiableSegPtr(g_GenomeIndexB)->lo = minB;
		MC.AI.GetModifiableSegPtr(g_GenomeIndexB)->hi = maxB;

		const AlnInfo &MCAI = MC.AI;
		if (MCAI.GetSeg(g_GenomeIndexA).lo < Region.GetModifiableSegPtr(g_GenomeIndexA)->lo)
			Region.GetModifiableSegPtr(g_GenomeIndexA)->lo = MCAI.GetSeg(g_GenomeIndexA).lo;

		if (MCAI.GetSeg(g_GenomeIndexA).hi > Region.GetModifiableSegPtr(g_GenomeIndexA)->hi)
			Region.GetModifiableSegPtr(g_GenomeIndexA)->hi = MCAI.GetSeg(g_GenomeIndexA).hi;

		if (MCAI.GetSeg(g_GenomeIndexB).lo < Region.GetModifiableSegPtr(g_GenomeIndexB)->lo)
			Region.GetModifiableSegPtr(g_GenomeIndexB)->lo = MCAI.GetSeg(g_GenomeIndexB).lo;

		if (MCAI.GetSeg(g_GenomeIndexB).hi > Region.GetModifiableSegPtr(g_GenomeIndexB)->hi)
			Region.GetModifiableSegPtr(g_GenomeIndexB)->hi = MCAI.GetSeg(g_GenomeIndexB).hi;
		}

// Convert MCs into vector of AIs (needed by OptimalCover)
	AIVec MCAIs;
	for (unsigned MCIndex = 0; MCIndex < MicroChainCount; ++MCIndex)
		{
		MCInfo &MC = MCs[MCIndex];
		MCAIs.push_back(MC.AI);
		}

// Find "optimal cover" of chains
	BoolVec Used;
	OptimalCover(Region, MCAIs, Used);

#if	TRACE
	Log("\n");
	Log("RepfixFilter4MicroChain, MCs=\n");
	LogMCs(MCs, Used, InAIs);
	Log("\n");
#endif

	const RevFile &In = GetFilterInputRevFile();
	for (unsigned MCIndex = 0; MCIndex < MicroChainCount; ++MCIndex)
		{
		const MCInfo &MC = MCs[MCIndex];
		bool MCUsed = Used[MCIndex];

		const IntVec &LocalAIIndexes = MC.AIIndexes;
		for_CIntVec(LocalAIIndexes, p)
			{
			unsigned LocalAIIndex = *p;
			unsigned AIIndex = AIIndexes[LocalAIIndex];
			if (MCUsed)
				g_ptrOut->CopyAln(In, AIIndex);
			else
				{
				++g_ChrPairDeletedCount;
				if (g_ptrDel != 0)
					g_ptrDel->CopyAln(In, AIIndex);
				}
			}
		}
	}
