#ifndef REVFILE_H
#define REVFILE_H

#include <stdio.h>
#include <map>
#include <vector>
#include "geninfo.h"
#include "intlist.h"
#include "revfile_disk.h"
#include "alninfo.h"
#include "iface.h"
#include <set>

#define VALIDATE	0
#define PROGRESS	0

#ifdef _MSC_VER
#pragma warning(disable : 4800)	// int->bool conversion
#endif

void RevComp(char *S, unsigned L);
void RevComp(StdStr &S);

// for getpid():
#ifdef	_MSC_VER
#include <process.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif

#define ReadData_(v)		ReadStdioFile(m_f, &v, (unsigned) sizeof(v))
#define WriteData_(v)	WriteStdioFile(m_f, &v, (unsigned) sizeof(v))

#define STATE(x, y)	if (m_State != x) Die("RevFile: function call in invalid state"); m_State = y;
#define STATEC(x)	if (m_State != x) Die("RevFile: function call in invalid state");

typedef vector<pair<unsigned, unsigned> > IntPairVec;
typedef map<GenChr, IntVec> GenChrToIntVecMap;
typedef vector<GenInfo> GenInfoVec;

struct IxLoHi
	{
	unsigned Lo;
	unsigned Hi;
	bool operator<(const IxLoHi &rhs) const
		{
		if (Lo < rhs.Lo)
			return true;
		return Hi < rhs.Hi;
		}
	};

// Letters are always from + strand, so if Plus=false
// then one letter should be rev-comp'd to recover the
// alignment.
struct PairwiseCol
	{
	unsigned GenomeIndex1;
	unsigned GenomeIndex2;
	unsigned ChrId1;
	unsigned ChrId2;
	unsigned Pos1;	// 0-based w.r.t. plus strand
	unsigned Pos2;	// 0-based w.r.t. plus strand
	char Letter1;	// plus strand letter
	char Letter2;	// plus strand letter
	bool Plus;		// false iff opposite strands aligned.

	bool HasGap() const
		{
		return Letter1 == '-' || Letter2 == '-';
		}

	bool IsMatch() const
		{
		char CompLetter(char c);

		if (HasGap())
			return false;
		if (Plus)
			return toupper(Letter1) == toupper(Letter2);
		else
			return toupper(Letter1) == toupper(CompLetter(Letter2));
		}

	bool operator<(const PairwiseCol &rhs) const
		{
		if (GenomeIndex1 != rhs.GenomeIndex1)
			return GenomeIndex1 < rhs.GenomeIndex1;
		if (GenomeIndex2 != rhs.GenomeIndex2)
			return GenomeIndex2 < rhs.GenomeIndex2;
		if (ChrId1 != rhs.ChrId1)
			return ChrId1 < rhs.ChrId1;
		if (ChrId2 != rhs.ChrId2)
			return ChrId2 < rhs.ChrId2;
		if (Plus != rhs.Plus)
			return (Plus ? 1 : 0) < (rhs.Plus ? 1 : 0); // avoid bool->int warning
		if (Pos1 != rhs.Pos1)
			return Pos1 < rhs.Pos1;
		return Plus ? Pos2 < rhs.Pos2 : Pos2 > rhs.Pos2;
		}
	};

typedef void (*ON_PAIRWISE_COL)(unsigned User, const PairwiseCol &Col);

class RevFile
	{
protected:
	enum RevFileState
		{
		Closed,
		Reading,
		Writing
		};

public:
	RevFile();
	~RevFile();
	void Close();
	const string &GetFileName() const;

// Members for reading existing file
// =================================
	void Open(const string &FileName);
	const GenInfoVec &GetGenomes() const;

	unsigned GetMagic() const;
	bool CanCompressAI(const AlnInfo &AI) const;

// Get AlnInfo, return AlnIndex.
	unsigned GetAlnInfo(unsigned AlnInfoIndex, AlnInfo &AI) const;
	unsigned GetAlnInfoX(unsigned AlnInfoIndex, AlnInfoX &AIX) const;

	unsigned GetAlnRows(unsigned AlnIndex, StdStrVec &Rows) const;

// GetGenomeCount/Info are redundant because could get
// via GetGenomes; provided for convenience.
	unsigned GetGenomeCount() const;
	const GenInfo &GetGenomeInfo(unsigned GenomeIndex) const;

	GenInfo &GetModifiableGenomeInfo(unsigned GenomeIndex);
	GenInfo *GetModifiableGenomeInfoPtr(const string &GenomeName,
	  unsigned *ptrGenomeIndex);

// GetGenomeName, GetChrCount and GetChrInfo are redundant because
// could get via GenInfo; provided for convenience.
	const string &GetGenomeName(unsigned GenomeIndex) const;
	unsigned GetGenomeIndex(const string &GenomeName) const;
	unsigned GetChrCount(unsigned GenomeIndex) const;
	const ChrInfo &GetChrInfo(unsigned GenomeIndex, unsigned ChrId) const;
	const ChrInfo *GetChrInfo(unsigned GenomeIndex, const string &ChrName) const;

	const string &GetChrName(unsigned GenomeIndex, unsigned ChrId) const;
	unsigned GetChrId(unsigned GenomeIndex, const string &ChrName) const;
	unsigned GetChrId(unsigned GenomeIndex, unsigned ChrIndex) const;
	unsigned GetChrIndex(unsigned GenomeIndex, unsigned ChrId) const;
	coord_t GetChrLength(unsigned GenomeIndex, unsigned ChrId) const;

// Members for enumerating sequence data.
	unsigned GetSeqCount() const;
	void GetSeqByIndex(unsigned SeqIndex, char *Seq, unsigned Bytes,
	  unsigned *ptrGenomeIndex, unsigned *ptrChrId) const;
	void GetSeqGenChrByIndex(unsigned SeqIndex, unsigned *ptrGenomeIndex,
	  unsigned *ptrChrId) const;
	coord_t GetSeqLengthByIndex(unsigned SeqIndex) const;

// Members for getting sequence data by genome & chromosome
	bool IsSeqPresent(unsigned GenomeIndex, unsigned ChrId) const;
	coord_t GetSeqLength(unsigned GenomeIndex, unsigned ChrId) const;
	void GetSeq(unsigned GenomeIndex, unsigned ChrId, char *Seq, unsigned Bytes) const;
	void GetSubSeq(unsigned GenomeIndex, unsigned ChrId, char *Seq,
	  unsigned From, unsigned Length, bool RevComp = false) const;
	void GetSubSeq2(unsigned GenomeIndex, unsigned ChrId, StdStr &Seq,
	  unsigned From, unsigned Length, bool RevComp = false) const;
	void GetSubSeq3(const GCSeg &Seg, StdStr &Seq) const;

// Not necessarily 1-1 correspondence between AlnInfo's & AlnRows's.
// Can have an AlnInfo without an explicit alignment, i.e. without
// rows, but not vice versa.
	unsigned GetAlnInfoCount() const;

// Basic query:
//		Input is (gen1, chr1), (gen2, chr2) ... (genN, chrN)
//			Each pair must be present, so is (gen1, chr1) AND (gen2, chr2) AND ...
//		Output is vector of AlnInfoIndexes.
	void GetAppendAlnInfoIndexesChr(unsigned GenIndex, unsigned ChrId,
	  IntSet &AlnInfoIndexes) const;

// Get AlnInfo[X]s given vector of AlnInfoIndexes.
// Current implemented by naive loop over GetAlnInfo(), but
// could be optimized for speed in future if needed.
	void GetAlnInfos(const IntVec &AIIndexes, AIVec &AIs) const;
	void GetAlnInfoXs(const IntVec &AIIndexes, AIXVec &AIXs) const;
	void GetAlnInfos(const IntSet &AIIndexes, AIVec &AIs) const;

// Special case: get all AlnInfo[X]s.
// Commonly needed operation, allows internal optimization later if desired.
	void GetAllAlnInfos(AIVec &AIs, bool ShowProgress = true) const;
	void GetAllAlnInfoXs(AIXVec &AIXs, bool ShowProgress = true) const;

	void GetAppendAIIxRangesChr(unsigned GenIndex, unsigned ChrId,
	  IntPairVec &AlnInfoIndexLoHis) const;
	void GetAppendAIIxRangesChr(unsigned GenIndex, unsigned ChrId,
	  IntPairSet &AlnInfoIndexLoHis) const;
	void GetAppendAlnInfoIndexesChr(unsigned GenIndex, unsigned ChrId,
	  IntVec &AlnInfoIndexes) const;

// Convert AlnInfoIndex <-> AlnIndex.
	unsigned GetAlnIndex(unsigned AlnInfoIndex) const;

	void GetActiveChrPairs(unsigned GenomeIndexA, unsigned GenomeIndexB,
	  set<IntPair> &ChrIdPairs) const;
	void GetActiveChrPairsSlow(unsigned GenomeIndexA, unsigned GenomeIndexB,
	  set<IntPair> &ChrIdPairs) const;

// Members for creating new file
// =============================
	void Create(const string &FileName, const GenInfoVec &Genomes);
	void Create(const string &FileName);

	unsigned AddGenome(const string &Name);

// Add a new alignment, return AlnInfoIndex (not AlnIndex!).
// AlnInfoIndex can be used to get AlnIndex if needed.
	unsigned AddAln(const AlnInfo &AI, const StdStrVec &Rows);

// Add a new sequence, return SeqIndex
	unsigned AddSeq(const string &GenomeName, const string &ChrName,
	  const char *Seq, coord_t Length);

// Check if segment lengths in AI.Segs[] match (ungapped) sequence
// lengths in Rows[], if not issue error / warning message.
	bool ValidateSegLengths(const AlnInfo &AI, const StdStrVec &Rows,
	  bool DieOnError = true);

	unsigned AddAlnInfo(const AlnInfo &AI);
	unsigned AddAlnInfo(const AlnInfo &AI, unsigned AlnIndex);
	void CopyAln(const RevFile &rhs, unsigned rhsAlnInfoIndex);

	void EnumPairwiseCols(unsigned GenomeIndex1, unsigned ChrId1,
	  unsigned GenomeIndex2, unsigned ChrId2,
	  ON_PAIRWISE_COL OnCol, unsigned User, bool ShowProgress = true) const;
	void GetPairwiseCols(unsigned GenomeIndex1, unsigned ChrId1,
	  unsigned GenomeIndex2, unsigned ChrId2,
	  set<PairwiseCol> &Cols) const;

private:
	void CloseReading();
	void CloseWriting();
	void AddHashEntry(unsigned GenomeIndex, unsigned ChrId, uint64 Pos,
	  unsigned AlnCount);
	HashTableEntry *FindHashEntry(unsigned GenomeIndex, unsigned ChrId) const;
	uint64 StoreBitRows(const StdStrVec &Rows);
	void GetAlnRowsInPlace(const DiskAlnInfoTOCEntryC &DATEC, StdStrVec &Rows) const;

	void AddToIndexList(unsigned GenomeIndex, unsigned ChrId,
	  unsigned AlnInfoIndex);
	void WriteHashTable();
	void WriteAlnInfoTOC();

	void WriteGenomes();
	void WriteAlnInfoSegs();
	void WriteAlnTOC();
	void WriteAlns();
	void WriteSeqs();

	void GetDATE(unsigned AlnInfoIndex, DiskAlnInfoTOCEntry &DATE) const;
	void GetDATEC(unsigned AlnInfoIndex, DiskAlnInfoTOCEntryC &DATEC) const;
	unsigned GetAlnInfoDATE(const DiskAlnInfoTOCEntry &DATE,
	  unsigned AlnInfoIndex, AlnInfo &AI) const;
	void GetAlnInfoDATECompressed(const DiskAlnInfoTOCEntryC &DATEC,
	  AlnInfo &AI) const;
	void AddAlnInfoCompress(const AlnInfo &AI, const byte *Rows = 0,
	  uint64 FilePosRows = UINT64_MAX);
	void GetAlnRowsCompressed(unsigned AlnIndex, StdStrVec &Rows) const;

private:
	string m_FileName;
	RevFileState m_State;
	GenInfoVec m_Genomes;
	unsigned m_AlnInfoCount;
	unsigned m_AlnCount;
	unsigned m_HashTableEntryCount;
	HashTableEntry *m_Hash;

// Valid when writing only
// Tmp files
	string m_AlnInfoTOCFileName;
	string m_AlnInfoSegsFileName;
	string m_AlnTOCFileName;
	string m_AlnsFileName;
	string m_SeqsFileName;
	FILE *m_fAlnInfoTOC;
	FILE *m_fAlnInfoSegs;
	FILE *m_fAlnTOC;
	FILE *m_fAlns;
	FILE *m_fSeqs;

// Map (GenIndex,ChrId) to SeqIndex.
	map<IntPair, unsigned> m_SeqTOC;
	vector<LongIntPair> m_SeqPosLength;

	RevFileHdr m_Hdr;
	FILE *m_f;

private:
	map<IntPair, set<IxLoHi> > m_GenChrToAIIxRanges;

// Valid when reading only
	mutable unsigned m_DATECacheFirstEntryIndex;
	mutable unsigned m_DATECacheSize;		// in entries, not bytes
	mutable unsigned m_DATECacheEntryCount;
	mutable DiskAlnInfoTOCEntry *m_DATECache;

public:
	mutable unsigned m_AICacheHitCount;
	mutable unsigned m_AICacheRefreshCount;
	unsigned m_CompressedAICount;
	unsigned m_InPlaceRowsCount;
	};

void Trunc64(byte *Target, uint64 i64, unsigned Bytes);
void Untrunc64(const byte *Buffer, unsigned Bytes, uint64 &i64);

#endif	// REVFILE_H
