#include "scoremx.h"
#include "utils.h"
#include <ctype.h>

/***
Symbols:

G, A, T, C: G, A, T, C
Purine (A or G): R
Pyrimidine (T or C): Y
A or T: W
G or C: S
A or C: M
G or T: K
A or T or C: H
G or C or T: B
G or A or C: V
G or A or T: D
G or A or T or C: N (obsolete: X)

Transition: A <-> T or C <-> G.
Transversion: any other substitution.
***/

SCOREMATRIX DNAScoreMx;

#define TRACE	0

const int MX_SIZE = 256;
const int MAX_LINE = 4096;
const int MAX_HEADINGS = 256;
const int MIN_HEADINGS = 4;
static unsigned char Heading[MAX_HEADINGS];
static int HeadingCount = 0;

void InitScoreMatrix(score_t Match, score_t Mismatch, score_t Wildcard)
	{
	score_t Transition = Mismatch;
	score_t Transversion = Mismatch;

	for (int i = 0; i < 256; ++i)
		for (int j = 0; j < 256; ++j)
			DNAScoreMx[i][j] = Transition;

	DNAScoreMx[(int) 'A'][(int) 'A'] = Match;
	DNAScoreMx[(int) 'C'][(int) 'C'] = Match;
	DNAScoreMx[(int) 'G'][(int) 'G'] = Match;
	DNAScoreMx[(int) 'T'][(int) 'T'] = Match;

	DNAScoreMx[(int) 'A'][(int) 'G'] = Transversion;
	DNAScoreMx[(int) 'G'][(int) 'A'] = Transversion;
	DNAScoreMx[(int) 'C'][(int) 'T'] = Transversion;
	DNAScoreMx[(int) 'T'][(int) 'C'] = Transversion;

	DNAScoreMx[(int) 'A'][(int) 'N'] = Wildcard;
	DNAScoreMx[(int) 'C'][(int) 'N'] = Wildcard;
	DNAScoreMx[(int) 'G'][(int) 'N'] = Wildcard;
	DNAScoreMx[(int) 'T'][(int) 'N'] = Wildcard;

	DNAScoreMx[(int) 'N'][(int) 'A'] = Wildcard;
	DNAScoreMx[(int) 'N'][(int) 'C'] = Wildcard;
	DNAScoreMx[(int) 'N'][(int) 'G'] = Wildcard;
	DNAScoreMx[(int) 'N'][(int) 'T'] = Wildcard;

	DNAScoreMx[(int) 'N'][(int) 'N'] = Wildcard;

	DNAScoreMx[(int) 'a'][(int) 'a'] = Match;
	DNAScoreMx[(int) 'c'][(int) 'c'] = Match;
	DNAScoreMx[(int) 'g'][(int) 'g'] = Match;
	DNAScoreMx[(int) 't'][(int) 't'] = Match;

	DNAScoreMx[(int) 'a'][(int) 'g'] = Transversion;
	DNAScoreMx[(int) 'g'][(int) 'a'] = Transversion;
	DNAScoreMx[(int) 'c'][(int) 't'] = Transversion;
	DNAScoreMx[(int) 't'][(int) 'c'] = Transversion;

	DNAScoreMx[(int) 'a'][(int) 'n'] = Wildcard;
	DNAScoreMx[(int) 'c'][(int) 'n'] = Wildcard;
	DNAScoreMx[(int) 'g'][(int) 'n'] = Wildcard;
	DNAScoreMx[(int) 't'][(int) 'n'] = Wildcard;

	DNAScoreMx[(int) 'n'][(int) 'a'] = Wildcard;
	DNAScoreMx[(int) 'n'][(int) 'c'] = Wildcard;
	DNAScoreMx[(int) 'n'][(int) 'g'] = Wildcard;
	DNAScoreMx[(int) 'n'][(int) 't'] = Wildcard;

	DNAScoreMx[(int) 'n'][(int) 'n'] = Wildcard;
	}

static void LogMx()
	{
	Log("Matrix\n");
	Log("     ");
	for (int i = 0; i < HeadingCount; ++i)
		{
		unsigned char c = Heading[i];
		Log("    %c", c);
		}
	Log("\n");

	for (int i = 0; i < HeadingCount; ++i)
		{
		unsigned char ci = Heading[i];
		Log("%c    ", ci);
		for (int j = 0; j < HeadingCount; ++j)
			{
			unsigned char cj = Heading[j];
#if	INTEGER_DP
			Log("%5d", DNAScoreMx[ci][cj]);
#else
			Log("%5g", DNAScoreMx[ci][cj]);
#endif
			}
		Log("\n");
		}
	Log("\n");
	}

//static unsigned MxCharToLetter(unsigned char c)
//	{
//	for (unsigned Letter = 0; Letter < (unsigned) HeadingCount; ++Letter)
//		if (Heading[Letter] == c)
//			return Letter;
//	Die("Letter '%c' has no heading", c);
//	return 0;
//	}

static void ParseGap(const char *Line, score_t *ptrGapOpen, score_t *ptrGapExtend)
	{
#if	INTEGER_DP
	int n = sscanf(Line, "GAP %d %d", ptrGapOpen, ptrGapExtend);
#else
	int n = sscanf(Line, "GAP %g %g", ptrGapOpen, ptrGapExtend);
#endif
	if (n != 2)
		Die("Invalid format for GAP line '%s'", Line);
	if (*ptrGapOpen < 0 || *ptrGapExtend < 0)
		Die("Gap penalties in matrix file must be >= 0");
	*ptrGapOpen *= -1;
	*ptrGapExtend *= -1;
	}

void ReadScoreMx(FILE *f, score_t *ptrGapOpen, score_t *ptrGapExtend)
	{
// Find column headers
	unsigned char Line[MAX_LINE];
	for (;;)
		{
		bool EndOfFile = ReadLineStdioFile(f, (char *) Line, sizeof(Line));
		if (EndOfFile)
			Die("Premature EOF in matrix file");

		if (Line[0] == '#')
			continue;
		else if (Line[0] == ' ')
			break;
		else if (Line[0] == 'G')
			ParseGap((char *) Line, ptrGapOpen, ptrGapExtend);
		else
			Die("Invalid line in matrix file: '%s'", Line);
		}

// Read column headers
	HeadingCount = 0;
	for (unsigned char *p = Line; *p; ++p)
		{
		unsigned char c = *p;
		if (!isspace(c))
			Heading[HeadingCount++] = c;
		}

	if (HeadingCount > 0 && Heading[HeadingCount-1] == '*')
		--HeadingCount;

	if (HeadingCount < MIN_HEADINGS)
		Die("Error in matrix file: < %d headers, line='%s'",
		  Line, MIN_HEADINGS);

#if TRACE
	{
	Log("ReadMx\n");
	Log("%d headings: ", HeadingCount);
	for (int i = 0; i < HeadingCount; ++i)
		Log("%c", Heading[i]);
	Log("\n");
	}
#endif

// Zero out matrix
	for (int i = 0; i < MX_SIZE; ++i)
		for (int j = 0; j < MX_SIZE; ++j)
			DNAScoreMx[i][j] = 0;

// Read data lines
	for (unsigned RowIndex = 0; RowIndex < (unsigned) HeadingCount; ++RowIndex)
		{
		bool EndOfFile = ReadLineStdioFile(f, (char *) Line, sizeof(Line));
		if (EndOfFile)
			Die("Premature EOF in matrix file");

#if	TRACE
		Log("Line=%s\n", Line);
#endif
		if (Line[0] == '#')
			continue;

		char c = Line[0];
#if	TRACE
		Log("Row char=%c\n", c);
#endif

		unsigned RowLetter = c;
#if	TRACE
		Log("Row letter = %u\n", RowLetter);
#endif

		unsigned char *p = Line + 1;
		unsigned char *maxp = p + strlen((char *) Line);
		for (unsigned Col = 0; Col < (unsigned) HeadingCount; ++Col)
			{
			if (p >= maxp)
				Die("Too few fields in line of matrix file: '%s'", Line);
			while (isspace(*p))
				++p;
			unsigned char *Value = p;
			while (!isspace(*p))
				++p;
#if	INTEGER_DP
			score_t v = atoi((char *) Value);
#else
			score_t v = (score_t) atof((char *) Value);
#endif
			unsigned char HeaderChar = Heading[Col];
			unsigned ColLetter = HeaderChar;
			DNAScoreMx[RowLetter][ColLetter] = v;
			p += 1;
			}
		}

//#if VERBOSE
	LogMx();
//#endif
	}
