#include "utils.h"
#include "six.h"
#include "progress.h"

#define TRACE	0

void TruncLabel(std::string &Label);
coord_t GetOverlapLength(coord_t loA, coord_t hiA, coord_t loB, coord_t hiB);

//#ifndef NDEBUG
//static void ValidateOverlaps(const RevFile &Rev, unsigned AIIndex,
//  unsigned GenomeIndex, unsigned ChrId, coord_t lo, coord_t hi)
//	{
//	AlnInfo AI;
//	Rev.GetAlnInfo(AIIndex, AI);
//	const std::vector<GCSeg> &Segs = AI.Segs;
//	const unsigned SegCount = (unsigned) Segs.size();
//	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
//		{
//		const GCSeg &Seg = Segs[SegIndex];
//		if (Seg.GenomeIndex != GenomeIndex || Seg.ChrId != ChrId)
//			continue;
//		if (GetOverlapLength(Seg.lo, Seg.hi, lo, hi) > 0)
//			return;
//		}
//	Die("ValidateOverlaps failed");
//	}
//#endif

static void LogSixNode(const GCSegX &SX)
	{
	Log("%6u  %5u  %10u  %10u  %c  %7u\n",
	  SX.Seg.GenomeIndex,
	  SX.Seg.ChrId,
	  (unsigned) SX.Seg.lo,
	  (unsigned) SX.Seg.hi,
	  SX.Seg.Plus ? '+' : '-',
	  SX.AIIndex);
	}

void LogSix(const Six &Index)
	{
	Log("Six %p File %s Gen %u Chr %u\n",
	  &Index,
	  Index.ptrRevFile->GetFileName().c_str(),
	  Index.GenIndex,
	  Index.ChrId);
	Log("Genome    Chr          lo          hi  +  AIIndex\n");
	Log("======  =====  ==========  ==========  =  =======\n");
	Index.Traverse(LogSixNode);
	}

void LogSixes(const RevFile &Rev, const SixVecVec &Indexes)
	{
	unsigned GenCount = Rev.GetGenomeCount();
	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		{
		const std::string &GenName = Rev.GetGenomeName(GenIndex);
		const GenInfo &GI = Rev.GetGenomeInfo(GenIndex);
		const unsigned ChrCount = Rev.GetChrCount(GenIndex);
		for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
			{
			const ChrInfo &CI = GI.GetChrInfoByIndex(ChrIndex);
			unsigned ChrId = CI.GetId();
			const std::string &ChrName = GI.GetChrName(ChrId);
			const Six &Index = Indexes[GenIndex][ChrIndex];
			Log("\n");
			Log("Six %u,%u %.16s, %.16s\n",
			  GenIndex,
			  ChrIndex,
			  GenName.c_str(),
			  ChrName.c_str());
			LogSix(Index);
			}
		}
	Log("\n\n");
	}

// Indexes[i][j] is the Six for genome i, chr j.
void MakeSixes(const RevFile &Rev, SixVecVec &Indexes)
	{
#if	TRACE
	Log("\n");
	Log("MakeSixes\n");
#endif
	Indexes.clear();

	Progress("Reading AIs in %s", Rev.GetFileName().c_str());

	AIVec AIs;
	Rev.GetAllAlnInfos(AIs);

	unsigned GenCount = Rev.GetGenomeCount();
	Indexes.resize(GenCount);
	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		{
		Six Index;
		unsigned ChrCount = Rev.GetChrCount(GenIndex);
		Indexes[GenIndex].resize(ChrCount);
		FORIX(Chr)
			{
			Six &Index = Indexes[GenIndex][ChrIndex];
			unsigned ChrId = Rev.GetChrId(GenIndex, ChrIndex);
			Index.ptrRevFile = &Rev;
			Index.GenIndex = GenIndex;
			Index.ChrId = ChrId;
			}
		}

	const unsigned AICount = SIZE(AIs);
	ProgressStart("Building interval indexes");
#if	TRACE
	Log("   Index  Genome    Chr          lo          hi  +  AIIndex\n");
	Log("========  ======  =====  ==========  ==========  =  =======\n");
#endif
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AICount);

		const AlnInfo AI = AIs[AIIndex];
		const std::vector<GCSeg> &Segs = AI.Segs;
		const unsigned SegCount = (unsigned) Segs.size();
		GCSegX SegX;
		SegX.AIIndex = AIIndex;
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = Segs[SegIndex];
			SegX.Seg = Seg;
			unsigned GenIndex = Seg.GenomeIndex;
			const GenInfo &GI = Rev.GetGenomeInfo(GenIndex);
			unsigned ChrIndex = GI.GetChrIndex(Seg.ChrId);
			Six &Index = Indexes[GenIndex][ChrIndex];
#if	TRACE
			Log("%8p  %6u  %5u  %10u  %10u  %c  %7u\n",
			  &Index,
			  GenIndex,
			  ChrIndex,
			  (unsigned) Seg.lo,
			  (unsigned) Seg.hi,
			  Seg.Plus ? '+' : '-',
			  SegX.AIIndex);
#endif
			Index.Add(SegX);
			}
		}
	ProgressDone();
#if	TRACE
	LogSixes(Rev, Indexes);
#endif
	}

void DeleteFromSix(Six &Index, const GCSegXVec &Segs)
	{
	const unsigned SegCount = (unsigned) Segs.size();
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSegX &Seg = Segs[SegIndex];
		Index.Delete(Seg);
		}
	}

void LogGCSegXs(const GCSegXVec &Segs)
	{
	Log("Genome    Chr          lo          hi  +  AIIndex\n");
	Log("======  =====  ==========  ==========  =  =======\n");
	const unsigned SegCount = SIZE(Segs);
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSegX &SX = Segs[SegIndex];
		Log("%6u  %5u  %10u  %10u  %c  %7u\n",
		  SX.Seg.GenomeIndex,
		  SX.Seg.ChrId,
		  (unsigned) SX.Seg.lo,
		  (unsigned) SX.Seg.hi,
		  SX.Seg.Plus ? '+' : '-',
		  SX.AIIndex);
		}
	}

//void SixGetAIIndexes(const RevFile &Rev, const SixVecVec &Indexes,
//  unsigned GenIndex,unsigned ChrId, coord_t lo, coord_t hi, IntVec &AIIndexes)
//	{
//	const unsigned GenCount = Rev.GetGenomeCount();
//	asserta(GenIndex < GenCount);
//	const SixVec &SV = Indexes[GenIndex];
//	const unsigned ChrIndex = Rev.GetChrIndex(GenIndex, ChrId);
//	asserta(ChrIndex < SIZE(SV));
//	const Six &Index = SV[ChrIndex];
//
//	GCSegXVec Segs;
//	SixGetOverlaps(Index, Interval(lo, hi), Segs);
//	const unsigned Count = SIZE(Segs);
//	AIIndexes.resize(Count);
//	for (unsigned i = 0; i < Count; ++i)
//		AIIndexes[i] = Segs[i].AIIndex;
//	}

void SixGetOverlaps(const Six &Index, const Interval &Iv, GCSegXVec &Segs)
	{
	Index.GetOverlaps(Iv, Segs);

#ifndef NDEBUG
	{
// Validate index -- segments must all overlap interval.
	const unsigned SegCount = SIZE(Segs);
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSegX &SX = Segs[SegIndex];
		coord_t Ov = GetOverlapLength(SX.Seg.lo, SX.Seg.hi, Iv.first, Iv.second);
		if (Ov == 0)
			{
			Log("\n");
			Log("Index:\n");
			LogSix(Index);
			Log("\n");
			Log("Results:\n");
			LogGCSegXs(Segs);
			Log("\n");
			Log("Result %u,%u %u-%u %c does not overlap with query %u - %u\n",
			  SX.Seg.GenomeIndex,
			  SX.Seg.ChrId,
			  (unsigned) SX.Seg.lo,
			  (unsigned) SX.Seg.hi,
			  SX.Seg.Plus ? '+' : '-',
			  (unsigned) Iv.first,
			  (unsigned) Iv.second);
			Die("SixGetOverlaps validation failed");
			}
		}
	}
#endif
	}
