#include "utils.h"
#include "intlist.h"
#include "align_params.h"
#include <stack>

#define TRACE			0
#define TRACE_OPTS		0

/***
George's sliding window method for extracting high-scoring
regions from a long alignment while minimizing the
[>X][-X][>X} problem.

Slide a window across the alignment. Within each window,
repeatedly find optimal sub-alignments until none found
above the thresholds for acceptance. Output is then the
union of all such sub-alignments from all windows.
***/

score_t BestSubString(const score_t s[], unsigned ls, unsigned *ptrBestStart,
  unsigned *ptrBestEnd);
score_t GetAlnRuns(const char *Row1, const char *Row2, unsigned Length,
  IntVec &RunStarts, IntVec &RunEnds, std::vector<score_t> &RunScores);
void LogRuns(const char *Row1, const char *Row2, unsigned /* Length */,
  const IntVec &RunStarts, const IntVec &RunEnds,
  const std::vector<score_t> &RunScores);
bool AcceptAln(score_t Score, unsigned Length);

// Start or stop of high-scoring region
struct RunBounds
	{
	unsigned FirstRun;
	unsigned LastRun;
	};

#if	TRACE
static void LogStack(const std::stack<RunBounds> &RunBoundsStack)
	{
	std::stack<RunBounds> Copy = RunBoundsStack;
	Log("Stack size=%u\n", (unsigned) Copy.size());
	Log("Run From  Run To\n");
	Log("========  ======\n");
	//   12345678  123456

	while (Copy.size() > 0)
		{
		RunBounds RB = Copy.top();
		Log("%8u  %6u\n", RB.FirstRun, RB.LastRun);
		Copy.pop();
		}
	}
#endif	// TRACE

struct Bound
	{
	unsigned Pos;
	bool IsStart;
	};

static bool ByPos(const Bound &B1, const Bound &B2)
	{
// Want starts before ends to avoid coverage count < 0.
	if (B1.Pos == B2.Pos)
		return B1.IsStart && !B2.IsStart ? true : false;
	return B1.Pos < B2.Pos;
	}

static void DeleteShortBadRegions(const IntVec &InColStarts, 
  const IntVec &InColEnds, unsigned MinBadLength, IntVec &OutColStarts,
  IntVec &OutColEnds)
	{
	assert(InColStarts.size() == InColEnds.size());

	OutColStarts.clear();
	OutColEnds.clear();

	unsigned InCount = (unsigned) InColStarts.size();
	if (InCount == 0)
		return;

	unsigned LastEnd = UINT_MAX;
	unsigned LastStart = UINT_MAX;
	for (unsigned InIndex = 0; InIndex < InCount; ++InIndex)
		{
		unsigned Start = InColStarts[InIndex];
		unsigned End = InColEnds[InIndex];

		if (InIndex == 0)
			{
			LastEnd = InColEnds[0];
			LastStart = Start;
			continue;
			}
		assert(Start > LastEnd);

/***
 LastStart         LastEnd    Start       End
    |                  |        |          |
    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                       <------->
                       BadLength
***/
		unsigned BadLength = Start - LastEnd - 1;
		if (BadLength >= MinBadLength)
			{
			OutColStarts.push_back(LastStart);
			OutColEnds.push_back(LastEnd);
			LastStart = Start;
			}
		LastEnd = End;
		}

	assert(LastStart != UINT_MAX);
	assert(LastEnd != UINT_MAX);

	OutColStarts.push_back(LastStart);
	OutColEnds.push_back(LastEnd);

	assert(OutColStarts.size() == OutColEnds.size());
	assert(OutColStarts[0] == InColStarts[0]);
	assert(OutColEnds[OutColEnds.size() - 1] == InColEnds[InColEnds.size() - 1]);
	}

static void GetUnion(const IntVec &ColStarts, const IntVec &ColEnds,
  IntVec &UStarts, IntVec &UEnds)
	{
#if	TRACE
	Log("\n");
	Log("GetUnion\n");
#endif
	assert(ColStarts.size() == ColEnds.size());

	UStarts.clear();
	UEnds.clear();

	const unsigned RegionCount = (unsigned) ColStarts.size();
	const unsigned BoundCount = 2*RegionCount;

	std::vector<Bound> Bounds;
	Bounds.reserve(BoundCount);

	for (unsigned RegionIndex = 0; RegionIndex < RegionCount; ++RegionIndex)
		{
		Bound B;
		B.Pos = ColStarts[RegionIndex];
		B.IsStart = true;
		Bounds.push_back(B);

		B.Pos = ColEnds[RegionIndex];
		B.IsStart = false;
		Bounds.push_back(B);
		}

	std::sort(Bounds.begin(), Bounds.end(), ByPos);

	unsigned CoverageCount = 0;
	for (unsigned BoundIndex = 0; BoundIndex < BoundCount; ++BoundIndex)
		{
		const Bound &B = Bounds[BoundIndex];
#if	TRACE
		Log("Pos %u %s Cov %u\n", B.Pos, B.IsStart ? "Start" : "End", CoverageCount);
#endif
		if (B.IsStart)
			{
			if (CoverageCount == 0)
				{
#if	TRACE
				Log("STARTS %u\n", B.Pos);
#endif
				UStarts.push_back(B.Pos);
				}
			++CoverageCount;
			}
		else
			{
			assert(CoverageCount > 0);
			--CoverageCount;
			if (CoverageCount == 0)
				UEnds.push_back(B.Pos);
#if	TRACE
				Log("  ENDS %u\n", B.Pos);
#endif
			}
		}
	assert(CoverageCount == 0);
	assert(UStarts.size() == UEnds.size());
	}

static void GetGoodRegions(const char *Row1, const char *Row2,
  unsigned WindowStartCol, unsigned WindowEndCol,
  IntVec &GoodStarts, IntVec &GoodEnds)
	{
	assert(WindowEndCol >= WindowStartCol);
	unsigned WindowColCount = WindowEndCol - WindowStartCol + 1;
	if (WindowColCount == 0)
		return;

	const char *Row1W = Row1 + WindowStartCol;
	const char *Row2W = Row2 + WindowStartCol;

#if	TRACE || TRACE_OPTS
	{
	unsigned L = WindowColCount;
	if (L > 1024)
		L = 1024;
	Log("\n");
	Log("GetGoodRegions cols %u - %u\n", WindowStartCol, WindowEndCol);
	
	Log("%.*s\n", L, Row1W);
	Log("%.*s\n", L, Row2W);
	}
#endif

	IntVec RunStarts;
	IntVec RunEnds;
	std::vector<score_t> RunScores;
	GetAlnRuns(Row1W, Row2W, WindowColCount, RunStarts, RunEnds, RunScores);
#if	TRACE
	LogRuns(Row1W, Row2W, WindowColCount, RunStarts, RunEnds, RunScores);
#endif

	assert(RunStarts.size() == RunScores.size());
	assert(RunEnds.size() == RunScores.size());
	const unsigned RunCount = (unsigned) RunStarts.size();
	assert(RunCount > 0);

// Convert from STL to C vector for BestSubString.
	score_t *RunScores2 = myalloc(score_t, RunCount);
	for (unsigned i = 0; i < RunCount; ++i)
		RunScores2[i] = RunScores[i];

/***
Maintain a stack of segments to check for good regions.
ColStarts out as the whole alignment. When a good region
is found, it typically creates two regions to be explored
again (to the left and right).
***/
	std::stack<RunBounds> RunBoundsStack;

	RunBounds All;
	All.FirstRun = 0;
	All.LastRun = RunCount - 1;

	RunBoundsStack.push(All);

	while (!RunBoundsStack.empty())
		{
#if	TRACE
		LogStack(RunBoundsStack);
#endif
		RunBounds Region = RunBoundsStack.top();
		RunBoundsStack.pop();

		unsigned FirstRun = Region.FirstRun;
		unsigned LastRun = Region.LastRun;
		assert(LastRun < RunCount);
		assert(LastRun >= FirstRun);
		unsigned ThisRunCount = LastRun - FirstRun + 1;

		unsigned BSS_BestRunStart;
		unsigned BSS_BestRunEnd;
#if	TRACE
		{
		Log("Before BestSubString FirstRun=%u ThisRunCount=%u\n", FirstRun, ThisRunCount);
		for (unsigned i = 0; i < ThisRunCount; ++i)
			Log(" [%u]=%d", i, (RunScores2 + FirstRun)[i]);
		}
#endif
		score_t BestScore = BestSubString(RunScores2 + FirstRun, ThisRunCount,
		  &BSS_BestRunStart, &BSS_BestRunEnd);
#if	TRACE
		Log("FirstRun=%u LastRun=%u\n", FirstRun, LastRun);
		Log("BestScore = %d\n", BestScore);
#endif
		if (BestScore < 0)
			continue;

		assert(BSS_BestRunStart < ThisRunCount);
		assert(BSS_BestRunEnd < ThisRunCount);

	// BSS coordinates for BestSubString, which are offset
	// relative to full alignment.
		unsigned BestRunStart = FirstRun + BSS_BestRunStart;
		unsigned BestRunEnd = FirstRun + BSS_BestRunEnd;

		unsigned BestColStart = RunStarts[BestRunStart];
		unsigned BestColEnd = RunEnds[BestRunEnd];

		unsigned GlobalColStart = WindowStartCol + BestColStart;
		unsigned GlobalColEnd = WindowStartCol + BestColEnd;
		
		unsigned BestColCount = BestColEnd - BestColStart + 1;

#if	TRACE
		Log("BSS_BestRunStart=%u  BSS_BestRunEnd=%u ThisRunCount=%u\n",
		  BSS_BestRunStart, BSS_BestRunEnd, ThisRunCount);
		Log("BestRunStart=%u  BestRunEnd=%u\n", BestRunStart, BestRunEnd);
		Log("BestColStart=%u  BestColEnd=%u  BestColCount=%u\n",
		  BestColStart, BestColEnd, BestColCount);
		Log("GlobalColStart=%u  GlobalColEnd=%u\n",
		  GlobalColStart, GlobalColEnd);
#endif
		assert(BestRunStart < RunCount);
		assert(BestRunEnd < RunCount);
		assert(BestRunEnd >= BestRunStart);
		assert(RunScores[BestRunStart] >= 0);
		assert(RunScores[BestRunEnd] >= 0);
		assert(BestColStart < WindowColCount);
		assert(BestColEnd < WindowColCount);
		assert(BestColEnd >= BestColStart);
		assert(Row1[GlobalColStart] != '-');
		assert(Row2[GlobalColStart] != '-');
		assert(Row1[GlobalColEnd] != '-');
		assert(Row2[GlobalColEnd] != '-');
		assert(BestColCount <= WindowColCount);

		if (!AcceptAln(BestScore, BestColCount))
			continue;

#if	TRACE_OPTS
	{
	Log("\n");
	Log("Best region:\n");
	Log("%.*s\n", WindowColCount, Row1W);
	Log("%.*s\n", WindowColCount, Row2W);
	for (unsigned i = 0; i < BestColStart; ++i)
		Log(" ");
	for (unsigned i = BestColStart; i <= BestColEnd; ++i)
		Log("*");
	Log("\n");
	}
#endif

		GoodStarts.push_back(GlobalColStart);
		GoodEnds.push_back(GlobalColEnd);
#if	TRACE
		Log("Add good region GlobalColStart=%u GlobalColEnd=%u\n",
		  GlobalColStart, GlobalColEnd);
#endif

	// Preceding sub-alignment to be revisited
		if (BestRunStart > FirstRun)
			{
			RunBounds Left;
			Left.FirstRun = FirstRun;
			Left.LastRun = BestRunStart - 1;
#if	TRACE
			Log("Push left runs %u - %u\n", Left.FirstRun, Left.LastRun);
#endif
			RunBoundsStack.push(Left);
			}

	// Following sub-alignment to be revisited
		if (BestRunEnd < RunCount - 1)
			{
			RunBounds Right;
			Right.FirstRun = BestRunEnd + 1;
			Right.LastRun = RunCount - 1;
#if	TRACE
			Log("Push right runs %u - %u\n", Right.FirstRun, Right.LastRun);
#endif
			RunBoundsStack.push(Right);
			}
		}

	myfree(RunScores2);
	}

void Slide(const char *Row1, const char *Row2, unsigned AlnLength,
  unsigned WindowLength, unsigned WindowOffset, unsigned MinBadLength,
  IntVec &ColStarts, IntVec &ColEnds)
	{
#if	TRACE
	{
	Log("\n");
	Log("\n");

	unsigned L = AlnLength;
	if (L > 1024)
		L = 1024;
	Log("Slide AlnLength=%u WindowLength=%u WindowOffset=%u\n",
	  AlnLength, WindowLength, WindowOffset);
	Log("%.*s\n", L, Row1);
	Log("%.*s\n", L, Row2);
	}
#endif

	ColStarts.clear();
	ColEnds.clear();
	if (AlnLength == 0)
		return;

	if (WindowLength == 0 || WindowOffset == 0)
		Die("Smooth: must have window length > 0 and offset > 0");

	IntVec GoodStarts;
	IntVec GoodEnds;
	unsigned WindowStart = 0;
	bool Done = false;
	for (;;)
		{
		if (WindowStart >= AlnLength || Done)
			break;
		unsigned WindowEnd = WindowStart + WindowLength - 1;
		if (WindowEnd >= AlnLength)
			{
		/***
		Special case: make sure we do one window that ends
		on the last column to avoid artifacts at the end.
		***/
			WindowEnd = AlnLength - 1;
			if (WindowEnd >= WindowLength)
				WindowStart = (WindowEnd + 1) - WindowLength;
			else
				WindowStart = 0;
			Done = true;
			}
		unsigned Length = WindowEnd - WindowStart + 1;
		assert(Length <= WindowLength);
#if	TRACE
		Log("\n");
		Log("WINDOW cols %u - %u, length %u\n", WindowStart, WindowEnd, Length);
		Log("\n");
#endif
		GetGoodRegions(Row1, Row2, WindowStart, WindowEnd, GoodStarts, GoodEnds);

		WindowStart += WindowOffset;
		}

	IntVec GoodColStarts;
	IntVec GoodColEnds;
	GetUnion(GoodStarts, GoodEnds, GoodColStarts, GoodColEnds);
#if TRACE
	{
	Log("\n");
	Log("Good regions (after union, before mergebad):\n");
	Log("Index  Start    End  s  e\n");
	Log("=====  =====  =====  =  =\n");
	for (unsigned i = 0; i < (unsigned) GoodColStarts.size(); ++i)
		{
		unsigned Start = GoodColStarts[i];
		unsigned End = GoodColEnds[i];
		Log("%5u  %5u  %5u  %c  %c\n", i, Start, End, Row1[Start], Row2[End]);
		}
	}
#endif

// Verify good regions begin and end with a match
#ifndef NDEBUG
	{
	const unsigned Count = (unsigned) GoodColStarts.size();
	for (unsigned Index = 0; Index < Count; ++Index)
		{
		unsigned Start = GoodColStarts[Index];
		unsigned End = GoodColEnds[Index];

		char s1 = Row1[Start];
		char s2 = Row2[Start];
		assert(s1 != '-');
		assert(s2 != '-');

		char e1 = Row1[End];
		char e2 = Row2[End];
		assert(e1 != '-');
		assert(e2 != '-');
		}
	}
#endif

	DeleteShortBadRegions(GoodColStarts, GoodColEnds, MinBadLength,
	  ColStarts, ColEnds);

#if TRACE
	{
	Log("\n");
	Log("Good regions (after deleteshortbadregions):\n");
	Log("Index  Start    End  s  e\n");
	Log("=====  =====  =====  =  =\n");
	for (unsigned i = 0; i < (unsigned) ColStarts.size(); ++i)
		{
		unsigned Start = ColStarts[i];
		unsigned End = ColEnds[i];
		Log("%5u  %5u  %5u  %c  %c\n", i, Start, End, Row1[Start], Row2[End]);
		}
	}
#endif

// Re-verify good regions begin and end with a match
#ifndef NDEBUG
	{
	const unsigned Count = (unsigned) ColStarts.size();
	for (unsigned Index = 0; Index < Count; ++Index)
		{
		unsigned Start = ColStarts[Index];
		unsigned End = ColEnds[Index];

		char s1 = Row1[Start];
		char s2 = Row2[Start];
		if (s1 == '-')
			Die("ColStart[%u] = %u, Row1[%u] = '-'",
			  Index, Start, Start);
		if (s2 == '-')
			Die("ColStart[%u] = %u, Row2[%u] = '-'",
			  Index, Start, Start);

		char e1 = Row1[End];
		char e2 = Row2[End];
		if (e1 == '-')
			Die("ColEnd[%u] = %u, Row1[%u] = '-'",
			  Index, End, End);
		if (e2 == '-')
			Die("ColStart[%u] = %u, Row2[%u] = '-'",
			  Index, End, End);
		}
	}
#endif

// Validate in range, sorted & disjoint
#ifndef NDEBUG
	const unsigned Count = (unsigned) ColStarts.size();
	for (unsigned Index = 0; Index < Count; ++Index)
		{
		if (ColStarts[Index] >= AlnLength)
			Die("Slide: ColStarts[%u] = %u, AlnLength = %u",
			  Index, ColStarts[Index], AlnLength);
		if (ColEnds[Index] >= AlnLength)
			Die("Slide: ColEnds[%u] = %u, AlnLength = %u",
			  Index, ColEnds[Index], AlnLength);
		if (Index > 0)
			{
			if (ColStarts[Index] <= ColEnds[Index-1])
				Die("Slide: ColStarts[%u] = %u <= ColEnds[%u-1] = %u",
				  Index, ColStarts[Index], Index, ColEnds[Index-1]);
			}
		if (ColStarts[Index] > ColEnds[Index])
			Die("Slide: ColStarts[%u] = %u > ColEnds[%u] = %u",
			  Index, ColStarts[Index], Index, ColEnds[Index]);
		}
#endif
	}
