#define TRACE	0

#include "utils.h"
#include "evomodel.h"
#include "evolver.h"
#include "myrand.h"
#include "progress.h"
#include "ratemgr.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);
uint8 GetRandProbIndex(double MeanProb);

unsigned Round3Up(unsigned i)
	{
	unsigned r = i + (3 - i%3)%3;
	asserta(r%3 == 0);
	return r;
	}

unsigned Round3Down(unsigned i)
	{
	unsigned r = i - i%3;
	asserta(r%3 == 0);
	return r;
	}

EVENT_RESULT Evolver::Try_CreateCDS()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	vector<gpos_t> CDSLengths;
	const unsigned GeneCount = SIZE(Genes);
	for (unsigned i = 0; i < GeneCount; ++i)
		{
		const GeneInfo &Gene = Genes[i];
		const unsigned CDSCount = SIZE(Gene.CDSs);
		for (unsigned j = 0; j < CDSCount; ++j)
			{
			const CDSInfo &CDS = Gene.CDSs[j];
			CDSLengths.push_back(CDS.Length);
			}
		}

	const unsigned N = SIZE(CDSLengths);
	if (N == 0)
		EV_FAILN(CreateCDS, NoCDSs)

	unsigned r = RandInt(N);
	gpos_t CDSLength = CDSLengths[r];
	CDSLength += (3 - CDSLength%3)%3;
	asserta(CDSLength%3 == 0);

	GeneInfo Gene;
	gpos_t NewDonorPos = GPOS_T_MAX;
	gpos_t NewAcceptorPos = GPOS_T_MAX;
	unsigned Try = 1;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(CreateCDS, NoCandidates)

		bool Found = GetCandCDS(Gene, CDSLength, NewDonorPos, NewAcceptorPos);
		if (Found)
			break;
		++Try;
		}

	bool Strand = Gene.Plus;

// Find end of prev exon to determine codex
	gpos_t PrevCDSPos = GetAdjExonPos(NewAcceptorPos, !Strand);
	asserta(PrevCDSPos != GPOS_T_MAX);
	const BaseInfo &PrevCDSBase = GetBase(PrevCDSPos);
	int8 Codex = PrevCDSBase.Codex;

// Fully constrain new acceptor
	BaseInfo &FirstAcceptorBase = GetBase(NewAcceptorPos);
	asserta(FirstAcceptorBase.IsIntron());
	FirstAcceptorBase.SetSplice();

	gpos_t SecondAcceptorPos = GetAdjPos(NewAcceptorPos, Strand);
	BaseInfo &SecondAcceptorBase = GetBase(SecondAcceptorPos);
	asserta(SecondAcceptorBase.IsIntron());
	SecondAcceptorBase.SetSplice();

// First pos in new CDS
	gpos_t Pos = GetAdjPos(SecondAcceptorPos, Strand);
	m_CurrentEventPos = Pos;
	unsigned BaseCount = 0;
#if	TRACE
	Gene.LogMe();
	Log("NAP %"PF" NDP %"PF"\n", NewAcceptorPos, NewDonorPos);
#endif
	for (;;)
		{
		if (Pos == NewDonorPos)
			break;
		++BaseCount;
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert base %u before: ", BaseCount);
		Base.LogMe();
#endif
		asserta(Base.IsIntron());
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
		Codex = IncCodexStrand(Codex, Strand);
		Base.Codex = Codex;
		Base.SetCDS();
#if TRACE
		Log(" .. after ");
		Base.LogMe(true);
#endif
		Pos = GetAdjPos(Pos, Strand);
		}
	asserta(BaseCount > 0);
	asserta(BaseCount%3 == 0);

// Fully constrain new Donor
	BaseInfo &FirstDonorBase = GetBase(NewDonorPos);
	asserta(FirstDonorBase.IsIntron());
	FirstDonorBase.SetSplice();

	gpos_t SecondDonorPos = GetAdjPos(NewDonorPos, Strand);
	BaseInfo &SecondDonorBase = GetBase(SecondDonorPos);
	asserta(SecondDonorBase.IsIntron());
	SecondDonorBase.SetSplice();

	GeneInfo Gene2;
	GetGene(NewAcceptorPos, Gene2);
	unsigned ErrCount = ValidateGene(Gene2);
	if (ErrCount > 0)
		Die("Try_CreateCDS: ValidateGene failed");

	EV_ACCEPTN(CreateCDS, NewAcceptorPos)
	}

bool Evolver::GetCandCDS(const GeneInfo &Gene, gpos_t CDSLength,
  gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const
	{
	const unsigned IntronCount = SIZE(Gene.Introns);
	const unsigned ExonCount = SIZE(Gene.Exons);
	asserta(ExonCount == IntronCount + 1);

// Try introns in random order, otherwise bias towards
// 5' end of gene.
	vector<unsigned> IntronIndexes;
	for (unsigned i = 0; i < IntronCount; ++i)
		IntronIndexes.push_back(i);
	RandomizeVec(IntronIndexes);

	for (unsigned j = 0; j < IntronCount; ++j)
		if (GetCandCDSIntron(Gene, CDSLength, IntronIndexes[j],
		  NewDonorPos, NewAcceptorPos))
			return true;
	return false;
	}

/***
New CDS exons are always created between coding exons.
In other words, we never create a new first or last CDS.
However, new first or last CDSs can arise by START and STOP
codon moves into preceding / following all-UTR exons.

|<------CDS----->|<-------Intron--------->|<----CDS---->|
                      GT|<--Frame-->|AG
                       |             |
                     New             New
                Acceptor             Donor

The requirements are:

	New introns must be >= minimum length.
	Splice sites must match current bases.
	New CDS length must be multiple of 3 and
	within slop bases of target length.
	Frame must not contain a STOP codon.
	"Split" codons at start and end of new frame
	must not be STOP.
***/
bool Evolver::GetCandCDSIntron(const GeneInfo &Gene, gpos_t TargetCDSLength,
  unsigned IntronIndex, gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const
	{
	bool Strand = Gene.Plus;

	asserta(IntronIndex + 1 < SIZE(Gene.Exons));
	asserta(IntronIndex < SIZE(Gene.Introns));
	const IntronInfo &Intron = Gene.Introns[IntronIndex];
	if (Intron.Length < 3*P_MinIntronLength)
		{
#if	TRACE
		Log("FAIL: Short intron\n");
#endif
		return false;
		}

	string FractCodonStart;
	string FractCodonStart2;
	string FractCodonEnd;
	string FractCodonEnd2;

// Preceding exon might be UTR, in which case punt
	bool StartStop = false;
	if (!GetFractCodonBeforeIntron(Intron.Pos, FractCodonStart, Strand, StartStop) || StartStop)
		return false;

// Following exon might be UTR, in which case punt
	if (!GetFractCodonAfterIntron(Intron.Pos, FractCodonEnd, Strand, StartStop) || StartStop)
		return false;

	unsigned SumLengths = SIZE(FractCodonStart) + SIZE(FractCodonEnd);
	asserta(SumLengths == 0 || SumLengths == 3);

	string IntronSeq;
	GetSubSeq(Intron.Pos, Intron.Length, IntronSeq, Strand);

// Donor is first two bases
	string DonorPair;
	DonorPair = IntronSeq.substr(0, 2);

// Acceptor is last two bases
	string AcceptorPair;
	AcceptorPair = IntronSeq.substr(Intron.Length - 2, 2);

/***
          <-----------IntronLength---------------->
                         <------------D----------->
          <--Intron-->GT|<--Frame-->|AG<--Intron-->
                       | <--------->|
                     Max     Min    New
                Acceptor     New    Donor
				  Offset     CDS
                             Length
***/

	unsigned IntronLength = (unsigned) Intron.Length;
	if (TargetCDSLength <= P_CreateExonLengthSlop)
		return false;

	unsigned MinNewCDSLength = (unsigned) (TargetCDSLength - P_CreateExonLengthSlop);
	unsigned MaxNewCDSLength = (unsigned) (TargetCDSLength + P_CreateExonLengthSlop);

	MinNewCDSLength = Round3Up(MinNewCDSLength);
	MaxNewCDSLength = Round3Down(MaxNewCDSLength);

	unsigned MinAcceptorOffset = P_MinIntronLength + 2;
	unsigned D = MinNewCDSLength + 2 + P_MinIntronLength;
	if (IntronLength <= D)
		return false;

	unsigned MaxAcceptorOffset = IntronLength - D;
	unsigned MaxDonorOffset = IntronLength - P_MinIntronLength - 2;

// Consider candidate acceptors in random order
	vector<unsigned> AcceptorOffsets;
	for (unsigned AcceptorOffset = MinAcceptorOffset;
	  AcceptorOffset <= MaxAcceptorOffset; ++AcceptorOffset)
		{
		string Pair = IntronSeq.substr(AcceptorOffset, 2);
		if (Pair == AcceptorPair)
			AcceptorOffsets.push_back(AcceptorOffset);
		}

	RandomizeVec(AcceptorOffsets);
	if (AcceptorOffsets.size() > 256)
		AcceptorOffsets.resize(256);

	const unsigned N = SIZE(AcceptorOffsets);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned AcceptorOffset = AcceptorOffsets[i];
		unsigned MinDonorOffset = AcceptorOffset + 2 + MinNewCDSLength;
		unsigned HiDonorOffset = AcceptorOffset + 2 + MaxNewCDSLength;
#if	TRACE
		Log("Donor %s Acceptor %s AccOff %u MinDO %u MaxDO %u HiDO %u MinCDS %u Max CDS %u\n",
		  DonorPair.c_str(),
		  AcceptorPair.c_str(),
		  AcceptorOffset,
		  MinDonorOffset,
		  MaxDonorOffset,
		  HiDonorOffset,
		  MinNewCDSLength,
		  MaxNewCDSLength);
#endif
		if (HiDonorOffset > MaxDonorOffset)
			HiDonorOffset = MaxDonorOffset;
		for (unsigned DonorOffset = MinDonorOffset; DonorOffset <= MaxDonorOffset;
		  DonorOffset += 3)
			{
			string Pair = IntronSeq.substr(DonorOffset, 2);
			if (Pair == DonorPair)
				{
				asserta(DonorOffset > AcceptorOffset + 2);
				unsigned FrameLength = DonorOffset - AcceptorOffset - 2;
				string FrameSeq =
				  FractCodonStart + IntronSeq.substr(AcceptorOffset+2, FrameLength)
				  + FractCodonEnd;
				asserta(SIZE(FrameSeq)%3 == 0);
				if (!HasStopCodon(FrameSeq))
					{
					gpos_t StartPos =
					  Strand ? Intron.Pos : GetRangeEndPos(Intron.Pos, Intron.Length);
					NewDonorPos = StartPos;
					for (unsigned k = 0; k < DonorOffset; ++k)
						NewDonorPos = GetAdjPos(NewDonorPos, Strand);

					NewAcceptorPos = StartPos;
					for (unsigned k = 0; k < AcceptorOffset; ++k)
						NewAcceptorPos = GetAdjPos(NewAcceptorPos, Strand);
#if	TRACE
					Log("FOUND\n");
#endif
					return true;
					}
				}
			}
		}

	return false;
	}
