#define TRACE	0

#include "evolver.h"
#include "myrand.h"
#include "evomodel.h"

uint8 GetRandProbIndex(double MAC);

EVENT_RESULT Evolver::Try_CreateTermUTR()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	const unsigned GeneCount = SIZE(Genes);
	vector<gpos_t> UTRLengths;
	for (unsigned i = 0; i < GeneCount; ++i)
		{
		const GeneInfo &Gene = Genes[i];
		const unsigned UTRCount = SIZE(Gene.UTRs);
		for (unsigned j = 0; j < UTRCount; ++j)
			{
			const UTRInfo &UTR = Gene.UTRs[j];
			UTRLengths.push_back(UTR.Length);
			}
		}

	const unsigned N = SIZE(UTRLengths);
	unsigned r = RandInt(N);
	gpos_t UTRLength = UTRLengths[r];

	GeneInfo Gene;
	gpos_t UTRStartPos = GPOS_T_MAX;
	gpos_t NewDonorPos = GPOS_T_MAX;
	gpos_t NewAcceptorPos = GPOS_T_MAX;
	bool FoundFirst = false;
	bool FoundLast = false;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			break;

		bool TryFirstBeforeLast = RandTrue(0.5);
		if (TryFirstBeforeLast)
			{
			FoundFirst = GetCandFirstUTR(Gene, UTRLength, UTRStartPos,
			  NewDonorPos, NewAcceptorPos);
			if (FoundFirst)
				break;
			FoundLast = GetCandLastUTR(Gene, UTRLength, UTRStartPos,
			  NewDonorPos, NewAcceptorPos);
			if (FoundLast)
				break;
			}
		else
			{
			FoundFirst = GetCandLastUTR(Gene, UTRLength, UTRStartPos,
			  NewDonorPos, NewAcceptorPos);
			if (FoundFirst)
				break;
			FoundLast = GetCandFirstUTR(Gene, UTRLength, UTRStartPos,
			  NewDonorPos, NewAcceptorPos);
			if (FoundLast)
				break;
			}
		}
	if (!FoundFirst && !FoundLast)
		EV_FAILN(CreateTermUTR, NoCandidates)

	bool Strand = Gene.Plus;

	gpos_t Pos = UTRStartPos;
	m_CurrentEventPos = Pos;
#if	TRACE
	Gene.LogMe();
	Log("UTRStartPos %"PF" NDP %"PF" NAP %"PF"\n",
	  UTRStartPos, NewDonorPos, NewAcceptorPos);
#endif
	for (gpos_t i = 0; i < UTRLength; ++i)
		{
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert UTR base before: ");
		Base.LogMe();
#endif
		asserta(Base.IsUntranscribedGene());
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
		Base.SetUTR();
#if TRACE
		Log(" .. after ");
		Base.LogMe(true);
#endif
		Pos = GetAdjPos(Pos, Strand);
		}

// Fully constrain new Donor
	BaseInfo &FirstDonorBase = GetBase(NewDonorPos);
	FirstDonorBase.SetIntron();
	FirstDonorBase.SetSplice();

	gpos_t SecondDonorPos = GetAdjPos(NewDonorPos, Strand);
	BaseInfo &SecondDonorBase = GetBase(SecondDonorPos);
	SecondDonorBase.SetIntron();
	SecondDonorBase.SetSplice();

// Fully constrain new Acceptor
	BaseInfo &FirstAcceptorBase = GetBase(NewAcceptorPos);
	FirstAcceptorBase.SetIntron();
	FirstAcceptorBase.SetSplice();

	gpos_t SecondAcceptorPos = GetAdjPos(NewAcceptorPos, Strand);
	BaseInfo &SecondAcceptorBase = GetBase(SecondAcceptorPos);
	SecondAcceptorBase.SetIntron();
	SecondAcceptorBase.SetSplice();

	Pos = GetAdjPos(SecondDonorPos, Strand);
	unsigned BaseCount = 0;
	for (;;)
		{
		if (Pos == NewAcceptorPos)
			break;
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert intron base before: ");
		Base.LogMe();
#endif
		asserta(!Base.IsExon() && !Base.IsIntron());
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
		Base.SetIntron();
		++BaseCount;
#if TRACE
		Log(" .. after ");
		Base.LogMe(true);
#endif
		Pos = GetAdjPos(Pos, Strand);
		}
	asserta(BaseCount >= P_MinIntronLength);

	GeneInfo Gene2;
	GetGene(NewAcceptorPos, Gene2);
	unsigned ErrCount = ValidateGene(Gene2);
	if (ErrCount > 0)
		Die("Try_CreateTermUTR: ValidateGene failed");

	EV_ACCEPTN(CreateTermUTR, NewAcceptorPos)
	}

/***
    Gene start              Donor         Acceptor
       |                      |             |
       ....... [New first UTR]AG............GT[Old first exon] ... 
       <==>                     <New intron>
	   margin
***/
bool Evolver::GetCandFirstUTR(const GeneInfo &Gene, gpos_t UTRLength,
  gpos_t &UTRStartPos, gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetCandFirstUTR\n");
	Gene.LogMe();
	LogGene(Gene);
#endif

	const unsigned ExonCount = SIZE(Gene.Exons);
	if (ExonCount == 0)
		return false;

	bool Strand = Gene.Plus;
	const ExonInfo &FirstExon = Gene.GetExonX(0);

	gpos_t GeneStartPos =
	  Strand ? Gene.Pos : GetRangeEndPos(Gene.Pos, Gene.Length);

	gpos_t FirstExonStartPos =
	  Strand ? FirstExon.Pos : GetRangeEndPos(FirstExon.Pos, FirstExon.Length);

	if (GeneStartPos == FirstExonStartPos)
		return false;

	string Seq;
	for (gpos_t Pos = GeneStartPos; Pos != FirstExonStartPos;
	  Pos = GetAdjPos(Pos, Strand))
		{
		const BaseInfo &Base = GetBase(Pos);
		asserta(Base.IsGene());
		asserta(!Base.IsExon() && !Base.IsIntron());
		char Letter = Base.GetStrandLetter(Strand);
		Seq.push_back(Letter);
		}

	unsigned SeqLength = SIZE(Seq);
/***

  <--------------------------SeqLength--------------------->
  <--------------DonorOffset----------->
  .........................<-UTRLength->AG<--New intron-->GT<--OldFirstExon-->
  |                        |            |                 |
  GeneStartPos             UTRStartPos  DonorPos          AcceptorPos

***/
	unsigned MinDonorOffset = UTRLength;
	unsigned d = P_MinIntronLength + 4;
	if (SeqLength <= d)
		return false;
	unsigned MaxDonorOffset = SeqLength - d;

	vector<unsigned> DonorOffsets;
	for (unsigned Offset = MinDonorOffset; Offset <= MaxDonorOffset; ++Offset)
		if (Seq.substr(Offset, 2) == "AG")
			DonorOffsets.push_back(Offset);

	unsigned N = SIZE(DonorOffsets);
	if (N == 0)
		return false;

	unsigned r = RandInt(N);
	unsigned DonorOffset = DonorOffsets[r];
	asserta(DonorOffset >= UTRLength);
	unsigned UTROffset = DonorOffset - UTRLength;
	UTRStartPos = GeneStartPos;
	for (unsigned i = 0; i < UTROffset; ++i)
		UTRStartPos = GetAdjPos(UTRStartPos, Strand);

	NewDonorPos = GeneStartPos;
	for (unsigned i = 0; i < DonorOffset; ++i)
		NewDonorPos = GetAdjPos(NewDonorPos, Strand);

	NewAcceptorPos = FirstExonStartPos;
	NewAcceptorPos = GetAdjPos(NewAcceptorPos, !Strand);
	NewAcceptorPos = GetAdjPos(NewAcceptorPos, !Strand);
	return true;
	}

/***
                  Donor         Acceptor
                    |             |
 ... [Old last exon]AG............GT[New last UTR]
                      <New intron>
***/
bool Evolver::GetCandLastUTR(const GeneInfo &Gene, gpos_t UTRLength,
  gpos_t &UTRStartPos, gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetCandLastUTR\n");
	Gene.LogMe();
	LogGene(Gene);
#endif

	const unsigned ExonCount = SIZE(Gene.Exons);
	if (ExonCount == 0)
		return false;

	bool Strand = Gene.Plus;
	const ExonInfo &LastExon = Gene.GetExonX(ExonCount-1);
	
	gpos_t GeneEndPos =
	  Strand ? GetRangeEndPos(Gene.Pos, Gene.Length) : Gene.Pos;

	gpos_t LastExonEndPos =
	  Strand ? GetRangeEndPos(LastExon.Pos, LastExon.Length) : LastExon.Pos;

	if (GeneEndPos == LastExonEndPos)
		return false;

	string Seq;
	gpos_t Pos = GetAdjPos(LastExonEndPos, Strand);
	for (;;)
		{
		if (Pos == GeneEndPos)
			break;
		const BaseInfo &Base = GetBase(Pos);
		asserta(Base.IsGene());
		asserta(!Base.IsExon() && !Base.IsIntron());
		char Letter = Base.GetStrandLetter(Strand);
		Seq.push_back(Letter);
		Pos = GetAdjPos(Pos, Strand);
		}

	unsigned SeqLength = SIZE(Seq);
	if (SeqLength < P_MinIntronLength + 4)
		return false;

/***
          <---------------------------SeqLength----------------------->
          <--MaxAcceptorOffset--><----------------d------------------->
<LastExon>AG<-----New intron---->GT<-UTRLength->.......................
          |                      |                                    |
   DonorPos            AcceptorPos                           GeneEndPos
***/

	unsigned MinAcceptorOffset = P_MinIntronLength + 2;
	unsigned d = (unsigned) UTRLength + 2;
	if (SeqLength <= d)
		return false;
	unsigned MaxAcceptorOffset = SeqLength - d;

	vector<unsigned> AcceptorOffsets;
	for (unsigned Offset = MinAcceptorOffset; Offset <= MaxAcceptorOffset; ++Offset)
		if (Seq.substr(Offset, 2) == "GT")
			AcceptorOffsets.push_back(Offset);

	unsigned N = SIZE(AcceptorOffsets);
	if (N == 0)
		return false;

	unsigned r = RandInt(N);
	unsigned AcceptorOffset = AcceptorOffsets[r];

	NewDonorPos = GetAdjPos(LastExonEndPos, Strand);

	NewAcceptorPos = NewDonorPos;
	for (unsigned i = 0; i < AcceptorOffset; ++i)
		NewAcceptorPos = GetAdjPos(NewAcceptorPos, Strand);

	UTRStartPos = NewAcceptorPos;
	UTRStartPos = GetAdjPos(UTRStartPos, Strand);
	UTRStartPos = GetAdjPos(UTRStartPos, Strand);

	return true;
	}
