#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);

EVENT_RESULT Evolver::Try_DeleteUTR()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	unsigned ExonIndexToDelete = UINT_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(DeleteUTR, NoUTRs)

		ExonIndexToDelete = GetCandDeleteUTRExonIndex(Gene);
		if (ExonIndexToDelete != UINT_MAX)
			break;
		}
	if (ExonIndexToDelete == UINT_MAX)
		EV_FAILN(DeleteUTR, NoUTRs)

	m_CurrentEventPos = Gene.Pos;
	const ExonInfo &Exon = Gene.Exons[ExonIndexToDelete];
	asserta(Exon.AllUTR);
	unsigned UTRIndexToDelete = Exon.UTRIndex;
	asserta(UTRIndexToDelete < SIZE(Gene.UTRs));

	const unsigned UTRCount = SIZE(Gene.UTRs);
	const unsigned ExonCount = SIZE(Gene.Exons);
	asserta(UTRIndexToDelete < UTRCount);
	const UTRInfo &UTR = Gene.UTRs[UTRIndexToDelete];

	unsigned ExonIndex = UTR.ExonIndex;
	asserta(ExonIndex < ExonCount);

	bool FirstUTR = (ExonIndex == 0);
	bool LastUTR = (ExonIndex == ExonCount - 1);
	bool FirstOrLastUTR = (FirstUTR || LastUTR);

#if	TRACE
	Log("\n");
	Gene.LogMe();
	LogGene(Gene);
	Log("Exon to del %u, UTR to del %u, first UTR %c, last UTR %c\n",
	  ExonIndexToDelete, UTRIndexToDelete, tof(FirstUTR), tof(LastUTR));
	UTR.LogMe(true);
#endif

	asserta(UTR.Pos == Exon.Pos && UTR.Length == Exon.Length);

	gpos_t Pos = UTR.Pos;
	for (gpos_t i = 0; i < UTR.Length; ++i)
		{
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert UTR base pos %"PF, Pos);
		Base.LogMe();
#endif
		if (FirstOrLastUTR)
			Base.SetUntranscribedGene();
		else
			Base.SetIntron();
		Base.Deconstrain();
#if	TRACE
		Log(" => ");
		Base.LogMe(true);
#endif
		Pos = GetNextPos(Pos);
		}

// If terminal UTR, intron becomes untranscribed
	if (FirstOrLastUTR)
		{
		gpos_t Pos = GPOS_T_MAX;
		bool Plus = true;
		if (FirstUTR)
			{
		// First UTR in evolved order.
		// De-intronize following intron in evolved order.
			Plus = true;
			Pos = GetRangeEndPos(UTR.Pos, UTR.Length);
			}
		else
			{
			asserta(LastUTR);
		// Last UTR in evolved order.
		// De-intronize preceding intron in evolved order.
			Plus = false;
			Pos = UTR.Pos;
			}
#if	TRACE
		Log("First or last, intron -> untranscribed Pos=%"PF" Plus %c\n",
		  Pos, pom(Plus));
#endif
		for (;;)
			{
			Pos = GetAdjPos(Pos, Plus);
			BaseInfo &Base = GetBase(Pos);
#if	TRACE
			Log("Convert intron base ");
			Base.LogMe();
#endif
			if (!Base.IsIntron())
				{
#if	TRACE
				Log("-- not intron, done\n");
#endif
				break;
				}
			Base.SetUntranscribedGene();
#if	TRACE
			Log(" => ");
			Base.LogMe(true);
#endif
			}
		}
	else
		{
		asserta(!FirstOrLastUTR);

		{
	// Deconstrain preceding splice site
		gpos_t Pos = GetPrevPos(UTR.Pos);
		BaseInfo &Base1 = GetBase(Pos);
#if	TRACE
		Log("Deconstrain prec. splice ");
		Base1.LogMe();
#endif
		Base1.SetIntron();
		Base1.Deconstrain();
#if	TRACE
		Log(" => ");
		Base1.LogMe(true);
#endif
		Pos = GetPrevPos(Pos);
		BaseInfo &Base2 = GetBase(Pos);
#if	TRACE
		Log("Deconstrain prec. splice ");
		Base2.LogMe();
#endif
		Base2.SetIntron();
		Base2.Deconstrain();
#if	TRACE
		Log(" => ");
		Base2.LogMe(true);
#endif
		}

		{
	// Deconstrain following splice site
		gpos_t Pos = GetRangeEndPos(UTR.Pos, UTR.Length);
		Pos = GetNextPos(Pos);
		BaseInfo &Base1 = GetBase(Pos);
#if	TRACE
		Log("Deconstrain following splice ");
		Base1.LogMe();
#endif
		asserta(Base1.IsSplice());
		Base1.SetIntron();
		Base1.Deconstrain();
#if	TRACE
		Log(" => ");
		Base1.LogMe(true);
#endif
		Pos = GetNextPos(Pos);
		BaseInfo &Base2 = GetBase(Pos);
#if	TRACE
		Log("Deconstrain following splice ");
		Base2.LogMe();
#endif
		asserta(Base2.IsSplice());
		Base2.SetIntron();
		Base2.Deconstrain();
#if	TRACE
		Log(" => ");
		Base2.LogMe(true);
#endif
		}
		}

	GeneInfo NewGene;
	GetGene(UTR.Pos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("DeleteUTR: Validate gene failed");

	EV_ACCEPTN(DeleteUTR, UTR.Pos)
	}

unsigned Evolver::GetCandDeleteUTRExonIndex(const GeneInfo &Gene) const
	{
	gpos_t FirstUTRBaseCount = 0;
	gpos_t LastUTRBaseCount = 0;
	GetUTRBaseCounts(Gene, FirstUTRBaseCount, LastUTRBaseCount);

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);
	for (unsigned j = 0; j < SIZE(ExonIndexes); ++j)
		{
		unsigned ExonIndex = ExonIndexes[j];
		const ExonInfo &Exon = Gene.Exons[ExonIndex];
		if (!Exon.AllUTR)
			continue;
		gpos_t UTRBaseCount = GetUTRBaseCount(Gene, ExonIndex);
		if (UTRBaseCount - Exon.Length < P_MinUTRBases)
			continue;
		return ExonIndex;
		}
	return UINT_MAX;
	}

bool Evolver::IsFirstUTR(const GeneInfo &Gene, unsigned ExonIndex) const
	{
	bool FirstUTR = true;
	for (unsigned i = 0; i <= ExonIndex; ++i)
		{
		const ExonInfo &Exon = Gene.Exons[i];
		if (Exon.AllCoding)
			{
			FirstUTR = false;
			break;
			}
		}
	return FirstUTR;
	}
