#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);

EVENT_RESULT Evolver::Try_MoveAcceptorIntoCDS()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldAcceptorPos = GPOS_T_MAX;
	gpos_t NewAcceptorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveAcceptorIntoCDS, NoAcceptors)

		bool Found = GetMoveAcceptorIntoCDSCand(Gene, OldAcceptorPos, NewAcceptorPos);
		if (Found)
			break;
		}

	m_CurrentEventPos = OldAcceptorPos;
	bool Strand = Gene.Plus;
#if	TRACE
	Log("Select gene %u old pos %"PF" new pos %"PF"\n",
	  Gene.Index, OldAcceptorPos, NewAcceptorPos);
	Gene.LogMe();
#endif

	string OldPair;
	string NewPair;

	gpos_t SecondOldAcceptorPos = GetAdjPos(OldAcceptorPos, Strand);
	gpos_t SecondNewAcceptorPos = GetAdjPos(NewAcceptorPos, Strand);

	BaseInfo &OldBase1 = GetBase(OldAcceptorPos);
	BaseInfo &NewBase1 = GetBase(NewAcceptorPos);

	BaseInfo &OldBase2 = GetBase(SecondOldAcceptorPos);
	BaseInfo &NewBase2 = GetBase(SecondNewAcceptorPos);

	asserta(OldBase1.IsIntron());
	asserta(OldBase2.IsIntron());

	asserta(OldBase1.IsSplice());
	asserta(OldBase2.IsSplice());

	asserta(NewBase1.IsCDS());
	asserta(NewBase2.IsCDS());

	OldPair.push_back(OldBase1.GetStrandLetter(Strand));
	NewPair.push_back(NewBase1.GetStrandLetter(Strand));

	OldPair.push_back(OldBase2.GetStrandLetter(Strand));
	NewPair.push_back(NewBase2.GetStrandLetter(Strand));

	asserta(OldPair == NewPair);

	unsigned BaseCount = 0;
	gpos_t SecondAcceptorPos = GetAdjPos(OldAcceptorPos, Strand);
	gpos_t FirstCDSPos = GetAdjPos(SecondAcceptorPos, Strand);
	const BaseInfo &FirstCDSBase = GetBase(FirstCDSPos);
	asserta(FirstCDSBase.IsCDS());
#if	TRACE
	Log("\n");
	Log("OAP %"PF" OAP2 %"PF" NAP %"PF" NAP2 %"PF"\n",
	  OldAcceptorPos, SecondOldAcceptorPos, NewAcceptorPos, SecondNewAcceptorPos);
#endif

	gpos_t Pos = FirstCDSPos;
	for (;;)
		{
		if (Pos == NewAcceptorPos)
			break;
		BaseInfo &Base = GetBase(Pos);
		++BaseCount;
#if	TRACE
		Log("Convert base %u pos %"PF", before: ", Pos, BaseCount);
		Base.LogMe();
#endif
		asserta(Base.IsCDS());
		Base.SetIntron();
		Base.Deconstrain();
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		Pos = GetAdjPos(Pos, Strand);
		}
	asserta(BaseCount > 0);
	asserta((BaseCount + 2)%3 == 0);
	asserta(BaseCount <= P_MaxMoveSpliceBases);

	OldBase1.SetIntron();
	OldBase2.SetIntron();

	OldBase1.Deconstrain();
	OldBase2.Deconstrain();

	NewBase1.SetSplice();
	NewBase2.SetSplice();

	GeneInfo NewGene;
	GetGene(NewAcceptorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveAcceptorIntoCDS: Validate gene failed");

	EV_ACCEPTN(MoveAcceptorIntoCDS, NewAcceptorPos)
	}

// Acceptor is sequence at end of intron.
// Find sequence in CDS: <intron><3n bases>XX
// where XX is the last two bases in the intron and no new stop
// codon is created that spans the enlarged intron.
bool Evolver::GetMoveAcceptorIntoCDSCand(const GeneInfo &Gene,
  gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveAcceptorIntoCDSCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif
	bool Strand = Gene.Plus;
	const unsigned ExonCount = SIZE(Gene.Exons);
	if (ExonCount == 1)
		return false;

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);

	for (unsigned j = 0; j < ExonCount; ++j)
		{
		unsigned ExonIndex = ExonIndexes[j];
		const ExonInfo &Exon = Gene.Exons[ExonIndex];
		if (!Exon.AllCoding)
			{
#if	TRACE
			Log("FAIL: AllCoding\n");
#endif
			continue;
			}

		if (Strand && ExonIndex == 0)
			continue;
		if (!Strand && ExonIndex + 1 == ExonCount)
			continue;

		gpos_t FirstExonPos =
		  Strand ? Exon.Pos : GetRangeEndPos(Exon.Pos, Exon.Length);

		const BaseInfo &FirstExonBase = GetBase(FirstExonPos);
		asserta(FirstExonBase.IsExon() && FirstExonBase.IsCDS());

		gpos_t SecondOldAcceptorPos = GetAdjPos(FirstExonPos, !Strand);
		OldAcceptorPos = GetAdjPos(SecondOldAcceptorPos, !Strand);

		const BaseInfo &FirstOldAcceptorBase = GetBase(OldAcceptorPos);
		const BaseInfo &SecondOldAcceptorBase = GetBase(SecondOldAcceptorPos);
		asserta(FirstOldAcceptorBase.IsSplice());
		asserta(SecondOldAcceptorBase.IsSplice());

#if	TRACE
		Log("GetMoveAcceptorIntoCDSCand\n");
		Log("Exon: ");
		Exon.LogMe(true);
		Log("FEB  ");
		FirstExonBase.LogMe(true);
		Log("FOAB ");
		FirstOldAcceptorBase.LogMe(true);
		Log("SOAB ");
		SecondOldAcceptorBase.LogMe(true);
#endif

		asserta(FirstOldAcceptorBase.IsIntron());
		asserta(SecondOldAcceptorBase.IsIntron());

		asserta(FirstOldAcceptorBase.IsSplice());
		asserta(SecondOldAcceptorBase.IsSplice());

		string FractCodonPrefix;
		const ExonInfo &PrevExon =
		  (Strand ? Gene.Exons[ExonIndex-1] : Gene.Exons[ExonIndex+1]);
		bool StartStop = false;
		bool Ok = GetFractCodonExonEnd(PrevExon, FractCodonPrefix, Strand, StartStop);
		if (!Ok || StartStop)
			{
#if	TRACE
			Log("FAIL: GetFractCodonExonEnd OK=%c StartStop=%c\n", Ok, StartStop);
#endif
			continue;
			}

		string OldAcceptorPair;
		OldAcceptorPair.push_back(FirstOldAcceptorBase.GetStrandLetter(Strand));
		OldAcceptorPair.push_back(SecondOldAcceptorBase.GetStrandLetter(Strand));

#if TRACE
		{
		Log("\n");
		Log("Exon: ");
		Exon.LogMe(true);
		Log("OldAcceptorPos %"PF" %"PF"\n", OldAcceptorPos, SecondOldAcceptorPos);
		Log("FirstExonBase: ");
		FirstExonBase.LogMe(true);
		Log("FirstOldAcceptorBase: ");
		FirstOldAcceptorBase.LogMe(true);
		Log("SecondOldAcceptorBase: ");
		SecondOldAcceptorBase.LogMe(true);
		Log("OldAcceptorPair %s\n", OldAcceptorPair.c_str());
		}
#endif
		asserta(FirstOldAcceptorBase.IsIntron());
		asserta(SecondOldAcceptorBase.IsIntron());

		gpos_t MaxDist = P_MaxMoveSpliceBases;
		if (MaxDist + 8 >= Exon.Length)
			{
			if (Exon.Length <= 16)
				return false;
			MaxDist = Exon.Length - 8;
			}
		unsigned MaxCodons = (unsigned) (MaxDist/3);

		NewAcceptorPos = OldAcceptorPos;
		for (unsigned k = 0; k <= MaxCodons; ++k)
			{
			if (NewAcceptorPos == Gene.StopCodonFirstBasePos)
				return false;
			const BaseInfo &Base1 = GetBase(NewAcceptorPos);
			if (k > 0 && !Base1.IsCDS())
				{
#if	TRACE
				Log("FAIL: Base1 not CDS\n");
#endif
				break;
				}

			gpos_t Pos2 = GetAdjPos(NewAcceptorPos, Strand);
			if (Pos2 == Gene.StopCodonFirstBasePos)
				return false;
			const BaseInfo &Base2 = GetBase(Pos2);
			if (k > 0 && !Base2.IsCDS())
				{
#if	TRACE
				Log("FAIL: Base2 not CDS\n");
#endif
				break;
				}

			string Pair;
			Pair.push_back(Base1.GetStrandLetter(Strand));
			Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
			Log("\n");
			Log("Base1: ");
			Base1.LogMe(true);
			Log("Base2: ");
			Base2.LogMe(true);
			Log("Base3: ");
			Log("Pair %s\n", Pair.c_str());
#endif
			if (k > 0 && Pair == OldAcceptorPair)
				{
#if	TRACE
				Log("PAIR MATCH\n");
#endif
				if (!FractCodonPrefix.empty())
					{
					string NewCodon(FractCodonPrefix);
					gpos_t Pos = NewAcceptorPos;
					Pos = GetAdjPos(Pos, Strand);
					while (NewCodon.length() < 3)
						{
						Pos = GetAdjPos(Pos, Strand);
						const BaseInfo &Base = GetBase(Pos);
						if (!Base.IsCDS())
							goto NextExon;
						NewCodon.push_back(Base.GetStrandLetter(Strand));
						}
					asserta(NewCodon.length() == 3);
#if TRACE
					Log("NewCodon=%s NewAcceptorPos=%"PF" Strand=%c\n",
					  NewCodon.c_str(), NewAcceptorPos, pom(Strand));
#endif
					if (IsStopCodon(NewCodon))
						goto NextExon;
					}
#if TRACE
				Log("FOUND\n");
#endif
				return true;
				}
			gpos_t Pos3 = GetAdjPos(Pos2, Strand);
			if (Pos2 == Gene.StopCodonFirstBasePos)
				return false;
			const BaseInfo &Base3 = GetBase(Pos3);
			if (!Base3.IsCDS())
				{
#if	TRACE
				Log("FAIL: Base3 not CDS\n");
#endif
				goto NextExon;	
				}
			NewAcceptorPos = GetAdjPos(Pos3, Strand);
			}
	NextExon:
		;
		}

	return false;
	}
