#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);
uint8 GetRandProbIndex(double MAC);

EVENT_RESULT Evolver::Try_MoveCDSDonorIntoIntron()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldDonorPos = GPOS_T_MAX;
	gpos_t NewDonorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveCDSDonorIntoIntron, NoDonors)

		bool Found = GetMoveCDSDonorIntronCand(Gene, OldDonorPos, NewDonorPos);
		if (Found)
			break;
		}
	m_CurrentEventPos = OldDonorPos;

#if	TRACE
	Log("Select gene %u old pos %"PF" new pos %"PF"\n",
	  Gene.Index, OldDonorPos, NewDonorPos);
	Gene.LogMe();
#endif

	bool Strand = Gene.Plus;

	string OldPair;
	string NewPair;

	GetNextPair(OldDonorPos, OldPair, Strand);
	GetNextPair(NewDonorPos, NewPair, Strand);
	asserta(OldPair == NewPair);

	unsigned BaseCount = 0;
	gpos_t LastCDSPos = GetAdjPos(OldDonorPos, !Strand);
	const BaseInfo &LastCDSBase = GetBase(LastCDSPos);
	asserta(LastCDSBase.IsCDS());
	int8 Codex = LastCDSBase.Codex;
	gpos_t Pos = OldDonorPos;
	for (;;)
		{
		if (Pos == NewDonorPos)
			break;
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert base %u, before: ", BaseCount + 1);
		Base.LogMe();
#endif
		asserta(Base.IsIntron());
		Codex = IncCodexStrand(Codex, Strand);
		Base.Codex = Codex;
		Base.SetCDS();
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		++BaseCount;
		Pos = GetAdjPos(Pos, Strand);
		}
	asserta(BaseCount > 0);
	asserta(BaseCount%3 == 0);
	asserta(BaseCount <= P_MaxMoveSpliceBases);

// No mutations allowed in new donor sites
	BaseInfo &Base1 = GetBase(Pos);
	asserta(Base1.IsIntron());
	Base1.SetSplice();

	Pos = GetAdjPos(Pos, Strand);
	BaseInfo &Base2 = GetBase(Pos);
	asserta(Base2.IsIntron());
	Base2.SetSplice();

	GeneInfo NewGene;
	GetGene(NewDonorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveCDSDonorIntoIntron: Validate gene failed");

	EV_ACCEPTN(MoveCDSDonorIntoIntron, NewDonorPos)
	}

// Donor is sequence at start of intron.
// Find sequence: <intron start><3n bases with no in-frame STOP>XX
// where XX is the first two bases in the intron.
// Must ensure that no new STOP codon is created that spans
// the enlarged intron.
bool Evolver::GetMoveCDSDonorIntronCand(const GeneInfo &Gene,
  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveDonorIntronCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);

	for (unsigned j = 0; j < SIZE(ExonIndexes); ++j)
		if (GetMoveCDSDonorIntronCandExon(Gene, ExonIndexes[j],
			  OldDonorPos, NewDonorPos))
			return true;

	return false;
	}

bool Evolver::GetMoveCDSDonorIntronCandExon(const GeneInfo &Gene,
  unsigned ExonIndex, gpos_t &OldDonorPos, gpos_t &NewDonorPos) const
	{
	const ExonInfo &Exon = Gene.Exons[ExonIndex];
	if (!Exon.AllCoding)
		return false;

	bool Strand = Gene.Plus;
	if (Strand && ExonIndex + 1 == SIZE(Gene.Exons))
		return false;
	if (!Strand && ExonIndex == 0)
		return false;
	unsigned NextExonIndex = (Strand ? ExonIndex + 1 : ExonIndex - 1);
	const ExonInfo &NextExon = Gene.Exons[NextExonIndex];
	if (!NextExon.AllCoding)
		return false;

	unsigned IntronIndex = (Strand ? ExonIndex : ExonIndex - 1);
	const IntronInfo &Intron = Gene.Introns[IntronIndex];
	OldDonorPos =
	  Strand ? Intron.Pos : GetRangeEndPos(Intron.Pos, Intron.Length);
	const BaseInfo &OldFirstDonorBase = GetBase(OldDonorPos);
#if	TRACE
	Log("\n");
	Log("Intron: ");
	Intron.LogMe(true);
	Log("OldFirstDonorBase: ");
	OldFirstDonorBase.LogMe(true);
#endif
	asserta(OldFirstDonorBase.IsIntron());
	asserta(OldFirstDonorBase.IsSplice());

	string OldDonorPair;
	GetNextPair(OldDonorPos, OldDonorPair, Strand);

	string FractCodonPrefix;
	string FractCodonSuffix;
	bool StartStop;
	GetFractCodonBeforeIntron(OldDonorPos, FractCodonPrefix, Strand, StartStop);
	if (StartStop)
		return false;
	GetFractCodonAfterIntron(OldDonorPos, FractCodonSuffix, Strand, StartStop);
	if (StartStop)
		return false;

	unsigned SumLengths = SIZE(FractCodonPrefix) + SIZE(FractCodonSuffix);
	asserta(SumLengths == 0 || SumLengths == 3);

	NewDonorPos = OldDonorPos;
	string Codon(FractCodonPrefix);
	for (unsigned CodonsMoved = 0; CodonsMoved < P_MaxMoveSpliceBases/3;
	  ++CodonsMoved)
		{
		const BaseInfo &Base1 = GetBase(NewDonorPos);
		if (!Base1.IsIntron())
			return false;

		gpos_t Pos2 = GetAdjPos(NewDonorPos, Strand);
		const BaseInfo &Base2 = GetBase(Pos2);
		if (!Base2.IsIntron())
			return false;

		gpos_t Pos3 = GetAdjPos(Pos2, Strand);
		const BaseInfo &Base3 = GetBase(Pos3);
		if (!Base3.IsIntron())
			return false;

		string Pair;
		Pair.push_back(Base1.GetStrandLetter(Strand));
		Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
		Log("\n");
		Log("NDP %"PF" Pos2 %"PF" Pos3 %"PF"\n", NewDonorPos, Pos2, Pos3);
		Log("Base1: ");
		Base1.LogMe(true);
		Log("Base2: ");
		Base2.LogMe(true);
		Log("Base3: ");
		Base3.LogMe(true);
		Log("Pair %s OldDonor %s\n", Pair.c_str(), OldDonorTriplet.c_str());
#endif

		if (CodonsMoved > 0 && Pair == OldDonorPair)
			{
			gpos_t BasesMoved = 3*CodonsMoved;
			asserta(BasesMoved < Intron.Length);
			if (Intron.Length - BasesMoved < P_MinIntronLength)
				return false;

			if (!FractCodonSuffix.empty())
				{
				string Codon(FractCodonSuffix);
				gpos_t Pos = NewDonorPos;
				while (Codon.size() < 3)
					{
					Pos = GetAdjPos(Pos, !Strand);
					const BaseInfo &Base = GetBase(Pos);
					Codon = Base.GetStrandLetter(Strand) + Codon;
					}
				if (IsStopCodon(Codon))
					return false;
				}
			return true;
			}

		Codon.push_back(Base1.GetStrandLetter(Strand));
		if (Codon.size() == 3)
			{
			if (IsStopCodon(Codon))
				return false;
			Codon.clear();
			}
		Codon.push_back(Base2.GetStrandLetter(Strand));
		if (Codon.size() == 3)
			{
			if (IsStopCodon(Codon))
				return false;
			Codon.clear();
			}
		Codon.push_back(Base3.GetStrandLetter(Strand));
		if (Codon.size() == 3)
			{
			if (IsStopCodon(Codon))
				return false;
			Codon.clear();
			}

		NewDonorPos = GetAdjPos(Pos3, Strand);
		}
	return false;
	}
