#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);

EVENT_RESULT Evolver::Try_MoveDonorIntoCDS()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldDonorPos = GPOS_T_MAX;
	gpos_t NewDonorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveDonorIntoCDS, NoDonors)

		bool Found = GetMoveDonorIntoCDSCand(Gene, OldDonorPos, NewDonorPos);
		if (Found)
			break;
		}
	m_CurrentEventPos = OldDonorPos;
	bool Strand = Gene.Plus;

	string OldPair;
	string NewPair;

	gpos_t Old2 = GetAdjPos(OldDonorPos, Strand);
	gpos_t New2 = GetAdjPos(NewDonorPos, Strand);

	BaseInfo &OldBase1 = GetBase(OldDonorPos);
	BaseInfo &NewBase1 = GetBase(NewDonorPos);

	BaseInfo &OldBase2 = GetBase(Old2);
	BaseInfo &NewBase2 = GetBase(New2);

	asserta(OldBase1.IsIntron());
	asserta(OldBase2.IsIntron());

	asserta(NewBase1.IsCDS());
	asserta(NewBase2.IsCDS());

	OldPair.push_back(OldBase1.GetStrandLetter(Strand));
	NewPair.push_back(NewBase1.GetStrandLetter(Strand));

	OldPair.push_back(OldBase2.GetStrandLetter(Strand));
	NewPair.push_back(NewBase2.GetStrandLetter(Strand));

	asserta(OldPair == NewPair);

	unsigned BaseCount = 0;
	gpos_t LastCDSPos = GetAdjPos(OldDonorPos, !Strand);
	const BaseInfo &LastCDSBase = GetBase(LastCDSPos);
	asserta(LastCDSBase.IsCDS());
	gpos_t Pos = NewDonorPos;
	for (;;)
		{
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert base %u, before: ", BaseCount + 1);
		Base.LogMe();
#endif
		asserta(Base.IsCDS());
		Base.SetIntron();
		Base.Deconstrain();
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		++BaseCount;
		Pos = GetAdjPos(Pos, Strand);
		if (Pos == OldDonorPos)
			break;
		}
	asserta(BaseCount > 0);
	asserta(BaseCount%3 == 0);
	asserta(BaseCount <= P_MaxMoveSpliceBases);

	OldBase1.SetIntron();
	OldBase2.SetIntron();

// No mutations allowed in new Donor site
	NewBase1.SetSplice();
	NewBase2.SetSplice();

	GeneInfo NewGene;
	GetGene(NewDonorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveDonorIntoCDS: Validate gene failed");

	EV_ACCEPTN(MoveDonorIntoCDS, NewDonorPos)
	}

bool Evolver::GetMoveDonorIntoCDSCand(const GeneInfo &Gene, gpos_t &OldDonorPos,
  gpos_t &NewDonorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveDonorIntoCDSCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);

	for (unsigned j = 0; j < SIZE(ExonIndexes); ++j)
		if (GetMoveDonorIntoCDSCandExon(Gene, ExonIndexes[j],
		  OldDonorPos, NewDonorPos))
			return true;

	return false;
	}

// Donor is pair of bases at start of intron.
// Find sequence in CDS: XX<3n bases><intron>
// where XX is the first two bases in the intron and no new
// stop codon is created that spans the enlarged intron.
bool Evolver::GetMoveDonorIntoCDSCandExon(const GeneInfo &Gene, unsigned ExonIndex,
  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const
	{
	bool Strand = Gene.Plus;

	const unsigned ExonCount = SIZE(Gene.Exons);

	const ExonInfo &Exon = Gene.Exons[ExonIndex];
	if (!Exon.AllCoding)
		return false;

// All exons except the last have donor immediately following.
	if (Strand && ExonIndex + 1 == ExonCount)
		return false;
	if (!Strand && ExonIndex == 0)
		return false;

	const ExonInfo &NextExon =
	  (Strand ? Gene.Exons[ExonIndex+1] : Gene.Exons[ExonIndex-1]);
	if (NextExon.CDSIndex == UINT_MAX)
		return false;

	gpos_t LastExonPos =
	  Strand ? GetRangeEndPos(Exon.Pos, Exon.Length) : Exon.Pos;

	const BaseInfo &LastExonBase = GetBase(LastExonPos);
	asserta(LastExonBase.IsExon());

	OldDonorPos = GetAdjPos(LastExonPos, Strand);
	gpos_t SecondOldDonorPos = GetAdjPos(OldDonorPos, Strand);

	const BaseInfo &FirstOldDonorBase = GetBase(OldDonorPos);
	const BaseInfo &SecondOldDonorBase = GetBase(SecondOldDonorPos);

	asserta(FirstOldDonorBase.IsIntron());
	asserta(SecondOldDonorBase.IsIntron());

	asserta(FirstOldDonorBase.IsSplice());
	asserta(SecondOldDonorBase.IsSplice());

	string FractCodonSuffix;
	bool StartStop;
	bool Ok = GetFractCodonExonStart(NextExon, FractCodonSuffix, Strand, StartStop);
	if (!Ok)
		return false;
	if (StartStop)
		return false;

	string OldDonorPair;
	OldDonorPair.push_back(FirstOldDonorBase.GetStrandLetter(Strand));
	OldDonorPair.push_back(SecondOldDonorBase.GetStrandLetter(Strand));

#if TRACE
	{
	Log("\n");
	Log("Exon: ");
	Exon.LogMe(true);
	Log("OldDonorPos %"PF"\n", OldDonorPos);
	Log("LastExonBase: ");
	LastExonBase.LogMe(true);
	Log("FirstOldDonorBase: ");
	FirstOldDonorBase.LogMe(true);
	Log("SecondOldDonorBase: ");
	SecondOldDonorBase.LogMe(true);
	Log("OldDonorPair %s\n", OldDonorPair.c_str());
	}
#endif
	asserta(FirstOldDonorBase.IsIntron());
	asserta(SecondOldDonorBase.IsIntron());

	gpos_t MaxCodons = P_MaxMoveSpliceBases/3;
	if (MaxCodons*3 + 8 >= Exon.Length)
		{
		if (Exon.Length < 16)
			return false;
		MaxCodons = (Exon.Length - 8)/3;
		}
	NewDonorPos = OldDonorPos;
	for (unsigned k = 0; k < MaxCodons; ++k)
		{
		NewDonorPos = GetAdjPos(NewDonorPos, !Strand);
		if (NewDonorPos == Gene.StartCodonLastBasePos)
			return false;

		const BaseInfo &Base3 = GetBase(NewDonorPos);
		if (!Base3.IsCDS())
			return false;

		NewDonorPos = GetAdjPos(NewDonorPos, !Strand);
		if (NewDonorPos == Gene.StartCodonLastBasePos)
			return false;
		const BaseInfo &Base2 = GetBase(NewDonorPos);
		if (!Base2.IsCDS())
			return false;

		NewDonorPos = GetAdjPos(NewDonorPos, !Strand);
		if (NewDonorPos == Gene.StartCodonLastBasePos)
			return false;
		const BaseInfo &Base1 = GetBase(NewDonorPos);
		if (!Base1.IsCDS())
			return false;

		string Pair;
		Pair.push_back(Base1.GetStrandLetter(Strand));
		Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
		Log("\n");
		Log("Base1: ");
		Base1.LogMe(true);
		Log("Base2: ");
		Base2.LogMe(true);
		Log("Base3: ");
		Base3.LogMe(true);
		Log("Pair %s\n", Pair.c_str());
#endif

		if (Pair == OldDonorPair)
			{
#if	TRACE
			Log("PAIR MATCH\n");
#endif
			if (!FractCodonSuffix.empty())
				{
				string NewCodon(FractCodonSuffix);
				while (NewCodon.length() < 3)
					{
					gpos_t Pos = NewDonorPos;
					while (NewCodon.length() < 3)
						{
						Pos = GetAdjPos(Pos, !Strand);
						const BaseInfo &Base = GetBase(Pos);
						if (!Base.IsCDS())
							return false;
						NewCodon = Base.GetStrandLetter(Strand) + NewCodon;
						}
					asserta(NewCodon.length() == 3);
					if (IsStopCodon(NewCodon))
						return false;
					}
				}
#if TRACE
			Log("FOUND\n");
#endif
			return true;
			}
		}
	return false;
	}
