#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);
uint8 GetRandProbIndex(double MAC);

EVENT_RESULT Evolver::Try_MoveUTRDonorIntoIntron()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldDonorPos = GPOS_T_MAX;
	gpos_t NewDonorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveUTRDonorIntoIntron, NoDonors)

		bool Found = GetMoveUTRDonorIntoIntronCand(Gene, OldDonorPos,
		  NewDonorPos);
		if (Found)
			break;
		}

	m_CurrentEventPos = OldDonorPos;

	bool Strand = Gene.Plus;

	string OldPair;
	string NewPair;

	gpos_t Old2 = GetAdjPos(OldDonorPos, Strand);
	gpos_t New2 = GetAdjPos(NewDonorPos, Strand);

	BaseInfo &OldBase1 = GetBase(OldDonorPos);
	BaseInfo &NewBase1 = GetBase(NewDonorPos);

	BaseInfo &OldBase2 = GetBase(Old2);
	BaseInfo &NewBase2 = GetBase(New2);

	asserta(OldBase1.IsIntron());
	asserta(OldBase2.IsIntron());

	asserta(NewBase1.IsIntron());
	asserta(NewBase2.IsIntron());

	OldPair.push_back(OldBase1.GetStrandLetter(Strand));
	NewPair.push_back(NewBase1.GetStrandLetter(Strand));

	OldPair.push_back(OldBase2.GetStrandLetter(Strand));
	NewPair.push_back(NewBase2.GetStrandLetter(Strand));

#if	TRACE
	Log("ODP %"PF" NDP %"PF" Old %s New %s\n",
	  OldDonorPos, NewDonorPos, OldPair.c_str(), NewPair.c_str());
#endif

	asserta(OldPair == NewPair);

	unsigned BaseCount = 0;
	gpos_t Pos = OldDonorPos;
	for (;;)
		{
		BaseInfo &Base = GetBase(Pos);
		++BaseCount;
#if	TRACE
		Log("Convert base %u, before: ", BaseCount);
		Base.LogMe();
#endif
		asserta(Base.IsIntron());
		Base.SetUTR();
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		Pos = GetAdjPos(Pos, Strand);
		if (Pos == NewDonorPos)
			break;
		}
	asserta(BaseCount > 0);
	asserta(BaseCount <= P_MaxMoveSpliceBases);

	NewBase1.SetSplice();
	NewBase2.SetSplice();

	GeneInfo NewGene;
	GetGene(NewDonorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveUTRDonorIntoIntron: Validate gene failed");

	EV_ACCEPTN(MoveUTRDonorIntoIntron, OldDonorPos)
	}

bool Evolver::GetMoveUTRDonorIntoIntronCand(const GeneInfo &Gene,
  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveUTRDonorIntoIntronCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);
	for (unsigned j = 0; j < SIZE(ExonIndexes); ++j)
		if (GetMoveUTRDonorIntoIntronCandExon(Gene, ExonIndexes[j],
		  OldDonorPos, NewDonorPos))
			return true;

	return false;
	}

/***
Donor is pair of bases at start of intron, DD=Donor, AA=Donor:

   <---Exon---> DD Intron AA <---Exon---> DD Intron AA <---Exon--->
***/
bool Evolver::GetMoveUTRDonorIntoIntronCandExon(const GeneInfo &Gene,
  unsigned ExonIndex, gpos_t &OldDonorPos, gpos_t &NewDonorPos) const
	{
	bool Strand = Gene.Plus;
	const unsigned ExonCount = SIZE(Gene.Exons);

//There is a Donor following end of all exons, except last exon in
//transcription (X) order.
	if (Strand && ExonIndex + 1 == ExonCount)
		return false;
	if (!Strand && ExonIndex == 0)
		return false;

	const ExonInfo &Exon = Gene.Exons[ExonIndex];
	unsigned IntronIndex = (Strand ? ExonIndex : ExonIndex - 1);
	const IntronInfo &Intron = Gene.Introns[IntronIndex];
	if (Intron.Length <= P_MinIntronLength)
		return false;

#if TRACE
	Log("\n");
	Log("GetMoveUTRDonorIntoIntronCandExon\n");
	Log("Exon: ");
	Exon.LogMe(true);
	Log("Intron: ");
	Intron.LogMe(true);
#endif

	if (!Exon.AllUTR)
		return false;

	gpos_t LastExonPos =
	  Strand ? GetRangeEndPos(Exon.Pos, Exon.Length) : Exon.Pos;

	const BaseInfo &FirstExonBase = GetBase(LastExonPos);
	asserta(FirstExonBase.IsUTR());

	OldDonorPos = GetAdjPos(LastExonPos, Strand);
	gpos_t SecondOldDonorPos = GetAdjPos(OldDonorPos, Strand);

	const BaseInfo &FirstOldDonorBase = GetBase(OldDonorPos);
	const BaseInfo &SecondOldDonorBase = GetBase(SecondOldDonorPos);

	asserta(FirstOldDonorBase.IsIntron());
	asserta(SecondOldDonorBase.IsIntron());

	asserta(FirstOldDonorBase.IsSplice());
	asserta(SecondOldDonorBase.IsSplice());

	string OldDonorPair;
	OldDonorPair.push_back(FirstOldDonorBase.GetStrandLetter(Strand));
	OldDonorPair.push_back(SecondOldDonorBase.GetStrandLetter(Strand));

#if TRACE
	{
	Log("\n");
	Log("OldDonorPos %"PF"\n", OldDonorPos);
	Log("FirstExonBase: ");
	FirstExonBase.LogMe(true);
	Log("FirstOldDonorBase: ");
	FirstOldDonorBase.LogMe(true);
	Log("SecondOldDonorBase: ");
	SecondOldDonorBase.LogMe(true);
	Log("OldDonorPair %s\n", OldDonorPair.c_str());
	}
#endif
	asserta(FirstOldDonorBase.IsIntron());
	asserta(SecondOldDonorBase.IsIntron());

	gpos_t MaxDist = P_MaxMoveSpliceBases;
	if (Intron.Length < P_MinIntronLength + MaxDist)
		MaxDist = Intron.Length - P_MinIntronLength;
#if	TRACE
	Log("Intron length %"PF" MinLen %"PF" MaxDist %"PF"\n",
	  Intron.Length, P_MinIntronLength, MaxDist);
#endif
	gpos_t Pos1 = GetAdjPos(OldDonorPos, Strand);
	for (gpos_t Dist = 1; Dist < MaxDist; ++Dist)
		{
		gpos_t Pos2 = GetAdjPos(Pos1, Strand);

		const BaseInfo &Base1 = GetBase(Pos1);
		const BaseInfo &Base2 = GetBase(Pos2);
	
#if	TRACE
		Log("Pos1 %"PF" NewDonorPos %"PF"\n", Pos1, Pos2);
		Log("Base1: ");
		Base1.LogMe(true);
		Log("Base2: ");
		Base2.LogMe(true);
#endif
		asserta(Base1.IsIntron());
		asserta(Base2.IsIntron());

		string Pair;
		Pair.push_back(Base1.GetStrandLetter(Strand));
		Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
		Log("Pair %s Old %s\n", Pair.c_str(), OldDonorPair.c_str());
#endif
		if (Pair == OldDonorPair)
			{
			NewDonorPos = Pos1;
#if	TRACE
			Log("FOUND NDP %"PF" Dist %"PF"\n", NewDonorPos, Dist);
#endif
			return true;
			}

		Pos1 = Pos2;
		}
	return false;
	}
