#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

uint8 ProbToProbIndex(double Prob);

EVENT_RESULT Evolver::Try_MoveUTRTerm()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

TRC
	Log("Try_MoveUTRTerm\n");
ENDTRC
	for (;;)
		{
		GeneInfo Gene;
		m_CurrentEventPos = Gene.Pos;

		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveUTRTerm, NoCandidates)

		bool Done = TryMoveUTRTermGene(Gene);
		if (Done)
			EV_ACCEPTN(MoveUTRTerm, Gene.Pos)
		}
	}

void Evolver::GetUTRBaseCounts(const GeneInfo &Gene, gpos_t &FirstUTRBaseCount,
  gpos_t &LastUTRBaseCount) const
	{
	FirstUTRBaseCount = 0;
	LastUTRBaseCount = 0;

	const unsigned UTRCount = SIZE(Gene.UTRs);
	bool FoundCDS = false;
	const unsigned ExonCount = SIZE(Gene.Exons);
	for (unsigned i = 0; i < ExonCount; ++i)
		{
		const ExonInfo &Exon = Gene.Exons[i];
		unsigned UTRIndex = Exon.UTRIndex;
		if (UTRIndex != UINT_MAX)
			{
			asserta(UTRIndex < UTRCount);
			const UTRInfo &UTR = Gene.UTRs[UTRIndex];
			if (FoundCDS)
				LastUTRBaseCount += UTR.Length;
			else
				FirstUTRBaseCount += UTR.Length;
			}
		if (Exon.CDSIndex != UINT_MAX)
			FoundCDS = true;
		}
	}

bool Evolver::TryMoveUTRTermGene(const GeneInfo &Gene)
	{
TRC
	Log("\n");
	Log("TryMoveUTRTermGene\n");
	LogGene(Gene);
ENDTRC
	const unsigned UTRCount = SIZE(Gene.UTRs);
	if (UTRCount == 0)
		return false;

	double dDist = RandNorm(P_MoveUTRTermMeanDist, P_MoveUTRTermDistStdDev) + 0.5;
	unsigned Dist = (unsigned) dDist;

	bool Contract = RandTrue(0.5);
	bool First = RandTrue(0.5);

	gpos_t FirstUTRBaseCount = 0;
	gpos_t LastUTRBaseCount = 0;
	GetUTRBaseCounts(Gene, FirstUTRBaseCount, LastUTRBaseCount);
	gpos_t UTRBaseCount = First ? FirstUTRBaseCount : LastUTRBaseCount;
	if (Contract && UTRBaseCount < P_MinUTRBases + Dist)
		return false;

TRC
	Log("FirstUTRBaseCount  %"PF"\n", FirstUTRBaseCount);
	Log("LastUTRBaseCount   %"PF"\n", LastUTRBaseCount);
	Log("First %c Contract %c\n", tof(First), tof(Contract));
ENDTRC

	if (Contract)
		{
		if (First)
			{
			if (FirstUTRBaseCount < P_MinUTRBases + Dist)
				return false;
			}
		else
			{
			if (LastUTRBaseCount < P_MinUTRBases + Dist)
				return false;
			}
		}

	unsigned UTRIndex = First ? 0 : UTRCount - 1;
	const UTRInfo &UTR = Gene.UTRs[UTRIndex];
	unsigned ExonIndex = UTR.ExonIndex;
	asserta(ExonIndex < SIZE(Gene.Exons));
	const ExonInfo &Exon = Gene.Exons[ExonIndex];
	if (Exon.CDSIndex != UINT_MAX)
		return false;

// Set Pos to position of terminal UTR base in gene in evolved
// order. If First, is the first base, otherwise the last. Strand
// is ignored; if e.g. First && !Gene.Plus then Pos is the last
// transcribed UTR base.
	gpos_t Pos = First ? UTR.Pos : GetRangeEndPos(UTR.Pos, UTR.Length);
	BaseInfo &EndBase = GetBase(Pos);

TRC
	Log("EndPos %"PF"\n", Pos);
	Log("EndBase: ");
	EndBase.LogMe(true);
ENDTRC

	asserta(EndBase.IsUTR());

	if (Contract)
		{
		gpos_t Count = 0;
		while (Count < Dist)
			{
			BaseInfo &Base = GetBase(Pos);
TRC
			Log("Contract Pos=%"PF" ", Pos);
			Base.LogMe();
ENDTRC
		// Count UTR bases only, but convert
		// both UTR and intron bases
			if (Base.IsUTR())
				++Count;
			Base.SetUntranscribedGene();
			Base.Deconstrain();
TRC
			Log(" => ");
			Base.LogMe(true);
ENDTRC
			Pos = GetAdjPos(Pos, First);
			asserta(Pos != GPOS_T_MAX);
			}
	// Nasty special case: if last base converted
	// is exactly at end of exon, must convert
	// the following / preceding intron also.
		BaseInfo *Base = &GetBase(Pos);
		while (Base->IsIntron())
			{
			Base->SetUntranscribedGene();
			Base->Deconstrain();
TRC
			Log("Contract (intron) Pos=%"PF" ", Pos);
			Log(" => ");
			Base->LogMe(true);
ENDTRC
			Pos = GetAdjPos(Pos, First);
			asserta(Pos != GPOS_T_MAX);
			Base = &GetBase(Pos);
			}
		}
	else
		{
		assert(!Contract);
		double Prob = UTR.MeanAccProb;
		uint8 ProbIx = ProbToProbIndex(Prob);
		if (ProbIx == NEUTRAL_PROBIX)
			ProbIx = MAX_NON_NEUTRAL_PROBIX;
		for (gpos_t i = 0; i < Dist; ++i)
			{
			Pos = GetAdjPos(Pos, !First);
			BaseInfo &Base = GetBase(Pos);
TRC
			Log("Expand Pos=%"PF" ", Pos);
			Base.LogMe();
ENDTRC
			if (!Base.IsUntranscribedGene())
				break;
			Base.SetUTR();
			Base.ProbIx = ProbIx;
TRC
			Log(" => ");
			Base.LogMe(true);
ENDTRC
			}
		}

	GeneInfo Gene2;
	GetGene(Gene.Pos, Gene2);
	unsigned ErrCount = ValidateGene(Gene2);
	if (ErrCount > 0)
		Die("TryMoveUTRTermGene, ValidateGene failed");
	return true;
	}

gpos_t Evolver::GetUTRBaseCount(const GeneInfo &Gene, unsigned ExonIndex) const
	{
	gpos_t FirstUTRBaseCount = 0;
	gpos_t LastUTRBaseCount = 0;
	GetUTRBaseCounts(Gene, FirstUTRBaseCount, LastUTRBaseCount);
	bool FirstUTR = IsFirstUTR(Gene, ExonIndex);
	return FirstUTR ? LastUTRBaseCount : FirstUTRBaseCount;
	}
