#include "evolver.h"
#include "ratemgr.h"

/***
There is a complication due to the special case where the
copy point is too close to the end of the array in which
case you'll copy bases that don't belong, e.g.:

	ACGT ACGT ACGT ACGT AC^GT GATTACA...

Here, ACGT (or any circular permutation) is the repeated element,
GATTACA... is "random" stuff following the array, and ^ is the
copy point. If you don't check, you'll copy the four bases GT GA
starting at ^, which will mess up the repeated element.

The process is: (1) pick a random tandem base, set M to the motif
length, (2) back up M bases, abandon the attempt if we back up past
the beginning of the array, (3) tandem copy M bases. This has the
correct behavior: it gives a uniform random distribution over all
tandem bases except the last M-1 bases in each array, which are not
in fact Cands for being a copy point.
***/

EVENT_RESULT Evolver::Try_TandemExpand()
	{
	gpos_t Pos = GetRandomTandemPos();
	if (Pos == GPOS_T_MAX)
		EV_FAILN(TandemExpand, GetRandomTandemPos)

	gpos_t TandemMotifLength = GetTandemMotifLength(Pos);

	if (!RangeExists(Pos, TandemMotifLength))
		EV_FAILN(TandemExpand, MotifRangeExists)

// Back up (see comment above)
	gpos_t StartPos = Pos;
	m_CurrentEventPos = Pos;
	for (gpos_t i = 0; i < TandemMotifLength; ++i)
		{
		gpos_t PrevPos = GetPrevPos(StartPos);
		if (PrevPos == GPOS_T_MAX)
			EV_FAILN(TandemExpand, FindMotifStart)

		gpos_t PrevTandemMotifLength = GetTandemMotifLength(PrevPos);
		if (PrevTandemMotifLength != TandemMotifLength)
			EV_FAILN(TandemExpand, FindMotifStart)

		StartPos = PrevPos;
		}

	string Seq;
	GetSubSeq(StartPos, TandemMotifLength, Seq);
	gpos_t InsertPos = GetRangeEndPos(StartPos, TandemMotifLength);
	ACCEPT_INSERT(TandemExpand, InsertPos, Seq)

	vector<gpos_t> Genes;
	vector<gpos_t> CopiedGenes;
	Copy(StartPos, InsertPos, TandemMotifLength, Genes, CopiedGenes);
	gpos_t NewMotifPos = GetNextPos(InsertPos);

	SetTandemMotifLength(NewMotifPos, TandemMotifLength, TandemMotifLength);

	RecomputeTandemRates();
	EV_ACCEPT(TandemExpand, InsertPos, TandemMotifLength)
	}
