#include "utils.h"
#include "revfile.h"
#include "getopts.h"
#include "progress.h"

static unsigned g_ErrorCount;
static unsigned g_SegCount;

bool ValidateAI(const RevFile &Rev, unsigned AIIndex, const AlnInfo &AI)
	{
	bool OK = true;
	unsigned GenomeCount = Rev.GetGenomeCount();
	unsigned SegCount = (unsigned) AI.Segs.size();
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		++g_SegCount;
		const GCSeg &Seg = AI.Segs[SegIndex];

		unsigned GenomeIndex = Seg.GenomeIndex;
		unsigned ChrId = Seg.ChrId;

		if (GenomeIndex >= GenomeCount)
			{
			++g_ErrorCount;
			OK = false;
			LogAI(AI);
			Log("\n");
			Log("  AIIndex %u GenomeIndex %u >= GenomeCount %u\n",
			  AIIndex, GenomeIndex, GenomeCount);
			continue;
			}

		const GenInfo &GI = Rev.GetGenomeInfo(GenomeIndex);
		const ChrInfo *ptrChrInfo = GI.GetChrInfoPtr(ChrId);
		if (ptrChrInfo == 0)
			{
			++g_ErrorCount;
			OK = false;
			LogAI(AI);
			Log("\n");
			Log("  AIIndex %u Seg[%u], ChrId %u not found in genome %u\n",
			  AIIndex, SegIndex, ChrId, GenomeIndex);
			continue;
			}

		if (Seg.lo > Seg.hi)
			{
			++g_ErrorCount;
			OK = false;
			LogAI(AI);
			Log("\n");
			Log("  AIIndex %u Seg[%u] lo=%u > hi=%u\n",
			  AIIndex, SegIndex, (unsigned) Seg.lo, (unsigned) Seg.hi);
			continue;
			}

		coord_t ChrLength = ptrChrInfo->GetLength();
		if (Seg.hi >= ChrLength)
			{
			++g_ErrorCount;
			OK = false;
			const std::string &ChrName = ptrChrInfo->GetName();
			LogAI(AI);
			Log("\n");
			Log("  AIIndex %u Seg[%u] hi=%u > ChrLength=%u (%s)\n",
			  AIIndex,
			  SegIndex,
			  (unsigned) Seg.hi,
			  (unsigned) ChrLength,
			  ChrName.c_str());
			continue;
			}
		}
	return OK;
	}

void ValidateAIs()
	{
	const char *FileName = RequiredValueOpt("valais");
	bool NoMem = FlagOpt("nomem");

	RevFile Rev;
	Rev.Open(FileName);

	AIVec AIs;
	AlnInfo AI;
	if (!NoMem)
	{
		Progress("Reading AIs");
		Rev.GetAllAlnInfos(AIs);
	}

	ProgressStart("Validating");
	const unsigned AICount = Rev.GetAlnInfoCount();
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
	{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AICount);
		if (NoMem)
		{
			Rev.GetAlnInfo(AIIndex, AI);
			ValidateAI(Rev, AIIndex, AI);
		}
		else
			ValidateAI(Rev, AIIndex, AIs[AIIndex]);
	}
	ProgressDone();
	Progress("%u errors found in %u AIs, %u segs",
	  g_ErrorCount, AICount, g_SegCount);
	}
