#include <vector>
#include <map>
#include <string>
#include <iterator>
#include <algorithm>
#include "revfile.h"
#include "intlist.h"
#include "types.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"

/* Some globals */

/* The input files */
extern RevFile InFiles[2];
extern RevFile ValFile;

/* RefGenome[i] is the index of the 'reference genome' in the i-th revfile */
extern int RefGenome[2];
/* OtherGenome[i] is the index of the 'other genome' in the i-th revfile */
extern int OtherGenome[2];
/* ValGenome[i] is the index of OtherGenome[i] in ValFile */
extern int ValGenome[2];

extern unsigned MaxErrs;
extern const char *GFFFileName;

/* End of important globals */

typedef std::pair<unsigned,unsigned> ChrPosPair;
typedef std::multiset<std::pair<ChrPosPair, ChrPosPair> > AlnMap;


void FillMap(RevFile &In, unsigned FromGenome, unsigned ToGenome, AlnMap &AMap)
{
	unsigned AlnCount = In.GetAlnInfoCount();
	for (unsigned i = 0; i != AlnCount; ++i)
	{
		AlnInfo AI;
		unsigned AlnIndex = In.GetAlnInfo(i, AI);

		if (AI.GetSegCount() != 2)
			Die("Invalid number of segments (%u) for alignment #%u", AI.GetSegCount(), i);

		const GCSeg &FromSeg = AI.GetSeg(FromGenome);
		const GCSeg &ToSeg = AI.GetSeg(ToGenome);

		ChrPosPair FromCPP = std::make_pair(FromSeg.ChrId, FromSeg.Plus ? FromSeg.lo : FromSeg.hi);
		ChrPosPair ToCPP = std::make_pair(ToSeg.ChrId, ToSeg.Plus ? ToSeg.lo : ToSeg.hi);

		int From1 = FromSeg.Plus ? 1 : -1;
		int To1 = ToSeg.Plus ? 1 : -1;

		StdStrVec Rows;
		In.GetAlnRows(AlnIndex, Rows);

		unsigned FromIndex = AI.GetSegIndex(FromGenome);
		unsigned ToIndex = AI.GetSegIndex(ToGenome);

		for (unsigned j = 0; j != Rows[0].size(); ++j)
		{
			if ((Rows[FromIndex][j] != '-') && (Rows[ToIndex][j] != '-'))
				AMap.insert(std::make_pair(FromCPP, ToCPP));
			if (Rows[FromIndex][j] != '-')
				FromCPP.second += From1;
			if (Rows[ToIndex][j] != '-')
				ToCPP.second += To1;
		}
	}
}

void TransMap(AlnMap &Map1to2, AlnMap &Map1to3, AlnMap &Map2to3)
{
	AlnMap::const_iterator i = Map1to2.begin();
	AlnMap::const_iterator j = Map1to3.begin();

	while ((i != Map1to2.end()) && (j != Map1to3.end()))
	{
		if (i->first == j->first)
		{
			AlnMap::const_iterator i2 = i;
			AlnMap::const_iterator j2 = j;

			while ((i2 != Map1to2.end()) && (i2->first == i->first))
				++i2;
			while ((j2 != Map1to3.end()) && (j2->first == j->first))
				++j2;

			for (AlnMap::const_iterator i_iter = i; i_iter != i2; ++i_iter)
			{
				for (AlnMap::const_iterator j_iter = j; j_iter != j2; ++j_iter)
				{
					Map2to3.insert(std::make_pair(i_iter->second, j_iter->second));
				}
			}

			i = i2;
			j = j2;
		}
		else if (i->first < j->first)
			++i;
		else
			++j;
	}
}

class SimpleGFFRecord
{
	private:
	
	unsigned CountErrs;
	
	public:
	
	bool valid;
	bool fromFirstFile;
	ChrPosPair StartFrom;
	ChrPosPair StopFrom;
	ChrPosPair StartTo;
	ChrPosPair StopTo;

	SimpleGFFRecord() : CountErrs(0), valid(false), fromFirstFile(false) { }

	void OutputToFile(FILE *fout)
	{
		if (fout == NULL)
			return;

		if (MaxErrs > 0)
		{
			if (CountErrs == MaxErrs)
				return;
			else
				++CountErrs;
		}

		fprintf(fout, "%u\tvalta\t%s\t%u\t%u\t.\t.\t.\ttarget %u %u %u\n", StartFrom.first, fromFirstFile ? "only_in_first" : "only_in_second", StartFrom.second, StopFrom.second, StartTo.first, StartTo.second, StopTo.second);
	}
};

void CompareMaps(AlnMap &DerivedMap, AlnMap &ValMap)
{
	AlnMap::const_iterator i = DerivedMap.begin();
	AlnMap::const_iterator j = ValMap.begin();

	FILE *fout = NULL;
	if (GFFFileName != NULL)
		fout = OpenStdioFile(GFFFileName, true);

	SimpleGFFRecord Rec;

	unsigned CountI = 0, CountJ = 0;

	while ((i != DerivedMap.end()) || (j != ValMap.end()))
	{
		if ((i != DerivedMap.end()) && (j != ValMap.end()) && (*i == *j))
		{
			++i;
			++j;
		}
		else if ((j == ValMap.end()) || ((i != DerivedMap.end()) && (*i < *j)))
		{
			if (Rec.valid)
			{
				if ((Rec.fromFirstFile == true) && (Rec.StopFrom.first == i->first.first) && (Rec.StopFrom.second + 1 == i->first.second) &&
				   (Rec.StopTo.first == i->second.first) && (Rec.StopTo.second + 1 == i->second.second))
				{
					++(Rec.StopTo.second);
					++(Rec.StopFrom.second);
				}
				else
				{
					Rec.OutputToFile(fout);
					Rec.StartFrom = Rec.StopFrom = i->first;
					Rec.StartTo = Rec.StopTo = i->second;
				}
			}
			else
			{
				Rec.valid = true;
				Rec.fromFirstFile = true;
				Rec.StartFrom = Rec.StopFrom = i->first;
				Rec.StartTo = Rec.StopTo = i->second;
			}
			++CountI;
			++i;
			//printf("Only in input: (%u,%u) -> (%u, %u)\n", i->first.first, i->first.second, i->second.first, i->second.second);
		}
		else
		{
			assert ((i == DerivedMap.end()) || ((j != ValMap.end()) && (*i > *j)));
			if (Rec.valid)
			{
				if ((Rec.fromFirstFile == false) && (Rec.StopFrom.first == j->first.first) && (Rec.StopFrom.second + 1 == j->first.second) &&
				   (Rec.StopTo.first == j->second.first) && (Rec.StopTo.second + 1 == j->second.second))
				{
					++(Rec.StopTo.second);
					++(Rec.StopFrom.second);
				}
				else
				{
					Rec.OutputToFile(fout);
					Rec.StartFrom = Rec.StopFrom = j->first;
					Rec.StartTo = Rec.StopTo = j->second;
				}
			}
			else
			{
				Rec.valid = true;
				Rec.fromFirstFile = false;
				Rec.StartFrom = Rec.StopFrom = j->first;
				Rec.StartTo = Rec.StopTo = j->second;
			}
			++CountJ;
			++j;
			//printf("Only in output: (%u,%u) -> (%u, %u)\n", j->first.first, j->first.second, j->second.first, j->second.second);
		}
	}
	if (Rec.valid)
		Rec.OutputToFile(fout);
	if (fout != NULL)
		CloseStdioFile(fout);
	Progress("Finished; extra column counts: %u and %u", CountI, CountJ);
}

void ValAlign()
{
	AlnMap RefToOtherMap[2];
	AlnMap OtherToOtherMap;
	AlnMap ValMap;

	Progress("Reading 1st file");
	FillMap(InFiles[0], RefGenome[0], OtherGenome[0], RefToOtherMap[0]);
	Progress("Reading 2nd file");
	FillMap(InFiles[1], RefGenome[1], OtherGenome[1], RefToOtherMap[1]);

	Progress("Making transitive map");
	TransMap(RefToOtherMap[0], RefToOtherMap[1], OtherToOtherMap);
	RefToOtherMap[0].clear();
	RefToOtherMap[1].clear();

	Progress("Reading 3rd file");
	FillMap(ValFile, ValGenome[0], ValGenome[1], ValMap);

	Progress("Comparing maps");
	CompareMaps(OtherToOtherMap, ValMap);
}

