#include <vector>
#include <map>
#include <string>
#include <iterator>
#include <algorithm>
#include "revfile.h"
#include "intlist.h"
#include "types.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"

/* Some globals */

/* The input files */
extern RevFile InFiles[2];
extern RevFile ValFile;

/* RefGenome[i] is the index of the 'reference genome' in the i-th revfile */
extern int RefGenome[2];
/* OtherGenome[i] is the index of the 'other genome' in the i-th revfile */
extern int OtherGenome[2];
/* ValGenome[i] is the index of OtherGenome[i] in ValFile */
extern int ValGenome[2];

extern unsigned MaxErrs;
extern const char *GFFFileName;

/* End of important globals */

typedef unsigned ChrPosPair;
typedef std::vector<std::pair<ChrPosPair, ChrPosPair> > AlnMap;

void FillMap(RevFile &In, unsigned FromGenome, unsigned ToGenome, AlnMap &AMap)
{
	AMap.clear();
	AMap.reserve(1000000);
	unsigned AlnCount = In.GetAlnInfoCount();
	for (unsigned i = 0; i != AlnCount; ++i)
	{
		AlnInfo AI;
		unsigned AlnIndex = In.GetAlnInfo(i, AI);

		if (AI.GetSegCount() != 2)
			Die("Invalid number of segments (%u) for alignment #%u", AI.GetSegCount(), i);

		const GCSeg &FromSeg = AI.GetSeg(FromGenome);
		const GCSeg &ToSeg = AI.GetSeg(ToGenome);

		ChrPosPair FromCPP = FromSeg.Plus ? FromSeg.lo : FromSeg.hi;
		ChrPosPair ToCPP = ToSeg.Plus ? ToSeg.lo : ToSeg.hi;

		int From1 = FromSeg.Plus ? 1 : -1;
		int To1 = ToSeg.Plus ? 1 : -1;

		StdStrVec Rows;
		In.GetAlnRows(AlnIndex, Rows);

		unsigned FromIndex = AI.GetSegIndex(FromGenome);
		unsigned ToIndex = AI.GetSegIndex(ToGenome);

		for (unsigned j = 0; j != Rows[0].size(); ++j)
		{
			if ((Rows[FromIndex][j] != '-') && (Rows[ToIndex][j] != '-'))
				AMap.push_back(std::make_pair(FromCPP, ToCPP));
			if (Rows[FromIndex][j] != '-')
				FromCPP += From1;
			if (Rows[ToIndex][j] != '-')
				ToCPP += To1;
		}
	}
	Progress("Sorting...");
	std::sort(AMap.begin(), AMap.end());
}

bool CheckStdioEOF(FILE *f)
{
	int tmp = fgetc(f);

	if (tmp == EOF)
		return true;
	else
	{
		ungetc(tmp, f);
		return false;
	}
}

void TransMap(const char *file1, const char *file2, AlnMap &AMap)
{
	AMap.clear();
	AMap.reserve(1000000);

	FILE *f1 = OpenStdioFile(file1);
	FILE *f2 = OpenStdioFile(file2);

	std::pair<unsigned, unsigned> i;
	std::pair<unsigned, unsigned> j;

	if ((CheckStdioEOF(f1)) || (CheckStdioEOF(f2)))
		return;

	ReadStdioFile(f1, &i, sizeof(i));
	ReadStdioFile(f2, &j, sizeof(j));

	bool ValidI = true, ValidJ = true;

	while (ValidI && ValidJ)
	{
		if (i.first == j.first)
		{
			AlnMap Map1;
			AlnMap Map2;

			Map1.reserve(64);
			Map2.reserve(64);

			Map1.push_back(i);
			Map2.push_back(j);

			ValidI = false;
			while (!CheckStdioEOF(f1))
			{
				ReadStdioFile(f1, &i, sizeof(i));
				if (i.first == Map1[0].first)
					Map1.push_back(i);
				else
				{
					ValidI = true;
					break;
				}
			}

			ValidJ = false;
			while (!CheckStdioEOF(f2))
			{
				ReadStdioFile(f2, &j, sizeof(j));
				if (j.first == Map2[0].first)
					Map2.push_back(j);
				else
				{
					ValidJ = true;
					break;
				}
			}

			for (AlnMap::const_iterator i_iter = Map1.begin(); i_iter != Map1.end(); ++i_iter)
			{
				for (AlnMap::const_iterator j_iter = Map2.begin(); j_iter != Map2.end(); ++j_iter)
				{
					AMap.push_back(std::make_pair(i_iter->second, j_iter->second));
				}
			}
		}
		else if (i.first < j.first)
		{
			if (!CheckStdioEOF(f1))
			{
				ReadStdioFile(f1, &i, sizeof(i));
				ValidI = true;
			}
			else
			{
				ValidI = false;
			}
		}
		else
		{
			if (!CheckStdioEOF(f2))
			{
				ReadStdioFile(f2, &j, sizeof(j));
				ValidJ = true;
			}
			else
			{
				ValidJ = false;
			}
		}
	}

	Progress("Sorting...");
	std::sort(AMap.begin(), AMap.end());

	CloseStdioFile(f1);
	CloseStdioFile(f2);
}

class SimpleGFFRecord
{
	private:
	
	unsigned CountErrs;
	
	public:
	
	bool valid;
	bool fromFirstFile;
	ChrPosPair StartFrom;
	ChrPosPair StartTo;
	ChrPosPair StopFrom;
	ChrPosPair StopTo;

	SimpleGFFRecord() : CountErrs(0), valid(false), fromFirstFile(false), StartFrom(0), StartTo(0), StopFrom(0), StopTo(0) { }

	void OutputToFile(FILE *fout)
	{
		if (fout == NULL)
			return;

		if (MaxErrs > 0)
		{
			if (CountErrs == MaxErrs)
				return;
			else
				++CountErrs;
		}

		fprintf(fout, "0\tvalta\t%s\t%u\t%u\t.\t.\t.\ttarget 0 %u %u\n", fromFirstFile ? "only_in_first" : "only_in_second", StartFrom, StopFrom, StartTo, StopTo);
	}
};

void CompareMaps(const char *DerivedMapFile, AlnMap &ValMap)
{
	FILE *fin = OpenStdioFile(DerivedMapFile);
	std::pair<unsigned, unsigned> i;
	bool ValidI = false;

	if (!CheckStdioEOF(fin))
	{
		ValidI = true;
		ReadStdioFile(fin, &i, sizeof(i));
	}

	AlnMap::const_iterator j = ValMap.begin();

	FILE *fout = NULL;
	if (GFFFileName != NULL)
		fout = CreateStdioFile(GFFFileName);

	SimpleGFFRecord Rec;

	unsigned CountI = 0, CountJ = 0;

	while ((ValidI) || (j != ValMap.end()))
	{
		if ((ValidI) && (j != ValMap.end()) && (i == *j))
		{
			if (!CheckStdioEOF(fin))
			{
				ValidI = true;
				ReadStdioFile(fin, &i, sizeof(i));
			}
			else
				ValidI = false;
			++j;
		}
		else if ((j == ValMap.end()) || ((ValidI) && (i < *j)))
		{
			if (Rec.valid)
			{
				if ((Rec.fromFirstFile == true) && (Rec.StopFrom + 1 == i.first) && (Rec.StopTo + 1 == i.second))
				{
					++(Rec.StopTo);
					++(Rec.StopFrom);
				}
				else
				{
					Rec.OutputToFile(fout);
					Rec.fromFirstFile = true;
					Rec.StartFrom = Rec.StopFrom = i.first;
					Rec.StartTo = Rec.StopTo = i.second;
				}
			}
			else
			{
				Rec.valid = true;
				Rec.fromFirstFile = true;
				Rec.StartFrom = Rec.StopFrom = i.first;
				Rec.StartTo = Rec.StopTo = i.second;
			}
			++CountI;
			if (!CheckStdioEOF(fin))
			{
				ValidI = true;
				ReadStdioFile(fin, &i, sizeof(i));
			}
			else
				ValidI = false;
		}
		else
		{
			assert ((!ValidI) || ((j != ValMap.end()) && (i > *j)));
			if (Rec.valid)
			{
				if ((Rec.fromFirstFile == false) && (Rec.StopFrom + 1 == j->first) && (Rec.StopTo + 1 == j->second))
				{
					++(Rec.StopTo);
					++(Rec.StopFrom);
				}
				else
				{
					Rec.OutputToFile(fout);
					Rec.fromFirstFile = false;
					Rec.StartFrom = Rec.StopFrom = j->first;
					Rec.StartTo = Rec.StopTo = j->second;
				}
			}
			else
			{
				Rec.valid = true;
				Rec.fromFirstFile = false;
				Rec.StartFrom = Rec.StopFrom = j->first;
				Rec.StartTo = Rec.StopTo = j->second;
			}
			++CountJ;
			++j;
		}
	}
	if (Rec.valid)
		Rec.OutputToFile(fout);
	if (fout != NULL)
		CloseStdioFile(fout);
	CloseStdioFile(fin);
	Progress("Finished; extra column counts: %u and %u", CountI, CountJ);
}

void WriteMap(const char *MapFile, AlnMap &AMap)
{
	FILE *fout = CreateStdioFile(MapFile);

	for (AlnMap::const_iterator ci = AMap.begin(); ci != AMap.end(); ++ci)
	{
		WriteStdioFile(fout, &*ci, sizeof(std::pair<unsigned, unsigned>));
	}

	CloseStdioFile(fout);
}

void ReadMap(const char *MapFile, AlnMap &AMap)
{
	AMap.clear();

	FILE *fin = OpenStdioFile(MapFile);

	std::pair<unsigned, unsigned> UP;
	while(!CheckStdioEOF(fin))
	{
		ReadStdioFile(fin, &UP, sizeof(UP));
		AMap.push_back(UP);
	}

	CloseStdioFile(fin);
}

void ValAlignDisk()
{
	if ((InFiles[0].GetChrCount(0) != 1) || (InFiles[0].GetChrCount(1) != 1))
		Die("Input file #1 does not have a single chromosome");

	if ((InFiles[1].GetChrCount(0) != 1) || (InFiles[1].GetChrCount(1) != 1))
		Die("Input file #2 does not have a single chromosome");

	if ((ValFile.GetChrCount(0) != 1) || (ValFile.GetChrCount(1) != 1))
		Die("Input file #3 does not have a single chromosome");

	AlnMap AMap;

	std::string Tmp1, Tmp2, Tmp3;
	MakeTmpFileName("valta", Tmp1);
	MakeTmpFileName("valta", Tmp2);
	MakeTmpFileName("valta", Tmp3);

	Progress("Reading 1st file...");
	FillMap(InFiles[0], RefGenome[0], OtherGenome[0], AMap);
	Progress("Writing to disk...");
	WriteMap(Tmp1.c_str(), AMap);

	Progress("Reading 2nd file...");
	FillMap(InFiles[1], RefGenome[1], OtherGenome[1], AMap);
	Progress("Writing to disk...");
	WriteMap(Tmp2.c_str(), AMap);

	Progress("Making transitive map...");
	TransMap(Tmp1.c_str(), Tmp2.c_str(), AMap);
	Progress("Writing to disk...");
	WriteMap(Tmp3.c_str(), AMap);

	Progress("Reading 3rd file");
	FillMap(ValFile, ValGenome[0], ValGenome[1], AMap);

	Progress("Comparing maps");
	CompareMaps(Tmp3.c_str(), AMap);

	DeleteStdioFile(Tmp1.c_str());
	DeleteStdioFile(Tmp2.c_str());
	DeleteStdioFile(Tmp3.c_str());
}
