#include "wordhist.h"
#include "fastastr.h"
#include "utils.h"
#include <limits.h>
#include <math.h>

uint64 CalcWordCount(unsigned WordSize, bool Xlat);

WordHist::WordHist()
	{
	m_Counts = 0;
	m_WordCount = 0;
	m_Alpha = ALPHA_Invalid;
	m_WordSize = 0;
	}

WordHist::~WordHist()
	{
	Free();
	}

void WordHist::Free()
	{
	if (m_Counts != 0)
		myfree(m_Counts);
	m_Counts = 0;
	m_WordCount = 0;
	m_Alpha = ALPHA_Invalid;
	m_WordSize = 0;
	}

void WordHist::Init(ALPHA Alpha, unsigned WordSize)
	{
	m_Alpha = Alpha;
	m_WordSize = WordSize;
	m_WordCount = CalcWordCount(Alpha, WordSize);

	if (m_WordCount > UINT_MAX)
		Die("WordHist::Create, alpha=%s, wordsize=%u, word count %s too long, overflows unsigned int",
		  AlphaToStr(Alpha), WordSize, uint64tostr(m_WordCount));

	uint64 lBytes = m_WordCount*sizeof(COUNT_TYPE);
	off_t Bytes = (off_t) lBytes;
	if ((uint64) Bytes != lBytes)
		Die("Alpha=%s, wordsize=%u, histogram file size overflows off_t",
		  AlphaToStr(Alpha), WordSize);

	m_Counts = myalloc(COUNT_TYPE, m_WordCount);
	memset(m_Counts, 0, (size_t) (m_WordCount*sizeof(COUNT_TYPE)));
	}

void WordHist::IncWordCount(uint64 Word)
	{
	if (Word >= m_WordCount)
		Die("WordHist::IncWordCount(0x%x), word > max", (unsigned) Word);
	if (m_Counts[Word] < COUNT_TYPE_MAX)
		++(m_Counts[Word]);
	}

void WordHist::SetWordCount(uint64 Word, unsigned Count)
	{
	if (Word >= m_WordCount)
		Die("WordHist::SetWordCount(0x%x, %u), word > max", (unsigned) Word, Count);
	if (Count > COUNT_TYPE_MAX)
		Die("WordHist::SetWordCount(0x%x, %u), count > max", (unsigned) Word, Count);
	m_Counts[Word] = Count;
	}

unsigned WordHist::GetCount(uint64 Word) const
	{
	return m_Counts[Word];
	}

uint64 WordHist::GetTotalCount() const
	{
	uint64 Count = 0;
	for (uint64 Word = 0; Word < m_WordCount; ++Word)
		Count += m_Counts[Word];
	return Count;
	}

void WordHist::FromFile(const char *FileName)
	{
	FILE *f = OpenStdioFile(FileName);

	WordHistHdr Hdr;
	ReadStdioFile(f, 0, &Hdr, sizeof(Hdr));
	if (Hdr.Magic != WordHistMagic)
		Die("Invalid hist file, bad magic");

	m_Alpha = Hdr.Alpha;
	m_WordSize = Hdr.WordSize;
	m_GenomeSize = Hdr.GenomeSize;
	m_WordCount = CalcWordCount(m_Alpha, m_WordSize);

// Bytes in histogram data array
	uint64 lBytes = m_WordCount*sizeof(COUNT_TYPE);
	off_t Bytes = (off_t) lBytes;
	if ((uint64) Bytes != lBytes)
		Die("Histogram file size overflows off_t");

	uint64 ExpectedFileSize = sizeof(WordHistHdr) + Bytes;
	off_t FileSize = GetStdioFileSize(f);

	if ((uint64) FileSize != ExpectedFileSize)
		{
		char s1[64];
		char s2[64];
		Die("Alpha %s, word count %s, size of hist file should be %s, is %s",
		  AlphaToStr(m_Alpha),
		  uint64tostr(m_WordCount),
		  uint64tostr(ExpectedFileSize, s1),
		  uint64tostr(FileSize, s2));
		}

	m_Counts = myalloc(COUNT_TYPE, m_WordCount);
	ReadStdioFile(f, (off_t) sizeof(Hdr), m_Counts, Bytes);
	CloseStdioFile(f);
	}

void WordHist::ToFile(const char *FileName, uint64 GenomeSize) const
	{
	FILE *f = CreateStdioFile(FileName);
	if (m_Alpha == ALPHA_Invalid)
		Die("WordHist::ToFile, not initialized");

	WordHistHdr Hdr;
	Hdr.Magic = 0;
	Hdr.Alpha = m_Alpha;
	Hdr.WordSize = m_WordSize;
	Hdr.GenomeSize = GenomeSize;

// Write header with incorrect magic,
// thus file will be invalid unless all
// data correctly written.
	WriteStdioFile(f, 0, &Hdr, sizeof(Hdr));

// Bytes in histogram data array
	uint64 lBytes = m_WordCount*sizeof(COUNT_TYPE);
	off_t Bytes = (off_t) lBytes;
	if ((uint64) Bytes != lBytes)
		Die("Histogram file size overflows off_t");

	WriteStdioFile(f, (off_t) sizeof(Hdr), m_Counts, Bytes);

// Re-write header with correct magic
	Hdr.Magic = WordHistMagic;
	WriteStdioFile(f, 0, &Hdr, sizeof(Hdr));
	CloseStdioFile(f);
	}
