#include "revfile.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"

const unsigned MAX_LINE = 1024;

struct GenChrPair
	{
	unsigned GenId0;
	unsigned ChrId0;
	unsigned GenId1;
	unsigned ChrId1;

	bool operator<(const GenChrPair &rhs) const
		{
		if (GenId0 != rhs.GenId0)
			return GenId0 < rhs.GenId0;
		if (ChrId0 != rhs.ChrId0)
			return ChrId0 < rhs.ChrId0;
		if (GenId1 != rhs.GenId1)
			return GenId1 < rhs.GenId1;
		return ChrId1 < rhs.ChrId1;
		}
	};

static void CopyChrPair(const RevFile &In, unsigned GenId0, unsigned ChrId0,
  unsigned GenId1, unsigned ChrId1, RevFile &Out)
	{
	IntVec AIIndexes;
	std::vector<AlnInfo> AIs;

	In.GetAppendAlnInfoIndexesChr(GenId0, ChrId0, AIIndexes);
	In.GetAlnInfos(AIIndexes, AIs);

	unsigned CountWithoutAln = 0;
	unsigned CountWithAln = 0;

	const unsigned AICount = (unsigned) AIIndexes.size();
	for (unsigned i = 0; i < AICount; ++i)
		{
		unsigned AIIndex = AIIndexes[i];
		unsigned AlnIndex = In.GetAlnIndex(AIIndex);

		AlnInfo AI;
		In.GetAlnInfo(AIIndex, AI);

		if (AI.Segs.size() != 2)
			Die("Not pairwise, AI.Segs.size() = %u", (unsigned) AI.Segs.size());

		const GCSeg &Seg0 = AI.GetSeg(0);
		const GCSeg &Seg1 = AI.GetSeg(1);

		if (Seg0.GenomeIndex != GenId0 || Seg0.ChrId != ChrId0)
			Die("Query failed");

		if (Seg1.GenomeIndex != GenId1 || Seg1.ChrId != ChrId1)
			continue;

		if (AlnIndex == UINT_MAX)
			{
			Out.AddAlnInfo(AI);
			++CountWithoutAln;
			}
		else
			{
			StdStrVec Rows;
			In.GetAlnRows(AlnIndex, Rows);
			Out.AddAln(AI, Rows);
			++CountWithAln;
			}
		}
	Progress("Gen,Chr %u,%u - %u,%u: %u with alignments, %u without",
	  GenId0,
	  ChrId0,
	  GenId1,
	  ChrId1,
	  CountWithAln,
	  CountWithoutAln);
	}

void XChrs()
	{
	const char *InFileName = RequiredValueOpt("xchrs");
	const char *OutFileName = RequiredValueOpt("out");
	const char *ChrFileName = RequiredValueOpt("chrfile");

	Progress("Opening %s", InFileName);
	RevFile In;
	In.Open(InFileName);

	Progress("Opening %s", ChrFileName);
	FILE *fChrFile = OpenStdioFile(ChrFileName);

	Progress("Creating %s", OutFileName);
	RevFile Out;
	const std::vector<GenInfo> &Genomes = In.GetGenomes();
	Out.Create(OutFileName, Genomes);

	std::set<GenChrPair> GenChrPairs;
	for (;;)
		{
		char Line[MAX_LINE];
		bool EndOfFile = ReadLineStdioFile(fChrFile, Line, sizeof(Line));
		if (EndOfFile)
			break;

		GenChrPair P;
		int n = sscanf(Line, "%u %u %u %u", &P.GenId0, &P.ChrId0, &P.GenId1, &P.ChrId1);
		if (n != 4)
			Die("Didn't get 4 fields: '%s'", Line);

		if (GenChrPairs.find(P) != GenChrPairs.end())
			{
			Warning("Pair specified twice: %u %u %u %u",
			  P.GenId0, P.ChrId0, P.GenId1, P.ChrId1);
			continue;
			}

		GenChrPairs.insert(P);

		CopyChrPair(In, P.GenId0, P.ChrId0, P.GenId1, P.ChrId1, Out);
		}

	In.Close();
	CloseStdioFile(fChrFile);

	Progress("Writing output data");
	Out.Close();
	}
