#define TRACE	0

#include "utils.h"
#include "revfile.h"
#include "gff2.h"
#include "getopts.h"
#include "progress.h"
#include "ivtree.h"
#include "lohi.h"

// TODO: linear interpolation to alignment columns
// TODO: option to xfer or not when paralagous

coord_t GetOverlapLength(coord_t loA, coord_t hiA, coord_t loB, coord_t hiB);

static unsigned g_FromGenomeIndex;
static unsigned g_FromChrId;
static unsigned g_ToGenomeIndex;
static vector<AlnInfoX> g_AIXs;
static IvTree<LoHi> g_Ix;
static unsigned g_NoOverlapCount;
static unsigned g_PartialOverlapCount;
static unsigned g_CoveredCount;
static const RevFile *g_Rev;

static void BuildIx(const RevFile &Rev, const string &ChrName)
	{
	Progress("Build index for %.32s", ChrName.c_str());

	g_Ix.Clear();
	g_AIXs.clear();

	g_FromChrId = Rev.GetChrId(g_FromGenomeIndex, ChrName);

	IntVec AIIndexes;
	Rev.GetAppendAlnInfoIndexesChr(g_FromGenomeIndex, g_FromChrId, AIIndexes);

	Rev.GetAlnInfoXs(AIIndexes, g_AIXs);

	const unsigned AIXCount = SIZE(g_AIXs);
	for (unsigned AIXIndex = 0; AIXIndex < AIXCount; ++AIXIndex)
		{
		const AlnInfoX &AIX = g_AIXs[AIXIndex];
		const GCSegVec &Segs = AIX.AI.Segs;
		const unsigned SegCount = SIZE(Segs);
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = Segs[SegIndex];
			if (Seg.GenomeIndex == g_FromGenomeIndex && Seg.ChrId == g_FromChrId)
				{
				LoHi LH;
				LH.Lo = Seg.lo;
				LH.Hi = Seg.hi;
				LH.Index = AIXIndex;

				g_Ix.Add(LH);
				}
			}
		}
	}

static void XferRec(FILE *f, const GFFRecord2 &Rec)
	{
#if	TRACE
	Log("\n");
	Rec.LogMe(true);
#endif
	asserta(Rec.Start > 0);
	asserta(Rec.Start <= Rec.End);
	gpos_t FromRec_lo = Rec.Start - 1;
	gpos_t FromRec_hi = Rec.End - 1;

	vector<LoHi> Results;
	g_Ix.GetOverlaps(pair<gpos_t, gpos_t>(FromRec_lo, FromRec_hi), Results);
#if	TRACE
	{
	Log("Overlaps: ");
	for (vector<LoHi>::const_iterator p = Results.begin(); p != Results.end(); ++p)
		{
		const LoHi &LH = *p;
		Log("; ");
		LH.LogMe();
		}
	Log("\n");
	}
#endif

	if (Results.empty())
		{
		++g_NoOverlapCount;
		return;
		}

	for (vector<LoHi>::const_iterator p = Results.begin(); p != Results.end(); ++p)
		{
		const LoHi &LH = *p;
		unsigned Index = LH.Index;
		asserta(Index < SIZE(g_AIXs));
		const AlnInfoX &AIX = g_AIXs[Index];
		const GCSegVec &Segs = AIX.AI.Segs;
		const unsigned SegCount = SIZE(Segs);
		unsigned ToSegIndex = UINT_MAX;
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = Segs[SegIndex];
			if (Seg.GenomeIndex == g_ToGenomeIndex)
				{
				ToSegIndex = SegIndex;
				break;
				}
			}
		if (ToSegIndex == UINT_MAX)
			{
			Log("\n");
			AIX.AI.LogMe();
			Warning("Target genome not found");
			continue;
			}
		const GCSeg &ToSeg = Segs[ToSegIndex];
#if	TRACE
		Log("ToSeg: ");
		ToSeg.LogMe(true);
#endif

		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &FromSeg = Segs[SegIndex];
			if (FromSeg.GenomeIndex == g_FromGenomeIndex && FromSeg.ChrId == g_FromChrId &&
			  GetOverlapLength(FromRec_lo, FromRec_hi, FromSeg.lo, FromSeg.hi) > 0)
				{
#if	TRACE
				Log("FromSeg: ");
				FromSeg.LogMe(true);
#endif
				gpos_t From_lo = FromRec_lo;
				gpos_t From_hi = FromRec_hi;
				bool Partial = false;
				if (From_lo < FromSeg.lo)
					{
					Partial = true;
					From_lo = FromSeg.lo;
					}
				if (From_hi > FromSeg.hi)
					{
					Partial = true;
					From_hi = FromSeg.hi;
					}

				if (Partial)
					++g_PartialOverlapCount;
				else
					++g_CoveredCount;

				gpos_t To_lo = ToSeg.lo + (From_lo - FromSeg.lo);
				gpos_t To_hi = ToSeg.lo + (From_hi - FromSeg.lo);
				
				GFFRecord2 OutRec = Rec;
				OutRec.SeqName = g_Rev->GetChrName(g_ToGenomeIndex, ToSeg.ChrId);
				if (Rec.Strand == '+' || Rec.Strand == '-')
					{
					if (FromSeg.Plus == ToSeg.Plus)
						OutRec.Strand = Rec.Strand;
					else
						OutRec.Strand = (Rec.Strand == '+' ? '-' : '+');
					}
				else
					OutRec.Strand = Rec.Strand;

				char Str[512];
				sprintf(Str, " xferfrom \"%s %u %u\";", Rec.SeqName.c_str(), From_lo, From_hi);
				OutRec.Attrs += Str;
				  
				OutRec.Start = To_lo + 1;
				OutRec.End = To_hi + 1;
#if	TRACE
				Log("OutRec: ");
				OutRec.LogMe(true);
#endif
				WriteGFFRec(f, OutRec);
				}
			}
		}
	}

void XferAnnotsVec(const vector<GFFRecord2> &InRecs, const RevFile &AlnRev,
  unsigned FromGenomeIndex, unsigned ToGenomeIndex, FILE *f)
	{
	g_Rev = &AlnRev;
	g_FromGenomeIndex = FromGenomeIndex;
	g_ToGenomeIndex = ToGenomeIndex;

	vector<GFFRecord2> Recs(InRecs);
	sort(Recs.begin(), Recs.end());

	string CurrChrName;
	for (vector<GFFRecord2>::const_iterator p = Recs.begin(); p != Recs.end(); ++p)
		{
		const GFFRecord2 &Rec = *p;
		if (Rec.SeqName != CurrChrName)
			{
			CurrChrName = Rec.SeqName;
			BuildIx(AlnRev, CurrChrName);
			}
		XferRec(f, Rec);
		}

	Progress("xfer: %u complete annots, %u partial", g_CoveredCount, g_PartialOverlapCount);
	}

void XferAnnots()
	{
	const char *InFileName = RequiredValueOpt("xferannots");
	const char *RevFileName = RequiredValueOpt("aln");
	const char *FromGenomeName = ValueOpt("fromgenome");
	const char *ToGenomeName = ValueOpt("togenome");
	const char *OutFileName = RequiredValueOpt("out");

	RevFile AlnRev;
	AlnRev.Open(RevFileName);

	unsigned FromGenomeIndex = 0;
	unsigned ToGenomeIndex = 1;

	IntOpt("fromgenix", &FromGenomeIndex);
	IntOpt("togenix", &ToGenomeIndex);

	if (FromGenomeName != 0)
		FromGenomeIndex = AlnRev.GetGenomeIndex(FromGenomeName);
	if (ToGenomeName != 0)
		ToGenomeIndex = AlnRev.GetGenomeIndex(ToGenomeName);

	FILE *fIn = OpenStdioFile(InFileName);
	FILE *fOut = CreateStdioFile(OutFileName);

	vector<GFFRecord2> InRecs;
	GFFRecord2 Rec;
	ProgressStartFile(fIn, "Reading input");
	unsigned Count = 0;
	while (GetNextGFFRecord(fIn, Rec))
		{
		if (Count++%100 == 0)
			ProgressStepFile();
		InRecs.push_back(Rec);
		}
	ProgressDone();

	XferAnnotsVec(InRecs, AlnRev, FromGenomeIndex, ToGenomeIndex, fOut);

	CloseStdioFile(fIn);
	CloseStdioFile(fOut);
	}
