#ifndef gene_h
#define gene_h

#include <vector>

using namespace std;

struct ExonInfo
	{
	gpos_t Pos;
	gpos_t Length;
	double MeanAccProb;
	bool AllCoding;
	bool AllUTR;
	unsigned CDSIndex;
	unsigned UTRIndex;

	ExonInfo()
		{
		Clear();
		}

	void Clear()
		{
		Pos = GPOS_T_MAX;
		Length = GPOS_T_MAX;
		MeanAccProb = FLT_MAX;
		CDSIndex = UINT_MAX;
		UTRIndex = UINT_MAX;
		AllCoding = false;
		AllUTR = false;
		}

	void LogMe(bool NL = false) const
		{
		Log("Exon Pos %"PF" Len %"PF" MAC %5.3f AllCDS %c UTR %c",
		  Pos, Length, MeanAccProb, tof(AllCoding), tof(AllUTR));
		if (CDSIndex != UINT_MAX)
			Log(" CDS %u", CDSIndex);
		if (NL)
			Log("\n");
		}
	};

struct CDSInfo
	{
	gpos_t Pos;
	gpos_t Length;
	double MeanAccProb;
	unsigned ExonIndex;
	uint8 Frame;
	int8 Codex;

	CDSInfo()
		{
		Clear();
		}

	void Clear()
		{
		Pos = GPOS_T_MAX;
		Length = GPOS_T_MAX;
		MeanAccProb = FLT_MAX;
		ExonIndex = UINT_MAX;
		Frame = UCHAR_MAX;
		Codex = SCHAR_MAX;
		}

	void LogMe(bool NL = false) const
		{
		Log("CDS Pos %"PF" Length %"PF" MAC %.3g Exon %u Frame %u Codex %d",
		  Pos, Length, MeanAccProb, ExonIndex, Frame, Codex);
		if (NL)
			Log("\n");
		}
	};

struct UTRInfo
	{
	gpos_t Pos;
	gpos_t Length;
	double MeanAccProb;
	unsigned ExonIndex;

	UTRInfo()
		{
		Clear();
		}

	void Clear()
		{
		Pos = GPOS_T_MAX;
		Length = GPOS_T_MAX;
		MeanAccProb = FLT_MAX;
		ExonIndex = UINT_MAX;
		}
	void LogMe(bool NL = false) const
		{
		Log("UTR Pos %"PF" Length %"PF" MAC %.3g Exon %u",
		  Pos, Length, MeanAccProb, ExonIndex);
		if (NL)
			Log("\n");
		}
	};

struct IntronInfo
	{
	gpos_t Pos;
	gpos_t Length;

	IntronInfo()
		{
		Clear();
		}

	void Clear()
		{
		Pos = GPOS_T_MAX;
		Length = GPOS_T_MAX;
		}
	void LogMe(bool NL = false) const
		{
		Log("Intron Pos %"PF" Length %"PF, Pos, Length);
		if (NL)
			Log("\n");
		}
	};

/***
In GeneInfo, ???Info vector indexes increase
in evolved sequence order. The Get???X() functions
return in transcription order, i.e. reversed iff
on minus strand (Plus=false).
Pos is always first position of given element in
evolved order.
***/
struct GeneInfo
	{
	unsigned Index;
	gpos_t Pos;
	gpos_t Length;
	gpos_t EndPos;
	bool Plus;
	vector<ExonInfo> Exons;
	vector<IntronInfo> Introns;
	vector<CDSInfo> CDSs;
	vector<UTRInfo> UTRs;
	gpos_t StartCodonFirstBasePos;
	gpos_t StartCodonLastBasePos;
	gpos_t StopCodonFirstBasePos;
	gpos_t StopCodonLastBasePos;
	double MeanAccProb;

	GeneInfo()
		{
		Clear();
		}

	void Clear()
		{
		Index = UINT_MAX;
		Pos = GPOS_T_MAX;
		EndPos = GPOS_T_MAX;
		Length = GPOS_T_MAX;
		Plus = false;
		Exons.clear();
		Introns.clear();
		CDSs.clear();
		UTRs.clear();
		StartCodonFirstBasePos = GPOS_T_MAX;
		StartCodonLastBasePos = GPOS_T_MAX;
		StopCodonFirstBasePos = GPOS_T_MAX;
		StopCodonLastBasePos = GPOS_T_MAX;
		MeanAccProb = GPOS_T_MAX;
		}

	void CDSOffsetToPos(gpos_t Offset, unsigned &CDSIndex,
	  unsigned &CDSPos, gpos_t &Pos) const;

	const ExonInfo &GetExonX(unsigned i) const
		{
		assert(i < SIZE(Exons));
		return Plus ? Exons[i] : Exons[SIZE(Exons) - i - 1];
		}

	const IntronInfo &GetIntronX(unsigned i) const
		{
		assert(i < SIZE(Introns));
		return Plus ? Introns[i] : Introns[SIZE(Introns) - i - 1];
		}

	const CDSInfo &GetCDSX(unsigned i) const
		{
		assert(i < SIZE(CDSs));
		return Plus ? CDSs[i] : CDSs[SIZE(CDSs) - i - 1];
		}

	void LogMe() const
		{
		Log("Index %u Pos %"PF" Length %"PF" %c Start %"PF" Stop %"PF"\n",
		  Index, Pos, Length, pom(Plus), StartCodonFirstBasePos, StopCodonFirstBasePos);
		for (unsigned i = 0; i < SIZE(Exons); ++i)
			{
			const ExonInfo &Exon = Exons[i];
			Log("[%u] ", i);
			Exon.LogMe(true);
			}
		for (unsigned i = 0; i < SIZE(Introns); ++i)
			{
			const IntronInfo &Intron = Introns[i];
			Log("[%u] ", i);
			Intron.LogMe(true);
			}
		for (unsigned i = 0; i < SIZE(CDSs); ++i)
			{
			const CDSInfo &CDS = CDSs[i];
			Log("[%u] ", i);
			CDS.LogMe(true);
			}
		for (unsigned i = 0; i < SIZE(UTRs); ++i)
			{
			const UTRInfo &UTR = UTRs[i];
			Log("[%u] ", i);
			UTR.LogMe(true);
			}
		}
	};

#endif // gene_h
